; s i m f w . s c h
;
;   Copyright (C) 1995 The University of Oregon
;   Copyright (C) 1995 Lars Thomas Hansen <lth@cs.uoregon.edu>
;
;   Permission to use, copy, and/or distribute this software and its
;   documentation for any purpose and with or without fee is hereby granted,
;   provided that both the above copyright notice and this permission notice
;   appear in all copies and derived works.  
;
;   This software is provided ``as is'' without express or implied warranty.
;
; lth@cs.uoregon.edu / August 15, 1995
; $Id$
;
; Simulation framework based on threads, for the STk threads package.
; ALPHA VERSION!
;
; INTERFACE
;  (setup-simulation)
;    Initialize simulation system; call this before anything else.
;
;  (start-simulation)
;    After creating processes and so on, call this to get things going.
;
;  (end-simulation)
;    Call this to end the simulation.
;
;  (simspawn thunk)
;    Spawn a simulation task. The task suspends by calling one of the
;    sleep procedures or blocking on a 'get' from a simulation queue.
;
;  (sleep-until time)
;    Sleep until 'time' is the current time.
;
;  (sleep-for units)
;    Sleep for 'units' time units.
;
;  (integers starting-with)
;    Return a generator of integers.
;
;  (make-sim-queue)
;    Return a queue which sleeps a task if it is empty.
;
;  (random-stream dist low high)
;    Return a generator of random numbers between low and high (integers),
;    with the given distribution. Currently only 'uniform' is accepted.
;
;  (msg type format arg ... )
;    Print a message. Type can be used to control what messages are printed,
;    through the msgctl interface.
;
; BUGS
;  - This should maybe be a module?
;  - Too many hacks
;  - ^C necessary after end-simulation
;
; COMMENTS
;  The code is a little complicated in order to accomodate dynamic 
;  (that is, during the simulation) task creation and deletion. It's 
;  unclear whether that actually works, it needs testing.

(require "threads.sch")
(require "protectq.sch")
(require "synch.sch")

; Sleeping / global clock

(define clock-mutex #f)     ; protects 'clock', 'idlecount', 'taskcount'.
(define clock 0)            ; current value of clock
(define taskcount 0)        ; number of tasks in system
(define idlecount 0)        ; tasks currently idle
(define idle-condition #f)  ; clock task blocks on this
(define cqueue #f)          ; sorted queue of waiting tasks
(define waiting 0)
(define wait-sem #f)

(define (setup-simulation)
  (begin-tasking)
  (set-random-seed! 37)
  (set! cqueue (make-protected-blocking-sorted-queue
		(lambda (a b)
		  (< (cdr a) (cdr b)))))
  (set! clock-mutex (make-mutex))
  (set! idle-condition (make-event))
  (set! clock 0)
  (set! taskcount 0)
  (set! idlecount 0)
  (set! waiting 0)
  (set! wait-sem (make-semaphore 0))
  #t)

(define (start-simulation)
  (msg 'info "STARTING SIMULATION.")
  (spawn clocktask)
  (start-all))

; FIXME.
(define (end-simulation)
  (msg 'info "ENDING SIMULATION. PRESS ^C.")
  (end-tasking))

(define (wait-for-start)
  (call-with-mutex clock-mutex
    (lambda ()
      (set! waiting (+ waiting 1))))
  (semaphore-down wait-sem))

(define (start-all)
  (wait-condition
   (lambda ()
     (call-with-mutex clock-mutex
       (lambda ()
	 (= waiting taskcount)))))
  (do ((i waiting (- i 1)))
      ((zero? i) #t)
    (semaphore-up wait-sem)))

; A simulation task is like any other task, except that when it is spawned,
; it adds one to the number of tasks in the system, and when it terminates,
; it subtracts one from the number of tasks. This allows the clock task
; to kick in only when all tasks in the system are idle.

(define (simspawn proc)
  (set! taskcount (+ taskcount 1))
  (spawn (lambda ()
	   (wait-for-start)
	   (proc)
	   (call-with-mutex clock-mutex
	     (lambda ()
	       (msg 'debug "Task terminates.")
	       (set! idlecount (+ idlecount 1))
	       (release-clocktask-maybe))))))

(define (release-clocktask-maybe)
  (if (= taskcount idlecount)
      (event-signal idle-condition)))
  
(define (clocktask)
  (msg 'debug "Clock task #1")
  (event-wait idle-condition)
  (msg 'debug "Clock task #2")
  (check-termination-and-error)
  (msg 'debug "Clock task #3")
  (call-with-mutex clock-mutex
    (lambda ()
      (msg 'debug "Clock task #4")
      (let* ((s (qget* cqueue)))
	(set! clock (cdr (car s)))
	(msg 'debug "[clock] advanced to ~a" clock)
	(for-each (lambda (t) (semaphore-up (car t))) s))))
  (clocktask))

(define (schedule-wakeup sem time)
  (call-with-mutex clock-mutex
    (lambda ()
      (qput cqueue (cons sem time)))))

(define (idle-me sem)
  (call-with-mutex clock-mutex
    (lambda ()
      (set! idlecount (+ idlecount 1))
      (release-clocktask-maybe)
      (mutex-signal-and-semaphore-down clock-mutex sem)
      (set! idlecount (- idlecount 1)))))

(define (check-termination-and-error)
  ; This procedure should only ever be called when all tasks are idling.
  (if (not (= idlecount taskcount))
      (begin (msg 'info "INCONSISTENCY: IMPOSSIBLE SITUATION.")
	     (end-simulation)))

  ; If all tasks are idling but none are waiting for a time to arrive,
  ; then all are waiting on a lock.
  (if (and (qempty? cqueue) (not (zero? taskcount)))
      (begin (msg 'info "DEADLOCK / ALL WAITING.")
	     (end-simulation)))

  ; If there are no more tasks, terminate.
  (if (zero? taskcount)
      (begin (msg 'info "NO MORE TASKS.")
	     (end-simulation))))

; FIXME: need to implement qpush.

(define (qget* queue)
  (let ((first (qget queue)))
    (let loop ((l '()))
      (if (qempty? queue)
	  (cons first l)
	  (let ((next (qget queue)))
	    (if (= (cdr first) (cdr next))
		(loop (cons next l))
		(begin (qpush queue next)
		       (cons first l))))))))

(define (sleep-until n)
  (let ((s (make-semaphore 0)))
    (schedule-wakeup s n)
    (msg 'debug "Sleeping on clock.")
    (idle-me s)))

(define (sleep-for n)
  (sleep-until (+ n (call-with-mutex clock-mutex
		      (lambda ()
		        clock)))))

(define (sleep-on-semaphore sem)
  (msg 'debug "Sleeping on semaphore.")
  (idle-me sem))

(define (wakeup-from-semaphore sem)
  (semaphore-up sem))

; Stream of integers starting with n.

(define (integers n)
  (lambda ()
    (let ((k n))
      (set! n (+ n 1))
      k)))

; Stream of random numbers from some distribution.
;
; FIXME: This is not right: for reproducibility, there should be a 
; random state associated with each stream produced, but here 
; there is one, global state maintained by STk. We can fix this.

(define (random-stream type lo hi)
  (cond ((eq? type 'uniform)
	 (lambda ()
	   (+ lo (random (- hi lo)))))
	(else
	 (error "Random-stream: don't know how to deal with: " type))))

(define random
  (let ((r random))
    (atomic-lambda (n)
      (r n))))

; Another flavor of blocking queue, integrated with queueing system.

(define (make-sim-queue)
  (define q   (make-basic-queue))
  (define sem (make-semaphore 0))
  (define m   (make-mutex))

  (define (get)
    (sleep-on-semaphore sem)
    (call-with-mutex m
      (lambda ()
	(qget q))))

  (define (put item)
    (call-with-mutex m
      (lambda ()
	(qput q item)))
    (wakeup-from-semaphore sem))

  (define (empty?)
    (call-with-mutex m
      (lambda ()
	(qempty? q))))

  (define (data)
    (call-with-mutex m
      (lambda ()
	(qdata q))))

  (define (push)
    (call-with-mutex m
      (lambda ()
	(qpush q))))

  (lambda (selector)
    (case selector
      ((get pop) get)
      ((put) put)
      ((push) push)
      ((empty?) empty?)
      ((data) data)
      (($inspect) (q '$inspect))
      (else ???))))

(define *msgmap* #f)

(define (msg type fmt . args)
  (let ((m (apply format #f fmt args)))
    (if (memq type *msgmap*)
	(format #t "[time ~a] ~a~%" clock m))))

(define msgctl 
  (let ((msgs '(debug info urgent)))
    (lambda (level)
      (set! *msgmap* (or (memq level msgs) '())))))

(msgctl 'debug)


; Just spin in the scheduler until some event becomes true. Not robust,
; but works here.

(define (wait-condition thunk)
  (let loop ()
    (if (not (thunk))
	(begin (surrender-timeslice)
	       (loop)))))

; Knows too much...

(define (make-event) 
  (let ((m (make-mutex)))
    (mutex-wait m)
    m))

(define (event-wait e)
  (mutex-wait e))

(define (event-signal e)
  (call-without-interrupts
    (lambda ()
      (if (not (null? (qdata (vector-ref e 2))))
	  (mutex-signal e)))))

(provide "simfw.sch")

; eof
