; a c t o r s . s c h
;
;   Copyright (C) 1995 The University of Oregon
;   Copyright (C) 1995 Lars Thomas Hansen <lth@cs.uoregon.edu>
;
;   Permission to use, copy, and/or distribute this software and its
;   documentation for any purpose and with or without fee is hereby granted,
;   provided that both the above copyright notice and this permission notice
;   appear in all copies and derived works.  
;
;   This software is provided ``as is'' without express or implied warranty.
;
; An implementation of the actor model of computation, on top of the 
; STk tasking system.
;
; lth@cs.uoregon.edu / July 21, 1995
; $Id$
;
; DESCRIPTION
;   The macro 'actor' takes a parameter list (for initial parameters) 
;   followed by a sequence of definitions, and finally a lambda expression
;   of one argument, and evaluates to an actor template. The procedure 
;   'make-actor' takes a template and some initial arguments and creates 
;   an actor. The procedure 'send' sends a message to an actor, which the 
;   actor will receive by its lambda expression being invoked with the 
;   message as a parameter.
;
;   Procedures lexically located in the body of the actor can change the
;   actor's behavior by assigning lambda expressions of one argument to 
;   the variable *behavior*. 
;
;   Procedures lexically located in the body of the actor can obtain a handle
;   on the actor by referencing the variable *self*. This value will never
;   change for the lifetime of the actor.
;
; EXAMPLE
;   (define t 
;     (actor (acctno)
;       (define balance 0)
;
;       (lambda (msg)
;         (case (car msg)
;   	   ((withdraw) (set! balance (- balance (cadr msg))))
;	   ((deposit)  (set! balance (+ balance (cadr msg))))
;	   ((query)    (send (cadr msg) (list 'balance acctno balance)))))))
;
;   (define r
;     (actor ()
;       (lambda (msg)
;         (case (car msg)
;           ((query)   (send (cadr msg) (list 'query *self*)))
;           ((balance) (format #t "Balance on ~a is ~a~%"
;		  	      (cadr msg) (caddr msg)))))))
;
;   (define acct         (make-actor t 11001300794))
;   (define receptionist (make-actor r))
;
;   (send receptionist `(query ,acct))
;
; REFERENCES
;   Gul Agha. "Actors - A model of Concurrent Computation in Distributed
;   Systems." MIT Press, 1986.
;
; BUGS
; - This code has been hacked together; it may change as I understand actors
;   better. In particular, the interaction between 'become' and 'receive'
;   is a little dubious here (I think).
; - You must call begin-tasking before you use the actor package.

(require "threads.sch")
(require "protectq.sch")

(define-macro (actor args . e)
  (let* ((r    (reverse e))
	 (proc (car r))
	 (defs (reverse (cdr r))))
    `(lambda ,args
       ,@defs
       (define *behavior* ,proc)
       (define *q* (make-protected-blocking-queue))
       (define *loop* (lambda ()
			(let ((msg (qget *q*)))
			  (*behavior* msg)
			  (*loop*))))
       (define *self* (lambda (msg)
			(qput *q* msg)))
       (spawn *loop*)
       *self*)))
                  
(define (make-actor x . args) (apply x args))
(define (send actor msg) (actor msg))

(provide "actors.sch")

; eof
