; e n g i n e . s c h
;
;   Copyright (C) 1995 The University of Oregon
;   Copyright (C) 1995 Lars Thomas Hansen <lth@cs.uoregon.edu>
;
;   Permission to use, copy, and/or distribute this software and its
;   documentation for any purpose and with or without fee is hereby granted,
;   provided that both the above copyright notice and this permission notice
;   appear in all copies and derived works.  
;
;   This software is provided ``as is'' without express or implied warranty.
;
; lth@cs.uoregon.edu / July 26, 1995
; $Id: engine.sch,v 1.1 1995/07/26 18:29:59 lth Exp lth $
;
; An implementation of the engine abstraction, as implemented in Chez Scheme.
;
; For STk with tasking extensions.
;
; DESCRIPTION
;   make-engine takes a thunk and returns an engine. An engine is a procedure
;   of three arguments: Ticks, End, and Stop.
;    'Ticks' is an integer.
;    'End' is a procedure of two arguments, return value and remaining ticks.
;    'Stop' is a procedure of one argument, a new engine.
;
;   To run an engine, pass it the three arguments. If the thunk returns,
;   then 'end' is called with the return value of the thunk and the amount
;   of remaining fuel. If the time slice expires, then 'stop' is called with
;   an engine representing the rest of the suspended computation; to continue,
;   call the new engine.
;
; NOTES
;   In this implementation, the old engine remains valid and can be run again.
;   Short of any observable side effects, it is a snapshot of the suspended
;   computation and running it again will result in the same result as the 
;   first run.
;
;   In this implementation, an engine blocking on input will block the entire
;   process. This can be fixed by having the engine call 'stop' with more
;   than one argument, or by having an engine accept another procedure,
;   call it 'resource-suspend', which is called with a new engine and the
;   resource on which the engine is suspended, for some reasonable abstract
;   definition of resource.
;
; REFERENCES
;   R. Kent Dybvig. "The Scheme Programming Language". Prentice-Hall, 1987.

; Requires: nothing.

(define make-engine
  (let ()

    (define tick-factor 10)
    (define engines-initialized #f)
    (define out-cont #f)

    (define (do-make-engine thunk)
      (lambda (ticks end stop)
	(let ((v (call-with-current-continuation
		  (lambda (k)
		    (set! out-cont k)
		    (run-engine ticks thunk)))))
	  (if (pair? v)
	      (end (car v) (cadr v))
	      (stop (do-make-engine (lambda () (v #f))))))))

    (define (run-engine ticks thunk)
      (tasking/enable-interrupts (* ticks tick-factor))
      (let* ((r     (thunk))
	     (ticks (tasking/disable-interrupts)))
	(out-cont (list r ticks))))

    (define (initialize-engines)
      (if (not engines-initialized)
	  (begin (tasking/setup-tasking
		  (lambda ()
		    (call-with-current-continuation
		     (lambda (k)
		       (out-cont k))))
		  (lambda (event fd)
		    #f))
		 (set! engines-initialized #t))))
  
    (lambda (thunk)
      (initialize-engines)
      (do-make-engine thunk))))

(provide "engine.sch")

; eof
