; m o d u l e . s c h
;
;   Copyright (C) 1995 The University of Oregon
;   Copyright (C) 1995 Lars Thomas Hansen <lth@cs.uoregon.edu>
;
;   Permission to use, copy, and/or distribute this software and its
;   documentation for any purpose and with or without fee is hereby granted,
;   provided that both the above copyright notice and this permission notice
;   appear in all copies and derived works.  
;
;   This software is provided ``as is'' without express or implied warranty.
;
; lth@cs.uoregon.edu / July 7, 1995
; $Id: module.sch,v 1.2 1995/07/26 18:29:59 lth Exp lth $
;
; Syntax-based module systems for STk.
;
; DESCRIPTION
;   We define a macro MODULE. A module is a scope which exports some 
;   values.
;
; USAGE
;   To define a module, use the following form:
;
;     (module NAME (EXPORT ...) EXPR ...)
; 
;   where NAME is an identifier (the name of the module), each EXPORT is 
;   either an identifier or a list of two identifiers denoting internal
;   and external names, respectively (see below), and each EXPR is an
;   expression.
;
;   Each export which is only an identifier can be considered an abbreviation
;   for a two-element list where both elements are that identifier. We 
;   therefore consider an export to be a two-tuple (I E) where I is the
;   internal name and E is the external name. 'I' names a quantity defined
;   in the body of the module; E denotes the name it will be known by outside
;   the module.
;
;   A module specification expands to a toplevel BEGIN expression
;   containing only definitions; for example, the module
;
;     (module foo (bar (mybaz baz)) 
;       (define (bar) ...)
;       (define (bum) ...) 
;       (define (mybaz) ...)
;       e ...)
;
;   becomes
; 
;     (begin (define foo.bar #f)
;            (define foo.baz #f)
;            (define G1
;              (let ((G2 (let ()
;                          (define (bar) ...)
;                          (define (bum) ...)
;                          (define (mybaz) ...)
;                          e ...
;                          (vector bar mybaz))))
;                (set! foo.bar (vector-ref G2 0))
;                (set! foo.baz (vector-ref G1 0))
;                (unspecified))))
;
;   where G1 and G2 are anonymous symbols. The deal with the vector makes
;   the macro a little more hygienic, but there's still a long way to go.
;
;   The rest of the program can subsequently use foo.bar and foo.baz.
;
; BUGS (in order of importance)
; - Only values can be exported, not locations. As a consequence, there
;   can be no 'public variables' in a module -- all variable access must
;   be through accessor functions. 
; - Macros are not hygienic.
; - Modules can be used as top-level constructs only.
; - A number of common scheme procedures are effectively reserved and may
;   not be redefined.
; - Any name modsys:* is reserved.
;
; COMMENTS
; - Personally I think that the inability to export a location is a feature.
; - The top-level-only problem can be solved; see for example the solution
;   used by the monitor system (in monitor.sch).
; - Problems with capturing common names can be solved by some renaming.

(define modsys:separator ".")

(define (modsys:extname m n)
  (string->symbol
   (string-append (symbol->string m)
		  modsys:separator
		  (symbol->string (if (symbol? n) n (cadr n))))))

(define (modsys:intname n)
  (if (symbol? n) n (car n)))

(define (modsys:iota n)
  (define (loop i)
    (if (= i n)
	'()
	(cons i (loop (+ i 1)))))
  (loop 0))

(define-macro (module name exportlist . exprs)
  (let ((lname (gensym)))
    `(begin ,@(map (lambda (n)
		     `(define ,(modsys:extname name n) #f))
		   exportlist)
	    (define ,(gensym)
	      (let ((,lname (let ()
			      ,@exprs
			      (vector ,@(map modsys:intname exportlist)))))
		,@(map (lambda (n k)
			 `(set! ,(modsys:extname name n)
				(vector-ref ,lname ,k)))
		       exportlist
		       (modsys:iota (length exportlist))))))))

(provide "module.sch")

; eof
