; m o n i t o r . s c h
;
;   Copyright (C) 1995 The University of Oregon
;   Copyright (C) 1995 Lars Thomas Hansen <lth@cs.uoregon.edu>
;
;   Permission to use, copy, and/or distribute this software and its
;   documentation for any purpose and with or without fee is hereby granted,
;   provided that both the above copyright notice and this permission notice
;   appear in all copies and derived works.  
;
;   This software is provided ``as is'' without express or implied warranty.
;
; lth@cs.uoregon.edu / July 7, 1995 / July 27, 1995
; $Id: monitor.sch,v 1.2 1995/07/26 18:29:59 lth Exp lth $
;
; Syntax-based monitors based on module system.
; For STk with tasking extensions.
;
; DESCRIPTION
;   We define a macro MONITOR which is like a module (q.v.) where external
;   access to each exported value is controlled by a mutex.
;
; FEATURES
;   There are two procedures of no arguments, called MONITOR-LOCK and
;   MONITOR-UNLOCK, which are visible inside the monitor. They can be used
;   to unlock and then re-lock the monitor from inside the monitor. This
;   is sometimes useful; for example, if a monitor controls a list of
;   procedures (hooks, say) which has to be run and the decision to run
;   them is made inside the monitor, and the hooks must be run with the
;   monitor unlocked, then the most localized mechanism for accomplishing
;   this is to unlock the monitor, run the hooks, and re-lock the monitor.
;   This is not without its dangers, but it works. Use with care.
;
; BUGS and DISCUSSION
; - It may be that a monitor should take some initial parameters which
;   will be its initialization values. This is for convenience only; an
;   init export does just as well.
;
; - A number of common scheme procedures are effectively reserved and may
;   not have their global values redefined.
;
; - The use of 'apply' in the monitor is undesirable, as apply is generally
;   slow (although it may not make much difference in STk). An alternative
;   would be to parse the body of the monitor and duplicate the
;   parameter list. It's doable, but hairy: lots of special cases with
;   nested BEGINs, dotted argument lists, and so on.
;
; - Only values can be exported, not locations. As a consequence, there
;   can be no 'public variables' in a monitor -- all variable access must
;   be through accessor functions. 
;
; - Not hygienic: the reserved global names may also not be redefined in the
;   module.
;
; - The module is complicated by its ability to be nested inside modules
;   and other monitors. For an example of a top-level-only structure, see
;   the module system.

(require "synch.sch")
(require "module.sch")

(define (monsys:wrap mname p)
  (let ((args (gensym))
	(r    (gensym)))
    `(lambda ,args
       (mutex-wait ,mname)
       (let ((,r (apply ,p ,args)))
	 (mutex-signal ,mname)
	 ,r))))

(define-macro (monitor name exportlist . exprs)
  (let ((lname (gensym))
	(mname (gensym))
	(gname (gensym))
	(p     (gensym)))
    `(begin ,@(map (lambda (n k)
		     `(define ,(modsys:extname name n)
			(letrec ((,p (lambda args
				    (set! ,p (,gname ,k))
				    (apply ,p args))))
			  (lambda args
			    (apply ,p args)))))
		   exportlist
		   (modsys:iota (length exportlist)))
	    (define (MONITOR-UNLOCK) (mutex-signal ,mname))
	    (define (MONITOR-LOCK)   (mutex-wait ,mname))
	    (define ,gname
	      (letrec ((,lname (let ()
			      ,@exprs
			      (vector ,@(map (lambda (n)
					       (monsys:wrap
						mname
						(modsys:intname n)))
					     exportlist))))
		    (,mname (make-mutex)))
		(lambda (n)
		  (vector-ref ,lname n)))))))

(provide "monitor.sch")

; eof
