/*
 * Stuff to include operators in the call graph.
 * lth@cs.uoregon.edu / november/december, 1994
 */
 
/*
 * Ask: lth@cs.uoregon.edu
 *
 * KNOWN BUGS & PROBLEMS
 *
 * 1.  Consider
 *
 *       class Y { ...
 *         T operator[]( int );
 *       };
 *       class X { ...
 *         Y operator[]( int );  // note: Y is not the same as X
 *       };
 *       ...
 *       X x; ...; x[ 5 ][ 7 ];  // of type T.
 *
 *    This is not parsed correctly; it should parse as
 *
 *      (x.operator[](5)).operator[](7)
 *
 *    but does not. As a consequence, array operators are not always 
 *    handled correctly unless the caller of this module takes special
 *    care.
 *
 * 2. The comma operator is not implemented in pc++ 1.x (known bug) and
 *    is therefore not handled here yet.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#include "sage++user.h"
#include "sage++callgraph.h"
#include "operator.h"

static int sametype( SgType *, SgType * );
static int sametype( SgTypeList *, SgTypeList * );

/* The intuition for the unary attribute is that it distinguishes
 * between unary and binary forms for operators that have both,
 * but is irrelevant for operators that do not.
 */
struct xx {
  char *name;
  int op;
  int unary;  /* 0 = no, 1 = yes, 2 = don't care */
} opnames[] = { { "+", ADD_OP, 0 },
	        { "+", UNARY_ADD_OP, 1 },
		{ "-", SUBT_OP, 0 },
		{ "-", MINUS_OP, 1 },
		{ "*", MULT_OP, 0 },
		{ "*", DEREF_OP, 1 },
		{ "/", DIV_OP, 2 }, 
		{ "%", MOD_OP, 2 },
		{ "^", XOR_OP, 2 },
		{ "&", BITAND_OP, 0 },
		{ "&", ADDRESS_OP, 1 },
		{ "|", BITOR_OP, 2 },
		{ "~", BIT_COMPLEMENT_OP, 1 },
		{ "!", NOT_OP, 1 },
		{ "=", ASSGN_OP },
		{ "<", LT_OP, 2 },
		{ ">", GT_OP, 2 },
		{ "+=", PLUS_ASSGN_OP, 2 },
		{ "-=", MINUS_ASSGN_OP, 2 },
		{ "*=", MULT_ASSGN_OP, 2 },
		{ "/=", DIV_ASSGN_OP, 2 },
		{ "%=", MOD_ASSGN_OP, 2 },
		{ "^=", XOR_ASSGN_OP, 2 },
		{ "&=", AND_ASSGN_OP, 2 },
		{ "|=", IOR_ASSGN_OP, 2 },
		{ "<<", LSHIFT_OP, 2 },
		{ ">>", RSHIFT_OP, 2 },
		{ ">>=", RSHIFT_ASSGN_OP, 2 },
		{ "<<=", LSHIFT_ASSGN_OP, 2 },
		{ "==", EQ_OP, 2 },
		{ "!=", NE_OP, 2 },
		{ "<=", LTEQL_OP, 2 },
		{ ">=", GTEQL_OP, 2 },
		{ "&&", AND_OP, 2 },
		{ "||", OR_OP, 2 },
		{ "++", PLUSPLUS_OP, 2 },
		{ "--", MINUSMINUS_OP, 2 },
		{ "->*", ARROWSTAR_OP, 2 },
		{ "->", POINTST_OP, 2 },    // "points to"??? What a choice.
		{ "()", FUNCTION_OP, 2 },
		{ "[]", ARRAY_OP, 2 },
		{ "new", NEW_OP, 2 },
		{ "delete", DELETE_OP, 2 },
		{ 0, 0 }
		};

/* 
 * Symbol classification: given an operator name, returns the unnormalized operator 
 * code for the operator, or 0.
 *
 * This code is probably slow. Binary search or some form of hashing or trie would
 * be an improvement.
 */
int isOperatorName( SgSymbol *s, int unary )
{
  char *id;
  int i;

  if (s == 0) return 0;
  id = s->identifier();
  if (id == 0) return 0;
  for ( i = 0 ; 
        opnames[i].name != 0 
     && (strcmp( id, opnames[i].name ) != 0 || (opnames[i].unary != 2 && opnames[i].unary != unary))
        ; i++ )
    ;
  return opnames[i].op;
}

SgTypeList *buildtype( SgExpression *e )
{
  if (e == 0) return 0;
  switch (e->variant()) {
  case EXPR_LIST :
    if (e->lhs() == 0 || e->lhs()->type() == 0) return 0;
    if (e->rhs() == 0) return 0;
    return new SgTypeList( e->lhs()->type(), buildtype( e->rhs() ) );
  case FUNCTION_OP :
    return buildtype( e->lhs() );
  default :
    if (e->type() == 0) return 0;
    return new SgTypeList( e->type(), 0 );
  }
}

/*
 * Operator table stuff
 */
optbl::optbl()
{
  int i;

  for ( i = 0 ; i < sizeof( tbl )/sizeof( tbl[0] ); i++ )
    tbl[i] = 0;
}

optbl::~optbl()
{
  int i;
  entry *e, *f;

  for ( i = 0 ; i < sizeof( tbl )/sizeof( tbl[0] ); i++ ) {
    e = tbl[i];
    while (e != 0) {
      f = e;
      e = e->next;
      if (f->rhs != 0) delete f->rhs;
      delete f;
    }
  }
}

const char *optbl::opname( int op )
{
  int i;

  op = unnormalize( op );
  for ( i = 0 ; opnames[i].name != 0 && op != opnames[i].op ; i++ )
    ;
  if (opnames[i].name != 0) return opnames[i].name;
  return "(INVALID)";
}

int optbl::checkop( int op )
{
  if (op < OPTBL_FIRST_OP || op > OPTBL_LAST_OP) {
    warning( "Operator out of range: %d (ignored).", op );
    return 0;
  }
  return 1;
}

int optbl::unnormalize( int op )
{
  return op + OPTBL_FIRST_OP;
}

int optbl::normalize( int op )
{
  return op - OPTBL_FIRST_OP;
}

void optbl::add_operator( int op, SgType *lhs, SgTypeList *rhs, SgCallGraphFunRec *funrec )
{
  entry *e;

  if (!checkop( op )) return;
  op = normalize( op );
  
  e = new entry;
  if (e == 0) panic( "optbl: Out of memory!" );
  e->next = tbl[ op ];
  tbl[ op ] = e;
  e->lhs = lhs;
  e->rhs = rhs;
  e->funrec = funrec;
}

/* Returns 0 if not found or if error. If error, warning is printed.
 * Error can be operator out of range, or multiple match.
 */
SgCallGraphFunRec *optbl::match( int op, SgType *lhs, SgTypeList *rhs )
{
  entry *e;
  int found;
  SgCallGraphFunRec *funrec;

  if (!checkop( op )) return 0;
  op = normalize( op );

  e = tbl[ op ];
  if (e == 0) return 0;

  funrec = 0;
  found = 0;

  while (e != 0) {
    if (exact_match( lhs, rhs, e )) {
      if (!found) {
	funrec = e->funrec;
	found = 1;
      }
      else {
	warning( "type error: ambiguous operator: %s\n", opname( op ) );
	return 0;
      }
    }
    e = e->next;
  }

  if (!found) {
    e = tbl[ op ];
    while (e != 0) {
      if (approx_match( lhs, rhs, e, op )) {
	if (!found) {
	  funrec = e->funrec;
	  found = 1;
	}
	else {
	  warning( "type error: ambiguous operator." );
	  return 0;
	}
      }
      e = e->next;
    }
  }

  return funrec;
}

/* exact_match() returns 1 if the argument types exactly match the entry */
int optbl::exact_match( SgType *lhs, SgTypeList *rhs, optbl::entry *e )
{
  return ((lhs == 0 && e->lhs == 0) || sametype( lhs, e->lhs ))
      && ((rhs == 0 && e->rhs == 0) || sametype( rhs, e->rhs ));
}

/* Approx_match() returns 1 if the argument types can be cast to match the entry. 
 * For the time being, we deal only with inheritance, not with user-defined
 * conversion operators or built-in type casts.
 *
 * All operators except = are inherited.
 */
int optbl::approx_match( SgType *lhs, SgTypeList *rhs, optbl::entry *e, int op )
{
  SgClassType *t;
  SgDerivedType *dt;
  SgPointerType *pt;
  SgType *l = lhs;

  if (unnormalize( op ) != ASSGN_OP) {
again:
    if (l == 0) goto error;
    if (t = isSgClassType( l )) {
      l = find_operator_in_base( t, opname( op ) );
      if (l) return exact_match( l, rhs, e );
    }
    else if (dt = isSgDerivedType( l )) {  // special case #1
      if (dt->typeName() == 0) goto error;
      if (l = dt->typeName()->type()) goto again;
      goto error;
    }
    else if ((pt = isSgPointerType( l )) && (pt->indirection() < 0)) { // special case #2
      int ind = pt->indirection();
      l = pt->baseType();
      while ((pt = isSgPointerType( l )) && ind < 0) {
	// While indirection is negative, we we loop until it reaches 0, 
	// at which point more indirection is not munched.
	ind += pt->indirection();
	l = pt->baseType();
      }
      goto again;
    }
    else if (isSgPointerType( l ))
      return 0;
    else 
      goto error;
  }
  return 0;
error:
#if 0
  // This is way obnoxious in some of the benchmarks; useful for debugging.
  printf( "operator.C: warning: failed for operator %d, lhs=%d\n",
	  unnormalize( op ), (lhs ? lhs->variant() : -1) );
#endif
  return 0;
}

/*
 * Essentially, since we can depend on the shadowing rule, we need to 
 * find the one superclass (in some defined search order), if any,
 * which contains a definition for the operator. That definition is
 * then required to match the argument types, possibly with applied 
 * conversions.
 */
SgType *optbl::find_operator_in_base( SgClassType *t, const char *opname )
{
  SgClassStmt *s;
  int i, j, k, l;
  
  if (t == 0) return 0;
  s = isSgClassStmt( t->structureDecl() );

  for ( i = 0, k = s->numberOfSuperClasses() ; i < k ; i++ ) {
    SgClassSymb *cs = isSgClassSymb( s->superClass( i ) );
    
    for ( j = 0, l = cs->numberOfFields() ; j < l ; j++ ) {
      if (strcmp( cs->field(j)->identifier(), opname ) == 0)
	return cs->type();
    }
  }

  for ( i = 0, k = s->numberOfSuperClasses() ; i < k ; i++ ) {
    SgType *r = find_operator_in_base( isSgClassType( isSgClassSymb( s->superClass( i ) )->type() ), opname );
    if (r) return r;
  }

  return 0;
}

void optbl::warning( char *fmt, ... )
{
  va_list args;

  va_start( args, fmt );
  vfprintf( stderr, fmt, args );
  va_end( args );
  fprintf( stderr, "\n" );
}

void optbl::panic( char *fmt, ... )
{
  va_list args;

  va_start( args, fmt );
  vfprintf( stderr, fmt, args );
  va_end( args );
  fprintf( stderr, "\n" );
  exit( 1 );
}

static int sametype( SgTypeList *t1, SgTypeList *t2 )
{
  return (t1 == 0 && t2 == 0) 
      || (   t1 != 0
          && t2 != 0
          && sametype( t1->type, t2->type ) 
          && sametype( t1->next, t2->next ));
}

/* This needs to embody the five rules better than it does now. */
static int sametype( SgType *t1, SgType *t2 )
{
  SgReferenceType *rt;
  SgPointerType *pt1, *pt2;
  SgDerivedType *dt;

  if (t1 == 0 || t2 == 0) 
    return 0;
  else if (t1 == t2 )
    return 1;
  else if (rt = isSgReferenceType( t1 ))
    return sametype( rt->baseType(), t2 );
  else if (rt = isSgReferenceType( t2 ))
    return sametype( t1, rt->baseType() );
  else if ((pt1 = isSgPointerType( t1 )) && (pt2 = isSgPointerType( t2 )))
    return sametype( pt1->baseType(), pt2->baseType() );
  else if (isSgPointerType( t1 ) || isSgPointerType( t2 ))
    return 0;
  else if (isSgDescriptType( t1 ))
    return sametype( t1->baseType(), t2 );
  else if (isSgDescriptType( t2 ))
    return sametype( t1, t2->baseType() );
  else if (dt = isSgDerivedType( t1 )) {
    if (dt->typeName() == 0) printf( "PANIC!!" );
    return sametype( dt->typeName()->type(), t2 );
  }
  else if (dt = isSgDerivedType( t2 )) {
    if (dt->typeName() == 0) printf( "PANIC!!" );
    return sametype( t1, dt->typeName()->type() );
  }
  else if (t1->symbol() != 0 && t1->symbol() == t2->symbol())
    return 1;
  else 
    return 0;
}

/* eof */
