/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993                  */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/*
 * Ask: lth@cs.uoregon.edu, mainly. 
 *
 * Rumor has it this code was originaly written by Dennis Gannon and then
 * hacked by Christof Braun. Lars Hansen has since changed it considerably
 * and added tons of features (and hacks).
 *
 * As far as memory management is concerned, this code is a royal mess.
 *
 * DESCRIPTION
 *  The main interface consists of the SgCallGraph data type and the two
 *  functions GenCallTree and computeClosures. To use the module, first 
 *  declare a variable of the data type, then call GenCallTree to generate
 *  the graph. If you call computeClosure on the graph, attributes are
 *  propagated (vaguely similar to a transitive closure, whence the name).
 *
 *  See comments in GenCallTree and computeClosures for the details.
 *
 * HACK No. 1
 *  The Sage C++ typechecker is broken and does not handle the following
 *  case correctly:
 *    void f();
 *    void f() { ... }
 *    main() { f(); }
 *  The reference to f() in main() is resolved as a reference to the symbol
 *  created in line 1, not to the one in line 2, and there are two different
 *  symbols for f(). Internal declarations have the same problem.
 *
 *  In order to solve this problem temporarily (the only right thing is to
 *  fix the typechecker) the following hack is in place.
 *
 *  On the first pass, when the function tables are built, lookup is 
 *  performed _by name_ rather than _by symbol_. Every name is canonicalized
 *  (i.e., prepended the necessary number of scope operators and class
 *  names) to the form ::C1::C2 ... ::Cn::f, where the Ci are class names.
 *  Notice the initial ::. This name is then hashed, and the symbol from
 *  which the name was created is given the correct function record.
 *  This way, a declaration and the corresponding definition will share a
 *  function record, and the typechecker bug no longer matters.
 *
 *  The functionality given by this hack is accessed only by calling
 *  GenCallTree with the second argument (use_fundecl_hack) set to 1.
 *
 *  The hack does not handle obscure cases like the following:
 *    class C { int f() { int v(); v(); }; int v() { ... }; };
 *  where the forward declaration of v in C::f will be resolved as
 *  ::v (unless the typechecker got it right) rather than ::C::v.
 *
 * HACK No. 2
 *  Hack No. 1 does not get overloading right; it looks only at the name
 *  of the function and disregards its parameters. Hack No. 2 fixes this
 *  in part by distinguishing also on the number of arguments to the
 *  procedure, thereby getting many overloading cases (but certainly not
 *  all) right.
 *
 *  Hack No. 2 leverages off Hack No. 1. Disabling the former disables
 *  both.
 *
 *  A summary of the way this is implemented: each function record is on
 *  a circular list of function records which define function (symbols) which
 *  hash to the same name. (I.e., all ::C::D::f are on the same list.) Each
 *  of the records exists independently in the function record hashtable, too,
 *  hashed on the symbol pointer, like in the good old days.
 *  Each function record has two fields minargs and maxargs which determines
 *  the range of the number of arguments accepted by that function. There may
 *  be >1 function records on the same list with overlapping ranges; all but
 *  one of these has the member variable 'ignoreme' set to 1. Lookup during
 *  pass2 is always done with symbol and number of arguments. The symbol's
 *  address is hashed (just like the old version did), the function record
 *  is found, and a search on the circular list is performed for a record
 *  which has the right range of arguments and ignoreme=0.
 *
 *  The purpose of the ignoreme variable (which allows multiple overloaded
 *  functions with the same number of arguments) is to allow all functions
 *  to be in the callgraph's tables but only one function to be found
 *  for a given number of arguments during lookup.
 *
 * KNOWN BUGS
 *  1) Does not deal with overloading. This should probably not be done
 *     until the C++ parser gets overloading right anyway, at which point
 *     we should get it (and much else in this module) for free. Many
 *     cases are handled correctly due to hack #2, though.
 *
 *  2) Does not deal with operator arguments of the type (*foo); these
 *     structures are represented improperly by the compiler (IMO), and
 *     with incorrect types.
 *
 *  3) Handles only one file.
 *
 *  4) Storage management is completely out of hand.
 */

// Set to 1 to get some performance measurements.
// Tested on Solaris with CC only.
#define DEBUG 0

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include "sage++user.h"
#include "sage++callgraph.h"
#include "operator.h"

#include "compatible.h"
#if DEBUG
// On solaris, only -- other compilers may barf. Sigh.
extern "C" int gettimeofday( struct timeval * );
#endif
#ifdef SYS5
#include <string.h>
#else
#include <strings.h>
#endif

#if DEBUG
int hashcnt[11];
#endif

SgFile *globalF;    // hack (sdiwan)

struct Funinfo {
  Funinfo( FundeclId *i, int a, int b ) { id = i; min = a; max = b; }
  FundeclId *id;
  int min;
  int max;
};

struct SgCallGraphHashRec {
  SgCallGraphHashRec();
  ~SgCallGraphHashRec();
  SgSymbol *name;
  SgCallGraphFunRec *frec;
  SgCallGraphHashRec *next;
  Funinfo *info;
};

SgCallGraphHashRec::SgCallGraphHashRec()
{
  frec = 0;
  next = 0;
  info = 0;
}

SgCallGraphHashRec::~SgCallGraphHashRec()
{
  if (info) delete info;
  if (next) delete next;
}

///////////////////////////////////////////////////////////////////////////
//
// Hack to deal with function declaration / overloading problems.

#define FUNDECL_HASH 2003

// Fundecl/Fundeclhash make up an auxiliary data structure used during the
// first pass of call graph construction.
struct FundeclId {
  char *name;
  unsigned long int h;
  SgCallGraphFunRec *frec;
  FundeclId *next;

  FundeclId( const char * );
  ~FundeclId();
};

FundeclId::FundeclId( const char *n )
{
  name = new char[ strlen( n ) + 1 ];
  strcpy( name, n );
  h = 0;
  frec = 0;
  next = 0;
}

FundeclId::~FundeclId()
{
  delete [] name;
  if (next) delete next;
}

struct Fundeclhash {
  Fundeclhash();
  ~Fundeclhash();
  FundeclId *make_id( SgFunctionSymb *s, SgCallGraph * );
  SgCallGraphFunRec *lookup( FundeclId *, int, int );
  void add( FundeclId *, SgCallGraphFunRec * );

private:
  FundeclId *tbl[ FUNDECL_HASH ];
  static unsigned long int hash( const char * );
};

Fundeclhash::Fundeclhash()
{
  for ( int i = 0 ; i < FUNDECL_HASH ; i++ ) tbl[i] = 0;
}

Fundeclhash::~Fundeclhash()
{
  for ( int i = 0; i < FUNDECL_HASH ; i++ ) delete tbl[i];
}

// The identifier will look like ::C1::...::Cn::f/v where Ci is a 
// class name (i may be 0) and v is the number of arguments (a nonnegative
// decimal integer).
FundeclId *Fundeclhash::make_id( SgFunctionSymb *s, SgCallGraph *cg )
{
  char buf[ 1024 ];
  char *p, *i;
  int l;
  unsigned long h;
  FundeclId *id;
  SgStatement *d;
  SgClassStmt *cs;
  Funinfo *info;

  // Look for cached data
  if (info = cg->locateInfoInTable( s ))
    return info->id;

  // Must construct.
  p = buf+sizeof( buf )-1;
  buf[sizeof(buf)-1] = 0;
  i = s->identifier();
  l = strlen( i );
  p -= l;
  strncpy( p, i, l );
  p -= 2;
  strncpy( p, "::", 2 );

  // Should use declaredInStmt here, but it is _very_ slow, so we
  // use scope and handle the special case when the scope is returned
  // as the class stmt.
  // cs = isSgClassStmt( s->declaredInStmt()->controlParent() );

  d = s->scope();
  if (!(cs = isSgClassStmt( d )))
    cs = isSgClassStmt( d->controlParent() );

  while (cs) {
    i = cs->name()->identifier();
    l = strlen( i );
    p -= l;
    strncpy( p, i, l );
    p -= 2;
    strncpy( p, "::", 2 );
    cs = isSgClassStmt( cs->controlParent() );
  }

//*DEBUG*/printf( "Name is <%s/%d>\n", p, s->numberOfParameters() );
  h = hash( p );
  id = tbl[h];
#if DEBUG
  int cnt = 0; // count collisions
#endif
  while ( id && strcmp( id->name, p ) != 0 ) {
    id = id->next;
#if DEBUG
    cnt++;
#endif
  }
#if DEBUG
  if (cnt > 10) cnt = 10;
  hashcnt[cnt]++;
#endif
  if (!id) {
    id = new FundeclId( p );
    id->h = h;
    id->next = tbl[h];
    tbl[h] = id;
  }
  return id;
}

SgCallGraphFunRec *Fundeclhash::lookup( FundeclId *i, int min, int max )
{
  SgCallGraphFunRec *f;

  if (i->frec == 0) return 0;
  f = i->frec;
  do {
    if (f->minargs == min && f->maxargs == max && !f->ignoreme) return f;
    f = f->altnext;
  } while (f != i->frec);
  return 0;
}

void Fundeclhash::add( FundeclId *i, SgCallGraphFunRec *f )
{
  if (i->frec == 0) {
//*DEBUG*/printf( "Empty: %s %d %d\n", i->name, f->minargs, f->maxargs );
    f->altnext = f;
    f->altprev = f;
    i->frec = f;
  }
  else {
//*DEBUG*/printf( "Nonempty: %s %d %d\n", i->name, f->minargs, f->maxargs );
    f->altnext = i->frec->altnext;
    f->altprev = i->frec;
    f->altnext->altprev = f;
    f->altprev->altnext = f;
  }
}

// This is hashpjw() from the Red Dragon Book.
// Its result is always no more than 24 bits wide, and this code
// should work also on 64-bit systems.
unsigned long int Fundeclhash::hash( const char *p )
{
  unsigned long h = 0, g;
  for ( ; *p ; p++ ) {
    h = (h << 4) + *p;
    if (g = h & 0xF0000000L) {
      h = h ^ (g >> 24);
      h = h ^ g;
    }
  }
  return h % FUNDECL_HASH;
}
// end of code for function declaration/overloading hack
//
///////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////
//
// A stack ADT to deal with environments for destructors.

// symnode outside the class because it is also used stand-alone
struct symnode {
  SgSymbol *sym;
  symnode  *next;
};

class envstack {
  struct stacknode {
    SgStatement *st;
    symnode *syms;
    stacknode *next;
  };
  
  stacknode *stk;
  SgCallGraph *graph;
public:
  envstack( SgCallGraph*g )
    {
      stk = 0;
      graph = g;
    }
  
  SgStatement *top()
    { if (empty())
	fprintf( stderr,
		"Error: sgCallGraph: envstack::top(): stack empty.\n" );
      return (empty() ? 0 : stk->st);
    }
  
  void pop( SgStatement * );
  
  int  empty()  
    { return stk == 0; }
  
  void push( SgStatement *st )
    { stacknode *p = new stacknode;
      p->st = st;
      p->syms = 0;
      p->next = stk;
      stk = p;
    }
  
  void addsym( SgSymbol *sym )
    { if (stk == 0)
	fprintf( stderr,
		"Error: sgCallGraph: envstack::addsym(): stack empty.\n" );
      else {
	symnode *p = new symnode;
	p->sym = sym;
	p->next = stk->syms;
	stk->syms = p;
      }
    }
};

void envstack::pop( SgStatement *st )
{
  stacknode *r;
  symnode *p, *q;
  
  if (stk == 0) {
    fprintf( stderr,
	    "Error: sgCallGraph: envstack::pop(): stack empty.\n" );
    exit( 1 );
  }
  
  p = stk->syms;
  while (p != 0) {
    graph->updateDestructorsFor( st, 0, p->sym );
    q = p;
    p = p->next;
    delete q;
  }
  
  r = stk; 
  stk = stk->next; 
  delete r; 
}

// End stack ADT
///////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////
//
// Function Record hash table -- should perhaps be templatized?
//
// FUNDECL_HACK (Hack No. 2):
// Each entry in the hash table is a member of a circular linked list
// of function records that has the same _name_ as the function for that
// entry. However, each of the functions takes a different number
// of arguments. The fields minargs and maxargs holds the argument range.
// This allows us to distinguish overloaded functions on the number
// of arguments.
// The circular list (and subsequent searching of it) is necessary to
// work around a bug in the Sage typechecker which gets the symbols wrong.

void SgCallGraph::inithash()
{
  for( int i = 0; i < SGMAX_HASH; i++)
    hash_table[i] = 0;
}

// Compute the minimum and maximum number of arguments to the procedure
// whose symbol is passed.
//
// - For procedures with optional arguments, min is the number of
//   fixed arguments, max is the full number of arguments.
// - For varargs, min is the number of fixed arguments, max is
//   very large (10000).
//
// [To think that this function was originally 1 line long...]

void SgCallGraph::computeArguments( SgFunctionSymb *s, int *min, int *max )
{
  int np;
  int i;
//  SgExprListExp *el;
  SgExpression *el;
  SgExpression *e;
  SgVarRefExp *ve;
  SgSymbol *ss;
  SgVarDeclStmt *sv;
  SgStatement *stmt;
  Funinfo *info;
  int v;

  // Look for cached info...
  if (info = locateInfoInTable( s )) {
    *min = info->min;
    *max = info->max;
    return;
  }

  np = s->numberOfParameters();
  if (fortran) {
    *min = *max = np;
    return;
  }
  *min = 0; *max = 0;

  // The use of declaredInStmt() makes things _very_ slow ( O( n^2 ) );
  // instead, a prepass annotates each symbol with the statement it was
  // declared in, and we use that info here.
  //  if ((stmt = s->declaredInStmt()) == 0) return;  // declaration not found

  if ((stmt = (SgStatement*)(s->attributeValue( 0 ))) == 0) return;

  // We may have a list of declarations and must scan it for the right 
  // prototype. Here, sv->expr(0) is an expr_list where the nodes
  // are function_ref expressions (or not).
  if (sv = isSgVarDeclStmt( stmt )) {
    //el = isSgExprListExp( sv->expr(0) );
    el = sv->expr(0);
    e = 0;
    while (el != 0) {
      e = el->lhs();
      //el = isSgExprListExp( el->rhs() );
      el = el->rhs();
      // functions return the strangest things...
      while (e && ((v = e->variant()) == DEREF_OP || v == ADDRESS_OP))
	e = e->lhs();
      if (e && e->variant() == FUNCTION_REF && e->symbol() == s) break;
      e = 0;
    }
    if (e == 0) return; // declaration not found
  }
  else if (isSgFuncHedrStmt( stmt )) {
    e = stmt->expr(0)->lhs();
    while (e && ((v = e->variant()) == DEREF_OP || v == ADDRESS_OP))
      e = e->lhs();
    if (!e || e->variant() != FUNCTION_REF) return; // declaration not found
  }
  else {
    // Have observed behavior where the defining stmt is IF_NODE (var 105) or
    // EXPR_STMT_NODE (var 173). These occur when a function is called
    // before it is declared (i.e., the function is in fact 'declared' in
    // an expression!). The only reasonable explanation seems to be that
    // the typechecker is broken, since such constructs are illegal C++. 
    // The reason this has not been discovered is that while the depfile,
    // as it exists after parsing, is broken, the unparsed file is usually
    // fine due to the constructed #include "runtime.h" which takes care
    // of the undefined constructs as far as the real C++ compiler is 
    // concerned.
    //
    // I love it.
    //
    // A workaround would be to declare the symbol here, with an argument
    // list consisting only of `...'.
    //
    // The error message is turned off by default since it does not seem
    // to matter in practice and it is _really_ obnoxious when turned on.
#if 0
    fprintf( stderr,
	    "SgCallGraph: Warning: computeArguments#1: %s/%d: variant %d.\n",
	    s->identifier(), np, stmt->variant() );
#endif
    return;
  }

  // Get parameter list.
  // e is a function_ref expression whose lhs is an expr_list which
  // is the parameter list.
  //el = isSgExprListExp( e->lhs() );
  el = e->lhs();

  // Traverse parameter list, calculate number of arguments.
  for (i = 0 ; i < np && el != 0 ; i++ ) {
    e = el->lhs();
    //el = isSgExprListExp( el->rhs() );
    el = el->rhs();
    if (i == np - 1
	&& (ve = isSgVarRefExp( e )) 
	&& (ss = ve->symbol())
	&& strcmp( ss->identifier(), "..." ) == 0) {
      *max = 10000;
      return;
    }
    else if (e->variant() == ASSGN_OP) {
      // In a correct program, all following args are default args.
      *max = np;
      return;
    }
    else {
      (*min)++;
      (*max)++;
    }
  }
  if (i < np && el == 0 || i == np && el != 0) {
    // This is normally an error: the number of parameters differs
    // from the length of the parameter list. However, it is valid in 
    // one case: when np == 0, i == 0, and el has length 1, and the
    // parameter is 'void'.
    //
    // So, we must check this case. The representation for 'void' in
    // the function list is a VAR_REF with a type which is T_VOID.

    if (i == 0
	&& np == 0
	&& el->rhs() == 0
	&& el->lhs() != 0
	&& el->lhs()->variant() == VAR_REF
	&& el->lhs()->type() == SgTypeVoid()) {
      return; // ignore it
    }
    while (el != 0) {
      i++;
      el = el->rhs();
    }
    fprintf( stderr, 
	    "SgCallGraph: Error: computeArguments#2: %s: Sage++ says %d"
	    " formals; %d were found.\n",
	    s->identifier(), np, i );
  }
}

SgCallGraphFunRec *SgCallGraph::locateFunctionInTable(SgSymbol *s)
{
  unsigned long int hi = hashSymbol(s);
#if 1
  SgCallGraphHashRec *r;

  r = hash_table[hi];
  while (r && r->name != s) r = r->next;
  return (r ? r->frec : 0);
#else
// OLD CODE
  SgCallGraphFunRec *f;
 
  if(f = hash_table[hi]){
    while(f->s != s)
      if(!(f = f->next_hash))
	return NULL;
    return f;
  }
  return NULL;
#endif
}

Funinfo *SgCallGraph::locateInfoInTable( SgSymbol *s )
{
  unsigned long int hi = hashSymbol(s);
  SgCallGraphHashRec *r;

  r = hash_table[hi];
  while (r && r->name != s) r = r->next;
  return (r ? r->info : 0);
}

SgCallGraphFunRec *SgCallGraph::locateFunctionInTable( SgSymbol *s, int n )
{
  if (fundecl_hack && n != -1) {
    unsigned long int hi = hashSymbol(s);
    SgCallGraphHashRec *r;
    SgCallGraphFunRec *f;
    
    r = hash_table[hi];
    while (r && r->name != s) r = r->next;
    if (r) {
      f = r->frec;
      do {
	if (f->minargs <= n && n <= f->maxargs && !f->ignoreme) return f; 
	f = f->altnext;
      } while (f != r->frec);
    }
    return 0;
  }
  else
    return locateFunctionInTable( s );

}

// NOTE! This may not modify f, since f may be shared: several symbols
// may hash to f. This means that the hash table structure must be 
// decoupled from the frec structure. Hence new code.

void SgCallGraph::insertInHashTable(SgSymbol *s, SgCallGraphFunRec *f, Funinfo *info )
{
  unsigned long int hi = hashSymbol(s);

#if 1
  // Old code inserted at end of chain, should not matter.
  SgCallGraphHashRec *r = new SgCallGraphHashRec();

//printf( "adding %s %d %d\n", s->identifier(), f->minargs, f->maxargs );
  r->name = s;
  r->frec = f;
  r->info = info;
  r->next = hash_table[hi];
  hash_table[hi] = r;
#else
// OLD CODE
  SgCallGraphFunRec *f2;

  f->next_hash = NULL; 
  if(!(f2 = hash_table[hi]))
    hash_table[hi] = f;
  else{
    while(f2->next_hash){
      f2 = f2->next_hash;
    }
    f2->next_hash = f;
  }
#endif
}

unsigned long int SgCallGraph::hashSymbol(SgSymbol *s)
{
  return (((unsigned long int)(void *)s) % SGMAX_HASH);
}

// End function record hash table
//
///////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////
//
// Old, unmaintained, and presumably unused code

#if CG_INCLUDE_OBSOLETE_CODE
SgSymbol *SgCallGraph::function(int i)
{ // i-th function in table (0 = first)
  SgCallGraphFunRec *f=FunListHead;
  static SgCallGraphFunRec *lastf=FunListHead;
  static int first_time=1;
  if(first_time){
    first_time=0;
    fprintf(stderr, "SgCallGraph:: Using obsolete function:\n"
	    "--    SgSymbol *function(int)\n");
  }
  if(i < num_funs_in_table){
    if(i >= lastf->id)i-=(f=lastf)->id;
    for(;i>0;i--)f = f->next;
    return f->s;
  }
  else return NULL;
}

SgStatement *SgCallGraph::functionBody(int i)
{ // i-th function in table (0 = first)
  SgCallGraphFunRec *f=FunListHead;
  static SgCallGraphFunRec *lastf=FunListHead;
  static first_time=1;
  if(first_time){
    first_time=0;
    fprintf(stderr, "SgCallGraph:: Using obsolete function:\n"
	    "--    SgStatement *functionBody(int)\n");
  }
  if(i < num_funs_in_table){
    if(i >= lastf->id)i-=(f=lastf)->id;
    for(;i>0;i--)f = f->next;
    return f->body;
  }
  else return NULL;
}

void SgCallGraph::printTableEntry(int i)
{
  SgSymbol *s = function(i);
  static first_time=1;
  if(!first_time){
    first_time=0;
    fprintf(stderr, "SgCallGraph:: Using obsolete function:\n"
	    "--    void printTableEntry(int)\n");
  }
  printFunctionEntry(s);
}

SgStatement *SgCallGraph::callSiteStmt(SgSymbol *fname, int i)
{ // stmt of i-th call of funname
  SgCallSiteList *cs;
  SgCallGraphFunRec *f;
  
  static int first_time=1;
  if(first_time){
    first_time=0;
    fprintf(stderr, "SgCallGraph:: Using obsolete function:\n"
	    "--    SgStatement *callSiteStmt(SgSymbol *, int)\n");
  }
  if((f = locateFunctionInTable(fname)) && (i < f->Num_Call_Sites)){
    cs = f->callSites->next;
    for(;i>0;i--)cs = cs->next;
    return cs->stmt;
  }
  else return NULL;
#if 0
  if((f = locateFunctionInTable(fname)) && (i < f->Num_Call_Sites))
    return f->callsite[i];
  else return NULL;
#endif
}
	
SgExpression *SgCallGraph::callSiteExpr(SgSymbol *fname, int i)
{ // expression of i-th call
  SgCallSiteList *cs;
  SgCallGraphFunRec *f;

  static int first_time=1;
  if(first_time){
    first_time=0;
    fprintf(stderr, "SgCallGraph:: Using obsolete function:\n"
	    "--    SgExpression *callSiteExpr(SgSymbol *, int)\n");
  }
  if((f = locateFunctionInTable(fname)) && (i < f->Num_Call_Sites)){
    cs = f->callSites->next;
    for(;i>0;i--)cs = cs->next;
    return cs->expr;
  }
  else return NULL;

#if 0
  if((f = locateFunctionInTable(fname)) && (i < f->Num_Call_Sites))
    return f->callexpr[i];
  else return NULL;
#endif
}
	
SgSymbol *SgCallGraph::calledFunction(SgSymbol *funname, int i)
{ // i-th function called in funname
  SgCallGraphFunRecList *cf;
  SgCallGraphFunRec *f;
  
  static int first_time=1;
  if(first_time){
    first_time=0;
    fprintf(stderr, "SgCallGraph:: Using obsolete function:\n"
	    "--    SgSymbol *calledFunction(SgSymbol *, int)\n");
  }
  if((f = locateFunctionInTable(funname)) && (i < f->Num_Call_List)){
    cf = f->callList->next;
    for(;i>0;i--)cf = cf->next;
    return cf->fr->s;
  }
  else return NULL;

#if 0
  if((f = locateFunctionInTable(funname)) && (i < f->Num_Call_List)) 
    return f->callList[i]->s;
  else return NULL;
#endif
}
#endif // #if CG_INCLUDE_OBSOLETE_CODE

// End Old Code
///////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////
//
// This is the source code for sage++callgraph 

int SgCallGraph::numberOfFunctionsInGraph(){ return num_funs_in_table; }

SgSymbol *SgCallGraph::firstFunction()
{
  return (currentFun = FunListHead)?currentFun->s:NULL;
}

SgSymbol *SgCallGraph::nextFunction()
{
  if(currentFun)
    return (currentFun = currentFun->next)?currentFun->s:NULL;
  else
    return NULL;
}

SgStatement *SgCallGraph::functionBody(SgSymbol *fname)
{
  SgCallGraphFunRec *f;  
  return (f=locateFunctionInTable(fname))?f->body:NULL;
}

SgSymbol *SgCallGraph::resolveSymbol( SgSymbol *s )
{
  SgFunctionSymb *fs;

  if (fs = isSgFunctionSymb( s )) {
    int min, max;
    SgCallGraphFunRec *f;

    computeArguments( fs, &min, &max );
    return (f = locateFunctionInTable( fs, min )) ? f->s : s;
  }
  else
    return s;
}

int SgCallGraph::functionId(SgSymbol *fname)
{
  SgCallGraphFunRec *f;  
  return (f=locateFunctionInTable(fname))?f->id:-1;
}

int SgCallGraph::numberOfCallSites(SgSymbol *funname)
{ // number of call sites for funname
  SgCallGraphFunRec *f;
  if(f = locateFunctionInTable(funname)) return f->Num_Call_Sites;
  else return 0;
}

SgStatement *SgCallGraph::firstCallSiteStmt(SgSymbol *fname)
{ // stmt of first call of fname
  SgCallGraphFunRec *f;
  if((f = locateFunctionInTable(fname)) && (f->Num_Call_Sites>0))
    return (f->currentCallSite = f->callSites->next)->stmt;
  else return NULL;
}

SgStatement *SgCallGraph::nextCallSiteStmt(SgSymbol *fname)
{ // stmt of next call of fname
  SgCallGraphFunRec *f;
  if((f = locateFunctionInTable(fname)) 
     && f->currentCallSite
     && (f->currentCallSite!=f->callSites))
    return (f->currentCallSite = f->currentCallSite->next)->stmt;
  else return NULL;
}

							     
SgExpression *SgCallGraph::firstCallSiteExpr(SgSymbol *fname)
{ // expression of first call
  SgCallGraphFunRec *f;
  if((f = locateFunctionInTable(fname)) && (f->Num_Call_Sites>0))
    return (f->currentCallExpr = f->callSites->next)->expr;
  else return NULL;
}

SgExpression *SgCallGraph::nextCallSiteExpr(SgSymbol *fname)
{ // expression of next call
  SgCallGraphFunRec *f;
  if((f = locateFunctionInTable(fname)) 
     && f->currentCallExpr
     && (f->currentCallExpr!=f->callSites))
    return (f->currentCallExpr = f->currentCallExpr->next)->expr;
  else return NULL;
}

SgSymbol *SgCallGraph::firstCalledFunction(SgSymbol *fname)
{ // first function called in fname
  SgCallGraphFunRec *f;
  if((f = locateFunctionInTable(fname)) && (f->Num_Call_List>0))
    return (f->currentFunCall = f->callList->next)->fr->s;
  else return NULL;
}

SgSymbol *SgCallGraph::nextCalledFunction(SgSymbol *fname)
{ // next function called in fname
  SgCallGraphFunRec *f;
  if((f = locateFunctionInTable(fname)) 
     && f->currentFunCall
     && (f->currentFunCall!=f->callList))
    return (f->currentFunCall = f->currentFunCall->next)->fr->s;
  else return NULL;
}

SgStatement *SgCallGraph::SgCalledFunctionStmt(SgSymbol *fname)
{ //get statement of current called function
  SgCallGraphFunRec *f;
  if((f = locateFunctionInTable(fname)) && f->currentFunCall)
    return (f->currentFunCall)->stmt;
  else return NULL;
}

SgExpression *SgCallGraph::SgCalledFunctionExpr(SgSymbol *fname)
{ //get expression of current called function
  SgCallGraphFunRec *f;
  if((f = locateFunctionInTable(fname)) && f->currentFunCall)
    return (f->currentFunCall)->expr;
  else return NULL;
}

int SgCallGraph::numberOfFunsCalledFrom(SgSymbol *funname)
{
  // how many call sites in funname
  SgCallGraphFunRec *f;
  if(f = locateFunctionInTable(funname)) return f->Num_Call_List;
  else return 0;
}
	
int SgCallGraph::isACollectionFunc(SgSymbol* funname)
{ // 1 if funname is a method of a collection.
  SgCallGraphFunRec *f;
  return (f = locateFunctionInTable(funname))
      && (f->type != SGMOE_FUN)
      && f->isCollection;
}

int SgCallGraph::isAMethodOfElement(SgSymbol* funname)
{ // 1 if funname is a method of a method of an elemnt.
  SgCallGraphFunRec *f;
  return (f = locateFunctionInTable(funname)) && (f->type == SGMOE_FUN);
}
	
int SgCallGraph::isCalledInSeq(SgSymbol* funname)
{ // 1 if funname is called in a sequential sect.
  SgCallGraphFunRec *f;
  if(f = locateFunctionInTable(funname)) return f->calledInSeq;
  else return 0;
}
	
int SgCallGraph::isCalledInPar(SgSymbol* funname)
{ // 1 if funname is called in parallel code
  SgCallGraphFunRec *f;
  if(f = locateFunctionInTable(funname)) return f->calledInPar;
  else return 0;
}
	
int SgCallGraph::minLevel(SgSymbol* funname)
{ // minimal level function is called
  SgCallGraphFunRec *f;
  if(f = locateFunctionInTable(funname)) return f->minlevel;
  else return -1;
}
	
int SgCallGraph::maxLevel(SgSymbol* funname)
{ // maximal level function is called
  SgCallGraphFunRec *f;
  if(f = locateFunctionInTable(funname)) return f->maxlevel;
  else return -1;
}
	
void SgCallGraph::init( SgFile *f )
{
  num_funs_in_table = 0;
  currentFun = FunListHead = FunListTail = NULL;
  main = 0;
  mainFunRec = 0;
  inithash();
  fortran = f->languageType() == ForSrc;
  cplusplus = f->languageType() == CSrc;
}
			
void SgCallGraph::printFunctionEntry(SgSymbol *s)
{
   SgCallSiteList *cs;
   SgCallGraphFunRecList *cf;
   SgCallGraphFunRec *t = locateFunctionInTable(s);
   
   if(!t){
       printf("function %s not found in table\n", s->identifier());
       return;
   }

   printf("---------------------------------------------------------\n");
   printf("[%d] name: %s\n",t->id,t->s->identifier());
   switch(t->type){
	case SGMOE_FUN: 
		printf("Method Of ELement.");
		if(t->className) 
		  printf(" Collection: %s\n", t->className->identifier());
		else printf("\n");
		break;
	case SGMOC_FUN:
                if(t->isCollection) printf("Method of Collection\n");
		else printf("Class member function\n");
		if(t->className) 
		  printf(" Class: %s\n", t->className->identifier());
		else printf("\n");
		break;
	default:
		printf("ordinary function\n");
	}
   if(t->calledInPar) printf("invoked in parallel!\n");
   else if(t->calledInSeq) printf("part of main thread \n");
   if(t->calledInPar && t->calledInSeq) printf("WARNING: both types!!\n");

   printf("call sites:");
   if(t->Num_Call_Sites == 0) printf("none\n");
   for(cs = t->callSites; cs; cs=cs->next) cs->stmt->unparsestdout();
   printf("which calls:");
   if(t->Num_Call_List == 0) printf(" nothing\n");
   else printf("\n");
   for(cf = t->callList; cf; cf=cf->next) 
                    printf("          %s\n", cf->fr->s->identifier());
   printf("\n");
}


//
// true if s is  same class as c or a super class of c.
//
int isSameClassOrSuperClass(SgSymbol *c, SgSymbol *s){
   SgStatement *body;
   SgClassStmt *t;
   int i;
   if(!c || !s) return 0;
   if(c == s) return 1;
   if((body = c->body()) && (t = isSgClassStmt(body))){
	for(i = 0; i < t->numberOfSuperClasses(); i++){
		if(isSameClassOrSuperClass(t->superClass(i), s)){
			 return 1;
			}
		}
	}
    return 0;
}
	
SgCallGraphFunRec * SgCallGraph::lookForFunctionOpForClass(SgSymbol *c) {
    if(!c) return NULL;;
    for(SgCallGraphFunRec *f=FunListHead; f; f=f->next){
	if (isSameClassOrSuperClass(c,f->className) &&
            (!strcmp(f-> s->identifier(),"()")))
          return f;
	}
    return NULL;
}
//
// the following function takes a function call reference (statment, expr) to
// a function t and updates the table with the reference info, i.e. who called
// whom.
//

void SgCallGraph::updateFunctionTableConnections(SgCallGraphFunRec *t,
						 SgStatement *stat,
						 SgExpression *e,
						 int atfront )
{
  SgStatement *st;
  SgCallGraphFunRec *t2 = NULL;   
  SgFunctionSymb *name;

  if(t){
    SgCallSiteList *cs = new SgCallSiteList;
    cs->stmt = stat;
    cs->expr = e;
    if(t->Num_Call_Sites++){
      cs->next = t->callSites->next;
      t->callSites->next = cs;
      t->callSites = cs;
    }
    else
      t->callSites = cs->next = cs;
    st = stat;
    // This is what it used to be, but that is too aggressive.
    //st = stat->controlParent();
    while(st 
	  && (st->variant() != FUNC_HEDR)
	  && (st->variant() != GLOBAL)
	  && (st->variant() != PROC_HEDR)
	  && (st->variant() != PROS_HEDR)
	  && (st->variant() != PROG_HEDR))
      st = st->controlParent();

    // Have a procedure header; find it.
    if (st && (st->variant() == FUNC_HEDR
	      || st->variant() == PROC_HEDR
	      || st->variant() == PROG_HEDR
	      || st->variant() == PROS_HEDR))
      name = isSgFunctionSymb( st->symbol() );
    else if (st && st->variant() == GLOBAL)
      name = isSgFunctionSymb( main );
    else {
      fprintf( stderr, 
	       "SgCallGraph: Warning: bogus control parent structure.\n" );
    }

    if (fundecl_hack) {
      if (!name)
	t2 = 0;
      else {
	int min, max;
	FundeclId *id = fundecltbl->make_id( name, this );
	computeArguments( name, &min, &max );
	t2 = fundecltbl->lookup( id, min, max );
      }
    }
    else
      t2 = locateFunctionInTable( name );

    if(!t2){
      fprintf( stderr, "SgCallGraph: Warning: could not locate call: \n");
      stat->unparsestdout();
    }
    else{
      SgCallGraphFunRecList *fr = new SgCallGraphFunRecList;
      fr->fr = t;
      fr->stmt = stat;
      fr->expr = e;
      if(t2->Num_Call_List++){
	fr->next = t2->callList->next;
	t2->callList->next = fr;
	if (!atfront) t2->callList = fr;  /* circular list... */
      }
      else
	t2->callList = fr->next = fr;
    }
  }
  else { /* 
	   printf("unknown function in expression:");
	   e->unparsestdout(); printf("\n"); */
  }
}

//
// the following function builds a new, simplified type for a given
// type generated by the compiler.  Its main function is to cancel
// out the occurrences of adjacent ref and deref operators.
// note: this must be used with care.
// it strips off attributes like const, global, sync, etc.
//

SgType *makeReducedType(SgType *t)
{
  int done, sign;
  SgPointerType *p;
  SgDerivedType *d;
  SgDescriptType *des;
  SgSymbol *s;

  if(!t) return NULL;
  done = 0;
  sign = 0;
  while(!done){
    if(des = isSgDescriptType(t)){
      //printf("descript found \n");
      if(des->hasBaseType() && des->baseType()){
	if((des->modifierFlag() & BIT_LONG) ||
	   (des->modifierFlag() & BIT_UNSIGNED) ||
	   (des->modifierFlag() & BIT_SIGNED) ||
	   (des->modifierFlag() & BIT_SHORT))
	  done = 1;
	else t = des->baseType();
      }
      else done = 1;
    }
    else if(d = isSgDerivedType(t)){
      // printf("derived found \n");
      if((s = d->typeName()) && s->type()) t = s->type();
      else done = 1;
    }
    else if(isSgArrayType(t)){
      //printf("array found \n");
      if(t->baseType()){
	t = t->baseType();
      }
      else done = 1;                        
    }
    else if(p = isSgPointerType(t)){
      //printf("ptr found \n");
      if(p->baseType()){
	sign += p->indirection();
	t = p->baseType();
      }
      else done = 1;
    }
    else done = 1;
  }
  // now add back balance of pointer refs.
  //  printf("sign = %d\n", sign);
  while(sign){
    p = new SgPointerType(*t);
    if(sign < 0){
      p->setIndirection(-1);
      sign++;
    }
    else{
      p->setIndirection(1);
      sign--;
    }
    t = p;
  }
  return t;
}

// To find collection references we need to look at types
// to do this we need to reduce the type expression.
// there may be several layers of dereferences of pointer types
// and typedefs and type modifiers that hide the true type.  
// this functions tries to reduce these to the true type
// of the object.
//

SgType *findTrueType(SgExpression *e)
{
  SgType *t;
  SgPointerType *p, *p2;
  SgDerivedType *d;
  SgDescriptType *des;
  SgSymbol *s;
  int done = 0;
  
  if(!e) return NULL;
  if(!(t = e->type())) return NULL;
  while(!done) {
    if(des = isSgDescriptType(t)){
      if(des->hasBaseType() && des->baseType()) t = des->baseType();
      else done = 1;
    }
    else if(d = isSgDerivedType(t)){
      if((s = d->typeName()) && s->type()) t = s->type();
      else done = 1;
    }
    /*  need to fix baseType for array
	else if(ar = isSgArrayType(t)){
	if((p2 = isSgPointerType(ar->baseType())) &&
	(1 + p2->indirection() == 0)) t = p2->baseType();
	else done = 1;                        
	}
	*/
    else if(p = isSgPointerType(t)){
      if((p2 = isSgPointerType(p->baseType())) &&
	 (p->indirection() + p2->indirection() == 0)) t = p2->baseType();
      else if((d = isSgDerivedType(p->baseType())) && 
	      (s = d->typeName()) &&
	      (p2 = isSgPointerType(s->type())) &&
	      (p->indirection() + p2->indirection() == 0) )
	t = p2->baseType();
      else done = 1;
    }
    else done = 1;
  }
  return t;
}

SgSymbol *lookForClassOrCollectionSymbol( SgType *type,
					 int stop_at_typedefs = 0 )
{
  SgDescriptType *des;
  SgDerivedType *dt;
  SgArrayType *at;
  SgClassSymb *cs;
  SgDerivedCollectionType *ct;
  
  if (type == 0) 
    return 0;
  else if (dt = isSgDerivedType( type )) {
    if (cs = isSgClassSymb( dt->typeName() ))
      return cs;
    else
      return lookForClassOrCollectionSymbol( dt->typeName()->type(),
					    stop_at_typedefs );
  }
  else if (ct = isSgDerivedCollectionType( type )) {
    return ct->collectionName();
  }
  else if (at = isSgArrayType( type ))
    return lookForClassOrCollectionSymbol( at->baseType(), stop_at_typedefs );
  else if (!stop_at_typedefs 
	   && (des = isSgDescriptType( type )) 
	   && (des->modifierFlag() & BIT_TYPEDEF)) {
    if(des->hasBaseType() && des->baseType()) 
      return lookForClassOrCollectionSymbol( des->baseType(), 
					    stop_at_typedefs );
    else
      return 0;
  }
  else
    return 0;
}

// Find collection or class symbol

static SgClassSymb *lookForClassSymbol( SgType *type, int stop_at_typedefs=0 )
{
  return
    isSgClassSymb( lookForClassOrCollectionSymbol( type, stop_at_typedefs ) );
}

// This procedure finds all the constructors which will be _immediately_
// called when an object of class cs is created, and associates them in the
// call graph with the given statement and expression. The expression may
// be 0. "Immediately called" means 
// - the constructor if there is one; transitivity deals with the rest
// - the constructors of members of class type, if there are any, and so on,
//   recursively.

void SgCallGraph::updateConstructorsFor( SgStatement *stat, 
					SgExpression *e, 
					SgSymbol *sym,
					int nargs,
					int skip,
					int atfront )
{
  int i, j, k;
  char *classid;
  SgClassStmt *clst;
  SgClassType *t;
  SgClassSymb *cs;
  
  cs = isSgClassSymb( sym );
  classid = cs->identifier();
  j = cs->numberOfFields();
  
  // direct constructor 
  if (!skip) {
    for ( i = 0 ; i < j ; i++ ) {
      if (strcmp( classid, cs->field(i)->identifier() ) == 0) {
	SgCallGraphFunRec *t = locateFunctionInTable( cs->field(i), nargs );
	updateFunctionTableConnections( t, stat, e, atfront );
	return;
      }
    }
  }
  
  // constructors for bases
  // FIXME: should take into account ordered, parameterized list in
  // class header.
  if ((t = isSgClassType( cs->type() ))
      && (clst = isSgClassStmt( t->structureDecl() ))) {
    for ( i = 0, k = clst->numberOfSuperClasses() ; i < k ; i++ ) {
      SgClassSymb *csx = isSgClassSymb( clst->superClass( i ) );
      if (csx)
	updateConstructorsFor( stat, e, csx, -1, 0, atfront );
    }
  }
  
  // constructors for members of class type
  for ( i = 0 ; i < j ; i++ ) {
    SgClassSymb *cs2;
    
    if (isSgFieldSymb(cs->field(i))
	&& (cs2 = lookForClassSymbol( cs->field(i)->type() )))
      updateConstructorsFor( stat, e, cs2, -1, 0, atfront );
  }
}

void SgCallGraph::updateDestructorsFor( SgStatement *stat, SgExpression *e,
				       SgSymbol *sym, int skip )
{
  int i, j, k;
  char *classid, *id;
  SgClassStmt *clst;
  SgClassType *t;
  SgClassSymb *cs;
  
  cs = isSgClassSymb( sym );
  j = cs->numberOfFields();
  classid = cs->identifier();
  
  // direct destructors
  if (!skip) {
    for ( i = 0; i < j ; i++ ) {
      id = cs->field(i)->identifier();
      if (*id == '~' && strcmp( classid, id+1 ) == 0) {
	SgCallGraphFunRec *t = locateFunctionInTable( cs->field(i) ); // FIXME
	updateFunctionTableConnections( t, stat, e );
	return;
      }
    }
  }
  
  // destructors for bases
  if ((t = isSgClassType( cs->type() ))
      && (clst = isSgClassStmt( t->structureDecl() ))) {
    for ( i = 0, k = clst->numberOfSuperClasses() ; i < k ; i++ ) {
      SgClassSymb *csx = isSgClassSymb( clst->superClass( i ) );
      if (csx)
	updateDestructorsFor( stat, e, csx, 0 );
    }
  }
  
  // destructors for members of class type
  for ( i = 0 ; i < j ; i++ ) {
    SgClassSymb *cs2;
    
    if (isSgFieldSymb(cs->field(i))
	&& (cs2 = lookForClassSymbol( cs->field(i)->type() ))) {
      updateDestructorsFor( stat, e, cs2, 0 );
    }
  }
}

// This is a utility function that searches an expression in a statement
// for calls to functions.  when they are found they are recored
// in the table.
void SgCallGraph::findFunctionCalls( SgStatement *stat,
				     SgExpression *e,
				     optbl& operators )
{
  SgFunctionCallExp *f;
  SgFuncPntrExp *fop;
  SgNewExp *newe;
  int done = 0;
  
  if(!e) return;
  if(f = isSgFunctionCallExp(e)){
    SgSymbol *s = f->funName();
    SgCallGraphFunRec *t = locateFunctionInTable(s, f->numberOfArgs() );
    updateFunctionTableConnections(t, stat, e);
  }
  else if(fop = isSgFuncPntrExp(e)){
    /* This should not need to be here, but it cannot be removed yet */
    /*  of the form (*funptr)(args) or Coll(args)->...
	look for a collection type.
	*/
    SgExpression *funptr = fop->funExp();
    SgExpression *ref;
    SgType *type;
    SgDerivedCollectionType *colref;
    while(funptr->variant() == EXPR_LIST) funptr = funptr->lhs();
    type = findTrueType(funptr);
    /*  while(funptr->variant() == DEREF_OP) funptr = funptr->lhs(); */
    ref = funptr;
    if(ref && type)
      if(colref = isSgDerivedCollectionType(type)){
	/* have a collection and a () ref. */
	/*
	   printf("collection refernce: "); ref->unparsestdout();
	   printf(" with  type:%s <",  colref->collectionName()->identifier());
	   if(elemtype = isSgDerivedType(colref->elementClass())){
	   SgSymbol *elemsym = elemtype->typeName();
	   printf("%s >\n", elemsym->identifier());
	   }
	   else printf(" ?? >\n");
	   */
	SgCallGraphFunRec *t =
	  lookForFunctionOpForClass(colref->collectionName());
	updateFunctionTableConnections(t, stat, fop);
      }
  }
  else if (newe = isSgNewExp( e )) {
    // Basic variations:
    //   new Class-name
    //   new Class-name( a1, ... )
    //   new Class-name[ E, ... ]...
    // Also note that depending on the class definition, constructors 
    // for members and bases may be called. In addition, operator new 
    // or ::operator new may be called (if defined). And, constructors 
    // may be overloaded.
    //
    // There are also interesting ones like new C::D (for a class D defined
    // inside class C) with above variations.
    SgCastExp *ce;
    if (ce = isSgCastExp( newe->lhs()) ) {
      SgType *type;

      if (type = ce->type()) {
	SgClassSymb *cs;
	int nargs = 0;
	SgExpression *e;
	// figure the number of actual arguments.
	// weirdness: arglist may be pointer to empty expr-list node,
	// and this means 0 args.
	e = ce->lhs(); // get arglist
	if (e && e->lhs()) // handle weirdness
	  while (e) { e = e->rhs(); nargs++; }

	if (cs = lookForClassSymbol( type ))
	  updateConstructorsFor( stat, newe, cs, nargs );
      }
    }
    else
      fprintf( stderr, "sgCallGraph: Warning: weird NEW expression.\n" );
  }
  else if (isSgDeleteExp( e )) {
    SgClassSymb *cs;
    SgPointerType *pt;
    
    // Must do subexpressions first to deal with weird shit like 'delete new X'
    
    if (e->lhs()) findFunctionCalls( stat, e->lhs(), operators );
    if (e->rhs()) findFunctionCalls( stat, e->rhs(), operators );
    done = 1;
    
    // lhs() is the expression to be deleted. It is a pointer expression.
    // If it points to a class object or array of class objects, we win.
    
    if (pt = isSgPointerType( e->lhs()->type() )) {
      if (cs = lookForClassSymbol( pt->baseType() )) 
	updateDestructorsFor( stat, e, cs );
    }
    else
      fprintf( stderr, "sgCallGraph: Warning: Weird DELETE expression.\n" );
  }
  
  if (!done) {
    if(e->lhs())findFunctionCalls(stat, e->lhs(), operators );
    if(e->rhs())findFunctionCalls(stat, e->rhs(), operators );
  }
  
  // important to do this last so the subexprs have been taken care of.
  { SgCallGraphFunRec *funrec;
    SgTypeList *rhs;
    SgType *lhs;
    
    if (e->variant() == ARRAY_REF) {
      // This is a hack to deal with a bug in the v1.x parser; the parser
      // does not convert all array references with overloaded subscripting 
      // operator to variant ARRAY_OP. 
      // The bug is further described (as #1) in the header of operator.C.
      rhs = buildtype( e->lhs() );
      if (funrec = operators.match( ARRAY_OP, e->symbol()->type(), rhs ))
	updateFunctionTableConnections(funrec, stat, e);
      if (rhs != 0) delete rhs;
    }
    else if (e->variant() == POINTST_OP) {
      // -> has unusual semantics: it is treated as a _unary_ operator.
      if (funrec = operators.match( POINTST_OP, e->lhs()->type(), 0 ))
	updateFunctionTableConnections( funrec, stat, e );
    }
    else if (e->variant() == PLUSPLUS_OP || e->variant() == MINUSMINUS_OP) {
      lhs = (e->lhs() ? e->lhs()->type() : 0);
      rhs = (e->rhs() ? new SgTypeList( e->rhs()->type(), 0 ) : 0);
      if (funrec = operators.match( e->variant(), lhs, rhs ))
	updateFunctionTableConnections( funrec, stat, e );
      if (rhs != 0) delete rhs;
    }
    else {
      lhs = (e->lhs() ? e->lhs()->type() : 0);
      rhs = buildtype( e->rhs() );
      if (funrec = operators.match( e->variant(), lhs, rhs ))
	updateFunctionTableConnections(funrec, stat, e);
      if (rhs != 0) delete rhs;
    }
  }
}


// Return the main symbol's address. May be 0.
SgSymbol *SgCallGraph::mainSymbol()
{
  return main;
}

SgCallGraphFunRec::SgCallGraphFunRec()
{
  id = 0;
  Num_Call_Sites = 0;
  Num_Call_List = 0;
  callList = NULL;
  callSites = NULL;
  currentCallSite = NULL;
  currentCallExpr = NULL;
  currentFunCall = NULL;
  isCollection = 0;
  calledInPar = 0;
  calledInSeq = 0;
  className = 0;
  minlevel = -1;
  maxlevel = -1;
  minargs = 0;
  maxargs = 0;
  ignoreme = 0;
  ident = 0;
  body = 0;
  next = 0;
  s = 0;
}

void SgCallGraph::addFunctionRecord( SgCallGraphFunRec *frec )
{
  frec->id = num_funs_in_table++;
  if (!FunListHead)
    FunListHead = frec;
  else
    FunListTail->next = frec;
  FunListTail = frec;
}

// Add a function symbol to the table of known functions.
//
// But only do it if the symbol refers to a function definition, not
// to a declaration! This is a workaround for a Sage symbol table/type check
// problem, where declarations have symbols which are not the same as
// the symbols in the definition.

SgCallGraphFunRec *SgCallGraph::addToTable( SgSymbol *sym,
					   int type,
					   int isCollection,
					   SgSymbol *className /* ,
					   Fundeclhash *fundecltbl */ )
{
  SgFunctionSymb *s;
  SgCallGraphFunRec *frec = 0;
  FundeclId *id;
  int min, max;
  int mustignore = 0;

  s = isSgFunctionSymb( sym );
  if (s == 0) {
    fprintf( stderr, "sgCallGraph: bogus symbol (%d) in addToTable\n",
	    sym->variant() );
    exit( 1 );
  }

  if (fundecl_hack) {
    id = fundecltbl->make_id( s, this );
    computeArguments( s, &min, &max );
    frec = fundecltbl->lookup( id, min, max );
  }
  else 
    if (frec = locateFunctionInTable( s )) return frec;

  if (!frec) {
    // Create record
    if(!(frec = new SgCallGraphFunRec)) {
      Message("Out of memory in sgCallGraph\n",0);
      exit(1);
    }
    
    frec->s = s;
    frec->type = type;
    frec->isCollection = isCollection;
    frec->className = className;

    if (fundecl_hack) {
      frec->minargs = min;
      frec->maxargs = max;
      frec->ignoreme = mustignore;
      frec->ident = id;
    }

    addFunctionRecord( frec );
    
    if (fundecl_hack)
      fundecltbl->add( id, frec );
  }

  if (!locateFunctionInTable( s )) {
    if (fundecl_hack)
      insertInHashTable( s, frec, new Funinfo( id, min, max ) );
    else
      insertInHashTable( s, frec );
  }
  
  return frec;
}

void SgCallGraph::cloneFunctionRecord( SgCallGraphFunRec *frec,
				      SgFuncHedrStmt *fhst )
{
  SgCallGraphFunRec *frec2 = new SgCallGraphFunRec;

  frec2->s = frec->s;
  frec2->ident = frec->ident;
  frec2->maxargs = frec->maxargs;
  frec2->minargs = frec->minargs;
  frec2->ignoreme = 1;
  frec2->s = isSgFunctionSymb( fhst->symbol() );
  frec2->body = frec2->s->body();
  addFunctionRecord( frec2 );
  if (fundecl_hack)
    fundecltbl->add( frec2->ident, frec2 );
}

// computeClosures traverses the graph and looks at who calls whom.

void SgCallGraph::computeClosures()
{
  int changed, j, level;
  SgCallGraphFunRec *theMain = NULL;
  SgCallGraphFunRec *f;
  
//*DEBUG*/printf( "computing the transitive closure\n" );
  for(f=FunListHead; f; f=f->next)
    f->flag = 0;
  
  // Find the functions that are called as part of the main 
  // (sequential) thread
  theMain = mainFunRec;
  // if there is no main procedure (as may be the case with library
  // and/or header files), then we create a "dummy" main procedure
  // and continue on ... (sdiwan,jvillaci)
  if (!theMain) {
	 SgType *foo = globalF->firstType();  // check for variant=DEFAULT
	 while (foo) {
		if (foo->variant() == DEFAULT) {
		    break;
		}
		foo=foo->next();
	 }
	 theMain = new SgCallGraphFunRec;
	 theMain->s = 
		new SgFunctionSymb(fortran ? PROGRAM_NAME : FUNCTION_NAME,
					    "_implicit_dummy_main",
					    *foo,
//					    *(globalF->SgTypeWithId(8)),
					    *(globalF->firstStatement()));
	 theMain->ident = new FundeclId("");
	 theMain->maxargs = 0;
	 theMain->minargs = 0;
	 theMain->ignoreme = 0;
	 theMain->body = 
		new SgStatement(fortran ? PROG_HEDR : FUNC_HEDR, 0, theMain->s, 0, 0, 0);
	 theMain->body->setControlParent(globalF->firstStatement());
	 addFunctionRecord( theMain );
	 if (fundecl_hack)
		fundecltbl->add( theMain->ident, theMain );
	 if (fortran) {   // add extra node for the implicit END statement in Fortran
		theMain->next = new SgCallGraphFunRec;
		theMain->next->ident = new FundeclId("");
		theMain->next->maxargs = 0;
		theMain->next->minargs = 0;
		theMain->next->ignoreme = 0;
		theMain->next->body = 
		    new SgStatement(CONTROL_END, 0, 0, 0, 0, 0);
		theMain->next->body->setControlParent(theMain->body);
	 }
	 main = theMain->s;
	 mainFunRec = theMain;
  }
  if (theMain){
//*DEBUG*/printf( "Loop nest 1\n" );
    changed = 1;
    theMain->calledInSeq = 1;
    while(changed) {
      changed = 0;
      for(f=FunListHead; f; f = f->next){
	if(!f->flag && f->calledInSeq){
	  SgCallGraphFunRecList *fl=f->callList;
	  for(j = 0; j < f->Num_Call_List; j++,fl=fl->next)
	    if((fl->fr->type != SGMOE_FUN) && 
	       (fl->fr->isCollection == 0))
	      fl->fr->calledInSeq = 1;
	  changed = 1;
	  f->flag = 1;
	}
      }
    }
  } else {
	 Message("no main procedure found in file, exiting ...", 0);
	 exit(1);
  }
  //
  // find the functions that are called in the parallel regions.
  //
  for(f=FunListHead; f; f=f->next){
    f->flag = 0;
    if((f->type == SGMOE_FUN) || f->isCollection) f->calledInPar = 1;
  }
  changed = 1;
//*DEBUG*/printf( "Loop nest 2\n" );
  while(changed){
    changed = 0;
    for(f=FunListHead; f; f=f->next){
      if(!f->flag && f->calledInPar){
	SgCallGraphFunRecList *fl=f->callList;
	for(j = 0; j < f->Num_Call_List; j++,fl=fl->next)
	  fl->fr->calledInPar = 1;
	changed = 1;
	f->flag = 1;
      }
    }
  }
  //
  // compute levels
  //
  for(f=FunListHead; f; f=f->next) {
    f->flag = 0;
  }
  theMain->minlevel = 0;
  theMain->maxlevel = 0;
  theMain->flag = 1;
  changed = 1;
  level   = 0;
//*DEBUG*/printf( "Loop nest 3\n" );
  enum { MUST_VISIT = 1, HAS_VISITED = 2 };
  while(changed) {
    level++;
    changed = 0;
    for(f=FunListHead; f; f=f->next) {
      if(f->flag & MUST_VISIT) {
//*DEBUG*/printf( "Visiting %s\n", f->s->identifier() );	
	f->flag &= ~MUST_VISIT;
	SgCallGraphFunRecList *fl=f->callList;
	for(j = 0; j < f->Num_Call_List; j++,fl=fl->next) {
	  if(fl->fr->minlevel==-1) fl->fr->minlevel = level;
	  if (!(fl->fr->flag & HAS_VISITED)) {
//*DEBUG*/printf( "Marking %s\n", fl->fr->s->identifier() );
	    fl->fr->maxlevel = level;
	    fl->fr->flag = MUST_VISIT;
	    changed = 1;
	  }
	}
	f->flag |= HAS_VISITED;
      }
    }
  }
//*DEBUG*/printf( "computeClosures done.\n" );
}

/*
 * loopOverStatements()
 *
 * This is pass 2 of the call graph. A single pass is made over all
 * statements, and for each statement, all function calls in the expression
 * are inspected and recorded as calls from the function in which the call
 * exists to the function which the call names. In addition, function
 * definitions are remembered, and constructor and destructor calls are
 * inferred and remembered.
 */
void SgCallGraph::loopOverStatements( SgStatement *first,
				     int C,
				     int FORTRAN,
				     optbl& operators )
{
  SgStatement *st;
  envstack stk( this );
  symnode *glob = 0; int globcnt = 0;
  SgStatement *prevstmt = 0;
  
  for( st = first ; st; st = st->lexNext()) {
    SgFuncHedrStmt *fnst;
    SgVarDeclStmt *vdst;
    
    // For each function definition, remember its body. This means either:
    // - setting the body field of an existing declaration
    // - cloning an existing declaration/definition, and setting the
    //   body of the clone. The clone has ignoreme==1.
    // The purpose of remembering is to supply the body to callers of
    // the functionBody() member function. A very conventient side effect
    // (_not_ by accident) is to have an entry in the function table for
    // each defined function.

    if (fnst = isSgFuncHedrStmt( st )) {
      SgCallGraphFunRec *frec;
      frec = locateFunctionInTable( fnst->symbol(),
				    fnst->numberOfParameters() );
      if (frec) {
	if (frec->body) {
	  // must create a new symbol record with ignoreme = 1
	  cloneFunctionRecord( frec, fnst );
	}
	else {
	  frec->s = isSgFunctionSymb( fnst->symbol() ); // keeps tags unique
	  frec->body = fnst->symbol()->body();
	}
      }
      else
	fprintf( stderr, "sgCallGraph: Error: No function record for %s/%d\n",
		fnst->symbol()->identifier(),
		fnst->numberOfParameters() );
    }

    // For each C function: if it is a constructor or destructor function,
    // annotate it with the constructors and destructors which it will call.
    // (Note that 'C' means 'C++' in Sage.)

    if (C && (fnst = isSgFuncHedrStmt( st ))) {
      SgSymbol *s = &fnst->name();
      SgMemberFuncSymb *mf;
      
      SgSymbol *cn;
      if ((mf = isSgMemberFuncSymb( s )) && (cn = mf->className())) {
	char *mfi = mf->identifier();
	char *cni = cn->identifier();
	  
	// These statements annotate the constructor/destructor with those
	// of members and bases
	if (strcmp( cni, mfi ) == 0)
	  updateConstructorsFor( fnst, 0, cn, 0, 1 );
	else if (*mfi == '~' && strcmp( cni, mfi+1 ) == 0)
	  updateDestructorsFor( fnst, 0, cn, 1 );
      }
    }
    
    // The following IF nest deals with declared variables of class type; its
    // purpose is to discover calls to constructors and destructors.
    //
    // Assumptions
    // 1. The first statement in any dep file is a GLOBAL statement
    // 2. Every statement except the GLOBAL statement has a chain of
    //    controlParent() ending up at the GLOBAL statement.
    // 3. In the lexNext() order, controlParents behave in a stack-like
    //    fashion.
    
    if (C) {
      // The following statements deal with pushing and popping environments
      if (st->variant() == GLOBAL) {
	stk.push( st );
      }
      else if (st->controlParent() != stk.top()) {
	if (st->controlParent() == prevstmt) {
	  stk.push( prevstmt );
	}
	else {
	  while (stk.top() != 0 && stk.top() != st->controlParent())
	    stk.pop( prevstmt );
	}
      }
      prevstmt = st;
      
      // Code to do constructors and destructors
      if (vdst = isSgVarDeclStmt( st )) {
	// For each variable declaration: if we declare a variable of
	// class/collection type, insert constructor calls and remember 
	// destructor calls for the end of the block or the end of the program.
	//
	// However, looping over statements we encounter variables of 
	// class/collection type declared inside classes... these should 
	// be ignored here!
	SgSymbol *cs;
	
	if (cs = lookForClassOrCollectionSymbol( vdst->type() )) {
	  // basically a class type
	  int i, j;
	  // Some loops over AST contents run from 1..n, some from 0..n-1.
	  // Could we please be consistent??!? (dammit!)
	  for ( i = 0, j = vdst->numberOfSymbols() ; i < j ; i++ ) {
	    SgStatement *cst;
	    
	    if (cs == lookForClassOrCollectionSymbol( vdst->symbol(i)->type(),
						     1 )) {
	      // Got a variable of class or collection type w/o significant
	      // modifiers.
	      cst = st;
	      while (1) {
		int v = cst->variant();
		if (v != GLOBAL && 
		    v != FUNC_HEDR && 
		    v != CLASS_DECL && 
		    v != DERIVED_CLASS_DECL &&
		    v != STRUCT_DECL &&
		    v != UNION_DECL &&
		    v != COLLECTION_DECL &&
		    v != TEMPLATE_DECL &&
		    v != TECLASS_DECL) {
		  cst = cst->controlParent();
		}
		else {
		  break;
		}
	      }
	      if (cst->variant() == FUNC_HEDR) {
		SgExpression *e = vdst->var( i );
		int nargs = 0;
		if (e->variant() == CLASSINIT_OP) {
		  // here, rhs is parameter list -- walk it to find the length.
		  e = e->rhs();
		  while (e) {
		    nargs++;
		    e = e->rhs();
		  }
		}
		updateConstructorsFor( vdst, 0, cs, nargs );
		stk.addsym( cs );
	      }
	      else if (cst->variant() == GLOBAL) {
		symnode *p = new symnode;
		p->sym = cs;
		p->next = glob;
		glob = p;
		globcnt++;
	      }
	    }
	  }
	}
      }
    } // end if (C)
    
    // Fortran and Fortran M stuff: look for procedure calls, and process
    // calls in several variations.

    if (FORTRAN) {
      SgCallStmt *callst;
      SgProsCallStmt *pcallst;
      SgProsCallLctn *plcallst;
      SgProsCallSubm *pscallst;
      
      if (callst = isSgCallStmt( st )) {
	// CALL foo
	SgCallGraphFunRec *t = 
	  locateFunctionInTable( callst->name(), callst->numberOfArgs() );
	updateFunctionTableConnections( t, st, 0 );
      }
      else if (pcallst = isSgProsCallStmt( st )) {
	// Fortran M: PROCESSCALL foo
	SgCallGraphFunRec *t =
	  locateFunctionInTable( pcallst->name(), pcallst->numberOfArgs() );
	updateFunctionTableConnections( t, st, 0 );
      }
      else if (plcallst = isSgProsCallLctn( st )) {
	// Fortran M: PROCESSCALL foo( ... ) LOCATION( ... )
	SgCallGraphFunRec *t =
	  locateFunctionInTable( plcallst->name(), plcallst->numberOfArgs() );
	updateFunctionTableConnections( t, st, 0 );
      }
      else if (pscallst = isSgProsCallSubm( st )) {
	// Fortran M: PROCESSCALL foo( ... ) SUBMACHINE( ... )
	SgCallGraphFunRec *t =
	  locateFunctionInTable( pscallst->name(), pscallst->numberOfArgs() );
	updateFunctionTableConnections( t, st, 0 );
      }	
    }
    
    // General recursive treatment for all languages: for all expressions
    // in the statement, look for function calls in the expression.

    findFunctionCalls(st, st->expr(0), operators );
    findFunctionCalls(st, st->expr(1), operators );
    findFunctionCalls(st, st->expr(2), operators );
  }
  
  // Cleanup after the last function in the file: annotate the main procedure
  // with constructors and destructors for global objects.

  if (C && main) {
    // Loop over global symbols: for each symbol, push it onto both the 
    // front and end of the main function
    while ( glob ) {
      symnode *p;
      
      updateConstructorsFor( main->body(), 0, glob->sym, 0, 0, 1 );
      updateDestructorsFor( main->body(), 0, glob->sym );
      p = glob;
      glob = glob->next;
      delete p;
    }
  }
  else if (C && !main && globcnt) {
    fprintf( stderr,
	    "SgCallGraph: Warning: no main procedure; "
	    "%d global object%s will not be accounted for.\n", 
	    globcnt, (globcnt > 1 ? "s" : "") );
  }
}

/* 
 * buildFunctionTable()
 *
 * This is pass1 of the call graph constructor. It builds a table of
 * the available functions.
 *
 * The table is built by making a single pass over the symbol table 
 * and recording each symbol/frec pair (note, however, HACK No. 1, 
 * in the program heading). 
 */
void SgCallGraph::buildFunctionTable( SgFile *f,
				     optbl &operators /*,
				     Fundeclhash &fdtbl */ )
{
  SgSymbol *s;
  SgFunctionSymb *fname, *fs;
  SgMemberFuncSymb *memfname;
  SgCallGraphFunRec *funrec;
    
  for( s = f->firstSymbol() ; s; s = s->next()){
    if(memfname = isSgMemberFuncSymb(s)) {
      funrec =
	addToTable( memfname, 
		   (memfname->isMethodOfElement() ? SGMOE_FUN : SGMOC_FUN),
		   memfname->className() &&
		   (memfname->className()->variant() == COLLECTION_NAME
		    || memfname->className()->variant() == TECLASS_NAME),
		   memfname->className() /* ,
		   &fdtbl */ );
    }
    else if (fname = isSgFunctionSymb(s)) {
      // isSgFunctionSymb() covers FORTRAN procedures and main programs as well
      funrec = addToTable( fname, SGNORMAL_FUN, 0, 0 /*, &fdtbl */ );
    }
    else if (s->variant() == PROCESS_NAME) {
      // Fortran M _PROCESS_ declaration
      funrec = addToTable( s, SGNORMAL_FUN, 0, 0 /*, &fdtbl */ );
    }
    
    if (funrec == 0) continue;	// function declaration (???)
    
    // Find the main procedure ('main' or 'Processor_main')
    // FIXME: a PROGRAM statement is not required in Fortran, but this
    //        code requires one to find the main program.
    if (isSgFunctionSymb( s ) && !isSgMemberFuncSymb( s )) {
      int doit = 0;
      if (f->languageType() == CSrc)
	doit = (strcmp( s->identifier(), "Processor_Main" ) == 0
		|| strcmp( s->identifier(), "main" ) == 0);
      else if (f->languageType() == ForSrc) 
        doit = (s->variant() == PROGRAM_NAME );
      else {
	fprintf( stderr, "SgCallGraph: Error: unknown language type %d\n",
		f->languageType() );
	exit( 1 );
      }
      
      if (doit) {
	if (main && (strcmp(s->identifier(), main->identifier()) != 0))
	  fprintf( stderr, "SgCallGraph: Warning: several main procedures "
		  "detected in program.\n" );
	main = s;
	mainFunRec = funrec;
      }
    }
    
    // Add overloaded operators to operator table.
    
    if ((fs = isSgFunctionSymb( s )) && f->languageType() == CSrc) {
      // FunctionSymb covers memberfuncs, too.
      SgMemberFuncSymb *mf;
      int op, i;
      SgTypeList *rhs = 0;
      SgType *lhs = 0;
      int np, mfs, isunary;
      
      np = fs->numberOfParameters();
      mfs = (isSgMemberFuncSymb( s ) != 0);
      isunary = (!mfs && np == 1) || (mfs && np == 0);
      
      if (op = isOperatorName( s, isunary )) {
	if (mf = isSgMemberFuncSymb( s )) {
	  if (strcmp( mf->identifier(), "->" ) == 0) {
	    // -> is _unary_.
	    lhs = new SgPointerType( mf->className()->type() );
	  }
	  else if (strcmp( mf->identifier(), "++" ) == 0
		   || strcmp( mf->identifier(), "--" ) == 0) {
	    if (mf->numberOfParameters() == 0) {
	      lhs = mf->className()->type();
	    }
	    else {
	      rhs = new SgTypeList( mf->className()->type(), 0 );
	    }
	  }
	  else {
	    if (strcmp( mf->identifier(), "->*" ) == 0)
	      lhs = new SgPointerType( mf->className()->type() );
	    else
	      lhs = mf->className()->type();
	    for ( i = mf->numberOfParameters() - 1 ; i >= 0 ; i-- )
	      rhs = new SgTypeList( mf->parameter(i)->type(), rhs );
	  }
	}
	else if (fs->numberOfParameters() >= 1) {
	  lhs = fs->parameter(0)->type();
	  if (strcmp( fs->identifier(), "->" ) == 0)
	    lhs = new SgPointerType( lhs );
	  else if (strcmp( fs->identifier(), "++" ) == 0
		   || strcmp( fs->identifier(), "--" ) == 0) {
	    if (fs->numberOfParameters() == 2) {
	      rhs = new SgTypeList( lhs, 0 );
	      lhs = 0;
	    }
	  }
	  else {
	    if (strcmp( fs->identifier(), "->*" ) == 0)
	      lhs = new SgPointerType( lhs );
	    for ( i = fs->numberOfParameters() - 1 ; i >= 1 ; i-- )
	      rhs = new SgTypeList( fs->parameter(i)->type(), rhs );
	  }
	}
	else {
	  fprintf( stderr, "Warning: global 0-ary operator is not legal!\n" );
	  lhs = 0;
	}
	if (lhs != 0 || rhs != 0)
	  operators.add_operator( op, lhs, rhs, funrec );
      }
    }
  }				// end of symbol-stepping for loop
}

#if DEBUG
void difftime( struct timeval t1 )
{
  struct timeval t2;
  long sec, usec;

  gettimeofday( &t2 );
  usec = t2.tv_usec - t1.tv_usec;
  sec = t2.tv_sec - t1.tv_sec;
  if (usec < 0) {
    usec += 1000000;
    sec--;
  }
  fprintf( stderr, "Elapsed time: %ld.%ld\n", sec, usec );
}
#endif

/*
 *  GenCallTree()
 *
 *  The call graph module works in two passes. The first pass builds a
 *  table which maps function symbols to function records, where a function
 *  record is a data type specific to this module. The second
 *  pass creates, for each function X, a list in X's function record 
 *  containing the symbols of all the functions called from X.
 */

void SgCallGraph::GenCallTree( SgFile *f, int use_fundecl_hack )
{
  fundecl_hack = use_fundecl_hack;
  init( f );
  GenCallTree2( f );
  globalF = f;       // hack (sdiwan)
}

void SgCallGraph::GenCallTree2( SgFile *f  )
{
  optbl operators;

  if (fundecl_hack) {
    fundecltbl = new Fundeclhash;
    
    // This prepass annotates each symbol with the statement it was 
    // declared in, making it possible to find the declaration statement
    // of the symbol in constant time. In contrast, declaredInStmt() uses
    // roughly quadratic time.

    SgStatement *st;
    SgVarDeclStmt *vdst;
    SgFuncHedrStmt *fhst;
    for ( st = f->firstStatement() ; st ; st = st->lexNext() ) {
      if (vdst = isSgVarDeclStmt( st )) {
	int i, n;
	for ( i = 0, n = vdst->numberOfSymbols() ; i < n ; i++ )
	  vdst->symbol(i)->addAttribute( 0, (void*)st, 0 );
      }
      else if (fhst = isSgFuncHedrStmt( st )) { 
	fhst->symbol()->addAttribute( 0, (void*)st, 0 );
      }
    }
  }

  buildFunctionTable( f, operators /*, fdtbl */ );

  loopOverStatements( f->firstStatement(),
		     f->languageType() == CSrc,
		     f->languageType() == ForSrc,
		     operators );

  if (fundecl_hack)
    delete fundecltbl;
}

/******************** end of class SgCallGraph **********************/
