#include "structs.h"
#include "match.h"
#include "scan.h"
#include "expr.h"
#include "event.h"
#include "abstract.h"
#include "set.h"
#include "strs.h"


/*
 ***********************************************************************
 * FUNCTION(S):    Access functions for a list of abstract events.     *
 *                 + get_element_from_abstract_list()                  *
 *                 + get_size_of_abstract_list()                       *
 *                 + truncate_abstract_list()                          *
 *                 + free_abstract_event_list()                        *
 *                 + abstract_list_concat()                            *
 * DESCRIPTION:    All the functions are wrappers. Calls related       *
 *                 functions in file "list.c". Probably inefficient, so*
 *                 changing them to macros may help, if that becomes an*
 *                 issue.                                              *
 * NOTES:                                                              *
 ***********************************************************************
*/  


int get_size_of_abstract_list (ABSTRACT_EVENT_LIST *ael)
{
  return get_size_of_list ((LINKED_LIST *) ael);
}


ABSTRACT_EVENT *get_element_from_abstract_list 
  (ABSTRACT_EVENT_LIST *ael, int n)
{
  ABSTRACT_EVENT  *ae = 
    (ABSTRACT_EVENT *) get_element_from_list ((LINKED_LIST *)ael, n);
  return ae;
}


void truncate_abstract_list (ABSTRACT_EVENT_LIST *ael, int n)
{
  truncate_linked_list((LINKED_LIST *)ael, n);
}


void free_abstract_event_list (ABSTRACT_EVENT_LIST *ael)
{
  remove_all_nodes_from_list ((LINKED_LIST *)ael);
  free_list ((LINKED_LIST *) ael);
}


void abstract_list_concat 
  (ABSTRACT_EVENT_LIST **head,  ABSTRACT_EVENT_LIST *new)
{
  linked_list_concat((LINKED_LIST **) head, (LINKED_LIST *) new);
  return;
}

void add_node_to_abstract_list (ABSTRACT_EVENT_LIST **ael, ABSTRACT_EVENT *ae)
{
  add_node_to_list((LINKED_LIST **) ael, (void *) ae);
}



/*
 ***********************************************************************
 * FUNCTION:     allocate_abstract_event()                             *
 * DESCRIPTION:  allocates and initializes a node in the match tree    *
 * NOTES:                                                              *
 ***********************************************************************
*/

ABSTRACT_EVENT *allocate_abstract_event (char *orig, char *name, char *marker,
					 ABSTRACT_EVENT *parent_node)
{
  ABSTRACT_EVENT *ae = (ABSTRACT_EVENT *) malloc (sizeof (ABSTRACT_EVENT));
  ae->error  = 0;
  ae->status = ON;
  ae->dummy = 0;
  ae->cnt = 1;

  ae->orig = StrAlloc (StrLen(orig)+1);
  StrCpy(ae->orig, orig);
  
  if (name) {
    ae->name = StrAlloc (StrLen (name) + MAX_PROC_DIGIT);
    StrCpy(ae->name, name);
  }
  else
    ae->name = NULL;

  if (marker) {
    ae->marker = StrAlloc (StrLen(marker) + 1);
    StrCpy (ae->marker, marker);
  }
  else
    ae->marker = NULL;
  
  ae->prev = NULL;
  ae->down = NULL;
  ae->next = NULL;
  ae->attr = NULL;
  ae->winfo = NULL;
  ae->ep = NULL;
  ae->parent = parent_node;
  ae->ave = NULL;
  ae->clps = NULL;
  ae->slide_clps = NULL;
  return ae;
}


/*
 ***********************************************************************
 * FUNCTION:      free_abstract_event()                                *
 * DESCRIPTION:   Deallocates the whole subtree rooted at the given    *
 *                abstract event. This is consistent with the definit- *
 *                ion of abstract events. An abstract event is a group *
 *                primitive events, and is found by matching them in a *
 *                user-defined way. So the primitive events are return-*
 *                ed to the trace.                                     *
 * NOTES:         Recursive                                            *
 ***********************************************************************
*/

void free_abstract_event (ABSTRACT_EVENT *ae)
{
  if (ae == NULL)
    return;

  ae->ep = NULL;
  free_abstract_event (ae->down);
  free_abstract_event (ae->next);
  free(ae);
  return;
}


/*
 ***********************************************************************
 * FUNCTION:     append_to_sibling_abstract_event()                    *
 * DESCRIPTION:  While building the match tree, the siblings at any    *
 *               level of the match tree are connected together by the *
 *               "next" and "prev" pointers. The siblings arise as the *
 *               "chain" and "pt-chain" definitions are matched.       *
 * CALLER:       o match_chain_definition()                            *
 *               o match_ptchain_definition()                          *
 * NOTES:        Callers are in file "match.c"                         *
 ***********************************************************************
*/

void append_to_sibling_abstract_event
(ABSTRACT_EVENT **ths_ae, ABSTRACT_EVENT *nxt_ae)
{
  ABSTRACT_EVENT *ae = (*ths_ae);

  if (nxt_ae == NULL)
    return;

  if (ae == NULL) {
    (*ths_ae) = nxt_ae;
    return;
  }
  
  while (ae->next) 
    ae = ae->next;

  ae->next = nxt_ae;
  nxt_ae->prev = ae;

  return;
}

/*
 ***********************************************************************
 * FUNCTION:    does_abstract_event_contain_error()                    *
 * DESCRIPTION: Returns "true" if the match tree has an error, "false" *
 *              otherwise. Necessary to stop the matching process, and *
 *              to draw the match tree in the correct color later.     *
 * CALLER:      o match_chain_definition()                             *
 *              o match_pchain_definition()                            *
 *              o match_ptchain_definition()                           *
 * NOTES:       Caller functions are in "match.c"                      *
 ***********************************************************************
*/

boolean does_abstract_event_contain_error (ABSTRACT_EVENT *ae)
{

  if (ae == NULL)
    return FALSE_;
  if (ae->error)
    return TRUE_;
  return (does_abstract_event_contain_error(ae->down) || 
	  does_abstract_event_contain_error (ae->next));
}


/*
 ***********************************************************************
 * FUNCTION:     get_process_set_of_abstract_event()                   *
 * DESCRIPTION:  Given a list of abstract event, finds the "core" set  *
 *               of processes that constitute the abstract events.     *
 *               Duplicate ids are removed. The process set is sorted  *
 *               in ascending order.                                   *
 * CALLER:                                                             *
 * NOTES:                                                              *
 ***********************************************************************
*/

CORE_SET_OBJ 
  *get_process_set_of_abstract_event_list (ABSTRACT_EVENT_LIST *ael)
{
  CORE_SET_OBJ *cso = NULL;
  INT_LINKED_LIST *il = NULL;
  int size = get_size_of_abstract_list (ael);
  int i;
  for (i=0; i<size; i++) {
    ABSTRACT_EVENT *ae = get_element_from_abstract_list (ael, i);

    /* We do not want to include the sibling of "ae" */
    if (ae->down)
      get_process_id_list (&il, ae->down);
    else    /* may be primitive event */
      get_process_id_list (&il, ae);
  }

  cso = make_core_set ((LINKED_LIST *)il);
  return cso;
}


CORE_SET_OBJ 
  *get_process_set_of_abstract_event (ABSTRACT_EVENT *ae)
{
  CORE_SET_OBJ *cso = NULL;
  INT_LINKED_LIST *il = NULL;

  /* We do not want to include the sibling of "ae" */
  if (ae->down)
    get_process_id_list (&il, ae->down);
  else    /* may be primitive event */
    get_process_id_list (&il, ae);

  cso = make_core_set ((LINKED_LIST *)il);
  return cso;
}




void GetPointers (ABSTRACT_EVENT *ae, char *dest, ABSTRACT_EVENT_LIST **ael,
		  int flag)
{
  if (ae == NULL)
    return;
  
  if (!StrCmp(ae->name, dest)) {
    add_node_to_abstract_list (ael, ae);
    GetPointers (ae->next, dest, ael, 1);
    return;
  }
  else {
    GetPointers (ae->down, dest, ael,1);
    if (flag)
      GetPointers (ae->next, dest, ael,1);
    return;
  }
}

int get_pid (ABSTRACT_EVENT *ae, char *name)
{
  if (ae == NULL)
    return 1;
  if (!StrCmp(ae->name, name))
    return (PID(ae->ep));
  else
    return (get_pid (ae->down, name) * get_pid(ae->next, name));
}
  
    

ATTRIBUTE *allocate_attribute (char *name)
{
  ATTRIBUTE *attr = (ATTRIBUTE *) malloc (sizeof(ATTRIBUTE));
  attr->name = StrAlloc (StrLen(name)+1);
  StrCpy(attr->name, name);
  attr->val_list = NULL;
  attr->color = -1;
  attr->next = NULL;
  return attr;
}



INT_LINKED_LIST *get_attr_value (ABSTRACT_EVENT *ae, char *name)
{
  ATTRIBUTE *attr = NULL; 

  /* first check the abstract event */
  if (ae)
    attr = ae->attr;
  if (attr != NULL) {
    while (attr) {
      if (!StrCmp(attr->name, name))
	return attr->val_list;
      attr = attr->next;
    }
  }
  
  /* now check the primitive event */
  if (ae && ae->ep)
    attr = ae->ep->attr;  
  if (attr != NULL) {
    while (attr) {
      if (!StrCmp(attr->name, name))
	return attr->val_list;
      attr = attr->next;
    }
  }

  return NULL;
}

ATTRIBUTE *get_attribute_with_name (ABSTRACT_EVENT *ae, char *aname)
{
  ATTRIBUTE *atr = NULL;
  if (ae)
    atr = ae->attr;
  while (atr) {
    if (!StrCmp(atr->name, aname))
      return atr;
    atr = atr->next;
  }

  /* now check the primitive event */
  if (ae && ae->ep) 
    atr = ae->ep->attr;
  while (atr) {
    if (!StrCmp(atr->name, aname))
      return atr;
    atr = atr->next;
  }    
  
  return NULL;

}


void *apply_attr (void *obj, void *str)
{
  LINKED_LIST *new = NULL;
  ABSTRACT_EVENT *ae = (ABSTRACT_EVENT *) obj;
  char *name = (char *) str;
  VAL_OBJ *vo = allocate_value_object(0);  /* Initialize to zero */

  if (!StrCmp(name,"pid"))
    vo->value = PID(ae->ep);
  else if (!StrCmp(name, "ltime"))
    vo->value = LTIME(ae->ep);
  else if (!StrCmp(name,"mid"))
    vo->value = MsgID(ae->ep);
  else {
    ATTRIBUTE *attr = get_attribute_with_name (ae, name);
    if (attr) {
      /* make a copy of the list */
      LINKED_LIST *ll = NULL;
      int size = get_size_of_list(attr->val_list);
      int i;

      for (i=0; i<size; i++) {
	VAL_OBJ *vo = allocate_value_object(0);
	VAL_OBJ *v1 = (VAL_OBJ *) get_element_from_list (attr->val_list, i);
	vo->value = v1->value;
	add_node_to_list (&ll, (void *)vo);
      }
      return (void *) ll;
    }
    else
      return (void *) NULL;
  }
  add_node_to_list (&new, (void *) vo);
  return (void *) new;
}


     
ABSTRACT_EVENT_LIST *GetTailPointer (ABSTRACT_EVENT_LIST *ael, int *cnt,
				     ABSTRACT_EVENT_LIST *start)
{
  ABSTRACT_EVENT_LIST *tl = start;
  if (ael == NULL)
    return NULL;

  while (tl->next) {
    (*cnt)++;
    tl = tl->next;
  }
  
  (*cnt)++;
  return tl;
}


    
						

void restore_abstract_event (ABSTRACT_EVENT *ae)
{
  if (ae == NULL)
    return;
  
  if (ae->ep) {
    Active(ae->ep) = 1;
    ae->ep = NULL;
  }

  restore_abstract_event (ae->down);
  restore_abstract_event (ae->next);

  return;
}




boolean in_collapse_child (ABSTRACT_EVENT *ae, COLLAPSE_NODE *cc, char *attr)
{
  COLLAPSE_NODE *tcc = cc;
  INT_LINKED_LIST *v1 = get_attr_value (ae, attr);
  
  for (; tcc != NULL; tcc = tcc->next) {
    int size = get_size_of_abstract_list (tcc->ael);
    ABSTRACT_EVENT *tae = get_element_from_abstract_list (tcc->ael, 0);
    INT_LINKED_LIST *v2 = get_attr_value (tae, attr);
    if (list_equal (v1, v2) == TRUE_) {
      ABSTRACT_EVENT_LIST *ael = NULL;
      add_node_to_abstract_list (&ael, ae);
      add_abstract_list_to_collapse_child (tcc, ael);
      tcc->total = size+1;
      return TRUE_;
    }
  }
  return FALSE_;
}
  
boolean IsAttrInEvent (ABSTRACT_EVENT *ae, char *aname)
{
  ATTRIBUTE *attr = ae->attr;
  if (attr == NULL)
    return FALSE_;
  
  while (attr) {
    if (!StrCmp(attr->name, aname))
      return TRUE_;
    attr = attr->next;
  }
  return FALSE_;
}
  
void GetCandidatesForPartition (ABSTRACT_EVENT *ae, char *aname, 
				ABSTRACT_EVENT_LIST **ael, int flag)
{
  if (ae == NULL)
    return;

  if (IsAttrInEvent(ae, aname) == TRUE_) {
    add_node_to_abstract_list (ael, ae);
    if (flag)
      GetCandidatesForPartition (ae->next, aname, ael, 1);
    return;
  }
  else {
    GetCandidatesForPartition (ae->down, aname, ael, 1);
    if (flag)
      GetCandidatesForPartition (ae->next, aname, ael, 1);
    return;
  }
}

void add_collapse_node_to_abstract_event (COLLAPSE_NODE *cn, ABSTRACT_EVENT *ae)
{
  COLLAPSE_NODE *tmp = ae->clps;
  if (tmp == NULL) {
    ae->clps = cn;
    return;
  }
  else {
    while (tmp->next)
      tmp = tmp->next;
    tmp->next = cn;
  }
  return;
}


COLLAPSE_NODE *allocate_collapse_node ()
{
  COLLAPSE_NODE *cc = (COLLAPSE_NODE *) malloc (sizeof (COLLAPSE_NODE));
  cc->ave     = NULL;
  cc->total = cc->part = 1;
  cc->my_id   = cc->par_id  = -1;
  cc->pattern = NULL;
  cc->pae     = NULL;
  cc->parent  = NULL;
  cc->down    = NULL;
  cc->next   = NULL;
  cc->winfo   = NULL;
  cc->scn = NULL;
  cc->ael     = NULL;
  return cc;
}
