/*
 ********************************************************************
 * FILE:           array.c                                          *
 * PROGRAMMER:     Joydip Kundu                                     *
 * ORGANIZATION:   University of Massachusetts, Amherst             *
 *                 University of Oregon, Eugene                     *
 * FUNCTIONS:      + reverse()                                      *
 *                 + int2str()                                      *
 *                 + id2arr()                                       *
 *                 + stl2char()                                     *
 *                 + find_name()                                    *
 *                 + find_size()                                    *
 * NOTES:          Support functions for handling identifier arrays *
 *                 in Ariadne. In Ariadne each element of the array *
 *                 is handled as separate identifiers. A detailed   *
 *                 description can be found in "expr.c"             *
 ********************************************************************
*/

#include <stdio.h>
#include "structs.h"
#include "strs.h"

/*
 ********************************************************************
 * FUNCTION:     reverse()                                          *
 * DESCRIPTION:  Given a string, reverses it.                       *
 ********************************************************************
*/

void reverse (char s[])
{ 
  int c,i,j;
  for (i=0, j=StrLen(s)-1; i<j; i++, j--) {
    c = s[i];
    s[i] = s[j];
    s[j] = c;
  }
}

/*
 *************************************************************************
 * FUNCTION:     int2str()                                               *
 * DESCRIPTION:  Given an integer, returns a string representation of the*
 *               same.                                                   *
 *************************************************************************
*/

char *int2str(int n, int size)
{
  int i = 1;
  char *s = StrAlloc (size);
  if (n<0)
    printf("Very clever LoneStar! But no negative array index is allowed.\n");

  s[0] = '[';
  do {
    s[i++] = n%10 + '0';
  } while ((n /= 10) > 0);
  s[i++] = ']';
  s[i] = '\0';
  reverse(s);
  return s;
}

/*
 ***************************************************************************
 * FUNCTION:     id2arr()                                                  *
 * DESCRIPTION:  Given an string "name", and an integer "i", returns a new *
 *               string "name[i]" e.g., called with "iter" and "5", returns*
 *               "iter[5]".                                                *
 ***************************************************************************
*/

char *id2arr(char* name, int x)
{
  int size_name, size_x;
  char *str_new, *str_x;
  
  if (name)
    size_name = StrLen(name) + 1;
  else
    size_name = 1;

  size_x = find_size(x) + 4;
  
  str_x = int2str(x, size_x);
  str_new = StrAlloc(size_x + size_name);
  str_new[0] = '\0';
  
  if (name)
    StrCat(str_new, name);
  if (str_x)
    StrCat(str_new, str_x);
  return str_new;
}

/*
 ***********************************************************************
 * FUNCTION:     stl2char()                                            *
 * DESCRIPTION:  Given an expression tree, evaluates it and returns a  *
 *               string representation of the integer result, e.g.,    *
 *               given a SYN_TREE_LIST representation of i+3*j where   *
 *               i=3, j=5, will return a string "18".                  *
 ***********************************************************************
*/

char *stl2char (SYN_TREE_LIST *stl)
{
  char *str = NULL;
  if (stl == NULL)
    return str;
  while (stl)
    {
      int x = get_int (evaluate_int_list(stl, NULL));
      char *s = id2arr (str,x);
      free(str);
      str = s;
      stl = stl->next;
    }
  return str;
}

/*
 ***********************************************************************
 * FUNCTION:     find_name()                                           *
 * DESCRIPTION:  Given an unresolved identifier, evaluates the express-*
 *               ion tree associated with it, and returns a string rep-*
 *               resenting the name.                                   *
 ***********************************************************************
*/

char *find_name (IDE *left)
{
  char *stl_char = stl2char (left->stl);
  int  stl_size = 0;
  int  id_size = 0;
  char *new = NULL;

  if (stl_char)
    stl_size = StrLen(stl_char);

  if (left->id_chars)
    id_size = StrLen (left->id_chars);

  new = StrAlloc (id_size + stl_size + 1);
  new[0] = '\0';
  StrCat(new,left->id_chars);
  StrCat(new,stl_char);
  return new;
}

/*
 ***********************************************************************
 * FUNCTION:     find_size()                                           *
 * DESCRIPTION:  Given an integer, finds the number of characters req- *
 *               uired to hold the string representation of the same.  *
 *               Will return 3 if n = 18, 4 if n = 259, and so on.     *
 ***********************************************************************
*/

int find_size (int n)
{
  int x = 0;
  do {
    x++;
  } while ( (n /= 10) > 0);
  return x;
}

