/*
 *************************************************************************
 * FILE:         assert.c                                                *
 * PROGRAMMER:   Joydip Kundu                                            *
 * ORGANIZATION: University of Massachusetts, Amherst                    *
 *               University of Oregon, Eugene                            *
 * DESCRIPTION:  Functions related to checking logical time assertions   *
 *               over abstract events: precedes and parallels            *
 * FUNCTIONS:    + assert_relationship_between_abstract_events()         *
 *               + check_precedes()                                      *
 *               + follow_event_trail()                                  *
 *               + set_up_early_late_set()                               *
 *               + get_early_late_set()                                  *
 * NOTES:                                                                *
 *************************************************************************
*/

#include <stdio.h>
#include <stdlib.h>
#include "structs.h"

#include "abstract.h"
#include "global.h"
#include "event.h"
#include "match.h"
#include "set.h"


/*
 ************************************************************************
 * Include file dependencies:                                           *
 *   structs.h :       Data Structures                                  *
 *   abstract.h:       get_process_set_of_abstract_event()              *
 *   global.h:         get_total_procs()                                *
 *   event.h:          PID, LTIME etc.                                  *
 *   match.h:          ComputeSoSet()                                   *
 *   set.h:            get_core_set_section()                           *
 ************************************************************************
*/             


#define PRECEDES_  0
#define PARALLELS_  1



void get_early_late_set (EVENT **epp, EVENT *ep, int what)
{
  int pid = PID(ep);
  
  if (epp[pid] == NULL) {
    epp[pid] = ep;
    return;
  }

  switch (what) {
  case 0:
    if (LTIME (ep) < LTIME(epp[pid]))
      epp[pid] = ep;
    break;
  case 1:
    if (LTIME (ep) > LTIME(epp[pid]))
      epp[pid] = ep;
    break;
  }
  return;
}

/*
 ************************************************************************
 * FUNCTION:     set_up_early_late_set()                                *
 * DESCRIPTION:  "early(ae)" = {"x" such that pid(x) belongs to the set *
 *                             S_core(ae), and "x" is a primitive event *
 *                             in "ae", and for every "y" != "x" such   *
 *                             pid(y) == pid(x), and "y" is a primitive *
 *                             event in "ae", ltime(x) < ltime(y)}      *
 *               "late(ae)" = {"x" such that pid(x) belongs to the set  *
 *                             S_core(ae), and "x" is a primitive event *
 *                             in "ae", and for every "y" != "x" such   *
 *                             pid(y) == pid(x), and "y" is a primitive *
 *                             event in "ae", ltime(x) > ltime(y)}      *
 *                MaxLtime[] = maximum (ltime(late(ae)))                *
 * DESIGN:                                                              *
 * CALLER:                                                              *
 * NOTES:                                                               *
 ************************************************************************
*/

void set_up_early_late_set (ABSTRACT_EVENT *ae, EVENT **early,
			EVENT **late, int *MaxLtime, int next)
{
  if ((ae == NULL) || (early == NULL) || (late == NULL))     /* sanity check */
    return;

  if (ae->status == ON && ae->ep) {
    get_early_late_set (early, ae->ep, 0);
    get_early_late_set (late, ae->ep, 1);
    if ((*MaxLtime) < LTIME(ae->ep))
      (*MaxLtime) = LTIME(ae->ep);
  }

  if (ae->status == ON)
    set_up_early_late_set (ae->down, early, late, MaxLtime, 1);
  if (next)
    set_up_early_late_set (ae->next, early, late, MaxLtime, 1);
}


/*
 ****************************************************************************
 * FUNCTION:      follow_event_trail()                                      *
 * DESCRIPTION:   Called when we are trying to check relationship between   *
 *                two abstract events A->B say. We first take early(A) and  *
 *                late(B), and follow the event trails starting for early(A)*
 *                to check whether there is any event that precedes an      *
 *                event in late(B). Next we take early(B) and late(A), and  *
 *                check the same.                                           *
 *                cutoff determines when to end the search.                 *
 * DESIGN:                                                                  *
 * CALLER:                                                                  *
 * NOTES:                                                                   *
 ****************************************************************************
*/


boolean follow_event_trail (EVENT *ep, EVENT **lhi, int cutoff)
{
  EVENT *tmp = ep;
  int pid;
  SO_SET *ws;
  EVENT_PTR_LIST *epl;
  boolean result;

  if (ep == NULL)       /* default */
    return FALSE_;

  if (LTIME(ep) > cutoff)  /* end the search, cutoff = Max(LTIME(late)) */
    return FALSE_;
  
  if (lhi[PID(ep)]) {
    if (LTIME(ep) < LTIME(lhi[PID(ep)]))
      return TRUE_;
    else
      return FALSE_;
  }

  result = follow_event_trail (ep->next, lhi, cutoff);
  if (result)
    return result;

  /* if the event in question is a Write or a Multicast, we need 
     to follow the message edges */

  if (!StrCmp(get_event_type(ep),"W") || !StrCmp(get_event_type(ep),"M")) {
    if (SoSet(ep) == NULL)
      SoSet(ep) = ComputeSoSet(ep);
    ws = SoSet(ep);

    epl = ws->epl;
    while (epl != NULL) {
      result = follow_event_trail (epl->ep, lhi, cutoff);
      if (result)
	return result;
      epl = epl->next;
    }
  }
  return FALSE_;
}


/*
 ***********************************************************************
 * FUNCTION:      check_precedes()                                     *
 * DESCRIPTION:   to check whether A --> B, where                      *
 *                "elo"  = early(A)                                    *
 *                "slo"  = "core" set of processes involved in A       *
 *                "lhi"  = late (B)                                    *
 *                "shi"  = "core" set of processes involved in B       *
 *                "cutoff" = max(ltime(shi))                           *
 *                "context"   = PARALLEL_ or PRECEDES_                 *
 * DESIGN:                                                             *
 * CALLER:        check_relationship_between_abstract_events()         *
 * NOTES:                                                              *
 ***********************************************************************
*/

boolean check_precedes (EVENT **elo, CORE_SET_OBJ *slo, 
			EVENT **lhi, CORE_SET_OBJ *shi, 
			int cutoff, int context)
{
  int i;
  CORE_SET_OBJ *cso = get_core_set_section (slo, shi);
  int total_procs = get_total_procs();
  /* If we are doing this in order to assert parallel, and the sets
     intersect, then one has to precede the other. So we return TRUE_
     regardless. Parallel has to fail */

  if ((context == PARALLELS_) && (CoreSetSize(cso) > 0))
    return TRUE_;
    
  for (i=0; i<cso->size; i++) {
    if ( LTIME(elo[cso->elem[i]]) < LTIME(lhi[cso->elem[i]]) )
      return TRUE_;
  }
  
  for (i=0; i<total_procs; i++) {
    if (elo[i] == NULL)
      continue;
    
    if (follow_event_trail(elo[i], lhi, cutoff) == TRUE_)
      return TRUE_;
  }
  
  return FALSE_;
}


/*
 **********************************************************************
 * FUNCTION:     assert_relation ()                                   *
 * DESCRIPTION:  Given an abstract event (ae), a relation (rel), and  *
 *               filter (wrt), it checks whether "rel" holds between  *
 *               the children of "ae"                                 *
 * DESIGN:       + Get all the children of ae                         *
 *               + Get the process-set of each child                  *
 *               + Compute the early and the late set for each child  *
 *               + Now check for relationship                         *
 **********************************************************************
*/

boolean assert_relationship_between_abstract_events
(ABSTRACT_EVENT *ae, WRT_LIST *wrt, OPER rel)
{
  int i,j;
  int child_cnt = count_children_of_abstract_event(ae);

  CORE_SET_OBJ **cso =
    (CORE_SET_OBJ **) calloc (child_cnt, sizeof(CORE_SET_OBJ **));
  EVENT ***early =
    (EVENT ***) calloc (child_cnt, sizeof(EVENT ***));
  EVENT ***late =
    (EVENT ***) calloc (child_cnt, sizeof(EVENT ***));

  int *MaxLtime = (int *)  malloc (child_cnt * sizeof (int ));

  /* for each child of "ae" we need to compute the early and late
     sets, and the process sets that the child spans */

  ABSTRACT_EVENT *child; 
  ABSTRACT_EVENT *tae = ae;
  ABSTRACT_EVENT *tmp;
  boolean res_lo, res_hi;
  int total_procs =  get_total_procs();   /* total number of processes */

  /* Go down until you find a real abstract event. This is required, since
     otherwise we may turn off whole subtress unintentionally */

  while (tae->dummy == 1)
    tae = tae->down;

  switch_off_unlisted_events (ae, wrt);

  /* Note that precedes and parallels are assertions on a list of abstract 
     events. In our match tree implementation, siblings are connected
     in a linked list. In order to check the relationships, we need to
     compute the "early" and "late" sets for each abstract events in the
     list. To ensure, we compute the sets in the right fashion, we need
     to first break the list, and rejoin it later. */

  for (i=0; i<child_cnt; i++) {
    /* keep track of the next pointer */
    tmp = tae->next;
      
    /* break the list */
    tae->next = NULL;
    
    /* allocate "early" and "late" sets */
    early[i] = (EVENT **) calloc (total_procs, sizeof (EVENT **));
    late[i]  = (EVENT **) calloc (total_procs, sizeof (EVENT **));

    /* initialize "early" and "late" sets */
    for (j=0; j<total_procs; j++)
      early[i][j] = late[i][j] = (EVENT *) NULL;

    /* Get the core process set of "ae" */
    cso[i] = get_process_set_of_abstract_event (tae);

    /* set up "early" and "late" sets, and the cutoff logical time */
    set_up_early_late_set (tae, early[i], late[i], &MaxLtime[i], 1);
    
    /* rejoin the siblings */
    if (tae)
      tae->next = tmp;

    /* go to the next sibling */
    tae = tae->next;
  }

  /* "ae" remains unchanged */
  tae = ae;
  
  /* skip over dummy nodes */
  while (tae->dummy == 1)
    tae = tae->down;

  /* child_cnt tells us how many abstract events are there in the list */
  switch(rel){
    /* parallels on "n" abstract events require n(n-1)/2 comparisons */
  case OperParallel:
    for (i=0; i<child_cnt; i++) {
      for (j=1; j<child_cnt; j++) {
	ABSTRACT_EVENT *ttae = tae->next;
	if (i==j)
	  continue;
	res_lo = 
	  check_precedes(early[i], cso[i], late[j], cso[j], 
			 MaxLtime[j], PARALLELS_);
	res_hi = 
	  check_precedes(early[j], cso[j], late[i], cso[i], 
			 MaxLtime[i], PARALLELS_);
	
	if (res_lo || res_hi) {
	  printf ("\n <ptch> \"%s\" is not parallel to \"%s\".\n",
		  tae->marker, ttae->marker);
	  return FALSE_;
	}
      }
      tae = tae->next;
    }
    return TRUE_;
    
  case OperPrecede:
    /* Precedes is a transitive relation, so only (n-1) comparisons */
    for (i=0; i<child_cnt-1; i++) {
      res_lo = 
	check_precedes (early[i], cso[i], late[i+1], cso[i+1], 
			MaxLtime[i+1], PRECEDES_);
      res_hi = 
	check_precedes (early[i+1], cso[i+1], late[i], cso[i], 
			MaxLtime[i], PRECEDES_);
      if (!res_lo || res_hi) {
	printf("\n <ptch> \"%s\" does not precede \"%s\".\n",
	       tae->marker, tae->next->marker);
	return FALSE_;
      }
      tae = tae->next;
    }
    return TRUE_;
  }
}


