/*
 *********************************************************************
 * FILE:          break.c                                            *
 * PROGRAMMER:    Joydip Kundu                                       *
 * ORGANIZATION:  University of Massachusetts                        *
 *                University of Oregon                               *
 * DESCRIPTION:   functions related to setting abstract event-based  *
 *                breakpoints                                        *
 * FUNCTIONS:     + allocate_break_object()                          *
 *                + is_current_breakpoint_compatible()               *
 *********************************************************************
*/


#include <stdio.h>
#include "structs.h"
#include "set.h"
#include "list.h"
#include "expr.h"
#include "event.h"
#include "poeg.h"
#include "match.h"
#include "assert.h"
#include "abstract.h"

/******   GLOBAL VARIABLES *******/
CORE_SET_OBJ *S_core, *S_extra;
extern FILE *GTraceFile;
int backup = 1;       /* keeps track of breakpoint readjustment */

typedef struct break_object {
  EVENT_PTR_LIST *epl;
} BREAK_OBJ;

typedef LINKED_LIST BREAK_LIST;

BREAK_LIST *GBreakPoints = NULL;


BREAK_OBJ *allocate_break_object()
{
  BREAK_OBJ *bo = (BREAK_OBJ *) malloc(sizeof(BREAK_OBJ));
  bo->epl = NULL;
  return bo;
}

/*
 ***********************************************************************
 * FUNCTION:     is_current_breakpoint_compatible()                    *
 * DESCRITION:   Checks whether user request can be satisfied without  *
 *               causing a deadlock.                                   *
 * DESIGN:       + For each breakpoint in the list, we have a list of  *
 *               events in each process that is the footprint of the   *
 *               breakpoint in that process.                           *
 *               + Compare for each process the footprint of the curre-*
 *               nt request with that of the previous request.         *
 *               + If there is a conflict, reject the current request  *
 * CALLER:       
 * NOTES:
 ***********************************************************************
*/

int is_current_breakpoint_compatible (EVENT_PTR_LIST *epl)
{
  int before = 0, after = 0;  /* variables to check conflict */
  int i = 0, size = 0;
  
  /* The very first request is compatible, by definition */
  if (GBreakPoints == NULL)
    return 1;

  /* For each of the breakpoint in the list, we need to check compatibility */
  size = get_size_of_list ((LINKED_LIST *) GBreakPoints);
  for (i=0; i<size; i++) {
    BREAK_OBJ *bo = get_element_from_list((LINKED_LIST *)GBreakPoints, i);
    EVENT_PTR_LIST *bepl = bo->epl;   /* already set breakpoint */
    EVENT_PTR_LIST *mepl = epl;       /* current request */

    /* Our breakpoints are truly global. Each breakpoint must have a 
       footprint in each process. So we do not need to check for 
       process identifiers within the while loop */
    while (bepl) {
      if (LTIME(bepl->ep) >  LTIME(mepl->ep))
	after = 1;
      else if (LTIME(bepl->ep) < LTIME(mepl->ep))
	before = 1;
      bepl = bepl->next;
      mepl = mepl->next;
    }
    /* The following check can be done within the while loop. But is it 
       worth it? Depends on whether one expects the breakpoints to 
       conflict or not. We expect experienced user, who would most likely
       not set conflicting breakpoints. So the test is performed outside. */

    if (after == 1 && before == 1)
      return 0;
    after = before = 0;
  }
  /* No conflict was found */
  return 1;
}
      
  
/*
 ***********************************************************************
 * FUNCTION:     get_event_list_for_current_breakpoint()               *
 * DESCRIPTION:  
 * DESIGN:
 * CALLER:
 * GLOBALS:        o GNumProcs
 *          
 * NOTES:
 ***********************************************************************
*/

EVENT_PTR_LIST *get_event_list_for_current_breakpoint ()
{
  int pid;
  EVENT_PTR_LIST *head = NULL;
  
  /* Starting at process 0, we attempt to find the footprint of the 
     current breakpoint at each process. Note breakpoint_flag is 
     cleared after a breakpoint is set */

  for (pid=0; pid<GNumProcs; pid++) {
    EVENT_PTR_LIST *epl = (EVENT_PTR_LIST *) malloc(sizeof(EVENT_PTR_LIST));
    EVENT *ep;
    epl->next = NULL;
    epl->ep = NULL;
    ep = get_first_event_on_proc(pid);
    while(ep && get_event_breakpoint_flag(ep) == BREAK_NONE)
      ep = ep->next;
    
    /* sanity check */
    if (ep == NULL) {  /* This should never happen */
      /* Well this could happen if some process is not logged */
      printf("Current breakpoint has no footprint on process %d\n", pid);
      printf("Aborting checking compatibility ...\n");
      printf("FILE: break.c, FUNCTION: get_event_list_for_current_breakpoint()\n");
      continue;
    }
    epl->ep = ep;
    epl->next = head;
    head = epl;
  }
  return head;
}
    

void add_node_to_sorted_int_list (INT_LINKED_LIST **il, int value)
{
  VAL_OBJ *nvo;
  int size = get_size_of_list(*il);
  int i;
  for (i=0; i<size; i++) {
    VAL_OBJ *vo = (VAL_OBJ *) get_element_from_list (*il, i);
    if (value < vo->value) {
      nvo = allocate_value_object (value);
      insert_node_in_int_list (il, nvo, i);
      return;
    }
    else if (value == vo->value)
      return;
  }
  /* If we come here, either the list is empty or all elements are 
     less than value */
  nvo = allocate_value_object (value);
  add_node_to_int_list (il, nvo);
}
  

void get_process_id_list (INT_LINKED_LIST **il, ABSTRACT_EVENT *ae)
{
  if (ae->next)
    get_process_id_list (il, ae->next);

  if (ae->down)
    get_process_id_list (il, ae->down);

  if (ae->ep && ae->status == ON) {
    int pid = get_event_pid(ae->ep);
    add_node_to_sorted_int_list (il, pid);
  }
}
    


/*
 **********************************************************************
 * FUNCTION:        highlight_events_in_breakpoint_cut()              *
 * DESCRIPTION:     for all the events before the breakpoint, set     *
 *                  event.breakpoint_cut = BEFORE_BREAK_CUT           *
 *                  the default is                                    *
 *                    event.breakpoint_cut == NO_BREAK_CUT            *
 **********************************************************************
*/

void highlight_events_in_breakpoint_cut (int pid)
{
  EVENT *ep = get_first_event_on_proc(pid);
  while (ep && get_event_breakpoint_flag(ep) == BREAK_NONE) {
    set_event_breakpoint_cut(ep, BEFORE_BREAK_CUT);
    ep = get_event_next (ep);
  }
  if (ep && get_event_breakpoint_flag(ep) == BREAK_AFTER_EVENT)
    set_event_breakpoint_cut(ep, BEFORE_BREAK_CUT);
  return;
}


void make_breakpoint_in_process_unique (int pid, int when)
{
  EVENT *ep;
  

  if (when == BREAK_BEFORE_EVENT) { 
    ep = get_first_event_on_proc(pid);
    while (ep && get_event_breakpoint_flag(ep) != BREAK_BEFORE_EVENT) {
      set_event_breakpoint_cut(ep, BEFORE_BREAK_CUT);
      ep = get_event_next(ep);
    }
    if (ep)
      ep = get_event_next(ep);
    while (ep) {
      set_event_breakpoint_flag(ep, BREAK_NONE);
      ep = get_event_next(ep);
    }
  }
  else { /* when == BREAK_AFTER_EVENT */
    ep = get_last_event_on_proc (pid);
    while (ep && get_event_breakpoint_flag(ep) != BREAK_AFTER_EVENT)
      ep = get_event_prev(ep);

    if (ep) {
      set_event_breakpoint_cut(ep, BEFORE_BREAK_CUT);
      ep = get_event_prev(ep);
    }
      
    while (ep) {
      set_event_breakpoint_flag(ep, BREAK_NONE);
      set_event_breakpoint_cut(ep, BEFORE_BREAK_CUT);
      ep = get_event_prev(ep);
    }
  }      
}

  
  
void add_break_point_for_unrelated_processes ()
{
  int total_proc = get_total_procs();
  int pid;
  for (pid=0; pid<total_proc; pid++) {
    EVENT *ep = get_first_event_on_proc(pid);
    while (ep && (EnumType(ep) == Phase))
      ep = get_event_next(ep);
    if (ep && (get_event_breakpoint_cut(ep) == NONE_BREAK_CUT))
      set_event_breakpoint_flag(ep, BREAK_BEFORE_EVENT);
  }
}




void switch_off_events_in_breakpoint_cut (EVENT *ep)
{
  if (get_event_breakpoint_cut(ep) == BEFORE_BREAK_CUT)
    set_event_breakpoint_cut (ep, AFTER_BREAK_CUT);
  
  ep = get_event_next (ep);
  
  while (get_event_breakpoint_cut(ep) == BEFORE_BREAK_CUT) {
    set_event_breakpoint_cut (ep, AFTER_BREAK_CUT);
    set_event_breakpoint_flag (ep, BREAK_NONE);
    ep = get_event_next(ep);
  }
  return;
}

void make_me_the_only_breakpoint (EVENT *ep) 
{
  EVENT *tep = get_first_event_on_proc (PID(ep));
  while (tep != ep) {
    set_event_breakpoint_flag(tep, BREAK_NONE);
    tep = get_event_next(tep);
  }
}

int consistency_check (int pid)
{
  char *msg1 = "breakpoint in process %d needs to be moved back\n";
  char *msg2 = "since the corresponding send in process %d is not\n";
  char *msg3 = "in the previous cut\n";
  int x;
  EVENT *ep = get_first_event_on_proc (pid);

  while ((ep != NULL) && get_event_breakpoint_cut(ep) == BEFORE_BREAK_CUT) { 
    /* I am before the breakpoint */
    if (!StrCmp(get_event_type(ep), "R") && (OID(ep) >= 0) ) {
      SO_SET *so = ComputeSoSet(ep);
      EVENT *snd = so->epl->ep;  /* Get the matching send */
      if (get_event_breakpoint_cut(snd) != BEFORE_BREAK_CUT) {
	/* But my corresponding send is NOT in my cut */
	int snd_pid = get_event_pid (snd);
	CORE_SET_OBJ *tso;
	INT_LINKED_LIST *il = NULL;
	VAL_OBJ *vo = allocate_value_object (snd_pid);
	add_node_to_int_list(&il, vo);
	tso = make_core_set (il);

	if (is_element_in_set (tso, S_core) == 1) {
	  printf("bpoint moved in process %d due to process %d\n", 
		 pid, snd_pid);
	  set_event_breakpoint_flag (ep, BREAK_BEFORE_EVENT); 
	  switch_off_events_in_breakpoint_cut (ep);
	  /* We may need to backup other breakpoints in other processes */
	  backup++;
	  free_core_set_object (tso);
	  return 0;
	}
	else {
	  S_extra = get_core_set_union (S_extra, tso);
	  /* We need to advance the breakpoint frontier in snd_pid */
	  set_event_breakpoint_flag (snd, BREAK_AFTER_EVENT);
	  highlight_events_in_breakpoint_cut (snd_pid);
	  x = consistency_check(snd_pid);
	  if (x == 0) {
	    set_event_breakpoint_cut (ep, BREAK_BEFORE_EVENT);
	    switch_off_events_in_breakpoint_cut (ep);
	    /* We may need to backup other breakpoints in other processes */
	    backup++;
	    free_core_set_object (tso);
	    return 0;
	  }
	  /* otherwise check if there was an earlier breakpoint */
	  make_me_the_only_breakpoint (snd);
	}
	free_core_set_object(tso);
      }
    }
    ep = get_event_next(ep);
  }
  return 1;
}
	
	
void tentative_setting_of_breakpoint(ABSTRACT_EVENT *ae, int when)
{
  int cnt = 0;
  int i;
  EVENT **early, **late;
  CORE_SET_OBJ *cso = NULL;
  int max_logical_time = 0;

  /* cnt will hold the max pid of the prim event in ae */
  count_switched_on_primitive_events(ae, &cnt);  


  early = (EVENT **) calloc (cnt+1, sizeof(EVENT **));
  late  = (EVENT **) calloc (cnt+1, sizeof(EVENT **));

  for (i=0; i<=cnt; i++) 
    early[i] = late[i] = (EVENT *) NULL;

  set_up_early_late_set (ae, early, late, &max_logical_time, 0);

  if (when == BREAK_AFTER_EVENT) 
    for (i=0; i<=cnt; i++) {
      if (late[i])
	set_event_breakpoint_flag(late[i], BREAK_AFTER_EVENT);
    }
  else if (when == BREAK_BEFORE_EVENT)
    for (i=0; i<=cnt; i++) {
      if (early[i])
	set_event_breakpoint_flag(early[i], BREAK_BEFORE_EVENT);
    }
}

void add_new_breakpoint()
{
}
  

/*
 ***********************************************************************
 * FUNCTION:         break_after_select_abstract_event()               *
 * DESCRIPTION:      Given an abstract event tries to set a consistent *
 *                   breakpoint immediately before/ after it. Failing  *
 *                   so, sets the latest consistent breakpoint at an   *
 *                   earlier state. Notifies user as to what went wrong*
 * DESIGN:           + Get the set of processes spanned by the abstract*
 *                   event, call this  S_core                          *
 *                   + Initialize S_other to empty set                 *
 *                   + Highlight the events in S_core to the tentative *
 *                   consistent cut that the user requested.           *
 *                   + foreach process in S_core,                      *
 *                        o for each Read event before the breakpoint  *
 *                          cut, find the corresponding Write event.   *
 *                        o if the Write event belongs to a process in *
 *                          S_Core, and is before the breakpoint,      *
 *                          proceed, otherwise we need to set a break  *
 *                                                                     *
 ***********************************************************************
*/

void break_with_select_abstract_event (ABSTRACT_EVENT_LIST *ael, int when)
{
  EVENT_PTR_LIST *epl = NULL;
  int pid, size;
  int i;
  /* First get the set of processes that form the abstract event */
  S_core = get_process_set_of_abstract_event_list (ael);
  S_extra = NULL;

  /* Initialize backup */
  backup = 1;
  

  /* Set a tentative breakpoint */
  size = get_size_of_abstract_list (ael);
  for (i=0; i<size; i++) {
    ABSTRACT_EVENT *ae = get_element_from_abstract_list (ael, i);
    tentative_setting_of_breakpoint(ae, when);
  }

  size = CoreSetSize(S_core);
  /* Remove multiple breakpoints from the same process */
  /* Highlight all the events in the process before the breakpoint */

  for (pid=0; pid<size; pid++) {
    make_breakpoint_in_process_unique (S_core->elem[pid], when);
    highlight_events_in_breakpoint_cut (S_core->elem[pid]);
  }
  while (backup != 0) {
    backup = 0;
    for (pid=0; pid<size; pid++)
      consistency_check (S_core->elem[pid]);
  }

  /* Some processes that are unrelated should be stopped at their earliest */
  add_break_point_for_unrelated_processes();

  /* Is the current breakpoint compatible with the previously set ones */
  epl = get_event_list_for_current_breakpoint();
  if (is_current_breakpoint_compatible(epl) == 1) {
    BREAK_OBJ *bo = allocate_break_object();
    bo->epl = epl;
    add_node_to_list((LINKED_LIST **)&GBreakPoints, (void *) epl);
    add_new_breakpoint();
    store_trace_file();
  }
  else {
    printf("New breakpoint although consistent overlaps with others\n");
    printf("Breakpoint not set\n");
  }
  
}
