/*
 *************************************************************************
 * FILE:         decl.c                                                  *
 * PROGRAMMER:   Joydip Kundu                                            *
 * ORGANIZATION: University of Massachusetts, Amherst                    *
 *               University of Oregon, Eugene                            *
 * DESCRIPTION:  Functions related to handling "set", "chain", "pchain", *
 *               and "ptchain" declarations. Note that both "chain" and  *
 *               "ptchain" refer to same object type TREE_OBJ            *
 * FUNCTIONS:    + allocate_decl()                                       *
 *               + free_set_object()                                     *
 *               + free_tree_object()                                    *
 *               + free_pchain_object()                                  *
 * NOTES:                                                                *
 *************************************************************************
*/


/******************** include files **************************************/

#include "structs.h"
#include "array.h"
#include "expr.h"
#include "scan.h"
#include "set.h"
#include "strs.h"




/*
 ************************************************************************
 * Include file dependencies:                                           *
 *   structs.h ------->Data Structures                                  *
 *   array.h --------->find_name()                                      *
 *   expr.h ---------->get_int()                                        *
 *                     evaluate_int_list()
 *   scan.h ---------->id_find()                                        *
 *   set.h ----------->make_core_set()                                  *
 *                     get_core_set_union()                             *
 *                     get_core_set_section()                           *
 *                     get_core_set_direct_product()                    *
 ************************************************************************
*/             


/*
 *********************************************************************
 * FUNCTION(S):   Allocation routines.                               *
 *                o allocate_decl(): DECL node allocator             *
 *                o allocate_ide() : IDE node allocator              *
 *                o allocate_imap(): Index Map Object allocator      *
 *                o allocate_set():  Named set allocator             *
 *                o allocate_tree_object(): chain/pt-chain allocator *
 *                o allocate_pchain_object(): pchain allocator       *
 *********************************************************************
*/

DECL *allocate_decl (DECL_KIND dk)
{
  DECL *d = (DECL *) malloc (sizeof(DECL));
  d->decl_next = NULL;
  d->decl_kind = dk;
  d->decl_object = NULL;
  return d;
}

IDE *allocate_ide()
{
  IDE *ide = (IDE *) malloc (sizeof(IDE));
  ide->decl_kind = DeclUnknown;
  ide->id_chars = NULL;
  ide->stl = NULL;
  return ide;
}

IMAP_OBJ *allocate_imap_object (SYN_TREE_LIST *stl)
{
  IMAP_OBJ *imo = (IMAP_OBJ *) malloc (sizeof(IMAP_OBJ));
  imo->stl = stl;
  return imo;
}


SET_OBJ *allocate_set()
{
  int i;
  SET_OBJ *so = (SET_OBJ *) malloc (sizeof(SET_OBJ));
  so->condsyn = NULL;
  so->ide = NULL;
  so->map = NULL;
  so->name = NULL;
  so->core = NULL;
  return so;
}

TREE_OBJ *allocate_tree_obj( DECL_KIND dk)
{
  TREE_OBJ *temp = (TREE_OBJ *) malloc (sizeof (TREE_OBJ));
  temp->decl_kind = dk;
  temp->ide = NULL;
  temp->oper = OperNone;
  temp->cnt = 1;
  temp->flag = -1;
  temp->name = temp->marker = NULL;
  temp->down = temp->next = NULL;
  temp->wrt = NULL;
  temp->dfl = NULL;
  return temp;
}


PCH_OBJ *allocate_pchain()
{
  PCH_OBJ *po = (PCH_OBJ *) malloc (sizeof (PCH_OBJ));
  po->chain = NULL;
  po->set = NULL;
  return po;
}


/*
 *************************************************************************
 * FUNCTION(S):   Deallocates objects pointed by the current declaration *
 *                of an identifier.                                      *
 *                o free_set_object():    frees a set object             *
 *                o free_tree_object():   frees a chain/pt-chain object  *
 *                o free_pchain_object(): frees a p-chain object         *
 * NOTES:         Order of the function definitions is important.        *
 *************************************************************************
*/

void free_set_object (SET_OBJ *so)
{
  free (so);
  return;
}


void free_tree_object (TREE_OBJ *co)
{
  TREE_OBJ *dwn;
  TREE_OBJ *nxt;
  if (!co)
    return;
  dwn = co->down;
  nxt = co->next;
  co->down = co->next = NULL;
  free(co);
  free_tree_object(dwn);
  free_tree_object(nxt);
  return;
}

void free_pchain_object (PCH_OBJ *po)
{
  TREE_OBJ *co;
  SET_OBJ *so;

  if (!po) 
    return;
  co = po->chain;
  so = po->set;
  
  po->chain = NULL;
  po->set = NULL;
  free(po);
  free_tree_object(co);
  free_set_object(so);
  return;
}

 
/* 
 ************************************************************************
 * FUNCTION:     free_previous_declaration()                            *
 * DESCRIPTION:  Called when a variable is redeclared. Ariadne being    *
 *               an interpretive environment allows for dynamic redecl- *
 *               -aration. The previous declaration is reclaimed, and   *
 *               hence lost.                                            *
 ************************************************************************
*/
  
void free_previous_declaration(IDENT *id)
{
  switch (id->id_decl->decl_kind) {
  case DeclSet:
    {
      SET_OBJ *so = (SET_OBJ *) (id->id_decl->decl_object);
      free_set_object (so);
      return;
    }
  case DefPchain:
    {
      PCH_OBJ *po = (PCH_OBJ *) (id->id_decl->decl_object);
      free_pchain_object (po);
      return;
    }
  case DeclChain:
  case DefTchain:
    {
      TREE_OBJ *to = (TREE_OBJ *) (id->id_decl->decl_object);
      free_tree_object (to);
      return;
    }
  case DeclMatch:
    {
      ABSTRACT_EVENT *ae = (ABSTRACT_EVENT *) (id->id_decl->decl_object);
      free_abstract_event (ae);
      return;
    }
  default:
    return;
  }
}
	


void create_decl_int (IDENT *id)
{
  id->id_decl = allocate_decl(DeclInt);
  id->id_decl->decl_object = (OBJECT *) malloc (sizeof (OBJECT));
  return;
}

/*
 **********************************************************************
 * FUNCTION(S):   Replication routines.                               *
 *                o replicate_set_object()                            *
 *                o replicate_tree_object()                           *
 *                o replicate_pchain_object()                         *
 * DESCRIPTION:   Why replicate?                                      *
 *                Consider the statement:                             *
 *                   for (i= 0 to 7 by 1) (* Modula-3 for loop *)     *
 *                    begin                                           *
 *                     set A[i] = {i .. 3*i};                         *
 *                    end                                             *
 *                 Within the for-statement, we for a set object      *
 *                 corresponding to "A[i]". As we walk through the    *
 *                 index variable "i", we replicate "A[i]" and then   *
 *                 evaluate it to "A[0]", "A[2]", ..., "A[6]". WIthout*
 *                 replication the original definition of "A[i]" would*
 *                 be lost. Same reasons for chain, p-chain and       *
 *                 pt-chain definitions.                              *
 * NOTES:          I am not sure, whether a better solution exists    *
 **********************************************************************
*/

SET_OBJ *replicate_set_object (SET_OBJ *so)
{
  SET_OBJ *new = allocate_set();
  new->condsyn = so->condsyn;
  new->map = so->map;
  new->ide = so->ide;
  return new;
}


TREE_OBJ *replicate_tree_object (TREE_OBJ *to, DECL_KIND dk)
{
  TREE_OBJ *new;
  if (to == NULL)
    return (TREE_OBJ *) NULL;
  if (dk != to->decl_kind) {
    printf("Do you know that the implementation may have a bug!");
    return (TREE_OBJ *) NULL;
  }
  new = allocate_tree_obj(dk);
  new->oper = to->oper;
  new->cnt = to->cnt;
  new->ide = to->ide;
  new->name = NULL;
  new->wrt = to->wrt;
  new->dfl = to->dfl;
  if (to->marker) {
    new->marker = StrAlloc (StrLen (to->marker) + 1);
    strcpy (new->marker, to->marker);
  }
  else
    new->marker = NULL;

  new->flag = to->flag;

  new->down = replicate_tree_object (to->down, dk);
  new->next = replicate_tree_object (to->next, dk);
  return new;
}

PCH_OBJ *replicate_pchain_object (PCH_OBJ *po)
{
  SET_OBJ *new_set = replicate_set_object (po->set);
  TREE_OBJ *new_chain = replicate_tree_object (po->chain, DeclChain);
  PCH_OBJ *new = allocate_pchain();
  new->chain = new_chain;
  new->set = new_set;
  new->on_kind = po->on_kind;
  return new;
}


/*
 *********************************************************************
 * FUNCTION(S):  Interpretation of different definitions.            *
 *               o interpret_set_object()                            *
 *               o interpret_tree_object()                           *
 *               o interpret_pchain_object()                         *
 *               o interpret_identifier_tobe_defined()               *
 * DESCRIPTION:  Resolves an unresolved definition.                  *
 * NOTES:        Recursive.                                          *
 *********************************************************************
*/

boolean interpret_tree_object (TREE_OBJ *to)
{
  if (to == NULL)
    return TRUE_;

  if (to->ide) {
    char *new = find_name(to->ide);
    IDENT *id = id_find (new);

    if (check_error(new, to->decl_kind) == FALSE_)
      return FALSE_;

    if (to->decl_kind == DeclChain)
      to->flag = id->id_reserved;
    else if ( (id->id_decl->decl_kind == DeclPchain))
      to->flag = 1;

    to->name = new;
    to->ide = NULL;
    
    if (to->marker == NULL) {
      to->marker = StrAlloc(StrLen(new)+1);
      strcpy(to->marker, new);
    }
  }
  if (interpret_tree_object(to->down) && interpret_tree_object(to->next))
    return TRUE_;
  else
    return FALSE_;
}

boolean interpret_set_object (SET_OBJ *so)
{
  SYN_TREE_LIST *stl = so->condsyn;
  int i =  0;

  if (so->ide) {
    SET_OBJ *tso;
    char *new = find_name(so->ide);
    IDENT *id = id_find (new);
    if (error_check_for_identifier(id, new) == 0)
      return FALSE_;
    if (id->id_decl->decl_kind != DeclSet) {
/*      printf(terse_feedback, "%s is not a set.\n", id->id_chars); */
      printf ("%s is not a set.\n", id->id_chars);
      return FALSE_;
    }
            
    so->name = new;
    so->ide = NULL;
    return TRUE_;
  }

  if (stl->next == NULL) { /* we do not have comma separated list */
    /* we have either a..b, or set A = {0}; types */
    INT_LINKED_LIST *loc = evaluate_int_list (stl, NULL);
    /* first element is dimension */
    if (loc->next) {
      so->core = make_core_set (loc->next);
      so->core->dim = get_int (loc);
    }
    else {  /* if we have just one element */
      so->core = make_core_set (loc);
      so->core->dim = 1;
    }

    so->core->size /= so->core->dim;
  }
  else {
    so->core = make_core_set (evaluate_int_list(stl,NULL));
  }
  stl = stl->next;


  while (stl)
    {
      CORE_SET_OBJ *t1 = so->core;
      CORE_SET_OBJ *t2 = make_core_set (evaluate_int_list(stl,NULL));
      CORE_SET_OBJ *t3 = get_core_set_union (t1, t2);
      so->core = t3;
      stl = stl->next;
      free_core_set_object(t1);
      free_core_set_object(t2);
    }

  return TRUE_;
}


boolean interpret_pchain (PCH_OBJ *po)
{
  if (interpret_tree_object(po->chain) && interpret_set_object(po->set))
    return TRUE_;
  else
    return FALSE_;
}

/*
 ***********************************************************************
 * FUNCTION:    interpret_identifier_tobe_defined()                    *
 * DESCRIPTION: Resolves an unresolved identifier in a defintion.      *
 *              The unresolved identfier forms the left hand side of   *
 *              a definition.                                          *
 * NOTES:       "level" determines the scope of the identifier. In     *
 *              Ariadne, one can assign values to identifiers within   *
 *              a "chain" or a "ptchain" expression. These identifiers *
 *              stay alive only within the matching context.           *
 ***********************************************************************
*/

IDENT *interpret_identifier_tobe_defined (IDE *left, int level)
{
  IDENT *id; 
  DECL *d;
  char *new = find_name(left);
  id = id_find_insert(new);

  if (id->id_decl == NULL || level) {
    d = allocate_decl(left->decl_kind);
    d->decl_next = id->id_decl;
    id->id_decl = d;
  }
  id->id_decl->decl_kind = left->decl_kind;
  return id;
}


/*
 ************************************************************************
 * FUNCTION(S): Handles different kinds of chain/pt-chain definitions.  *
 *              o build_tree_item(): chain or pt-chain                  *
 *              o build_unresolved_tree_item(): Array element           *
 *              o concat_tree_objects(): chain or pt-chain              *
 *              o build_special_chain_symbol(): @, PUSH_, POP_          *
 * DESCRIPTION: Called when a single item in a "chain" or a "ptchain"   *
 *              declaration is encountered.                             *
 ************************************************************************
*/

TREE_OBJ *build_tree_item (TREE_OBJ *obj, int cnt, WRT_LIST *wrt)
{
  TREE_OBJ *t = allocate_tree_obj(obj->decl_kind);
  t->down = obj;
  t->cnt = cnt;
  t->wrt = wrt;
  return t;
}

TREE_OBJ *build_unresolved_tree_item (IDE *ide, DECL_KIND dk)
{
  TREE_OBJ* temp = allocate_tree_obj(dk);
  temp->ide = ide;
  return temp;
}



void concat_tree_objects (TREE_OBJ *head, TREE_OBJ *tail)
{
  TREE_OBJ *temp = head;
  while (temp->next != NULL)
    temp = temp->next;
  temp->next = tail;
  return;
}

TREE_OBJ* build_special_chain_symbol (OPER op)
{
  TREE_OBJ *c = allocate_tree_obj(DeclChain);
  c->oper = op;
  return c;
}




/*
 ************************************************************************
 * FUNCTION(S):   Handles different kinds of set definitions.           *
 *                o get_set_union():  set A = B union C;                *
 *                o get_set_intersection: set A = B section C;          *
 *                o get_set_direct_product: set A = B cross C;          *
 ************************************************************************
*/

CORE_SET_OBJ *get_set_union (SET_OBJ *s1, SET_OBJ *s2)
{
  return get_core_set_union(CoreSet(s1), CoreSet(s2));
}


CORE_SET_OBJ *get_set_intersection (SET_OBJ *s1, SET_OBJ *s2)
{
  return get_core_set_section(CoreSet(s1), CoreSet(s2));
}

CORE_SET_OBJ *get_set_direct_product (SET_OBJ *s1, SET_OBJ *s2)
{
  return get_core_set_direct_product (CoreSet(s1), CoreSet(s2));
}


