/*
 ***********************************************************************
 * FILE:         event.c                                               *
 * PROGRAMMER:   Joydip Kundu                                          *
 * ORGANIZATION: University of Massachusetts, Amherst                  *
 *               University of Oregon, Eugene                          *
 * DESCRIPTION:  Functions related to allocation and deallocation of   *
 *               events, and setting and retrieving of different fields*
 * FUNCTIONS:    + allocate_event()                                    *
 *               + get_event_pid()                                     *
 *               + get_event_otherid()                                 *
 *               + get_event_logical_time()                            *
 *               + get_event_breakpoint_flag()                         *
 *               + get_event_next()                                    *
 *               + get_event_prev()                                    *
 *               + get_event_name()                                    *
 *               + get_event_type()                                    *
 *               + set_event_logical_time()                            *
 *               + increment_event_breakpoint_count()                  *
 *               + decrement_event_breakpoint_count()                  * 
 * NOTES:                                                              *
 ***********************************************************************
*/

#include <stdio.h>
#include "structs.h"


/*
 ***********************************************************************
 * FUNCTION:        Allocation routines                                *
 *                  + allocate_event()                                 *
 *                  + allocate_event_ptr_list()                        *
 * CALLER:                                                             *
 * NOTES:                                                              *
 ***********************************************************************
*/  

EVENT *allocate_event()
{
  EVENT *ep = (EVENT *) malloc (sizeof(EVENT));
  ep->myid = ep->otherid = -1;
  sprintf(ep->event_type, "U");
  ep->event_tag[0] = '\0';
  ep->event_size = -1;
  
  ep->instr_flag = 0;
  ep->event_name[0] = '\0';
  ep->ltime = ep->msg_id = -1;
  ep->bp_flag = BREAK_NONE;
  ep->bp_cut = NONE_BREAK_CUT;
  ep->bpcount = 0;
  ep->error = -1;

  ep->active = TRUE_;
  ep->filtered = FALSE_;
  
  ep->so_set = NULL;

  ep->prev = ep->next = NULL;
  ep->attr = NULL;
  return ep;
}

EVENT_PTR_LIST *allocate_event_ptr_list (EVENT *ep)
{
  EVENT_PTR_LIST *epl = (EVENT_PTR_LIST *) malloc (sizeof (EVENT_PTR_LIST));
  epl->ep = ep;
  epl->next = NULL;
  return epl;
}



/*
 ***********************************************************************
 * FUNCTION(S):    retrieval routines for different fields in an event *
 *                 + get_event_pid()                                   *
 *                 + get_event_otherid()                               *
 *                 + get_event_logical_time()                          *
 * NOTES:                                                              *
 ***********************************************************************
*/

int get_event_pid (EVENT *ep)
{
  return ep->myid;
}

int get_event_otherid (EVENT *ep)
{
  return ep->otherid;
}

int get_event_logical_time (EVENT *ep)
{
  return ep->ltime;
}

int get_event_breakpoint_flag (EVENT *ep)
{
  return ep->bp_flag;
}


int get_event_breakpoint_cut (EVENT *ep)
{
  return ep->bp_cut;
}

EVENT *get_event_next (EVENT *ep)
{
  if (ep)
    return ep->next;
  else
    return NULL;
}

EVENT *get_event_prev (EVENT *ep)
{
  if (ep)
    return ep->prev;
  else
    return NULL;
}

char *get_event_name (EVENT *ep)
{
  return ep->event_name;
}

char *get_event_type (EVENT *ep)
{
  return ep->event_type;
}


/*
 ***********************************************************************
 * FUNCTION(S):    + set_event_logical_time()                          *
 *                 + set_event_breakpoint_flag()                       *
 *                 + set_event_breakpoint_cut()                        *
 * DESCRIPTION:    setting different fields of the event               *
 * CALLER:                                                             *
 * NOTES:                                                              *
 ***********************************************************************
*/

void set_event_logical_time (EVENT *ep, int value)
{
  ep->ltime = value;
  return;
}

void set_event_breakpoint_flag (EVENT *ep, int value)
{
  ep->bp_flag = value;
  return;
}

void set_event_breakpoint_cut (EVENT *ep, int value)
{
  ep->bp_cut = value;
  return;
}

EVENT_TYPE EnumType(EVENT *ep)
{
  if (!StrCmp(ep->event_type,"W"))
    return Write;
  else if (!StrCmp(ep->event_type,"R"))
    return Read;
  else if (!StrCmp(ep->event_type,"M"))
    return MCast;
  else if (!StrCmp(ep->event_type,"P"))
    return Phase;
  else if  (!StrCmp(ep->event_type,"U"))
    return UserDefined;
  else
    return Unknown;
}



/*
 ***********************************************************************
 * FUNCTION(S):   + increment_breakpoint_count()                       *
 *                + decrement_breakpoint_count()                       *
 * DESCRIPTION:   We use a single integer to code the number of break- *
 *                points that can be set in the context of a primitive *
 *                event. With 32 bit integer, the higher order 16 bits *
 *                are used to keep track of breakpoints set before the *
 *                primitive event, and the lower order 16 bits for     *
 *                those set after the primitive event. This allows for *
 *                2^16 + 2^16 total breakpoints.                       *
 * CALLER:                                                             *
 * NOTES:                                                              *
 ***********************************************************************
*/

void increment_breakpoint_count (EVENT *ep, int where)
{
  int before;
  int after; 
  
  if (ep == NULL) {    /* sanity check */
    printf("WARNING: add_to_break_point_count() passed a NULL event\n");
    return;
  }

  before = after = ep->bpcount;
  before = (before >> 16);
  after  = after & 0x00ff;

  if (where == BREAK_BEFORE_EVENT)
    before++;
  else if (where == BREAK_AFTER_EVENT)
    after++;
  
  ep->bpcount = ((before << 16) + after);
  return;
}


void decrement_breakpoint_count (EVENT *ep, int where)
{
  int before; 
  int after;
  if (ep == NULL)
    return;

  before = after = ep->bpcount;
  before = (before >> 16);
  after  = after & 0x00ff;

  if (where == BREAK_BEFORE_EVENT) {
    if (before == 0)
      return;
    else
      before--;
  }
  else if (where == BREAK_AFTER_EVENT) {
    if (after == 0) 
      return;
    else
      after--;
  }
  
  ep->bpcount = ((before << 16) + after);
  return;
}
