#include "structs.h"
#include "abstract.h"
#include "list.h"
#include "scan.h"
#include "set.h"
#include "strs.h"

extern boolean list_equal (INT_LINKED_LIST *v1, INT_LINKED_LIST *v2);

/*
 ************************************************************************
 * Include file dependencies:                                           *
 *   structs.h :       Data Structures                                  *
 *   abstract.h:       evaluate_abstract_list()                         *
 *                     apply_attr()                                     *
 *   list.h:           compute_new_list()                               *
 *                     add_node_to_list()                               *
 *                     insert_node_in_list()                            *
 *   scan.h:           id_find()                                        *
 *   set.h:            get_cross_product_in_list()                      *
 *                     get_intersection_in_list()                       *
 *                     get_union_in_list()                              *
 *                     get_int_list_from_set()                          *
 ************************************************************************
*/             

INT_LINKED_LIST 
*get_intersection_of_list (INT_LINKED_LIST *left, INT_LINKED_LIST *right)
{
  CORE_SET_OBJ *cleft = NULL;
  CORE_SET_OBJ *cright = NULL;
  CORE_SET_OBJ *cso;
  INT_LINKED_LIST *result, *tmp;
  if (left == NULL || right == NULL)
    return NULL;

  cleft = make_core_set (left);
  cright = make_core_set (right);
  
  cleft->dim = 1;
  cright->dim = 1;
  cso = get_core_set_section (cleft, cright);
  tmp = get_int_list_from_set (cso);
  result = tmp->next;
  if (result)
    result->prev = NULL;
  tmp->next = NULL;
  free(tmp->contents);
  free(tmp);
  free_core_set_object (cleft);
  free_core_set_object (cright);
  free_core_set_object (cso);
  return result;
}
  
void remove_duplicate (INT_LINKED_LIST *il)
{
  int len = get_size_of_list ((LINKED_LIST *) il);
  int i;
  if (len <= 1)
    return;

  for (i=0; i<len; i++) {
    VAL_OBJ *v1 = (VAL_OBJ *) get_element_from_list (il, i);
    int j;
    
    for (j=i+1; j<len; j++) {
      VAL_OBJ *v2 = (VAL_OBJ *) get_element_from_list (il, j);
      if (v1->value == v2->value) {
	/* remove v2 */
	delete_node_from_list (&il, (void *)v2);
	len--;
	free(v2);
      }
    }
  }
}


/*
 *************************************************************************
 * FUNCTION(S):    Access functions for list of integers (INT_LIST)      *
 *                 + insert_node_in_int_list()                           *
 *                 + add_node_to_int_list()                              *
 * DESCRIPTION:    Calls functions in "list.c"                           *
 *************************************************************************
*/

void insert_node_in_int_list (INT_LINKED_LIST **il, VAL_OBJ *vo, int pos)
{
  insert_node_in_list ((LINKED_LIST **) il, (void *) vo, pos);
}


void add_node_to_int_list (INT_LINKED_LIST **il, VAL_OBJ *vo)
{
  add_node_to_list ((LINKED_LIST **) il, (void *) vo);
}


/*
 ************************************************************************
 * FUNCTION:     allocate_value_object()                                *
 ************************************************************************
*/

VAL_OBJ *allocate_value_object (int n)
{
  VAL_OBJ *vo = (VAL_OBJ *) malloc (sizeof(VAL_OBJ));
  vo->value = n;
  return vo;
}


SYN_TREE *allocate_syntree()
{
  SYN_TREE *st = (SYN_TREE *) malloc (sizeof (SYN_TREE));
  st->left = st->right = NULL;
  st->name = NULL;
  st->value = -1;
  st->oper = ApplyNone;
  return st;
}


/*
 *********************************************************************
 * FUNCTION(S):  Functions related to various list operators.        *
 *               + equal_value(e1, e2)    e1 == e2                   *
 *               + geq_value(e1, e2)      e1 >= e2                   *
 *               + gt_value(e1, e2)       e1 > e2                    *
 *               + leq_value(e1, e2)      e1 <= e2                   *
 *               + lt_value(e1, e2)       e1 < e2                    *
 * DESCRIPTION:  Binary operators                                    *
 *********************************************************************
*/

boolean equal_value (void *e1, void *e2)
{
  VAL_OBJ *v1 = (VAL_OBJ *) e1;
  VAL_OBJ *v2 = (VAL_OBJ *) e2;

  if (v1 == NULL || v2 == NULL)
    return FALSE_;

  if (v1->value != v2->value)
    return FALSE_;

  return TRUE_;
}

boolean geq_value (void *e1, void *e2)
{
  VAL_OBJ *v1 = (VAL_OBJ *) e1;
  VAL_OBJ *v2 = (VAL_OBJ *) e2;

  if (v1 == NULL || v2 == NULL)
    return FALSE_;

  if (v1->value >= v2->value)
    return TRUE_;

  return FALSE_;
}


boolean gt_value (void *e1, void *e2)
{
  VAL_OBJ *v1 = (VAL_OBJ *) e1;
  VAL_OBJ *v2 = (VAL_OBJ *) e2;

  if (v1 == NULL || v2 == NULL)
    return FALSE_;

  if (v1->value > v2->value)
    return TRUE_;

  return FALSE_;
}

boolean leq_value (void *e1, void *e2)
{
  VAL_OBJ *v1 = (VAL_OBJ *) e1;
  VAL_OBJ *v2 = (VAL_OBJ *) e2;

  if (v1 == NULL || v2 == NULL)
    return FALSE_;

  if (v1->value <= v2->value)
    return TRUE_;

  return FALSE_;
}
  
boolean lt_value (void *e1, void *e2)
{
  VAL_OBJ *v1 = (VAL_OBJ *) e1;
  VAL_OBJ *v2 = (VAL_OBJ *) e2;

  if (v1 == NULL || v2 == NULL)
    return FALSE_;

  if (v1->value < v2->value)
    return TRUE_;

  return FALSE_;
}



/*
 ***************************************************************************
 * FUNCTION:      build_id_node()                                          *
 * DESCRIPTION:   Returns a SYN_TREE pointer. The reason is as follows:    *
 *                Ariadne does not require the user to declare an array    *
 *                beforehand. Array elements A[1], A[2], etc are not acce- *
 *                with an offset from a base, as is done in standard com-  *
 *                piling; instead they are treated as separate identifiers.*
 *                This works in our case, since the structural hierarchy of*
 *                "chain-pchain-ptchain" supports parallelism and sequence *
 *                quite well to the point where array constructs are not   *
 *                truly required. So when we see a definition/use A[i], we *
 *                need to build a SYN_TREE for "i", and substitute its     *
 *                value later, when we process the definition.             *
 * NOTES:                                                                  *
 ***************************************************************************
*/

SYN_TREE *build_id_node (IDENT *id)
{
  SYN_TREE *st = allocate_syntree ();
  st->name = StrAlloc(StrLen(id->id_chars)+1);
  strcpy(st->name, id->id_chars);
  return st;
}

/*
 **************************************************************************
 * FUNCTION:       build_int_node()                                       *
 * DESCRIPTION:                                                           *
 **************************************************************************
*/

SYN_TREE *build_int_node (int i)
{
  SYN_TREE *st = allocate_syntree();
  st->value = i;
  return st;
}


SYN_TREE *build_syntax_tree
  (SYN_TREE_LIST *st_left, SYN_TREE_LIST *st_right, ExprOper op, char* name)
{
  SYN_TREE *st = allocate_syntree();
  st->oper = op;
  st->left = st_left;
  st->right = st_right;
  st->value = -1;
  if (name) {
    st->name = StrAlloc (StrLen(name)+1);
    strcpy(st->name, name);
  }
  else
    st->name = NULL;
  return st;
}

/*
 *************************************************************************
 * FUNCTION:     add_syntree_list_to_ide()                               *
 * DESCRIPTION:  IDE is a partial IDENT. At the parsing phase, Ariadne   *
 *               cannout resolve A[i], and so it makes a IDE node for "A"*
 *               and connect a SYN_TREE_LIST corresponding to the "i" to *
 *               IDE. The IDE is resolved as an IDENT when "i" is evalu- *
 *               ated later.                                             *
 *************************************************************************
*/

void add_syntree_list_to_ide (IDE *ide, SYN_TREE_LIST *stl)
{
  SYN_TREE_LIST *temp = ide->stl;
  if (temp == NULL) {
    ide->stl = stl;
    return;
  }
  
  while (temp->next)
    temp = temp->next;

  temp->next = stl;
  return;
}


/* scope_of_context may be LOCAL_ or GLOBAL */

INT_LINKED_LIST *evaluate_int_list 
(SYN_TREE_LIST *stl, ABSTRACT_EVENT *context)
{
  int prod, sum;
  IDENT *id;
  ABSTRACT_EVENT_LIST *ael;
  SYN_TREE_LIST *s1, *s2;
  INT_LINKED_LIST *p1, *p2, *p3, *til;
  INT_LINKED_LIST *result = NULL;
  INT_LINKED_LIST *result2 = NULL;
  int val, size;
  int size1, size2;
  LINKED_LIST *pll;

  int e1, e2, by,i;
  VAL_OBJ *vo = NULL;

  if (stl == NULL)
    return NULL;

  switch (stl->st->oper) {
  case ApplySet:
    e1 = 
      get_int (evaluate_int_list(stl->st->left, context));
    e2 = 
      get_int (evaluate_int_list(stl->st->right, context));
    by = stl->st->value;
    vo = allocate_value_object (1);
    add_node_to_list (&result, vo);

    for (i=e1; i<=e2; i+=by) {
      vo = allocate_value_object (i);
      add_node_to_list(&result, vo);
    }
    return result;
  case ApplyCross:
    p1 = evaluate_int_list (stl->st->left, context);
    p2 = evaluate_int_list (stl->st->right, context);
    result = get_cross_product_in_list (p1, p2);
    return result;
  case ApplySection:
    p1 = evaluate_int_list (stl->st->left, context);
    p2 = evaluate_int_list (stl->st->right, context);
    result = get_intersection_in_list (p1, p2);
    return result;
  case ApplyUnion:
    p1 = evaluate_int_list (stl->st->left, context);
    p2 = evaluate_int_list (stl->st->right, context);
    result = get_union_in_list (p1,p2);
    return result;
  case ApplyAbs:
    e1 = get_int (evaluate_int_list(stl->st->left, context));
    if (e1 >= 0)
      vo = allocate_value_object(e1);
    else
      vo = allocate_value_object(-e1);

    add_node_to_list (&result, vo);
    return result;

  case ApplyAdd:
    e1 = get_int (evaluate_int_list(stl->st->left, context));
    e2 = get_int (evaluate_int_list(stl->st->right, context));
    vo = allocate_value_object(e1+e2);
    add_node_to_list (&result, vo);
    return result;

  case ApplySub:
    e1 = get_int (evaluate_int_list(stl->st->left, context));
    e2 = get_int (evaluate_int_list(stl->st->right, context));
    vo = allocate_value_object(e1-e2);
    add_node_to_list (&result, vo);
    return result;

  case ApplyMult:
    e1 = get_int (evaluate_int_list(stl->st->left, context));
    e2 = get_int (evaluate_int_list(stl->st->right, context));
    vo = allocate_value_object(e1*e2);
    add_node_to_list (&result, vo);
    return result;

  case ApplyDiv:
    e1 = get_int (evaluate_int_list(stl->st->left, context));
    e2 = get_int (evaluate_int_list(stl->st->right, context));

    if (e2 != 0)
      vo = allocate_value_object(e1/e2);
    add_node_to_list (&result, vo);
    return result;
    
  case ApplyDistinct:
    pll = evaluate_int_list (stl->st->left, context);
    val = are_elements_distinct ((LINKED_LIST *) pll, equal_value);
    vo = allocate_value_object (val);
    add_node_to_list (&result, vo);
    return result;

  case ApplyDisjoint:
    s1 = stl->st->left;
    if (s1)
      s2 = s1->next;
    else
      s2 = NULL;
    p1 = (INT_LINKED_LIST *) evaluate_int_list (s1, context);
    p2 = (INT_LINKED_LIST *) evaluate_int_list (s2, context);
    size1 = get_size_of_list ((LINKED_LIST *)p1);
    size2 = get_size_of_list ((LINKED_LIST *)p2);
    printf("length before: size(p1)= %d, size(p2)= %d ", size1, size2);
    printf("p1: size = %d\n", size1);
    for (i=0; i<size1; i++) {
      VAL_OBJ *x = (VAL_OBJ *) get_element_from_list((LINKED_LIST *)p1, i);
      if (x)
      printf("%d ", x->value);
    }
    printf("\n");

    printf("p2: size = %d\n", size2);
    for (i=0; i<size2; i++) {
      VAL_OBJ *x = (VAL_OBJ *) get_element_from_list((LINKED_LIST *)p2, i);
      if (x) 
      printf("%d ", x->value);
    }
    printf("\n");
      
    remove_duplicate (p1);
    remove_duplicate (p2);

    size1 = get_size_of_list ((LINKED_LIST *)p1);
    size2 = get_size_of_list ((LINKED_LIST *)p2);

    printf("length after %d ", size1);
    for (i=0; i<size1; i++) {
      VAL_OBJ *x = (VAL_OBJ *) get_element_from_list((LINKED_LIST *)p1, i);
      if (x)
	printf("%d ", x->value); 
    }
    printf("\n");

    linked_list_concat((LINKED_LIST **) &p1, (LINKED_LIST *) p2);

    size = get_size_of_list ((LINKED_LIST *)p1);
    printf("length after concat %d ", size);
    for (i=0; i<size; i++) {
      VAL_OBJ *x = (VAL_OBJ *) get_element_from_list((LINKED_LIST *)p1, i);
      /* printf("%d ", x->value); */
    }
    printf("\n");

    val = are_elements_distinct ((LINKED_LIST *) p1, equal_value);
    vo = allocate_value_object (val);
    add_node_to_list (&result, vo);
    return result;

  case ApplySuperSet:
    /* A is a superset of B */
    s1 = stl->st->left;
    if (s1) {
      s2 = s1->next;
      s1->next = NULL;
    }
    else
      s2 = NULL;
    p1 = (INT_LINKED_LIST *) evaluate_int_list (s1, context);
    p2 = (INT_LINKED_LIST *) evaluate_int_list (s2, context);

    /* Don't ask me I just work here -Joydip */
    if (s2)
      s1->next = s2;
    remove_duplicate (p1);
    remove_duplicate (p2);

    p3 = get_intersection_of_list (p1, p2);
    /* p3 should be equal to p2 */

    if (list_equal (p2, p3) == TRUE_)
      val = 1;
    else
      val = 0;
    vo = allocate_value_object (val);
    add_node_to_list (&result, vo);
    return result;
  case ApplySubSet:
    /* A is a subset of B */
    s1 = stl->st->left;
    if (s1)
      s2 = s1->next;
    else
      s2 = NULL;
    p1 = (INT_LINKED_LIST *) evaluate_int_list (s1, context); /* A */
    p2 = (INT_LINKED_LIST *) evaluate_int_list (s2, context); /* B */

    remove_duplicate (p1);
    remove_duplicate (p2);

    p3 = get_intersection_of_list (p1, p2);
    /* p3 should be equal to p2 */

    if (list_equal (p1, p3) == TRUE_)
      val = 1;
    else
      val = 0;
    vo = allocate_value_object (val);
    add_node_to_list (&result, vo);
    return result;
  case ApplySize:
    pll = evaluate_int_list (stl->st->left, context);
    val = get_size_of_list (pll);
    vo = allocate_value_object (val);
    add_node_to_list (&result, vo);
    return result;

  case ApplyEqual:
    pll = evaluate_int_list (stl->st->left, context);
    val = check_ordered_relation (pll, equal_value);
    vo = allocate_value_object (val);
    add_node_to_list (&result, vo);
    return result;

  case ApplyGec:
    pll = evaluate_int_list (stl->st->left, context);
    val = check_ordered_relation (pll, geq_value);
    vo = allocate_value_object (val);
    add_node_to_list (&result, vo);
    return result;

  case ApplyGtc:
    pll = evaluate_int_list (stl->st->left, context);
    val = check_ordered_relation (pll, gt_value);
    vo = allocate_value_object (val);
    add_node_to_list (&result, vo);
    return result;

  case ApplyLec:
    pll = evaluate_int_list (stl->st->left, context);
    val = check_ordered_relation (pll, leq_value);
    vo = allocate_value_object (val);
    add_node_to_list (&result, vo);
    return result;

  case ApplyLtc:
    pll = evaluate_int_list (stl->st->left, context);
    val = check_ordered_relation (pll, lt_value);
    vo = allocate_value_object (val);
    add_node_to_list (&result, vo);
    return result;

  case ApplyProd:
    val = 1;
    pll = evaluate_int_list (stl->st->left, context);
    size = get_size_of_list(pll);
    for (til = pll; til; til = til->next)
      val *= get_int(til);

    vo = allocate_value_object (val);
    add_node_to_list (&result, vo);
    return result;

  case ApplySum:
    val = 0;
    pll = evaluate_int_list (stl->st->left, context);
    size = get_size_of_list(pll);
    for (til = pll; til; til = til->next)
      val += get_int(til);

    vo = allocate_value_object (val);
    add_node_to_list (&result, vo);
    return result;
  case ApplyNone:
    if (stl->st->name) {
      /* May be a variable with int value, or an attribute name */
      id = id_find(stl->st->name);
      if (id->id_reserved != 2) {
	/* it may be a set or an integer */
	if (id->id_decl == NULL) {
	  printf("%s has not been defined\n", id->id_chars);
	  return NULL;
	}
	if (id->id_decl->decl_kind != DeclSet) {
	  /* it may be NULL */
	  val = id->id_decl->decl_object->val_obj.value;
	  vo = allocate_value_object (val);
	  add_node_to_list (&result, vo);
/*	  return result; */
	}
	else { /* it is a set */
	  SET_OBJ *so = (SET_OBJ *) id->id_decl->decl_object;
	  result =  get_int_list_from_set(so->core);
	  /* return result; */
	}
      }

      ael = evaluate_abstract_list (stl->st->left, context);
      result = 
	(INT_LINKED_LIST *) 
	  compute_new_list((LINKED_LIST *) ael, apply_attr, 
			   (void *) stl->st->name);
      /* return result; */
    }
    else {
      val = stl->st->value;
      vo = allocate_value_object (val);
      add_node_to_list (&result, vo);
      /* return result; */
    }
    result2 = evaluate_int_list (stl->next, context);
    linked_list_concat (&result, result2);
    return result;
  }
}
    
    
boolean list_equal (INT_LINKED_LIST *v1, INT_LINKED_LIST *v2)
{
  int i;
  int s1 = get_size_of_list(v1);
  int s2 = get_size_of_list(v2);

  if ( (v1==NULL) && (v2==NULL) ) {
    printf("Null value in attribute\n");
    return TRUE_;
  }

  if (s1 != s2)
    return FALSE_;

  for (i=0; i<s1; i++) {
    VAL_OBJ *p1 = (VAL_OBJ *) get_element_from_list (v1, i);
    VAL_OBJ *p2 = (VAL_OBJ *) get_element_from_list (v2, i);

    if ( (p1 == NULL) || (p2 == NULL) )
      return FALSE_;
    
    if ( (p1->value) != (p2->value) )
      return FALSE_;
  }

  return TRUE_;
}

int get_int (INT_LINKED_LIST *il)
{
  int *pval = (int *) get_element_from_list(il, 0);
  int value;
  if (pval == NULL)
    value = -1;
  else 
    value = *pval;
  return value;
}
