#include "structs.h"


LINKED_LIST *allocate_linked_list()
{
  LINKED_LIST *this = (LINKED_LIST *) malloc(sizeof(LINKED_LIST));
  this->prev = this->next = NULL;
  this->contents = NULL;
  return this;
}



/*
 ***********************************************************************
 * FUNCTION:         add_node_to_list()                                *
 * DESCRIPTION:      Adds a node to a linked list.                     *
 ***********************************************************************
*/

void add_node_to_list (LINKED_LIST **ll, void *contents)
{
  LINKED_LIST *tmp = (*ll);
  LINKED_LIST *new = allocate_linked_list();
  new->contents = contents;
  
  if (tmp == NULL) {
    (*ll) = new;
    return;
  }

  while (tmp->next)
    tmp = tmp->next;
  
  tmp->next = new;
  new->prev = tmp;
  return;
}


/*
 *********************************************************************
 * FUNCTION:    delete_node_from_list()                              *
 * DESCRIPTION: Given an element pointer, deletes the list item that *
 *              points to the element. The element itself is not     *
 *              freed, since we do not know whether there are other  *
 *              pointers pointing to it. The programmer has to ensure*
 *              that the element is freed, when it is unneeded.      *
 *********************************************************************
*/

void delete_node_from_list(LINKED_LIST **head, void *contents)
{
  LINKED_LIST *ll = *head;
  /* If the list is empty do nothing */
  if (ll == NULL)
    return;
  
  /* Try to find the element */
  while (ll && ll->contents != contents)
    ll = ll->next;
  
  /* ll now points to the node to be deleted. This should work 
   even when the list contains just 1 element */
  if (ll->prev == NULL)
    (*head) = ll->next;
  else 
    ll->prev->next = ll->next;

  if (ll->next)
    ll->next->prev = ll->prev;

  /* Now free the link node but not the object */
  ll->contents = NULL;
  ll->prev = ll->next = NULL;
  free(ll);
}


 
void apply_transform_on_elements (LINKED_LIST *ll, void (*transform)()) 
{
  while (ll) {
    (*transform)(ll->contents);
    ll = ll->next;
  }
  return;
}
/*
LINKED_LIST *compute_new_list 
  (LINKED_LIST *ll, void* (*list_maker)(void *, void *),  void *arg1)
{
  LINKED_LIST *new = NULL;
  while (ll) {
    void *contents = (void *) (*list_maker)(ll->contents, arg1);
    add_node_to_list (&new, contents);
    ll = ll->next;
  }
  return new;
}
  
*/



/*
 *********************************************************************
 * FUNCTION:    linked_list_concat ()                                *
 * DESCRIPTION: concatenates two linked lists.                       *
 *********************************************************************
*/

void linked_list_concat (LINKED_LIST **head,  LINKED_LIST *new)
{
  LINKED_LIST *tmp = *head;
  if (*head == NULL) {
    (*head) = new;
    return;
  }

  if (new == NULL)
    return;
  while (tmp->next) 
    tmp = tmp->next;
  tmp->next = new;
  new->prev = tmp;
  return;
}

LINKED_LIST *compute_new_list 
  (LINKED_LIST *ll, void* (*list_maker)(void *, void *),  void *arg1)
{
  LINKED_LIST *new = NULL;
  while (ll) {
    LINKED_LIST *result = (LINKED_LIST *) (*list_maker)(ll->contents, arg1);
    linked_list_concat (&new, result);
    ll = ll->next;
  }
  return new;
}

/*
 *********************************************************************
 * FUNCTION:     get_size_of_list()                                  *
 *********************************************************************
*/

int get_size_of_list (LINKED_LIST *ll)
{
  int cnt = 0;
  while (ll) {
    ++cnt;
    ll = ll->next;
  }
  return cnt;
}


void *get_element_from_list (LINKED_LIST *ll, int pos)
{
  int i = 0;
  while (i<pos && ll) {
    ll = ll->next;
    i++;
  }

  if (ll)
    return ll->contents;
  else
    return NULL;
}
  

int are_elements_distinct (LINKED_LIST *ll, boolean (*match)(void *, void *) )
{
  LINKED_LIST *p1, *p2;
  for (p1 = ll; p1; p1 = p1->next) {
    p2 = p1->next;
    while (p2) {
      if ((*match)(p1->contents, p2->contents) == TRUE_)
	return 0;
      p2 = p2->next;
    }
  }
  return 1;
}
    

int check_ordered_relation 
  (LINKED_LIST *ll, boolean (*match)(void *, void *) )
{
  LINKED_LIST *p1 = ll;
  for (p1 = ll; p1 && p1->next; p1 = p1->next) {
    if ((*match)(p1->contents, p1->next->contents) == FALSE_)
      return 0;
  }
  return 1;
}

void truncate_linked_list (LINKED_LIST *ll, int n)
{
  int i;
  LINKED_LIST *tail;

  if (get_size_of_list(ll) <= n)
    return;

  for (i=0; i<n; i++) 
    ll = ll->next;
  
  tail = ll->next;
  tail->prev = NULL;
  ll->next = NULL;

  /* We need to add code to delete tail */
}
  
void insert_node_in_list (LINKED_LIST **ll, void *contents, int pos)
{
  int i;
  LINKED_LIST *tmp = (*ll);
  LINKED_LIST *new = allocate_linked_list();
  new->contents = contents;

  if (tmp == NULL) {
    (*ll) = new;
    return;
  }

  for (i=0; i<pos; i++)
    tmp = tmp->next;

  /* tmp points to the element before which new will be inserted */
  new->next = tmp;
  new->prev = tmp->prev;
  
  tmp->prev = new;

  if (new->prev)
    new->prev->next = new;
  else
    (*ll) = new;
}

void remove_all_nodes_from_list (LINKED_LIST *ll)
{
 LINKED_LIST *p = ll;
 while (p) {
   p->contents = NULL;
   p = p->next;
 }
 return;
}

void free_list (LINKED_LIST *ll)
{
  LINKED_LIST *p, *q;
  
  p = ll;
  while (p) {
    q = p->next;
    p->prev = NULL;
    free(p);
    p = q;
  }
  ll = NULL;
  return;
}
