#include <stdio.h>
#include "jkmacros.h"
#include "structs.h"
#include "abstract.h"
#include "event.h"
#include "expr.h"
#include "global.h"
#include "init.h"
#include "list.h"
#include "strs.h"


#define FORMAT_VERSION 1


int total_procs;  
int GMaxMsgID  = 0;
char CurrentFileName[64];


EVENT **GFirstEventPtr;
EVENT_PTR_LIST **GEvntQ;
EVENT_PTR_LIST **GEvntStack;
EVENT **GExecGraph;           /* current event pointer array */
EVENT **GTailGraph;
FILE *GTraceFile;
FILE *GDumpFile;
HASH_TABLE  *GEventHashTab;
extern int StackNum;
int ltime[MAX_PROCS];


/*
 ********************************************************************
 * FUNCTION(S):   Getting events on a particular process.           *
 *                o get_first_event_on_proc()                       *
 *                o get_current_event_on_proc()                     *
 * DESCRIPTION:   + GFirstEventPtr keeps track of the first event   *
 *                on a process.                                     *
 *                + GExecGraph keeps track of the event pointer as  *
 *                it moves with the progress of the matching algo.  *
 ********************************************************************
*/

EVENT *get_first_event_on_proc (int pid)
{
  if (pid < total_procs)
    return GFirstEventPtr[pid];
  else
    return NULL;
}

EVENT *get_last_event_on_proc (int pid)
{
  if (pid < total_procs)
    return GTailGraph[pid];
  else
    return NULL;
}

EVENT *get_current_event_on_proc (int pid)
{
  if (pid < total_procs)
    return GExecGraph[pid];
  else
    return NULL;
}



/*
 ***********************************************************************
 * FUNCTION:      get_msg_id()                                         *
 * DESCRIPTION:   Given a process identifier (pid) and the local count *
 *                saying the number of messages sent so far by the     *
 *                process, it returns a unique identifier              *
 * NOTE:          Assumes maximum number of processes = 256            *
 *                Uses the first 12 bit as process identifier          *
 *                With the current crop of machines, we will be safe   *
 *                for a while.                                         *
 *                Note msg_id is a signed int (need to be changed).    *
 ***********************************************************************
*/

int get_msg_id (int pid, int count)
{
  int msg_id = (pid << 20)+count;
  return msg_id;
}

/*
 ***********************************************************************
 * FUNCTION:      set_msg_id_of_matching_read()                        *
 * DESCRIPTION:   finds the matching read of a "W" or "M" event, and   *
 *                sets its message-identifier to be that of the origi- *
 *                nating "W" or "M"                                    *
 * DESIGN:        + if it is a "W", we know the process that received  *
 *                the "W". Look for the first "R" that is from "me"    *
 *                whose msg_id has not yet been set. Set it.           *
 *                + if it is a Multicast, then for all other processes *
 *                do the same for all processes except me.             *
 *                                                                     *
 * NOTE:          Assumes FIFO order of message delivery or processor  *
 *                consistency.                                         *
 ***********************************************************************
*/

void set_msg_id_of_matching_read (EVENT *ep)
{
  int msgid = MsgID(ep);
  int otherid = OID(ep);
  int myid = PID(ep);
  int total_procs = get_total_procs();
  int pid;

  if (EnumType(ep) == MCast) { /* multicast */
    for(pid=0; pid<total_procs; pid++) {
      EVENT *tep = get_first_event_on_proc(pid);
      if (pid == myid)
	continue;
      while (tep) {
	if ((!StrCmp(EventType(tep), "R")) && (OID(tep) == myid) &&
	    (MsgID(tep) == AA_NONE)) {
	  MsgID(tep) = msgid;
	  break;
	}
	tep = tep->next;
      }
    }
  }
  else {
    EVENT *tep = get_first_event_on_proc(otherid);
    while (tep) {
      if (!StrCmp(EventType(tep), "R") && (OID(tep) == myid) &&
	  (MsgID(tep) == AA_NONE)) {
	MsgID(tep) = msgid;
	break;
      }
      tep = tep->next;
    }
  }
  
}

/*
 ***********************************************************************
 * FUNCTION:      set_logical_time()                                   *
 * DESCRIPTION:   sets up logical time for the current process         *
 * DESIGN:        + start at the beginning                             *
 *                + if event is not a "R", then logical time can be    *
 *                set by incrementing ltime[pid]                       *
 *                + if event is a "R", get the logical time of the     *
 *                matching "W" or "M". If that time is greater than my *
 *                local clock. Set local clock = Logical time of (W/M) *
 *                Set logical time by incrementing local clock.        *
 *                                                                     *
 * NOTE:          Implements Lamport's algorithm for incrementing local*
 *                clocks.                                              *
 ***********************************************************************
*/


void set_logical_time(int pid) 
{
  EVENT *ep = get_current_event_on_proc(pid);
  
  while (ep) {
    if (LTIME(ep) == AA_NONE) {
      if ((EnumType(ep) == Read) && OID(ep) >= 0) {
	int cand = get_logical_time_from_other_proc (OID(ep), MsgID(ep));
	if (cand > ltime[pid])
	  ltime[pid] = cand;
      }
      LTIME(ep) = ++ltime[pid];
    }
    ep = ep->next;
  }
}


/*
 ***********************************************************************
 * FUNCTION:      get_logical_time_from_other_proc()                   *
 * DESCRIPTION:                                                        *
 * DESIGN:        + start at the beginning                             *
 *                + if event is not a "R", and logical time is not set *
 *                set it by incrementing the local clock.              *
 *                + if event is a "R", get the logical time of the     *
 *                matching "W" or "M". If that time is greater than my *
 *                local clock. Set local clock = Logical time of (W/M) *
 *                Set logical time by incrementing local clock.        *
 *                                                                     *
 * NOTE:          set_logical_time() and get_logical_time_from_other_  *
 *                proc() together simulates the program execution.     *
 ***********************************************************************
*/

int get_logical_time_from_other_proc (int pid, int msgid)
{
  EVENT *ep = get_current_event_on_proc(pid);
  
  while(ep && (MsgID(ep) != msgid)) {
    if (LTIME(ep) == AA_NONE) {
      /* For Reads we have to make sure it is not a dummy read */
      if ((EnumType(ep) == Read) && (OID(ep) >= 0)) {
	int cand = get_logical_time_from_other_proc (OID(ep), MsgID(ep));
	if (cand > ltime[pid])
	  ltime[pid] = cand;
      }
      LTIME(ep) = ++ltime[pid];
    }
    ep = ep->next;
  }
  
  if (ep) {
    if (LTIME(ep) == AA_NONE)
      LTIME(ep) = ++ltime[pid];
    
    return (LTIME(ep));
  }
  else {
    printf("Trace has an orphan Receive\n");
    return -1;
  }
}
    
      

/*
 ***********************************************************************
 * FUNCTION:      process_trace_for_ariadne()                          *
 * DESCRIPTION:   Adds logical time, and message identifier for Ariadne*
 ***********************************************************************
*/
      
void process_trace_for_ariadne ()
{
  int pid;
  int total_procs = get_total_procs();
  
  /* initialize logical time */
  for (pid=0; pid<total_procs; pid++)
    ltime[pid] = 0;

  /* Number all writes, Multicasts and Reads */
  for (pid=0; pid<total_procs; pid++) {
    int cnt = 1;
    EVENT *ep = get_first_event_on_proc(pid);
    while (ep) {
      if (!StrCmp(EventType(ep), "W") || !StrCmp(EventType(ep), "M") || 
	  (!StrCmp(EventType(ep), "R") && (OID(ep) == AA_UNDEFINED))) {
	++cnt;
	MsgID(ep) = get_msg_id (pid, cnt);
	set_msg_id_of_matching_read (ep);
      }
      ep = get_event_next(ep);
    }
  }

  /* Now set the logical times: the hard part */
  for (pid=0; pid<total_procs; pid++)
    set_logical_time (pid);
    
  
}

/*
 ***********************************************************************
 * FUNCTION:      insert_event_in_exec_graph()                         *
 * DESCRIPTION:   o Inserts event in the Execution Graph (GExecGraph)  *
 *                o Updates GTailGraph                                 *
 ***********************************************************************
*/

void insert_event_in_exec_graph (EVENT *ep)
{
  int pid = PID(ep);
  switch (EnumType(ep))
    {
    case Read:
    case Write:
    case MCast:
      if (MsgID(ep) > GMaxMsgID)
	GMaxMsgID = MsgID(ep);
      break;
    }

  if (GExecGraph[pid] == NULL) {
    GExecGraph[pid] = ep;
    NextEvent(ep) = NULL;
    PrevEvent(ep) = NULL;
    GTailGraph[pid] = ep;
  }
  else {
    NextEvent(ep) = NULL;
    PrevEvent(ep) = GTailGraph[pid];
    NextEvent(GTailGraph[pid]) = ep;
    GTailGraph[pid] = ep;
  }
  return;
}

/*
 ***************************************************************
 * FUNCTION:    free_exec_graph()                              *
 * DESCRIPTION: deallocates all data structures related to the *
 *              program execution graph.                       *
 *              o GFirstEventPtr                               *
 *              o GExecGraph                                   *
 *              o GTailGraph                                   *
 ***************************************************************
*/

void free_exec_graph (int max_pid)
{
  int pid;
  EVENT *p, *q;
  for (pid=0; pid<max_pid; pid++) {
    if (GFirstEventPtr[pid] == NULL)
      continue;
    else {
      p = GFirstEventPtr[pid];
      while (p) {
	q = p->next;
	free(p);
	p = q;
      }
      GExecGraph[pid] = NULL;
      GTailGraph[pid] = NULL;
      GFirstEventPtr[pid] = NULL;
    }
  }
  free(GExecGraph);
  free(GTailGraph);
  free(GFirstEventPtr);
  GExecGraph = GTailGraph = GFirstEventPtr = NULL;
  return;
}

void FreeEventQ (int num_procs)
{
  int pid;
  for (pid=0; pid<num_procs; pid++)
    free (GEvntQ[pid]);
  free (GEvntQ);
  return;
}

void FreeEvntStack (int num_procs)
{
  int pid;
  for (pid=0; pid<num_procs; pid++)
    free (GEvntStack[pid]);
  free (GEvntStack);
  return;
}


     
EVENT_PTR_LIST **AllocateEventQ (int num_procs)
{
  int pid;
  EVENT_PTR_LIST **pepl =
    (EVENT_PTR_LIST **) calloc(num_procs, sizeof(EVENT_PTR_LIST **));

  for (pid=0; pid<num_procs; pid++)
    pepl[pid] = 0;
  return pepl;
}

char **AllocateCharBuf (int level_cnt)
{
  int lvl;
  char **buf = (char **) calloc (level_cnt, sizeof (char **));

  for (lvl=0; lvl<level_cnt; lvl++)
    {
      buf[lvl] = StrAlloc (1000);
      buf[lvl][0] = '\0';
    }
  return buf;
}

EVENT **AllocateExecGraph (int num_procs)
{
  int pid;
  EVENT  **pep = 
    (EVENT **) calloc(num_procs, sizeof(EVENT**));
  
  for (pid=0; pid<num_procs; pid++)
    pep[pid] = NULL;

  return pep;
}

      

  
/*
 ******************************************************************
 * FUNCTIONS:     Related to the hash table                       *
 *                o add_hash_table_entry()                        *
 *                o allocate_hash_tabel()                         *
 *                o build_hash_table()                            *
 * DESCRIPTION:   Each event pointer is kept in a hash table. So  *
 *                that finding out matching Reads for Write and   *
 *                Multicast can be easy.                          *
 * NOTES:         Another way may be foregoing this, and check    *
 *                each process to find matching reads. Memory is  *
 *                at premium in our application.                  *
 ******************************************************************
*/

void add_hash_table_entry (EVENT *ep, HASH_TABLE *ht)
{
  int pos = HashEventFunc (ep);
  
  AbortIf ((pos<0 || pos >= ht->size), "Invalid hash value.\n");

  if ( (ht->buckets)[pos] != NULL)
    {
      EVENT_PTR_LIST *epl = ht->buckets[pos];

      while (epl->next)
	epl = epl->next;
      
      epl->next = allocate_event_ptr_list (ep);
    }
  else
    (ht->buckets)[pos] = allocate_event_ptr_list (ep);
  return;
}


HASH_TABLE *allocate_hash_table (int size)
{
  int i;
  HASH_TABLE *ht = (HASH_TABLE *) malloc (sizeof (HASH_TABLE));

  ht->size = size;
  ht->buckets = 
    (EVENT_PTR_LIST **) calloc (size, sizeof(EVENT_PTR_LIST **));
  
  for (i=0; i<size; i++)
    (ht->buckets)[i] = NULL;
  return ht;
}
			       
void build_hash_table(HASH_TABLE *ht)
{
  int pid;
  EVENT *ep;
  
  for (pid=0; pid<total_procs; pid++) {
    for (ep = GExecGraph[pid]; ep != NULL; ep = NextEvent(ep)) {
      switch (EnumType(ep)) {
      case Read:
      case Write:
      case MCast:
	add_hash_table_entry(ep, ht);
	break;
      }
    }
  }
  return;
}




int ReadHeader (FILE *fp)
{
  int version, num_procs;

  fscanf(fp, "%d %d\n", &version, &num_procs);
  
  if (version != FORMAT_VERSION)
    return -1;
  else
    return num_procs;
}

EVENT_TYPE GetEvEnumType (char *ev_type)
{
  if (!StrCmp(ev_type, "R"))
    return Read;
  if (!StrCmp(ev_type, "W"))
    return Write;
  if (!StrCmp(ev_type, "M"))
    return MCast;
  if (!StrCmp(ev_type, "P"))
    return Phase;

  return Unknown;
}
  

/*
 *********************************************************************
 * FUNCTION:      get_event_desc_object()                            *
 * DESCRIPTION:   Given the name of an event returns the description *
 *                object associated with it, if any                  *
 *********************************************************************
*/

EVENT_DESC_OBJ *get_event_desc_object (char *event_name)
{
  int size = get_size_of_event_desc_list (GEventDescList);
  int i;
  
  for (i=0; i<size; i++) {
    EVENT_DESC_OBJ *edo = 
      get_element_from_event_desc_list (GEventDescList, i);
    if (!StrCmp(edo->name, event_name))
      return edo;
  }
  return NULL;
}

/*
 ********************************************************************
 * FUNCTION:        read_additional_attributes()                    *
 * DESCRIPTION:     Used for reading state information (mostly)     *
 *                  User needs to declare the additional attributes *
 *                  and the events they are associated with in the  *
 *                  init file.                                      *
 * DESIGN:          + Get the event description from GEventIndexList*
 ********************************************************************
*/

void read_additional_attributes (FILE *fp, EVENT *ep)
{
  int match = 0;
  int value;
  EVENT_DESC_OBJ *edo = get_event_desc_object (EventType(ep));

  /* error check */
  if (edo == NULL) {
    printf("Event %s has only primitive attributes\n", EventType(ep));
    printf("Use <add field> command to add new attributes\n");
    printf("Ignoring extra fields\n");
    while (1) {
      char c = fgetc(fp);
      if (c == '\n')
	return;
    }
  }

  /* If the event has no tag, then we read; else we see whether tags match */
  if (edo->utag == NULL) 
    match = 1;
  else if (!StrCmp(edo->utag, ep->event_tag))
    match = 1;
  

  if (match) {
    int i;
    int size = get_size_of_list ((LINKED_LIST *) edo->eal);
    for (i=0; i<size; i++) {
      EVENT_ATTR_OBJ *eao = 
	get_element_from_list ((LINKED_LIST *) edo->eal, i);
      ATTRIBUTE *attr = allocate_attribute (eao->name);
      LINKED_LIST *list = NULL;
      VAL_OBJ *vo;
      int ival;
      attr->next = ep->attr;
      ep->attr = attr;

      /* currently we read only integers for both int and float */
      /* kept because floats are space consuming */

      if (!StrCmp(eao->atype, "int"))
	fscanf(fp, "%d", &ival);
      else if (!StrCmp(eao->atype, "float"))
	fscanf(fp, "%d", &ival);
      
      vo = allocate_value_object (ival);
      add_node_to_list (&list, (void *) vo);
      attr->val_list = list;
    }
  }
  else { /* match == 0 */
    printf("Event type does not match\n");
  }
  return;
}


/*
 ******************************************************************
 * FUNCTION:      get_next_event_from_trace_file()                *
 * DESCRIPTION:   Obvious                         
 ******************************************************************
*/

EVENT *get_next_event_from_trace_file (FILE *fp)
{
  int is_there_more;
  EVENT *ep = allocate_event();
  int r = fscanf( fp, "%d %d %s %s %d %d %d", &PID(ep), &OID(ep),
		 &EventType(ep), &EventTag(ep), &(ep->event_size),
		 &(ep->bp_flag), &is_there_more);
  if (r == -1) {
    free(ep);
    return NULL;
  }

  if (is_there_more)
    read_additional_attributes (fp, ep);


  if (EnumType(ep)==Phase)
    Active(ep) = 0;


  if (StrCmp(EventTag(ep), "none")) {  /* if tag != none */
    sprintf(ep->event_name, "%s_%s", EventType(ep), EventTag(ep));
  }
  else
    sprintf(ep->event_name, "%s", EventType(ep));

  return ep;
}


  

void MyLoadTraceFile (char *LPhaseName, int LPhaseIter, char *RPhaseName,
		      int RPhaseIter)
{
  EVENT *ep;    /* EventCursor */
  boolean begin[MAX_PROCS], end[MAX_PROCS]; /* if begin[i] is TRUE_, then
					       load PID "i" events unless
					       end[i] is TRUE_ */
  int pid, phase_id;
  boolean b = FALSE_;

  if (!StrCmp("START", LPhaseName))
    b = TRUE_;


  for (pid=0; pid<total_procs; pid++)
    {
      begin[pid] = b;
      end[pid]   = FALSE_;
    }



  
  /**********************************************************************
   * Read the Event_Set members till you exhaust the file. Very coarse- *
   * grained filtering is in effect, as specified by the left and the   *
   * right cursors. Only events between these two cursors will be load- *
   * ed using "InsertEvent()".                                          *
   *********************************************************************/

  while (ep = get_next_event_from_trace_file (GTraceFile))
    {
      int pid = PID(ep);
      
      if (EnumType(ep) == Phase)
	{
	  AbortIf (!StrCmp(PhaseName(ep), "START"),
		       "Phase Name 'START' disallowed in tracefile");
	  AbortIf (!StrCmp(PhaseName(ep), "END"),
		       "Phase name 'END' disallowed in tracefile");
	  
	  /**************************************************
	   * Check whether begin[pid] needs to be turned ON *
	   *************************************************/

	  if ( (begin[pid] == FALSE_) && !StrCmp(PhaseName(ep), LPhaseName) &&
	      (PhaseIteration(ep) == LPhaseIter))
	    begin[pid] = TRUE_;

	  
	  /************************************************
	   * Check whether end[pid] needs to be turned ON *
	   ************************************************/
	      
	  if ((end[pid] == FALSE_) && !StrCmp(PhaseName(ep), RPhaseName) &&
	      (PhaseIteration(ep) == RPhaseIter))
	    end[pid] = TRUE_;
	}
      
      /*******************************************************
       * Acceptable events needs to be inserted in the graph *
       *******************************************************/

      if (begin[pid] && !end[pid])
	insert_event_in_exec_graph (ep);
      else
	free(ep);
    }

  return;
}


/**************************************************************************
 * This function is called when a "load" command is encountered. It opens *
 * the file, reads the header, and calls "MyLoadTraceFile()" to read the  *
 * events in, closes the file, and then connects up Reads and Writes by   *
 * building a hash table of MsgIDs. The actual WriteSet connections are   *
 * evaluated lazily by the function "ComputeWriteSet()".                  *
 *************************************************************************/

boolean LoadTraceFile (char *FileName, char *LPhaseName, int LPhaseIter,
		       char *RPhaseName, int RPhaseIter)
{
  int pid;
  char filename[80];
  sprintf(filename, "%s.btrace", FileName);


  if (GTraceFile)
    fclose (GTraceFile);

  free_exec_graph (total_procs);
/*  free_event_q (GNumProcs + STACK_SIZE);
    free_event_stack (STACK_SIZE); */

/*
  if (GTraceFile)
    {
      fclose (GTraceFile);
      FreeExecGraph(total_procs);
      FreeEventQ(total_procs + STACK_SIZE);
      FreeEventStack(STACK_SIZE);
    }
*/
  sprintf(CurrentFileName, "%s", filename);
  if ((GTraceFile = fopen(filename, "rb")) == 0) {
      if ((GTraceFile = fopen(FileName, "rb")) == 0) {
	printf("Error: can't open file %s\n", filename);
	return FALSE_;
      }
    }

  total_procs = ReadHeader (GTraceFile);
  set_total_procs(total_procs);

  if (total_procs == -1) {
    printf("Error: found wrong version number in file.\n");
    printf("Expecting version # %d\n", FORMAT_VERSION);
    fclose(GTraceFile);
    return FALSE_;
  }

  if (total_procs == -2) {
    printf("Error: poblem reading data file.\n");
    fclose(GTraceFile);
    return FALSE_;  
  }

  /* There are somthings we cannot handle */
  if ((total_procs <= 0) || (total_procs > 10000)) {
    printf("Error: It is a folly to use too many processors\n");
    printf("       It is a folly to use too few (especially < 0)\n");
    fclose(GTraceFile);
    return FALSE_;
  }

  GFirstEventPtr = AllocateExecGraph(total_procs);
  GExecGraph = AllocateExecGraph(total_procs);
  GTailGraph = AllocateExecGraph(total_procs);


  StackNum = total_procs;
  GEvntQ = AllocateEventQ (total_procs + STACK_SIZE);
  GEvntStack = AllocateEventQ (STACK_SIZE);
/*  GSuccTchBuf = AllocateCharBuf (STACK_SIZE); */
  MyLoadTraceFile (LPhaseName, LPhaseIter, RPhaseName, RPhaseIter);
  fclose (GTraceFile);

  for (pid=0; pid<total_procs; pid++) 
    GFirstEventPtr[pid] = GExecGraph[pid];


  GDumpFile = fopen ("mfile", "w+");

  /* Process the trace to ensure it is compatible with Ariadne */
  process_trace_for_ariadne();


  /* Now build the hash table */
  GEventHashTab = allocate_hash_table(119771); 
  build_hash_table(GEventHashTab);
  

  return TRUE_;
}


 
void store_trace_file ()
{
  EVENT *ep;
  int pid;

  FILE *fp = fopen("ave.trace.bpoint", "w");
  if (fp == NULL)
    printf("could not open file ave.trace.bpoint\n");
  fprintf(fp,"1 %d\n",total_procs);
  for (pid=0; pid<total_procs; pid++) {
    ep = get_first_event_on_proc(pid);
    while (ep) {
      fprintf(fp, "%d %d %s %s %d %d 0\n",
	      PID(ep), OID(ep), EventType(ep), EventTag(ep), ep->event_size,
	      get_event_breakpoint_flag(ep));
      ep = ep->next;
	/* Sameer : note 0 added as iflag to keep compatibility with traces */
    }
  }
  fclose(fp);
}
  
