/*
 *************************************************************************
 * FILE:         scan.c                                                  *
 * PROGRAMMER:   Joydip Kundu                                            *
 * ORGANIZATION: University of Massachusetts, Amherst                    *
 *               University of Oregon, Eugene                            *
 * DESCRIPTION:  Functions related to checking logical time assertions   *
 *               over abstract events: precedes and parallels            *
 * FUNCTIONS:    + keycode()                                             *
 *               + id_find_insert()                                      *
 *               + id_find()                                             *
 *               + append()                                              *
 *               + default_scan_init()                                   *
 *               + scaninit()                                            *
 *               + ave_scan_init()                                       *
 *               + process_ident()                                       *
 *               + process_marker_ident()                                *
 *               + process_numeric()                                     *
 *               + process_comment()                                     *
 *               + escape()                                              *
 *               + yylex()                                               *
 *               + get_size_of_wrt_list                                  *
 *               + add_reserved_words()                                  *
 *               + get_utag_from_name()                                  *
 *               + allocate_event_desc_object()                          *
 *               + allocate_event_attribute_object()                     *
 *               + get_name_without_utag()                               *
 *               + add_node_to_event_attribute_list()                    * 
 *               + add_node_to_event_desc_list()                         *
 *               + add_to_event_desc()                                   *
 *               + add_additional_fields()                               *
 *               + get_element_from_event_desc_list()                    *
 *               + get_size_of_event_desc_list()                         *
 * NOTES:                                                                *
 *************************************************************************
*/


#include "std.h"
#include <stdio.h>
#include "xstdio.h"

#include "structs.h"
#include "parse.tab.h"
#include "default.h"
#include "strs.h"
#include "global.h"
#include "decl.h"
#include "plot1.h"
#include "canvas.h"
#include "list.h"


/*
 ************************************************************************
 * Include file dependencies:                                           *
 *   structs.h :       Data Structures                                  *
 *   parse.tab.h:      Tokens for keywords                              *
 *   default.h:        Default values used in default_scan_init()       *
 *   strs.h:           StrLen(), StrCmp()                               *
 *   global.h:         GInputBuf, GInputBufCnt                          *
 *   plot1.h:          Variables in default_scan_init()                 *
 *   decl.h:           allocate_decl()                                  *
 *   list.h:                                                            *
 ************************************************************************
*/             


# define EOF_ 0  /* special end-of-file token (0 value required by bison) */

/*
  Keyword table and lookup/extract routine
*/
typedef struct keyword {
  char *symbol;
  int   code;
} KEYWORD;
static KEYWORD keytab [] = {
  "abs",        ABS_,
  "add",        ADD_,
  "after",      AFTER_,
  "and",        AND_,  
  "before",     BEFORE_,
  "begin",      BEGIN_,
  "break",      BREAK_,
  "by",         BY_,
  "ch",         CHAIN_,
  "child",      CHILD_,
  "compute",    COMPUTE_,
  "const",      CONST_,
  "count",      COUNT_,
  "create",     CREATE_,
  "cross",      CROSS_,
  "disjoint",   DISJOINT_,
  "distinct",   DISTINCT_,
  "dump",       DUMP_,
  "end",        END_,
  "eq",         EQUAL_,
  "etype",      EVENT_,
  "exec",       EXEC_,
  "explain",    EXPLAIN_,
  "fif",        FIF_,
  "first",      FIRST_,
  "float",      FLOAT_,
  "for",        FOR_,
  "foreach",    FOREACH_,
  "gec",        GEC_,
  "gtc",        GTC_,
  "imap",       IMAP_,
  "in",         IN_,
  "info",       INFO_,
  "int",        INTEGER_,
  "last",       LAST_,
  "leaf",       LEAF_,
  "lec",        LEC_,
  "left",       LEFT_,
  "load",       LOAD_,
  "ltc",        LTC_,
  "marker",     MARKER_,
  "match",      MATCH_,
  "mlc",        MLC_,
  "mrc",        MRC_,
  "onall",      ONALL_,
  "onsome",     ONSOME_,
  "option",     OPTION_,
  "or",         OR_,
  "owners",     OWNERS_,
  "parent",     PARENT_,
  "part",       PART_,
  "pch",        PCHAIN_,
  "print",      PRINT_,
  "prod",       PRODUCT_,
  "quit",       QUIT_,
  "relation",   RELATION_,
  "remainder",  REMAINDER_,
  "remove",     REMOVE_,
  "report",     REPORT_,
  "reset",      RESET_,
  "restore",    RESTORE_,
  "right",      RIGHT_,
  "section",    SECTION_,
  "set",        SET_,
  "show",       SHOW_,
  "sibling",    SIBLING_,
  "sif",        SIF_,
  "sizeof",     SIZE_,
  "source",     SOURCE_,
  "subset",     SUBSET_,
  "sum",        SUM_,
  "superset",   SUPERSET_,
  "tch",        TCHAIN_,
  "to",         TO_,
  "union",      UNION_,
  "utag",       UTAG_,
  "view",       VIEW_,
  "with",       WITH_,
  "wrt",        WRT_
  };

static int nkeys = (sizeof keytab) / (sizeof (KEYWORD));


/*
 ********************************************************************
 * FUNCTION:      keycode()                                         *
 * DESCRIPTION:   Does binary search of keyword table and returns   *
 *                -1 if not found, otherwise the token number.      *
 ********************************************************************
*/

int keycode (char *s)
{
  int lo = 0;
  int hi = nkeys - 1;
  int mid, cmp;

  while (lo <= hi) {
    mid = (lo + hi) >> 1;
    cmp = StrCmp (s, keytab[mid].symbol);
    if (cmp == 0)
      return keytab[mid].code;
    else if (cmp < 0)
      hi = mid - 1;
    else
      lo = mid + 1;
  }
  return -1;
}

/*  reserved words, fields and tags table and size */

static char **reswords;
static int nreswords;
static char **resfields;
static int nresfields;
static char  **restags;
static int nrestags;


# define IDTAB_SIZE 119
static IDENT *(idtab[IDTAB_SIZE]);



/*
 *********************************************************************
 * FUNCTIONS:     identifier lookup                                  *
 *                + id_find_insert(): if identifier already in       *
 *                table, returns the pointer; else insert identifier *
 *                in table, and return pointer.                      *
 *                + id_find() : if identifier in table, returns      *
 *                pointer; else returns NULL                         *
 *********************************************************************
*/

IDENT *id_find_insert (char *name)
{
  IDENT **p;
  char *q;
  int h;

  /* calculate hash code */
  for (h = 0, q = name; *q; h += *q++);

  /* probe and find matching entry, or leave p pointing to slot
     at which to chain a new IDENT block */
  for (p = &idtab[h % IDTAB_SIZE]; *p; p = &((*p)->id_next))
    if (StrCmp (name, (*p)->id_chars) == 0)
      return *p;

  /* allocate and fill in new IDENT block */
  *p = (IDENT *) malloc (sizeof (IDENT));
  (*p)->id_next = 0;
  (*p)->id_reserved = 0;
  (*p)->id_decl = (DECL *) NULL;
  (*p)->id_chars = StrAlloc (strlen (name) + 1);
  strcpy ((*p)->id_chars, name);
  return *p;
}

IDENT *id_find (char *name)
/* identifier lookup */
{
  IDENT **p;
  char *q;
  int h;

  /* calculate hash code */
  for (h = 0, q = name; *q; h += *q++);

  /* probe and find matching entry, or leave p pointing to slot
     at which to chain a new IDENT block */
  for (p = &idtab[h % IDTAB_SIZE]; *p; p = &((*p)->id_next))
    if (StrCmp (name, (*p)->id_chars) == 0)
      return *p;
  
  /* We did not find the id. Return Null */
  *p = NULL;
  return *p;
}



/**************************************************************************
  tokbuf is a pointer to a buffer for token characters
  tokbufsize is its size, which may be increased if the buffer overflows
  tokbufcnt is the number of characters currently in the buffer
 **************************************************************************/

static char *tokbuf;
static int   tokbufsize = 0;
static int   tokbufcnt;
# define INITBUFSIZE 100

/*
  current holds the result of the most recent fgetc call; it is an int
  so that we can represent EOF, too.

  The approach is as follows: current always holds the character *about to
  be processed* (the lookahead); it may be EOF or a real character; advance
  moves to the next character and, for convenience, returns it, too (and is
  actually defined as a macro because it is so simple).

  yylineno (a global) keeps track of the line number for error messages
  We use eol for errors detected when end-of-line is reached, so the
  error message will not print the following line number
*/


static int current = 0;
int        yylineno = 1;
static int eol = 0;


#define advance() {(current = getc (stdin)); if (eol) {++yylineno; eol=0;} if (current == '\n') eol=1;} 


/*
#define advance() {(current = GInputBuf[GInputBufCnt++]); if (eol) {++yylineno; eol=0;} if (current == '\n') eol = 1;}
*/

/* Adds a character to the token buffer, growing the buffer with realloc
   if the buffer is full (we never shrink the buffer) */

void append (char c)
{
  if (tokbufcnt == tokbufsize)
    tokbuf = (char *) realloc (tokbuf, tokbufsize += INITBUFSIZE);
  tokbuf[tokbufcnt++] = c;
}


/*
 ********************************************************************
 * FUNCTION:      default_scan_init()                               *
 * DESCRIPTION:   initializes global variables with default values  *
 * NOTES:         default.h holds default values                    *
 ********************************************************************
*/

void default_scan_init()
{
  g_max_color =  MAX_COLOR;
  g_color_collapsed = COLOR_COLLAPSED;
  g_color_error = COLOR_ERROR;
  g_color_normal = COLOR_NORMAL;
  g_scatter_plot_x_sep = SCATTER_PLOT_X_SEP;
  g_scatter_plot_y_sep = SCATTER_PLOT_Y_SEP;
  g_scatter_plot_max_x_dim = SCATTER_PLOT_MAX_COL;
  g_scatter_plot_max_y_dim = SCATTER_PLOT_MAX_ROW;
  g_line_plot_x_sep = LINE_PLOT_X_SEP;
  g_ctg_stat_x_sep = CTG_STAT_X_SEP;
  g_ctg_stat_y_sep = CTG_STAT_Y_SEP;
  g_ctg_stat_width = CTG_STAT_WIDTH;
  g_ctg_stat_height = CTG_STAT_HEIGHT;

}  

/* The (only!) initialization routine for the scanner. It:
   - clears the hash table chains (identifier tables)
   - inserts the reserved words into the identifier table
   - creates an initial token buffer
   - reads in the first character of the input */

void scaninit ()
{
  int i,j;
  char str[80];
  IDENT *id;
/*  FILE *fp = fopen ("ave.init.file", "rw");
    fscanf(fp, "%d\n", &nreswords); */
  nreswords = 4;
  reswords = (char **) calloc (nreswords, sizeof (char **));
  reswords[0] = StrAlloc (2);
  sprintf(reswords[0], "W");
  reswords[1] = StrAlloc (2);
  sprintf(reswords[1], "R");
  reswords[2] = StrAlloc (2);
  sprintf(reswords[2], "M");
  reswords[3] = StrAlloc (5);
  sprintf(reswords[3], "proc");

  nresfields = 4;
  resfields = (char **) calloc (nresfields, sizeof(char **));
  resfields[0] = StrAlloc (4);
  sprintf(resfields[0], "pid");
  resfields[1] = StrAlloc (4);
  sprintf(resfields[1], "oid");
  resfields[2] = StrAlloc (6);
  sprintf(resfields[2], "ltime");
  resfields[3] = StrAlloc (5);
  sprintf(resfields[3], "utag");

  nrestags = 0;

  for (i =  IDTAB_SIZE; --i >= 0;  idtab[i] = 0);
  for (i = 0; i < nreswords; ++i) {
  id = id_find_insert (reswords[i]);
  id->id_reserved = 1; 
  } 

  for (i = 0; i < nresfields; ++i) {
  id = id_find_insert (resfields[i]);
  id->id_reserved = 2; 
  } 
  
  for (j=0; j<nrestags; j++) {
    for (i=0; i<nreswords; i++) {
      char str[64];
      sprintf(str,"%s_%s",reswords[i], restags[j]);
      id = id_find_insert (str);
      id->id_reserved = 1;
    }
  }

  id = id_find_insert("select");
  id->id_decl = allocate_decl(DeclMatch);
  tokbuf = StrAlloc (INITBUFSIZE);
  default_scan_init();
}


void AveScanInit()
{
  advance();
}

/*
  The scanner is supported by a variety of process_xxx routines for handling
  the more complex token forms. Each has its own conventions, described in the
  comment preceding it.
*/



/* Routine to process identifier. It
   - expects current to hold the first letter of an identifier or keyword;
     (must be in A-Z or a-z)
   - leaves current as first character after the identifier/keyword
   - returns the token code ID_ or the token code for the keyword
   - sets yylval to point to the IDENT struct if the token is an identifier */

int process_ident ()
{
  /* keeps track of case to avoid checking keywords unless all upper case */
  int all_lower = ('a' <= current && current <= 'z');
  int keyc;

  for (;;) {
    append (current);
    advance ();
    if ('a' <= current && current <= 'z');
    else if (('A' <= current && current <= 'Z') ||
	     ('0' <= current && current <= '9') ||
	     current == '_')
      all_lower = 0;
    else
      break;
  }
  append (0);
  if (all_lower && ((keyc = keycode (tokbuf)) > 0))
    return keyc;
  yylval.id = id_find_insert (tokbuf);
  return ID_;
}


/* Routine to process match_identifier. It
   - expects current to hold '_'
   - leaves current as first character after the identifier/keyword
   - returns the token code _ID or the token code for the keyword
   - sets yylval to point to the IDENT struct if the token is an identifier */

int process_match_ident ()
{

  for (;;) {
    append (current);
    advance ();
    if (('a' <= current && current <= 'z') ||
	('A' <= current && current <= 'Z') ||
	('0' <= current && current <= '9') ||
	current == '_');
    else
      break;
  }
  append (0);
  yylval.id = id_find_insert (tokbuf);
  return _ID;
}



/* Routine to process numeric literal. It
   - expects current to be the first digit of a numeric literal, which must
     be in 0-9
   - leaves current as the character following the literal
   - detects and handles based literals by calling process_based
   - handles decimal literals itself (puts value in yylval, returns INT_)
   - returns 0 on error */

int process_numeric ()
{
  yylval.ival = current - '0';
  for (;;) {
    append (current);
    advance ();
    if ('0' <= current && current <= '9')
      yylval.ival = yylval.ival * 10 + (current - '0');
    else
      return INT_;
  }
}


/* Routine to process comment. It
   - expects current to be the * of (* that opened the comment
   - leaves current as the character after the *) that finished the comment
   - handles nested comments, unterminated comments */

void process_comment ()
{
  int comlevel = 1;
  int prev;

  advance ();
  for (;;) {
    prev = current;
    advance ();
    if (prev == '*' && current == ')') {
      if (--comlevel == 0) {
	advance ();
	return;
      }
    }
    else if (prev == '(' && current == '*')
      ++comlevel;
    else if (current == EOF) {
      yyerror ("Unterminated comment at eof.");
      return;
    }
  }
}


/* Helper routine for process_text; handles escape sequence. It
   - expects current to be the \ that starts an escape sequence
   - appends the resulting character (if any) to tokbuf
   - leaves current as the character after the escape sequence
*/

void escape ()
{
  advance ();
  if (current == EOF)
    yyerror ("End of file in text escape string.");
  else {
    char *inputs  = "\\\'\"fnrt";
    char *outputs = "\\\'\"\f\n\r\t";
    char *p = index (inputs, current);
    if (p)
      append (outputs[p - inputs]);
    else
      yyerror ("Illegal character in escape sequence; code %d",
	      current);
    advance ();
  }
}


/*
  To give a simpler switch statement in yylex, we build a table that specifies
  for each ASCII code what to do. The single character operators have their
  token codes and can be returned immediately. The illegal, white space, or
  more complex token are classed according to their handling, as detailed
  below. Note that these names *start* with _ whereas the token names *end*
  with _. Note also that all these entries are negative.
*/
# define _BAD	-1	/* illegal character */
# define _SKIP	-2	/* white space character */
# define _LET	-3	/* letters */
# define _DIGIT	-4	/* digits */
# define _LT	-5	/* < */
# define _GT	-6	/* > */
# define _LPAR	-7	/* ( */
# define _DOT	-8	/* . */
# define _COLON -9	/* : */
# define _DQUOT -10	/* " */
# define _PLUS  -11     /* + */
#define  _USC   -12     /* _ */
/*
  This table has 8 entries per line and has one entry for each possible
  byte value, 0 through 255.
*/
static int check[256] = {
  _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,
  _BAD,  _SKIP,  _SKIP,  _SKIP,  _SKIP,  _SKIP,   _BAD,   _BAD,
  _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,
  _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,
 _SKIP,    '!', _DQUOT,    '#',   _BAD,   _BAD,   _BAD,   _BAD,
 _LPAR,    ')',    '*',  _PLUS,    ',',    '-',   _DOT,    '/',
_DIGIT, _DIGIT, _DIGIT, _DIGIT, _DIGIT, _DIGIT, _DIGIT, _DIGIT,
_DIGIT, _DIGIT, _COLON,    ';',    _LT,    '=',    _GT,   _BAD,
   '@',   _LET,   _LET,   _LET,   _LET,   _LET,   _LET,   _LET,
  _LET,   _LET,   _LET,   _LET,   _LET,   _LET,   _LET,   _LET,
  _LET,   _LET,   _LET,   _LET,   _LET,   _LET,   _LET,   _LET,
  _LET,   _LET,   _LET,    '[',   _BAD,    ']',    '^',   _USC,
  _BAD,   _LET,   _LET,   _LET,   _LET,   _LET,   _LET,   _LET,
  _LET,   _LET,   _LET,   _LET,   _LET,   _LET,   _LET,   _LET,
  _LET,   _LET,   _LET,   _LET,   _LET,   _LET,   _LET,   _LET,
  _LET,   _LET,   _LET,    '{',    '|',    '}',   _BAD,   _BAD,
  _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,
  _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,
  _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,
  _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,
  _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,
  _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,
  _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,
  _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,
  _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,
  _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,
  _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,
  _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,
  _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,
  _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,
  _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,
  _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD,   _BAD };


/* The main scanning routine, yylex returns a token code and leaves any
   associated value in yylval. It expects that current is already set to the
   first character of the next token (or white space, etc.), and leaves it
   that way, too. */
int yylex ()
{
  int cls; /* character classification, from class table */

  tokbufcnt = 0;	/* makes the token buffer empty */

  for (;;) {
    if (current == EOF)
      return EOF_;
    if ((cls = check[current]) >= 0) {
      /* unambiguous single character operator token;
	 don't forget to advance! */
      advance ();
      return cls;
    }
    switch (cls) {	/* the harder cases */
    default:
    case _BAD:
      yyerror ("Bad input character; code = 0x%2x", current);
      /* fall through */
    case _SKIP:
      advance ();
      break;
    case _LET:
      return process_ident ();
    case _DIGIT:
      return process_numeric ();
    case _USC:
      return process_match_ident();
    case _LT:	/* < <= */
      advance ();
      if (current == '+')
	{
	  advance();
	  return PUSH_;
	}
      else if (current == '=')
	{
	  advance();
	  return LE_;
	}
      else
	return '<';
    case _GT:	/* > >= */
      advance ();
      if (current != '=')
	return '>';
      advance ();
      return GE_;
    case _PLUS:  /* +  +> */
      advance();
      if (current != '>')
	return '+';
      advance();
      return POP_;
    case _LPAR: /* ( (* */
      advance ();
      if (current != '*')
	return '(';
      process_comment ();
      break;
    case _DOT:	/* . .. */
      advance ();
      if (current != '.')
	return '.';
      advance ();
      return DOTDOT_;
    case _COLON: /* : := */
      advance ();
      if (current == '=') {
	advance();
	return COLON_EQ_;
      }
      else if (current == ':') {
	advance();
	return COLON_COLON_;
      }
      else
	return ':';
/*      if (current != '=')
	return ':';
      advance ();
      return COLON_EQ_; */
      
    }
  }
}

int get_size_of_wrt_list (WRT_LIST *wrt)
{
  int cnt = 0;
  if (wrt == NULL)
    return 0;
  while (wrt) {
    ++cnt;
    wrt = wrt->next;
  }
  return cnt;
}


/*
 ******************************************************************
 * FUNCTION:      add_reserved_words()                            *
 * DESCRIPTION:   called when an statement of the following forms *
 *                are encountered (AddStmt in parse.y)            *
 *                   add event synch, barrier;                    *
 *                   add utag  pivot, elim, shut;                 *
 *                   add float x, y to barrier;                   *
 * DESIGN:        + Get the number of reserved words to be added  *
 *                + 
 ******************************************************************
*/

void add_reserved_words (WRT_LIST *wrt, char *name)
{
  int i,j;
  int count = get_size_of_wrt_list (wrt);
  int size;
  char ***tmp;
  IDENT *id;

  if (!StrCmp(name, "event")) {
    tmp = &reswords;
    size = nreswords;
    nreswords += count;
    *tmp = (char **) realloc (reswords, (size+count)* sizeof(char **));
  }
  else if (!StrCmp(name, "utag")) {
    tmp = &restags;
    size = nrestags;
    nrestags += count;
    if (restags == NULL)
      *tmp = (char **) calloc ((size+count), sizeof(char **));
    else
      *tmp = (char **) realloc (restags, (size+count)* sizeof(char **));
  }
  else if (!StrCmp(name, "float") || (!StrCmp(name, "int"))) {
    tmp = &resfields;
    size = nresfields;
    nresfields += count;
    *tmp = (char **) realloc (resfields, (size+count)* sizeof(char **));
  }
  else {  /* error */
    printf("Expecting add event/float/int/utag list_of_identifiers\n");
    printf("%s should not be used\n", wrt->marker);
    return;
  }
  


  for (i=size; i<count+size; i++) {
    (*tmp)[i] = StrAlloc (strlen(wrt->marker)+1);
    sprintf((*tmp)[i], "%s", wrt->marker);
    id = id_find_insert (wrt->marker);
    id->id_reserved = 1;
    wrt = wrt->next;
  }

  for (j=0; j<nrestags; j++) {
    for (i=0; i<nreswords; i++) {
      char str[64];
      sprintf(str, "%s_%s", reswords[i], restags[j]);
      id = id_find_insert (str);
      id->id_reserved = 1;
    }
  }
  return;
}

/*
 ********************************************************************
 * FUNCTION:     get_utag_from_name()                               *
 * DESCRIPTION:  Given a name W_query, returns "query" if it is a   *
 *               user-defined tag. Returns Null if "query" is NOT a *
 *               user-defined tag.                                  *
 ********************************************************************
*/

char *get_utag_from_name (char *name)
{
  int i;

  for (i=0; i<nrestags; i++) {
    char *pos = strstr(name, restags[i]);
    if (pos != NULL) 
      return pos;
  }
  return NULL;
}





EVENT_DESC_LIST *GEventDescList = NULL;

/*
 *******************************************************************
 * FUNCTION:      allocate_event_desc_object()                     *
 * DESCRIPTION:   
 *******************************************************************
*/
			
EVENT_DESC_OBJ *allocate_event_desc_object (char *name, char *utag)
{
  EVENT_DESC_OBJ *this = 
    (EVENT_DESC_OBJ *) malloc (sizeof(EVENT_DESC_OBJ));
  if (name) {
    this->name = StrAlloc (strlen(name) + 1);
    sprintf(this->name, "%s", name);
  }

  if (utag) {
    this->utag = StrAlloc (strlen(utag)+1);
    sprintf(this->utag, "%s", utag);
  }

  this->eal = NULL;
  return this;
}


/*
 ***************************************************************
 * FUNCTION:      allocate_event_attribute_object()            *
 * DESCRIPTION:                                                *
 ***************************************************************
*/

EVENT_ATTR_OBJ *allocate_event_attribute_object(char *name, char *atype)
{
  EVENT_ATTR_OBJ *this = (EVENT_ATTR_OBJ *) (malloc (sizeof(EVENT_ATTR_OBJ)));
  if (name) {
    this->name = StrAlloc(strlen(name)+1);
    sprintf(this->name, "%s", name);
  }
  else
    this->name = NULL;
  
  if (atype){
    this->atype = StrAlloc(strlen(atype)+1);
    sprintf(this->atype, "%s", atype);
  }
  else
    this->atype = NULL;

  return this;
}






void add_attribute_to_event_desc_tab (char *event, char *tag)
{
}
  


/*
 *******************************************************************
 * FUNCTION:       get_name_without_utag()                         *
 * DESCRIPTION:    Given a name W_query, where "query" is a user-  *
 *                 defined tag, returns "W".                       *
 *******************************************************************
*/


char *get_name_without_utag (char *name, char *utag)
{
  int x = StrLen(name);
  int y = StrLen(utag);
  char *tmp;

  if (name == NULL)  /* sanity check: probably not needed */
    return NULL;

  if (utag == NULL) {   /* sanity check */
    tmp = StrAlloc (StrLen(name)+1);
    sprintf(tmp, "%s", name);
  }
  else {  /* name = tmp_utag */
    tmp = StrAlloc (x-y);
    strncpy(tmp, name, x-y-1);
    tmp[x-y] = '\0';
  }
  return tmp;
}
  

void add_node_to_event_attribute_list (EVENT_ATTR_LIST **eal,
				       EVENT_ATTR_OBJ *eao)
{
  add_node_to_list ((LINKED_LIST **)eal, (void *) eao);
}


void add_node_to_event_desc_list (EVENT_DESC_LIST **eil,
				   EVENT_DESC_OBJ *eio)
{
  add_node_to_list ((LINKED_LIST **) eil, (void *) eio);
}
      

void add_to_event_desc (char *name, WRT_LIST *alist, char *atype)
{
  char *etag = get_utag_from_name (name);
  char *ename = get_name_without_utag (name, etag);
  EVENT_DESC_OBJ *eio = allocate_event_desc_object(ename, etag);

  add_node_to_event_desc_list (&GEventDescList, eio);

  while (alist) {
    EVENT_ATTR_OBJ *eao 
      = allocate_event_attribute_object (alist->marker, atype);
    add_node_to_event_attribute_list (&(eio->eal), eao);
    alist = alist->next;
  }
  return;
}


void add_additional_fields (char *atype, WRT_LIST *alist, WRT_LIST *elist)
{
  int count = get_size_of_wrt_list (alist);
  char ***tmp = &resfields;
  int i;
  WRT_LIST *ap = alist;
  IDENT *id;

  *tmp = (char **) realloc (resfields, (nresfields+count)*sizeof(char **));

  for (i=nresfields; i<count+nresfields; i++) { 
    (*tmp)[i] = StrAlloc (strlen(ap->marker)+1);
    sprintf((*tmp)[i], "%s", ap->marker);
    id = id_find_insert (ap->marker);
    id->id_reserved = 2;
    ap = ap->next;
  }

  while (elist) {
    add_to_event_desc (elist->marker, alist, atype);
    elist = elist->next;
  }
}


EVENT_DESC_OBJ *get_element_from_event_desc_list (EVENT_DESC_LIST *edl, int i)
{
 EVENT_DESC_OBJ *edo = 
   (EVENT_DESC_OBJ *) get_element_from_list ((LINKED_LIST *) edl, i);
 return edo;
}

int get_size_of_event_desc_list (EVENT_DESC_LIST *edl)
{
  return get_size_of_list ((LINKED_LIST *) edl);
}

