#include <stdio.h>

#include "structs.h"
#include "decl.h"            
#include "expr.h"
#include "scan.h"
#include "set.h"


/*
 ************************************************************************
 * Include file dependencies:                                           *
 *   structs.h :       Data Structures                                  *
 *   decl.h:           create_decl_int()                                *
 *                     interpret_identifier_tobe_defined()              *
 *   expr.h:           allocate_value_object()                          *
 *                     add_node_to_int_list()                           *
 *                     evaluate_int_list()                              *
 *   scan.h:           id_find_insert()                                 *
 *   set.h:            get_core_set_section()                           *
 ************************************************************************
*/             




/*
 **********************************************************************
 * ****                Methods for CORE_SET_OBJ                  **** *
 *                                                                    *
 * allocate_core_set_object(): allocator                              *
 * free_core_set_object()    : destructor                             *
 * set_size_of_core_set()    : sets size parameter                    *
 * set_dim_of_core_set()     : sets dimension                         *
 * allocate_set_element()    : alloates elements for core set         *
 *                                                                    *
 **********************************************************************
*/

CORE_SET_OBJ *allocate_core_set_object()
{
  CORE_SET_OBJ *this = (CORE_SET_OBJ *) malloc(sizeof(CORE_SET_OBJ));
  CoreSetDim(this) = -1;
  CoreSetSize(this) = -1;
  CoreSetTail(this) = 0;
  CoreSetElem(this) = NULL;
  return this;
}

void free_core_set_object (CORE_SET_OBJ *this)
{
  if (this) {
    if (this->elem)
      free(this->elem);
    free(this);
  }
  return;
}

void set_size_of_core_set (CORE_SET_OBJ *set, int size)
{
  set->size = size;
  return;
}

void set_dim_of_core_set (CORE_SET_OBJ *set, int dim)
{
  set->dim = dim;
  return;
}


/*
 ******************************************************************
 * allocate_set_elem(): allocates appropriate size array of ints  *
 *                      initializes to -1                         *
 ******************************************************************
*/

int *allocate_set_element (int dim, int size)
{
  int *elem = (int *) malloc(dim*size*sizeof(int));
  int i;
  for (i=0; i<dim*size; i++)
    elem[i] = -1;
  return elem;
}

/*
 ********************************************************************
 * copy_core_set(): copies the "src" set to the "dest" set. Assumes *
 *     dest set is large enough to hold the additional elements con-*
 *     tained in the "src" set.                                     *
 ********************************************************************
*/

void copy_core_set (CORE_SET_OBJ *dest, CORE_SET_OBJ *src)
{
  int i;

  for (i=0; i<CoreSetDim(src)*CoreSetSize(src); i++) {
    CoreSetNthElem(dest, CoreSetTail(dest)) = CoreSetNthElem(src, i);
    CoreSetTail(dest)++;
  }
  
  return;
}


/*
 **********************************************************************
 * is_element_empty(): Checks whether an element contains -1 or not   *
 *      Note that this routine may need alteration if this convention *
 *      is changed.                                                   *
 **********************************************************************
*/

boolean is_element_empty (CORE_SET_OBJ *elem, int pos)
{
  int i;
  for (i=0; i<CoreSetDim(elem); i++) {
    if (CoreSetNthElem(elem, pos*CoreSetDim(elem)+i) == -1)
      return TRUE_;
  }
  return FALSE_;
}    

/*
 ***********************************************************************
 * is_element_equal(): Given two elements check whether they are equal *
 ***********************************************************************
*/

boolean is_element_equal (CORE_SET_OBJ *e1, CORE_SET_OBJ *e2)
{
  int i;
  if (CoreSetDim(e1) != CoreSetDim(e2))
    return FALSE_;
  for (i=0; i<CoreSetDim(e1); i++)
    if (CoreSetNthElem(e1,i) != CoreSetNthElem(e2,i))
      return FALSE_;
  return TRUE_;
}
  



/*
 **********************************************************************
 * get_core_set_elem(): Given a core set, and a position parameter n, *
 *      returns the n-th element in the set, NULL otherwise.          *
 **********************************************************************
*/

CORE_SET_OBJ *get_core_set_elem (CORE_SET_OBJ *set, int pos) 
{
  int i;
  CORE_SET_OBJ *this = allocate_core_set_object();
  CoreSetDim(this) = CoreSetDim(set);
  CoreSetSize(this) = 1;
  CoreSetElem(this) =  
    allocate_set_element (CoreSetDim(this), CoreSetSize(this));

  /* Check for invalid element request */
  if (pos > CoreSetSize(set))
    return NULL;

  for (i=0; i<CoreSetDim(this); i++) 
    CoreSetNthElem(this,i) = CoreSetNthElem(set, CoreSetDim(set)*pos+i);

  return this;
}


/*
 ********************************************************************
 * get_set_elem() :  Given a named set, and a position parameter n, *
 *                   returns the n-th element in the set.           *
 ********************************************************************
*/

CORE_SET_OBJ *get_set_elem (SET_OBJ *nso, int pos)
{
  return get_core_set_elem (CoreSet(nso), pos);
}




/*
 ****************************************************************************
 * get_elem _cross_product(): Given two core set elements returns the cross *
 *     product of the two. e.g., given (1,2) and (3) will return (1,2,3)    *
 ****************************************************************************
*/
  
CORE_SET_OBJ *get_elem_direct_product (CORE_SET_OBJ *e1, CORE_SET_OBJ *e2)
{
  int i;
  CORE_SET_OBJ *this = allocate_core_set_object();
  CoreSetDim(this) = CoreSetDim(e1) + CoreSetDim(e2);
  CoreSetSize(this) = 1;
  CoreSetElem(this) = (int *) malloc (CoreSetDim(this)*sizeof(int));
  
  for(i=0; i<CoreSetDim(e1); i++)
    CoreSetNthElem(this,i) = CoreSetNthElem(e1,i);

  for(i=CoreSetDim(e1); i<CoreSetDim(this); i++)
    CoreSetNthElem(this,i) = CoreSetNthElem(e2,i-CoreSetDim(e1));

  return this;
}


/*
 ************************************************************************
 * get_set_cross_product(): takes two named set, and returns the cross- *
 *    product of the two. Let A = {(1,1), (1,2), (1,3)} and B = {1,2,3} *
 *    then it returns  { (1,1,1), (1,1,2), (1,1,3), (1,2,1), (1,2,2),   *
 *    (1,2,3), (1,3,1), (1,3,2), (1,3,3) }                              *
 ************************************************************************
*/



CORE_SET_OBJ *get_core_set_direct_product (CORE_SET_OBJ *s1, CORE_SET_OBJ *s2)
{
  int i,j;
  CORE_SET_OBJ *this = allocate_core_set_object();
  CoreSetSize(this) = CoreSetSize(s1) * CoreSetSize(s2); 
  CoreSetDim(this)  = CoreSetDim(s1) + CoreSetDim(s2);
  CoreSetElem(this) = allocate_set_element (CoreSetDim(this), CoreSetSize(this));
  
  for (i=0; i<CoreSetSize(s1); i++){
    CORE_SET_OBJ *t1 = get_core_set_elem (s1,i);
    for (j=0; j<CoreSetSize(s2); j++) {
      CORE_SET_OBJ *t2 = get_core_set_elem(s2,j);
      CORE_SET_OBJ *t3 = get_elem_direct_product (t1, t2);
      copy_core_set(this, t3);
      free_core_set_object (t2);
      free_core_set_object (t3);
    }
    free_core_set_object (t1);
  }
  return this;
}
      


  
/*
 ********************************************************************
 * is_element_in_set(): Given an element and a set, returns TRUE_ if *
 *    the element is in the set.                                    *
 ********************************************************************
*/

boolean is_element_in_set (CORE_SET_OBJ *elem, CORE_SET_OBJ *set)
{
  int i;
  if (CoreSetDim(elem) != CoreSetDim(set))
    return FALSE_;

  if (CoreSetSize(set) <= 0)
    return FALSE_;

  for (i=0; i<CoreSetSize(set); i++) {
    CORE_SET_OBJ *tmp = get_core_set_elem (set, i);
    if (is_element_equal (elem, tmp) == TRUE_) {
      free_core_set_object(tmp);
      return TRUE_;
    }
    free_core_set_object(tmp);
  }
  return FALSE_;
}
      
  
/*
 ***********************************************************************
 * get_core_set_union(): Given two core sets returns the union after   *
 *      removing duplicates.                                           *
 ***********************************************************************
*/
  
CORE_SET_OBJ *get_core_set_union (CORE_SET_OBJ *s1, CORE_SET_OBJ *s2)
{
  int dup_cnt = 0;
  int i;
  CORE_SET_OBJ *this;

  if ((s1 == NULL) && (s2 != NULL)) {
    this = allocate_core_set_object();
    set_size_of_core_set (this, CoreSetSize(s2));
    set_dim_of_core_set (this, CoreSetDim(s2));
    CoreSetElem(this) =
      allocate_set_element (CoreSetDim(s2),CoreSetSize(s2));
    copy_core_set(this, s2);
    return this;
  }	  
  else if ((s2==NULL) && (s1 != NULL))  {
    this = allocate_core_set_object();
    set_size_of_core_set (this, CoreSetSize(s1));
    set_dim_of_core_set (this, CoreSetDim(s1));
    CoreSetElem(this) =
      allocate_set_element (CoreSetDim(s1),CoreSetSize(s1));
    copy_core_set(this, s1);
    return this;
  }	   
  else if ((s1==NULL) && (s2==NULL))
    return NULL;
  
  if (CoreSetDim(s1) != CoreSetDim(s2))
    return NULL;

  this = allocate_core_set_object();

  set_size_of_core_set (this, CoreSetSize(s1)+CoreSetSize(s2));
  set_dim_of_core_set (this, CoreSetDim(s1));

  CoreSetElem(this) =
    allocate_set_element (CoreSetDim(s1), 
			  CoreSetSize(s1)+CoreSetSize(s2));

  /* Copy the first core set immediately */
  copy_core_set (this, s1);
  
  /* We have to lookout for duplicates in the second case */
  for (i=0; i<CoreSetSize(s2); i++) {
    CORE_SET_OBJ *tmp = get_core_set_elem(s2,i);
    if (is_element_in_set(tmp, s1) == FALSE_)
      copy_core_set (this, tmp);
    else
      dup_cnt++;
    free_core_set_object(tmp);
  }
  
  set_size_of_core_set (this, CoreSetSize(this)-dup_cnt);
  return this;
}

  
/*
 ***********************************************************************
 * get_core_set_section(): Given two core sets returns the intersection*
 ***********************************************************************
*/


    
CORE_SET_OBJ *get_core_set_section (CORE_SET_OBJ *s1, CORE_SET_OBJ *s2)
{
  int i,j;  
  int cnt = 0;
  CORE_SET_OBJ *this = allocate_core_set_object();

  /* Check whether dimensions are compatible */
  if (CoreSetDim(s1) != CoreSetDim(s2)) {
    set_size_of_core_set (this, 0);
    return this;
  }


  set_dim_of_core_set (this, CoreSetDim(s1));
  set_size_of_core_set (this, CoreSetSize(s1));
  CoreSetElem(this) = allocate_set_element(CoreSetDim(this), CoreSetSize(this));

  for (i=0; i<CoreSetSize(s1); i++) {
    CORE_SET_OBJ *tmp = get_core_set_elem(s1,i);
    for (j=0; j<CoreSetSize(s2); j++) {
      if (is_element_in_set(tmp, s2) == TRUE_) {
	cnt++;
	copy_core_set (this, tmp);
	break;
      }
    }
    free_core_set_object(tmp);
  }
  CoreSetSize(this) = cnt;
  CoreSetDim(this)  = CoreSetDim(s1);
  return this;
}
    
    

    
CORE_SET_OBJ *make_core_set (LINKED_LIST *il)
{
  int i;
  int size = get_size_of_list(il);
  CORE_SET_OBJ *this = allocate_core_set_object();
  this->dim = 1;
  this->size = size;

  if (size == 0)
    return NULL;
  this->elem = allocate_set_element(1,size);
  
  for (i=0; i<size; i++) {
    this->elem[i] = get_int(il);
    il = il->next;
  }
  return this;
}




void set_index_expression (int index, int value)
{
  char name[20];
  IDENT *id;
  VAL_OBJ *vo;

  sprintf(name,"index%d",index);
  id = id_find_insert (name);
  if (id->id_decl == NULL) 
    create_decl_int(id);
  
  vo = (VAL_OBJ *) id->id_decl->decl_object;
  vo->value = value;
  return;
}

int get_index_expression (int index)
{
  char name[20];
  IDENT *id;
  VAL_OBJ *vo;
  sprintf(name,"index%d",index);
  id = id_find (name);
  vo = (VAL_OBJ *) id->id_decl->decl_object;
  return vo->value;
}
    
  

int get_process_identifier (SET_OBJ *so, int index)
{
  CORE_SET_OBJ *cso;
  int i = 0, cnt = 0, pid = -1;


  if (so == NULL) {
    printf("get_process_identifier: NULL pointer. PANIC\n");
    return -1;
  }
  cso = so->core;
  for (i=index*cso->dim; i<(index+1)*cso->dim; i++) {
    ++cnt;
    set_index_expression (cnt, cso->elem[i]);
  }
  
  if (so->map) {
    IDENT *iid = interpret_identifier_tobe_defined(so->map, 0);
    IMAP_OBJ *imo = (IMAP_OBJ *) iid->id_decl->decl_object;
    pid = get_int (evaluate_int_list (imo->stl, NULL));
  }
  else
    pid = get_index_expression (cnt);
  return pid;
}


INT_LINKED_LIST *get_int_list_from_set (CORE_SET_OBJ *cso)
{
  int i;
  int dim = cso->dim;
  int size = cso->size;
  INT_LINKED_LIST *ill = NULL;

  /* First element should be the dimension */
  VAL_OBJ *vo = allocate_value_object (dim);
  add_node_to_int_list (&ill, vo);

  for (i=0; i<size*dim; i++) {
    vo = allocate_value_object (cso->elem[i]);
    add_node_to_int_list (&ill, vo);
  }
  return ill;
}


INT_LINKED_LIST 
  *get_cross_product_in_list (INT_LINKED_LIST *left, INT_LINKED_LIST *right)
{
  INT_LINKED_LIST *locleft = left->next;
  INT_LINKED_LIST *locright = right->next;
  CORE_SET_OBJ *cleft = make_core_set (locleft);
  CORE_SET_OBJ *cright = make_core_set (locright);
  CORE_SET_OBJ *cso;
  INT_LINKED_LIST *result;
  cleft->dim = get_int(left);
  cright->dim = get_int(right);
  cso = get_core_set_direct_product (cleft, cright);
  result = get_int_list_from_set (cso);
  free_core_set_object (cleft);
  free_core_set_object (cright);
  free_core_set_object (cso);
  return result;
}

INT_LINKED_LIST 
  *get_intersection_in_list (INT_LINKED_LIST *left, INT_LINKED_LIST *right)
{
/* INT_LINKED_LIST *locleft = left->next;
  INT_LINKED_LIST *locright = right->next; */
  CORE_SET_OBJ *cleft = NULL;
  CORE_SET_OBJ *cright = NULL;
  CORE_SET_OBJ *cso;
  INT_LINKED_LIST *result;
  if (left == NULL || right == NULL)
    return NULL;
  if (left->next == NULL || right->next == NULL)
    return NULL;
  cleft = make_core_set (left->next);
  cright = make_core_set (right->next);
  
  cleft->dim = get_int(left);
  cright->dim = get_int(right);
  cso = get_core_set_section (cleft, cright);
  result = get_int_list_from_set (cso);
  free_core_set_object (cleft);
  free_core_set_object (cright);
  free_core_set_object (cso);
  return result;
}

INT_LINKED_LIST 
  *get_union_in_list (INT_LINKED_LIST *left, INT_LINKED_LIST *right)
{
  INT_LINKED_LIST *locleft = left->next;
  INT_LINKED_LIST *locright = right->next; 
  CORE_SET_OBJ *cleft = make_core_set (locleft);
  CORE_SET_OBJ *cright = make_core_set (locright);
  CORE_SET_OBJ *cso;
  INT_LINKED_LIST *result;
  cleft->dim = get_int(left);
  cright->dim = get_int(right);
  cso = get_core_set_union (cleft, cright);
  result = get_int_list_from_set (cso);
  free_core_set_object (cleft);
  free_core_set_object (cright);
  free_core_set_object (cso);
  return result;
}
					    

