/*
 *********************************************************************
 * FILE:          stmts.c                                            *
 * PROGRAMMER:    Joydip Kundu                                       *
 * ORGANIZATION:  University of Massachusetts, Amherst               *
 *                University of Oregon, Eugene                       *
 * FUNCTIONS:     o print_identifier()       Print Statement         *
 *                o set_global_constants()   Const Statement         *
 *                o restore_match_variable() Restore Statement       *
 *                o count_listed_events()    Count Statement         *
 *                o get_owner_process_set()  Owner Statement         *
 *                o compute_attribute_for_named_nodes()              *
 *                                           Compute Statement       *
 *                o partition_abstract_event_wrt_attribute()         *
 *                                           Partition Statement     *
 * NOTES:      
 *********************************************************************
*/


#include <stdio.h>
#include "structs.h"
#include "scan.h"
#include "symtab.h"
#include "abstract.h"
#include "plot.h"
#include "expr.h"
#include "global.h"
#include "ave_interf.h"
#include "strs.h"
#include "mtree.h"

/*
 *********************************************************************
 * FUNCTION:       print_identifier()                                *
 * DESCRIPTION:    handle statements of the form: print (x)          *
 * NOTES:          x is computed by assignment statement of the form *
 *                 int x = count (select wrt [W]);                   *
 *********************************************************************
*/

void print_identifier (IDENT *id)
{
  DECL *d = id->id_decl;
  
  if (d == NULL) {
    printf("%s has no current declaration\n", id->id_chars);
    return;
  }
}



/*
 ********************************************************************
 * FUNCTION:      set_global_constants()                            *
 * DESCRIPTION:   initializes global variables with user-supplied   *
 *                values                                            *
 * NOTES:         default.h holds default values                    *
 ********************************************************************
*/

void set_global_constants (IDENT *id, int value)
{
  char *name = id->id_chars;
  if (!StrCmp(name, "max_color"))
    g_max_color = value;
  else   if (!StrCmp(name, "g_color_collapsed"))
    g_color_collapsed = value;
  else   if (!StrCmp(name, "g_color_error"))
    g_color_error = value;
  else   if (!StrCmp(name, "g_color_normal"))
    g_color_normal = value;
  else if (!StrCmp(name, "scatter_plot_x_sep"))
    g_scatter_plot_x_sep = value;
  else if (!StrCmp(name, "scatter_plot_y_sep"))
    g_scatter_plot_y_sep = value;
  else if (!StrCmp(name, "scatter_plot_max_x_dim"))
    g_scatter_plot_max_x_dim = value;
  else if (!StrCmp(name, "scatter_plot_max_y_dim"))
    g_scatter_plot_max_y_dim = value;
  else if (!StrCmp(name, "line_plot_x_sep"))
    g_line_plot_x_sep = value;
  else if (!StrCmp(name, "ctg_stat_x_sep"))
    g_ctg_stat_x_sep = value;
  else if (!StrCmp(name, "ctg_stat_y_sep"))
    g_ctg_stat_y_sep = value;
  else if (!StrCmp(name, "ctg_stat_width"))
    g_ctg_stat_width = value;
  else if (!StrCmp(name, "ctg_stat_height"))
    g_ctg_stat_height = value;
  return;
}


/*
 ***********************************************************************
 * FUNCTION:     restore_match_variable()                              *
 * DESCRIPTION:  identifier should either be a match variable, or the  *
 *               "select" where the match variable is selected.        *
 * DESIGN:                                                             *
 * CALLER:
 * NOTES:
 ***********************************************************************
*/
  
void restore_match_variable (IDENT *id)
{
  ABSTRACT_EVENT *ae;

  if (id->id_decl->decl_kind != DeclMatch)
    {
      printf ("You can only restore a match variable\n");
      return;
    }
  ae = (ABSTRACT_EVENT *) id->id_decl->decl_object;
  id->id_decl->decl_object = (OBJECT *) NULL;
  restore_abstract_event (ae);
  free_abstract_event (ae);
  return;
}

/*
 ***********************************************************************
 * FUNCTION:
 * DESCRITION:
 * DESIGN:
 * CALLER:
 * NOTES:          + count(select wrt [W, R]); 
 *                 + count(ffts wrt [msg]);
 ***********************************************************************
*/


int count_listed_events (IDENT *id, WRT_LIST *wrt)
{
  ABSTRACT_EVENT *ae;
  int cnt = 0;

  if (!StrCmp(id->id_chars, "select")) 
    ae = GSelectAbstractEvent;
  else if (id->id_decl->decl_kind == DeclMatch) 
    ae = (ABSTRACT_EVENT *) id->id_decl->decl_object;
  else {
    printf("The argument to Count must be a match variable.\n");
    printf("%s is not a match variable\n", id->id_chars);
    return;
  }

  switch_off_unlisted_events (ae, wrt);
  count_switched_on_events (ae, &cnt);
  switch_on_all_events (ae);
  printf("%d events found in the match\n", cnt);
  return cnt;
}


CORE_SET_OBJ *get_owner_process_set (IDENT *id, WRT_LIST *wrt)
{
  CORE_SET_OBJ *cso = NULL;
  /* swicth off ulisted events */
  /* get_process_id_list */
  /* make set from intlist */
  /* return the core_set */
  return cso;
}



/*
 *****************************************************************************
 * FUNCTION:      compute_attribute_for_named_nodes()                        *
 * DESCRIPTION:   + called in the context of an instruction like:            *
 *                  with <mvar> foreach <dest> compute <attr> = <stl> ;      *
 *                + <mvar> can be "select" or a match variable. One cannot   *
 *                use the name of any other abstract events. The reason is,  *
 *                except for the "match variables" other names in the tree   *
 *                are definitions and hence are either of chain. p-chain or  *
 *                pt-chains.                                                 *
 *****************************************************************************
*/

void compute_attribute_for_named_nodes (IDENT *mvar, char *dest, 
		      char *attr, SYN_TREE_LIST *stl)
{
  /* obsessive initialization */
  ABSTRACT_EVENT *ae = NULL;
  ABSTRACT_EVENT_LIST *ael = NULL;
  ABSTRACT_EVENT_LIST *tmp = NULL;
  INT_LINKED_LIST *il = NULL;
  DECL *d = mvar->id_decl;
  int value = 0;
  IDENT *id = id_find_insert (attr);

  id->id_reserved = 2;

  /* Check whether "mvar" is actually a match variable or "select" */
  if (d->decl_kind != DeclMatch) {
    printf("WARNING: %s is not a match variable. Command ignored.\n",
	   mvar->id_chars);
    return;
  }
  ae = (ABSTRACT_EVENT *) mvar->id_decl->decl_object;

  /* GetPointers (ae, dest, &ael, 0); */
  ael = get_named_nodes (dest, ae, 0);
  tmp = ael;
  while(tmp) {
    ABSTRACT_EVENT *ae = get_element_from_abstract_list (tmp, 0);
    ATTRIBUTE *new = allocate_attribute (attr);
    il = evaluate_int_list (stl, ae);
    new->next = ae->attr;
    ae->attr = new;
    ae->attr->val_list = il;
    tmp = tmp->next;
  } 
}


/*
 *******************************************************************************
 * FUNCTION:      partition_abstract_event_wrt_attribute()                     *
 * DESCRIPTION:   with "context" foreach "dest" create part "new" wrt "aname"; *
 * 
 *******************************************************************************
*/

void partition_abstract_event_wrt_attribute
(ABSTRACT_EVENT *context, char *dest, char *new, char *aname)
{
  ABSTRACT_EVENT_LIST *dest_list = NULL;
  ABSTRACT_EVENT_LIST  *cand_list = NULL;
  ABSTRACT_EVENT_LIST  *dlt, *clt;
  int i, size;
  /* dest_list contains the list of abstract events which will get a new
     collapse node attached to it */

  /* GetPointers (context, dest, &dest_list, 0); */
  dest_list = get_named_nodes (dest, context, 0);


  /* For each abstract node, we need to form the partitions */
  dlt = dest_list;
  while (dlt) {
    int total = 0;
    int part = 0;
    COLLAPSE_NODE *cn = allocate_collapse_node();
    ABSTRACT_EVENT *ae = get_element_from_abstract_list (dlt, 0);
    cn->pae = ae;
    cn->par_id = cn->my_id = 0;
    cn->ave = allocate_ave_str ();
    cn->pattern = StrAlloc (StrLen (new) + 1);
    sprintf(cn->pattern, "%s", new);

    add_collapse_node_to_abstract_event (cn, ae);


    /* For each abstract event, identify the list of candidate abstract events */
    GetCandidatesForPartition (ae, aname, &cand_list, 0);  
    size = get_size_of_abstract_list (cand_list);
    for (i=0; i<size; i++) {
/*    for (clt = cand_list; clt != NULL; clt = clt->next) { */
      ABSTRACT_EVENT *tae = get_element_from_abstract_list (cand_list, i);
      COLLAPSE_NODE *cc  = cn->down;
      if (tae->error)
	break;
      if ((cc == NULL) || (in_collapse_child (tae, cc, aname) == FALSE_)) {
	INT_LINKED_LIST *v1 = get_attr_value (tae, aname);
	int x = get_int(v1);
	char str[64];
	sprintf(str,"(%s = %d)", aname, x);
	++part;
	add_partition_for_view (cn, tae, str);
      }
      ++total;
    }
    cn->total = total;
    cn->part  = part; 
    dlt = dlt->next;
  }
}


void add_marker_for_named_nodes (ABSTRACT_EVENT_LIST *root_ael, char *dest, 
		      SYN_TREE_LIST *dest_stl, char *marker)
{
  /* obsessive initialization */
  int root_cnt = get_size_of_abstract_list (root_ael);
  int len = StrLen(marker)+1;
  int cnt = 0;
  int i = 0;

  for (i=0; i<root_cnt; i++) {
    ABSTRACT_EVENT *ae = get_element_from_abstract_list (root_ael, i);
    ABSTRACT_EVENT_LIST *ael = 
      get_named_nodes_with_filter (dest, dest_stl, ae, 0);
    int dest_cnt = get_size_of_abstract_list (ael);
    int j = 0;

    for (j=0; j<dest_cnt; j++) {
      ABSTRACT_EVENT *tae = get_element_from_abstract_list (ael, j);
      free(tae->marker);
      tae->marker = StrAlloc(len);
      sprintf(tae->marker, "%s", marker);
    } 
    cnt += dest_cnt;
  }
  printf("Marker added to %d events\n", cnt);
}


void new_compute_attribute_for_named_nodes (ABSTRACT_EVENT_LIST *root_ael, 
     char *dest, SYN_TREE_LIST *dest_stl, char *attr_name, 
     SYN_TREE_LIST *attr_stl)
{
  /* obsessive initialization */
  ABSTRACT_EVENT *ae = NULL;
  ABSTRACT_EVENT_LIST *tmp = NULL;
  ABSTRACT_EVENT_LIST *ael = NULL;
  int len = get_size_of_abstract_list (root_ael);
  int i;

  for (i=0; i<len; i++) {
    ae   = get_element_from_abstract_list (root_ael, i);
    ael = get_named_nodes_with_filter (dest, dest_stl, ae, 0);
    tmp = ael;
    while(tmp) {
      ABSTRACT_EVENT *tae = get_element_from_abstract_list (tmp, 0);
      ATTRIBUTE *new_attr = allocate_attribute (attr_name);
      INT_LINKED_LIST *il = evaluate_int_list (attr_stl, tae);

      /* if il == NULL, it may be that we reference the parent */
      /* this will work for now but we need to add more tests */
      if (il == NULL && len > 1)
	il = evaluate_int_list (attr_stl, ae);
      new_attr->val_list = il;
      new_attr->next = tae->attr;
      tae->attr = new_attr;
      tmp = tmp->next;
    }
  }
}




