/*
 *************************************************************************
 * FILE:         support.c                                               *
 * PROGRAMMER:   Joydip Kundu                                            *
 * ORGANIZATION: University of Massachusetts, Amherst                    *
 *               University of Oregon, Eugene                            *
 * DESCRIPTION:  Support functions related to the match tree             *
 * FUNCTIONS:    + is_marker_in_list()                                   *
 *               + turn_off_abstract_events()                            *
 *               + turn_on_all_events()                                  *
 *               + count_switched_on_primitive_events()                  *
 *               + count_children_of_abstract_event()                    *
 *               + count_switched_on_events()                            *
 *               + count_listed_events()                                 *
 *               + get_owner_process_set()                               *
 *                                                                       *
 * NOTES:                                                                *
 *************************************************************************
*/


#include <stdio.h>
#include "structs.h"
#include "event.h"


/*
 ***************************************************************************
 * FUNCTION:       is_marker_in_list()                                     *
 * DESCRIPTION:    Given a list of identifiers "wrt", checks whether it    *
 *                 contains "marker".                                      *
 * DESIGN:                                                                 *
 * CALLER:         turn_off_abstract_event()                               *
 * NOTES:          Used to check, whether an abstract event in them match  *
 *                 tree should be considered for the current query or not. *
 ***************************************************************************
*/

boolean is_marker_in_list (char *marker, WRT_LIST *wrt)
{
  if (marker == NULL)
    return FALSE_;

  /* we are turning events off if we get a "FALSE_" response. We should
     not turn anything off if the wrt_list is empty.Thus ... */
  if (wrt == NULL)  
    return TRUE_;

  while (wrt) {
    if (!StrCmp(marker, wrt->marker)) 
      return TRUE_;
    wrt = wrt->next;
  }
  return FALSE_;
}

/*
 ***********************************************************************
 * FUNCTION:     switch_off_unlisted_events()                            *
 * DESCRIPTION:  Turns off abstract events that are not named in the   *
 *               list of events given in "wrt"                         *
 * DESIGN:
 * CALLER:                                                             *
 * NOTES:                                                              *
 ***********************************************************************
*/


boolean switch_off_unlisted_events (ABSTRACT_EVENT *ae, WRT_LIST *wrt)
{
  boolean r1, r2;
  r1 = TRUE_;

  if ((ae == NULL) || (wrt == NULL))     /* termination condition */
    return FALSE_;

  if ((is_marker_in_list (ae->marker, wrt) == FALSE_) &&
      (is_marker_in_list (ae->name, wrt) == FALSE_)) {
    r1 = switch_off_unlisted_events (ae->down, wrt);
    if (r1 == FALSE_)
      ae->status = OFF;
  }
  else
    ae->status = ON;
  
  r2 = switch_off_unlisted_events (ae->next, wrt);
/* old 
  return (r1 || r2);
 */
  if ((r1 == TRUE_) || (r2 == TRUE_))
          return TRUE_ ;   
  else
          return FALSE_ ;   
}

/*
 *************************************************************************
 * FUNCTION:       switch_on_all_events()                                  *
 * DESCRIPTION:    for each node in the match tree, sets ae->status = ON *
 * NOTES:          Needs to be called after evaluating expressions like  *
 *                 count (select wrt WrtList)                            *
 *************************************************************************
*/

void switch_on_all_events (ABSTRACT_EVENT *ae)
{
  if (ae == NULL)
    return;
  ae->status = ON;
  switch_on_all_events (ae->down);
  switch_on_all_events (ae->next);
}


void count_switched_on_primitive_events (ABSTRACT_EVENT *ae, int *cnt)
{
  if (ae == NULL)
    return;

  if (ae->status == OFF)
    return;
  
  if (ae->ep && (*cnt < PID(ae->ep)))
      (*cnt) = PID(ae->ep);
      
  count_switched_on_primitive_events (ae->down, cnt);
  count_switched_on_primitive_events (ae->next, cnt);
  return;
}


/*
 ***********************************************************************
 * FUNCTION:     count_children_of_abstract_event()                    *
 * DESCRIPTION:  returns the children of a node in the match tree      *
 * CALLER:                                                             *
 * NOTES:                                                              *
 ***********************************************************************
*/

int count_children_of_abstract_event (ABSTRACT_EVENT *ae)
{
  int cnt = 0;
  ABSTRACT_EVENT *tae = ae->down;
  
  while (tae) {
    ++cnt;
    tae = tae->next;
  }
  return cnt;
}

/*
 ***********************************************************************
 * FUNCTION:     count_switched_on_events()                            *
 * DESCRIPTION:  (primary) statement:                                  *
 * CALLER:
 * NOTES:
 ***********************************************************************
*/
void count_switched_on_events (ABSTRACT_EVENT *ae, int *cnt)
{
  if (ae == NULL)
    return;
  
  if (ae->status == ON)
    ++cnt;

  count_switched_on_events (ae->down, cnt);
  count_switched_on_events (ae->next, cnt);
}





void print_abstract_event (ABSTRACT_EVENT *ae)
{
  if (ae == NULL)
    return;

  printf("ave = %x, proc = %d, error = %d, cnt = %d, dummy = %d\n",
	 ae->ave, ae->proc, ae->error, ae->cnt, ae->dummy);
  printf("status = %d, orig = %s, name = %s\n",
	 ae->status, ae->orig, ae->name);
  if (ae->down)
    print_abstract_event (ae->down);
}

void print_tree_object (TREE_OBJ *to)
{
  if (to == NULL)
    return;

  if (to->name)
    printf("tree_name = %s, flag = %d\n", to->name, to->flag);
  if (to->next)
    print_tree_object(to->next);

  if (to->down)
    print_tree_object(to->down);
}
