/*
 *  ----------------------------------
 *  Darryl Brown
 *  University of Oregon pC++/Sage++
 *
 *  brkAccess.C
 *
 *  This program is a library that will interface with the parallel
 *  program to gain runtime data access and runtime control.
 *  
 *  ----------------------------------
 *  ----------------------------------
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "astrlib.h"
#include "typeApi.h"
#include "clServ.h"
#include "cargo.h"


/*
 * ---------------------------------------------
 * MACROS
 * ---------------------------------------------
 */

#define MAXVARS 100

/*
 * ---------------------------------------------
 * Globals 
 * ---------------------------------------------
 */

static int SocketID;
static int VarListIdx = 0;
static char *VarList[MAXVARS];

/*
 * ---------------------------------------------
 * Functions
 * ---------------------------------------------
 */

/*
 * ----------------------
 * int initServerConnection(int socketNum);
 *     initializes server connection given a socket number.
 * ----------------------
 */
int initServerConnection(int socketNum) {
  /*
   * socketNum - number of socket to open server connection
   *             from.
   */
  SocketID = TRNS_OpenServer(socketNum);
  return SocketID;
}


/*
 * ----------------------
 * int initClientConnection(int socketNum);
 *     initializes server connection given a socket number.
 * ----------------------
 */
int initClientConnection(int socketNum, char *host) {
  /*
   * socketNum - number of socket to open server connection
   *             from.
   */
  SocketID = TRNS_OpenClient(socketNum, host);
  return SocketID;
}

/*
 * ----------------------
 * int sendData(void *ms, int numbytest);
 *     send a package of data over the previously
 *     opened connection.
 * ----------------------
 */
int sendData(void *ms, int numbytes) {
  /*
   * ms - data to send.
   * numbytes - number of bytes in ms to send.
   */
  aPackage pkg;

  pkg.data = ms;
  pkg.length = numbytes;
  return CRG_SendPackage(SocketID, &pkg);
}


/*
 * ----------------------
 * int sendText(char *ms);
 *     sends textual data over the previously
 *     opened connection.
 * ----------------------
 */
int sendText(char *ms) {
  /* ms - textual message to send. */
  return sendData((void *)ms, strlen(ms)+1);
}

/*
 * ----------------------
 * CRG_package *receiveData();
 *     receives data package over previously opened
 *     connection.
 * ----------------------
 */
aPackage * receiveData() {
  return CRG_ReceivePackage(SocketID);
}


/*
 * ----------------------
 * void clearVars();
 *     clears the data structures that keep track
 *     of the variable instances.  This is necessary
 *     because at each breakpoint these instances
 *     may have changed, and so should be updated.
 * ----------------------
 */
static void clearVars() {
  int i;
  for (i = 0; i < VarListIdx; i++) {
    free((VarList[i]));
  }
  VarListIdx = 0;
}

/*
 * ----------------------
 * int pushVar(char *name);
 *     pushes a new variables onto a structure that
 *     keeps track of the current variable instantiations.
 * ----------------------
 */
static int pushVar(char *name) {
  /*  name - name of collection variable to push onto stack. */
  VarList[VarListIdx] = strsave(name);
  VarListIdx++;
  return   VarListIdx-1;
}


/*
 * ----------------------
 * int findIthCollection(char *varname);
 *     given a variable name, finds the FIRST occurence 
 *     of a collection with that name in the variable 
 *     list.  Note that since collections can be in different
 *     scopes, they could share the same name, and therefore 
 *     they would occur multiple times in the instantiation
 *     list.
 * ----------------------
 */
int findIthCollection(char *varname) {
  /* varname - variable name to find in variable list. */
  int i;
  for (i = 0; i < VarListIdx; i ++) {
    if (strsame(varname, VarList[i])) {
      return i;
    }
  }
  return -1;
}


/*
 * ----------------------
 * void printCurrentCollections();
 *     print current collection variable instantiations
 *     stored.
 * ----------------------
 */
void printCurrentCollections(){
  int i;
  for (i = 0; i < VarListIdx; i ++) {
    printf("(%d-->) %s\n", i, VarList[i]);
  }
}

/*
 * ----------------------
 * static void parseVar(char *line) 
 *     parse a line that may contain variable instantiation
 *     information.
 * ----------------------
*/
static void parseVar(char *line) {
  char *tmp, *vname;
  strtokStruct *tokStr;
  int cnt = 0;

  vname = NULL;
  tokStr = initstrtok(line);
  
  while (tmp = sstrtok(tokStr, " ")) {

    /* make sure this isn't garbage. */
    if (cnt == 1) {
      if (!strsame(tmp, "<")) {
	return;
      }
    }
    cnt++;
    vname = tmp;
  }
   /* make double sure this isn't garbage. */
  if (cnt < 4) {
    return;
  }
  if (vname && *vname) {
    pushVar(vname);
  }
  freestrtok(tokStr);
}
  

/*
 * ----------------------
 * int updateCurrentCollections()
 *     request and receive current collection variable
 *     instantiations from the parallel program.
 * ----------------------
 */
int updateCurrentCollections() {
  char retstr[1000], *line, newline[3];
  aPackage *pkg;
  strtokStruct *tokStr;

  sendText("getVars");
  pkg = receiveData();

  if (strsame("Program Exiting", (char *)pkg->data)) {
    printf("detected program is finished\n");
    sendText("done OK");
    return 0;
  }
  
  /* parse result */
  clearVars();
  sprintf(newline, "\n");
  strncpy(retstr, (char *) pkg->data, pkg->length - 1);
  tokStr = initstrtok(retstr);
  while (line = sstrtok(tokStr, newline)) {
    parseVar(line);
    line = "";
  }
  return 1;
}

/*
 * ----------------------
 * char * printCollections()
 *     prints a list of the collections
 * ----------------------
 */
char *printCollections() {
  char retstr[1000];
  int i;
  
  strcpy(retstr, "");
  for (i = 0;i < VarListIdx; i++) {
    strcat(retstr, VarList[i]);
    strcat(retstr, " ");
  }
  return strdup(retstr);
}


/*
 * ----------------------
 * void getTypes()
 *     request and receive type information from
 *     the executing parallel program.
 * ----------------------
 */
void getTypes() {
  char retstr[1000], *line, newline[3];
  aPackage *pkg;
  strtokStruct *tokStr;

  sendText("getTypes");
  pkg = receiveData();

  /* parse result */
  sprintf(newline, "\n");
  strncpy(retstr, (char *) pkg->data, pkg->length - 1);
  tokStr = initstrtok(retstr);
/*
  printf("first char %c, returnString %s\n", 
	 (char) (pkg->data), retstr);
*/
  while (line = sstrtok(tokStr, newline)) {
    TA_putType(line);
    line = "";
  }
}


/*
 * ----------------------
 * char * printTypes();
 *     print the types currently stored (could be none).
 * ----------------------
 */
char * printTypes() {
  char *pbuf, store[10000];
  int i;
  for (i = 0; i < TA_CompListIdx; i++) {
    pbuf = TA_writeComposite(TA_CompositeList[i]);
    strcat(store, pbuf);
    free (pbuf);
    strcat(store, "\n");
  }
  
  /* print the list of type descriptions */
  for (i = 0; i < TA_TypeListIdx; i++) {
    pbuf = TA_writeTypeDesc(TA_typeDescList[i]);
    strcat(store, pbuf);
    free(pbuf);
    strcat(store, "\n");
  }
  
  return strsave(store);
}



/*
 * ----------------------
 * void getSourceInfo()
 *     request and receive current source code line number
 *     information.
 * ----------------------
 */
char *getSourceInfo() {
  aPackage *pkg;

  sendText("getFuncInfo");
  pkg = receiveData();
  if (strsame("Program Exiting", (char *) pkg->data)) {
    printf("detected program is finished\n");
    sendText("done OK");
  }
  return strdup((char *) pkg->data);
}


/*
 * ----------------------
 * char *getVars()
 *     request and receive current source code line number
 *     information.
 * ----------------------
 */
char *getVars() {
  aPackage *pkg;

  sendText("getVars");
  pkg = receiveData();
  if (strsame("Program Exiting", (char *) pkg->data)) {
    printf("detected program is finished\n");
    sendText("done OK");
  }
  return strdup((char *) pkg->data);
}



/*
 * ----------------------
 * char * getUserDefFuncs()
 *     this func retrieves a list of all the user defined functions
 *     in the program.  "User defined" functions are those which the user
 *     specifies by prepending the string "UserDefined_" onto the 
 *     function name.  Note that this does not include the methods, 
 *     which can be list using the getUserMemFuncs routine following
 *     this routine.
 * ----------------------
 */
char * getUserDefFuncs() {
  aPackage *pkg;

  sendText("getUserFuncs");
  pkg = receiveData();
  return strdup((char *) pkg->data);
}



/*
 * ----------------------
 * char * getUserDefMethods()
 *     this func retrieves a list of all the user defined methods
 *     in the program.  "User defined" methods are those which the user
 *     specifies by prepending the string "UserDefined_" onto the 
 *     method name.  Note that this does not include the user defined
 *     functions which can be list using the getUserFuncs routine.
 * ----------------------
 */
char * getUserDefMethods() {
  aPackage *pkg;

  sendText("getUserMemFuncs");
  pkg = receiveData();
  return strdup((char *) pkg->data);
}



/*
 * ----------------------
 * void skipNBrkpnts(int numskip);
 *     execute program, skipping 'numskip' breakpoints.
 * ----------------------
 */
void skipNBrkpnts(int numskip) {
  /* numskip - number of breakpoints to skip. */
  char buf[30];
  sprintf(buf, "setContinue %d", numskip);
  sendText(buf);
  /* recieve blank answering packet. */
  receiveData();
}
  

/*
 * ----------------------
 * void nextBrkpoint();
 *     continue to the next breakpoint;
 * ----------------------
 */
void nextBrkpoint() {
  sendText("continue");
  /* recieve blank answering packet. */
  receiveData();
}


/*
 * ----------------------
 * void terminateProgram();
 *     terminate parallel program execution.
 * ----------------------
 */
void terminateProgram() {
  sendText("exit");
}




/*
 * ----------------------
 * void callAccessFuncByNum(int collnum, int elemNum, char *fname);
 *     this func calls one of the collection element access functions
 *     automatically generated by the BREEZY pre-compilation process.
 *     It calls the collection by unique number in the list of 
 *     instance variables.
 * ----------------------
 */
void callAccessFuncByNum(int collnum, int elemNumi, 
			 int elemNumj, char *fname) {
  /*
   * collNum - number in the collection variable list 'VarList'
   *            of collection to access.
   * elemNumi - index of the first dimension of the element to
   *            access.
   * elemNumj - index of the second dimension of the element to
   *            access.  IF THERE IS NO SECOND DIMENSION, THIS 
   *            PARAMETER MUST BE A NEGATIVE NUMBER.
   * fname    - the function name to access (i.e. the name of 
   *            the element class variable).
   */
  char buf[100];
  char *blank = "";

  if (!fname) {
    /*
     * no field name is specified
     * so give a blank for this specifier;
     */
    fname = blank;
  }
  if (elemNumj >= 0) { 
    /* 2-dimensional data request */
    sprintf(buf, "getCollElem %d %d %d %s",
	    collnum, elemNumi, elemNumj, fname);
  } else {
    /* 1-Dimensional data request */
    sprintf(buf, "getCollElem %d %d %s",
	    collnum, elemNumi, fname);
  }
  sendText(buf);
}


/*
 * ----------------------
 * void callAccessFuncByName(char *collName, int elemNum, char *fname);
 *     this func calls one of the collection element access functions
 *     automatically generated by the BREEZY pre-compilation process.
 *     It calls the FIRST occurence of a collection with that name in 
 *     the variable list.  Note that since collections can be in different
 *     scopes, they could share the same name, and therefore 
 *     they would occur multiple times in the instantiation
 *     list.  See 'callAccessFuncByNum' for calling a specific 
 *     collection instance by number instead of name.
 * ----------------------
 */
void callAccessFuncByName(char *collName, int elemNumi, 
			  int elemNumj, char *fname) {
  /*
   * collName - variable name of collection to access.
   * elemNumi - index of the first dimension of the element to
   *            access.
   * elemNumj - index of the second dimension of the element to
   *            access.  IF THERE IS NO SECOND DIMENSION, THIS 
   *            PARAMETER MUST BE A NEGATIVE NUMBER.
   * fname    - the function name to access (i.e. the name of 
   *            the element class variable).
   */
  callAccessFuncByNum(findIthCollection(collName), elemNumi,
		      elemNumj, fname);
}



/*
 * ----------------------
 *  void callUserMemFunc(int collnum, 
 *		     int elemnumi, int elemnumj, char *methodname);
 *     this func calls a user defined element method in the parallel
 *     program.  "User defined" methods are those which the user
 *     specifies by prepending the string "UserDefined_" onto the 
 *     method name.  The BREEZY precompilation process identifies
 *     these methods and prepares runtime access to them.
 * ----------------------
 */
void callUserMemFunc(int collnum, 
		     int elemnumi, int elemnumj, char *methodname) {
  /*
   * collNum - number in the collection variable list 'VarList'
   *            of collection to access.
   * elemNumi - index of the first dimension of the element to
   *            access.
   * elemNumj - index of the second dimension of the element to
   *            access.  IF THERE IS NO SECOND DIMENSION, THIS 
   *            PARAMETER MUST BE A NEGATIVE NUMBER.
   * methodname - the method name to access, _WITHOUT_ the prepended
   *            "UserDefined_" string.
   */
  char buf[100];

  if (elemnumj > 0) {
    /* 2D  */

    sprintf(buf, "callUserMemFunc %d %d %d %s",
	    collnum, elemnumi, elemnumj, methodname);
  } else {
    /* 1D  */
    sprintf(buf, "callUserMemFunc %d %d %s",
	    collnum, elemnumi, methodname);
  }
  sendText(buf);
}


/*
 * ----------------------
 * void callUserFunc(char *funcname);
 *     this func calls a user defined c function in the parallel
 *     program.  "User defined" functions are those which the user
 *     specifies by prepending the string "UserDefined_" onto the 
 *     function name.  The BREEZY precompilation process identifies
 *     these functions and prepares runtime access to them.
 * ----------------------
 */
void callUserFunc(char *funcname) {
  /*
   * funcname - the function name to access, _WITHOUT_ the prepended
   *            "UserDefined_" string.
   */
  char buf[100];

  sprintf(buf, "callUserFunc %s",
	  funcname);
  sendText(buf);
}

/*
 * ----------------------
 * void callUserMemFuncOverAll(int collnum, char *memfunc);
 *     this func calls the user defined member function for all elements
 *     of the specified collection in the parallel program.  
 *     "User defined" methods are those which the user
 *     specifies by prepending the string "UserDefined_" onto the 
 *     method name.  Note that this does not include the user defined
 *     -functions- which can be list using the getUserFuncs routine.
 * ----------------------
 */
void callUserMemFuncOverAll(int collnum, char *memfunc) {
  char buf[30];
  sprintf(buf, "callUserMemFuncOverAll %d %s", collnum, memfunc);
  sendText(buf);
}

/*
 * ----------------------
 * void callAccessFuncOverAll(int collnum);
 *     this func calls the generated member function for accessing the
 *     fields of the element of the specified collection.  It makes
 *     this call for all elements in the collection.  So, there will
 *     be as many packages sent back by the parallel program as there
 *     are elements in the collection.
 * ----------------------
 */
void callAccessFuncOverAll(int collnum, char *fname) {
  char buf[30];
  sprintf(buf, "getAllElems %d %s", collnum, fname);
  sendText(buf);
}


