/*------------------------------------------------------------
 *  Darryl Brown
 *  
 *  pC++ tools research group, University of Oregon
 *  October, 1990
 *  
 *  cargo.C
 *
 *  An implementation of a packaging system for sending/receiving
 *  data over connections using sockets.
 *
 *------------------------------------------------------------
 */


/*------------------------------------------------------------
 * INCLUDE FILES
 *
 */

#include "cargo.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
/*------------------------------------------------------------
 * GLOBALS
 *
 */

/*  A number that designates the next currently available socket.
 */
static int CRG_ConnectionNum[MAX_CONNECTIONS];
/*   The shared message allocated for shared type.
 */
static aPackage *CRG_SharedPackage;

#if 0
/* currently unused, so If-defed out... */
/*  The host name (or IP address) of the connected computer.
 */
static char CRG_ConnectionHostName[16];
/*  The host name (or IP address) of the current computer.
 */
static char CRG_MyHostName[16];
#endif


/*
 *  -----------------------------------
 *  MACROS  
 *  -----------------------------------
 *
 */
#define DEBUG 1
#define CRG_PROF 0

/*------------------------------------------------------------
 *  Interface Functions:
 * 
 *  aPackage *CRG_SendPackage(int connNum, aPackage *msg)
 *    Sends a message over connection, preceded by control
 *    information.
 *
 *  aPackage *CRG_ReceivePackage(int connNum)
 *    Receieves a package over connection.  Note that it mallocs 
 *    all the space necessary for the 
 *
 *  aPackage *CRG_InitPackage(packageTypes packType);
 *    Initializes a package so that it can be packed with 
 *    one or more calls to CRG_Pack.
 *
 *  aPackage *CRG_Pack(aPackage *msg, void *data, int numBytes)
 *    Packs new information onto the message, msg.
 *
 *  void CRG_RecyclePackage(aPackage *pkg)
 *    Frees up the space allocated for a package
 *
 * int CRG_SendText(register int connNum, char *ms)
 *     sends textual data over the previously
 *     opened connection.
 *
 * char *CRG_ReceiveText(register int connNum)
 *     receives textual data over the previously
 *     opened connection.
 *
 *------------------------------------------------------------
 */



/*------------------------------------------------------------
 * INTERNAL FUNCTIONS
 *------------------------------------------------------------
 */
int nextConnNum() {
  int i = 0;

  while ((CRG_ConnectionNum[i] != 0) && (i < MAX_CONNECTIONS)) {
    i++;
  }

  if (i == MAX_CONNECTIONS) {
    perror("Exceeded maximum Number of Connections: MAX_CONNECTIONS)\n");
  }
  
  return i;
}

void * biggerPackage(aPackage *pkg) {
  pkg->data = 
    TRNS_addBufferBlock(pkg->data, pkg->bufLength, pkg->length);
  pkg->bufLength += TRNS_BUFBLOCK;
  return pkg->data;
}

/*------------------------------------------------------------
 * EXTERNAL FUNCTIONS
 *------------------------------------------------------------
 */


/*
 *------------------------------------------------------------
 *  aPackage *CRG_InitPackage(packageTypes packType);
 *    Initializes a package so that it can be packed with 
 *    one or more calls to CRG_Pack.
 *
 *------------------------------------------------------------
 */
aPackage  *CRG_InitPackage(int packageType) {
 /* see cargo.h for package type defs */
  aPackage *pkg;


  switch (packageType) {
  case CRG_shared: {
    pkg = CRG_SharedPackage;
  }
  case CRG_unique: {
    if ((pkg = (aPackage *) malloc(sizeof(aPackage))) == NULL) {
      perror("malloc failed\n");
    }
  }
  }
  pkg->length = 0;
  pkg->bufLength = 0;
  pkg->packageType = packageType;
  return pkg;
}

/*------------------------------------------------------------
 *  int CRG_SendPackage(int connNum, aPackage *msg)
 *    Sends a message over connection, preceded by control
 *    information.
 *
 *------------------------------------------------------------
 */
int CRG_SendPackage(int connNum, aPackage *pkg) {
  char *controlBuf, buf[CRG_HDR_LEN+2];
  aMessage tmpMessage;

  tmpMessage.length = CRG_HDR_LEN + pkg->length;
  if ((controlBuf = malloc (tmpMessage.length+2)) == NULL) {
    perror("malloc failed\n");
  }
  /* store header in controlBuf */
  tmpMessage.data = controlBuf;

  sprintf(buf, "%d         ", pkg->length);
  strncpy(controlBuf, buf, CRG_HDR_LEN);

  /* now, pack rest of message behind header */
  controlBuf += CRG_HDR_LEN;
  memcpy(controlBuf, pkg->data, pkg->length);

  /* send off package */
  if (TRNS_Send(connNum, &tmpMessage) < tmpMessage.length) {
    perror("Couldn't send package.\n");
  }

#if DEBUG
  printf("\n*really* SENT:  %s\n\n", (char *) pkg->data);
#endif
  
  return pkg->length;
}

/*------------------------------------------------------------
 *  aPackage *CRG_ReceivePackage(int connNum)
 *    Receieves a package over connection.  Note that it mallocs 
 *    all the space necessary for the 
 *
 *------------------------------------------------------------
 */
aPackage * CRG_ReceivePackage(int connNum) {
  int cNum;
  aPackage *pkg;
  aMessage *msg;
  char *cstr;
  char buf[CRG_HDR_LEN+1];

  msg = TRNS_Receive(connNum, CRG_HDR_LEN);
  pkg = CRG_InitPackage(CRG_unique);
  if (msg->length < CRG_HDR_LEN) {
    perror("Couldn't receive header information.\n");
  }
  
  /* cstr = strtok((char *) msg->data, "#"); */
  cstr = (char *) msg->data;
  if (sscanf(cstr, "%d", &cNum) < 1) {
    strncpy(buf, cstr, CRG_HDR_LEN);
    printf("Warning, header information not integer: %s.\n",
	   buf);
    pkg->data = NULL;
    pkg->length = 0;
    TRNS_RecycleMessage(msg);
    return pkg;
  }
  TRNS_RecycleMessage(msg);
  
#if DEBUG 
  printf("CRG debug: recieving package content (%d bytes)\n",
	 cNum);
#endif

  msg = TRNS_Receive(connNum, cNum);

  if ( msg->length < cNum) {
    perror("Problem receiving package contents.\n");
    pkg->data = NULL;
    pkg->length = 0;
  } else {
    CRG_Pack(pkg, msg->data, msg->length);
  }

#if DEBUG
  printf("\nCRG debug:  RECEIVED:  %s\n\n", (char *) msg->data);
#endif

  TRNS_RecycleMessage(msg);

  return pkg;



}

/*
 *------------------------------------------------------------
 *  int CRG_Pack(aPackage *pkg, void *data, int numBytes)
 *    Packs new information onto the package, pkg.
 *
 *------------------------------------------------------------
 */
int CRG_Pack(aPackage *pkg, void *data, int numBytes) {
  void *tmpBuf;

  if ((pkg->length + numBytes) > pkg->bufLength) {
    biggerPackage(pkg);
    CRG_Pack(pkg, data, numBytes);
  } else {
    /* copy buffer contents over */
    tmpBuf = ((char *) pkg->data) + pkg->length; 
#if 0
    printf("ptr %d data %s numBytes %d", tmpBuf,
	   (char *)data, numBytes);
#endif
    memcpy(tmpBuf, data, numBytes);
    pkg->length += numBytes;
  }
  return pkg->length;
}

/* ------------------------------------------------------------
 *  void CRG_RecyclePackage(aPackage *pkg)
 *    Frees up the space allocated for a package
 *
 */
void CRG_RecyclePackage(aPackage *pkg) {
  if (pkg->length > 0) {
    free(pkg->data);
  }
  free(pkg);
}


/*
 * ----------------------
 * int CRG_SendText(register int connNum, char *ms)
 *     sends textual data over the previously
 *     opened connection.
 * ----------------------
 */
int CRG_SendText(register int connNum, char *ms) {
  /* ms - textual message to send. */
  aPackage *pkg;
  int retval;
  pkg = CRG_InitPackage(CRG_unique);

  CRG_Pack(pkg, ms, (strlen(ms)+1));
  retval = CRG_SendPackage(connNum, pkg);
  CRG_RecyclePackage(pkg);
  return retval;
}


/*
 * ----------------------
 * char * CRG_ReceiveText(register int connNum, char *ms);
 *     sends textual data over the previously
 *     opened connection.
 * ----------------------
 */
char * CRG_ReceiveText(register int connNum) {
  /* ms - textual message to send. */
  char *ms;
  aPackage *pkg;

  pkg = CRG_ReceivePackage(connNum);
  if ((ms = malloc (pkg->length)) == NULL) {
    perror("malloc failed\n");
  }
  strcpy(ms, (char *)pkg->data);
  CRG_RecyclePackage(pkg);
  return ms;
}






