/*------------------------------------------------------------
 *  Darryl Brown
 *  
 *  pC++ tools research group, University of Oregon
 *  October, 1990
 *  
 *  transport.c
 *
 *  An implementation of a low level data transfer library using
 *  sockets.
 *
 *------------------------------------------------------------
 *------------------------------------------------------------
 */

/*------------------------------------------------------------
 * INCLUDE FILES
 *
 */
#if CONFIG_ARCH==rs6000
#define _NO_BITFIELDS
#endif
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h> 
#include <stdio.h>
#include <stdlib.h>
#include "transport.h"

#include "compatible.h"
#if SYS5
#include <string.h>
#else
#include <strings.h>
#endif

/* only needed on sgi machines... */
#ifdef __sgi
#include <bstring.h>
#endif

/*
 *------------------------------------------------------------
 *  Macros
 *
 */
#ifndef INADDR_NONE
#  define INADDR_NONE ((u_long)-1)    /* -1 return */
#endif

#define DO_SPIN_READ 0
#define WAIT_TIME 5
#define DEBUG 0
#define TRNS_PROF 0


/*------------------------------------------------------------
 *
 *  Interface Functions:
 *
 *  Int TRNS_OpenClient(int portNum, char *host)
 *    Opens a new data client connection using portNum to host, returns 
 *    an ID that can be used in other data transer functions.
 *
 *  Int TRNS_OpenServer(int portNum)
 *    Opens a new data server connection using portNum, returns 
 *    an ID that can be used in other data transer functions.
 *
 *  void TRNS_CloseConnection(int ConnNum)
 *    Closes the connection designated by ConnNum.
 *
 *  void TRNS_Send(int ConnNum, void *data, int dataLen)
 *    Writes data into connection specified by ConnNum (of length
 *    dataLen bytes).
 *
 *  void *TRNS_Receive(int ConnNum) 
 *    Reads all data from connection specified by ConnNum. 
 *
 *
 *  void TRNS_RecycleMessage(aMessage *msg)
 *    Frees up the space allocated for a message
 *
 *------------------------------------------------------------
 */

/*------------------------------------------------------------
 *  GLOBALS
 *------------------------------------------------------------
 */

/*
 *------------------------------------------------------------
 *  INTERNAL FUNCTIONS
 *------------------------------------------------------------
 */

static void *chk_malloc(int size) {
  void *tmp;

  if ((tmp = (void *) malloc(size)) == NULL) {
    perror("Malloc failed (chk_malloc proc)\n");
  } else {
    return tmp;
  }
  return tmp;
}

/* ------------------------------------------------------------
 *  void *TRNS_addBufferBlock(void *buf, int bufLen, int copyLen)
 *    Increases buffer size, and copies copylen bytes from old buf
 *    to new buf.  Returns the new buffer pointer.
 *
 */
void *TRNS_addBufferBlock(void *buf, int bufLen, int copyLen) {
  void *tmpBuf;

  if ((tmpBuf = (void *) malloc(bufLen + TRNS_BUFBLOCK)) == NULL) {
    perror("malloc failed\n.");
  }

  /* copy buffer contents over */
  memcpy(tmpBuf, buf, copyLen);

  if (copyLen > 0) {
    free(buf);
  }

  return tmpBuf;
}

/*
 *------------------------------------------------------------
 *  EXTERNAL FUNCTIONS
 *------------------------------------------------------------
 */

/*------------------------------------------------------------
 *
 *  Int TRNS_OpenClient(int portNum, char *hostName)
 *    Opens a new data client connection using portNum to hostName, 
 *    returns an ID that can be used in other data transer functions.
 *
 */
int TRNS_OpenClient(int portNum, char *hostName) {
  struct hostent *host;
  int sockfd, waited = 0, connVar;
  struct sockaddr_in serv_addr;
  struct protoent *pp;
  int one = 1;

  memset((char *) &serv_addr, 0, sizeof(serv_addr));
  serv_addr.sin_family = AF_INET;
  serv_addr.sin_port=htons(portNum);

  /* -- interpret hostName as ip address, if it fails -- */
  /* -- try interpreting as hostname                        -- */
  if ( (serv_addr.sin_addr.s_addr=inet_addr(hostName)) == INADDR_NONE ) {
    if ((host = gethostbyname(hostName)) == NULL) {
      fprintf(stderr, "client:can't get ip address from name %s\n",
              host);
      fflush(stdout);
      return -1;
    }
    memcpy((char*)&serv_addr.sin_addr, host->h_addr, host->h_length);

#if DEBUG
    fprintf (stderr, "%s translated to %s\n", hostName,
                                              inet_ntoa(serv_addr.sin_addr));
#endif
  }
  
  do {

    if((sockfd=socket(AF_INET,SOCK_STREAM,0))<0)
      perror("TRNS_ERROR (server): can't open stream socket");

    
    /* no buffers; */
    pp = getprotobyname( "tcp" );
    if( !pp ) {
	perror( "getprotobyname" );
	exit( 1 );
    }
    if (setsockopt(sockfd, pp->p_proto, TCP_NODELAY, (char *) &one, sizeof 
		   (one)) == -1) {
      printf("setsockopt TCP_NODELAY client connection failed.\n");
    }
    if ( setsockopt( sockfd, SOL_SOCKET, SO_REUSEADDR,
		    (char *) &one, sizeof(one)) == -1) {
      printf("setsockopt sndbuf client connection failed.\n");
    }
#if 0
    if (setsockopt(sockfd, SOL_SOCKET, SO_SNDBUF, vtmp, sizeof (int)) == -1) {
      printf("setsockopt sndbuf client connection failed.\n");
    }
    
    if (setsockopt(sockfd, SOL_SOCKET, SO_RCVBUF, vtmp, sizeof (int)) == -1) {
      printf("setsockopt rcvbuf  client connection failed.\n");
    }
#endif    
    if (waited == 1) {
      /* waiting for binding to come active; */
      fprintf(stderr, "client:can't connect to server, \n");
      fprintf(stderr, "\ttrying again in %d seconds.\n", WAIT_TIME);  
      sleep(WAIT_TIME);
    }
    connVar = connect(sockfd,(struct sockaddr *) 
		       &serv_addr,sizeof(serv_addr));
#if DEBUG
    printf("connection variable: %d\n", connVar);
#endif
    if (connVar<0) waited = 1;
  } while (connVar<0);
    
  if (waited) {
    printf("\n\tSocket bound.\n");
    fflush(stdout);
  }
  return sockfd;
}

/*------------------------------------------------------------
 *
 *  int TRNS_OpenServer(int portNum)
 *    Opens a new data server connection using portNum, returns 
 *    an ID that can be used in other data transer functions.
 *
 */
int TRNS_OpenServer(int portNum) {
  int sockfd, newsockfd, clilen, waited, one = 1;
  struct sockaddr_in cli_addr, serv_addr;
  struct protoent *pp;

  /* IPPROTO_TCP  */
  if ((sockfd=socket(AF_INET,SOCK_STREAM,0))<0) {
    perror("TRNS_ERROR (server): can't open stream socket");
  }

  memset((char *) &serv_addr, 0, sizeof(serv_addr));
  serv_addr.sin_family = AF_INET;
  serv_addr.sin_addr.s_addr=htonl(INADDR_ANY);
  serv_addr.sin_port = htons(portNum);

  if( (pp = getprotobyname( "tcp" )) == NULL) {
    perror( "getprotobyname" );
    exit( 1 );
  }

  if (setsockopt(sockfd, pp->p_proto, TCP_NODELAY, (char *) &one, sizeof 
		 one) == -1) {
    perror("setsockopt TCP_NODELAY server connection failed.\n");
    exit( 1 );
  }

  if ( setsockopt( sockfd, SOL_SOCKET, SO_REUSEADDR,
		  (char *) &one, sizeof(one)) == -1) {
    printf("setsockopt sndbuf client connection failed.\n");
  }
#if 0
  /* no buffers; */
  
  if (setsockopt(sockfd, SOL_SOCKET, SO_SNDBUF, vtmp, sizeof (int)) == -1) {
    printf("setsockopt sndbuf serverconnection failed.\n");
  }
  if (setsockopt(sockfd, SOL_SOCKET, SO_RCVBUF, vtmp, sizeof (int)) == -1) {
    printf("setsockopt rcvbuf serverconnection failed.\n");
  }
#endif 

#if DEBUG
  printf("about to try to bind to socket %d\n", portNum);
#endif

  waited = 0;
  while(bind(sockfd,(struct sockaddr *)&serv_addr,sizeof(serv_addr))<0) {
    /* waiting for binding to come active; */
    fprintf(stderr, "server:can't bind local address, \n");
    fprintf(stderr, "\ttrying again in %d seconds.\n", WAIT_TIME);  
    sleep(WAIT_TIME);
    waited++;
  }

#if DEBUG
  printf("bound to socket %d\n", portNum);
#endif

  if (waited) {
    printf("\n\tSocket bound.\n");
    fflush(stdout);
  } else {
#if DEBUG
    /* always print this message when DEBUG is true; */
    printf("\n\tSocket bound.\n");
    fflush(stdout);
#endif
  }
  
  if (listen(sockfd,5) == -1) {
    perror("listen");
    exit(1);
  }

  clilen=sizeof(cli_addr);

  newsockfd=accept(sockfd,(struct sockaddr *) &cli_addr,&clilen);

#if DEBUG
  printf("\n\tConnection established.\n");
  fflush(stdout);
#endif

  if(newsockfd<0)
    perror("TRNS_ERROR (server): can't open stream socket");

  return newsockfd;
}

/*------------------------------------------------------------
 *  void TRNS_CloseConnection(int ConnNum)
 *    Closes the connection designated by ConnNum.
 *
 */
void TRNS_CloseConnection(int connNum) {
  close (connNum);
}
     
/* ------------------------------------------------------------
 *  int TRNS_Send(int connNum, aMessage *msg)
 *    Writes msg into connection specified by ConnNum
 *
 */
int TRNS_Send(register int connNum, aMessage *msg) {
  register int nleft, nwritten, nbytes;
  register char *data;
#if TRNS_PROF  
  static int mnum = 0;
  static int bnum = 0;
#endif

  data = (char *) msg->data;
  nbytes = nleft = msg->length;

  /* loop until written all bytes */ 
  while (nleft > 0) {
    nwritten=write(connNum, (char *) data, nleft);
    if (nwritten<=0)
      return (nwritten);
    nleft-=nwritten;
    data += nwritten;
  }


#if DEBUG
  printf("SENT %d bytes... ", msg->length);
#endif
#if TRNS_PROF  
  mnum++;
  bnum = bnum + nbytes - nleft;
  printf(" ... sent %d msgs, %d bytes... ", mnum,
	 bnum);
#endif
  return (nbytes - nleft);
}


/*------------------------------------------------------------
 *  msg *TRNS_Receive(int connNum, int dataLen) 
 *    Reads dataLen bytes of data from connection specified by ConnNum. 
 *    
 */
aMessage *TRNS_Receive(register int connNum, int dataLen){
  register int nleft, nread, buflen;
  void *c, *ptr;
  aMessage *msg;
  struct timeval timeout = {0,0};

#if TRNS_PROF  
  static int mnum = 0;
  static int bnum = 0;
#endif

  msg = (aMessage *) chk_malloc(sizeof (aMessage));

  nleft = dataLen;
  nread = 0;
  buflen = 0;
  while (buflen < dataLen) {
    c = TRNS_addBufferBlock(c, buflen, 0);
    buflen = buflen+ TRNS_BUFBLOCK;
  }
  while (nleft > 0) {
    ptr = (((char *)c)+(dataLen-nleft));

#if DO_SPIN_READ
    /* this implements non-blocking read... */
    FD_ZERO(&fd);
    FD_SET(connNum, &fd);
    while (select(FD_SETSIZE, &fd, NULL, NULL, &timeout) == 0)
      {
	FD_ZERO(&fd);
	FD_SET(connNum, &fd);
	timeout->tv_sec = 0;
	timeout->tv_usec = 0;
      };
#endif

    nread = read(connNum, ptr, dataLen);
    
    /* check to see if all of the bytes have been read by the 
     * socket, or if an error has occurred (nread < 0)
     */
    if (nread < 0)  {
      msg->data = NULL;
      msg->length = nread;
      return msg;
    } else if (nread == 0) {
      break;
    }
    nleft -= nread;
    buflen += MESSAGE_BUF_SIZE;
  }
  
  msg->data = (char *) chk_malloc(dataLen);
  msg->length = dataLen - nleft;
  memcpy(msg->data, c, dataLen);
  free(c);


#if DEBUG
  printf("RECEIVED %d bytes... ", msg->length);
#endif
#if TRNS_PROF  
  mnum++;
  bnum = bnum + dataLen;
  printf(" ... received %d msgs, %d bytes... ", mnum,
	 bnum);
#endif

  return msg;
}


     
/* ------------------------------------------------------------
 *  void TRNS_RecycleMessage(aMessage *msg)
 *    Frees up the space allocated for a message
 *
 */
void TRNS_RecycleMessage(aMessage *msg) {
  if (msg->length > 0) {
    free(msg->data);
  }
  free(msg);
}


