//  ----------------------------------
//  Darryl Brown
//  University of Oregon pC++/Sage++
//
//  TItypeDesc.C  - 
//
//  This module is called from instr.C.  It instruments a pC++
//  program so that parallel data from the program can later 
//  be accesses by extern applications.  These applications
//  "connect" to the program through an object passing 
//  library (cargo.c) that works over the network (using sockets).
//  
//  ----------------------------------
//  ----------------------------------

#include <stdio.h>
#include <stdlib.h>

#include "compatible.h"
#ifdef SYS5
#include <string.h>
#else
#include <strings.h>
#endif
  
#include "sage++user.h"
#include "sageUtils.h"
#include "typeApi.h"
#include "astrlib.h"


/*
 *  -----------------------------------
 *  MACROS  
 *  -----------------------------------
 *
 */

#define DEBUG 0
#define USR_FUNC_STR "UserDefined_"
#define USR_FUNC 0
#define ACC_FUNC 1


/*
 *  -----------------------------------
 *  GLOBALS
 *  -----------------------------------
 *
 */

TA_nameList  *CollectionQueue;
TA_nameList  *ElementQueue;




/*
 *  -----------------------------------
 *  FORWARD DECLARATIONS
 *  -----------------------------------
 *
 */
SgStatement * writeElementAccessFunction(SgFile *f, 
					 TA_composite *comp);
void buildType(char *compName, SgFile *aFile);
void writeDeepCompositeAccessFunc(SgFile *f, 
				  char *parentClass,
				  char *childClass,
				  char *childName);

/*
 *  -----------------------------------
 *  FUNCTIONS
 *  -----------------------------------
 *
 */
void brk_mySageInit() {
  ElementQueue = TA_createNameList();
  CollectionQueue = TA_createNameList();
}

/* 
 * TA_descriptEnum sageToDescEnum(int mflag) 
 *    converts an integer sage descriptor flag into 
 *    TA_descriptEnum type 
 */
TA_descriptEnum sageToDescEnum(int mflag) {

  if (mflag&&BIT_VOLATILE) {
    return d_volatile;
  }
  if (mflag&&BIT_REGISTER) {
    return d_register;
  }
  if (mflag&&BIT_LONG) {
    return d_long;
  }
  if (mflag&&BIT_SHORT) {
    return d_short;
  }
  if (mflag&&BIT_UNSIGNED) {
    return d_unsigned;
  }
  if (mflag&&BIT_SIGNED) {
    return d_signed;
  }
  return d_nodescriptor;
}


/* 
 * TA_typeEnum sageTotypeEnum(int variant) 
 *    converts an sage type variant
 *    into a TA_typeEnum 
 */
TA_typeEnum sageTotypeEnum(int variant) {

  switch(variant) {

  case T_INT:
    return t_int;
  case T_DOUBLE:
    return t_double;
  case T_FLOAT:
    return t_float;
  case T_CHAR:
    return t_char;
  case T_CLASS:
    return t_class;
  case T_DERIVED_CLASS:
    return t_class;
  case T_COLLECTION:
    return t_collection;
  case T_DERIVED_COLLECTION:
    return t_collection;
  case T_DERIVED_TYPE:
    return t_derived_type;
  case T_STRUCT:
    return t_struct;
  case T_UNION:
    return t_union;
  case T_DESCRIPT:
    return t_notype;
  }
  return t_notype;
}



/* 
 * int typeEnumToSage(TA_typeEnum variant);
 *    converts an sage typeEnum variant
 *    into a sage variant
 */
int typeEnumToSage(TA_typeEnum variant) {

  switch(variant) {

  case t_int:
    return T_INT;
  case t_double:
    return T_DOUBLE;
  case t_float:
    return T_FLOAT;
  case t_char:
    return T_CHAR;
  case t_class:
    return T_CLASS;
  case t_collection:
    return T_COLLECTION;
  case t_derived_type:
    return T_DERIVED_TYPE;
  case t_struct:
    return T_STRUCT;
  case t_union:
    return T_UNION;
  }
  return T_UNKNOWN;
}


/* 
 * TA_typeDesc * sageToTA_typeDesc(SgType *ty,  SgFile *f) 
 *    converts a sage SgType into a TA_typeDesc
 *    structure.
 */
TA_typeDesc * sageTotypeDesc(SgType *ty, SgFile *f) {
  TA_typeEnum te;
  TA_typeDesc *td;
  TA_descriptEnum de;
  SgDescriptType *sdt;
  SgExpression *se;
  char *buf;
  int dm;
  TA_dimType *d;
  SgDerivedType *sageDT;
  SgSymbol *typeSym;
  SgClassSymb *clSymb;

  if (sdt = isSgDescriptType(ty)) {
    de = sageToDescEnum(sdt->modifierFlag());
    td = sageTotypeDesc(sdt->baseType(), f);
    td->desc = de;
    return td;
  }
  
  if (isSgPointerType(ty)) {
    td = sageTotypeDesc(ty->baseType(), f);
    td->isPointer++;
    return td;
  } 

  if (isSgArrayType(ty)) {
    // set dimension parameters...;
    td = sageTotypeDesc(ty->baseType(), f);
    SgArrayType *at;
    at = (SgArrayType *) ty;
    d = TA_createDimType();
    for (int i = 0; i < at->dimension(); i++) {
      se = at->sizeInDim(i);
      buf =  unparseExp(se);
      dm = evalStr(buf);
#if DEBUG      
      printf("dim[%d] is %s (%d)\n", i, buf, dm);
#endif      
      TA_addDim(d, dm);
      free(buf);
    } 
    td->dims = d;
    return td;
  }    
  
  if ((sageDT = isSgDerivedType(ty)) &&
      (typeSym = sageDT->typeName()) &&
      (clSymb = isSgClassSymb(typeSym))) {
    td = TA_createTypeDesc();
    td->name = strsave(clSymb->identifier());
    td->thisType = t_class;
    buildType(td->name, f);
    return td;
#if DEBUG      
    printf("derived Type %s\n", td->name);
#endif
  }
  
  if (isSgClassType(ty)) {
    td = TA_createTypeDesc();
    td->name = strsave(ty->symbol()->identifier());
    td->thisType = sageTotypeEnum(ty->variant());
#if DEBUG      
    printf("class Type %s\n", td->name);
#endif
    if (TA_isComposite(td->thisType)) {
      buildType(td->name, f);
    }
    return td;
  }

  td = TA_createTypeDesc();
  td->thisType = te = sageTotypeEnum(ty->variant());
  return td;


}

/*
 * void buildType(char *compName, SgFile *aFile) {
 *     Build simplified representation of composite data
 *     structure sage to TA_composite type (see typeApi.h).
 */
void buildType(char *compName, SgFile *aFile) {
  int i;
  SgClassSymb *cSym;
  SgFieldSymb *flSym;
  TA_composite *comp;
  TA_field *f, *pf;

  cSym = (SgClassSymb *) 
    findDefSymb(compName, aFile);
  if (cSym == NULL)
    return;

  comp = TA_createComposite();
  comp->type = 
    sageTotypeEnum(cSym->type()->variant());
  comp->name = strsave(compName);
  pf = NULL;

#if DEBUG
  printf("composing %s\n\n", compName);
#endif

  for (i = 0; i <= cSym->numberOfFields(); i++) {
    if (flSym = isSgFieldSymb(cSym->field(i))) {
      
      f = TA_createField();
      f->name = strsave(flSym->identifier());
#if DEBUG
      printf("creating field %s\n", 
	     flSym->identifier());
#endif
      f->type = sageTotypeDesc(flSym->type(), aFile);
      if (pf != NULL) {
	/* not the first one, so add it to the list */
	pf->next = f;
      } else {
	/* it is the first one, so let the composite
	 * point to it */
	comp->fields = f;
      }

      /* if this field is a class, we are going to have to 
       * instrument it as well, so put it in the type list.
       */
	
      pf = f;
    }
  }
  TA_putComposite(comp);
#if DEBUG      
  printf("done with %s.\n", compName);
#endif
}

/*  
 * void addElemAndCollection(char *collName,
 *                        char *elemName,
 *			  SgFile *f)
 *     Adds collection and element name pairs 
 *     to respective lists CollectionQueue and
 *     ElementQueue.
 */
void addElemAndCollection(char *collName, char *elemName,
			  SgFile *f) {
  TA_nameEntry *cn, *en;

  cn = CollectionQueue->first;
  en = ElementQueue->first;
  while (cn != NULL && en != NULL) {
    if (strsame(cn->name, collName) && 
	strsame(en->name, elemName) ) {
      return ;
    }
    cn = cn->next; 
    en = en->next; 
  }
	
  /* store collection name in queue. */
  TA_pushName(collName, CollectionQueue);
#if DEBUG      
  printf("added %s to collec queue, %s to elem queue\n", collName, elemName);
#endif
  /* store element name in queue. */
  TA_pushName(elemName, ElementQueue);
  buildType(elemName, f);
}


/*  
 * SgStatement *addPtrSelectionCond(SgFile *f, 
 *				 TA_field *fld,
 *				 TA_composite *parent,
 *				 SgStatement *fhs, 
 *				 SgSymbol *elemSymb)
 *     Creates the sage representation for an if  
 *     statement that will check a fname parameter
 *     for equality with a particular field name
 *     in the element class, then call the 
 *     appropriate access function if they are the
 *     same. 
 */
SgStatement *addPtrSelectionCond(SgFile *f, 
				 char *fldName,
				 char *baseName,
				 TA_composite *parent,
				 SgStatement *fhs,
				 SgSymbol *elemSymb,
				 SgStatement *last) {
  // SgStatement *last will eventually be the last parameter
  int len;
  SgStatement *lst, *tmps;
  SgExpression *fexp, *fexp2;
  SgExpression *strCmpExp, *fldNameExp, *strLenExp;
  char buf[100];
  SgSymbol *vs;
  TA_field *fld;

  // --------------------------------------
  //   
  //  add code that selects a particular memberfunc
  //  (e.g. for basename "accessFunc" and field 'x'):
  //
  //     if (strsame(fname, "x")) {
  //       elementPtr->accessFunc_x();
  //     }
  //


  // check to see if field is a class...
  if (*fldName) {
    fld = parent->fields;
    while (fld) {
      if (strsame(fld->name, fldName)) {
	// check type to see if it is a class
	if (fld->type->thisType == t_class) {
	  
	  // special case if field is a class
	  
	  // --------------------------------------
	  //   
	  //  add code that selects this particular memberfunc,
	  //  and passes on the rest of the string after the
	  //  first word.
	  //
	  //     if (strncmp(fname, "x", strlen("x")) == 0) {
	  //       castedPtr->
	  //         brk_AccessFunc_simple_elem_test_class((fname 
	  //                                        + strlen("x") + 1));
	  //     }
	  //

	  // then we deal this special case for classes
	  // build strncomp, & add the fname param
	  //  strncmp(fname);
	  strCmpExp = buildFuncCallExp("wordSame", f);
	  vs = findVarDeclSymbFromStmt("name", fhs);
	  addParamToFuncExp(strCmpExp, vs);
	  
	  // add the fld name literal string
	  //  strncmp(fname, "x");
	  fldNameExp = new SgValueExp(fld->name);
	  addParamExpToFuncExp(strCmpExp, fldNameExp);
	  

	  sprintf(buf, "brk_AccessFunc_%s__%s", 
		  parent->name, fld->type->name);


	  lst = buildMembFuncCallStmt(parent->name,
					     buf,
					     f,
					     elemSymb);


	  
	  // now build the strlen func, & add the literal fld
	  // name expression to it.
	  // strlen("x");
	  len = strlen(fld->name);
	  strLenExp = new SgValueExp(len);
	  
	  fexp = new SgExpression(ADD_OP);
	  // the strlen("x");
	  fexp->setLhs(strLenExp->copy());
	  fexp->setRhs(*(new SgVarRefExp(*vs)));
	  
	  fexp2 = extractFuncExpFromHedr(lst);
	  addParamExpToFuncExp(fexp2, fexp);
#if DEBUG	  
	  printf("\nBUILD MEMBER FUNCTION CALL: \n");
	  lst->unparsestdout();
#endif	  
	  //create the if statemnet and add it 
	  // check for == NULL
	  //  (wordSame(fname, "x", strlen("x"))
	  lst = new SgIfStmt(*strCmpExp, *lst);
	  lastExecutableOfBlock(fhs)->insertStmtAfter(*lst);
	  return last;
	}
      }
      fld = fld->next;
    }
  }


  fexp2 = new SgValueExp(strsave(fldName));
  fexp = buildFuncCallExp("wordSame", f);
  addParamExpToFuncExp(fexp, fexp2);

  // find the name variable symbol
  vs = findVarDeclSymbFromStmt("name", fhs);
  addParamToFuncExp(fexp, vs);

  if (*fldName) {
    sprintf(buf, "%s_%s", baseName, fldName);
  } else {
    sprintf(buf, "%s", baseName, fldName);
  }
  tmps = buildMembFuncCallStmt(parent->name,
			       buf,
			       f,
			       elemSymb);
  //create the if statemnet and add it 
  lst = new SgIfStmt(*fexp, *tmps);
  last->insertStmtAfter(*lst);
  //fhs->lastDeclaration()->insertStmtAfter(*lst);
  return lst;
}




/* 
 * SgStatement * addTrimCode(SgFile *f, char *trimName
 *			  SgSymbol *lhValue, SgStatement *fhs)
 *      This functions add the code for a 
 *      call to trimLeadingSpaces function.
 */
SgStatement * addTrimCode(SgFile *f, char *trimName,
 			  SgSymbol *lhValue, SgStatement *fhs) {
  
  SgExpression *fexp;
  SgStatement *lst;

  fexp = buildFuncCallExp("trimLeadingSpaces", f);
  addParamExpToFuncExp(fexp, 
		       new 
		       SgVarRefExp(*
				   (
				    searchForParamInFuncStmt(fhs, 
							     trimName)
				    )));

  lst = buildAssignStmt(new SgVarRefExp(*lhValue),
			fexp);
  fhs->lastDeclaration()->insertStmtAfter(*lst);
  return lst;
}

/* 
 * SgStatement * addInitCode(SgFile *f,
 *			  SgSymbol *packSym)
 *      This functions add the code for a 
 *      CRG_InitPackage statement. 
 */
SgStatement * addInitCode(SgFile *f,
			  SgSymbol *packSym) {

  SgExpression *funcExp;
  SgStatement *lst;
  SgVarRefExp *fexp;
  
  // ---------------------------------;
  //
  // add function call, send package off:
  //
  //     pkg = CRG_InitPackage(CRG_shared);
  //

  funcExp = buildFuncCallExp("CRG_InitPackage", f);
  fexp = new SgVarRefExp(findVarDeclSymb("CRG_shared", f));
  addParamExpToFuncExp(funcExp, fexp);
  fexp = new SgVarRefExp(*packSym);
  lst = buildAssignStmt(fexp, funcExp);

  return lst;
}

/* 
 * SgStatement * addSendCode(SgFile *f,
 *			  SgSymbol *packSym)
 *      This functions add the code for a 
 *      CRG_SendPackage statement. 
 */
SgStatement * addSendCode(SgFile *f,
			  SgSymbol *packSym) {

  SgExpression *fexp;
  SgStatement *lst;
  SgVarRefExp *ve;
  
  // ---------------------------------;
  //
  // add function call, send package off:
  //
  //     CRG_SendPackage(brk_ConnectionNum, pkg);
  //
  lst = buildFuncCallStmt("CRG_SendPackage", f);
  fexp = extractFuncExpFromHedr(lst);  
  // add first parameter;
  ve = new SgVarRefExp(*(findVarDeclSymb("brk_ConnectionNum", f)));
  addParamExpToFuncExp(fexp, ve);
  ve = new SgVarRefExp(*(packSym));
  addParamExpToFuncExp(fexp, ve);

  return lst;
}


/* 
 * SgStatement * addPackCode(SgFile *f,
 *			  SgExpression *dataExp, 
 *			  SgSymbol *packSym)
 *      This adds code for packing a data expression
 *      into a package. 
 */
SgStatement * addPackCode(SgFile *f,
			  SgExpression *dataExp, 
			  SgSymbol *packSym) {

  SgExpression *fexp, *funcExp;
  SgStatement *lst;

  
  // ---------------------------------;
  //
  // add code to pack a data expression
  // (e.g. if 'buf' were the data expression):
  //
  //     CRG_Pack(pkg, buf, strlen(buf));
  //

  lst = buildFuncCallStmt("CRG_Pack", f);
  // add first parameter;
  funcExp = extractFuncExpFromHedr(lst);  
  fexp = new SgVarRefExp(*packSym);
  addParamExpToFuncExp(funcExp, fexp);
  // add second paramemter
  addParamExpToFuncExp(funcExp, dataExp);
  // build third paramemter...
  fexp = buildFuncCallExp("strlen", f);
  addParamExpToFuncExp(fexp, dataExp);
  // now add third parameter
  addParamExpToFuncExp(funcExp, fexp);

  return lst;
}

// -------------------------------------
// predeclare function for use here...
SgStatement * writeAccessToFields(TA_composite *comp,
			 SgFile *f, SgStatement *fhdr);
// -------------------------------------


/*
 *  void writePtrFieldAccessFunc(SgFile *f, 
 *			     TA_composite *parent,
 *			     TA_field *fld)
 *  this funciton generates the sage code to access a 
 *  the data in a pointer field.
 *
 */
void writePtrFieldAccessFunc(SgFile *f, 
			     TA_composite *parent,
			     TA_field *fld) {
  char *buf1, *tmpBuf1;
  SgStatement *st, *tmps, *lst;
  SgSymbol *bufsym, *fs, *loopvarsym, *packSym;
  SgExpression *fexp, *fexp2;
  SgType *ty, *voidTy, *charTy;
  SgValueExp *ve;
  SgArrayType *sat;
  SgFuncHedrStmt *fhs;
  SgArrayRefExp *arrRef;
  // SgVariableSymb *arrVar;
  SgSymbol *arrVar;
  SgForStmt *fstmt, *fs2, *firstFor;
  int i = 0;
  TA_typeDesc *at, *typeptr;
  int arrayDims[10];
  char *loopstr;
  SgVarRefExp *bufExp;

  buf1 = new char[100];
  typeptr = fld->type;
  at = fld->type;
  charTy = SgTypeChar();
  voidTy = SgTypeVoid();


  // -----------------------------------------
  // test to see if this is a class type field.  If it is, then
  // we need to just return.  This accessFunctions for this
  // field are built in writeDeepCompositeFunc, and they are 
  // called from writeAccessToFields.
  if (typeptr->thisType == t_class) {
    return;
  }


  // -----------------------------------------
  //
  // create the code for the new func declaration
  // and add it to the class (to access the field
  // 'p' e.g.):
  //
  //       void brk_AccessFunc_p();
  //
  //  would be added to the class specified by the
  //  parent parameter.
  //
  sprintf(buf1, "brk_AccessFunc_%s", fld->name);
  tmps = findCompDeclStmt(parent->name, f);
  if (funcDefinedInClass(tmps, buf1)) {
    return;
  }
  addFunctionToClass((SgClassStmt *) tmps, 
		     voidTy, buf1, PUBLIC_FIELD);

  // -----------------------------------------
  //
  // create the code for the new func definition,
  // adding it to the end of the program.  First
  // we create the header.   
  //
  //       void brk_AccessFunc_p() {
  //       }
  //
  fs = findVarDeclSymbFromStmt(buf1, tmps);
  fhs = (SgFuncHedrStmt *) buildFuncDefFromSymb(fs);

  st = lastStatement(f);
  st->insertStmtAfter(*fhs, *(f->firstStatement()));
    

  // -----------------------------------------
  //
  // now add some declarations to the func header:
  //
  //      aPackageStruct * pkg;
  //      char buf[100];
  // 
  ty = new SgPointerType(*(buildDerivedType("aPackage",
					    f)));
  packSym = declareVarInFunc(fhs, ty, "pkg");
  arrayDims[0] = 100;
  sat = (SgArrayType *) buildArrayType(1, arrayDims, charTy);
  bufsym = declareVarInFunc(fhs, sat, "buf");
  bufExp = new SgVarRefExp(*bufsym);

  st = findVarDeclStmt("buf", fhs);
  fexp = new SgExprListExp(*(new 
			     SgArrayRefExp(*bufsym,
					       *(new SgValueExp(100)))));
  st->setExpression(0, *fexp);

  // -----------------------------------------
  //
  // Now, add the executable statements...
  //
  st = fhs->lastDeclaration();
  
  // ---------------------------------;
  //
  // add the func call: 
  //  
  //   pkg = CRG_InitPackage(CRG_shared);
  //
  //
  lst =  addInitCode(f, packSym);
  st->insertStmtAfter(*lst, *fhs);
  st = lst;

  //  check to see if it is an array or not..
  if (at->dims && (at->dims->numDims > 0)) {

    // It is an array

    // ---------------------------------;
    //
    // add new code (e.g., to access int mx[2][3]):
    //
    //   for (int i0 = 0; i < 3; i++) {
    //     for (int i1 = 0; i < 3; i++) {
    //       sprintf(buf, "%d#", mx[i0][i1]);
    //       CRG_Pack(pkg, buf, strlen(buf));
    //     }
    //   }
    //
    arrVar = findMemb(parent->name, fld->name, f);
    arrRef = new SgArrayRefExp(*arrVar);
    tmps = NULL;
    for (i =  0; i < at->dims->numDims; i++) {
      loopstr = new char[3];
      sprintf(loopstr, "i%d", i);
      loopvarsym = declareVarInFunc(fhs, SgTypeInt(), loopstr);
      fstmt = (SgForStmt *) buildForStmt(0, at->dims->sizes[i], loopvarsym, NULL);
      if (tmps) {
	fs2 = (SgForStmt *) tmps;
    	fs2->set_body(*fstmt);
	tmps = (SgStatement *) fstmt;
      } else {
	firstFor = fstmt;
	tmps = (SgStatement *) fstmt;
	st->insertStmtAfter(*tmps);
      }
      /* st = tmps; */
      arrRef->addSubscript(*(new SgVarRefExp(*loopvarsym)));
    }

    // add a function call ;
    lst = buildFuncCallStmt("sprintf", f);
    // add first parameter;
    addParamToFuncCall("buf", charTy, lst);
    // add ac constant value parameter;
    // also put in any separators necessary (try spaces)
    sprintf(buf1, "%s ", TA_typeDescToFormat(typeptr));
    ve = new SgValueExp( strsave(buf1));
    fexp = extractFuncExpFromHedr(lst);  
    addParamExpToFuncExp(fexp, ve);
    // add last parameter;
    addParamExpToFuncExp(fexp, arrRef);
    // replace body..;
    st = lst;
    fstmt->set_body(*st);
    
    // add a function call ;
    lst = addPackCode(f, bufExp, packSym);
    st->insertStmtAfter(*lst, *fstmt);
    st = firstFor->lastNodeOfStmt();
  } else if (at->isPointer) {

    // It is a pointer (that is not an array)

    if ((at->thisType == t_struct) ||
	(at->thisType == t_class)) {
      //I havent handled this case yet, skip it.
      printf("trying to write a functions for a field that is a class, not implemented\n");

      //comp = TA_findComposite(at->name);
      //writeAccessToFields(comp, f, fhs);

    } else {
      
      // It is a non-composite type pointer;
      tmpBuf1 =  TA_typeDescToFormat(at);
 
      if ((strsame(tmpBuf1, "unknown")) || 
	  (strsame(tmpBuf1, ""))) {
	// ---------------------------------;
	//
        // if there is a problem with the format of the data
        // I cant handle it. 
        // 
	printf("Warning:\nthis member: ");
	printf("of type: ");
	printf("cannot be accesses\n----------------------\n\n");
      } else {

	// ---------------------------------;
        //
        // add new code (e.g., to access float *p):
        //
        //       sprintf(buf, "%9.3g#", *p);
        //       CRG_Pack(pkg, buf, strlen(buf));
        //
	lst = buildFuncCallStmt("sprintf", f);
	// add first parameter;
	addParamToFuncCall("buf", charTy, lst);
	// add ac constant value parameter;
	sprintf(buf1, "%s", tmpBuf1);
	ve = new SgValueExp(strsave(buf1));
	fexp = extractFuncExpFromHedr(lst);  
	addParamExpToFuncExp(fexp, ve);
	// add last parameter;
	ty = new SgType(typeEnumToSage(at->thisType));
	fs = findMemb(parent->name, fld->name, f);
	fexp2 = &(SgDerefOp(*(new SgVarRefExp(*fs))));
	addParamExpToFuncExp(fexp, fexp2);
	st->insertStmtAfter(*lst);
	st = lst;
	  
	// add a function call ;
	lst = addPackCode(f, bufExp, packSym);
	st->insertStmtAfter(*lst);
	st = lst;
      }
    }
  }

  // ---------------------------------;
  //
  // add last function call, send package off:
  //
  //     CRG_SendPackage(brk_ConnectionNum, pkg);
  //
  lst = addSendCode(f, packSym);

  st->insertStmtAfter(*lst, *fhs);
  st = lst;

#if DEBUG
  printf("written field access func...\n");
  fhs->unparsestdout();
#endif
}

/* 
 *  SgStatement * writeAccessToFields(TA_composite *comp,
 *			 SgFile *f, SgStatement *fhdr)
 *      It instruments access to each of the fields 
 *      in the class given a function header 
 *      pointer to which the new source code can 
 *      be added to.  It basically packages the data
 *      from all the fields of a class, and sends the
 *      package off.  
 */
SgStatement * writeAccessToFields(TA_composite *comp,
			 SgFile *f, SgStatement *fhdr) {

/* 
 * Given a function header, this function will add 
 * code like the following (given that the class 
 * has two fields in it: int x, *y):
 *
 *   char buf[100];
 *   aPackageStruct *pkg;
 *   *buf = 0;
 *
 *   pkg = CRG_InitPackage(CRG_shared);
 *   CRG_Pack(pkg, "$", strlen("$"));
 *   CRG_Pack(pkg, "#x#", strlen("#x#"));
 *   sprintf(buf, "%10d", x);
 *   CRG_Pack(pkg, buf, strlen(buf));
 *   CRG_Pack(pkg, "#y#", strlen("#y#"));
 *   sprintf(buf, "%10d", y);
 *   CRG_Pack(pkg, buf, strlen(buf));
 *   CRG_Pack(pkg, "$", strlen("$"));
 *   CRG_SendPackage(brk_ConnectionNum, pkg);
 * 
 */
  char *buf1;
  SgStatement *st, *lst;
  SgSymbol *bufsym, *packSym;
  SgExpression *fexp, *nullexp;
  SgType *ty, *charTy;
  SgValueExp *ve;
  TA_field *fld;
  int arrayDims[1];
  TA_typeDesc *ty2;
  SgVarRefExp *bufExp;


  st = fhdr;

  buf1 = new char[100];

  charTy = SgTypeChar();

  // ---------------------------------------
  //
  // Add some declarations:
  //
  //   char buf[100];
  //   aPackageStruct *pkg;
  //   

  ty = new 
    SgPointerType(*(buildDerivedType("aPackage", f)));
  // add new type to func;
  packSym = declareVarInFunc(fhdr, ty, "pkg");
  // get references for global ConnectionNum variable;

  // add array type to func;
  arrayDims[0] = 100;
  ty = buildArrayType(1, arrayDims, charTy);
  
  // add array type to func;
  bufsym = declareVarInFunc(fhdr, ty, "buf");
  st = findVarDeclStmt("buf", fhdr);
  fexp = new SgExprListExp(*(new 
			     SgArrayRefExp(*bufsym,
					   *(new SgValueExp(100)))));
  st->setExpression(0, *fexp);


  // -------------------------------------;
  // Executable statements;

  // --------------------------
  // create the line:
  // 
  //      *buf = 0
  // 
  st = fhdr->lastDeclaration();
  // assign null to strings..;
  nullexp = new SgValueExp((int) NULL);
  lst = buildAssignStmt(&(SgDerefOp(*(new SgVarRefExp(*bufsym)))), nullexp);
  st->insertStmtAfter(*lst);
  st = lst;
  
  // --------------------------
  // create the lines :
  // 
  //       pkg = CRG_InitPackage(CRG_shared);  
  //       CRG_Pack(pkg, "$", strlen("$"));
  //
  lst =  addInitCode(f, packSym);
  st->insertStmtAfter(*lst, *fhdr);
  st = lst;

  ve = new SgValueExp("$");
  lst = addPackCode(f, ve, packSym);
  st->insertStmtAfter(*lst);
  st = lst;

  // --------------------------
  // create the lines for each field in the class (e.g.):
  // 
  //      brk_Pack(pkg, "#start#", strlen("#start#"));
  //      sprintf(buf, "%d", start);
  //      brk_Pack(pkg, buf, strlen(buf));
  //

  bufExp = new SgVarRefExp(*(bufsym));

  fld = comp->fields;
  while (fld) {
    ty2 = fld->type;

    sprintf(buf1, "#%s#", fld->name);
    ve = new SgValueExp(strsave(buf1));
    lst = addPackCode(f, ve, packSym);
    st->insertStmtAfter(*lst);
    st = lst;
    
    // add a function call ;
    lst = buildFuncCallStmt("sprintf", f);
    // add first parameter;
    fexp = extractFuncExpFromHedr(lst);  
    addParamExpToFuncExp(fexp, (SgExpression *)bufExp);
    sprintf(buf1, "%s",  TA_typeDescToFormat(ty2));
    ve = new SgValueExp(strsave(buf1));
    fexp = extractFuncExpFromHedr(lst);  
    addParamExpToFuncExp(fexp, ve);
    // add last parameter;
    ty = new SgType(typeEnumToSage(ty2->thisType));
    addParamToFuncCall(fld->name, ty, lst);
    st->insertStmtAfter(*lst);
    st = lst;
    // put buf in return string;

    lst = addPackCode(f, (SgExpression *) bufExp, packSym);
    st->insertStmtAfter(*lst);
    st = lst;

    fld = fld->next;
  }

  fld = comp->fields;
  while (fld) {
    ty2 = fld->type;



    // ----------------------------------------
    // now, check to see if field is class or struct.
    // we need to create an extra access function for it
    // if so.
    //;  
    if (ty2->thisType  == t_class) {
#if DEBUG
      printf("adding recursive class access function %s in %s\n",
	     ty2->name, comp->name);
#endif

      // ----------------------------------------------;
      //
      // function declaration with arguments:
      //
      // void test_elem::brk_AccessFunc(char *name)
      //
      sprintf(buf1,"brk_AccessFunc");

      // find the statement where class is defined.
      lst = (SgClassStmt *) findCompDeclStmt(ty2->name, f);
#if DEBUG
      printf("added brk_AccessFunc decl to class %s\n",
	     ty2->name);
#endif
      // add the access function declaration to the class;
      if (!(funcDefinedInClass(lst, buf1))) {
	addFunctionToClass((SgClassStmt *) lst, SgTypeVoid(), 
			   buf1, 
			   PUBLIC_FIELD); 
	
	writeElementAccessFunction(f, TA_findComposite(ty2->name));
	writeDeepCompositeAccessFunc(f, 
				     comp->name,
				     ty2->name,
				     fld->name);
      }
    } else if (ty2->isPointer || (ty2->dims && (ty2->dims->numDims > 0))) {
      // ----------------------------------------
      // now, check to see if field is array, or pointer type.
      // we need to create an extra access function for it
      // if so.
      //;  
      writePtrFieldAccessFunc(f, comp, fld);
    }


    fld = fld->next;
  }

  // ----------------------------------------
  // lastly, send the package off:
  //
  //   CRG_Pack(pkg, "$", strlen("$"));
  //   CRG_SendPackage(pkg);
  //;

  ve = new SgValueExp("$");
  lst = addPackCode(f, (SgExpression *) ve, packSym);
  st->insertStmtAfter(*lst);
  st = lst;


  // ---------------------------------;
  //
  // add last function call, send package off:
  //
  //     CRG_SendPackage(brk_ConnectionNum, pkg);
  //
  lst = addSendCode(f, packSym);
  st->insertStmtAfter(*lst, *fhdr);
  st = lst;

#if DEBUG
  fhdr->unparsestdout();
#endif

  return st;
}

SgStatement *writeFieldSelections(TA_composite *comp, SgStatement *fhs,
				  SgStatement *st, SgFile *f,
				  SgSymbol *elemsym) {
  TA_field *af;
  SgStatement *lst;
  SgSymbol *nameVar;

  // ---------------------------------------------------
  // ---------------------------------------------------
  // 
  // Continue code: 
  //
  //   name = trim(fname);
  //  
  //   if (*name == '\0') {
  //     elemptr->accessFunc();
  //   }
  //   if (strsame(name, "x")) {
  //     elemptr->accessFunc_x();
  //   }
  //   if (strsame(name, "y")) {
  //     elemptr->accessFunc_y();
  //   } 
  //
  //
  //
  // now loop through fields, checking the
  // 'name' parameter against each field name.
  //
  
  // add name var to func;
  nameVar = declareVarInFunc(fhs, new SgPointerType(SgTypeChar()) ,
			     "name");
  if (st == fhs) {
    st = fhs->lastDeclaration();
  }

  af = comp->fields;
  while (af) {
    // check if the field is pointer or array type;
    if (af->type->isPointer || (af->type->dims && 
				(af->type->dims->numDims > 0))) {
      addPtrSelectionCond(f, af->name, "brk_AccessFunc", comp, 
			       fhs, elemsym, st);
    }
    af = af->next;
  }
  // check to see if name is NULL, call
  // access func for entire element if so.
  lst = addPtrSelectionCond(f, "", "brk_AccessFunc", comp,
			   fhs, elemsym, st);
  

  addTrimCode(f, "fname", nameVar, fhs);
  return lst;
}


/*  
 * void writeDeepCompositeAccessFunc(SgFile *f, 
 *				  char *parentClass,
 *				  char *childClass,
 *				  char *childName) 
 *     Creates the sage representation for a function
 *     that will allow access to a class declaration 
 *     within another class.
 */
void writeDeepCompositeAccessFunc(SgFile *f, 
				  char *parentClass,
				  char *childClass,
				  char *childName) {

/*
 * The function this will eventually create 
 * given the follwing conditions:
 *
 *  - parentClass name is "simple_elem" 
 *  - childClass name is "test_class"
 *  - member name of childClass "tstclass"
 *  - tstElem is a class with two fields: {int i,float f1[10]}
 * 
 * will look something like this   :
 *
 * void simple_elem::brk_AccessFunc_simple_elem_tstclass(char *name) {
 *                                       
 *   // - name is the character string description
 *   //    of the field requested.  If field is 
 *   //    is NULL, then return whole structure.
 *
 *   char *name;
 *   name = trimLeadingSpaces(fname);
 *   if (strsame(name, "")) {
 *     tstclass->accessFunc();
 *   }
 *   if (strsame(name, "i")) {
 *     tstclass->accessFunc_i();
 *   }
 *   if (strsame(name, "f1")) {
 *     tstclass->accessFunc_f1();
 *   }
 *     
 * }   
 *
 */ 
  char buf1[50];
  SgStatement *st;
  SgFuncHedrStmt *fhs;
  SgType *charPtr;
  SgSymbol *fs;
  SgVariableSymb *ps;
  SgExpression *exp;

  // ----------------------------------------------;
  //
  // function declaration with arguments:
  //
  // void simple_elem::brk_AccessFunc_simple_elem_tstclass(char *name)
  //
  sprintf(buf1,"brk_AccessFunc_%s__%s", parentClass,
	  childClass);

  if ((st = findFuncHedrStmt(buf1, f)) != NULL) {
    // this function has already been defined, return
    return;
  }

  // find the statement where class is defined.
  st = (SgClassStmt *) findCompDeclStmt(parentClass, f);
  // add the access function declaration to the class;
  if (funcDefinedInClass(st, buf1)) {
    return;
  }
  exp = addFunctionToClass((SgClassStmt *) st, SgTypeVoid(), 
			   buf1, 
			   PUBLIC_FIELD); 
  
  ps = new SgVariableSymb("fname",*(new SgPointerType(SgTypeChar())));
  addParamToFuncExp(exp, ps);

  // ----------------------------------------------;
  //
  // Note that this member function has already been 
  // declared above, we are defining the function here. 
  // (e.g. for Vector class):
  //
  // void Vector::brk_AccessFunc();
  //
  fs = findVarDeclSymbFromStmt(buf1, st);

  fhs = (SgFuncHedrStmt *) buildFuncDefFromSymb(fs);
  st = lastStatement(f);
  st->insertStmtAfter(*fhs, *(f->firstStatement()));


  charPtr = new SgPointerType(SgTypeChar());
  
  // add parameters to func hedr;
  // the field name (could be NULL);
  addParamToFuncStmt("fname", charPtr, fhs);

  //
  // ----------------------------------------------;
  
  writeFieldSelections(TA_findComposite(childClass), 
		       fhs,
		       fhs, 
		       f,
		       findMemb(parentClass, childName, f));



}


/*  
 * void writeCollectionAccessFunc(SgFile *f, 
 *				  TA_composite *inComp,
 *				  char *elementType)
 *     Creates the sage representation for a function
 *     that will allow access to a collection's 
 *     elements.  
 */
void writeCollectionAccessFunc(SgFile *f, 
				       char *compName,
				       char *elementName) {
/*
 * The function this will eventually create 
 * given the follwing conditions:
 *
 *   - Collection name is "DistVector" 
 *  - Element name is "s_elem"
 *  - s_elem is a class with two fields: {int x,y}
 * 
 * will look something like this   :
 *
 * void brk_AccessFuncDistVector__s_elem(void *objPtr, 
 *                                       int elemNum, 
 *                                       char *name) {
 *   // - objPtr points to the collection;
 *   // - elemNum is the element number to access;
 *   // - name is the character string description
 *   //    of the field requested.  If field is 
 *   //    is NULL, then return whole structure.
 *   DistVector <s_elem> *castedPtr;
 *    s_elem *elementPtr;
 *
 *   castedPtr = (DistVector *collPtr)->GetElement(eNum);
 *
 *   if (*name == NULL) {
 *     castedPtr->accessFunc();
 *   }
 *   if (strsame(name, "x")) {
 *     castedPtr->accessFunc_x();
 *   }
 *   if (strsame(name, "y")) {
 *     castedPtr->accessFunc_y();
 *   }
 *     
 * }   
 *
 */ 
  TA_composite *comp;
  char buf1[50];
  SgStatement *st, *lst; 
  SgSymbol *vs, *castsym, *elemsym, *objsym, *elemnumsym;
  SgPointerType *pt;
  SgExpression *fexp, *fexp2;
  SgFuncHedrStmt *fhs;
  SgType *charPtr, *voidTy, *elemType, *intty;

  // ----------------------------------------------;
  //
  // function declaration with arguments:
  //
  // void brk_AccessFuncDistVector__s_elem(void *objPtr, 
  //                                       int elemNum, 
  //                                       char *name) {
  //
  sprintf(buf1,"brk_AccessFunc%s__%s", compName,
	  elementName);
  charPtr = new SgPointerType(*(SgTypeChar()));
  voidTy = SgTypeVoid();
  fhs = buildFuncDefStmt(buf1, voidTy);

  st = lastStatement(f);
  st->insertStmtAfter(*fhs, *(f->firstStatement()));
  
  // add parameters to func hedr;
  pt = new SgPointerType(*SgTypeVoid());
  fexp2 = addParamToFuncStmt("objPtr", pt, fhs);
  intty = SgTypeInt();
  // element number parameter;
  fexp2 = addParamToFuncStmt("elemNum", intty, fhs);
  // the field name (could be NULL);
  fexp2 = addParamToFuncStmt("fname", charPtr, fhs);

  
  //
  // ----------------------------------------------;
  
  
  // ---------------------------------------------------
  // ---------------------------------------------------
  // 
  // Declarations...:                       
  //
  //   DistVector *castedPtr;
  //   s_elem *elementPtr;
  //

  //  find the TA_composite representing the element type.
  comp = TA_findComposite(elementName);

  // -------------------------------------;
  // declare the collection pointer
  // find the correct collection symbol;
  vs = findCollectionSymbol(compName, f);
  // and build a collection pointer;
  elemType = buildDerivedType(elementName, f);
#if DEBUG      
  printf("trying to make %s colleciton<%s> (%s)\n", vs->identifier(), 
	 elementName, compName);
#endif
  SgDerivedCollectionType *ctype;
  ctype = new SgDerivedCollectionType(*vs, *(vs->type()));
  ctype->setElementClass(*elemType);
  pt = new 
    SgPointerType(*(ctype));

  // -------------------------------------;
  // add a declaration to the function;
  castsym = declareVarInFunc(fhs, pt, "castedPtr");
  // add a declaration of the elementPtr to the function
  elemsym = declareVarInFunc(fhs, new SgPointerType(*elemType), 
			     "elementPtr");
  

  // ---------------------------------------------------
  // ---------------------------------------------------
  // 
  // Begin code segment after last declaration:
  //
  // castedPtr = (DistVector *) objPtr ;
  // castedPtr = (DistVector *collPtr)->GetElement(eNum);
  //
  //


  // ---------------------------------;
  //  get the last line of declarations in this function
  st = fhs->lastDeclaration();

  // ---------------------------------;
  // cast objPtr to the casted ptr;
  objsym = searchForParamInFuncStmt(fhs, "objPtr");
  lst  = composeCastAssignStatement(castsym, objsym);
  st->insertStmtAfter(*lst);
  st = lst;
  // ---------------------------------;
  // retrieve element;
  elemnumsym = searchForParamInFuncStmt(fhs, "elemNum");
  fexp = buildCollGetElemExp(&(SgDerefOp(*(new SgVarRefExp(*castsym)))), 
			     new SgVarRefExp(*elemnumsym));
  fexp2 = new SgVarRefExp(*elemsym);
  lst = buildAssignStmt(fexp2, fexp);
  st->insertStmtAfter(*lst);
  st = lst;
  

  // ---------------------------------------------------
  // the next function call adds the remaining code of this
  // function.
  // ---------------------------------------------------
  //  
  //   if (*name == '\0') {
  //     elemptr->accessFunc();
  //   }
  //   if (strsame(name, "x")) {
  //     elemptr->accessFunc_x();
  //   }
  //   if (strsame(name, "y")) {
  //     elemptr->accessFunc_y();
  //   } 
  //
  writeFieldSelections(comp, fhs, st, f, elemsym);

}



/*  
 * SgStatement * writeElementAccessFunction(SgFile *f, 
 *					 TA_composite *comp)
 *     Creates the sage representation for a function
 *     that provides access to a class used
 *     as an element of a collection or by an other 
 *     element class.
 */
SgStatement * writeElementAccessFunction(SgFile *f, 
					 TA_composite *comp) {
/*
 * The function this will eventually create 
 * given the follwing conditions:
 *
 *  - Element name is "s_elem"
 *  - s_elem is a class with two fields: {int x, *y}
 * 
 * will look something like this   :
 *
 * void s_elem::brk_AccessFunc() {
 *   char buf[100];
 *   aPackageStruct *pkg;
 *   *buf = 0;
 *
 *   pkg = CRG_InitPackage(CRG_shared);
 *   CRG_Pack(pkg, "$", strlen("$"));
 *   CRG_Pack(pkg, "#x#", strlen("#x#"));
 *   sprintf(buf, "%10d", x);
 *   CRG_Pack(pkg, buf, strlen(buf));
 *   CRG_Pack(pkg, "#y#", strlen("#y#"));
 *   sprintf(buf, "%10d", y);
 *   CRG_Pack(pkg, buf, strlen(buf));
 *   CRG_Pack(pkg, "$", strlen("$"));
 *   CRG_SendPackage(brk_ConnectionNum, pkg);
 * 
 * }   
 *
 *
 *  Note that the brunt of this work is done by 
 *  another function, writeAccessToFields()
 */ 
  SgStatement *st, *tmps;
  SgSymbol *fs;
  SgFuncHedrStmt *fhs;

  // ----------------------------------------------;
  //
  // Note that this member function has already been 
  // declared, we are defining the function here. 
  // (e.g. for Vector class):
  //
  // void Vector::brk_AccessFunc();
  //

  // need to get correct symbol for brk_AccessFunc???
  tmps = findCompDeclStmt(comp->name, f);
  fs = findVarDeclSymbFromStmt("brk_AccessFunc", tmps);

  fhs = (SgFuncHedrStmt *) buildFuncDefFromSymb(fs);
  st = lastStatement(f);
  st->insertStmtAfter(*fhs, *(f->firstStatement()));

  // ----------------------------------------------;
  //
  // write out the fields of this TA_composite.;
  //
  st = (SgStatement *) writeAccessToFields(comp, f, fhs);

  return fhs;
}


/*  
 * void writeUserDefAccessFunc(SgFile *f,
 *                             char *compName,
 *   		               char *elementName);
 *
 *     Creates the sage representation for a function
 *     that will allow access to user defined collection
 *     element member functions.  (We cant allow this with
 *     just collection member functions because that would
 *     be nested parallelism).
 */
void writeUserDefAccessFunc(SgFile *f,
			    char *compName,
			    char *elementName) {
/*
 * The function this will eventually create 
 * given the follwing conditions:
 *
 *   - Collection name is "DistVector" 
 *  - Element name is "s_elem"
 *  - s_elem is a class with two fields: {int x,y}
 * 
 * will look something like this   :
 *
 *   void brk_UserDef_DistVector__s_elem(void *objPtr, 
 *                                       int elemNum, 
 *                                       char *fname) {
 *   // - objPtr points to the collection;
 *   // - elemNum is the element number to access;
 *   // - name is the character string description
 *   //    of the field requested.  If field is 
 *   //    is NULL, then return whole structure.
 *   DistVector *castedPtr;
 *   s_elem *elementPtr;
 *
 *   castedPtr = (DistVector *collPtr)->GetElement(eNum);
 *
 *   if (strsame(membNname, "Sum")) {
 *     castedPtr->UserDefined_Sum();
 *   }
 * }   
 *
 */ 
  TA_composite *comp;
  SgStatement *st, *lst;
  SgSymbol *vs, *castsym, *objsym,
	        *elemnumsym, *elemsym, *nameVar ;
  SgClassSymb *clSym;
  SgMemberFuncSymb *membSym;
  SgPointerType *pt;
  char buf1[50];
  int i, len;
  SgType *charPtr, *voidTy, *elemType, *intty;
  SgFuncHedrStmt *fhs;
  SgExpression *fexp, *fexp2;

  // ----------------------------------------------;
  //
  // function declaration with arguments:
  //
  // void brk_UserDef_DistVector__s_elem(void *objPtr, 
  //                                       int elemNum, 
  //                                       char *name) {
  //
  sprintf(buf1,"brk_UserDef_%s_%s", compName,
	  elementName);
  charPtr = new SgPointerType(*(SgTypeChar()));
  voidTy = SgTypeVoid();
  fhs = buildFuncDefStmt(buf1, voidTy);

  st = lastStatement(f);
  st->insertStmtAfter(*fhs, *(f->firstStatement()));
  
  // add parameters to func hedr;
  pt = new SgPointerType(*SgTypeVoid());
  fexp2 = addParamToFuncStmt("objPtr", pt, fhs);
  intty = SgTypeInt();
  // element number parameter;
  fexp2 = addParamToFuncStmt("elemNum", intty, fhs);
  // the field name (could be NULL);
  fexp2 = addParamToFuncStmt("fname", charPtr, fhs);

  // ---------------------------------------------------
  // 
  // Declarations...:                       
  //
  //   DistVector *castedPtr;
  //   s_elem *elementPtr;
  //
  //  find the TA_composite representing the element type.
  comp = TA_findComposite(elementName);
  vs = findCollectionSymbol(compName, f);
  // and build a collection pointer;
  elemType = buildDerivedType(elementName, f);
#if DEBUG      
  printf("trying to make user def %s colleciton<%s> (%s)\n", vs->identifier(), 
	 elementName, compName);
#endif
  SgDerivedCollectionType *ctype;
  ctype = new SgDerivedCollectionType(*vs, *(vs->type()));
  ctype->setElementClass(*elemType);
  pt = new 
    SgPointerType(*(ctype));
  // add a declaration to the function;
  castsym = declareVarInFunc(fhs, pt, "castedPtr");
  // add a declaration of the elementPtr to the function
  elemsym = declareVarInFunc(fhs, new SgPointerType(*elemType), 
			     "elementPtr");
  
  

  // add name var to func;
  nameVar = declareVarInFunc(fhs, new SgPointerType(SgTypeChar()) ,
			     "name");

  // ---------------------------------------------------
  // 
  // Begin code segment after last declaration:
  //
  // castedPtr = (DistVector *) objPtr ;
  // elemenPtr = (DistVector * collPtr)->GetElement(eNum);
  //
  //
  st = fhs->lastDeclaration();
  // cast objPtr to the casted ptr;
  objsym = searchForParamInFuncStmt(fhs, "objPtr");
  lst  = composeCastAssignStatement(castsym, objsym);
  st->insertStmtAfter(*lst);
  st = lst;
  // retrieve element;
  elemnumsym = searchForParamInFuncStmt(fhs, "elemNum");
  fexp = buildCollGetElemExp(&(SgDerefOp(*(new SgVarRefExp(*castsym)))), 
			     new SgVarRefExp(*elemnumsym));
  fexp2 = new SgVarRefExp(*elemsym);
  lst = buildAssignStmt(fexp2, fexp);
  st->insertStmtAfter(*lst);
  st = lst;
  
  // ---------------------------------------------------
  // 
  // Continue code: 
  //
  //  
  //   if (strsame(fname, "Sum")) {
  //     castedPtr->UserDefined_Sum();
  //   } 
  //
  //
  // loop through fields, checking the
  // 'name' parameter against each 
  //  member func field name.
  //

  lst = findCompDeclStmt(elementName, f);
  if (clSym = isSgClassSymb(lst->symbol())) {
    i = 0;
    while (i <= clSym->numberOfFields()) {
      if (membSym = isSgMemberFuncSymb(clSym->field(i))) {
	len = strlen(USR_FUNC_STR);
	if (strncmp(membSym->identifier(), USR_FUNC_STR, len) == 0) {
	  // check if the field is pointer or array type;
	  st = addPtrSelectionCond(f, (membSym->identifier()+len), 
				   "UserDefined",
				   comp, fhs, elemsym, st);
	}
      }
      i++;
    }
  }


  addTrimCode(f, "fname", nameVar, fhs);
  
}



/*  
 * void addAccessFuncs(SgFile *f)
 *
 *     Adds all the functions necessary for accessing
 *     data in the parallel data structures.
 *
 */
void addAccessFuncs(SgFile *f) {
  
  SgClassStmt *st;
  SgType *ty;
  TA_nameEntry *elem, *coll;
  ty = SgTypeVoid();
  
  // ----------------------------------;
  //
  // For each name in the collection name queue
  // and corresponding name in the element name
  // queue, we need to write out access functions
  // to get to the data during runtime.
  //
  coll = CollectionQueue->first; 
  elem = ElementQueue->first;
  while (elem && coll) {
    // find the statement where class is defined.
    st = (SgClassStmt *) findCompDeclStmt(elem->name, f);
    // add the access function declaration to the class;
    if (!(funcDefinedInClass(st, "brk_AccessFunc"))) {
      addFunctionToClass((SgClassStmt *) st, ty, 
			 "brk_AccessFunc", 
			 PUBLIC_FIELD); 
      // add the sage representation code for the member function
      // to access the elements data
      writeElementAccessFunction(f, TA_findComposite(elem->name));
      // write the sage representation code for accessing a 
      // particular element of a collection
    }
    writeCollectionAccessFunc(f, coll->name,
			      elem->name); 
    writeUserDefAccessFunc(f, coll->name,
			   elem->name); 
    elem = elem->next;
    coll = coll->next;
  }
}



/*  
 * SgStatement *addFuncTableEntry(SgFile *f, 
 *				 char *funcCharName,
 *				 char *funcPtrName)
 *
 *     Creates the sage representation of a 
 *     function call that initializes one entry
 *     of a table that keeps track of string names
 *     and the corresponding function pointers.
 *
 */
SgStatement *addFuncTableEntry(SgFile *f, 
			       char *funcCharName,
			       char *funcPtrName,
			       int funcType) {
  SgStatement *lst;
  SgValueExp *ve; 
  SgExpression *fexp, *fexp2;
  SgSymbol *fs;
  char fcallBuf[40];

  // ----------------------------------------------
  //
  // add the code for a function table entry:
  //
  //      TA_putAcessFunc("Collection DistVector <s_elem>",
  //              brk_AccessFuncDistVector__s_elem);
  //
#if DEBUG      
  printf("I want to add an entry putfunc(\"%s\", %s)\n",
	 funcCharName,
	 funcPtrName);
#endif
  if (funcType == ACC_FUNC) {
    sprintf(fcallBuf, "TA_putAccessFunc");
  } else {
    // must be a USR_FUNC
    sprintf(fcallBuf, "TA_putUserFunc");
  }

  lst = buildFuncCallStmt(fcallBuf, f);

  // add a constant value parameter;
  fexp = extractFuncExpFromHedr(lst);  
  ve = new SgValueExp(strsave(funcCharName));
  addParamExpToFuncExp(fexp, ve);

  // add last parameter;
  fs = findFunctionSymb(funcPtrName, f);
  if (fs == NULL) {
    printf("Warning: Couldn't find function sybmol: %s \n", funcPtrName);
    return NULL;
  }
  fexp2 = new SgVarRefExp(*fs);
  addParamExpToFuncExp(fexp, fexp2);

  return lst;
}


/*  
 * void writeAccessFuncTable(SgFile *f)
 *
 *     Creates the sage representation of a 
 *     function that initializes a table of
 *     collection type names and the corresponging
 *     function that retrieves the data for that 
 *     collection.  This table will also include any
 *     user defined functions that begin with 
 *     whatever is in the USR_FUNC_STR macro (
 *     see top of this file).  The string follwing user
 *     defined in the function name will be used as
 *     the name entry in the table.
 *
 */
void writeAccessFuncTable(SgFile *f) {

/* Will create a function that puts all
 * collection function-collection function
 * name pairs into a storage databank for
 * runtime access.
 *
 * For only one collection-element pair, 
 * the function would look something like:
 *  
 *    void brk_AccessFunctionEntryInit() {
 *      TA_putAccessFunc("Collection DistVector <s_elem>",
 *              brk_AccessFuncDistVector__s_elem);
 *
 *    }
 *
 */
  TA_nameEntry *coll, *elem;
  char *buf1, funcCharName[100], funcPtrName[50];
  SgStatement *st, *lst, *sttmp;
  SgFuncHedrStmt *fhs, *fhdr;
  SgSymbol *fs;
  SgMemberFuncSymb *mfs;

  buf1 = new char[100];
  
  // ---------------------------------;
  //
  // create the function header:
  // 
  //    void brk_AccessFunctionEntryInit() {
  //    }
  //

  sprintf(buf1, "brk_AccessFunctionTableInit");
  
  fhs = buildFuncDefStmt(buf1, SgTypeVoid());
  st = lastStatement(f);
  st->insertStmtAfter(*fhs, *(f->firstStatement()));
  
  // ---------------------------------;
  //
  // Add the table initialization calls
  // 
  //      TA_putAccessFunc("Collection DistVector <s_elem>",
  //              brk_AccessFuncDistVector__s_elem);
  //
  st = fhs->lastDeclaration();
  coll = CollectionQueue->first;
  elem = ElementQueue->first;
  while (coll && elem) {
#if DEBUG      
    printf("Creating function entry for %s elem %s\n",
	   coll->name, elem->name);
#endif
    sprintf(funcCharName, "Collection %s <%s>", coll->name, elem->name);
    sprintf(funcPtrName, "brk_AccessFunc%s__%s", coll->name, elem->name);
    lst = addFuncTableEntry(f, funcCharName, funcPtrName, ACC_FUNC);
    if (lst) {
      st->insertStmtAfter(*lst);
      st = lst;
    }

    coll = coll->next;
    elem = elem->next; 
  }

  sttmp = f->firstStatement();
  int len;
  len = strlen(USR_FUNC_STR);
  while (sttmp) {
    if (fhdr = isSgFuncHedrStmt(sttmp)) {
      fs = fhdr->symbol();
      buf1 = fs->identifier();
      if (strncmp(buf1, USR_FUNC_STR, len) == 0) {
	buf1 += len;
	if (mfs = isSgMemberFuncSymb(fs)) {
	  //	  must put in each case of the collection/element pairs for the table 
          coll = CollectionQueue->first;
	  elem = ElementQueue->first;
	  while (coll && elem) {
	    if (strsame(elem->name, mfs->className()->identifier())) {
	      sprintf(funcCharName, "%s <%s>::%s", coll->name, elem->name, buf1);
	      sprintf(funcPtrName, "brk_UserDef_%s_%s",
		      coll->name, elem->name);
	      lst = addFuncTableEntry(f, funcCharName, funcPtrName, ACC_FUNC);
	      st->insertStmtAfter(*lst);
	      st = lst;
	    }
	    coll = coll->next;
	    elem = elem->next; 
	  }
	} else {
	  strcpy(funcPtrName, fs->identifier());
	  strcpy(funcCharName, buf1);
	  lst = addFuncTableEntry(f, funcCharName, funcPtrName, USR_FUNC);
	  st->insertStmtAfter(*lst);
	  st = lst;
	}
      }
    }
    sttmp = sttmp->lexNext();
  }
}


/*  
 * void brk_getCollections (SgFile *aFile) {
 *
 *     gets all the collection information set
 *     up for data access during runtime.
 *     This function is called from instr.C
 *
 */
void brk_getCollections (SgFile *aFile) {
  SgSymbol *classsymb;
  char *classname;
  SgStatement *s;
  SgType *elemType;
  SgDerivedCollectionType *d;
  char *elemname;
  SgCExpStmt *ass;
  SgVarDeclStmt *var;
  SgSymbol *varsymb;
  SgExprListExp *elist;
  SgFile *f;

  //  ---------------------------------------------
  // 
  // Loops through all statements, checking for collection
  // declarations.  These can occur in two places: 
  // 
  //   (1) in a variable declaration statement
  //   (2) in a "new" statement
  // 
  // So, we must check for both cases.  Once we have found
  // a collection, we add it and its element class to queues
  // for keeping track of each type respectively.  These element
  // queues build all the type representation necessary for 
  // creating access functions later on (after the loop).
  // 
  f = aFile;
  s = aFile->firstStatement();
  while ( s ) {
    if ( var = isSgVarDeclStmt (s) ) {
      for (int i=0; i<var->numberOfSymbols(); i++) {
	varsymb = var->symbol(i);
	if ( d = isSgDerivedCollectionType (varsymb->type()) ) {
	  /* -- Symbol is of type CollectionType<ElementType>    -- */
	  /* -- get type names                                   -- */
	  /* ------------------------------------------------------ */
	  classsymb = d->collectionName();
	  elemType  = d->elementClass();
	  classname = classsymb->identifier();
	  elemname  = (elemType->symbol())->identifier();
	  addElemAndCollection(classname, elemname, 
			       aFile);
	}
      }
    } else if ( (ass=isSgCExpStmt (s)) && 
	       (elist=isSgExprListExp(ass->expr()))) {
      SgExpression *e;
      SgNewExp *n;
      SgPointerType *t;
      
      for (int i=0; i<elist->length(); i++) {
	e = elist->elem(i);
	if ( (e->variant() == ASSGN_OP) &&
	    (n = isSgNewExp(e->rhs())) &&
	    (t = isSgPointerType(n->type())) &&
	    (d = isSgDerivedCollectionType(t->baseType()))
	    )  {
	  /* -- dynamic collection allocation Statement:         -- */
	  /* -- var = new CollectionType<ElementType>()          -- */
	  /* ------------------------------------------------------ */
	  classsymb = d->collectionName();
	  elemType  = d->elementClass();
	  classname = classsymb->identifier();
	  elemname  = (elemType->symbol())->identifier();
	  addElemAndCollection(classname, elemname, 
			       aFile);
	}
      }
    }
    s = s->lexNext();
  }

  //  ---------------------------------------------
  // 
  //     Adds all the functions necessary for accessing
  //     data in the parallel data structures.
  // 
  addAccessFuncs(f);

  //  ---------------------------------------------
  // 
  //     Adds a function that will initialize a table
  //     of function names and the function pointers
  //     associated with them.
  // 
  writeAccessFuncTable(f);
#if DEBUG
  aFile->saveDepFile("tempfile.dep");
#endif  
}


/*  
 *  void brk_printTIFFile()
 *
 *    Writes out a Type Information File that will
 *    be read in by the parallel program during startup
 *    phase.  This type information will then be given
 *    to any applications that will be connecting to the
 *    parallel program for runtime access to the parallel
 *    data.
 *
 */
void brk_printTIFFile() {
  int i;
  TA_composite *tmpc;
  TA_typeDesc *tmpt;
  char *buf;
  char fname[20];
  FILE *fptr;

  //  ---------------------------------------------
  // 
  // Open the file "TypeInfoFile.TIF"
  // 
  strcpy(fname, "TypeInfoFile.TIF");
  fptr = fopen(fname, "w");
  if (!fptr) {
    printf("file oepn of %s failed\n", fname);
    exit(1);
  }

  //  ---------------------------------------------
  // 
  //  write out all the types needed for accessing 
  //  parallel data at runtime.
  // 
  //  fprintf(fptr, "/*\n\n All Collections & classes which are elements of \n");
  //  fprintf(fptr, " a distributed collection \n\n*/\n");
  //  First, the collections and element classes they use
  i = 0;
  tmpc = TA_CompositeList[0];
  while (tmpc != NULL) {
    buf = TA_writeComposite(tmpc);
    fprintf(fptr, buf);
    fprintf(fptr, " \n");
    free(buf);
    i++;
    tmpc = TA_CompositeList[i];
  }
  // then, all the type descriptions used by the 
  // element classes that are non-primitive (but non-composite)
  // data types. 
  i = 0;
  tmpt = TA_typeDescList[0];
  while (tmpt != NULL) {
    buf = TA_writeTypeDesc(tmpt);
    fprintf(fptr, buf);
    fprintf(fptr, " ");
    free(buf);
    i++;
    tmpt = TA_typeDescList[i];
  }

}



