// c f g . C
// $Id: cfg.C,v 1.5 1996/11/11 18:34:16 kurtw Exp $
// Ask: lth@cs.uoregon.edu
//
// Copyright (C) 1995 The University of Oregon.
//
// Intra-procedural control-flow-graph analyzer for TAU.
// Supports C++ code only, right now.
//
// ***WARNING. This code is not reentrant. Avoid parallelism :-)
//
// Basic idea: There is a procedure compute_cfg() which takes a pointer
// to an SgFuncHedrStmt and returns a control-flow-graph (CFG) for that
// function's body. The graph consisists of CFGs linked by CFGEs (CFG edges).
// Each CFG contains a linked list of 'labels': a textual representation
// of the statements in that basic block[*]. Each CFGE contains one label:
// a textual representation of the condition for the control transfer along
// that edge. Unconditional transfers have the label U, IF-THEN transfers
// have the labels T and F for the branches, and switches are labelled with
// case values as much as possible.
//
// [*] 'Basic block' in an extended sense, where implicit branching in 
//     the source -- as in short-circuit boolean evaluation -- is not
//     taken into account.
//
// Bugs, etc.
//   The 'parentcount' field really serves two purposes: count of in-edges,
//   and count of actual parents in a graph with all backedges removed.
//   It might be useful to separate these two uses into separate variables.

#include <stdlib.h>
#include <stdio.h>
#include "sage++user.h"
#include "libSage++.h"
#include "cfg.h"

#define USE_SUNPARSE 1

inline int max( int a, int b ) { return (a > b ? a : b); }

///////////////////////////////////////////////////////////////////////////
// CFGE members

CFGE::CFGE( CFG *to, const char *l, CFGE *n )
{
  sink = to;
  next = n;
  label = new CFGTEXT( l, 0 );
}

CFGE::~CFGE()
{
  delete next;
  delete label;
}

///////////////////////////////////////////////////////////////////////////
// CFGTEXT members.

CFGTEXT::CFGTEXT( const char *s, CFGTEXT *n )
{
  label = new char[ strlen( s )+1 ];
  strcpy( label, s );
  next = n;
}

CFGTEXT::~CFGTEXT()
{
  fflush(stdout);
  delete [] label;
  if (next) delete next;
}

CFGTEXT::operator==( const char *s )
{
  return strcmp( label, s ) == 0;
}

///////////////////////////////////////////////////////////////////////////
// CFG member functions
//
// There is some magic with keeping track of allocated objects here.
// It works like this: The global `g_gclist' is initially 0. Since this code 
// is not reentrant, we can assume that `g_gclist' is private for each call to 
// compute_cfg(). It therefore contains a list of allocated nodes, linked
// in allocation order throught the member variable `gclink' in struct CFG.
// The last thing compute_cfg() does before returning is setting the `gclist'
// member of the root node to the value of g_gclist and g_gclist to 0
// so that it is ready for the next call. Now, the returned CFG node
// contains a list of all allocated nodes in the tree, whether reachable or
// not through the children lists). When the destructor for CFG is called,
// it deletes the node in the gclist link, eventually freeing all. It also
// deletes each CFGE list by deleting the first node, which cascades.
//
// There are two advantages to doing it this way. First, some nodes may
// be orphaned by the construction of the graph, and would be lost. Second,
// by avoiding the graph structure during destruction, it becomes much
// easier to destroy the graph.

static int nodecount;       // tracks number of created nodes, for labelling
static CFG *g_gclist = 0;   // tracks allocated nodes.

CFG::CFG()
{
  init(); 
}

CFG::CFG( const char *s, const int xtag )
{
  init();
  label( s, xtag );
}

void CFG::init()
{
  nodecount++;
  tag = -1;
  xtag = -1;
  childcount = parentcount = links = mustbranch = 0;
  dumped = 0;
  level = width = 0;
  thelabel = 0;
  children = 0;
  gclink = g_gclist;
  g_gclist = this;
  gclist = 0;
}

void CFG::label( const char *s )
{
  thelabel = new CFGTEXT( s, thelabel );
}

void CFG::label( const char *s, const int xtag )
{
  label( s );
  this->xtag = xtag;
}

int CFG::unlabelled()
{
  return thelabel == 0 || (thelabel->next == 0 && *thelabel == "[dummy]" );
}


// kurtw: fixed the deletion of the CFG:  gclist is set as a pointer to the 
// root of the CFG.  When the entire CFG is deleted, the old code deleting
// gclist starts an infinite loop.  Thus, I now delete gclink, which should
// cascade through ALL nodes of the graph, included unconnected ones.
CFG::~CFG()
{
  if (thelabel) delete thelabel;
  if (children) delete children;
#if 0
  if (gclist) delete gclist; else delete gclink;
#endif
  if (gclink) delete gclink;
}

///////////////////////////////////////////////////////////////////////////
//
// Compute_cfg() computes the control-flow-graph for the given function
// header statement and returns it. 

CFG  *CFG_mknodlist( SgStatement *, CFG *, CFG *, CFG *, CFG *, CFG * );
void CFG_compact( CFG * );
int  CFG_order_nodes( CFG *, int );
void CFG_annotate_nodes( CFG * );
void CFG_label_cleanup();
void CFG_remember_label( SgLabel *, CFG * );
CFG  *CFG_recover_label( SgLabel * );

CFG *compute_cfg( SgFuncHedrStmt *fhst )
{
  CFG *retnode;
  CFG *cfg;

  // First compute the graph.
  nodecount = 0;
  retnode = new CFG( "[end]", L_END );
  cfg = CFG_mknodlist( fhst->childList1(0), retnode, 0, 0, retnode, 0 );

  // Label the flow graph such that parents have lower-valued labels
  // than their children (ignoring backedges). There may be holes in
  // the labelling order.
  CFG_order_nodes( cfg, nodecount-1 );

  // Compact the graph by removing unnecessary edges and vertices.
//  CFG_compact( cfg );

  // There may be more nodes than are reachable (see comments in mknod())
  // so in order to always have the starting node be node 0, label it
  // explicitly. Node numbers aren't encoded in the edges, so the change
  // "propagates" everywhere.
  cfg->tag = 0;

  // Annotate CFG with level and width.
  CFG_annotate_nodes( cfg );

  // Setup for later destructor call; explained above.
  cfg->gclist = g_gclist;
  g_gclist = 0;

  // Tidy up GOTO labels.
  CFG_label_cleanup();

  return cfg;
}


///////////////////////////////////////////////////////////////////////////
// CFG Construction
//
// CFG_mknodlist() builds up a partial graph from a list of statements, 
// and it does it in reverse order.

CFG  *CFG_mknod( SgStatement *, CFG *, CFG *, CFG *, CFG *, CFG * );
CFG  *CFG_mknod2( SgStatement *, CFG *, CFG *, CFG *, CFG *, CFG * );
void CFG_mkedge( CFG *, CFG *, const char * );
const char *CFG_unparse_exp( SgExpression * );
const char *CFG_unparse_stmt( SgStatement * );

/* `child' is a statement in some childList.
 * `cont' is the statement continuation -- the next statement.
 * `contcont' is the block to jump to on a CONTINUE, or 0.
 * `breakcont' is the block to jump to on a BREAK, or 0.
 * `retcont' is the block to jump to on a RETURN.
 * `swnode' is the enclosing switch node, or 0.
 */
CFG *CFG_mknodlist( SgStatement *child, 
		   CFG *cont, CFG *contcont, CFG *breakcont, CFG *retcont,
		   CFG *swnode )
{
  if (child == 0)
    return cont;
  else {
    cont = CFG_mknodlist( child->nextInChildList(),
			  cont, contcont, breakcont, retcont, swnode );
    return CFG_mknod( child, cont, contcont, breakcont, retcont, swnode );
  }
}

CFG *CFG_mknod( SgStatement *stmt,
	        CFG *cont, CFG *contcont, CFG *breakcont, CFG *retcont,
	        CFG *swnode )
{
  CFG *node, *child1, *child2;

  { SgLabelStmt *ls;
    // Hahaha!! You thought you could use stmt->hasLabel(). Hahaha!!
    // (sigh...)
    if (ls = isSgLabelStmt( stmt )) {
      cont->mustbranch = 1;
      CFG_remember_label( ls->label(), cont );
    }
  }
  switch (stmt->variant()) {
    case IF_NODE : 
      child1 = CFG_mknod2( stmt->childList1(0),
			   cont, contcont, breakcont, retcont, swnode );
      child2 = CFG_mknod2( stmt->childList2(0),
			   cont, contcont, breakcont, retcont, swnode );
      node = new CFG( "IF", L_IF );
      CFG_mkedge( node, child1, "T" );
      CFG_mkedge( node, child2, "F" );
      return node;
    case WHILE_NODE :
    case FOR_NODE :
      if (stmt->variant() == FOR_NODE)
	node = new CFG( "FOR", L_FOR );
      else
	node = new CFG( "WHILE", L_WHILE );
      child1 = CFG_mknod2( stmt->childList1(0),
			   node, node, cont, retcont, swnode );
      CFG_mkedge( node, child1, "T" );
      CFG_mkedge( node, cont, "F" );
      return node;
    case DO_WHILE_NODE :
      node = new CFG( "DO-WHILE", L_DO_WHILE );
      child1 = CFG_mknod2( stmt->childList1(0),
			   node, node, cont, retcont, swnode );
      CFG_mkedge( node, cont, "F" );
      CFG_mkedge( node, child1, "T" );
      return child1;
    case SWITCH_NODE :
      node = new CFG( "SWITCH", L_SWITCH );
      child1 = CFG_mknod2( stmt->childList1(0),
			   cont, contcont, cont, retcont, node );
      if (node->tag == -2) {
	// default was given by user
	node->tag = -1;
      }
      else {
	// no default case was given by user
	// FIXME: would be reasonable to add this edge at end.
	CFG_mkedge( node, cont, "<default>" );
      }
      return node;
    case CASE_NODE :
      if (!swnode) {
	fprintf( stderr, "Warning: cfg: illegal use of `case'\n" );
	goto defcase;
      }
      cont->label( "CASE", L_CASE );
      cont->mustbranch = 1;
      CFG_mkedge( swnode, cont,
		 CFG_unparse_exp( isSgCaseOptionStmt( stmt )->caseRange(0)) );
      return cont;
    case DEFAULT_NODE :
      if (!swnode) {
	fprintf( stderr, "Warning: cfg: illegal use of `default'\n" );
	goto defcase;
      }
      cont->label( "DEFAULT", L_CASE );
      cont->mustbranch = 1;
      swnode->tag = -2;  // Magic; see case for SWITCH_NODE above.
      CFG_mkedge( swnode, cont, "<default>" );
      return cont;
    case BREAK_NODE :
      if (!breakcont) {
	fprintf( stderr, "Warning: cfg: illegal use of `break'\n" );
	goto defcase;
      }
      node = new CFG( "BREAK", L_BREAK );
      CFG_mkedge( node, breakcont, "U" );
      return node;
    case CONTINUE_NODE :
      if (!contcont) {
	fprintf( stderr, "Warning: cfg: illegal use of `continue'\n" );
	return cont;
      }
      node = new CFG( "CONTINUE", L_CONTINUE );
      CFG_mkedge( node, contcont, "U" );
      return node;
    case RETURN_NODE :
      node = new CFG( "RETURN", L_RETURN );
      CFG_mkedge( node, retcont, "U" );
      return node;
    case GOTO_NODE :
      { CFG *target;
	target = CFG_recover_label( isSgGotoStmt( stmt )->branchLabel() );
	node = new CFG( "GOTO", L_GOTO );
	CFG_mkedge( node, target, "U" );
	return node;
      }
    case CONTROL_END :
      // explicitly invisible.
      return cont;
    default :
    defcase :
      // Normal statements always equal the continuation, except when 
      // the continuation is a branch target, in which case we must also
      // jump.
      if (cont->mustbranch) {
	node = new CFG;
	CFG_mkedge( node, cont, "U" );
	cont = node;
      }
      cont->label( CFG_unparse_stmt( stmt ) );
      return cont;
  }
}

CFG *CFG_mknod2( SgStatement *slist,
		 CFG *cont, CFG *contcont, CFG *breakcont, CFG *retcont,
		 CFG *swnode )
{
  if (slist != 0) {
    CFG *cc = new CFG;
    CFG_mkedge( cc, cont, "U" );
    return CFG_mknodlist( slist, cc, contcont, breakcont, retcont, swnode );
  }
  else
    return cont;
}

void CFG_mkedge( CFG *from, CFG *to, const char *label )
{
  if (from == 0 || to == 0) {
    fprintf( stderr, "Warning: cfg: invalid edge.\n" );
    return;
  }
  from->children = new CFGE( to, label, from->children );
  from->childcount++;
  to->parentcount++;
}

///////////////////////////////////////////////////////////////////////////
//
// Unparsers which unparse to a string buffer rather than to a file.
// Really, Sage++ should support and do this, right now only expressions
// are supported.

// Unparse expression. Returns a string in a statically allocated buffer.

const char *CFG_unparse_exp( SgExpression *e )
{
  static char buf[ 2000 ];

#if USE_SUNPARSE
  e->sunparse( buf );
  return buf;
#else
  // This is a temporary fix which only handles constants properly.
  SgValueExp *ve;

  while (e && e->variant() == EXPR_LIST) e = e->lhs();
  if (!e)
    return "<expr>";
  else if (ve = isSgValueExp( e )) {
    // SgValueExp is _really_ strange.
    // FIXME: not quite right, as it does not deal with unsigned/long/short
    if (e->variant() == INT_VAL || e->variant() == LONG_INT_VAL) // Unportable
      sprintf( buf, "%d", ve->intValue() );
    else if (e->variant() == FLOAT_VAL)
      strcpy( buf, ve->floatValue() );
    else if (e->variant() == DOUBLE_VAL)
      strcpy( buf, ve->doubleValue() );
    else if (e->variant() == CHAR_VAL) {
      char c = ve->charValue();
      if (c == '\\')
	sprintf( buf, "'\\\\'" );
      else if (c < ' ')
	sprintf( buf, "'\\%03o'", c ); // FIXME: common escapes would be OK
      else
	sprintf( buf, "'%c'", c );
    }
    else
      strcpy( buf, "<expr>" );
    return buf;
  }
  else
    return "<expr>";
#endif
}

// Unparse statement. Returns a string in a statically allocated buffer.
// FIXME: Needs to be much more sophisticated.

const char *CFG_unparse_stmt( SgStatement *st )
{
  static char buf[256];

  switch (st->variant()) {
    case EXPR_STMT_NODE :
      sprintf( buf, "<EXPR>" );
      break;
    case ASSIGN_STAT :
      sprintf( buf, "<ASSIGN>" );
      break;
    default :
      sprintf( buf, "<STMT>" );
  }
  if (isSgVarDeclStmt( st ))
    sprintf( buf, "<DECL>" );
  return buf;
}

///////////////////////////////////////////////////////////////////////////
//
// Handling goto targets  --  much ado about nothing.
//
// recover_label returns the node to which to branch on a GOTO to the
// given label. However, this label might not have been seen yet. If
// so, we return a fictitious node, which will later be patched to goto
// the correct node.
//
// remember_label inserts a ( label, node ) pair in the list of branch
// targets. If the label has been assigned a temporary node, patch the
// temporary node with a jump to the real node, and insert the real
// node in the table.
//
// This simple strategy is nice only if coalescing of nodes is later
// performed.
//
// There are usually very few labels and goto's in code, so a linked
// list is sufficient.

struct CFGL {
  CFGL( SgLabel *, CFG *, CFGL * );
  ~CFGL();
  SgLabel *label;
  CFG *node;
  int fake;
  CFGL *next;
  CFGL *prev;
};

CFGL::CFGL( SgLabel *l, CFG *n, CFGL *p )
{
  fake = 0;
  next = p;
  prev = 0;
  label = l;
  node = n;
  if (p) p->prev = this;
}

CFGL::~CFGL()
{
  if (next) next->prev = prev;
  if (prev) prev->next = next;
}

static CFGL *labels = 0;

void CFG_remember_label( SgLabel *l, CFG *n )
{
  CFGL *p;

//*DEBUG*/ printf( "Remembering label %s\n", l->symbol()->identifier() );
  for ( p = labels ; p != 0 && p->label != l ; p = p->next )
    ;
  if (p == 0)
    labels = new CFGL( l, n, labels );
  else if (p->fake == 0) {
    fprintf( stderr, "Error: cfg: multiple labels: %s\n", 
	     l->symbol()->identifier() );
    exit( 1 );
  }
  else {
    labels = new CFGL( l, n, labels );
    CFG_mkedge( p->node, n, "U" );
    delete p;
  }
}

CFG *CFG_recover_label( SgLabel *label )
{
  CFGL *p;

//*DEBUG*/ printf( "Recovering label %s\n", label->symbol()->identifier() );
  for ( p = labels ; p != 0 && p->label != label ; p = p->next )
    ;
  if (p == 0) {
    labels = new CFGL( label, new CFG, labels ); // [dummy]
    labels->fake = 1;
    return labels->node;
  }
  else
    return p->node;
}

void CFG_label_cleanup()
{
  CFGL *p;

  while (labels) {
    p = labels;
    labels = labels->next;
    delete p;
  }
}

///////////////////////////////////////////////////////////////////////////
//
// CFG compaction
//
// A node X can be removed if it has no label and it has a single out-edge
// labelled "U". All pointers to X should be redirected to X's target.
//
// Unimplemented:
// One can also coalesce two blocks where one has an unconditional branch
// to the other, and that edge is the only edge entering the second block.
// This is not currently done. The way to do that in this framework is
// to move all stmt labels from the latter block into the former so as to make
// the latter unlabelled, and then perform the above form of compaction.
// This is simple to do, but can be confusing: control flow is no longer
// directly related to the source code. Consider this fragment:
//
//   do { if (x > y) x++; else continue; y++ } while (x < 5);
//
// With the second form of compaction, the flow graph will more resemble
// the following (equivalent) statement:
//
//   do { if (x > y) { x++; y++ } else continue; } while (x < 5);
//
// Some experiments should be used here; perhaps this should be a user-
// controllable switch (compact/not). There are other problems, too, like
// what to do with explicit control-flow statements (GOTO, BREAK) which
// trigger this optimization: they will suddenly end up in the middle of
// a block!
//
// [the 'parentcount' field is set up to enable this optimization.]

int CFG_compact2( CFG * );
void CFG_compact( CFG *cfg )
{
  while (CFG_compact2( cfg ))
    ;
}

// FIXME: this code is not right.
int CFG_compact2( CFG *cfg )
{
  int modified = 0;
  CFGE *p;

  if (cfg == 0) return 0;
  if (cfg->tag == -2) return 0;

  cfg->tag = -2;
  for ( p = cfg->children ; p ; p = p->next ) {
    if (p->sink->unlabelled()
     && p->sink->children != 0
     && p->sink->children->next == 0
     && *(p->sink->children->label) == "U") {
      p->sink = p->sink->children->sink;
      modified = 1;
    }
    else
      modified = modified || CFG_compact2( p->sink );
  }
  cfg->tag = -1;
  return modified;
}

///////////////////////////////////////////////////////////////////////////
//
// CFG labelling

// Compute the ordering of the control flow graph by doing a 
// postorder labelling. The resulting labelling will be such that
// parents have lower numbers than their children, and 0 is the unique
// entry node.
//
// Also reset the parentcount to represent the number of parents
// which are not at the source of a backedge.
//
// Node->tag == -1 => it's unvisited
// Node->tag == -2 => it's on the current path
// Node->tag >= 0  => it's been labelled

int CFG_order_nodes( CFG *node, int label )
{
  CFGE *p;

  node->tag = -2;
  node->parentcount = 0;

  for ( p = node->children ; p ; p = p->next ) {
    if (p->sink->tag == -1) label = CFG_order_nodes( p->sink, label );
    if (p->sink->tag != -2) p->sink->parentcount++;
  }

  node->tag = label;
  return label-1;
}

// Label each node with its level and weight, using an O( E ) algorithm.
// The algorithm works by flooding the tree, ignoring backedges, and
// following each edge only once. An edge is followed from a node only when
// all in-edges to the node have been traversed.
//
// Algorithm assumes that parentcount represents the number of parents
// we get by removing all backedges from the graph.

int CFG_annotate_level( CFG *, int );
void CFG_annotate_width( CFG * );

void CFG_annotate_nodes( CFG *node )
{
  node->parentcount = 1;  /* Root node */
  CFG_annotate_level( node, 0 );
  node->parentcount = 0;

  CFG_annotate_width( node );
}

// Annotate the flow graph with the level of each node. Assumes
// that the 'links' fields is 0 for each node, initially, and that
// the parentcount link is correct except for the root node, for
// which it must be set to 1 even though it's 0.
//
// A node is traversed only the last time it is reached; the algorithm
// works by flooding the graph and ignoring backedges.
//
// There are concessions to a "nice" layout in that certain control
// structures are treated specially.
//
// kurtw: I changed it so that the node->level is assigned a value on the
// first visit of the node, and subsequent visits reassign it as the max of
// the levels provided from each parent link.

int CFG_annotate_level( CFG *node, int level )
{
  CFGE *p, *e1, *e2;
  int m, k;

  node->links++;
  m = node->level = level = max(level, node->level);  // kurtw change
  if (node->links != node->parentcount) return 0;

  switch (node->xtag) {
  case L_WHILE :
  case L_FOR :
    /* Do the T branch first and then do the F branch with max+1; this
     * ensures that the F branch ends up "below" all of the T stuff.
     */
    e1 = node->children; e2 = node->children->next;
    if (*(e1->label) == "F") { p = e1; e1 = e2; e2 = p; }

    /* We know that neither edge is a backedge. (What about empty loops?) */
    k = CFG_annotate_level( e1->sink, level+1 );
    m = max( m, k );
    m = max( m, CFG_annotate_level( e2->sink, k+1 ) );
    break;
  case L_SWITCH :
    /* Do all the CASEs first and finally any non-CASE.  This ensures
     * that the fallthru is "below" all the cases.  
     *
     * FIXME? If one case falls through to another, then the latter
     * will be on a lower level.  This is correct, but not necessarily
     * desirable for drawing the graph.
     */
    k = 0;
    for ( p = node->children ; p ; p = p->next )
      if (p->sink->tag > node->tag && p->sink->xtag == L_CASE)
	k = max( k, CFG_annotate_level( p->sink, level+1 ) );
    m = max( m, k );
    for ( p = node->children ; p ; p = p->next )
      if (p->sink->tag > node->tag && p->sink->xtag != L_CASE)
	m = max( m, CFG_annotate_level( p->sink, k+1 ) );
    break;
  default:
    for ( p = node->children ; p ; p = p->next )
      if (p->sink->tag > node->tag)
	m = max( m, CFG_annotate_level( p->sink, level+1 ) );
    break;
  }

  return m;
}


// Annotate the flow graph with the 'width' of the node, which
// is 1 for a leaf or the sum of the widths for children which are
// not along backedges. Assumes that the 'width' field is 0 for each
// node, initially.
//
// This has to be a separate loop from annotate_node() because the
// recursion condition is different.
//
// Unfortunately, this algorithm does not do the right thing because
// separate (narrow) paths to the same (wide) subtree get added together,
// making the width of top-level nodes truly large.
//
// Perhaps a more intuitive notion of 'width' is the maximum number of
// nodes on a given 'level' of the graph, as calculated above.

void CFG_annotate_width( CFG *node )
{
#if 1
  int width = 0;
  CFGE *p;
  
  if (node->width != 0) return;
  for ( p = node->children ; p ; p = p->next ) {
    if (p->sink->tag > node->tag) {
      CFG_annotate_width( p->sink );
      width += p->sink->width;
    }
  }
  node->width = max( width, 1 );
#endif
}

// eof

