/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993                  */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

// May 7, 1996
//
// Rough history:
//  Written by Bernd Mohr.
//  Hacked by Darryl Brown and Lars T. Hansen.
//  Changed ("cleaned up", perhaps) by Lars T. Hansen, January 1996.
//    Major changes: removal of globals, procedurification of duplicated
//                   code, addition of server mode, splitting of main()
//                   and some other large procedures.

# include <stdio.h>
# include <stdlib.h>
# include <signal.h>

# include "sage++user.h"
# include "sage++callgraph.h"
# include "baseClasses.h"
# include "sage++classhierarchy.h"
# include "cfg.h"

# ifndef __GNUC__
#   include <unistd.h>
#else
extern "C" int unlink(const char *path);
extern "C" int access(char *path, int amode);
extern "C" int chdir(char *path);
# endif

# include "compatible.h"
# ifdef SYS5
#   include <string.h>
# else
#   include <strings.h>
# endif

# ifndef R_OK
#   define R_OK 04
# endif

/* -- external functions -------------------------------------------------- */
/* -- from: scanfile.c ---------------------------------------------------- */
extern "C" void openFile (char *path);
extern "C" void closeFile (void);
extern "C" void findFuncOrClass (int line, char **result);
extern "C" void findNextChild (char *name, char **result, int marknewpos);
extern "C" void findNextMember (char *name, int line, char **result);
extern "C" void setNextPos (int line, int pos, int endline);

/* -- from: namecheck.c --------------------------------------------------- */
extern "C" void initCommands (void);
extern "C" int isValidFunc (char *filename, char *classname, char *funcname);

/* -- from: readcomfile.y ------------------------------------------------- */
extern "C" int readComFile (char *filename);

/* -- auxillary functions ------------------------------------------------- */
void printUsage (char* execName)
{
  fprintf (stderr,"Usage: %s\n",execName);
  fprintf (stderr,"  [-help | -h] [-usage | -u] [-debug]\n");
  fprintf (stderr,"  [-verbose | -v] [(-mode | -m) (ascii | vt100 | wyse)]\n");
  fprintf (stderr,"  [(-level | -l) level] [-nolib | -n]\n");
  fprintf (stderr,"  [(-commands | -c) comfile]\n");
  fprintf (stderr,"  [-nohack] [-server] \n" );
  fprintf (stderr,"  [-dumpall | -dumptxt | -dumpcg | -dumpch | -dumpdebug]\n");
  fprintf (stderr,"  [-dumpcfg id | -dumpbdb]\n" );
  fprintf (stderr,"  depfile\n");
  fprintf (stderr,
     "\n'%s -help' prints description of switches, parameters\n", execName);
  exit (1);
}

void printHelp (char* execName)
{
  printf ("Usage: %s\n", execName);
  printf ("  [-h | -help]  Print this help text\n");
  printf ("  [-u | -usage]  Print usage information (command line switches)\n");
  printf ("  [-v | -verbose]  Turn on verbose mode \n");
  printf ("  [-debug]  Print useful debugging information\n");
  printf ("  [-n | -nolib]  Exclude library functions\n");
  printf ("  [(-m | -mode) (ascii | vt100 | wyse)]  Output mode\n");
  printf ("  [(-l | -level) level]  Maximal level printed\n");
  printf ("  [(-c | -commands) comfile]  Use instrumentation command file\n");
  printf ("  [-nohack]  Do not compensate for Sage bugs\n" );
  printf ("  [-server]  Become a server\n" );
  printf ("  [-dumpall | -dumptxt | -dumpcg | -dumpch | -dumpdebug]\n" );
  printf ("  [-dumpcfg id | -dumpbdb]\n" );
  printf ("  depfile\n");
  exit (1);
}

void doCommand( int, SgProject *, int );
void doServer( SgProject * );

/* -- global variables ---------------------------------------------------- */
# define M_ASCII 0
# define M_VT100 1
# define M_WYSE  2
# define MAXMODE 3

# ifndef TRUE
#   define FALSE 0
#   define TRUE  1
# endif

char *VERT[MAXMODE]   = { "|", "x", "6" };

char *INIT[MAXMODE]   = { "", "\033(0", "\033H\002" };
char *FINISH[MAXMODE] = { "", "\033(B", "\033H\003" };

char *SO[MAXMODE]  = { "*", "\033[7m", "\b\033G4" };
char *SE[MAXMODE]  = { "*", "\033[0m", "\033G0" };

char *NORMAL[MAXMODE] = { " |__", " tqq", " 4::" };
char *LAST[MAXMODE]   = { " |__", " mqq", " 1::" };

# define NODUMP  0

/* Call graph */
# define CGDUMP  1

/* Functions and classes with their text positions */
# define TXTDUMP 2

/* Everything */
# define ALLDUMP 3

/* Class hierarchy dump */
# define CHDUMP  4

/* Debug: even more than ALL! */
# define DBGDUMP 5

/* Control-flow graph for one procedure */
# define CFGDUMP 6

/* Browser data base (ftag,ffile,ctag,cline, minus some info, plus some info)*/
# define BDBDUMP 7

/* Values used during TXTDUMP/BDBDUMP to encode state. */
# define BDBFIND 98
# define TXTFIND 99

# define SCANNED 77

int mode = M_ASCII;            /* Pretty-print: terminal mode */
int maxlevel = 10;             /* Pretty-print: max nesting level */
int noLib    = FALSE;          /* Pretty-print: don't show lib funcs */
int noHid    = TRUE;           /* Pretty-print: don't show hidden funcs */
int showInst = FALSE;          /* Pretty-print: ? */

int debugMode = FALSE;         /* Debug mode or not; used in namecheck.c */
int verboseMode = FALSE;       /* Verbose mode or not */
int beServer = FALSE;          /* Server mode or not */
int isFortran = FALSE;         /* FORTRAN code or not */
int allLowercase = TRUE;       /* case of FORTRAN routine names */
int fundecl_hack = TRUE;       /* Enable Sage++ type error workarounds */
char *projFileName = NULL;     /* Name of project file */
char *depFileName = NULL;      /* Name of dep file */

# ifdef __GNUC__
void onintr (int signal)
{
  printf (FINISH[mode]);
  exit (1);
}
# endif

/* -- functions for call graph browsing ----------------------------------- */

static char *classtype[] = { "-", "COLL" };

static int hidden[256];
static int maxhidden = 0;

static int isHidden (int tag)
{
  for (int i=0; i<maxhidden; i++) if ( hidden[i] == tag ) return (TRUE);
  return (FALSE);
}

void compute_hidden( SgCallGraph *CG )
{
  int i;
  SgSymbol *f;
  char *p;

  // Flag internal functions as hidden
  for (f=CG->firstFunction(), i=0; f; f=CG->nextFunction(), i++) {
    p = f->identifier();
    if ( strcmp (p, "Profiler") == 0 ||
	strcmp (p, "~Profiler") == 0 ||
	strcmp (p, "PROFILE_COLL_NAMES") == 0 ||
	strncmp (p, "brk_", 4) == 0 ) {
      hidden[maxhidden++] = f->id();
    }
  }
}


/*----------------------------------------------------------------------*/

/* NOTE: DumpFunc encodes state in its filepath and dumpMode parameters. */

static int DumpFunc ( SgCallGraph *CG, SgSymbol *funcsymb, int used,
		      char **filepath, int *dumpMode )
{
  SgSymbol *classsymb;
  SgMemberFuncSymb *memsymb;
  SgStatement *body;
  SgSymbol *f;
  char *classname = "-", *functype = "Ord";
  char *filename;
  int funcline = 0, classtag = -2, lastline = 0, headline = 0;
  int isColl = 0;
  int numcalls = 0;
  char *ptr;
  char fbuf[64];

  if ( isHidden (funcsymb->id()) ) return TRUE;

  if ( memsymb = isSgMemberFuncSymb(funcsymb) ) {
    body      = memsymb->body();
    classsymb = memsymb->className();
    classname = classsymb->identifier();

    if ( strncmp (classname, "brk_", 4) == 0 ) return TRUE;

    classtag  = classsymb->id();
    if ( (isColl = (classsymb->variant() == COLLECTION_NAME)) ) {
      if ( memsymb->isMethodOfElement() )
        functype = "MoE";
      else
        functype = "MoC";
    }
    else
      functype = "Mem";
  }
  else {
    if ( funcsymb == CG->mainSymbol() ) functype = "Main";
    if ( funcsymb->variant() == PROCESS_NAME ) functype = "Prcs";
    body = funcsymb->body();
  }

  if ( body ) {
    funcline = body->lineNumber();
    filename = body->fileName();
    if ( isFortran ) {
      if ( body->lastDeclaration() )
        headline = body->lastDeclaration()->lineNumber();
      else
        headline = funcline;
      lastline = body->lastNodeOfStmt()->lineNumber();
    }
  }
  else {
    filename = "-";
  }

  // The first time into DumpFunc *dumpMode is set to TXTFIND; it gets
  // changed to TXTDUMP and remains so througout the dump, until the
  // next traversal of the depfile (see doCommand(), below). Therefore,
  // the following block of code is only executed once at each start
  // of a depfile traversal.
  //
  // *filepath is initially "" and gets set to the name of the source
  // file for the current function (if available) or "-" (if not
  // available); if the name changes then the new file (is available)
  // is brought into memory using the openFile() function.

  if ( *dumpMode == TXTFIND || *dumpMode == BDBFIND ) {
    if ( strcmp (filename, *filepath) != 0 ) {
      *dumpMode = (*dumpMode == TXTFIND ? TXTDUMP : BDBDUMP);
      *filepath = filename;  /* ASGN */
      if ( **filepath != '-' ) openFile (*filepath);
    }
    else
      return FALSE;
  }

  // Dump only if the current filename is the same as the open file.

  if ((*dumpMode == TXTDUMP || *dumpMode == BDBDUMP)
      && strcmp( filename, *filepath ) != 0) 
    return FALSE;

  if ( (*dumpMode == TXTDUMP || *dumpMode == BDBDUMP) && **filepath != '-' ) {
    if ( isFortran ) {
      sprintf (fbuf, "@ %d.0 %d.79 %d.0 %d.79", funcline, headline,
	       headline+1, lastline);
      ptr = fbuf;
      setNextPos (headline+1, 0, lastline);
    }
    else {
      findFuncOrClass (funcline, &ptr);
    }
  }
  else
    ptr = "";
     
  printf ("  ftag: %d %s\n", funcsymb->id(), funcsymb->identifier());
  printf (" ffile: %d %s %s\n", funcline, filename,
	  (*dumpMode!=BDBDUMP ? ptr : ""));

  if (*dumpMode == BDBDUMP) { 
    if (strcmp( functype, "Main" ) == 0)
      printf( " ftype: Main\n\n" );
    else
      printf( "\n" ); 
    return TRUE; 
  }

  printf ("fcalls: %d -", CG->numberOfCallSites(funcsymb));
  numcalls = 0;
  for (f=CG->firstCalledFunction(funcsymb); f;
       f=CG->nextCalledFunction(funcsymb)) {
    if ( ! isHidden (f->id()) || *dumpMode == DBGDUMP ) {
      printf (" %d", f->id());
      numcalls++;
    }
  }
  printf ("\n");
  if ( *dumpMode == TXTDUMP && **filepath != '-' ) {
    printf ("  fpos:");
    for (f=CG->firstCalledFunction(funcsymb); f;
	 f=CG->nextCalledFunction(funcsymb)) {
      if ( ! isHidden (f->id()) ) {
/* There is a subtle problem here. The call to findNextChild is only valid
   if the file in the buffer in scanfile.c (read in by openFile() above)
   is the right file for the function. This is apparently not always the case,
   but I don't yet know why. --lars
 */
	findNextChild (f->identifier(), &ptr, TRUE);
	printf (" { %s }", ptr);
      }
    }
    printf ("\n");
  } else if ( *dumpMode == CGDUMP && numcalls ) {
    printf (" fline:");
    for (f=CG->firstCalledFunction(funcsymb); f;
	 f=CG->nextCalledFunction(funcsymb)) {
      if ( ! isHidden (f->id()) )
	printf (" %d", CG->SgCalledFunctionStmt(funcsymb)->lineNumber());
    }
    printf ("\n");
  }
  printf ("fclass: %s %s %d\n", classname, classtype[isColl], classtag);
  printf (" ftype: %s %s %s\n", (CG->isCalledInPar(funcsymb) ||
				 strcmp(functype,"Prcs")==0 ? "par" : "seq"),
	  functype, (used ? "used" : "not"));
  if ( *dumpMode == DBGDUMP ) {
    printf ("fdebug: moe=%d moc=%d seq=%d par=%d min=%d max=%d\n",
	    CG->isAMethodOfElement(funcsymb),
	    CG->isACollectionFunc(funcsymb),
	    CG->isCalledInSeq(funcsymb),
	    CG->isCalledInPar(funcsymb),
	    CG->minLevel(funcsymb),
	    CG->maxLevel(funcsymb));
  }
  printf ("\n");
  return TRUE;
}

// TransTree takes a call graph and a procedure (func)
// and marks as reached all procedures called by func. In effect, it
// computes the function call closure with respect to one procedure.

void TransTree ( SgCallGraph *CG, SgSymbol *func, char *used )
{
  int id;
  SgSymbol *f;

  id = CG->functionId (func);
  if ( used[id] )
    return;
  else {
    used[id] = TRUE;
    for (f=CG->firstCalledFunction(func); f; f=CG->nextCalledFunction(func))
      TransTree ( CG, f, used );
  }
}

/* -- functions for call graph browsing ----------------------------------- */
static char *type1 = "?iuoe?iuoe";
static char *type2 = "fffffvvvvv";

char *printType (memberType type, SgSymbol *symb, char *cname)
{
  SgDescriptType *d;
  static char res[8];
  char *mname = symb->identifier();

  res[0] = '{';
  res[1] = type1[(int) type];
  res[2] = type2[(int) type];
  res[3] = ' ';
  if ( d = isSgDescriptType(symb->type()) ) {
    int attr = d->modifierFlag();
    if ( attr & BIT_VIRTUAL ) res[3] = 'v';
    // if ( attr & BIT_INLINE  ) res[4] = 'i'; now stored in bif node
  }
  if ( symb->declaredInStmt()->isInline() )
    res[4] = 'i';
  else
    res[4] = ' ';
  if ( strcmp (mname, cname) == 0 )
    res[5] = 'c';
  else if ( mname[0] == '~' && strcmp (mname+1, cname) == 0 )
    res[5] = 'd';
  else if ( (mname[0] < 'a' || mname[0] > 'z') &&
            (mname[0] < 'A' || mname[0] > 'Z') )
    res[5] = 'o';
  else
    res[5] = ' ';
  res[6] = '}';
  res[7] = '\0';

  return res;
}

void TransHierarchy (int level, classHierarchy *t, SgCallGraph *CG )
{
  classHierarchy *curr;
  classMember *mem;

  if ( level ) {
    if ( strcmp (t->className, "Profiler") == 0 ||
         strncmp (t->className, "brk_", 4) == 0 ) return;

    printf (" ctag: %d %s %s\n", t->classSymbol->id(), t->className,
                                 classtype[t->isCollection()]);
  }
  else
    printf (" ctag: -1 {} -\n");

  printf (" csub:");
  curr = (classHierarchy *) t->children->searchList();
  while ( curr ) {
    if ( strcmp (curr->className, "Profiler") != 0 &&
         strncmp (curr->className, "brk_", 4) != 0 )
      printf (" %d", curr->classSymbol->id());
    curr = (classHierarchy *) t->children->nextItem();
  }
  printf ("\n");

  mem = (classMember *) t->classMembers->searchList();
  while ( mem ) {
    if ( strncmp (mem->name, "brk_", 4) != 0 ) {
      printf (" cmem: %s %d %s\n", mem->name, 
	      (fundecl_hack
	       ? CG->resolveSymbol( mem->symbol )->id()
	       : mem->symbol->id() ),
              printType (mem->typeVariant, mem->symbol, t->className));
    }
    mem = (classMember *) t->classMembers->nextItem();
  }
  printf ("\n");

  curr = (classHierarchy *) t->children->searchList();
  while ( curr ) {
    TransHierarchy (level+1, curr, CG);
    curr = (classHierarchy *) t->children->nextItem();
  }
}

/* From what Bernd tells me, this stuff was written by Darryl before
 * there was any real class support in Sage, and could be rewritten
 * to use the now-existing support. --lars 
 */
/* filepath is the name of the file we are interested in dumping 
 * classes for.  Each class hierarchy is checked to see that it
 * comes from this file; if not, it is not dumped.
 * filepath is a bona fide file name.
 */
void DumpClasses (int level, classHierarchy *t, SgCallGraph *CG, 
		  char *filepath, int dumpMode )
{
  classHierarchy *curr;
  classMember *mem;
  char *ptr;

  if ( level ) {
    /* -- hide instr generated classes --------------------------- */
    if ( strcmp (t->className, "Profiler") == 0 ||
         strncmp (t->className, "brk_", 4) == 0 ) return;

    if ( strcmp (t->fileName(), filepath) == 0 ) {
      findFuncOrClass (t->lineNumber(), &ptr);
      printf ("  ctag: %d %s %s\n", t->classSymbol->id(), t->className,
                                    classtype[t->isCollection()]);
      printf (" cfile: %d %s %s\n", t->lineNumber(), t->fileName(), 
	      (dumpMode != BDBDUMP ? ptr : ""));

      if (dumpMode == BDBDUMP) { printf( "\n" ); goto cleanup; }

      printf (" cbase:");
      curr = (classHierarchy *) t->parents->searchList();
      while ( curr ) {
        if ( strcmp (curr->className, "Profiler") != 0 &&
             strncmp (curr->className, "brk_", 4) != 0 ) {
          findNextMember (curr->className, t->lineNumber(), &ptr);
          printf (" { %d %s }", curr->classSymbol->id(), ptr);
        }
        curr = (classHierarchy *) t->parents->nextItem();
      }
      printf ("\n");

      printf ("  cpos:");
      findNextChild ("public", &ptr, FALSE);          printf (" { %s }", ptr);
      findNextChild ("protected", &ptr, FALSE);       printf (" { %s }", ptr);
      findNextChild ("private", &ptr, FALSE);         printf (" { %s }", ptr);
      findNextChild ("MethodOfElement", &ptr, FALSE); printf (" { %s }", ptr);

      printf ("\n");

      printf (" cmpos:");
      mem = (classMember *) t->classMembers->searchList();
      while ( mem ) {
	if ( type2[mem->typeVariant] == 'f' ) { //is funtion
          if ( strncmp (mem->name, "brk_", 4) != 0 ) {
            findNextMember (mem->name, mem->declaration->lineNumber(), &ptr);
            printf (" { %d %s }",
		    (fundecl_hack 
		     ? CG->resolveSymbol( mem->symbol )->id()
		     : mem->symbol->id() ),
		    ptr);
          }
        }
        mem = (classMember *) t->classMembers->nextItem();
      }
      printf ("\n\n");
    }
  }

 cleanup:
  curr = (classHierarchy *) t->children->searchList();
  while ( curr ) {
    DumpClasses (level+1, curr, CG, filepath, dumpMode );
    curr = (classHierarchy *) t->children->nextItem();
  }
}

/*----------------------------------------------------------------------*/

/* Dump the control-flow graph of the function with the given tag.
 *
 * Format of output is a node line followed by an nchildren line. Repeats.
 *
 *   node: <nodetag> <nodelevel> <nodewidth> <nodelabel>
 *   nchildren: <count> - <nodetag> <label> ...
 *
 * A label is a string labelling the control-flow edge, typically "true",
 * "false", or a case value in a switch.
 *
 * By definition, the unique entry node has a tag of 0.
 */
void cfgdumpnode( CFG *node, FILE *out )
{
  CFGE *p;
  CFGTEXT *q;

  if (node->dumped) return;
  node->dumped=1;

  fprintf( out, "     node: %d %d %d", node->tag, node->level, node->width );
  for ( q = node->thelabel ; q ; q = q->next )
    fprintf( out, " %s", q->label );
  putc( '\n', out );
  fprintf( out, "nchildren: %d -", node->childcount );
  for ( p = node->children ; p ; p = p->next )
    fprintf( out, " %d %s", p->sink->tag, p->label->label );
  putc( '\n', out );
  putc( '\n', out );

  for ( p = node->children ; p ; p = p->next )
    cfgdumpnode( p->sink, out );
}

void cfgdump( SgFile *f, int tag, FILE *out )
{
  CFG *cfg;
  SgStatement *s;
  SgFuncHedrStmt *fh;

  for ( s = f->firstStatement() ; s ; s = s->lexNext() ) {
    if ((fh = isSgFuncHedrStmt( s )) && fh->name().id() == tag) break;
  }
  if (!fh) {
    if (beServer) {
      fprintf( out, "ERROR cfgdump: No such tag: %d\n", tag );
      fprintf( stderr, "ERROR cfgdump: No such tag: %d\n", tag );
      return;
    }
    else {
      fprintf( stderr, "Error: cfgdump: can't find tag %d\n", tag );
      exit( 1 );
    }
  }

  cfg = compute_cfg( fh );
  cfgdumpnode( cfg, out );
  delete cfg;
}

// Generate the call tree.
// Handles multiple files.

void do_gencalltree( SgCallGraph *CG, SgProject *proj, int fundecl_hack )
{
  int numfiles = proj->numberOfFiles();
  int i;

  if (numfiles == 0) {
    fprintf( stderr, "Error: no files in project!\n" );
    exit( 1 );
  }
  CG->GenCallTree( &(proj->file(0)), fundecl_hack );
  for ( i = 1 ; i < numfiles ; i++ )
    CG->GenCallTree2( &(proj->file(i)) );
}


/* -- M A I N  function --------------------------------------------------- */

void prettyGraph( SgProject * );

main( int argc, char **argv )
{
  SgProject *project = 0;
  char *p;
  FILE *pf = 0;
  int i, l;
  int cfgdumptag;
  int dumpMode = NODUMP;

  if (p = getenv("TERM")) {
    if ( (strcmp (p, "vt100") == 0) || (strcmp (p, "ansi") == 0) ||
         (strcmp (p, "xterm") == 0) )
      mode = M_VT100;
    else if ( (p[0] == 'w') && (p[1] == 'y') )
      mode = M_WYSE;
    else
      mode = M_ASCII;
  }

  /*************************************************************/
  /* Parse command line */
  /*************************************************************/

  if (argc == 1) {
    printUsage (argv[0]);
    /*NOTREACHED*/
  }
  else {
    initCommands ();
    for (i=1; i<argc; i++) {
      if ( argv[i][0] == '-' ) {
        // argument == option
        p = argv[i];
        if ( (!strcmp(p,"-help")) || (!strcmp(p,"-h")) ) {
          printHelp (argv[0]);
	  /*NOTREACHED*/
        }
        else if ( (!strcmp(p,"-usage")) || (!strcmp(p,"-u")) ) {
          printUsage (argv[0]);
	  /*NOTREACHED*/
        }
        else if ( !strcmp(p,"-debug") ) {
          debugMode = TRUE;
        }
        else if ( !strcmp(p,"-dumpall") ) {
          dumpMode = ALLDUMP;
        }
        else if ( !strcmp(p,"-dumptxt") ) {
          dumpMode = TXTDUMP;
        }
        else if ( !strcmp(p,"-dumpcg") ) {
          dumpMode = CGDUMP;
        }
        else if ( !strcmp(p,"-dumpch") ) {
          dumpMode = CHDUMP;
        }
        else if ( !strcmp(p,"-dumpdebug") ) {
          dumpMode = DBGDUMP;
        }
	else if ( !strcmp( p, "-dumpcfg" )) {
	  dumpMode = CFGDUMP;
	  if (i == argc-1 ) {
	    fprintf( stderr, "Error: not enough arguments for -dumpcfg\n" );
	    printUsage( argv[0] );
	    /*NOTREACHED*/
	  }
	  i++;
	  cfgdumptag = atoi( argv[i] );
	  if (cfgdumptag < 0) {
	    fprintf( stderr, "Error: invalid argument to -dumpcfg: %s\n",
		    argv[i] );
	    printUsage( argv[0] );
	    /*NOTREACHED*/
	  }
	}
	else if ( !strcmp( p, "-dumpbdb" )) {
	  dumpMode = BDBDUMP;
	}
        else if ( (!strcmp(p,"-verbose")) || (!strcmp(p,"-v")) ) {
          verboseMode = TRUE;
        }
        else if ( (!strcmp(p,"-nolib")) || (!strcmp(p,"-n")) ) {
          noLib = TRUE;
        }
        else if ( !strcmp(p,"-hidden") ) {
          noHid = FALSE;
        }
	else if ( strcmp( p, "-server" ) == 0) {
	  beServer = TRUE;
	}
        else if ( (!strcmp(p,"-mode")) || (!strcmp(p,"-m")) ) {
          if ( i == (argc-1) ) {
            fprintf (stderr,"ERROR: not enough arguments for %s switch\n", p);
            printUsage (argv[0]);
	    /*NOTREACHED*/
          }
          i++;
          switch ( argv[i][0] ) {
            case 'a':
            case 'A': mode = M_ASCII;
                      break;
            case 'v':
            case 'V': mode = M_VT100;
                      break;
            case 'w':
            case 'W': mode = M_WYSE;
                      break;
            default : /*ignore*/
                      break;
          }
        }
        else if ( (!strcmp(p,"-level")) || (!strcmp(p,"-l")) ) {
          if ( i == (argc-1) ) {
            fprintf (stderr,"ERROR: not enough arguments for %s switch\n", p);
            printUsage (argv[0]);
	    /*NOTREACHED*/
          }
          i++;
          maxlevel = atoi(argv[i]);
        }
        else if ( (!strcmp(p,"-commands")) || (!strcmp(p,"-c")) )
        {
          if ( i == (argc-1) )
          {
            fprintf (stderr,"ERROR: not enough arguments for %s switch\n", p);
	    printUsage (argv[0]);
	    /*NOTREACHED*/
          }
          i++;
          if ( verboseMode ) printf ("reading commands file %s\n", argv[i]);
          if ( ! readComFile (argv[i]) ) exit (1);
          showInst = TRUE;
        }
	else if (strcmp( p, "-nohack" ) == 0) {
	  // Turn off the 'fix'.
	  fundecl_hack = 0;
	}
        else {
          fprintf (stderr,"ERROR: unknown option `%s'\n", p);
          printUsage (argv[0]);
	  /*NOTREACHED*/
        }
      }
      else {
        // non-option argument: must be .dep filename
        depFileName = argv[i];

        // work around the fact, that Sage++ only works in the same
        // directory as the depfile (no full pathnames supported).
        if ( (p = strrchr (depFileName, '/')) != NULL ) {
          *p = '\0';
          chdir (depFileName);
          depFileName = ++p;
        }

        // check for depfile (extention == .dep)
        // we could check also for magic string "sage.dep", but we are lazy :-)
        l = strlen(depFileName);
        p = depFileName + l - 4;
        if ( strcmp (p, ".dep") != 0 ) {
          fprintf (stderr,"ERROR: `%s': no dep file\n", depFileName);
          exit (1);
        }

        // create Sage Project filename (and file if needed)
        *p = '\0';
        projFileName = new char[l+2];
        sprintf (projFileName, "%s.proj", depFileName);
        *p = '.';

        if ( access (projFileName, R_OK) == -1 ) {
          pf = fopen (projFileName, "w");
          fprintf (pf, "%s\n", depFileName);
          fclose (pf);
        }

        if ( verboseMode || debugMode )
          printf ("using project file %s\n", projFileName);
        project = new SgProject (projFileName);
      }
    }
  }

  if (!depFileName) {
    fprintf( stderr, "%s: depfile required.\n", argv[0] );
    printUsage( argv[0] );
    /*NOTREACHED*/
  }

  /************************************************************/
  /* Do it */
  /************************************************************/

  if ( verboseMode || debugMode )
    printf ("processing %s:\n", project->fileName(0));

  isFortran = (project->file(0).languageType() == ForSrc);
  if ( verboseMode || debugMode )
    printf ("type: %s\n", isFortran ? "Fortran" : "C");

  if ( dumpMode )
    doCommand( dumpMode, project, cfgdumptag );
  else if ( beServer )
    doServer( project );
  else 
    prettyGraph( project );
  
  if ( pf ) unlink( projFileName );

  return 0;
}

void doServer( SgProject *project )
{
  static struct {
    char *name;
    int  code;
  } cmdtbl[] = {{ "dumpall", ALLDUMP },
		{ "dumptxt", TXTDUMP },
		{ "dumpcg",  CGDUMP },
		{ "dumpch",  CHDUMP },
		{ "dumpdebug", DBGDUMP },
		{ "dumpcfg", CFGDUMP },      // Argument: function ID
		{ 0, -1 }};
  char buf[ 80 ];
  int tag, i;
  
  while (fgets( buf, sizeof(buf), stdin ) != 0) {
    i = 0;
    while (cmdtbl[i].name && 
	   strncmp( buf, cmdtbl[i].name, strlen( cmdtbl[i].name ) ) != 0 )
      i++;
    if (cmdtbl[i].code == -1) {
      printf( "ERROR Bad command: %s", buf );
      fprintf( stderr, "ERROR Bad command: %s", buf );
      continue;
    }
    if (cmdtbl[i].code == CFGDUMP)
      sscanf( buf, "dumpcfg %d", &tag );
    doCommand( cmdtbl[i].code, project, tag );
    printf( "END\n" );
  }
}

void calltree( SgCallGraph *CG, SgProject *project )
{
  static int has_CG = 0;

  if (!has_CG) {
    do_gencalltree( CG, project, fundecl_hack );
    has_CG = 1;
  }
}

void closure( SgCallGraph *CG )
{
  static int has_closure = 0;

  if (!has_closure) {
    CG->computeClosures();
    has_closure = 1;
  }
}

void classhierarchy( classHierarchy *CH, SgProject *project )
{
  static int has_CH = 0;

  if (!has_CH) {
    CH->determineClassHierarchy(&(project->file(0)));
    has_CH = 1;
  }
}

void chsort( classHierarchy *CH )
{
  static int has_CH_sort = 0;

  if (!has_CH_sort) {
    CH->sort();
    has_CH_sort = 1;
  }
}

char *boilerplate( SgCallGraph *CG )
{
  int num, i;
  char *used;
  SgSymbol *mainSymb;
  
  num = CG->numberOfFunctionsInGraph();
  printf ("%d\n\n", num);
  
  // Look for a main() procedure.
  //
  // 'used' is an array of booleans; used[i] is true iff procedure i is
  // reachable from the main procedure, where i is the index of
  // the procedure in the AST.
  
  used = new char[num];
  for ( i = 0 ; i < num ; i++ ) used[i] = FALSE;
  compute_hidden( CG );

  mainSymb = CG->mainSymbol();
  if (!mainSymb) {
    fprintf (stderr, "%s: no main function\n", depFileName );
    exit (1);
  }
  
  // Select the called procedures.
  TransTree( CG, mainSymb, used );
  return used;
}

void doCommand( int dumpMode, SgProject *project, int cfgdumptag )
{
  static classHierarchy CH;
  static SgCallGraph CG;

  SgSymbol *f;
  char *used = 0;
  int i;
  char *filepath = "";

  switch (dumpMode) {
  case CHDUMP :
    classhierarchy( &CH, project );
    chsort( &CH );
    if (fundecl_hack) calltree( &CG, project );
    TransHierarchy (0, &CH, &CG);
    break;
  case CFGDUMP :
    cfgdump( &(project->file(0)), cfgdumptag, stdout );
    break;
  case DBGDUMP :
    calltree( &CG, project );
    closure( &CG );
    for (f=CG.firstFunction(); f; f=CG.nextFunction())
      DumpFunc ( &CG, f, TRUE, &filepath, &dumpMode );
    break;
  case CGDUMP :
    calltree( &CG, project );
    closure( &CG );
    used = boilerplate( &CG );
    for (f=CG.firstFunction(), i=0; f; f=CG.nextFunction(), i++)
      if ( used[i] ) DumpFunc ( &CG, f, TRUE, &filepath, &dumpMode );
    break;
  case ALLDUMP :
    calltree( &CG, project );
    closure( &CG );
    used = boilerplate( &CG );
    for (f=CG.firstFunction(), i=0; f; f=CG.nextFunction(), i++)
      DumpFunc ( &CG, f, used[i], &filepath, &dumpMode );
    break;
  case TXTDUMP :
  case BDBDUMP :
    { int scanned, i, num;
      SgSymbol *f;
      
      calltree( &CG, project );
      closure( &CG );
      classhierarchy( &CH, project );
      chsort( &CH );

      used = boilerplate( &CG );
      num = CG.numberOfFunctionsInGraph();

      scanned = 0;
      while ( scanned != num ) {
	dumpMode = ( dumpMode == TXTDUMP ? TXTFIND : BDBFIND );
	for (f=CG.firstFunction(), i=0; f; f=CG.nextFunction(), i++) {
	  if (used[i] != SCANNED) {
	    if ( DumpFunc ( &CG, f, used[i], &filepath, &dumpMode ) ) {
	      used[i] = SCANNED;
	      scanned++;
	    }
	  }
	}
	if ( *filepath != '-' )
	  DumpClasses (0, &CH, &CG, filepath, dumpMode);
      }
      // closeFile ();
    }
    break;
  default :
    fprintf( stderr, "INTERNAL ERROR #1" );
    exit( 1 );
  }
  if (used != 0) delete [] used;
}


/***************************************************************************/

// Default dump format: a pretty hierarchical graph.

void PrintFunc (SgSymbol *funcsymb);
void PrintGraph (SgCallGraph *CG, SgSymbol *func, int level);

int *finish;

void prettyGraph( SgProject *project )
{
  SgSymbol *mainSymb, *f;
  char *p;
  int i;
  SgCallGraph CG;

  calltree( &CG, project );
  closure( &CG );
  mainSymb = CG.mainSymbol();
  if (!mainSymb) {
    fprintf (stderr, "%s: no main function\n", depFileName );
    exit (1);
  }
  
  compute_hidden( &CG );
  
# ifdef __GNUC__
  signal (SIGINT,  onintr);
  signal (SIGQUIT, onintr);
# endif
  
  finish = new int[maxlevel+1];
  for (i=0; i<=maxlevel; i++) finish[i] = TRUE;
  
  putchar ('\n');
  PrintFunc (mainSymb);
  printf ("\n %s%s%s", INIT[mode], VERT[mode], FINISH[mode]);
  PrintGraph (&CG, mainSymb, 0);
  putchar ('\n');
}

void PrintFunc (SgSymbol *funcsymb)
{
  SgSymbol *classsymb;
  SgMemberFuncSymb *memsymb;
  char *funcname, *classname = 0, *filename = 0;
  char *INS = "", *INE = "";
  int funcline = -1;
  int isMoE = 0, isColl = 0;

  funcname = funcsymb->identifier();
  if ( funcsymb->body() ) {
    funcline = funcsymb->body()->lineNumber();
    filename = funcsymb->body()->fileName();
  }

  if ( memsymb = isSgMemberFuncSymb(funcsymb) ) {
    classsymb = memsymb->className();
    classname = classsymb->identifier();
    if ( (isColl = (classsymb->variant() == COLLECTION_NAME)) )
      isMoE  = memsymb->isMethodOfElement();
  }

  if ( showInst &&
       isValidFunc (filename, classname, funcname) &&
       funcsymb->body() ) {
    INS = SO[mode];
    INE = SE[mode];
  }

  if ( classname ) {
    printf (" %s%s::%s%s", INS, classname, funcname, INE);
    if ( verboseMode )
      printf ("%s%s", (isColl ? " [Coll]" : ""), (isMoE ? " [MoE]" : ""));
  }
  else
    printf (" %s%s%s", INS, funcname, INE);
  if ( verboseMode && filename )
    printf (" [%s:%d]", filename, funcline);
}

void PrintGraph (SgCallGraph *CG, SgSymbol *func, int level)
{
  static int done = FALSE;
  int i, j, n, last, no;
  SgSymbol *f;

  if ( no = CG->numberOfFunsCalledFrom(func) ) {
    if (level >= maxlevel) {
      printf (" ...\n");
      return;
    }

    if ( noLib || noHid) {
      n = 0;
      for (f=CG->firstCalledFunction(func), i=0; f;
           f=CG->nextCalledFunction(func), i++) {
        if ( (!noLib || f->body()) && (!noHid || !isHidden(f->id())) ) {
          n++;
          last = i;
        }
      }
      if ( n == 0 ) {
        putchar ('\n');
        return;
      }
    }
    else
      last = no - 1;

    finish[level] = FALSE;

    putchar ('\n');
    if (level) {
      printf (INIT[mode]);
      for (j=0; j<=level; j++)
        printf (" %s  ", (finish[j] ? " " : VERT[mode]));
      putchar ('\n');
      printf (FINISH[mode]);
    }

    for (f=CG->firstCalledFunction(func), i=0; f;
         f=CG->nextCalledFunction(func), i++) {
      if ( (!noLib || f->body()) && (!noHid || !isHidden(f->id())) ) {
        printf (INIT[mode]);
        for (j=0; j<level; j++)
          printf (" %s  ", (finish[j] ? " " : VERT[mode]));
        if (i == last)
          printf (LAST[mode]);
        else
          printf (NORMAL[mode]);
        printf (FINISH[mode]);

        PrintFunc (f);
        done = FALSE;
        if (i == last) finish[level] = TRUE;
        PrintGraph (CG, f, level+1);
      }
    }
    if ( !done ) {
      if (level) {
        printf (INIT[mode]);
        for (j=0; j<level; j++)
          printf (" %s  ", (finish[j] ? " " : VERT[mode]));
        putchar ('\n');
        printf (FINISH[mode]);
      }
      done = TRUE;
    }
  }
  else {
    putchar ('\n');
  }
}

// eof
