// g e n s y m t a b . C
// Generate an extensive symbol table from a .dep file, for Sneezy/Sneaky.
//
// Copyright (C) 1995 The University of Oregon.
//
//        Author: Lars Thomas Hansen [lth@cs.uoregon.edu]
// Creation date: July 28, 1995
//
// $Id: gensymtab.C,v 1.1 1996/01/08 23:33:24 lth Exp $
//
// DESCRIPTION
//   This program takes a .dep file or a .proj file and generates symbol
//   table information on a Scheme-readable format. If a .proj file is
//   given, all .dep files are processed, although it's not clear how
//   meaningful this is.
//
//   For a .dep file PATH/foo.dep, output is written to the file
//   PATH/foo.symtab. Errors are written to stdout.
//
// USAGE
//   gensymtab depfilename
//   gensymtab projfilename
//
// OUTPUT FORMAT (evolving)
//   Output consists of type table entries and symbol table entries.
//
//   Type table entry
//     (type <type-code> <type-desc>)
//
//   Symbol table entry
//     (symbol <name> type <type-code> <scope>)
//     (symbol <name> variable <type-code> <scope>)
//
//   Explanations
//   - A <name> is a quoted string, denoting a program identifier.
//   - A <type-code> is a nonnegative integer.
//   - A <type-desc> is a recursive type description (defined elsewhere).
//   - A <scope> is a structure which describes the scope of the identifier.
//     It can take several formats, but it's always a list where the first
//     element is a tag:
//       (lex-scope <start-line> <start-pos> <end-line> <end-pos>)
//       (class-scope <type-code>)
//     In a sense, a class scope is a collection of lexical scopes: the
//     lexical scopes of the class declaration body along with the lexical 
//     scopes of all member functions. Which one works better is unclear 
//     right now.
//
//     There may be several entries for the same symbol with different
//     attributes, types, and scopes, representing different uses for that
//     symbol in the program.
//
// BUGS and UNRESOLVED ISSUES
//   Holes in the scope are represented implicitly, which may or may not
//   be OK. 
//
//   The representation can be a lot more compact and still scheme-readable.
//
//   The lexical scope should include (a reference to) the file name.
//
//   We might want to distinguish function declaration and function
//   definition, somehow, because the scope of the declaration typically is
//   different from the scope of the definition.
//
//   The attribute of a base class and a member symbol (public, protected, 
//   private) is not irrelevant to resolve scope issues, but is not currently 
//   recorded in the symbol table.
//
//   A fair bit of code in this program compensates for bugs in Sage++. Most
//   such snippets of code are labeled SAGEBUG.
//
// PERFORMANCE
//   On a SS 10/51: Processes the .dep file for simple.pc and all the sources
//   and writes the output file (200KB) in less than 2 seconds elapsed time.
//   (Local disk :-)
//
// TODO
// - Scope information must have file name or ID (there could be a FILE record
//   for each file).
// - There are still some 'unknown' entries for scope, suggesting all names
//   are not annotated or that some names get dumped too early.
// - Several unsupported and not-yet types.
// - Must run preprocessor to cope with scanning problems. Unfortunately, this
//   means obtaining include-file paths and definitions as run with the 
//   compiler, and it means reacting to #file and #line in the preprocessor 
//   output.
// - Fix build_type().
// - Scope resolution in scanning for ::, ., ->, ->*, .*

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include "sage++user.h"

#define DEBUG 1

char *progname;                              // value of argv[0]
int option_print_warnings = 1;

SgProject *open_project( char * );
void generate_symtab( SgFile *, char * );

int main( int argc, char **argv )
{
  SgProject *project;

  progname = argv[0];
  if (argc != 2) {
    fprintf( stderr, "Usage: %s depfilename\n"
                     "       %s projfilename\n", progname, progname );
    exit( 1 );
  }

  project = open_project( argv[1] );
  for ( int fno = 0 ; fno < project->numberOfFiles() ; fno++ ) {
    SgFile *file;
    char *name;

    file = &(project->file( fno ));
    name = project->fileName( fno );
    generate_symtab( file, name );
  }

  return 0;
}

/*-------------------------------------------------------------------------*/

#ifndef NAMELEN_MAX
#define NAMELEN_MAX 1024                     // max length of file name
#endif

#define EMPTY_SYM "*"                        // symbol denoting "no symbol"
#define QEMPTY_SYM "\"*\""

#define T_FUNCSIG   LOCAL  //hijacked, for now; must be unique eventually

SgType *clean_type( SgType * );
const char *clean_name( SgSymbol * );
const char *mbit( const long, const char * );
void structured_type( FILE *, SgType *, const char * );
void print_type( FILE *, SgType *, long flags = 0 );
void print_fsign( FILE *, SgFunctionSymb * );
void print_scope( FILE *, SgLabelSymb * );
void print_scope( FILE *, SgFunctionSymb * );
void print_scope( FILE *, SgFieldSymb * );
void print_scope( FILE *, SgVariableSymb * );
void print_scope( FILE *, SgSymbol * );
void print_type_scope( FILE *, SgSymbol * );
void scanner( char *infile, FILE *outfp );
SgType *build_type( SgExpression *e, SgSymbol ** );
SgExpression *build_type0( SgExpression *e );
int panic( char *, ... );
int valid_function_op( SgExpression *, SgType ** );

class SgFuncSigType : public SgType {
public:
  SgFuncSigType( SgFunctionSymb * );
  SgFunctionSymb *symbol() {
    return isSgFunctionSymb( ((SgType*)this)->symbol() );
  }
};

SgFuncSigType::SgFuncSigType( SgFunctionSymb *sym ) : SgType( T_FUNCSIG, sym )
{
  // nothing
}

SgFuncSigType *isSgFuncSigType( SgType *t )
{
  return (t->variant() == T_FUNCSIG ? (SgFuncSigType*)t : 0);
}

struct name_node {
  SgSymbol    *symbol;
  char        *name;
  SgType      *type;
  name_node   *next;
  name_node( SgSymbol *s, SgType *t ) {
    symbol = s;
    name = s->identifier();
    type = t;
    next = 0;
  }
  ~name_node() {
    if (next != 0) delete next;
  }
};

struct file_node {
  char *n;
  file_node *prev;
  file_node *next;
  file_node() {
    prev = this;
    next = this;
    n = 0;
  }
  file_node( char *name, file_node *after_this ) {
    n = name;
    next = after_this->next;
    prev = after_this;
    prev->next = this;
    next->prev = this;
  }
  ~file_node() {
    prev->next = 0;
    if (next != 0) delete next;
  }
  int sameas( char *fn ) {
    return n != 0 && (n == fn || strcmp( fn, n ) == 0);
  }
  char *name() {
    return (n ? n : "");
  }
};

#define ATTR_TYPE_PRINTED 1
#define ATTR_SYM_PRINTED  2
#define ATTR_FUNCSIG      3
#define ATTR_DECLAREDSTMT 4

void mark_as_printed_type( SgType *t )
{
  t->addAttribute( ATTR_TYPE_PRINTED );
}

int has_printed_type( SgType *t )
{
  return t->getAttribute( 0, ATTR_TYPE_PRINTED ) != 0;
}

void mark_as_printed_symbol( SgType *t )
{
  t->addAttribute( ATTR_SYM_PRINTED );
}

int has_printed_symbol( SgType *t )
{
  return t->getAttribute( 0, ATTR_SYM_PRINTED ) != 0;
}

void generate_symtab( SgFile *f, char *name )
{
  FILE *fp = stdout;
  name_node *first = 0, *last = 0;
  file_node *fnode = new file_node();

  // Annotate all function names with the statement in which the function
  // was declared. We use this later rather than SgSymbol::declaredInStmt(),
  // which is very slow. We need to discover the statement to figure out
  // the type of the function, which is not recorded in a coherent type
  // structure. It is probably a SAGEBUG for the symbol not to point to such
  // a type structure (and for it not to exist).
  //
  // This loop also discovers the names of the files used to create the
  // dep file (i.e., .pc file and all headers which were #included).
  //
  // Files are kept on a two-way circular linked list with a header
  // node, where the current node is that of the filename of the most
  // recently scanned statement; this is a very effective cache.

  for ( SgStatement *st = f->firstStatement() ; st != 0 ; st = st->lexNext() ){
    SgFuncHedrStmt *fhs;
    SgVarDeclStmt *vds;
    SgClassStmt *cs;

    if (fhs = isSgFuncHedrStmt( st )) {
      (fhs->name()).addAttribute( ATTR_DECLAREDSTMT, (void*)st, sizeof( st ) );
      // for each parameter symbol, set the attribute.
      int i;
      for ( i = 0 ; i < fhs->numberOfParameters() ; i++ )
	(fhs->parameter(i))->addAttribute( ATTR_DECLAREDSTMT,
					  (void*)st, sizeof( st ) );
    }
    else if (vds = isSgVarDeclStmt( st )) {
      for ( int i = 0 ; i < vds->numberOfSymbols() ; i++ )
	(vds->symbol(i))->addAttribute( ATTR_DECLAREDSTMT,
				       (void*)st, sizeof( st ) );
    }
    else if (cs = isSgClassStmt( st )) {
      // CLASS_DECL, STRUCT_DECL, UNION_DECL, ENUM_DECL, COLLECTION_DECL
      (cs->name())->addAttribute( ATTR_DECLAREDSTMT, (void*)st, sizeof( st ) );
    }

    file_node *p = fnode;
    char *fn = st->fileName();

    do {
      if (p->sameas( fn )) break;
      p = p->prev;
    } while (p != fnode);
    if (!p->sameas( fn )) p = new file_node( fn, fnode );
    fnode = p;
  }

  // Annotate all function types (the mostly-null type FUNCTION_T) with a newly
  // created type representing the signature of the function. A FUNCTION_T is
  // created when a function type is declared in a typedef, variable, or 
  // function parameter, but the complete type does not exist as a type 
  // structure, only as inferrable from the declaration, ultimately from a
  // FUNCTION_OP expression. It's probably a SAGEBUG for the complete signature
  // not to exist in the first place.
  //
  // What we are looking for is a FUNCTION_OP expression with an lhs which
  // looks like a valid function name expression: a sequence of pointers and
  // references (DEREF_OP and ADDRESS_OP) terminating in a VAR_REF. The
  // type of the VAR_REF is then the type node of the symbol (a chain of
  // T_REFERENCE and T_POINTER ending in a T_FUNCTION). The rhs, if non-null,
  // is an EXPR_LIST of parameter declarations (symbols and types).
  // A parameter is a structure containing one of the
  // following operators: deref, var-ref, address, function, and array.

  for ( SgExpression *e=f->firstExpression() ; e!=0 ; e=e->nextInExprTable() ){
    SgType *ft;
    if (e->variant() == FUNCTION_OP && valid_function_op( e, &ft )) {
      SgExpression *sign = build_type0( e->rhs() );
      if (e->rhs() != 0 && sign == 0) continue;
      ft->addAttribute( ATTR_FUNCSIG, (void*)sign, sizeof( sign ) );
    }
  }

  // SAGEBUG: For some reason symbol table traversal needs to go before type
  // table traversal, or strange errors result... looks like the attribute
  // code clobbers the tree.
  //
  // Anyhow, there is now a flow dependence between passes, so it's not
  // too bad.

  for ( SgSymbol *s = f->firstSymbol() ; s != 0 ; s = s->next() ) {
    SgVariableSymb *vs;
    SgFieldSymb *fs;
    SgFunctionSymb *fns;
    SgLabelSymb *ls;
    SgConstantSymb *cs;

    if (s->identifier() == 0 || *(s->identifier()) == 0) continue;
    if (vs = isSgVariableSymb( s )) {
      // "..." being a variable symbol looks like a SAGEBUG to me.
      if (strcmp( s->identifier(), "..." ) == 0) continue;

      SgType *t = clean_type( vs->type() );
      fprintf( fp, "(symbol \"%s\" variable %d ", vs->identifier(), t->id() );
      print_scope( fp, vs );
      fprintf( fp, ")\n" );
    }
    else if (fs = isSgFieldSymb( s )) {
      // SAGEBUG: In the current version of Sage++, ENUM_NAME is used 
      // as the variant of the enum tag, not for enum fields.
      // FIELD_NAME can be either enum fields or structure fields.

      if (fs->variant() == ENUM_NAME) continue; // handled as type.

      // SAGEBUG: If there is not a prototype in scope for a function
      // which is called from a member function (see comments in 
      // print_fsign() for how this happens) then the function will
      // erroneously appear as a field symbol of the class with a null
      // type, in addition to appearing as a regular function.

      if (fs->type() == 0) {
	if (option_print_warnings)
	  fprintf( stderr, "WARNING: no declaration for member function %s\n",
		  fs->identifier() );
	continue;
      }

      fprintf( fp, "(symbol \"%s\" field %d ",
	      fs->identifier(), fs->type()->id() );
      print_scope( fp, fs );
      fprintf( fp, ")\n" );
    }
    else if (fns = isSgFunctionSymb( s )) {
      SgType *t = new SgFuncSigType( fns );

      fprintf( fp, "(symbol \"%s\" function %d ", s->identifier(), t->id() );
      print_scope( fp, fns );
      fprintf( fp, ")\n" );
      s->addAttribute( ATTR_FUNCSIG, (void*)t, sizeof( t ) );
    }
    else if (ls = isSgLabelSymb( s )) {
      fprintf( fp, "(symbol \"%s\" label ", s->identifier() );
      print_scope( fp, ls );
      fprintf( fp, ")\n" );
    }
    else if (cs = isSgConstantSymb( s )) {
      fprintf( fp, "(symbol \"%s\" const <type> <value> <scope>)\n",
	      s->identifier() );
    }
    else if (s->variant() == TYPE_NAME && s->type() != 0) {
      fprintf( fp, "(symbol \"%s\" type %d ",
	      s->identifier(), clean_type( s->type() )->id() );
      print_type_scope( fp, s );
      fprintf( fp, ")\n" );
    }
  }

  // For all types, dump types, giving each type an identifying number,
  // and remembering the names of those types that have them.
  // ElementType is handled specially.

  for ( SgType *t = f->firstType() ; t != 0 ; t = t->next() ) {
    SgType *t2;
    SgDescriptType *dt;
    name_node *p;

    if (t->isTheElementType()) continue;
    t2 = clean_type( t );
    if ((dt = isSgDescriptType( t )) && (dt->modifierFlag() & BIT_TYPEDEF))
      continue;

    if (!has_printed_type( t2 )) {
      fprintf( fp, "(type %d ", t2->id() );
      print_type( fp, t2 );
      fprintf( fp, ")\n" );
      mark_as_printed_type( t2 );
    }

    if (t2->variant() == T_FUNCSIG) continue;
    // SAGEBUG: Some pointer types have symbols, for no good reason. 
    // Anyway, code subsumed by following statement.
    // if (t->variant() == T_POINTER) continue;
    if (!isSgClassType( t2 ) && t2->variant() != T_ENUM) continue;
    if (t->symbol() == 0 || t->symbol()->identifier() == 0) continue;
    if (strcmp( t->symbol()->identifier(), "" ) == 0) continue;

    // remember name and ID and statement it was declared in
    (p = new name_node( t->symbol(), t2 )) || panic( "no memory" );
    if (last)
      last->next = p;
    else
      first = p;
    last = p;
  }

  for ( name_node *p = first ; p != 0 ; p = p->next ) {
    if (!has_printed_symbol( p->type )) {
      fprintf( fp, "(symbol \"%s\" type %d ", p->name, p->type->id() );
      print_type_scope( fp, p->symbol );
      fprintf( fp, ")\n" );
      mark_as_printed_symbol( p->type );
    }
  }

  fflush( fp );

  // For each file, do the identifier/scope scanning.
  { file_node *pp = fnode;
    do {
      if (*pp->name() != 0) {
#ifdef DEBUG
	fprintf( stderr, "; SCANNING: %s\n", pp->name() );
#endif
	scanner( pp->name(), fp );
      }
      pp = pp->next;
    } while (pp != fnode);
  }

  fflush( fp );

  delete first;
  delete fnode;
}

// Returns 1 if the expression can be understood as a function prototype
// in a function header.
//
// the lhs is a sequence of derefs and references ending in a var_ref
// the rhs is an expr_list or 0. Unfortunately (SAGEBUG), references are 
// encoded with the ADDRESS_OP variant. Fortunately, the representation
// of the declaration (*(&d))() is different from the expression (*(&d))():
// the latter has an EXPR_LIST as the lhs of the FUNCTION_OP.

int valid_function_op( SgExpression *e, SgType **ft )
{
  if (e->variant() != FUNCTION_OP) return 0;
  if (e->lhs() == 0) return 0;
  e = e->lhs();
  while (e && (e->variant() == DEREF_OP || e->variant() == ADDRESS_OP))
    e = e->lhs();
  if (e && e->variant() == VAR_REF) {
    SgType *t;
    t = e->type();
    while (t && (t->variant() == T_REFERENCE || t->variant() == T_POINTER))
      t = t->baseType();
    if (t && t->variant() == T_FUNCTION) {
      *ft = t;
      return 1;
    }
    return 0;
  }
  return 0;
}

SgExpression *build_type0( SgExpression *e )
{
  static SgSymbol *emptyname = 0;
  SgSymbol *name = 0;
  SgType *t;
  SgExpression *lhs, *rhs;

  if (e == 0) return 0;
  t = build_type( e->lhs(), &name );
  if (t == 0) return 0;
  if (name == 0) {
    if (!emptyname) emptyname = new SgSymbol( VARIABLE_NAME, "" );
    name = emptyname;
  }
  rhs = build_type0( e->rhs() );
  if (rhs == 0) return 0;
  lhs = new SgExpression( VAR_REF, 0, 0, name, t );
  return new SgExpression( EXPR_LIST, lhs, rhs, 0 );
}

// FIXME: Needs work. Way too primitive.

SgType *build_type( SgExpression *e, SgSymbol **name )
{
  SgType *t;
  int ptr = 0;
 Again:
  switch (e->variant()) {
  case DEREF_OP :
    ptr++;
    e = e->lhs();
    goto Again;
  case VAR_REF:
    *name = e->symbol();
    return e->type();
  case ADDRESS_OP :
    e = e->lhs();
    goto Again;
  case FUNCTION_OP :
    // Here we grab the T_FUNCTION, expecting it to be annotated eventually.
    // ?? name
// Plain WRONG. FIXME.
    return e->lhs()->lhs()->type();
  case ARRAY_REF :
    *name = e->symbol();
    t = e->type();
    while (ptr--)
      t = new SgPointerType( t );
    return t;
  default :
    fprintf( stderr, "build_type: very confused about (%d,%d)\n",
	    e->id(), e->variant() );
    // return SgTypeInt();
    return 0;
  }
}

void print_type( FILE *fp, SgType *t, long flags )
{
  // This is part sanity check and part SAGEBUG. The bug part is this: some
  // descriptor types in parameter lists of function types (not declarations
  // or definitions, only when the signature is used as a parameter type) do
  // not have a base type (it's 0), but when we assume it's an INT, everything
  // works fine, even the modifier flags.

  if (t == 0) {
//    fprintf( fp, "(ERROR null-type %ld)", flags );
    fprintf( fp, "(%sint)", mbit( flags, "s" ) );
    return;
  }
  switch (t->variant()) {
  case DEFAULT :
  case T_INT :
    fprintf( fp, "(%sint)", mbit( flags, "s" ) );
    break;
  case T_FLOAT :
    fprintf( fp, "(float)" );
    break;
  case T_DOUBLE :
    fprintf( fp, "(double)" );
    break;
  case T_COMPLEX :
    fprintf( fp, "(complex)" );
    break;
  case T_DCOMPLEX :
    fprintf( fp, "(double-complex)" );
    break;
  case T_CHAR :
    fprintf( fp, "(%schar)", mbit( flags, "" ) );
    break;
  case T_BOOL :
    fprintf( fp, "(bool)" );
    break;
  case T_STRING :
    fprintf( fp, "(string)");
    break;
  case T_ENUM :
    // FIXME: should we (can we) print the field names here?
    fprintf( fp, "(enum)" );
    break;
  case T_SUBRANGE :
    fprintf( fp, "(subrange)" ); // unsupported
    break;
  case T_LIST :
    fprintf( fp, "(list)" ); // unsupported
    break;
  case T_ARRAY : 
    {
      SgArrayType *at = isSgArrayType( t );
      fprintf( fp, "(array (" );
      for ( int i = at->dimension(), j=0 ; j < i ; j++ ) {
	SgExpression *size = at->sizeInDim( j );
	char buf[ 100 ];
	if (size == 0)
	  strcpy( buf, EMPTY_SYM );
	else
	  size->sunparse( buf );
	fprintf( fp, " %s", buf );
      }
      fprintf( fp, " ) " );
      print_type( fp, at->baseType() );
      fprintf( fp, ")" );
    }
    break;
  case T_RECORD :
    fprintf( fp, "(record)" );  // unsupported
    break;
  case T_ENUM_FIELD :
    fprintf( fp, "(enum-field)" );
    break;
  case T_UNKNOWN :
    fprintf( fp, "(unknown-type)" );  // unsupported
    break;
  case T_VOID :
    fprintf( fp, "(void)" );
    break;
  case T_DESCRIPT : 
    { long m = isSgDescriptType( t )->modifierFlag();

      if (m == 0 || (m & BIT_TYPEDEF)) {
	SgType *t2 = clean_type( t );
	if (t == t2)
	  fprintf( fp, "(type %d)", t->id() );
	else
	  print_type( fp, clean_type( t ) );
      }
      else
	print_type( fp, t->baseType(), m );
    }
    break;
  case T_FUNCTION :
    { SgExpression *plist;
      // The return type is encoded as the base type
      // The parameter list is stored as an expr in an attribute;
      // each expression is a var_ref where the type is the type
      // and the symbol, if not "", is the name.
      fprintf( fp, "(function () (" ); 
      plist = (SgExpression*)(t->attributeValue( 0, ATTR_FUNCSIG ));
//      if (!plist) { printf( "NOTE: plist is null.\n" ); }
      while (plist != 0) {
	// FIXME: handle '...'
	fprintf( fp, "(\"%s\" (type %d))",
		clean_name( plist->lhs()->symbol() ),
		clean_type( plist->lhs()->type() )->id() );
	plist = plist->rhs();
      }
      fprintf( fp, ")" );
      print_type( fp, t->baseType() );
      fprintf( fp, ")" );
    }
    break;
  case T_FUNCSIG :
    print_fsign( fp, isSgFuncSigType( t )->symbol() );
    break;
  case T_POINTER : 
    fprintf( fp, "(pointer " ); 
    print_type( fp, t->baseType() );
    fprintf( fp, ")");   
    break;
  case T_UNION      : structured_type( fp, t, "union" ); break;
  case T_STRUCT     : structured_type( fp, t, "struct" ); break;
  case T_CLASS      : structured_type( fp, t, "class" ); break;
  case T_COLLECTION : structured_type( fp, t, "collection" ); break;
  case T_DERIVED_CLASS : 
    { SgDerivedClassType *dt = isSgDerivedClassType( t );
      print_type( fp, clean_type( dt->typeName()->type() ) );
    }
    break;
  case T_DERIVED_COLLECTION :
    fprintf( fp, "(derived-collection)" );
    break;
  case T_DERIVED_TYPE :
    if (t->isTheElementType())
      fprintf( fp, "(element-type)" );
    else {
      t = clean_type( t );
      fprintf( fp, "(type %d)", t->id() );
    }
    break;
  case T_REFERENCE :
    { SgReferenceType *rt = isSgReferenceType( t );
      fprintf( fp, "(ref " );
      print_type( fp, rt->baseType() );
      fprintf( fp, ")" );
    }
    break;
  case T_DERIVED_TEMPLATE :
  case T_MEMBER_POINTER :
  case T_TECLASS :
  case T_GATE :
  case T_EVENT :
  case T_SEQUENCE :
  case T_AUTO :
  case T_IDL_SEQ :
  case T_IDL_STRING :
  case T_IDL_OCTET :
  case T_IDL_ANY :
    fprintf( fp, "(not-yet)" ); 
    break;
  default :
    fprintf( fp, "(error %d)", t->variant() );
    break;
  }
}

void print_baseclasses( FILE *fp, SgClassType *ct )
{
  // SAGEBUG: No way to do this in Sage++; there should be.
  // The base class list is an expression list hung off the ll_ptr2 field
  // of the Template substructure of the BIF node pointed to by the 
  // original_class field of the derived_class substructure of the type
  // node. Go to h/symb.h and h/bif.h for more answers.
  PTR_LLND e = 
    ct->thetype->entry.derived_class.original_class->entry.Template.ll_ptr2;
  SgExprListExp *el;
  
  if (e == 0) {
    fprintf( fp, "()" );
    return;
  }
  
  el = new SgExprListExp( e ); 
  fprintf( fp, "(" );
  while (el != 0) {
    fprintf( fp, "%d ", clean_type( el->lhs()->symbol()->type() )->id() );
    el = isSgExprListExp( el->rhs() );
  }
  fprintf( fp, ")" );
}

void structured_type( FILE *fp, SgType *t, const char *tag )
{
  int i;
  SgSymbol *s;
  SgClassType *ct;
  
  ct = isSgClassType( t );
  if (ct == 0) panic( "class type is 0." );
  fprintf( fp, "(%s ", tag );
  print_baseclasses( fp, ct );
  i = 0;
  while ((s = ct->fieldSymb( i++ )) != 0) {
    fprintf( fp, "(%s ", clean_name( s ) );
    if (isSgMemberFuncSymb( s )) {
      SgType *t = (SgType*)(s->attributeValue( 0, ATTR_FUNCSIG ));
      if (t != 0)
	fprintf( fp, "(type %d)", t->id() );
      else
	fprintf( fp, "(type -1)" );
    }
    else
      print_type( fp, s->type() );
    fprintf( fp, ") ");
  }
  fprintf( fp, ")" );
}

// Given function symbol, print its signature.
// Signature is printed as (tag (attr ...) ((id type) ...) type)
// where tag is FUNCTION or MEMBERFUNCTION and the attrs are STATIC, 
// EXTERN, INLINE, and FRIEND.

void print_fsign( FILE *fp, SgFunctionSymb *fs )
{
  int i;
  SgStatement *stmt;
  SgMemberFuncSymb *mf;

  if (mf = isSgMemberFuncSymb( fs ))
    fprintf( fp, "(memberfunction %d (",
	    clean_type( mf->className()->type() )->id() );
  else
    fprintf( fp, "(function (" );

  // attributes: go to the statement, look at the bits set:
  // sgstatement::isstatic, isfriend, isinline, isextern.
  stmt = (SgStatement*)(fs->attributeValue( 0, ATTR_DECLAREDSTMT ));
  if (stmt == 0) {
    // SAGEBUG
    // There's not a prototype or declaration of this function in the file
    // scope. This happens in sage because of the assumption that 'runtime.h'
    // will be included eventually, providing the necessary declarations
    // for the C++ compiler. However, the declarations are not here now.
    // The way to fix this is to assume the function has a ... argument list
    // and extern attribute, and a return type of void. It's wrong, but
    // it works.
    if (option_print_warnings)
      fprintf( stderr, "WARNING: no declaration for function %s\n",
	      fs->identifier() );
    fprintf( fp, "extern) ((" QEMPTY_SYM "(dotdotdot))) (void)" );
  }
  else {
    if (stmt->isStatic()) fprintf( fp, "static ");
    if (stmt->isInline()) fprintf( fp, "inline ");
    if (stmt->isExtern()) fprintf( fp, "extern ");
    if (stmt->isFriend()) fprintf( fp, "friend ");
    fprintf( fp, ") (" );
    for ( i = 0 ; i < fs->numberOfParameters() ; i++ ) {
      const char *n = clean_name( fs->parameter(i) );
      if (strcmp( n, "..." ) != 0) {
	fprintf( fp, "(\"%s\" ", n ); 
	print_type( fp, fs->parameter(i)->type() );
	fprintf( fp, ")" );
      }
      else
	fprintf( fp, "(" QEMPTY_SYM " (dotdotdot))" );
    }
    fprintf( fp, ") ");
    print_type( fp, fs->type() );
  }
  fprintf( fp, ")" );
}

void print_scope( FILE *fp, SgLabelSymb *s )
{
  // Scope is enclosing function only
  print_scope( fp, (SgSymbol*)s );
}

void print_scope( FILE *fp, SgFunctionSymb *s )
{
  // Scope is either rest of file or class, depending on type of function
  print_scope( fp, (SgSymbol*)s );
}

void print_scope( FILE *fp, SgFieldSymb *s )
{
  // Scope is the enclosing structure (class, collection, enum, union, struct)
  // and member functions, if applicable.
  print_scope( fp, (SgSymbol*)s );
}

void print_scope( FILE *fp, SgVariableSymb *s )
{
  // Scope is rest of file or rest of block, depending.
  // Formal parameters have the entire function as scope.
  print_scope( fp, (SgSymbol*)s );
}

void print_scope( FILE *fp, SgSymbol *s )
{
  SgStatement *stmt;

  stmt = (SgStatement*)(s->attributeValue( 0, ATTR_DECLAREDSTMT ));
  if (stmt == 0)
    fprintf( fp, "(unknown)" );
  else
    fprintf( fp, "(line %d)", stmt->lineNumber() );
}

void print_type_scope( FILE *fp, SgSymbol *s )
{
  // Scope of a type name, which comes in various forms
  // (typedefs, class names, whatnot).
  //fprintf( fp, "(unknown)" );
  if (s == 0)
    fprintf( fp, "(truly-unknown)" );
  else
    print_scope( fp, s );
}

SgType *clean_type( SgType *t0 )
{
  SgType *t = t0;

  if (t0 == 0) {
    fprintf( stderr, "WARNING: null type passed to clean_type\n" );
    return SgTypeInt();
  }
  do {
    while (t && t->variant() == T_DERIVED_TYPE && t->symbol() != 0)
      t = t->symbol()->type();
    while (t && t->variant() == T_DESCRIPT) {
      long m = isSgDescriptType( t )->modifierFlag();
      if (m == 0 || (m & BIT_TYPEDEF))
	t = t->baseType();
      else
	return t;
    }
  } while (t &&
	   (t->variant() == T_DERIVED_TYPE || t->variant() == T_DESCRIPT));
  
  // SAGEBUG: sometimes there is a symbol with a null type, resulting in a 
  // derived-type pointing to that symbol to have a null base type.
  //
  // It happens with the mysteriously named type 'vconstant', which
  // is used as a placeholder for the [] in a vector delete expression,
  // and with the ElementType type. The ElementType should be handled
  // elsewhere. The vconstant is handled here by returning (int).

  if (t == 0) {
    if (t0 != 0
	&& t0->variant() == T_DERIVED_TYPE
	&& strcmp( clean_name( t0->symbol() ), "vconstant" ) == 0)
      return SgTypeInt();

    if (option_print_warnings) {
      fprintf( stderr, "WARNING: null clean type based on %d (%d)",
	       t0->variant(), t0->isTheElementType() );
      if (t0->symbol() != 0 && t0->symbol()->identifier() != 0)
	fprintf( stderr, "%s", clean_name( t0->symbol() ) );
      fprintf( stderr, "\n" );
    }
    return SgTypeInt();
  }
  else
    return t;
}

const char *clean_name( SgSymbol *s )
{
  if (s == 0) return "*INVALID*";
  else if (s->identifier() == 0) return "*STRANGE*";
  else if (strcmp( s->identifier(), "" ) == 0) return EMPTY_SYM;
  else return s->identifier();
}

const char *mbit( const long flags, const char *def )
{
  static char b[ 100 ];
  *b = 0;
  if (flags & BIT_LONG) strcat( b, "long" );
  if (flags & BIT_SHORT) strcat( b, "short" );
  if (flags & BIT_SIGNED) if (strcmp( def, "s" ) != 0) strcat( b, "s" );
  if (flags & BIT_UNSIGNED) if (strcmp( def, "u" ) != 0) strcat( b, "u" );
  return b;
}

///////////////////////////////////////////////////////////////////////////
//
// Scope and identifier scanner

void scan_stream( FILE *, FILE * );
int keyword( char *, int );

void scanner( char *infile, FILE *outfp )
{
  FILE *fp;

  fp = fopen( infile, "r" );
  if (fp == 0) {
    fprintf( stderr, "ERROR: Unable to open file \"%s\" for scanning\n",
	    infile );
    return;
  }

  scan_stream( fp, outfp );

  fclose( fp );
}

// BUG: does not handle trigraphs.
// BUG: Complex macros are not handled well; we need to deal with \ at
//      the end of the line. The problem is that when a # line is seen,
//      it is discarded. But if such a line starts a macro with a {
//      which is ended on a subsequent line (which _will_ be scanned)
//      then there will be a scope underflow.
// BUG: #ifdefs can cause trouble. Consider an #if #else structure where
//      each branch contains an open brace but the close brace is "shared"
//      after the #endif. In this case we will get an excess of open scopes.
//      One can hack this, but preprocessing is the way to go... or
//      always choose the TRUE branch (ugh!).

void scan_stream( FILE *in, FILE *out )
{
  char buf[ 1024 ], file[ 1024 ];
  char *p;
  int lbegins;
  int line = 0;
  int depth = 0;
  int id_len;
  int id_col;
  int id_line;
  char *q_str;
  char id[ 128 ];
  int scanning_comment = 0;
  int qualify = 0;
  struct { int line, col; } scopes[ 100 ];
  int scopecnt = 0;
  int inparamlist;

#define hasident() (*id != 0)
#define topscope() (scopecnt == 0)

#define flushident() \
  do { \
    if (*id != 0 && inparamlist <= 1) \
      fprintf( out, "(ident \"%s\" %d %d)\n", id, id_line, id_col ); \
    *id = 0; id_len = 0; \
  } while(0)

#define saveident( p, len ) \
  do { int L = (len); id_len += L; strncat( id, (p), (len)+1 ); \
       id[id_len]=0; } while(0)

#define savequalify( s ) (q_str = (s))

#define addqualify() \
  do { if (id_len > 0) { strcat( id, q_str ); id_len += strlen( q_str ); } \
  } while(0)

#define beginparam()  fprintf( out, "(param/begin)\n" )
#define endparam()    fprintf( out, "(param/end)\n" )

  *id = 0;
  id_len = 0;
  inparamlist = 0;
  while (!feof( in )) {
    if (fgets( buf, sizeof( buf ), in ) == 0) break;
    line++;
    p = buf;
    lbegins = 1;

    if (scanning_comment) {
      lbegins = 0;
      while(1) {
	while (*p && *p != '*') p++;
	if (*p == 0) break;
	p++;
	if (*p == '/') {
	  scanning_comment = 0; 
	  break; 
	}
      }
    }

    while (*p) {
      if (lbegins) {
	flushident();
	while (isspace( *p )) p++;
      }
      if (lbegins && *p == '#') {
	p++;
	while (isspace( *p )) p++;
	if (strncmp( p, "line", 4 ) == 0) {
	  //line = 0; // FIXME
	}
	else if (strncmp( p, "file", 4 ) == 0) {
	  // strcpy( file, ... ); // do we need this? Probably. FIXME.
	}
	else if (strncmp( p, "pragma", 6 ) == 0) {
	  // maybe some special action with line numbers FIXME
	}
	*p = 0;
      }
      else if (isalpha( *p ) || *p == '_') {
	char *t = p;
	int c = p - buf;

	p++;
	while (isalnum( *p ) || *p == '_') p++;
	if (!keyword( t, p-t )) {
	  if (qualify) {
	    addqualify();
	    saveident( t, p-t );
	    qualify = 0;
	  }
	  else {
	    // Don't think two idents can follow each other anywhere, so no
	    // flushident() here.
	    id_line = line;
	    id_col = c;
	    saveident( t, p-t );
	  }
	}
	else
	  flushident();
      }
      else if (*p == '{') {
	flushident();
	scopes[scopecnt].line = line;
	scopes[scopecnt].col  = p-buf;
	scopecnt++;
	p++;
      }
      else if (*p == '}') {
	flushident();
	scopecnt--;
	if (scopecnt < 0) {
	  fprintf( stderr, "ERROR: line %d: Scope underflow.\n", line );
	  scopecnt = 0;
	}
	else {
	  fprintf( out, "(scope %d %d %d %d %d)\n", 
		  scopecnt,
		  scopes[scopecnt].line,
		  scopes[scopecnt].col,
		  line, 
		  p-buf );
	}
	p++;
      }
      else if (*p == '"') {
	flushident();
	p++;
	while (*p && *p != '"') {
	  if (*p == '\\') p++;
	  p++;
	}
	if (*p == 0)
	  fprintf( stderr, "ERROR: line %d: Bad string constant: %s\n", 
		  line, buf );
	else
	  p++;
      }
      else if (*p == '\'') {
	flushident();
	p++;
	if (!*p) continue;
	if (*p == '\\') p++;
	p++; 
	if (!*p) continue;
	p++;
      }
      else if (*p == '/') {
	p++;
	if (*p == '/')
	  while (*p) p++;
	else if (*p == '*') {
	  while (1) {
	    while (*p && *p != '*') p++;
	    if (!*p) {
	      scanning_comment = 1;
	      break;
	    }
	    p++;
	    if (*p == '/') break;
	  }
	}
	else
	  flushident();
      }
      else if (*p == '(' && hasident() && topscope() && !inparamlist) {
	// This is the beginning of a parameter list
	flushident();
	inparamlist = 1;
	beginparam();
	p++;
      }
      else if (*p == '(' && inparamlist) {
	// This might be a nested parameter list: if we have an
	// ident, it is. FIXME: must check. And propagate to ) case...
	// what about C (*f)( ... ): the f is not a parameter!
	flushident();
	inparamlist++;
	p++;
      }
      else if (*p == ')' && inparamlist) {
	flushident();
	inparamlist--;
	if (!inparamlist) endparam();
	p++;
      }
      else if (*p == ':' && *(p+1) == ':') {
	p+= 2;
	savequalify( "::" );
	qualify = 1;
      }
      else if (*p == '.' && *(p+1) == '*') {
	p += 2;
	savequalify( ".*" );
	qualify = 1;
      }
      else if (*p == '.' && !isdigit( *(p+1) ) && *(p+1) != '.') {
	p+=1;
	savequalify( "." );
	qualify = 1;
      }
      else if (*p == '-' && *(p+1) == '>') {
	p+= 2;
	savequalify( "->" );
	qualify = 1;
      }
      else if (*p == '-' && *(p+1) == '>' && *(p+2) == '*') {
	p+=3;
	savequalify( "->*" );
	qualify = 1;
      }
      else if (*p != 0) {
	flushident();
	p++;
      }
      lbegins = 0;
    }
  }
  if (scopecnt > 0)
    fprintf( stderr, "ERROR: Not all scopes were closed.\n" );
}


// Is it a C++ keyword?
// The list is from the ARM and probably needs updating. 
// Keep it alphabetic -- that's a binary search you're looking at!

int keyword( char *s, int len )
{
  static char *kwds[] =
    { "asm", "auto", "break", "case", "catch", "char",
      "class", "const", "continue", "default", "delete",
      "do", "double", "else", "enum", "extern", "float",
      "for", "friend", "goto", "if", "inline", "int", "long",
      "new", "operator", "private", "protected", "public",
      "register", "return", "short", "signed", "sizeof", "static", 
      "struct", "switch", "template", "this", "throw",
      "try", "typedef", "union", "unsigned", "virtual",
      "void", "volatile", "while" 
      };
  char buf[ 100 ];
  int first, last, mid, r;
  
  strncpy( buf, s, len );
  buf[len]=0;
  
  first = 0;
  last = sizeof( kwds )/sizeof(kwds[0])-1;
  while (first <= last) {
    mid = (first + last) / 2;
    r = strcmp( buf, kwds[mid] );
    if (r == 0) return 1;
    if (r < 0) last = mid-1; else first = mid+1;
  }
  return 0;
}

///////////////////////////////////////////////////////////////////////////
//
// Utilities

#ifndef __GNUC__
#include <unistd.h>
#else
extern "C" int access(char *path, int amode);
#endif
#ifndef R_OK
#   define R_OK 04
#endif

// Given the name of a project file or a dep file, open the project,
// and create the project if necessary. Terminate the program on error.
//
// This is general; should be used elsewhere too.
// Requires a global variable progname which is the value of argv[0].

SgProject *open_project( char *fn )
{
  SgProject *p;
  char proj_fn[ NAMELEN_MAX ];
  FILE *fp;
  int l;
  int isdep = 0;
  
  strcpy( proj_fn, fn );
  l = strlen( proj_fn );
  if (l > 4 && strcmp( proj_fn+l-4, ".dep") == 0) {
    proj_fn[l-4] = 0;
    strcat( proj_fn, ".proj" );
    isdep = 1;
  }
  
  if (access( proj_fn, R_OK ) == -1) {
    if (!isdep) {
      fprintf( stderr, "%s: project file %s does not exist.\n", 
	      progname,
	      proj_fn );
      exit( 1 );
    }
    fp = fopen( proj_fn, "w" );
    if (fp == 0 || fprintf( fp, "%s\n", fn ) < 0 || fclose( fp ) == EOF) {
      fprintf( stderr,
	      "%s: project file %s does not exist, and can't create.\n", 
	      progname,
	      proj_fn );
      exit( 1 );
    }
  }
  
  p = new SgProject( proj_fn );
  if (p == 0) {
    fprintf( stderr, "%s: can't create project for file: %s\n", 
	    progname,
	    proj_fn );
    exit( 1 );
  }
  
  return p;
}

// FIXME
int panic( char *fmt, ... )
{
  fprintf( stderr, "%s\n", fmt );
  exit( 1 );
  return 1;
}

// eof
