/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/


# include <stdio.h>
# include <stdlib.h>
# include <assert.h>
# include "sage++user.h"
# include "sage++callgraph.h"
# include "regexp.h"
# include "instr.h"


// including files for breakpoint debugging.;
#include "sageUtils.h"
#include "TItypeDesc.h"
  
#ifndef __GNUC__
#  include <unistd.h>
#else
extern "C" int unlink(const char *path);
extern "C" int access(char *path, int amode);
#endif

#include "compatible.h"
#ifdef SYS5
#include <string.h>
#else
#include <strings.h>
#endif

# ifndef R_OK
#   define R_OK 04
# endif

/* -- external functions -------------------------------------------------- */
/* -- from: readcomfile.y ------------------------------------------------- */
extern "C" int readComFile (char *filename);

/* -- from: namecheck.c --------------------------------------------------- */
extern "C" void initCommands (void);
extern "C" int isValidFunc (int minlevel, char *filename,
                            char *classname, char *funcname);
extern "C" int isNotTooSmall (int numstmts, int numchild, int isInline);

/* -- from: gendescr.c ---------------------------------------------------- */
extern "C" int newEventId (char *classname, char *funcname, int tag);
extern "C" void genTdl (char *tdlfile);
extern "C" void genEdf (char *edffile);
extern "C" void genFuncTable (char *tabfile, char *depfile, int* pcxx_tag);

/* -- internal functions -------------------------------------------------- */
static void instrFunctions (SgFile *, char *inFile, char *outFile);
static void instrument_coll_const (SgSymbol * fsymb, SgFile *file);
SgFunctionCallExp *
  insert_call_before( SgStatement *s, SgSymbol *f, int argp=0, int arg=0 );
SgFunctionCallExp *
  insert_call_after( SgStatement *s, SgSymbol *f, int argp=0, int arg=0 );
void instrument_parallel_ops( SgFile *f );

/* -- global variables ---------------------------------------------------- */
int verboseMode = FALSE;
int debugMode = FALSE;
int pcxx_tag[10];
int profileMode = FALSE;
int breakpointMode = FALSE;
int sneezyMode = FALSE;
int traceMode = FALSE;
static int numfunc = 0;

void printUsage (char* execName)
{
  fprintf (stderr,"Usage: %s\n",execName);
  fprintf (stderr,"    [-help | -h] [-usage | -u] [-debug]\n");
  fprintf (stderr,"    [-verbose | -v] [(-commands | -c) comfile]\n");
  fprintf (stderr,"    [(-tdl | -t) tdlfile] [(-o | -output) outfile]\n");
  fprintf (stderr,"    [(-edf | -e) edffile] [-profile | -p]\n");
  fprintf (stderr,"    [-breakpoint | -b]\n");
  fprintf (stderr,"    [-sneezy | -z]\n");
  fprintf (stderr,"    depfile\n");
  fprintf (stderr,
     "\n'%s -help' prints description of switches, parameters\n", execName);
  exit (1);
}

void printHelp (char* execName)
{
  printf ("Usage: %s\n", execName);
  printf ("  [-h | -help]  Print this help text\n");
  printf ("  [-u | -usage]  Print usage information (command line switches)\n");
  printf ("  [-v | -verbose]  Turn on Verbose Mode \n");
  printf ("  [-debug]  Print useful debugging information\n");
  printf ("  [(-c | -commands) comfile]  use instrumentation command file\n");
  printf ("  [(-t | -tdl) tdlfile]  Generate TDL description file\n");
  printf ("  [(-e | -edf) edffile]  Generate event descriptor file\n");
  printf ("  [(-o | -output) outfile]  Output dep file name\n");
  printf ("  [-p | -profile]  instrument for profiling\n");
  printf ("  [-b | -breakpoint]  instrument for breakpoint executive (Breezy)\n");
  printf ("  [-sneezy | -z] instrument for Sneezy\n");
  printf ("  depfile\n");
  exit (1);
}

main (int argc, char **argv)
{
  SgProject *project = 0;
  char *projFile = NULL;
  char *edfFile = NULL;
  char *tdlFile = NULL;
  char *outFile = NULL;
  char *depFile = NULL;
  char *p;
  FILE *pf = 0;
  int l;
  int musthavedep = 0;

  if (argc == 1)
    printUsage (argv[0]);

  initCommands ();

  for(int i = 1; i < argc; i++) {
    if ( argv[i][0] == '-' ) {
      p = argv[i];
      if ( (!strcmp(p,"-help")) || (!strcmp(p,"-h")) ) {
	printHelp (argv[0]);
      }
      else if ( (!strcmp(p,"-usage")) || (!strcmp(p,"-u")) ) {
	printUsage (argv[0]);
      }
      else if ( !strcmp(p,"-debug") ) {
	debugMode = TRUE;
      }
      else if ( (!strcmp(p,"-verbose")) || (!strcmp(p,"-v")) ) {
	verboseMode = TRUE;
      }
      else if ( (!strcmp(p,"-profile")) || (!strcmp(p,"-p")) ) {
	profileMode = TRUE;
	numfunc = 10;
	musthavedep = 1;
      }
      else if ( (!strcmp(p,"-breakpoint")) || (!strcmp(p,"-b")) ) {
	breakpointMode = 1;
	musthavedep = 1;
      }
      else if (strcmp( p, "-sneezy" ) == 0 || strcmp( p, "-z" ) == 0) {
	sneezyMode = TRUE;
	musthavedep = 1;
      }
      else if ( (!strcmp(p,"-commands")) || (!strcmp(p,"-c")) ) {
	if ( i == (argc-1) || *argv[i+1] == '-' ) {
	  fprintf (stderr,"ERROR: command file name required for -c switch\n");
	  printUsage (argv[0]);
	}
	i++;
	if ( verboseMode ) printf ("reading commands file %s\n", argv[i]);
	if ( ! readComFile (argv[i]) ) exit (1);
      }
      else if ( (!strcmp(p,"-output")) || (!strcmp(p,"-o")) ) {
	if ( i == (argc-1) || *argv[i+1] == '-' ) {
	  fprintf (stderr,"ERROR: output file name required for -o switch\n");
	  printUsage (argv[0]);
	}
	i++;
	outFile = argv[i];
      }
      else if ( (!strcmp(p,"-tdl")) || (!strcmp(p,"-t")) ) {
	if ( i == (argc-1) || *argv[i+1] == '-' ) {
	  fprintf (stderr,"ERROR: file name required for -t switch\n");
	  printUsage (argv[0]);
	}
	i++;
	tdlFile = argv[i];
	musthavedep = 1;
      }
      else if ( (!strcmp(p,"-edf")) || (!strcmp(p,"-e")) ) {
	if ( i == (argc-1) || *argv[i+1] == '-' ) {
	  fprintf (stderr,"ERROR: file name required for -e switch\n");
	  printUsage (argv[0]);
	}
	i++;
	edfFile = argv[i];
	musthavedep = 1;
      }
      else {
	fprintf (stderr,"ERROR: unknown option `%s'\n", p);
	printUsage (argv[0]);
      }
    }
    else if (depFile) {
      fprintf( stderr, "%s: Warning: Extraneous argument %s ignored.\n", 
	      argv[0],
	      argv[i] );
    }
    else {
      /* The basic idea: if there is a file name, then it should name the
       * depFile to use.  There may be no extension or the extension should
       * be ".dep"; we construct the full name of the depfile here.
       *
       * We also need to know the name of the project file.  This is derived
       * from the name of the depfile by changing the .dep to .proj.
       */
      depFile = argv[i];
      l = strlen(argv[i]);
      p = argv[i] + l - 4;
      if ( l < 4 || strcmp (p, ".dep") != 0 ) {
	fprintf (stderr,"ERROR: `%s': no dep file\n", argv[i]);
	exit (1);
      }
      else {
	*p = '\0';
	projFile = new char[l+2];
	sprintf (projFile, "%s.proj", argv[i]);
	*p = '.';
      }

      if ( access (projFile, R_OK) == -1 ) {
	pf = fopen (projFile, "w");
	fprintf (pf, "%s\n", depFile);
	fclose (pf);
      }

      if ( verboseMode || debugMode )
	printf ("using project file %s\n", projFile);
      project = new SgProject (projFile);
    }
  }

  if (musthavedep && !depFile) {
    fprintf( stderr, "ERROR: depfile required for the given option(s).\n" );
    printUsage( argv[0] );
  }

  if ( depFile ) {
    if ( ! outFile ) outFile = depFile;

    if ( verboseMode || debugMode )
      printf ("processing %s:\n", project->fileName(0));

    // -----------------------------------------------------------
    // creating type description file;
    if (breakpointMode) {

      // initialize sage items for breakpointing..;
      brk_mySageInit();

      SgFile *f;
      for (int j=0; j<project->numberOfFiles(); j++) {
	if (debugMode) {
	  printf("\t\tINSTRUMENTING PROJECT: %s\n\n", project->fileName(j));
	}
	f = &(project->file(j));
	brk_getCollections(f);
      }

      // generate
      // type description file and access functions.
      // write type description file...;
      brk_printTIFFile();
    }

    if (sneezyMode) {
      // FIXME: Sneezy
    }

    if (!profileMode && !sneezyMode && !breakpointMode) {
      if (edfFile == 0 && tdlFile == 0) {
	fprintf( stderr, "ERROR: no command.\n" );
	exit( 1 );
      }
      else
	traceMode = TRUE;
    }

    // -----------------------------------------------------------;

    instrFunctions (&(project->file(0)), project->fileName(0), outFile);

    if ( verboseMode ) printf ("\tgenerating dep file %s\n", outFile);
    project->file(0).saveDepFile(outFile);

    if ( pf ) unlink(projFile);
  }

  if ( profileMode )
  {
    pcxx_tag[1] = -2;
    pcxx_tag[2] = -2;
    pcxx_tag[3] = -2;

    if ( verboseMode ) printf ("generating functablefile profile.ftab\n");
    SgSymbol *bfunc = (project->file(0)).firstSymbol();
    while ( bfunc )
    {
      if ( isSgFunctionSymb(bfunc) )
      {
        if ( strcmp (bfunc->identifier(), "pcxx_Barrier") == 0 )
          pcxx_tag[1] = bfunc->id();
        else if ( strcmp (bfunc->identifier(), "tulip_Poll") == 0 )
          pcxx_tag[2] = bfunc->id();
        else if ( strcmp (bfunc->identifier(), "WhereIsData") == 0 )
          pcxx_tag[3] = bfunc->id();
      }
      bfunc = bfunc->next();
    }
    genFuncTable ("profile.ftab", outFile, pcxx_tag);
  }
  else if (traceMode) {
    if ( tdlFile ) {
      if ( verboseMode ) printf ("generating TDL file %s\n", tdlFile);
      genTdl (tdlFile);
    }

    if ( edfFile ) {
      if ( verboseMode ) printf ("generating edf file %s\n", edfFile);
      genEdf (edfFile);
    }
  }
  return 0;
}

/* For Sneezy, we need to instrument:
 *  - function entry and exit
 *  - begin and end parallel section
 *  - enter and exit element
 * Everything else should be taken care of in the RTS, _unless_ the 
 * translator does something magic with GetElement on shared-memory
 * systems.
 */
void instrFunctions (SgFile *file, char *inFile, char *outFile)
{
  SgCallGraph CG;
  SgDerivedType *profType;
  SgDerivedType *snzType;

  assert( profileMode || breakpointMode || sneezyMode || traceMode );

  if ( file->languageType() != CSrc ) {
    if (verboseMode || debugMode)
      fprintf( stderr, 
	      "Instrumenter pC++/instr: WARNING: Fortran not supported.\n" );
    return;
  }

  /* -- Initialize call graph module                             -- */
  /* -------------------------------------------------------------- */
  CG.GenCallTree(file);
  CG.computeClosures();

  if (profileMode || traceMode) {
    /* -- Search for class Profiler symbol                         -- */
    /* -------------------------------------------------------------- */
    SgSymbol *pclass = file->firstSymbol();
    while ( pclass ) {
      if ( isSgClassSymb(pclass) &&
	  (strcmp (pclass->identifier(), "Profiler") == 0) )
	break;
      pclass = pclass->next();
    }
    if ( !pclass ) {
      fprintf (stderr,"ERROR: `%s': class Profiler declaration missing\n",
	       inFile);
      exit (1);
    }
    profType = new SgDerivedType(*pclass);
  }

  /* -- Declarations for instrumenting functions                 -- */
  /* -------------------------------------------------------------- */
  SgStatement *func, *mainfunc, *s;
  SgSymbol *funcsymb, *classsymb, *profsymb, *breakpointsymb = NULL;
  SgMemberFuncSymb *memsymb;
  char *funcname, *classname, *filename;
  SgExpression *e1, *e2, *e3, *e, *args;
  int funcline;
  int isMoE, isColl;
  int numchild, numstmts, res, isInline, minlevel;
  SgSymbol *funcentry_symb = 0, *funcexit_symb = 0;

  if (sneezyMode) {
    funcentry_symb = 
      new SgFunctionSymb( FUNCTION_NAME, 
			 "snz_enter_function_event", 
			 *SgTypeVoid(),
			 *(file->firstStatement()) // GLOBAL node
			 );
    funcexit_symb =
      new SgFunctionSymb( FUNCTION_NAME,
			 "snz_exit_function_event",
			 *SgTypeVoid(),
			 *(file->firstStatement()) // GLOBAL node
			 );

    SgSymbol *zclass = file->firstSymbol();
    while ( zclass ) {
      if ( isSgClassSymb(zclass) &&
	  (strcmp (zclass->identifier(), "SneezyEntry") == 0) )
	break;
      zclass = zclass->next();
    }
    if ( !zclass ) {
      fprintf (stderr,"ERROR: `%s': class SneezyEntry declaration missing\n",
	       inFile);
      exit (1);
    }
    snzType = new SgDerivedType(*zclass);
  }

  /* -- skip through all functions off the file                  -- */
  /* -------------------------------------------------------------- */
  for (int i=0; i<file->numberOfFunctions(); i++)
    {
      /* -- determine type of function                             -- */
      /* ------------------------------------------------------------ */
      func = file->functions(i);
      funcsymb = func->symbol();
      funcname = funcsymb->identifier();
      funcline = func->lineNumber();
      filename = func->fileName();
      isMoE    = 0;

      if ( memsymb = isSgMemberFuncSymb(funcsymb) ) {
	classsymb = memsymb->className();
	classname = classsymb->identifier();
	if ( (isColl = (classsymb->variant() == COLLECTION_NAME)) )
	  isMoE  = memsymb->isMethodOfElement();
	
	if ( debugMode )
	  printf ("\t%4d:  %s%s%s::%s", funcline, (isColl ? "[Coll] " : ""),
		  (isMoE ? "[MoE] " : ""), classname, funcname);
      }
      else {
	classname = 0;
	isColl    = 0;

	if ( debugMode )
	  printf ("\t%4d:  %s", funcline, funcname);
      }


      /* -- determine "size" of function                           -- */
      /* ------------------------------------------------------------ */
      minlevel = CG.minLevel(funcsymb);
      numchild = CG.numberOfFunsCalledFrom(funcsymb);
      numstmts = 0;
      s = func;
      while ( s != func->lastNodeOfStmt() ) {
	numstmts++;
	s = s->lexNext();
      }
      if (funcsymb != funcentry_symb && funcsymb != funcexit_symb) {
	// WARNING: complexity of declaredInStmt is linear in size of scope!
	isInline = funcsymb->declaredInStmt()->isInline();
      }
      if ( debugMode ) 
	  printf (" <%ds/%dc> %s\n", numstmts, numchild, isInline?"inline":"");

      /* -- check whether function should be instrumented          -- */
      /* ------------------------------------------------------------ */
      if ( profileMode && (strcmp (funcname, "Processor_Main") == 0) ) {
	mainfunc    = func;
	pcxx_tag[0] = funcsymb->id();
      }
      else if ( strcmp (funcname, "PROFILE_COLL_NAMES") == 0 ) {
	profsymb = funcsymb;
      }
      else if (breakpointMode && 
	       strcmp (funcname, "brk_QUE_COLLECTION_INSTANCE") == 0 ) {
	if ( debugMode ) {
	  printf("found add_collection_instance symbol %s\n",
		 funcname);
	}
	breakpointsymb = funcsymb;
      }
      else if (sneezyMode && 
	       (funcsymb == funcentry_symb || funcsymb == funcexit_symb)) {
	// ignore
      }
      else if (sneezyMode &&
	       isValidFunc( minlevel, filename, classname, funcname ) &&
	       strcmp(funcname, "SneezyEntry") != 0 &&
	       strcmp(funcname, "~SneezyEntry") != 0 &&
	       (res = isNotTooSmall( numstmts, numchild, isInline )) ) {
	if (verboseMode) {
	  printf( "\tinstrumenting (for Sneezy) ");
	  if ( classname ) printf( "%s::", classname );
	  printf( "%s\n", funcname);
	}

	// This does not deal well with function exit.
	// insert_call_after(func, funcentry_symb, 1, funcsymb->id());

	SgSymbol *pcxx_Z = new SgVariableSymb("pcxx_Z", *snzType);
	pcxx_Z->declareTheSymbol(*func);
	e1 = new SgValueExp(funcsymb->id());

	SgExpression *classinit = new SgExpression(CLASSINIT_OP); 
	classinit->setRhs(*new SgExprListExp(*e1)); 
	classinit->setLhs(*new SgVarRefExp(*pcxx_Z));
	func->lexNext()->expr(0)->setLhs(*classinit);
      }
      else if ((profileMode || traceMode) &&
	       isValidFunc (minlevel, filename, classname, funcname) &&
	       (res = isNotTooSmall (numstmts, numchild, isInline)) ) {
	if ( debugMode ) printf ("\n");
	if ( verboseMode ) {
	  if ( classname )
	    printf ("\tinstrumenting %s::%s", classname, funcname);
	  else
	    printf ("\tinstrumenting %s", funcname);

	  if ( res == COUNT_ONLY )
	    printf (" [counting only]\n");
	  else
	    printf ("\n");
	}

	/* -- declare variable pcxx_P of type class Profiler       -- */
	/* ---------------------------------------------------------- */
	SgSymbol *pcxx_P = new SgVariableSymb("pcxx_P", *profType);
	pcxx_P->declareTheSymbol(*func);

	/* -- build 1. argument for constructor (eventID)          -- */
	/* ---------------------------------------------------------- */
	SgVariableSymb IDsymb("Ident");
	numfunc = newEventId (classname, funcname, funcsymb->id());
	e1 = new SgValueExp(numfunc);

	if ( ! profileMode ) {
	  /* -- build 2. Argument for MemberOfElement functions    -- */
	  /* -- normal functions get argument "0".                 -- */
	  /* -------------------------------------------------------- */
	  if ( isMoE )
	    e2 = new SgVarRefExp(IDsymb);
	  else
	    e2 = new SgValueExp(0);

	  /* -- build 3. Argument: id of function symbol           -- */
	  /* -------------------------------------------------------- */
	  e3 = new SgValueExp(funcsymb->id());

	  e = new SgExprListExp(*e3);
	  args = &SgExprListOp(*e2, *e);
	  args = &SgExprListOp(*e1, *args);
	}
	else if ( res == COUNT_ONLY ) {
	  /* -- build 2. Argument (constant char) so that          -- */
	  /* -- "count only" constructor is used.                  -- */
	  /* -------------------------------------------------------- */
	  e2 = new SgValueExp('C');
	  e = new SgExprListExp(*e2);
	  args = &SgExprListOp(*e1, *e);
	}
	else
	  args = new SgExprListExp(*e1);

	/* -- finish building "pcxx_P" constructor arguments and   -- */
	/* -- add them                                             -- */
	/* ---------------------------------------------------------- */
	SgExpression *classinit = new SgExpression(CLASSINIT_OP); 
	classinit->setRhs(*args); 
	SgExpression *pcxx_P_ref = new SgVarRefExp(*pcxx_P);
	classinit->setLhs(*pcxx_P_ref);
	SgStatement *st = func->lexNext();
	SgExpression *list = st->expr(0);
	list->setLhs(*classinit);
      }
      else if ( debugMode ) {
	printf ("\t++++++++++++++++++++ NO\n");
      }
    }

  if ( profileMode ) {
    /* -- instrument Processor_Main: 1.arg (0), 2.arg (numfunc)  -- */
    /* -- in the same way                                        -- */
    /* ------------------------------------------------------------ */
    SgSymbol *pcxx_P = new SgVariableSymb("pcxx_P", *profType);
    pcxx_P->declareTheSymbol(*mainfunc);
    e1 = new SgValueExp(0);	// 0
    e2 = new SgValueExp(numfunc+1); // n (= numfunc+1)
    e = new SgExprListExp(*e2);     // (n)
    args = &SgExprListOp(*e1, *e);  // (0, n)
    SgExpression *classinit = new SgExpression(CLASSINIT_OP); 
    classinit->setRhs(*args); 
    SgExpression *pcxx_P_ref = new SgVarRefExp(*pcxx_P);
    classinit->setLhs(*pcxx_P_ref);
    SgStatement *st = mainfunc->lexNext();
    SgExpression *list = st->expr(0);
    list->setLhs(*classinit);
  }
  
  if ( breakpointMode) {
    // only when breakpointMode == 2 do we need to instrument; (? --lth)
    instrument_coll_const(breakpointsymb, file);
  } else if (traceMode || profileMode ) {
    // for tracing and profiling;
    instrument_coll_const(profsymb, file);
  } else if (sneezyMode) {
    instrument_parallel_ops( file );
  }
}


/* FIXME: This does not catch "new C" in contexts other than on the rhs
 * of an assignment statement, notably not in contexts like
 *   ... f( new C ) ...
 *   if ((g = new C) == 0) ...  (and ditto ?: expression, and for/while)
 *   v = (cast_to_base)new C
 * FIXME: A lot of shared code here. Where are nested procedures when
 * we need them?
 */
void instrument_coll_const (SgSymbol * fsymb, SgFile *file) 
{
  
  /* fsymb is function symbol to insert at collection construction points */
  
  /* -- Instrument all collection variable declarations          -- */
  /* -------------------------------------------------------------- */
  SgSymbol *classsymb;
  char *classname;
  SgFunctionCallExp *call;
  SgExprListExp *exp;
  SgCExpStmt *c;
  SgValueExp *v;
  SgType *elemType;
  SgDerivedCollectionType *d;
  char *elemname;
  SgVarDeclStmt *var;
  SgCExpStmt *ass;
  SgSymbol *varsymb;
  SgExprListExp *elist;
  SgStatement *s;
  
  for ( s = file->firstStatement() ; s ; s = s->lexNext() ) {
    if ( var = isSgVarDeclStmt (s) ) {
      for (int i=0; i<var->numberOfSymbols(); i++) {
	varsymb = var->symbol(i);
	if ( d = isSgDerivedCollectionType (varsymb->type()) ) {
	  /* -- Symbol is of type CollectionType<ElementType>    -- */
	  /* -- get type names                                   -- */
	  /* ------------------------------------------------------ */
	  classsymb = d->collectionName();
	  classname = classsymb->identifier();
	  elemType  = d->elementClass();
	  elemname  = (elemType->symbol())->identifier();
	  if ( verboseMode )
	    printf ("\tfound %s<%s> %s\n",
		    classname, elemname, varsymb->identifier());
		  
	  /* -- Declare and insert call to function passed in <fsymb> -- */
	  /* -- pass all type and variable names                 -- */
	  /* ------------------------------------------------------ */
	  call = new SgFunctionCallExp(*fsymb);
	  v = new SgValueExp(classname);
	  call->addArg(*v);
	  delete(v);
	  v = new SgValueExp(elemname);
	  call->addArg(*v);
	  delete(v);
	  v = new SgValueExp(varsymb->identifier());
	  call->addArg(*v);
	  delete(v);
		  
	  exp = new SgExprListExp();
	  exp->setValue(*call);
	  c = new SgCExpStmt (*exp);
		  
	  var->insertStmtBefore(*c);
	}
      }
    }
    else if ( (ass=isSgCExpStmt (s)) && (elist=isSgExprListExp(ass->expr())) ){
      SgExpression *e;
      SgNewExp *n;
      SgPointerType *t;
      char *vname;
	  
      for (int i=0; i<elist->length(); i++) {
	e = elist->elem(i);
	if ((e->variant() == ASSGN_OP) &&
	    (n = isSgNewExp(e->rhs())) &&
	    (t = isSgPointerType(n->type())) &&
	    (d = isSgDerivedCollectionType(t->baseType())) ) {
	  /* -- dynamic collection allocation Statement:         -- */
	  /* -- var = new CollectionType<ElementType>()          -- */
	  /* ------------------------------------------------------ */
	  classsymb = d->collectionName();
	  classname = classsymb->identifier();
	  elemType  = d->elementClass();
	  elemname  = (elemType->symbol())->identifier();
	  vname     = new char[strlen(e->lhs()->symbol()->identifier())+2];
	  sprintf (vname, "*%s", e->lhs()->symbol()->identifier());
		  
	  if ( verboseMode )
	    printf ("\tfound %s<%s> %s\n", classname, elemname, vname);
		  
	  /* -- Declare and insert call to PROFILE_COLL_NAMES    -- */
	  /* -- pass all type and variable names                 -- */
	  /* ------------------------------------------------------ */
	  call = new SgFunctionCallExp(*fsymb);
	  v = new SgValueExp(classname);
	  call->addArg(*v);
	  delete(v);
	  v = new SgValueExp(elemname);
	  call->addArg(*v);
	  delete(v);
	  v = new SgValueExp(vname);
	  call->addArg(*v);
	  delete(v);
		  
	  exp = new SgExprListExp();
	  exp->setValue(*call);
	  c = new SgCExpStmt (*exp);
		  
	  ass->insertStmtBefore(*c);
	}
      }
    }
  }
}


/* Discovering whether an expression contains a collection invocation.
 * A collection invocation has the form X.func( args ) where X is a 
 * collection, or presumably (*X).func( args ) where X is a pointer to
 * a collection, X->func( args ) ditto, and so on.
 *
 * Note that X(e)->func( args ) is not a parallel invocation, whereas
 * X[e:e:e]->func( args ) _is_ one (or was that "."?) (but is that supported?)
 *
 * Following are some examples and anti-examples.

X.phello(argc,argv) 

becomes

2160-B (expr-stmt-node)
  |
11419-E (expr-list)
  |
11414-E (record-ref) - 11413-E (func-call "phello")
  |                      |
  |                    11416-E (expr-list) - 11418-E (expr-list)
  |                      |                      |
  |                      |                   11417-E (var-ref "argv")
  |                      |  
  |                    11415-E (var-ref "argc")
  |
11412-E (var-ref "X", type 1979-T (T_DERIVED_COLLECTION MyCollection))


(*Y).phello(argc,argv)

becomes

2086-B
  |
11440-E (expr-list)
  |
11435-E (record-ref) - 11434-E (func-call 1589-S)
  |                      |
  |                    11437-E (expr-list) - 11439-E (expr-list)
  |                      |                     |
  |                      |                   11438-E (var-ref 1597-S)
  |                      |
  |                    11436-E (var-ref 1596-S)
  |
11433-E (expr-list 1986-T)
  |
11432-E (deref-op 1986-T)
  |
11431-E (var-ref 1604-S 1984-T)


Y->phello(argc,argv)

becomes

2087-B (expr-stmt-node)
  |
11448-E (expr-list)
  |
11443-E (pointst_op) - 11442-E (func-call)
  |                      |
  |                    11445-E (expr-list) - 11447-E (expr-list)
  |                      |                     |
  |                      |                   11446-E (var-ref 1597-E)
  |                      |
  |                    11444-E (var-ref 1596-S)
  |
11441-E (var-ref 1604-S 1984-T (= ptr to derived))


---------------------------------------------------------------------------

Some anti-examples:

X(1)->hello()

becomes

2084-B (expr-stmt-node)
  |
11430-E (expr-list)
  |
11425-E (pointst_op) - 11424-E (func-call "hello")
  |
11423-E (function_op 1982-T == ptr 1978-T = "MyElement") - 11422-E (expr-list)
  |                                                          |
  |                                                        11421-E (int-val 1)
  |
11420-E (var-ref "X" type 1979-T)


(*X(1)).hello()

becomes

2085-B (expr-stmt-node)
  |
11443-E (expr-list)
  |
11438-E (record-ref) - 11437-E (func-call "hello")
  |
11436-E (expr-list 1984-T)
  |
11435-E (deref-op 1984-T)
  |
11434-E (function-op 1983-T) - 11433-E (expr-list)
  |                              |
  |                            11432-E (int-val 1)
  |
11431-E (var-ref "X" 1979-T)

Taken together, these lead us to believe that the expressions we are
looking for are either record-ref or pointst-op where the truetype of
the lhs is a derived collection type or pointer to derived collection
type and the rhs is a func-call.

There is one special case, namely when the collection is the type of the
symbol ThisCollection.  That's only valid in parallel sections and
implies a non-parallel call. (I think.)

*/

SgType *ThisCollectionType = 0;

int has_parallel_op( SgExpression *e )
{
  SgExpression *f;

  f = e;
  if (e == 0) return 0;
  if (e->variant() == RECORD_REF && e->rhs()->variant() == FUNC_CALL) {
    SgType *type;

    e = e->lhs();
    while(e && e->variant() == EXPR_LIST) e = e->lhs();
    if (!e) goto other;

    type = findTrueType(e);
    if (!type) goto other;

    return isSgDerivedCollectionType(type) != 0;
  }
  else if (e->variant() == POINTST_OP && e->rhs()->variant() == FUNC_CALL) {
    SgType *type;

    e = e->lhs();
    while(e && e->variant() == EXPR_LIST) e = e->lhs();
    if (!e) goto other;

    type = findTrueType(e);
    if (!type) goto other;

    return
      isSgPointerType(type) && 
      isSgDerivedCollectionType(type->baseType()) &&
      type != ThisCollectionType;
  }
 other:
  return has_parallel_op( f->lhs() ) || has_parallel_op( f->rhs() );
}


/* For each parallel operation, bracket it with
     snz_begin_parallel_event()
     ...
     snz_end_parallel_event()
   We could pass the collection ID here if it's useful.

   We also want to instrument with snz_enter_element and snz_exit_element
   but that does not seem feasible at this point, since the code for the
   implied loop is emitted later, during unparsing.  The unparser would have
   to do it, or we would have to introduce proxy element classes
   in the call chain, which seems extremely ugly.
 */
void instrument_parallel_ops( SgFile *f )
{
  SgSymbol *parentry_symb, *parexit_symb;

  assert( sneezyMode );

  parentry_symb = 
    new SgFunctionSymb( FUNCTION_NAME, 
		       "snz_begin_parallel_event",
		       *SgTypeVoid(),
		       *(f->firstStatement()) // GLOBAL node
		       );
  parexit_symb =
      new SgFunctionSymb( FUNCTION_NAME,
			 "snz_end_parallel_event",
			 *SgTypeVoid(),
			 *(f->firstStatement()) // GLOBAL node
			 );

  { /* Not all C++ compilers cope with scoping in a 'for' stmt. */

    /* Find the type of ThisCollection and remember it. */
    for ( SgSymbol *s = f->firstSymbol() ; s ; s = s->next() ) {
      if (s->identifier() && strcmp( s->identifier(), "ThisCollection" ) == 0){
	ThisCollectionType = s->type();
	break;
      }
    }
  }

  /* This loop checks every expression of every statement for a parallel
     call and brackets the statement if one is found.  There are subtleties
     ignored by this code, particularly parallel calls in the expressions
     of loop headers.  Consider:
        while ( X.doit(++i) )
	  ;
     which can be perfectly meaningful; the effect here would be to bracket
     the while statement with the _enter_ outside and the _exit_ inside,
     and that is wrong, very wrong.  FIXME.
     (Again, the effect can more easily be had by either instrumenting
      the sage libraries to do the enter/exit, or by considering the
      problem during unparsing, when more information is known.)
     (Anyway, the current pC++ compiler proceduces illegal code for the
      construct so I refuse to worry.)
   */
  for ( SgStatement *s = f->firstStatement() ; s ; s = s->lexNext() ) {
    if (has_parallel_op( s->expr(0) ) ||
	has_parallel_op( s->expr(1) ) ||
	has_parallel_op( s->expr(2) ) ) {
      insert_call_before( s, parentry_symb );
      insert_call_after( s, parexit_symb );
    }
  }
}

SgFunctionCallExp *
insert_call_before( SgStatement *s, SgSymbol *f, int argp, int arg )
{
  SgFunctionCallExp *call;

  call = new SgFunctionCallExp( *f );
  call->setType( SgTypeVoid() );
  if (argp) {
    SgValueExp *p;
    call->addArg( *(p = new SgValueExp( arg )) );
    delete p;
  }
  s->insertStmtBefore( *(new SgCExpStmt( *call ) ) );
  return call;
}

SgFunctionCallExp *
insert_call_after( SgStatement *s, SgSymbol *f, int argp, int arg )
{
  SgFunctionCallExp *call;

  call = new SgFunctionCallExp( *f );
  call->setType( SgTypeVoid() );
  if (argp) {
    SgValueExp *p;
    call->addArg( *(p = new SgValueExp( arg )) );
    delete p;
  }
  s->insertStmtAfter( *(new SgCExpStmt( *call ) ) );
  return call;
}

/* eof */
