/* m e r g e f . c               -- merge files based on an unparse map.
 * $Id: mergef.c,v 1.1 1996/01/08 23:33:25 lth Exp $
 *
 * lth@cs.uoregon.edu / January 5, 1996
 * Copyright (C) 1996 The University of Oregon.
 *
 * DESCRIPTION
 *   The goal is to take one file (the one named by the 'S' line in the map
 *   file) and its translation (temp.C) and merge them in such a way that
 *   the original source and the translation are intermingled in a reasonable
 *   way.
 *
 * USAGE
 *   mergef [ mapfile ] [ tempfile ]
 * where mapfile defaults to 'unparse.map' and tempfile defaults to 'temp.C'.
 * Mapfile must have extension '.map'; tempfile must have extension '.C'.
 *
 * OUTPUT FORMAT
 *   The resulting file contains directives which tell where the following
 *   lines (up to the next directive) came from: "## 0" means the lines came
 *   from the souce file, "## 1" means they came from the translated file.
 *   You can reconstruct the contents of both by running the following awk
 *   script:

   #!/bin/awk -f
   #
   # Use this program to detangle the output from mergef.  
   #
   # The default is to extract the original pC++ program (key == 0).  
   # You can however extract the generated program by giving the argument 
   # key=1 on the command line when running extract.

   BEGIN      { if (key != "") key = 0; state = 0; }
   /^\#\# 0/  { state = 0; next; }
   /^\#\# 1/  { state = 1; next; }
              { if (state == key) print }


 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <string.h>

char *progname;

int error( const char * );
void cannot( char *what, char *name );
int scan_to( FILE *in, int *line, int limit, FILE *out, int *lastwas, int now);

int main( int argc, char **argv )
{
  struct tbl_s {
    int tpos;
    int fpos;
  };

  FILE *tempfp, *mainfp, *mapfp;
  char *mapfile = "unparse.map";
  char *tempfile = "temp.C";
  char currfile[ 1024 ], mainfile[ 1024 ], line[ 1024 ];
  int last_templine, last_mainline, is_mainfile, lastwas, i, l;

  struct tbl_s *tbl;
  int tblcnt, tblsize, tblp;

  /* Parse arguments */

  for ( i = 1 ; i < argc ; i++ ) {
    l = strlen( argv[i] );
    if (i >= 2 && strcmp( argv[i]+l-2, ".C" ) == 0)
      tempfile = argv[i];
    else if (i >= 4 && strcmp( argv[i]+l-4, ".map" ) == 0)
      mapfile = argv[i];
    else {
      fprintf( stderr, "%s: Can't grok this argument: \"%s\"\n", 
	      argv[0], argv[i] );
      fprintf( stderr, "\nUsage: %s [ mapfile ] [ tempfile ]\n", argv[0] );
      exit( 1 );
    }
  }

  progname = argv[0];

  /* Loop over the mapping file, setting up the table of mappings. 
   * We throw away all mappings not directly related to the main file,
   * as they are not needed by this algorithm.
   */
  if ((mapfp = fopen( mapfile, "r" )) == 0) cannot( "fopen", mapfile );

  *currfile = 0;
  *mainfile = 0;
  is_mainfile = 0;

  tblcnt = 0;
  tblsize = 100;
  (tbl = (struct tbl_s *)malloc( 100*sizeof( struct tbl_s ) )) 
    || error( "malloc" );

  while (fgets( line, sizeof( line ), mapfp ) != 0) {
    if (*line == 'S') {
      sscanf( line, "S \"%[^\"]s\"", mainfile );
      is_mainfile = 1;
    }
    else if (*line == 'F') {
    AGAIN:
      sscanf( line, "F \"%[^\"]s\"", currfile );
      is_mainfile = (strcmp( currfile, mainfile ) == 0);
      if (!is_mainfile) {
	char *r;
	while ((r = fgets( line, sizeof( line ), mapfp )) != 0 && *line != 'F')
	  ;
	if (r) goto AGAIN;
      }
    }
    else if (*line == 'L') {
      int tpos, fpos;

      tblcnt++;
      sscanf( line, "L %d %d", &tpos, &fpos );
      if (tblcnt > tblsize) {
	tblsize *= 2;
	(tbl = (struct tbl_s *)realloc( tbl, tblsize*sizeof( struct tbl_s ) ))
	  || error( "malloc" );
      }
      tbl[tblcnt-1].tpos = tpos;
      tbl[tblcnt-1].fpos = fpos;
    }
    else if (*line == '#' || *line == '\n')
      ;
    else {
      fprintf( stderr, "%s: garbage line: <%s>\n", progname, line );
      exit( 1 );
    }
  }

  fclose( mapfp );

#ifdef DEBUG
  fprintf( stderr, "%d table entries\n", tblcnt );
  for ( tblp = 0 ; tblp < tblcnt ; tblp++ ) {
    fprintf( stderr, "L %d %d\n", tbl[tblp].tpos, tbl[tblp].fpos );
  }
#endif

  if ((tempfp = fopen( tempfile, "r" )) == 0) cannot( "open", tempfile );
  if ((mainfp = fopen( mainfile, "r" )) == 0) cannot( "fopen", mainfile );

  /* Loop over the two files and merge them. */

  tblp = 0;
  last_mainline = 0;
  last_templine = 0;
  lastwas = -1;
  while ( tblp < tblcnt ) {
    scan_to( mainfp, &last_mainline, tbl[tblp].fpos, stdout, &lastwas, 0 );
    scan_to( tempfp, &last_templine, tbl[tblp].tpos, stdout, &lastwas, 1 );
    tblp++;
  }

  /* Get the tails of both files. */

  scan_to( mainfp, 0, 0, stdout, &lastwas, 0 );
  scan_to( tempfp, 0, 0, stdout, &lastwas, 1 );

  fclose( mainfp );
  fclose( tempfp );

  exit( 0 );
}

int scan_to( FILE *in, int *line, int limit, FILE *out, int *lastwas, int new )
{
  char text[ 1024 ];
  int n = 0;

  limit--;

  if (*lastwas != new && (!line || *line < limit)) {
    fprintf( out, "## %d\n", new );
    *lastwas = new;
  }
  while (!line || *line < limit) {
    if (fgets( text, sizeof( text ), in ) == 0)
      if (!line) 
	break; 
      else 
	error( "fgets" );
    fputs( text, out );
    if (line) (*line)++;
    n++;
  }
  return n;
}

void cannot( char *what, char *name )
{
  fprintf( stderr, "%s: Can't %s \"%s\".\n", progname, what, name );
  exit( 1 );
}

int error( const char *str )
{
  perror( str );
  exit(0);
  /*NOTREACHED*/ /* return value is int to facilitate an idiom */
  return 0;
}

/* eof */
