%{

# include <stdio.h>
# include "regexp.h"
# include "instr.h"

/* We may use builtin alloca */
#include "compatible.h"
#ifdef _NEEDALLOCAH_
#  include <alloca.h>
#endif

static char *comfile;             /* name of command file        */
static char *errstring;           /* token that caused the error */
static int lineno  = 0;           /* input linenumber            */
static FILE *infile;              /* input filepointer           */
static int noError = TRUE;        /* input error indicator       */

extern void addCommand (int command, REXP *file, REXP *class, REXP *func);

%}

%union
{
  struct
  {
    char *string;
    int   num;
  }
  term;

  struct
  {
    REXP *file;
    REXP *class;
    REXP *func;
  }
  nonterm;

  REXP *rexp;

  LIMIT lmt;
}

%token <term> T_AND
%token <term> T_CHILDREN
%token <term> T_CLASS
%token <term> T_COUNT
%token <term> T_EXCLUDE
%token <term> T_FILE
%token <term> T_FUNCTION
%token <term> T_INCLUDE
%token <term> T_INLINE
%token <term> T_LEVEL
%token <term> T_ONLY
%token <term> T_OR
%token <term> T_SMALLER
%token <term> T_STATEMENTS
%token <term> T_THAN
%token <term> T_TO
%token <term> T_UP
  
%token <term> Identifier
%token <term> Number
%token <term> RegExpression
%token <term> T_EOC
%token <term> T_COLON
%token <term> T_DBL_COLON

%type <term> ExclInclCommand
%type <term> LogOp

%type <nonterm> Parameter
%type <nonterm> QualifiedName

%type <rexp> Name

%type <lmt> Limit

%%

Commands :
  Commands Commandline
  |
  /* empty */
  ;

Commandline :
  LevelCommand T_EOC
  |
  InlineCommand T_EOC
  |
  LimitCommand T_EOC
  |
  ExclInclCommand Parameter T_EOC
    {
      if ( noError )
      {
        addCommand ($1.num, $2.file, $2.class, $2.func);
      }
    }
  |
  T_EOC
    {}
  |
  error T_EOC
    {
      fprintf (stderr, "%s[%2d]: invalid commandline [near %s]\n",
               comfile, lineno, errstring);
      noError = FALSE;
      yyerrok;
    }
  ;

ExclInclCommand :
  T_EXCLUDE
    {
      $$.string = $1.string;
      $$.num    = COM_EXCLUDE;
    }
  |
  T_INCLUDE
    {
      $$.string = $1.string;
      $$.num    = COM_INCLUDE;
    }
  |
  Identifier
    {
      $$.string = $1.string;
      $$.num    = COM_NULL;

      fprintf (stderr, "%s[%2d]: invalid command `%s'\n",
               comfile, lineno+1, $1.string);
      noError = FALSE;
      yyerrok;
    }
  ;

Parameter :
  T_FILE Name
    {
      $$.file  = $2;
      $$.class = RNULL;
      $$.func  = RNULL;
    }
  |
  T_CLASS Name
    {
      $$.file  = RNULL;
      $$.class = $2;
      $$.func  = RNULL;
    }
  |
  T_FUNCTION Name
    {
      $$.file  = RNULL;
      $$.class = RNULL;
      $$.func  = $2;
    }
  |
  QualifiedName
    {
      $$ = $1;
    }
  ;

Name :
  Identifier
    {
      $$ = xalloc (1, REXP);
      $$->regexpr = (regexp *) 0;
      $$->regstr  = $1.string;
    }
  |
  RegExpression
    {
      $$ = xalloc (1, REXP);
      $$->regexpr = regcomp ($1.string);
      $$->regstr  = $1.string;
    }
  ;

QualifiedName :
  Name T_COLON Name T_DBL_COLON Name
    {
      $$.file  = $1;
      $$.class = $3;
      $$.func  = $5;
    }
  |
  Name T_DBL_COLON Name
    {
      $$.file  = RNULL;
      $$.class = $1;
      $$.func  = $3;
    }
  |
  Name T_COLON Name T_DBL_COLON
    {
      $$.file  = $1;
      $$.class = $3;
      $$.func  = RNULL;
    }
  |
  Name T_COLON Name
    {
      $$.file  = $1;
      $$.class = RNULL;
      $$.func  = $3;
    }
  |
  Name T_COLON
    {
      $$.file  = $1;
      $$.class = RNULL;
      $$.func  = RNULL;
    }
  |
  Name T_DBL_COLON
    {
      $$.file  = RNULL;
      $$.class = $1;
      $$.func  = RNULL;
    }
  |
  Name
    {
      $$.file  = RNULL;
      $$.class = RNULL;
      $$.func  = $1;
    }
  ;

LimitCommand :
  T_COUNT OptOnly Limit
    {
      count_limit.stmts = $3.stmts;
      count_limit.child = $3.child;
      count_limit.op    = $3.op;
    }
  |
  T_EXCLUDE Limit
    {
      excl_limit.stmts = $2.stmts;
      excl_limit.child = $2.child;
      excl_limit.op    = $2.op;
    }
  ;

Limit :
  T_SMALLER T_THAN Number T_STATEMENTS LogOp Number T_CHILDREN 
    {
      $$.stmts = $3.num;
      $$.child = $6.num;
      $$.op    = $5.num;
    }
  |
  T_SMALLER T_THAN Number T_CHILDREN LogOp Number T_STATEMENTS 
    {
      $$.stmts = $6.num;
      $$.child = $3.num;
      $$.op    = $5.num;
    }
  ;

OptOnly :
  T_ONLY
    {}
  |
  /* empty */
    {}
  ;

LogOp :
  T_AND
    {
      $$.num = OP_AND;
    }
  |
  T_OR
    {
      $$.num = OP_OR;
    }
  ;

InlineCommand :
  ExclInclCommand T_INLINE OptFunction
    {
      ignoreInlines = ( $1.num == COM_EXCLUDE );
    }
  ;

OptFunction :
  T_FUNCTION
    {}
  |
  /* empty */
    {}
  ;

LevelCommand :
  T_INCLUDE T_UP T_TO T_LEVEL Number
    {
      level_limit = $5.num;
    }
  ;

%%

int yyerror (char *msg)
{
  if ( strcmp (msg, "syntax error") != 0 )
  {
    fprintf (stderr, "%s[%2d]: yacc %s\n", comfile, lineno+1, msg);
    exit (1);
  }
  errstring = yylval.term.string;
}

void regerror (char *msg)
{
  fprintf (stderr, "%s[%2d]: regular expression error: %s\n",
           comfile, lineno+1, msg);
  noError = FALSE;
}

# if defined(sequent) || defined(butterfly)
char *bsearch (key, base, nel, size, compar)
char *key;
char *base;
int nel;
int size;
int (*compar) ();
{
  int i;

  for (i=0; i<nel; i++)
  {
    if ( (* compar) (key, base) == 0 ) return (base);
    base += size;
  }
  return (NULL);
}
# else
extern char *bsearch();
# endif


# define MAXSYMBOL 20

static struct symbol
{
  int   Token;      /* key word token number */
  char *Name;       /* key word              */
}
Symbols[MAXSYMBOL] =
{
  T_AND,        "and",
  T_CHILDREN,   "child",
  T_CHILDREN,   "children",
  T_CLASS,      "class",
  T_COUNT,      "count",
  T_EXCLUDE,    "exclude",
  T_FILE,       "file",
  T_FUNCTION,   "function",
  T_FUNCTION,   "functions",
  T_INCLUDE,    "include",
  T_INLINE,     "inline",
  T_LEVEL,      "level",
  T_ONLY,       "only",
  T_OR,         "or",
  T_SMALLER,    "smaller",
  T_STATEMENTS, "statement",
  T_STATEMENTS, "statements",
  T_THAN,       "than",
  T_TO,         "to",
  T_UP,         "up"
};

static int symbolCompare (struct symbol *s1, struct symbol *s2)
{
  return ( strcmp (s1->Name, s2->Name) );
}

static char *strSave (char *string)
{
  char *res;

  res = xalloc (strlen(string)+1, char);
  strcpy (res, string);
  return (res);
}

static int isSymbol (char *string, char **name)
{
  struct symbol *ptr;    /* -- pointer to element found --------------------- */
  struct symbol  symb;   /* -- element to search ---------------------------- */

  symb.Name = string;
  /* -- binary search in key word table (Symbols) --------------------------- */
  ptr = (struct symbol *) bsearch ((char *) (&symb),
                                   (char *) Symbols,
                                   MAXSYMBOL,
                                   sizeof(struct symbol),
                                   symbolCompare
                                  );
  if (ptr == NULL)
  {
    /* -- not found ------------------------------------ */
    *name = strSave (string);
    return (Identifier);
  }
  else
  {
    /* -- found: return key word token number ---------- */
    *name = ptr->Name;
    return (ptr->Token);
  }
}
static int ch;                      /* scanner character buffer      */
static int getnext = TRUE;          /* scanner lookahead indicator   */
static char  inbuf[256];            /* scanner input buffer          */

# define S 3
# define D 2
# define X 1
# define n 0

static char validch[] = {           /* valid identifier characters   */
/*      !  "  #  $  %  &  '  (  )  *  +  ,  -  .  /   */
        X, X, n, X, X, X, X, X, X, X, X, X, X, X, X,
/*   0  1  2  3  4  5  6  7  8  9  :  ;  <  =  >  ?   */
     D, D, D, D, D, D, D, D, D, D, n, X, n, X, n, X,
/*   @  A  B  C  D  E  F  G  H  I  J  K  L  M  N  O   */
     X, S, S, S, S, S, S, S, S, S, S, S, S, S, S, S,
/*   P  Q  R  S  T  U  V  W  X  Y  Z  [  \  ]  ^  _   */
     S, S, S, S, S, S, S, S, S, S, S, X, X, X, X, X,
/*   `  a  b  c  d  e  f  g  h  i  j  k  l  m  n  o   */
     X, S, S, S, S, S, S, S, S, S, S, S, S, S, S, S,
/*   p  q  r  s  t  u  v  w  x  y  z  {  |  }  ~      */
     S, S, S, S, S, S, S, S, S, S, S, X, X, X, X
};

# define ISVALID(ch,op,val) ((ch > ' ') && (ch <= '~') && (validch[ch - '!'] op val))

# undef X
# undef n

int yylex ()
{
  int i, n;

  do
  {
    if ( getnext ) ch = getc (infile);

    if ( ISVALID(ch,==,S) )
    {
      i = 0;
      do
      {
        inbuf[i++] = ch;
        ch = getc (infile);
      }
      while ( ISVALID(ch,>,0) );
      inbuf[i++] = '\0';
      getnext = FALSE;
      return ( yylval.term.num = isSymbol (inbuf, &yylval.term.string) );
    }
    else if ( ISVALID(ch,==,D) )
    {
      i = n = 0;
      do
      {
        inbuf[i++] = ch;
        n = n*10 + ch - '0';
        ch = getc (infile);
      }
      while ( ISVALID(ch,==,D) );
      inbuf[i++] = '\0';
      getnext = FALSE;
      yylval.term.num = n;
      yylval.term.string = strSave (inbuf);
      return ( Number );
    }
    else if (ch == '<')
    {
      getnext = TRUE;
      i = 0;
      while ( (ch = getc(infile)) != EOF )
      {
        if ( ch == '>' ) break;
        if ( ch == '\n' )
        {
          fprintf (stderr, "%s[%2d]: unterminated regular expression\n",
                   comfile, lineno+1);
          noError = FALSE;
          getnext = FALSE;
          break;
        }
        inbuf[i++] = ch;
      }
      inbuf[i++] = '\0';
      yylval.term.string = strSave (inbuf);
      return ( yylval.term.num = RegExpression );
    }
    else if (ch == ':')
    {
      if ( (ch = getc(infile)) != ':' )
      {
        getnext = FALSE;
        yylval.term.string = ":";
        return ( yylval.term.num = T_COLON );
      }
      else
      {
        getnext = TRUE;
        yylval.term.string = "::";
        return ( yylval.term.num = T_DBL_COLON );
      }
    }
    else
    {
      getnext = TRUE;
      switch ( ch )
      {
        case EOF : return ( 0 );
        case '#' : while ( getc(infile) != '\n' );   /* comment */
        case '\n': lineno++;
                   return ( T_EOC );
        default  : /* skip */
                   break;
      }
    }
  }
  while ( ch != EOF );

  return (0);
}

int readComFile (char *filename)
{
  comfile = filename;
  if ( (infile = fopen (comfile, "r")) == NULL )
  {
    perror (comfile);
    exit (1);
  }
  yyparse ();
  fclose (infile);
  return (noError);
}
