; Collections.
; A collection is an abstract data type with the following operations:
;
;  (make-collection typename type tag dim/dist align scope coll-no)
;                               => collection
;  (collection? obj)            => boolean
;  (find-collection-by-tag tag) => collection | #f
;  (find-collection-by-number number) => collection | #f
;  (all-collections)                  => ( collection ... )
;
;  (element-type coll)       => <type>
;  (element-size coll)       => <integer>
;  (collection-ptr coll cid) => <integer>
;  (dimension-list coll)     => ( <integer> ... )
;  (distribution coll)       => ( <dist-keyword> ... )
;  (alignment coll)          => <alignment-expression>
;
;
; Mainly for human consumption:
;  (typename coll)           => <string>, name of collection type
;  (tag coll)                => <string>, name of collection variable
;  (scope coll)              => <string>, name of procedure where declared

(require "utils.sch")

(define collection:type 1)
(define collection:ptr 2)
(define collection:dims 3)
(define collection:tag 4)
(define collection:dist 5)
(define collection:align 6)
(define collection:typename 7)
(define collection:scope 8)
(define collection:number 9)
(define collection:count 10)

(define make-collection #f)
(define destroy-collection #f)
(define find-collection-by-tag #f)
(define find-collection-by-number #f)
(define collection? #f)
(define all-collections #f)

(let ()

  (define collection-mutex (make-mutex))
  (define collection-list '())
  (define collection-tag (list 'collection))

  (set! make-collection 
	(lambda (typename type tag dim/dist align scope coll-no)
	  (call-with-mutex collection-mutex
	    (lambda ()
	      (let ((c (vector collection-tag
			       type
			       (make-extensible-vector 0 #f)
			       (map car dim/dist)
			       tag
			       (map cadr dim/dist)
			       align
			       typename
			       scope
			       coll-no
			       0)))
		(set! collection-list (cons c collection-list))
		c)))))

  (set! destroy-collection
	(lambda (coll)
	  (call-with-mutex collection-mutex
	    (lambda ()
	      (set! collection-list (remove-first! coll collection-list))))))

  (set! find-collection-by-tag 
	(lambda (the-tag)
	  (call-with-mutex collection-mutex
	    (lambda ()
	      (select-first (lambda (x) (equal? (tag x) the-tag))
			    collection-list)))))

  (set! find-collection-by-number
	(lambda (the-num)
	  (call-with-mutex collection-mutex
	    (lambda ()
	      (select-first (lambda (x) (equal? (collection-number x) the-num))
			    collection-list)))))

  (set! collection? 
	(lambda (obj) 
	  (and (vector? obj)
	       (> (vector-length obj) 0)
	       (eq? (vector-ref obj 0) collection-tag))))

  (set! all-collections
	(lambda ()
	  (call-with-mutex collection-mutex
	    (lambda ()
	      (list-copy collection-list)))))

  #t)


(define (element-type collection)
  (vector-ref collection collection:type))

(define (element-size collection)
  (struct-size (element-type collection)))

(define (collection-ptr collection context)
  (evector-ref (vector-ref collection collection:ptr) context))

(define (collection-ptr! collection context ptr)
  (evector-set! (vector-ref collection collection:ptr) context ptr))

(define (dimension-list collection)
  (vector-ref collection collection:dims))

(define (distribution collection)
  (vector-ref collection collection:dist))

(define (tag collection)
  (vector-ref collection collection:tag))

(define (alignment collection)
  (vector-ref collection collection:align))

(define (typename collection)
  (vector-ref collection collection:typename))

(define (scope collection)
  (vector-ref collection collection:scope))

(define (collection-number collection)
  (vector-ref collection collection:number))

(define (collection-numcontexts-incr! collection)
  (vector-set! collection collection:count
	       (+ (vector-ref collection collection:count) 1))
  (vector-ref collection collection:count))

(define (collection-numcontexts-decr! collection)
  (vector-set! collection collection:count
	       (- (vector-ref collection collection:count) 1))
  (vector-ref collection collection:count))

(provide "collection.sch")

; eof
