; Some code which grubs around the pC++ runtime.
;
; Lars Thomas Hansen [lth@cs.uoregon.edu]
; March 13, 1996.
;
; PRIMARY INTERFACE
;   (user-event-6xx cid tid data)  => unspecified
;   (thread-has-stopped cid tid) => unspecified
;   (thread-is-continuing cid tid) => unspecified
;   (inspect-collection-element collection idx ...)    => representation | #f
;   
; SECONDARY INTERFACE
;   (get-global-pointer collection idx ...)   => global pointer | #f
;   (get-element cid tid gptr len)  => bytevector
;   (get-element gptr len) => bytevector | #f
;
; DESCRIPTION
;   #f is returned when data cannot be obtained because no thread is 
;   available to service the request locally (we want to avoid causing
;   RPCs in the run-time).
;
;   There are some hooks:
;   - user-event-6xx should be called whenever a user event in the range 
;     600-699 is received.  These are magic events used by the debugger.
;   - thread-has-stopped should be called whenever a thread stops for any
;     undetermined amount of time.
;   - thread-is-continuing should be called whenever a thread is about
;     to continue; call it before you call client.continue.
;
;   A collection is an abstract data type; see collection.sch.
;
; TYPE DESCRIPTOR LAYOUT
;
; Primitive types (int, etc)
;   <type> --> (<primitive-type-name>)
;   <primitive-type-name> --> int | uint | long | ulong | short | ushort
;                           | char | uchar | schar
;
; Pointer types:
;   <type> --> (pointer <base-type>)
;   <base-type> --> <type>
;
; Structured types:
;   <type> -->  (struct <tag> <size> ( <field> ... ) )
;            |  (union <tag> <size> ( <field> ... ) )
;            |  (class <tag> <size> ( <field> ... ) )
;   <field> --> ( <name> <type> <offset> )
;   <name> --> <string>
;   <tag> --> <string>
;   <offset> --> <integer>
;
; Array types:
;   <type> --> (array <length> <base-type>)
;   <length> --> <integer>
;
; Enum types: 
;   <type> --> (enum <tag> <name> ...)
;
; Notes
; - All sizes are in bytes, as for sizeof.
; - Anonymous structures have tags which are empty strings.
; - The sizes of primitive and pointer types are known and constant
;   (not strictly ANSI, true for all platforms we care about).
; - In a union, all the fields have the same offset.
;
; BUGS
; - There are race conditions between code in this module and code which
;   indiscriminately continues threads.  Unfortunately, this code needs
;   to obtain a thread ID for its use, and it needs to know what's running
;   and what's not.  Furthermore, it depends on the thread _not_ being
;   continued between the time it obtains the thread id and the time it
;   uses it.  There's really not way to ascertain this short of synchronizing.
;   The implementation of thread-is-continuing can force such a 
;   synchronization, but it would be better to codify the practice and
;   make it a standard thing.
;
; - There is also a race condition in that when a stopped thread is selected
;   to obtain collection info, that context may have deleted its part of
;   the collection, whereas others have not yet done this, regardless of the
;   use of barriers within the RTS.  Our selection criteria should therefore
;   include the liveness of the collection in that context.

(require "utils.sch")
(require "synch.sch")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;

; Sizes of primitive types and poitners.
;
; These values depend on the target platform!!
; The current values are for pC++ running on the SGI PowerChallenge,
; compiled in 64-bit mode by the SGI C++ compiler 'CC'.
;
; There needs to be a way for the debugger to obtain these from the program.

(define enum-size 4)
(define int-size 4)
(define long-size 8)
(define short-size 2)
(define ptr-size 8)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;

; Application should call user-event-6xx when a user event in range 600-699
; is received.

; Problem: the client.read call cannot be done in the event handler, so
; we will have to spawn a thread here.  At the same time, we can't trust
; that the caller of user-event-6xx will not do something which races
; with that thread!  So, the user must pass a continuation which we
; will call with no arguments when we're done.  The user's event handler
; _must_ return to the event loop without doing anything else.  This is a
; real hassle.
;
; Format of event datum: (code context localptr length)

(define (user-event-6xx cid tid data k)
  (case (car data)
    ((600) (snz-get-context! cid (caddr data)) (k))
    ((601) (snz-get-local-addr! cid (caddr data)) (k))
    ((602) 
     (let ((ptr (caddr data))
	   (len (cadddr data)))
       (spawn
	(lambda ()
	  (let ((datum (read (open-input-string 
			      (decode-data 
			       (client.read cid tid (list cid ptr) len))))))
	    (snz-create-collection cid tid datum)
	    (k))))))
    ((603)
     (let ((ptr (caddr data))
	   (len (cadddr data)))
       (spawn
	(lambda ()
	  (let ((datum (read (open-input-string
			      (decode-data
			       (client.read cid tid (list cid ptr) len))))))
	    (snz-delete-collection cid tid datum)
	    (k))))))
    (else ???)))


; Given a collection and some indices, return a Scheme representation 
; of the named element, if possible, or #f if the element cannot be 
; obtained because no threads on the owning context are stopped.

(define (inspect-collection-element coll . rest)
  (let ((gptr (apply get-global-pointer coll rest)))
    (if (not gptr)
	#f
	(let ((tid (get-stopped-thread (car gptr))))
	  (if (not tid)
	      #f
	      (decode-type-and-data (element-type coll)
				    (get-element (car gptr)
						 tid
						 gptr
						 (element-size coll))))))))

; A general procedure which inspects all elements, varying 
; rightmost subscript fastest.

(define (inspect-collection coll proc)

  (define (inspect dims idx)
    (if (null? dims)
	(proc (apply inspect-collection-element coll idx) idx)
	(do ((i 0 (+ i 1)))
	    ((= i (car dims)) #t)
	  (inspect (cdr dims) (append idx (list i))))))

  (inspect (dimension-list coll) '()))


; Given a collection and some indices, return a global pointer to the element.
; The global pointer is a list: context and local address, both unsigned
; integers.  If the pointer cannot be obtained because the owning context
; is not available to be queried, return #f.
;
; The code is contorted because we do not want to incur inter-context
; communication during any of this (in fear of screwing up replay).  In
; a more practical system, we need to realize that there are multiple 
; communication channels in a system and that replay should not control
; all of them; in particular, there will be a channel for non-side-effecting
; system information flow and it needs to be unmonitored.

(define (get-global-pointer coll . rest)

  (define (mkargs cid)
    (cons (list sneezy.argtype:ulong (collection-ptr coll cid))
	  (map (lambda (idx)
		 (list sneezy.argtype:int idx))
	       rest)))

  (let ((cidtid (get-stopped-thread)))
    (if (not cidtid)
	#f
	(let* ((res (client.call-user-function (car cidtid)
					       (cdr cidtid)
					       (snz-get-context (car cidtid))
					       (mkargs (car cidtid))))
	       (ctx (decode-type-and-data '(int)
					  (decode-data (car res))))
	       (t   (get-stopped-thread ctx)))
	  (if (not t)
	      #f
	      (let* ((res (client.call-user-function ctx
						     t
						     (snz-get-local-addr ctx)
						     (mkargs ctx)))
		     (ptr (decode-type-and-data '(ulong)
						(decode-data (car res)))))
		(list ctx ptr)))))))


; Read an element from the global pointer.

(define (get-element . rest)
  (case (length rest)
    ((2) (let ((gptr (car rest))
	       (len  (cadr rest)))
	   (let ((t (get-stopped-thread (car gptr))))
	     (if (not t)
		 #f
		 (decode-data (client.read (car gptr) t gptr len))))))
    ((4) (decode-data (apply client.read rest)))
    (else ???)))


; Return a stopped thread as a pair (cid . tid).  
; If a context ID is given, return only a tid from that context.
; Return #f if there are none.
;
; The debugger must call thread-has-stopped and thread-is-continuing at
; the right places.

(define stopped-threads-mutex (make-mutex))
(define stopped-threads '())

(define (get-stopped-thread . rest)
  (call-with-mutex stopped-threads-mutex
    (lambda ()
      (cond ((null? stopped-threads) #f)
	    ((null? rest) (car stopped-threads))
	    (else (let ((probe (assq (car rest) stopped-threads)))
		    (if probe 
			(cdr probe)
			#f)))))))

(define (thread-has-stopped cid tid)
  (call-with-mutex stopped-threads-mutex
    (lambda ()
      (set! stopped-threads (cons (cons cid tid) stopped-threads)))))

(define (thread-is-continuing cid tid)
  (call-with-mutex stopped-threads-mutex
    (lambda ()
      (set! stopped-threads (remove-first! (cons cid tid) stopped-threads)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;

; The lock is not strictly necessary, as only one context will ever set
; any given slot, and references do not need to be synchronized.  However,
; it disambiguates any partial writes.

(define procptr-mutex (make-mutex))
(define get-context-vector (make-extensible-vector 10 #f))
(define get-local-addr-vector (make-extensible-vector 10 #f))

; Return address of get-context function in context ctx.

(define (snz-get-context ctx)
  (call-with-mutex procptr-mutex
    (lambda ()
      (evector-ref get-context-vector ctx))))

(define (snz-get-context! ctx addr)
  (call-with-mutex procptr-mutex
    (lambda ()
      (evector-set! get-context-vector ctx addr))))

; Return address of get-local-addr function in context ctx.

(define (snz-get-local-addr ctx)
  (call-with-mutex procptr-mutex
    (lambda ()
      (evector-ref get-local-addr-vector ctx))))

(define (snz-get-local-addr! ctx addr)
  (call-with-mutex procptr-mutex
    (lambda ()
      (evector-set! get-local-addr-vector ctx addr))))


(define coll-mutex (make-mutex))

; Data is (typename type tag dim/dist align scope coll-no ptr)

(define (snz-create-collection cid tid data)
  (assert (= tid 0) "snz-create-collection")
  (call-with-mutex coll-mutex
    (lambda ()
      (let* ((coll-data data)
	     (coll      (or (find-collection-by-number (list-ref coll-data 6))
			    (apply make-collection (but-last coll-data)))))
	(assert (collection? coll) "snz-create-collection")
	(collection-ptr! coll cid (list-ref coll-data 7))
	(collection-numcontexts-incr! coll)))))

; Data is (tag coll-no ptr)
;
; This is still wrong: if one context gets to create and delete before
; another gets to create, then we are perhaps doing the wrong thing?  
; The barriers in the implementation should take care of this problem.

(define (snz-delete-collection cid tid data)
  (assert (= tid 0) "snz-delete-collection")
  (call-with-mutex coll-mutex
    (lambda ()
      (let* ((coll-data data)
	     (coll      (find-collection-by-number (list-ref coll-data 1))))
	(assert (collection? coll) "snz-delete-collection")
	(collection-ptr! coll cid #f)
	(if (zero? (collection-numcontexts-decr! coll))
	    (destroy-collection coll))))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;

(define (decode-type-and-data element-type data)
  (let ((scanner (big-endian-element-scanner data)))

    (define get-int
      (case int-size
	((4) (scanner 'word-s))
	((8) (scanner 'dword-s))
	(else ???)))

    (define get-uint
      (case int-size
	((4) (scanner 'word-u))
	((8) (scanner 'dword-u))
	(else ???)))

    (define get-long
      (case long-size
	((4) (scanner 'word-s))
	((8) (scanner 'dword-s))
	(else ???)))

    (define get-ulong
      (case long-size
	((4) (scanner 'word-u))
	((8) (scanner 'dword-u))
	(else ???)))

    (define get-short
      (case short-size
	((2) (scanner 'hword-s))
	((4) (scanner 'word-s))
	((8) (scanner 'dword-s))
	(else ???)))

    (define get-ushort
      (case short-size
	((2) (scanner 'hword-u))
	((4) (scanner 'word-u))
	((8) (scanner 'dword-u))
	(else ???)))

    (define get-ptr
      (case ptr-size
	((4) (scanner 'word-u))
	((8) (scanner 'dword-u))
	(else ???)))

    (define get-enum
      (case enum-size
	((2) (scanner 'hword-u))
	((4) (scanner 'word-u))
	((8) (scanner 'dword-u))
	(else ???)))

    (define get-schar
      (scanner 'byte-s))

    (define get-uchar
      (scanner 'byte-u))

    (define (decoder t i)
      (cond ((null? t) #t)
	    ((symbol? (car t))
	     (case (car t)
	       ((int) (get-int i))
	       ((uint) (get-uint i))
	       ((long) (get-long i))
	       ((ulong) (get-ulong i))
	       ((short) (get-short i))
	       ((ushort) (get-ushort i))
	       ((char schar) (get-schar i))
	       ((uchar) (get-uchar i))
	       ((pointer) (get-ptr i))
	       ((enum) (get-enum i))
	       ((struct class union) (process (struct-fields t)))
	       ((array)
		(let* ((base-type (array-type t))
		       (base-size (type-size base-type))
		       (elements  (array-length t)))
		  (list->vector
		   (let loop ((l '()) (k 0))
		     (if (= i elements)
			 (reverse l)
			 (loop (cons (decoder base-type (+ i (* k base-size)))
				     l)
			       (+ k 1)))))))
	       (else (display "Bogus type: ")
		     (display t)
		     ???)))
	    (else ???)))

    (define (process field-list)
      (map (lambda (fld)
	     (list (car fld) (decoder (cadr fld) (caddr fld))))
	   field-list))

    (decoder element-type 0)))

(define (type-size t)
  (cond ((memq (car t) '(struct union class)) (struct-size base-type))
	((memq (car t) '(int uint)) int-size)
	((memq (car t) '(long ulong)) long-size)
	((memq (car t) '(short ushort)) short-size)
	((memq (car t) '(char uchar schar)) 1)
	((eq? (car t) 'enum) enum-size)
	((eq? (car t) 'pointer) pointer-size)
	((eq? (car t) 'array) (* (cadr t) (type-size (caddr t))))
	(else ???)))

(define struct-tag cadr)
(define struct-size caddr)
(define struct-fields cadddr)

(define array-length cadr)
(define array-type caddr)

; Grub data out from a data buffer.

(define big-endian-element-scanner
  (let ((two^64 (expt 2 64))
	(two^63-1 (- (expt 2 63) 1))
	(two^32 (expt 2 32))
	(two^31-1 (- (expt 2 31) 1))
	(two^16 (expt 2 16))
	(two^15-1 (- (expt 2 15) 1)))

    (lambda (data)

      (define (bytevector-ref bv i)
	(char->integer (string-ref bv i)))

      (define (get-unsigned-dword i)
	(+ (* two^32 (get-unsigned-word i))
	   (get-unsigned-word (+ i 4))))

      (define (get-unsigned-word i)
	(+ (* 256 (+ (* 256 (+ (* 256 (bytevector-ref data i))
			       (bytevector-ref data (+ i 1))))
		     (bytevector-ref data (+ i 2))))
	   (bytevector-ref data (+ i 3))))
      
      (define (get-unsigned-hword i)
	(+ (* 256 (bytevector-ref data i))
	   (bytevector-ref data (+ i 1))))

      (define (get-unsigned-byte i) (bytevector-ref data i))

      (define (fix i get hi+1 limit)
	(let ((n (get i)))
	  (if (> n limit)
	      (- hi+1 n)
	      n)))

      (define (get-signed-dword i) (fix i get-unsigned-word two^64 two^63-1))
      (define (get-signed-word i) (fix i get-unsigned-word two^32 two^31-1))
      (define (get-signed-hword i) (fix i get-unsigned-hword two^16 two^15-1))
      (define (get-signed-byte i) (fix i get-unsigned-byte 256 127))

      (lambda (key)
	(case key
	  ((word-u) get-unsigned-word)
	  ((word-s) get-signed-word)
	  ((dword-u) get-unsigned-dword)
	  ((dword-s) get-signed-dword)
	  ((hword-u) get-unsigned-hword)
	  ((hword-s) get-signed-hword)
	  ((byte-u) get-unsigned-byte)
	  ((byte-s) get-signed-byte)
	  (else ???))))))

(provide "inspect.sch")

; eof
