/* c i d t i d . c
 * $Id: cidtid.c,v 1.1 1996/05/01 20:09:35 lth Exp $
 *
 * Copyright (C) 1995 The University of Oregon.
 *
 *   Author: Lars Thomas Hansen [lth@cs.uoregon.edu]
 *  Created: November 15, 1995
 * Modified:
 *
 * The cidtid table is a table which maps a (CID,TID) pair to 
 * arbitrary data.
 *
 * The implementation is one linked list of tids per context, in a table
 * indexed by context.  This is really not optimal for somethink like
 * the KSR which has one context but a bunch of threads.  Ideally, the
 * lookup strategy should depend on the (dynamic) ratio of cids to tids.
 * A simple but useful strategy is to base the table on whichever there
 * is more of, cids or tids per cid.  We can discover this. It may also be
 * that a hash table working on some function of the cid and tid (say) is
 * perfectly adequate.
 */

#include <stdlib.h>
#include "cidtid.h"

struct bucket {
  int  key2;
  void *val;
  struct bucket *next;
};

struct snz_cidtid_t {
  int maxkey;
  struct bucket **buckets;
};

snz_cidtid_t *snz_alloc_cidtid( struct snz_proginfo_t *pi )
{
  int i;
  snz_cidtid_t *p;

  if ((p = malloc( sizeof( snz_cidtid_t ) )) == 0)
    snz_perror_panic( "malloc" );
  p->maxkey = pi->numcontexts-1;
  p->buckets =
    (struct bucket**)malloc( sizeof( struct bucket * )*pi->numcontexts );
  if (p->buckets == 0) 
    snz_perror_panic( "malloc" );
  for ( i = 0 ; i <= p->maxkey ; i++ )
    p->buckets[i] = 0;
  return p;
}

static struct bucket *find( snz_cidtid_t *map, int cid, int tid )
{
  int key1, key2;
  struct bucket *b;

  /* can be intelligent here */
  key1 = cid;
  key2 = tid;

  if (key1 > map->maxkey || key1 < 0) {
    /* oops */
    return 0;
  }
  b = map->buckets[ key1 ];
  while (b != 0 && b->key2 != key2) 
    b = b->next;
  return b;
}

static void add( snz_cidtid_t *map, int cid, int tid, void *data )
{
  int key1, key2;
  struct bucket *b;

  /* can be intelligent here */
  key1 = cid;
  key2 = tid;

  if (key1 > map->maxkey || key1 < 0) {
    /* oops */
    return;
  }
  if ((b = (struct bucket *)malloc( sizeof( struct bucket ) )) == 0)
    snz_perror_panic( "malloc" );
  b->next = map->buckets[ key1 ];
  map->buckets[ key1 ] = b;
  b->key2 = key2;
  b->val = data;
}

void snz_cidtid_insert( snz_cidtid_t *map, int cid, int tid, void *data )
{
  struct bucket *p;

  p = find( map, cid, tid );
  if (p == 0)
    add( map, cid, tid, data );
  else
    p->val = data;
}

void *snz_cidtid_lookup( snz_cidtid_t *map, int cid, int tid )
{
  struct bucket *p;

  p = find( map, cid, tid );
  if (p == 0) return 0;
  return p->val;
}

/* eof */
