/* c l i e n t - a p i . c
 * $Id: client-api.c,v 1.1 1996/05/01 20:09:37 lth Exp $
 *
 * Copyright (C) 1995 The University of Oregon
 *
 *   Author: Lars Thomas Hansen [lth@cs.uoregon.edu]
 *  Created: October 9, 1995
 * Modified:
 *
 * Sneezy connect code, C API implementation.
 *
 * DOCUMENTATION
 *  See the files "client-api.txt" and "client-api.h".
 */

#define CLIENT_SIDE

#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <string.h>

#include "client-api.h"
#include "socket.h"
#include "cidtid.h"

snz_conn_t *snz_make_connection( void )
{
  snz_conn_t *c;

  if ((c = (snz_conn_t*)malloc( sizeof( snz_conn_t ) )) == 0)
    snz_perror_panic( "snz_connect_socket: malloc" );
  snz_null_hstruct( &(c->handlers) );
  c->tag = SNZ_CONN_INVALID;
  return c;
}

/* Connect by socket. Right now only Lisp syntax is supported, but
 * other syntaxes can be supported by handshaking; see file 'notes'.
 */
void snz_connect_socket( snz_conn_t *c, char *ipaddr, int port )
{
  int fd;

  fd = connect_to_server( ipaddr, port );

  c->tag = SNZ_CONN_SOCK_LISP;
  c->u.socket.addr = strdup( ipaddr );
  c->u.socket.port = port;
  c->u.socket.fp_in = fdopen( fd, "r" );
  c->u.socket.fp_out = fdopen( fd, "w" );

  /* Handshaking code can go here, for now do Lisp only */

  /* Go to specific procedures to setup the connection structure */

  snz_connect_socket_lisp( c );

  c->sysmap = snz_alloc_cidtid( &(c->proginfo) );
}

int snz_close( snz_conn_t *c )
{
  c->close( c );
  c->close = 0;
  c->wait = 0;
  c->install = 0;
  c->uninstall = 0;
  c->cont = 0;
  c->terminate = 0;
  c->read = 0;
  c->write = 0;
  c->call_func = 0;
  c->read_reply = 0;
  free( c );
  return 0;
}

void snz_setup_handlers( snz_conn_t *c, snz_hstruct_t *h )
{
  memcpy( &(c->handlers), h, sizeof( snz_hstruct_t ) );
}

snz_proginfo_t *snz_program_info( snz_conn_t *c )
{
  return &(c->proginfo);
}

int snz_wait( snz_conn_t *c, int poll )
{
  return c->wait( c, poll );
}

void snz_continue( snz_conn_t *c, int cid, int tid )
{
  c->cont( c, cid, tid );
}

void snz_terminate( snz_conn_t *c, int cid, int tid )
{
  c->terminate( c, cid, tid );
}

void snz_install_handler( snz_conn_t *c, int cid, int tid, int etype ) 
{
  c->install( c, cid, tid, etype, etype+EID_OFFSET );
}

void snz_uninstall_handler( snz_conn_t *c, int cid, int tid, int etype )
{
  c->uninstall( c, cid, tid, etype );
}

void snz_read_data( snz_conn_t *c, int cid, int tid, snz_addr_t p, int len,
		    int (*f)( snz_conn_t *c, int cid, int tid, void *data ) )
{
  dispatch_read_request( c, cid, tid, p, len, f );
}

void snz_write_data( snz_conn_t *c, int cid, int tid, snz_addr_t p, int len,
		     void *data )
{
  if (data == 0)
    snz_panic( "snz_write_data: null data." );
  c->write( c, cid, tid, p, len, data );
}

void snz_call_func( snz_conn_t *c, int cid, int tid, snz_addr_t addr, 
		    snz_param_t *p,
		    int (*f)( snz_conn_t *c, int cid, int tid, void *data ) )
{
#if 0
  c->call_func( c, cid, tid, addr, p, f );
#endif
}


/**************************************************************************
 * User function call parameter setup.
 *
 * Currently, we deal only with integers, longs, strings, and raw pointers
 * obtained from the program itself.
 *
 * FIXME: There needs to be ways of passing other objects.
 */

snz_param_t *snz_alloc_param( void )
{ 
  snz_param_t *p;

  p = (snz_param_t*)malloc( sizeof( snz_param_t ) );
  if (p == 0) return 0;
  p->params = 0;
  return p;
}

void snz_clear_param( snz_param_t *p ) 
{ 
  int i;

  /* traverse params, free all string data. */
  for ( i = 0 ; i < p->params ; i++ )
    if (p->param[i].tag == SNZ_PARAM_STRING)
      if (p->param[i].u.s.copied) snz_str_free( p->param[i].u.s.data );
  /* no more params */
  p->params = 0;
}

void snz_free_param( snz_param_t *p ) 
{
  snz_clear_param( p );
  free( p );
}

void snz_param_int( snz_param_t *p, snz_int_t i )
{
  if (p->params == SNZ_MAX_PARAMS)
    snz_panic( "Too many args to program procedure call!\n" );
  p->param[p->params].tag = SNZ_PARAM_INT;
  p->param[p->params].u.i = i;
  p->params++;
}

void snz_param_long( snz_param_t *p, snz_long_t l )
{
  if (p->params == SNZ_MAX_PARAMS)
    snz_panic( "Too many args to program procedure call!\n" );
  p->param[p->params].tag = SNZ_PARAM_LONG;
  p->param[p->params].u.l = l;
  p->params++;
}

void snz_param_string( snz_param_t *p, snz_str_t *s, int must_copy ) 
{ 
  if (p->params == SNZ_MAX_PARAMS)
    snz_panic( "Too many args to program procedure call!\n" );
  p->param[p->params].tag = SNZ_PARAM_STRING;
  p->param[p->params].u.s.data = (must_copy? snzstr_dup( s ) : s);
  p->param[p->params].u.s.copied = 1;
  p->params++;
}

void snz_param_addr( snz_param_t *p, snz_addr_t ptr )
{
  if (p->params == SNZ_MAX_PARAMS)
    snz_panic( "Too many args to program procedure call!\n" );
  p->param[p->params].tag = SNZ_PARAM_ADDR;
  p->param[p->params].u.p = ptr;
  p->params++;
}


/**************************************************************************
 * Clear a handle structure in a portable way. 
 */
void snz_null_hstruct( snz_hstruct_t *h )
{
  h->ready = 0;
  h->err = 0;
  h->term = 0;
  h->bbarr = 0;
  h->ebarr = 0;
  h->bfetch = 0;
  h->efetch = 0;
  h->bput = 0;
  h->eput = 0;
  h->brpc = 0;
  h->erpc = 0;
  h->bsrpc = 0;
  h->esrpc = 0;
  h->bpar = 0;
  h->epar = 0;
  h->eelt = 0;
  h->xelt = 0;
  h->user = 0;
  h->bcc = 0;
  h->ecc = 0;
  h->bdc = 0;
  h->edc = 0;
  h->efun = 0;
  h->xfun = 0;
}

/* eof */
