/* c l i e n t - a p i . h
 * $Id: client-api.h,v 1.1 1996/05/01 20:09:38 lth Exp $
 *
 * Copyright (C) 1995 The University of Oregon
 *
 *   Author: Lars Thomas Hansen [lth@cs.uoregon.edu]
 *  Created: October 9-25, 1995
 * Modified:
 *
 * Sneezy client C API defintion. 
 *
 * NOTES
 *  This file defines a low-level C API for Sneezy. It's slightly more
 *  abstract than the raw Sneezy protocol and API; the advantage of hiding
 *  the lowest level is that it is possible to optimize the protocol for
 *  particular architectures and interconnection types on this level, but
 *  harder on the lowest level.
 *
 *  The parallel architecture may have data not representable on the
 *  client architecture, e.g. 64-bit ints or pointers, Cray floats, 
 *  one's complement integers, 16-bit chars, 9-bit bytes, and so on.
 *  This is dealt with by abstracting the representations of these
 *  types, providing packing, unpacking, operations, and so on. 
 *  The client discovers these dependencies at run time.
 */

#if defined( SERVER_SIDE ) || !defined(CLIENT_SIDE)
#error "client-api.h valid only on the client side."
#endif

#ifndef INCLUDED_CLIENT_API
#define INCLUDED_CLIENT_API

#include <stdio.h>

#ifdef __cplusplus
extern "C" {
#endif

/**************************************************************************
 *
 * TYPES
 */

#include "sneezy.h"

/* Event structure: you may initialize each member to a function pointer
 * or to null. If non-null, events of the given type will be signalled by
 * calling the handler function.
 *
 * The handler function must return 1 if another event is to be handled, 
 * 0 if snz_wait is to return to its caller.
 */
typedef struct snz_hstruct_t {
  int (*ready)( snz_conn_t *c, int cid, int tid );
  /* replies are not handled here, but as part of the command. */
  int (*err)( snz_conn_t *c, int cid, int tid, int code );
  int (*term)( snz_conn_t *c, int cid, int tid );
  int (*bbarr)( snz_conn_t *c, int cid, int tid, int bid );
  int (*ebarr)( snz_conn_t *c, int cid, int tid, int bid );
  int (*bfetch)( snz_conn_t *c, int cid, int tid, snz_gptr_t gs, 
		 snz_gptr_t gd, int len );
  int (*efetch)( snz_conn_t *c, int cid, int tid, snz_gptr_t gs,
		 snz_gptr_t gd, int len );
  int (*bput)( snz_conn_t *c, int cid, int tid, snz_gptr_t gs, 
	       snz_gptr_t gd, int len );
  int (*eput)( snz_conn_t *c, int cid, int tid, snz_gptr_t gs, 
	       snz_gptr_t gd, int len );

  int (*brpc)( snz_conn_t *c, int cid, int tid, snz_gptr_t param, 
	       int paramlen, snz_gptr_t elt, int code );
  int (*erpc)( snz_conn_t *c, int cid, int tid, snz_gptr_t param, 
	       int paramlen, snz_gptr_t elt, int code );
  /* Not currently in use */
  int (*bsrpc)( snz_conn_t *c, int cid, int tid, int src_tid, 
	        snz_gptr_t target, int code );
  /* Not currently in use */
  int (*esrpc)( snz_conn_t *c, int cid, int tid, int src_tid,
	        snz_gptr_t target );
  int (*bpar)( snz_conn_t *c, int cid, int tid );
  int (*epar)( snz_conn_t *c, int cid, int tid );
  int (*eelt)( snz_conn_t *c, int cid, int tid, snz_gptr_t elt );
  int (*xelt)( snz_conn_t *c, int cid, int tid, snz_gptr_t elt );
  int (*user)( snz_conn_t *c, int cid, int tid, int code, snz_gptr_t p, 
	       int len );
  int (*bcc)( snz_conn_t *c, int cid, int tid, int coll_id );
  int (*ecc)( snz_conn_t *c, int cid, int tid, int coll_id );
  int (*bdc)( snz_conn_t *c, int cid, int tid, int coll_id );
  int (*edc)( snz_conn_t *c, int cid, int tid, int coll_id );
  int (*efun)( snz_conn_t *c, int cid, int tid, int fun_id, snz_time_t time );
  int (*xfun)( snz_conn_t *c, int cid, int tid, int fun_id, snz_time_t time );
} snz_hstruct_t;


#include "client-low.h"


/**************************************************************************
 *
 * FUNCTION INTEFACE
 */

/* Create a connection structure, but don't connect yet */
snz_conn_t *snz_make_connection( void );


/* Connect to the agent. Right now, socket-based connections are all
 * that are supported. Returns 0 if some error occured, otherwise a 
 * pointer to a malloc'd structure which.
 */
void snz_connect_socket( snz_conn_t *c, char *ipaddr, int port );


/* Close a connection. Returns 1 on success, 0 on failure. */
int snz_close( snz_conn_t * );


/* Setup handlers for events. Note that this does not install handlers
 * in the agent/program, it only sets up event notification locally.
 */
void snz_setup_handlers( snz_conn_t *c, snz_hstruct_t *h );


/* Get the program information sent in the startup message.
 * See "proginfo.h" for a definition of snz_proginfo_t. 
*/
snz_proginfo_t *snz_program_info( snz_conn_t *c );


/* Waiting for one or more events; if poll==1 return immediately if
 * no event is available, and after the first event is handled regardless
 * of the handler's return code. Returns the number of events processed.
 */
int snz_wait( snz_conn_t *c, int poll );


/* These are the sneezy commands, broken out as functions. The first
 * argument is always a connection. In all cases they send the message
 * and return; if there is a reply, it must be gotten with snz_wait().
 *
 * Note that no event ID needs to be passed for install_handler; it is
 * made for you.
 *
 * Read, getparsectinfo, and getelemtype all have replies. Replies are
 * handled by passing explicit continuation functions to the procedures,
 * which will be called when results become available.  Data buffers
 * passed to the continuation have dynamic extent and may not be retained
 * by the called function; the data must be copied.  Data is always raw 
 * bytes.
 */
void snz_continue( snz_conn_t *c, int cid, int tid );

void snz_terminate( snz_conn_t *c, int cid, int tid );

void snz_install_handler( snz_conn_t *c, int cid, int tid, int etype );

void snz_uninstall_handler( snz_conn_t *c, int cid, int tid, int etype );

void snz_read_data( snz_conn_t *c, int cid, int tid, snz_addr_t p, int len, 
		    int (*f)( snz_conn_t *c, int cid, int tid, void *data ) );

void snz_write_data( snz_conn_t *c, int cid, int tid, snz_addr_t p, int len, 
		     void *data );

void snz_call_func( snz_conn_t *c, int cid, int tid, snz_addr_t addr, 
		    snz_param_t *p,
		    int (*f)( snz_conn_t *c, int cid, int tid, void *data ) );

/* Unimplemented */
/* snz_get_parsection_info( ??? ) */
/* snz_get_element_type (maybe) */
/* void snz_get_collection_info( conn, cid, tid, coll_id, 
                                 void (*f)( conn, cid, tid, dims, dimn[] ) );
 */
/* void snz_get_element_loc( conn, cid, tid, coll_id, dimn[],
                             void (*f)( conn, cid, tid, elt_cid ) );
 */


/* Parameters for snz_call_func(). */
snz_param_t *snz_alloc_param( void );
void snz_clear_param( snz_param_t *p );
void snz_free_param( snz_param_t *p );
void snz_param_int( snz_param_t *p, snz_int_t i );
void snz_param_long( snz_param_t *p, snz_long_t l );
void snz_param_string( snz_param_t *p, snz_str_t *s, int must_copy );
void snz_param_addr( snz_param_t *p, snz_addr_t ptr );


/* Set all fields in a hstruct to null (and no, you can't use bzero()). */
void snz_null_hstruct( snz_hstruct_t * );


#ifdef __cplusplus
}  /* end extern "C" */
#endif

#endif /* ifndef INCLUDED_CLIENT_API */

/* eof */
