/* c l i e n t - s o c k - l i s p . c
 * $Id: client-sock-lisp.c,v 1.1 1996/05/01 20:09:43 lth Exp $
 *
 * Copyright (C) 1995 The University of Oregon
 *
 *   Author: Lars Thomas Hansen [lth@cs.uoregon.edu]
 *  Created: October 25, 1995
 * Modified:
 *
 * Sneezy C API -- socket-specific code, for S-expression syntax.
 *
 * This all assumes single-threaded-ness.  In MT code, the file pointer 
 * and the read-pending queue must be protected by locks, at the very least.
 *
 * To handle replies:
 * We must maintain a queue of outstanding requests for cid/tid pairs,
 * because only one request can actually be outstanding with the agent
 * at any one time.  This is not a limitation in the agent but a 
 * limitation due to read requests not being tagged with a code, hence
 * making it impossible to distinguish replies.
 *
 * FIXME: should print warning when event is received for non-existing handler.
 */

#define CLIENT_SIDE
#define FULL_TYPECHECKS

#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <stropts.h>
#include <poll.h>

#include "client-api.h"
#include "client-reply.h"
#include "sexpr.h"
#include "socket.h"

/* STR is pre-empted by stropts.h on Solaris. */
#define SNZSTR2(x)    #x
#define SNZSTR(x)     SNZSTR2(x)

static void sock_close( snz_conn_t * );
static void sock_install( snz_conn_t *, int, int, int, int );
static void sock_uninstall( snz_conn_t *, int, int, int );
static void sock_continue( snz_conn_t *, int, int );
static void sock_terminate( snz_conn_t *, int, int );
static void sock_read( snz_conn_t *c, int, int, snz_addr_t, int );
static void sock_write( snz_conn_t *, int, int, snz_addr_t, int, void * );
static void sock_call_func( snz_conn_t *, int, int, snz_addr_t, snz_param_t *);
static int sock_wait( snz_conn_t *, int );
static int sock_read_reply( snz_conn_t *, int cid, int tid, void *data,
			    int (*f)( snz_conn_t *, int, int, void * ) );

/* Setup communication for Lisp syntax over socket. */
snz_conn_t *snz_connect_socket_lisp( snz_conn_t *c )
{
  sexpr_mem_t *mem = sexpr_alloc_mem();
  sexpr_obj_t *p, *q;

  c->close = sock_close;
  c->install = sock_install;
  c->uninstall = sock_uninstall;
  c->wait = sock_wait;
  c->cont = sock_continue;
  c->terminate = sock_terminate;
  c->read = sock_read;
  c->write = sock_write;
  c->call_func = sock_call_func;
  c->read_reply = sock_read_reply;

  /* FIXME!! Must initialize these fields properly! 
   * The data must be communicated in the startup message?
   * There's a bootstrap problem in doing so...
   */
  c->sizeof_int = sizeof(int);
  c->sizeof_long = sizeof(long);
  c->sizeof_addr = sizeof(void*);

  /* Wait for program info, Lisp syntax. */
  /* Structure is ( version ( ( cid threads ) ... ) ) */ /* FIXME */

  p = sexpr_parse( c, mem );
  if (!p)
    snz_panic( "snz_connect_socket_lisp: startup info: EOF on channel" );
  c->proginfo.version = val(num(car( p )));
  q = cadr(p);
  c->proginfo.numcontexts = length(q);
  c->proginfo.numthreads = (int*)calloc( c->proginfo.numcontexts, sizeof(int));
  if (c->proginfo.numthreads == 0) {
    snz_perror_panic( "snz_connect_socket_lisp: calloc" );
  }
  while (q != 0) { 
    c->proginfo.numthreads[ val(num(caar(q))) ] = val(num(cadar(q))); 
    q = cdr(q);
  }

  sexpr_release_mem( mem );
  return c;
}


static void sock_close( snz_conn_t *c )
{
  /* FIXME: more here to notify agent? */
  fclose( c->u.socket.fp_in );
  fclose( c->u.socket.fp_out );
}


static void sock_install( snz_conn_t *c, int cid, int tid, int etype, int eid )
{
  fprintf( c->u.socket.fp_out, "(" SNZSTR(SNZC_INSTALL_HANDLER) " %d %d (%d %d))", 
	   cid, tid, etype, eid );
  fflush( c->u.socket.fp_out );
}


static void sock_uninstall( snz_conn_t *c, int cid, int tid, int etype )
{
  fprintf( c->u.socket.fp_out, "(" SNZSTR(SNZC_UNINSTALL_HANDLER) " %d %d %d)",
	  cid, tid, etype );
  fflush( c->u.socket.fp_out );
}


static void sock_continue( snz_conn_t *c, int cid, int tid )
{
  fprintf( c->u.socket.fp_out, "(" SNZSTR(SNZC_CONTINUE) " %d %d)", cid, tid );
  fflush( c->u.socket.fp_out );
}


static void sock_terminate( snz_conn_t *c, int cid, int tid )
{
  fprintf( c->u.socket.fp_out, "(" SNZSTR(SNZC_TERMINATE) " %d %d)", cid, tid );
  fflush( c->u.socket.fp_out );
}


static void sock_read( snz_conn_t *c, int cid, int tid, snz_addr_t p, int len )
{
  char buf[ 20 ]; /*FIXME*/

  snz_addr2str( p, buf );
  fprintf( c->u.socket.fp_out, "(" SNZSTR(SNZC_READ) " %d %d (%d %s %d))",
	   cid, tid, cid, buf, len );
  fflush( c->u.socket.fp_out );
}


static void sock_write( snz_conn_t *c, int cid, int tid, snz_addr_t p,
		        int len, void *data )
{
  char *buf;
  char s[ 20 ]; /*FIXME*/

  buf = sexpr_encode_data( data, len );
  snz_addr2str( p, s );
  fprintf( c->u.socket.fp_out, "(" SNZSTR(SNZC_WRITE) " %d %d (%d %s %d \"%s\"))", 
	   cid, tid, cid, s, len, buf );
  free( buf );
  fflush( c->u.socket.fp_out );
}


/* FIXME */
/* need to be abstracted out */
/* need to setup reply */
static void sock_call_func( snz_conn_t *c, int cid, int tid, snz_addr_t f,
			    snz_param_t *params )
{
#if 0
  int i, j;
  char *s;
  FILE *fp;

  fp = c->u.socket.fp_out;
  fprintf( fp, "(call %d %d %lu %d", cid, tid, (unsigned long)f, params->pc );

  for ( i = 0; i < params->pc ; i++ ) {
    switch (params->p[i].tag) {
    case SNZ_PARAM_INT :
      snz_int2str( params->p[i].u.i, buf )
      fprintf( fp, " (%d %s)", params->p[i], buf );
      break;
    case SNZ_PARAM_LONG :
      snz_long2str( params->p[i].u.l, buf );
      fprintf( fp, " (%d %s)", params->p[i], buf );
      break;
    case SNZ_PARAM_STRING :
      putc( ' ', fp );
      putc( '"', fp );
      s = params->p[i].u.string.data;
      for ( j = 0 ; j < l ; j++ ) {
	if (s[j] == '"' || s[j] == '\\')
	  putc( '\\', fp );
	putc( s[j], fp );
      }
      putc( '"', fp );
      break;
    case SNZ_PARAM_ADDR :
      snz_addr2str( params->p[i].u.p, buf );
      fprintf( fp, " (%d %s)", params->p[i], buf );
      break;
    default :
      snz_panic( "CALL_FUNC: ARGS" );
      break;
    }
  }
  putc( ')', fp );
  fflush( fp );
#else
  snz_panic( __FILE__ " sock_call_func not implemented.");
#endif
}

snz_time_t decode_time( sexpr_obj_t *p )
{
  snz_time_t t;

  t.d = 0;       /* FIXME */
  return t;
}

/* What's in the data depends on the event, so the best we can do is
 * parse it and then make data structures as we call the handler.
 */
static int sock_wait( snz_conn_t *c, int do_poll )
{
  sexpr_mem_t *mem = sexpr_alloc_mem();
  sexpr_obj_t *p, *q;
  snz_time_t t;
  snz_gptr_t gs, gd;
  int count, r, n;

  /* This assumes there is no whitespace between messages! */ /* Can fix */
  if (do_poll && !ready( c->u.socket.fp_in, RDY_ANY )) return 0;

  count = 0;
  do {
    r = 1;
    count++;
    sexpr_clear_mem( mem );
    p = sexpr_parse( c, mem );
    if (!p)
      snz_panic( "sock_wait: EOF on channel." );

    /* Remember that in this implementation of the client API there is
     * a trivial mapping between etype and eid, so the dispatch is
     * very simple.
     */
    n = val(num(car(p)));
    if (n < EID_OFFSET) {
      switch (n) {
      case SNZEID_READY :
	if (!c->handlers.ready) snz_panic( "No READY handler." );
	r = c->handlers.ready( c, val(num(cadr(p))), val(num(caddr(p))) );
	break;
      case SNZEID_REPLY :  /* (eid cid tid data) */
	r = snz_reply( c, val(num(cadr(p))), val(num(caddr(p))), cadddr(p) );
	snz_message( "reply handler returned %d", r );
	break;
      default :
	snz_panic( "client-sock-lisp: Unknown: %d\n", n );
	break;
      }
    }
    else {
      switch (n-EID_OFFSET) {
      case SNZE_ERROR :		/* (eid cid tid code) */
	if (!c->handlers.err) break;
	r = c->handlers.err( c, val(num(cadr(p))), 
			    val(num(caddr(p))), val(num(cadddr(p))) );
	break;
      case SNZE_TERMINATE :	/* (eid cid tid) */
	if (!c->handlers.term) break;
	r = c->handlers.term( c, val(num(cadr(p))), val(num(caddr(p))) );
	break;
      case SNZE_BEGIN_BARRIER : /* (eid cid tid bid) */
	if (!c->handlers.bbarr) break;
	r = c->handlers.bbarr( c, val(num(cadr(p))), val(num(caddr(p))), 
			      val(num(cadddr(p))) );
	break;
      case SNZE_END_BARRIER :	/* (eid cid tid bid) */
	if (!c->handlers.ebarr) break;
	r = c->handlers.ebarr( c, val(num(cadr(p))), val(num(caddr(p))), 
			      val(num(cadddr(p))) );
	break;
      case SNZE_BEGIN_FETCH :	/* (eid cid tid (s.ctx s.p d.ctx d.p len)) */
	if (!c->handlers.bfetch) break;
	q = cadddr( p );
	gs = snz_mkgptr( val(num(car( q ))), addr(c,num(cadr( q ))) );
	gd = snz_mkgptr( val(num(caddr( q ))), addr(c,num(cadddr( q ))) );
	r = c->handlers.bfetch( c, val(num(cadr(p))), val(num(caddr(p))),
			       gs, gd, val(num(caddddr(q))) );
	break;
      case SNZE_END_FETCH :	/* (eid cid tid (s.ctx s.p d.ctx d.p len)) */
	if (!c->handlers.efetch) break;
	q = cadddr( p );
	gs = snz_mkgptr( val(num(car( q ))), addr(c,num(cadr( q ))) );
	gd = snz_mkgptr( val(num(caddr( q ))), addr(c,num(cadddr( q ))) );
	r = c->handlers.efetch( c, val(num(cadr(p))), val(num(caddr(p))),
			       gs, gd, val(num(caddddr(q))) );
	break;
      case SNZE_BEGIN_PUT :	/* (eid cid tid (s.ctx s.p d.ctx d.p len)) */
	if (!c->handlers.bput) break;
	q = cadddr( p );
	gs = snz_mkgptr( val(num(car( q ))), addr(c,num(cadr( q ))) );
	gd = snz_mkgptr( val(num(caddr( q ))), addr(c,num(cadddr( q ))) );
	r = c->handlers.bput( c, val(num(cadr(p))), val(num(caddr(p))),
			     gs, gd, val(num(caddddr(q))) );
	break;
      case SNZE_END_PUT :	/* (eid cid tid (s.ctx s.p d.ctx d.p len)) */
	if (!c->handlers.eput) break;
	q = cadddr( p );
	gs = snz_mkgptr( val(num(car( q ))), addr(c,num(cadr( q ))) );
	gd = snz_mkgptr( val(num(caddr( q ))), addr(c,num(cadddr( q ))) );
	r = c->handlers.eput( c, val(num(cadr(p))), val(num(caddr(p))),
			     gs, gd, val(num(caddddr(q))) );
	break;
      case SNZE_BEGIN_RPC : /* (eid cid tid (p.ctx p.p p.l e.ctx e.p code)) */
	if (!c->handlers.brpc) break;
	{ snz_gptr_t gp, ge;
	  int plen, code;
	  q = cadddr(p);
	  gp = snz_mkgptr( val(num(car(q))), addr(c,num(cadr(q))) );
	  plen = val(num(caddr(q)));
	  q = cdddr(q);
	  ge = snz_mkgptr( val(num(car(q))), addr(c,num(cadr(q))) );
	  code = val(num(caddr(q)));
	  r = c->handlers.brpc( c, val(num(cadr(p))), val(num(caddr(p))),
			       gp, plen, ge, code );
	}
	break;
      case SNZE_END_RPC : /* (eid cid tid (p.ctx p.p p.l e.ctx e.p code)) */
	if (!c->handlers.erpc) break;
	{ snz_gptr_t gp, ge;
	  int plen, code;
	  q = cadddr(p);
	  gp = snz_mkgptr( val(num(car(q))), addr(c,num(cadr(q))) );
	  plen = val(num(caddr(q)));
	  q = cdddr(q);
	  ge = snz_mkgptr( val(num(car(q))), addr(c,num(cadr(q))) );
	  code = val(num(caddr(q)));
	  r = c->handlers.erpc( c, val(num(cadr(p))), val(num(caddr(p))),
			       gp, plen, ge, code );
	}
	break;
      case SNZE_BEGIN_PARALLEL : /* (eid cid tid) */
	if (!c->handlers.bpar) break;
	r = c->handlers.bpar( c, val(num(cadr(p))), val(num(caddr(p))) );
	break;
      case SNZE_END_PARALLEL :	/* (eid cid tid) */
	if (!c->handlers.epar) break;
	r = c->handlers.epar( c, val(num(cadr(p))), val(num(caddr(p))) );
	break;
      case SNZE_ENTER_ELEMENT :	/* (eid cid tid (ctx p)) */
	if (!c->handlers.eelt) break;
	gs = snz_mkgptr( val(num(car(cadddr(p)))), addr(c,num(cadr(cadddr(p)))));
	r = c->handlers.eelt( c, val(num(cadr(p))), val(num(caddr(p))), gs );
	break;
      case SNZE_EXIT_ELEMENT :	/* (eid cid tid (ctx p)) */
	if (!c->handlers.xelt) break;
	gs = snz_mkgptr( val(num(car(cadddr(p)))), addr(c,num(cadr(cadddr(p)))));
	r = c->handlers.xelt( c, val(num(cadr(p))), val(num(caddr(p))), gs );
	break;
      case SNZE_BEGIN_SERVICE_RPC :
	if (!c->handlers.bsrpc) break;
	snz_panic( "client: begin-service-rpc not implemented." );
	break;
      case SNZE_END_SERVICE_RPC :
	if (!c->handlers.esrpc) break;
	snz_panic( "client: end-service-rpc not implemented." );
	break;
      case SNZE_USER_EVENT :	/* (eid cid tid (code ctx p len)) */
	if (!c->handlers.user) break;
	q = cadddr( p );
	gs = snz_mkgptr( val(num(cadr(q))), addr(c,num(caddr(q))) );
	r = c->handlers.user( c, val(num(cadr(p))), val(num(caddr(p))),
			     val(num(car(q))), gs, val(num(cadddr(q))) );
	break;
      case SNZE_BEGIN_CREATE_COLLECTION : /* (eid cid tid coll-id) */
	if (!c->handlers.bcc) break;
	r = c->handlers.bcc( c, val(num(cadr(p))), val(num(caddr(p))),
			    val(num(cadddr(p))) );
	break;
      case SNZE_END_CREATE_COLLECTION :	/* (eid cid tid coll-id) */
	if (!c->handlers.ecc) break;
	r = c->handlers.ecc( c, val(num(cadr(p))), val(num(caddr(p))),
			    val(num(cadddr(p))) );
	break;
      case SNZE_BEGIN_DELETE_COLLECTION : /* (eid cid tid coll-id) */
	if (!c->handlers.bdc) break;
	r = c->handlers.bdc( c, val(num(cadr(p))), val(num(caddr(p))),
			    val(num(cadddr(p))) );
	break;
      case SNZE_END_DELETE_COLLECTION :	/* (eid cid tid coll-id) */
	if (!c->handlers.edc) break;
	r = c->handlers.edc( c, val(num(cadr(p))), val(num(caddr(p))),
			    val(num(cadddr(p))) );
	break;
      case SNZE_ENTER_FUNCTION : /* (eid cid tid (fid time)) */
	if (!c->handlers.efun) break;
	q = cadddr(p);
	t = decode_time( cadr(q) );
	r = c->handlers.efun( c, val(num(cadr(p))), val(num(caddr(p))), 
			     val(num(car(q))), t );
	break;
      case SNZE_EXIT_FUNCTION :	/* (eid cid tid (fid time)) */
	if (!c->handlers.xfun) break;
	q = cadddr(p);
	t = decode_time( cadr(q) );
	r = c->handlers.xfun( c, val(num(cadr(p))), val(num(caddr(p))),
			     val(num(car(q))), t );
	break;
      default :
	snz_panic( "client-sock-lisp: Unknown: %d\n", n );
	break;
      }
    }
  } while (!do_poll && r);
  sexpr_release_mem( mem );
  return count;
}

/**************************************************************************/
/* Replies */

/* p is a string of encoded data. */
static int
sock_read_reply( snz_conn_t *c, int cid, int tid, void *data,
	         int (*f)( snz_conn_t *, int cid, int tid, void *d ) )
{
  sexpr_obj_t *p = (sexpr_obj_t *)data;
  unsigned len;
  int r;
  struct _sexpr_str s;
  void *buf;

  s = str(p);
  buf = sexpr_decode_data( s.data, &len );
  
  r = f( c, cid, tid, buf );
  free( buf );
  return r;
}

/* eof */
