/* p r o x y - s o c k - l i s p . c
 * $Id: proxy-sock-lisp.c,v 1.1 1996/05/01 20:09:51 lth Exp $
 *
 * Copyright (C) 1995 The University of Oregon.
 *
 *   Author: Lars Thomas Hansen [lth@cs.uoregon.edu]
 *  Created: November 8, 1995
 * Modified:
 *
 * S-expression I/O code for socket proxy.
 */

#define SERVER_SIDE
#define FULL_TYPECHECKS

#include <stdio.h>
#include "sneezy.h"
#include "sexpr.h"
#include "proxy-sock.h"

/* Read an s-expression encoding a command and parse it into the
 * provided 'cmd' buffer.
 *
 * FIXME: the length datum in the write request is redundant.
 */
int _proxy_read_lisp( snz_clnt_t *clnt, snz_cmd_t *cmd )
{
  static sexpr_mem_t *mem = 0;
  int cmdid, i;
  sexpr_obj_t *m, *q;
  FILE *fp;
  snz_arg_t *ap;

  fp = clnt->u.socket.fp_in;
  if (mem == 0)
    mem = sexpr_alloc_mem();
  else
    sexpr_clear_mem( mem );

  m = sexpr_parse( fp, mem );
  if (!m) {
    snz_message( "_proxy_read_lisp: EOF on client channel.");
    return 0;
  }

  cmdid = num(car(m));
  cmd->cmd = cmdid;
  switch (cmdid) {
  case SNZC_CONTINUE :  /* (continue cid tid ()) */
    cmd->cid = num(cadr(m));
    cmd->tid = num(caddr(m));
    break;
  case SNZC_TERMINATE : /* (terminate cid tid ()) */
    cmd->cid = num(cadr(m));
    cmd->tid = num(caddr(m));
    break;
  case SNZC_INSTALL_HANDLER :   /* (install cid tid (etype eid)) */
    cmd->cid = num(cadr(m));
    cmd->tid = num(caddr(m));
    q = cadddr(m);
    cmd->u.install.etype = num(car(q));
    cmd->u.install.eid = num(cadr(q));
    break;
  case SNZC_UNINSTALL_HANDLER : /* (uninstall cid tid etype) */
    cmd->cid = num(cadr(m));
    cmd->tid = num(caddr(m));
    cmd->u.uninstall.etype = num(cadddr(m));
    break;
  case SNZC_READ :      /* (read cid tid (ctx addr len)) */
    snz_message( "Proxy: READ" );
    cmd->cid = num(cadr(m));
    cmd->tid = num(caddr(m));
    q = cadddr(m);
    cmd->u.read.g.context = num(car(q));
    cmd->u.read.g.addr = (snz_addr_t)num(cadr(q));  /*FIXME*/
    cmd->u.read.len = num(caddr(q));
    break;
  case SNZC_WRITE :     /* (write cid tid (ctx addr len data)) */
    cmd->cid = num(cadr(m));
    cmd->tid = num(caddr(m));
    q = cadddr(m);
    snz_message( "Proxy: WRITE %s", str(cadddr(q)).data );
    cmd->u.write.g.context = num(car(q));
    cmd->u.write.g.addr = (snz_addr_t)num(cadr(q));  /*FIXME*/
    cmd->u.write.len = num(caddr(q));
    cmd->u.write.data = sexpr_decode_data( str(cadddr(q)).data, (unsigned*)0 );
    break;
  case SNZC_CALL_USER_FUNCTION : /* (cuf cid tid (fptr argc (tag data) ...)) */
    cmd->cid = num(cadr(m));
    cmd->tid = num(caddr(m));
    q = cadddr(m);
    cmd->u.call.fptr = (void(*)(void))num(car(q));
    cmd->u.call.argc = num(cadr(q));
    /* The argv structure is freed by the comm layer */
    ap = cmd->u.call.argv = 
      (snz_arg_t*)malloc(cmd->u.call.argc*sizeof(snz_arg_t));
    if (cmd->u.call.argc == 0)
      snz_panic( "proxy: CALL_USER_FUNCTION: out of memory." );
    q = cddr(q);
    for ( i = 0 ; i < cmd->u.call.argc ; i++ ) {
      int tag;
      sexpr_obj_t *datum;

      tag = num(caar(q));
      datum = cadar(q);
      q = cdr(q);
      ap[i].tag = tag;
      switch(tag) {
        case SNZAT_INT    : ap[i].u.i = (int)num(datum); break;
	case SNZAT_UINT   : ap[i].u.ui = (unsigned)num(datum); break;
	case SNZAT_LONG   : ap[i].u.l = num(datum); break;
	case SNZAT_ULONG  : ap[i].u.ul = (unsigned long)num(datum); break;
	case SNZAT_SHORT  : ap[i].u.s = (short)num(datum); break;
	case SNZAT_USHORT : ap[i].u.us = (unsigned short)num(datum); break;
	case SNZAT_CHAR   : ap[i].u.c = (char)num(datum); break;
	case SNZAT_UCHAR  : ap[i].u.uc = (unsigned char)num(datum); break;
	case SNZAT_SCHAR  : ap[i].u.sc = (signed char)num(datum); break;
	case SNZAT_CHARPTR: 
	  /* Note that we do _not_ dup the string here, on the assumption that
	   * since we're in the proxy, nothing strange will happen.
	   */
	  ap[i].u.str.cp = str(datum).data; 
	  ap[i].u.str.len = str(datum).length;
	  break;
      }
    }
    break;
#ifdef SUPPORT_CLIENTS_MULTIPLE
  case SNZC_GET_CLIENT_ID :
  case SNZC_GET_PROGRAM_STATE :
  case SNZC_TRANSFER_MASTER :
  case SNZC_SENDMSG :
  case SNZC_AUX_EVENTS :
  case SNZC_SIGNAL_EVENTS :
    break;
#endif
  default :
    snz_panic( "_proxy_read_lisp: unknown code: %d", num(car(m)) );
  }
  return 1;
}

/* Write a portable (textual) representation of the time to the FILE. */
static void fprintf_time( FILE *fp, snz_time_t t )
{
#ifdef TIME_IS_DOUBLE
  /* FIXME: print time as double */
  fprintf( fp, "0" );
#endif
#ifdef TIME_IS_LONG
  /* FIXME: print time as long */
  fprintf( fp, "0" );
#endif
#ifdef TIME_IS_STRUCT
  /* FIXME: print time as struct. */
  fprintf( fp, "0 0" );
#endif
}

/* Write an S-expression representation of the event to the FILE. */
void _proxy_write_lisp( snz_clnt_t *clnt, snz_event_t *e )
{
  int i;
  char *p, p1[SNZ_PTR_CHARS_MAX], p2[SNZ_PTR_CHARS_MAX];
  FILE *fp;

  fp = clnt->u.socket.fp_out;
  switch (e->etype) {
  case _SNZE_STARTUP :
    fprintf( fp, "(%d (", e->u.startup.info->version );
    for ( i = 0 ; i < e->u.startup.info->numcontexts ; i++ )
      fprintf( fp, "(%d %d)", i, e->u.startup.info->numthreads[i] );
    fprintf( fp, "))" );
    break;
  case _SNZE_READY :
    fprintf( fp, "(%d %d %d ())", SNZEID_READY, e->cid, e->tid );
    break;
  case _SNZE_REPLY :
    switch (e->u.reply.tag) {
      case _REPL_READ :
	fprintf( fp, "(%d %d %d ", SNZEID_REPLY, e->cid, e->tid );
	p = sexpr_encode_data( e->u.reply.data, e->u.reply.len );
	fprintf( fp, "\"%s\")", p );
	free( p );
	break;
      case _REPL_CALL :
	fprintf( fp, "(%d %d %d ())", SNZEID_REPLY, e->cid, e->tid );
	break;
      default :
	snz_panic( "_SNZE_REPLY: unknown code: %d", e->u.reply.tag );
	break;
    }
    break;
  case _SNZE_PUSH :
    fprintf( fp, "(%d %d %d ", SNZEID_PUSH, e->cid, e->tid );
    p = sexpr_encode_data( e->u.push.value, e->u.push.len );
    fprintf( fp, "\"%s\")", p );
    free( p );
    break;
  case SNZE_ERROR :
    fprintf( fp, "(%d %d %d %d)", e->eid, e->cid, e->tid, e->u.err.code ); 
    break;
  case SNZE_TERMINATE :
    fprintf( fp, "(%d %d %d ())", e->eid, e->cid, e->tid );
    break;
  case SNZE_BEGIN_BARRIER :
    fprintf( fp, "(%d %d %d %d)", e->eid, e->cid, e->tid, e->u.bbarr.bid );
    break;
  case SNZE_END_BARRIER :
    fprintf( fp, "(%d %d %d %d)", e->eid, e->cid, e->tid, e->u.ebarr.bid );
    break;
  case SNZE_BEGIN_FETCH :
    snz_ptr2str( e->u.bfetch.src.addr, p1 );
    snz_ptr2str( e->u.bfetch.dest.addr, p2 );
    fprintf( fp, "(%d %d %d (%d %s %d %s %d))", e->eid, e->cid, e->tid, 
	     e->u.bfetch.src.context, p1,
	     e->u.bfetch.dest.context, p2,
	     e->u.bfetch.len );
    break;
  case SNZE_END_FETCH :
    snz_ptr2str( e->u.efetch.src.addr, p1 );
    snz_ptr2str( e->u.efetch.dest.addr, p2 );
    fprintf( fp, "(%d %d %d (%d %s %d %s %d))", e->eid, e->cid, e->tid,
	     e->u.efetch.src.context, p1,
	     e->u.efetch.dest.context, p2,
	     e->u.efetch.len );
    break;
  case SNZE_BEGIN_PUT :
    snz_ptr2str( e->u.bput.src.addr, p1 );
    snz_ptr2str( e->u.bput.dest.addr, p2 );
    fprintf( fp, "(%d %d %d (%d %s %d %s %d))", e->eid, e->cid, e->tid,
	     e->u.bput.src.context, p1, 
	     e->u.bput.dest.context, p2,
	     e->u.bput.len );
    break;
  case SNZE_END_PUT :
    snz_ptr2str( e->u.eput.src.addr, p1 );
    snz_ptr2str( e->u.eput.dest.addr, p2 );
    fprintf( fp, "(%d %d %d (%d %s %d %s %d))", e->eid, e->cid, e->tid,
	     e->u.eput.src.context, p1,
	     e->u.eput.dest.context, p2,
	     e->u.bput.len );
    break;
  case SNZE_BEGIN_RPC :
    snz_ptr2str( e->u.brpc.param.addr, p1 );
    snz_ptr2str( e->u.brpc.elt.addr, p2 );
    fprintf( fp, "(%d %d %d (%d %s %d %d %s %d))", e->eid, e->cid, e->tid,
	    e->u.brpc.param.context, p1,
	    e->u.brpc.paramlen,
	    e->u.brpc.elt.context, p2,
	    e->u.brpc.code );
    break;
  case SNZE_END_RPC :
    snz_ptr2str( e->u.erpc.param.addr, p1 );
    snz_ptr2str( e->u.erpc.elt.addr, p2 );
    fprintf( fp, "(%d %d %d (%d %s %d %d %s %d))", e->eid, e->cid, e->tid,
	     e->u.erpc.param.context, p1,
	     e->u.erpc.paramlen,
	     e->u.erpc.elt.context, p2,
	     e->u.erpc.code );
    break;
#if 0
  case SNZE_BEGIN_SERVICE_RPC :
    fprintf( fp, "(%d %d %d (%d %d %d))", e->eid, e->cid, e->tid,
	    e->u.bsrpc.dest.context, e->u.bsrpc.dest.addr, e->u.bsrpc.code );
    break;
  case SNZE_END_SERVICE_RPC :
    fprintf( fp, "(%d %d %d (%d %d %d))", e->eid, e->cid, e->tid,
	    e->u.esrpc.dest.context, e->u.esrpc.dest.addr, e->u.esrpc.code );
    break;
#endif
  case SNZE_BEGIN_PARALLEL :
    fprintf( fp, "(%d %d %d ())", e->eid, e->cid, e->tid );
    break;
  case SNZE_END_PARALLEL :
    fprintf( fp, "(%d %d %d ())", e->eid, e->cid, e->tid );
    break;
  case SNZE_ENTER_ELEMENT :
    snz_ptr2str( e->u.eelt.elt.addr, p1 );
    fprintf( fp, "(%d %d %d (%d %s))", e->eid, e->cid, e->tid,
	     e->u.eelt.elt.context, p1 );
    break;
  case SNZE_EXIT_ELEMENT :
    snz_ptr2str( e->u.xelt.elt.addr, p1 );
    fprintf( fp, "(%d %d %d (%d %s))", e->eid, e->cid, e->tid,
	     e->u.xelt.elt.context, p1 );
    break;
  case SNZE_USER_EVENT :
    snz_ptr2str( e->u.user.p.addr, p1 );
    fprintf( fp, "(%d %d %d (%d %d %s %d))", e->eid, e->cid, e->tid, 
	    e->u.user.code, e->u.user.p.context, p1, e->u.user.len );
    break;
  case SNZE_BEGIN_CREATE_COLLECTION :
    fprintf( fp, "(%d %d %d %d)", e->eid, e->cid, e->tid, e->u.bcc.cid );
    break;
  case SNZE_END_CREATE_COLLECTION :
    fprintf( fp, "(%d %d %d %d)", e->eid, e->cid, e->tid, e->u.ecc.cid );
    break;
  case SNZE_BEGIN_DELETE_COLLECTION :
    fprintf( fp, "(%d %d %d %d)", e->eid, e->cid, e->tid, e->u.bdc.cid );
    break;
  case SNZE_END_DELETE_COLLECTION :
    fprintf( fp, "(%d %d %d %d)", e->eid, e->cid, e->tid, e->u.edc.cid );
    break;
  case SNZE_ENTER_FUNCTION :
    fprintf( fp, "(%d %d %d (%d ", e->eid, e->cid, e->tid,
	     e->u.efun.id );
    fprintf_time( fp, e->u.efun.time );
    fprintf( fp, "))" );
    break;
  case SNZE_EXIT_FUNCTION :
    fprintf( fp, "(%d %d %d (%d ", e->eid, e->cid, e->tid,
	     e->u.xfun.id );
    fprintf_time( fp, e->u.xfun.time );
    fprintf( fp, "))" );
    break;
#ifdef SUPPORT_CLIENTS_MULTIPLE
  case SNZE_SLAVE_READY :
  case SNZE_NEW_CLIENT :
  case SNZE_BECOME_MASTER :
  case SNZE_FOREIGN_EVENT :
  case SNZE_MESSAGE :
    break;
#endif
  default :
    snz_panic( "UNKNOWN ETYPE IN _proxy_write_lisp(): %d", e->etype );
    break;
  }
  fflush( fp );
}

/* eof */
