/* p r o x y - s o c k . c
 * $Id: proxy-sock.c,v 1.1 1996/05/01 20:09:54 lth Exp $
 *
 * Copyright (C) 1995 The University of Oregon.
 *
 *   Author: Lars Thomas Hansen [lth@cs.uoregon.edu], 
 *           based loosely on a design by Darryl Ivan Brown.
 *  Created: November 6, 1995
 * Modified:
 *
 * Sneezy proxy agent for communicating on sockets.
 *
 * Computational model: classic SPMD with separate processes all being
 * forked from a single initially executing program. In this (prototype)
 * implementation, the agent is a standalone process which communicates
 * with the parallel processes on binary sockets and with the clients on
 * some negotiated-protocol sockets.
 *
 * DESCRIPTION
 *   Snz_server_init_agent is called from the main() of the parallel program
 *   to create and initialize agent proxy.  It must be called before main()
 *   forks the worker processes.
 *
 *   Snz_server_init_context is called from the main() of the parallel program
 *   to connect each worker process to the agent proxy.  It must be called
 *   after main() has forked the worker processes.
 *
 * USAGE
 *   Compile with any number of the following options:
 *     SUPPORT_PROTO_LISP          S-expressions (portable)
 *     SUPPORT_PROTO_BIN           Uninterpreted (fastest)
 *     SUPPORT_PROTO_XDR           SunRPC XDR format (portable)
 *     SUPPORT_PROTO_BINEL32       Endian little 32-bit int IEEE FP (fast)
 *     SUPPORT_PROTO_BINBE32       Big endian 32-bit int IEEE FP (fast)
 *     SUPPORT_CLIENTS_MULTIPLE    Multi-client (expensive)
 *   At least one protocol must be selected.
 *
 * NOTE
 *   Only PROTO_LISP is supported right now.
 */

#define SERVER_SIDE

#define SUPPORT_PROTO_LISP
#define SUPPORT_PROTO_BIN

/* Plan 9 got it right. */
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <setjmp.h>
#include <signal.h>
#include <stropts.h>
#include <poll.h>
#include <string.h>          /* Not portable? */
#include <stdlib.h>
#include <errno.h>
#include <wait.h>

#include "sneezy.h"
#include "server-api.h"
#include "proxy-sock.h"
#include "server-low.h"
#include "socket.h"

extern int _proxy_portno;     /* common: the proxy listens on this one. */

static int num_contexts;     /* common: number of contexts. */
static int dead_contexts;    /* common: number of contexts that have died. */
static jmp_buf chldbuf;      /* for whenever all children die */

static void become_proxy( int client_port, int proxy_fd, int multi );
static void register_client( snz_clnt_t *c );
static snz_clnt_t *fd2client( int fd );
static int *connect_with_program( int proxy_fd, snz_proginfo_t * );
static void set_attr( int fd, int attr );
static void handle_sigchld( int );
static void add_fd( int fd );
static void remove_fd( int fd );
static int poll_data( void );
static int fdattr( int fd );
static void release_program( int *program_fds );

typedef struct {
  int  tag;
  char *negcode;
  int  (*read)( snz_clnt_t *, snz_cmd_t * );
  void (*write)( snz_clnt_t *, snz_event_t * );
} snz_protocol_t;

/* Supported protocols are selected by conditional compilation. */
static snz_protocol_t protocols[] = {

#ifdef SUPPORT_PROTO_LISP
  /* Textual s-expressions */
  { PROTO_LISP, "l", _proxy_read_lisp, _proxy_write_lisp },
#endif

#ifdef SUPPORT_PROTO_BIN
  /* Raw binary */
  { PROTO_BIN, "b", _proxy_read_bin, _proxy_write_bin },
#endif

#ifdef SUPPORT_PROTO_XDR
  /* 32-bit XDR */ /* code "x" */
#endif

#ifdef SUPPORT_PROTO_BINEL32
  /* Binary endian little 32 bits */ /* code "2" */
#endif

#ifdef SUPPORT_PROTO_BINBE32
  /* Binary big endian 32 bits */ /* code "3" */
#endif

};
static const int protocol_count = sizeof( protocols ) / sizeof( protocols[0] );

struct cidinfo {
  int cid;
  int nthreads;
};

/* Called from main() before the fork into multiple processes. 
 * If you want to use multiple clients, pass multi as 1. It's optional 
 * because multiple clients are expensive.
 */
void snz_server_init_agent( int agent_port, int multi )
{
  int k;
  int proxy_fd;

  snz_message( "INIT AGENT (%d)", getpid()  );
  num_contexts = _snz_total_nodes();
  if (num_contexts <= 0)
    snz_panic( "Unlikely number of contexts: %d", num_contexts );
  dead_contexts = 0;

  /* Trick: we open a server socket here so that both the parent and the
   * child can know the port number to use without any fancy IPC or
   * user-settable environment variables.
   */
  _proxy_portno = 0;
  open_server_socket( &_proxy_portno, &proxy_fd );

  k = fork();
  if (k == 0) {       /* In the child -- continue as parallel program. */
    close( proxy_fd );
    snz_message( "CHILD RETURNING (%d)", getpid() );
    return;
  }
  else if (k > 0) {   /* In the parent -- become proxy. */
    snz_message( "PARENT GOING ON (%d)", getpid() );
    become_proxy( agent_port, proxy_fd, multi );
    snz_message( "Proxy exiting." );
    exit( 0 );
  }

  snz_perror_panic( "fork" );
}

static void become_proxy( int client_port, int proxy_fd, int multi )
{
  int  (*accept_new_client)(int,snz_proginfo_t*) = 
    _proxy_accept_new_client_single_client;
  int (*route_to_program)(int*,snz_clnt_t*) = 
    _proxy_route_to_program_single_client;
  int (*route_to_clients)(int) =
    _proxy_route_to_clients_single_client;
  int client_fd, *program_fds, i;
  snz_proginfo_t info;
  snz_clnt_t *clnt;
  int client_count = 0;

  snz_message( "BECOME PROXY (%d)", getpid() );
#ifdef SUPPORT_CLIENTS_MULTIPLE
  if (multi) {
    route_to_program = _proxy_route_to_program_multiple_client;
    route_to_clients = _proxy_route_to_clients_multiple_client;
    accept_new_client = _proxy_accept_new_client_multiple_client;
  }
#else
  if (multi)
    snz_panic( "No compiled-in support for multiple clients." );
#endif

  /* POSIX only knows how (un)portable this is. */
  sigset( SIGCHLD, handle_sigchld );

  if (setjmp( chldbuf ) == 1) {
    snz_message( "All children terminated." );
    /* FIXME: should close this, that, and the other. */
    return;
  }

  program_fds = connect_with_program( proxy_fd, &info );
  close( proxy_fd );

  open_server_socket( &client_port, &client_fd );
  set_attr( client_fd, FD_CONNECT );
  add_fd( client_fd );

  /* Wait for client */
  /* Only client_fd is in set right now, so use poll_data(). */
  
  { int n;
    snz_message( "Waiting for primary client." );
    n = poll_data();
    snz_message( "Primary client (%d) is present.", n );
  }

  /* Add program fds to poll set */
  for ( i = 0 ; i < num_contexts ; i++ ) {
    set_attr( program_fds[i], FD_PROGRAM );
    add_fd( program_fds[i] );
  }

  while (1) {
    int rdyfd, fd;

    rdyfd = poll_data();

    /* FIXME: what about clients disconnecting? */

    switch (fdattr( rdyfd )) {
    case FD_CLIENT :  /* Message from client to a program process or to
		       * another client, depending on content of message.
		       */
      clnt = fd2client( rdyfd );
      do {
	if (!route_to_program( program_fds, clnt )) {
	  snz_message( "[client @fd %d has gone missing; terminating.]", 
		       rdyfd );
	  exit(0);
	}
      } while (ready( clnt->u.socket.fp_in, RDY_TRIVIAL ));
      break;
    case FD_PROGRAM : /* Message from a program process to clients */
      if (!route_to_clients( rdyfd )) {
	snz_message( "[context @fd %d has gone missing; removing it.]" );
	remove_fd( rdyfd );
      }
      break;
    case FD_CONNECT : /* New client! */
      fd = accept_new_client( accept_connection( rdyfd ), &info );
      add_fd( fd );
      client_count++;
      if (client_count == 1) release_program( program_fds );
      break;
    default:
      snz_panic( "POLL" );
      break;
    }
  }
}

/* All programs are blocking in read() for a synch. token. */

static void release_program( int *program_fds )
{
  int one = 1, i;

  snz_message( "Synchronizing write." );
  for ( i = 0 ; i < num_contexts ; i++ )
    write( program_fds[i], &one, sizeof(one) );
}

static void handle_sigchld( int dummy )
{
  dead_contexts++;
  wait((int*)0);
  if (dead_contexts == num_contexts)
    longjmp( chldbuf, 1 );
}

/**************************************************************************
 *
 * Maintain set of file descriptors for polling, and do the polling.
 * If necessary, these routines can be replaced with some which centralize
 * the management of the poll(), like the Tk poll routines.
 *
 * This should be the only place in the proxy where an indefinitely-blocking
 * read system call should happen.
 */

static int nfds = 0, maxfds = 0;
static struct pollfd *fds = 0;

static int poll_data( void )
{
  int r, i;

  /* poll() exists on Solaris, SunOS, HP-UX, and IRIX, at least. */
  /* One can also use select(), but poll() is nicer. */

  do {
    r = poll( fds, nfds, -1 );
  } while (r == -1 && errno == EINTR);

  if (r <= 0) 
    snz_perror_panic( "poll" );
  for ( i = 0 ; i < nfds ; i++ ) {
    if (fds[i].revents != 0) {
      fds[i].revents = 0;
      return fds[i].fd;
    }
  }
  snz_panic( "poll: impossible." );
  return -1;
}

static void add_fd( int fd )
{
  if (nfds == maxfds) {
    maxfds += 10;
    fds = (struct pollfd*)realloc( fds, sizeof( struct pollfd )*maxfds );
    if (fds == 0)
      snz_perror_panic( "realloc" );
  }
  fds[nfds].fd = fd;
  fds[nfds].events = POLLIN|POLLPRI|POLLRDNORM|POLLRDBAND;
  fds[nfds].revents = 0;
  nfds++;
}

static void remove_fd( int fd )
{
  int i;

  for ( i = 0 ; i < nfds && fds[i].fd != fd ; i++ )
    ;
  if (i == nfds)
    snz_panic( "remove_fd" );
  for ( i++ ; i < nfds ; i++ )
    fds[i-1] = fds[i];
}

/*************************************************************************
 *
 * Maintain a table of connections and their attributes, indexed by
 * descriptor.
 */

static struct attr {
  snz_clnt_t *clnt;
  int attr;
} *client_table = 0;
static int max_clients = 0;

/* Public because accept_new_client needs it, and is external. */
snz_clnt_t *_snz_new_client( int fd )
{
  snz_clnt_t *c;

  c = (snz_clnt_t *)malloc( sizeof( snz_clnt_t ) );
  c->read = 0;
  c->write = 0;
  c->fd = fd;
  c->tag = 0;
  c->protocol = 0;
  register_client( c );
  return c;
}

static void register_client( snz_clnt_t *c )
{
  set_attr( c->fd, FD_CLIENT );
  client_table[ c->fd ].clnt = c;
}

static void set_attr( int fd, int attr )
{
  if (fd >= max_clients) {
    max_clients = fd+1;
    client_table = (struct attr *)realloc( client_table,
					  sizeof(struct attr)*max_clients );
    if (client_table == 0)
      snz_perror_panic( "realloc" );
  }
  client_table[fd].attr = attr;
}

static snz_clnt_t *fd2client( int fd )
{
  return client_table[ fd ].clnt;
}

static int fdattr( int fd )
{
  return client_table[ fd ].attr;
}

/***************************************************************************
 *
 * Create connections.
 */

/* Assuming that proxy_fd has a valid server file descriptor, use it 
 * to wait for connections from the program processes. Return an array
 * of file descriptors of length num_contexts, one descriptor for each
 * open channel to a program process.
 */
static int *connect_with_program( int proxy_fd, snz_proginfo_t *info )
{
  int *program_fds, i, fd;
  struct cidinfo x;

  snz_message( "Connect_with_program %d contexts", num_contexts );

  info->version = 1;
  info->numcontexts = num_contexts;
  if ((info->numthreads = (int*)malloc( sizeof(int)*num_contexts )) == 0)
    snz_perror_panic( "malloc" );

  if ((program_fds = (int*)malloc( sizeof(int)*num_contexts )) == 0)
    snz_perror_panic( "malloc" );

  for ( i = 0 ; i < num_contexts ; i++ ) {
    fd = accept_connection( proxy_fd );
    if (read( fd, (void*)&x, sizeof( x ) ) < sizeof( x )) /*FIXME*/
      snz_perror_panic( "read" );
    snz_message( "fd %d context %d %d threads", fd, x.cid, x.nthreads );
    program_fds[x.cid] = fd;
    info->numthreads[x.cid] = x.nthreads;
  }

  return program_fds;
}


/* eof */
