/* s e r v e r - a p i . c
 * $Id: server-api.c,v 1.1 1996/05/01 20:09:55 lth Exp $
 *
 * Copyright (C) 1995 The University of Oregon
 *
 *   Author: Lars Thomas Hansen [lth@cs.uoregon.edu]
 *  Created: November 3, 1995
 * Modified:
 *
 * Machine-independent implementation of program (server) interface to 
 * Sneezy. The implementation of the virtual functions in the the 
 * snz_sconn_t structure can be found in server-remote.c or server-direct.c.
 */

#define SERVER_SIDE

#include "sneezy.h"
#include "server-api.h"
#include "server-low.h"

/* _snz_setup(), _snz_shutdown(), _snz_whoami(), _snz_get_event_id(),
 * and _snz_lookup_connection() depend on the computation and RTS model;
 * see e.g. server-spmd.c for their implementations.
 */
void snz_register_context( void )
{
  _snz_setup();
}

void snz_shutdown_context( void )
{
  _snz_shutdown();
}

void snz_whoami( int *context, int *thread )
{
  _snz_whoami( context, thread );
}

int snz_get_event_id( int event_type )
{
  int cid, tid;

  _snz_whoami( &cid, &tid );
  return _snz_get_event_id( cid, tid, event_type );
}

/* This can be faster if a single function call is used to lookup all
 * the data.
 */
#define PROLOGUE( _etype ) \
  int cid, tid, eid; \
  snz_sconn_t *c; \
  _snz_whoami( &cid, &tid ); \
  eid = _snz_get_event_id( cid, tid, _etype ); \
  if (eid == -1) return; \
  c = _snz_lookup_connection( cid, tid )

void snz_ready_event( void )
{
  int cid, tid, eid;
  snz_sconn_t *c;

  _snz_whoami( &cid, &tid );
  eid = SNZEID_READY;
  c = _snz_lookup_connection( cid, tid );
  c->ready( c, eid, cid, tid );
}

void snz_error_event( int code )
{
  PROLOGUE( SNZE_ERROR );
  c->err( c, eid, cid, tid, code );
}

void snz_terminate_event( void )
{
  PROLOGUE( SNZE_TERMINATE );
  c->term( c, eid, cid, tid );
}

void snz_begin_barrier_event( int barrier_id )
{
  PROLOGUE( SNZE_BEGIN_BARRIER );
  c->bbarr( c, eid, cid, tid, barrier_id );
}

void snz_end_barrier_event( int barrier_id )
{
  PROLOGUE( SNZE_END_BARRIER );
  c->ebarr( c, eid, cid, tid, barrier_id );
}

void snz_begin_fetch_event( int src_cid, void *src_addr, void *dest_addr, 
			    int len )
{
  snz_gptr_t gs, gd;
  PROLOGUE( SNZE_BEGIN_FETCH );
  gd = snz_mkgptr( cid, dest_addr );
  gs = snz_mkgptr( src_cid, src_addr );
  c->bfetch( c, eid, cid, tid, gs, gd, len );
}

void snz_end_fetch_event( int src_cid, void *src_addr, void *dest_addr,
			  int len )
{
  snz_gptr_t gs, gd;
  PROLOGUE( SNZE_END_FETCH );
  gd = snz_mkgptr( cid, dest_addr );
  gs = snz_mkgptr( src_cid, src_addr );
  c->efetch( c, eid, cid, tid, gs, gd, len );
}

void snz_begin_put_event( int dest_cid, void *dest_addr, void *src_addr,
			  int len )
{
  snz_gptr_t gs, gd;
  PROLOGUE( SNZE_BEGIN_PUT );
  gd = snz_mkgptr( dest_cid, dest_addr );
  gs = snz_mkgptr( cid, src_addr );
  c->bput( c, eid, cid, tid, gs, gd, len );
}

void snz_end_put_event( int dest_cid, void *dest_addr, void *src_addr,
		        int len )
{
  snz_gptr_t gs, gd;
  PROLOGUE( SNZE_BEGIN_PUT );
  gd = snz_mkgptr( dest_cid, dest_addr );
  gs = snz_mkgptr( cid, src_addr );
  c->eput( c, eid, cid, tid, gs, gd, len );
}

/* pC++ specific -- RPC on remote element */
/* function is some abstract 'function id', like a member pointer */
void snz_begin_rpc_event( int dest_cid, void *dest_addr, int function,
			  void *params, int plen )
{
  snz_gptr_t gd, gs;
  PROLOGUE( SNZE_BEGIN_RPC );
  gd = snz_mkgptr( dest_cid, dest_addr );
  gs = snz_mkgptr( cid, params );
  c->brpc( c, eid, cid, tid, gd, function, gs, plen );
}

/* Ditto */
void snz_end_rpc_event( int dest_cid, void *dest_addr, int function,
		        void *params, int plen )
{
  snz_gptr_t gd, gs;
  PROLOGUE( SNZE_END_RPC );
  gd = snz_mkgptr( dest_cid, dest_addr );
  gs = snz_mkgptr( cid, params );
  c->erpc( c, eid, cid, tid, gd, function, gs, plen );
}

void snz_begin_service_rpc_event( int src_cid, void *buf, int len )
{
  /*FIXME*/
  /* should this trigger on remote fetch/put requests? */
}

void snz_end_service_rpc_event( int src_cid, void *buf, int len )
{
  /*FIXME*/
  /* should this trigger on remote fetch/put requests? */
}

void snz_begin_parallel_event( void )
{
  PROLOGUE( SNZE_BEGIN_PARALLEL );
  c->bpar( c, eid, cid, tid );
}

void snz_end_parallel_event( void )
{
  PROLOGUE( SNZE_END_PARALLEL );
  c->epar( c, eid, cid, tid );
}

/* FIXME: do we want element index/indices here? */
void snz_enter_element_event( void *eltaddr )
{
  snz_gptr_t g;
  PROLOGUE( SNZE_ENTER_ELEMENT );
  g = snz_mkgptr( cid, eltaddr );
  c->eelt( c, eid, cid, tid, g );
}

void snz_exit_element_event( void *eltaddr )
{
  snz_gptr_t g;
  PROLOGUE( SNZE_EXIT_ELEMENT );
  g = snz_mkgptr( cid, eltaddr );
  c->xelt( c, eid, cid, tid, g );
}

void snz_user_event( int event_code, void *p, int len )
{
  snz_gptr_t g;
  PROLOGUE( SNZE_USER_EVENT );
  g = snz_mkgptr( cid, p );
  c->user( c, eid, cid, tid, event_code, g, len );
}

void snz_begin_create_collection_event( int cnum )
{
  PROLOGUE( SNZE_BEGIN_CREATE_COLLECTION );
  c->bcc( c, eid, cid, tid, cnum );
}

void snz_end_create_collection_event( int cnum )
{
  PROLOGUE( SNZE_END_CREATE_COLLECTION );
  c->ecc( c, eid, cid, tid, cnum );
}

void snz_begin_delete_collection_event( int cnum )
{
  PROLOGUE( SNZE_BEGIN_DELETE_COLLECTION );
  c->bdc( c, eid, cid, tid, cnum );
}

void snz_end_delete_collection_event( int cnum )
{
  PROLOGUE( SNZE_END_DELETE_COLLECTION );
  c->edc( c, eid, cid, tid, cnum );
}

void snz_enter_function_event( int function_id )
{
  snz_time_t t = _snz_get_time();
  PROLOGUE( SNZE_ENTER_FUNCTION );
  c->efun( c, eid, cid, tid, function_id, t );
}

void snz_exit_function_event( int function_id )
{
  snz_time_t t = _snz_get_time();
  PROLOGUE( SNZE_EXIT_FUNCTION );
  c->xfun( c, eid, cid, tid, function_id, t );
}

void snz_breakpoint_event( snz_addr_t *ip, int token )
{
#if 0
  PROLOGUE( SNZE_BREAKPOINT );
  c->brkpt( c, eid, cid, tid, ip );
  _snz_breakpoint_return( token );
#endif
}

void snz_push_value( int code, void *data, size_t len )
{
  int cid, tid, eid;
  snz_sconn_t *c;

  _snz_whoami( &cid, &tid );
  eid = SNZEID_PUSH;
  c = _snz_lookup_connection( cid, tid );
  c->push( c, eid, cid, tid, code, data, len );
}

/* eof */
