/* s e r v e r - s p m d . c
 * $Id: server-spmd.c,v 1.1 1996/05/01 20:10:00 lth Exp $
 *
 * Copyright (C) 1995 The University of Oregon
 *
 *   Author: Lars Thomas Hansen [lth@cs.uoregon.edu]
 *  Created: November 3, 1995
 * Modified: March 6, 1996 / lth: implemented call-user-function.
 *
 * Server code for pure (single thread per context) SPMD systems. 
 */

#define SERVER_SIDE

#include <stdlib.h>
#include <string.h>

#include "sneezy.h"
#include "server-low.h"
#include "proxy-sock.h"  /* for now */

static snz_sconn_t *my_connection;
static int my_node;
static int evtbl[ SNZ_MAX_EVENTS ];

static void cmd_terminate( snz_sconn_t *c, int cid, int tid );
static void cmd_install( snz_sconn_t *c, int cid, int tid, int etype, int eid);
static void cmd_uninstall( snz_sconn_t *c, int cid, int tid, int etype );
static void cmd_read( snz_sconn_t *c, snz_gptr_t g, int len, void *buf );
static void cmd_write( snz_sconn_t *c, snz_gptr_t g, int len, void *data );
static void cmd_call_func( snz_sconn_t *, void(*fptr)(void), int, snz_arg_t*);

void _snz_setup( void )
{
  snz_sconn_t *c;
  int i;

  my_node = _snz_spmd_nodenumber();
  snz_message( "_snz_setup: %d", my_node );
  c = my_connection = snz_server_init_context( my_node, 1 ); /*proxy-server.c*/

  /* Fixup connection to contain correct command handlers. */
  c->read_data = cmd_read;
  c->write_data = cmd_write;
  c->terminate = cmd_terminate;
  c->install = cmd_install;
  c->uninstall = cmd_uninstall;
  c->call_func = cmd_call_func;
  for ( i = 0 ; i < SNZ_MAX_EVENTS ; i++ )
    evtbl[i] = -1;
}

void _snz_shutdown( void )
{
  snz_server_shutdown_context( my_node, my_connection ); /*proxy-server.c*/
}

void _snz_whoami( int *cid, int *tid )
{
  *cid = my_node;
  *tid = 0;
}

snz_sconn_t *_snz_lookup_connection( int cid, int tid )
{
  return my_connection;
}

int _snz_get_event_id( cid, tid, event_type )
{
  return evtbl[ event_type ];
}

/* Command implementations */

static void cmd_terminate( snz_sconn_t *c, int cid, int tid )
{
  _snz_spmd_terminate_thread();   /* RTS dependent, see e.g. pcxx.c */
}

static void cmd_install( snz_sconn_t *c, int cid, int tid, int etype, int eid )
{
  snz_message( "[install %d %d]", etype, eid );
  evtbl[ etype ] = eid;
}

static void cmd_uninstall( snz_sconn_t *c, int cid, int tid, int etype )
{
  evtbl[ etype ] = -1;
}

static void cmd_read( snz_sconn_t *c, snz_gptr_t g, int len, void *buf )
{
  if (g.context != my_node)
    snz_panic( "REMOTE READ REQUEST: %d %p", g.context, (void*)g.addr );
  memcpy( (void*)buf, (void*)(g.addr), len );
}

static void cmd_write( snz_sconn_t *c, snz_gptr_t g, int len, void *buf )
{
  if (g.context != my_node)
    snz_panic( "REMOTE WRITE REQUEST: %d %p", g.context, (void*)g.addr );
  memcpy( (void*)(g.addr), buf, len );
}

/* This is impossible to do portably, in general.  We provide a 
   limited solution. Arguments are either int-sized or long-sized, and
   we provide for up to four of them, for arbitrary combinations.
   
   A fully general solution is an apply() function, but it is necessarily
   machine-dependent.  Every C library should have one.

   NOTE:
   This code is correct only if shorts and chars are always widened to
   ints at call boundaries.
 */
static void cmd_call_func( snz_sconn_t *c, void (*fptr)(void), int argc,
			   snz_arg_t *argv )
{
  unsigned int  iarg[4];
  unsigned long larg[4];
  int iargc, largc, i, k;

  if (argc > 4) 
    snz_panic( "server-spmd: CALL-USER-FUNCTION: too many args: %d (limit 4)", 
	       argc );
  iargc = largc = 0;
  for ( i = 0, k = 0; i < 4; i++ ) {
    k *= 2;
    if (i >= argc) continue;  /* HACK to make sure bits are in right place */
    switch (argv[i].tag) {
      case SNZAT_INT    : iarg[iargc++] = (unsigned)argv[i].u.i; break;
      case SNZAT_UINT   : iarg[iargc++] = (unsigned)argv[i].u.ui; break;
      case SNZAT_SHORT  : iarg[iargc++] = (unsigned)argv[i].u.s; break;
      case SNZAT_USHORT : iarg[iargc++] = (unsigned)argv[i].u.us; break;
      case SNZAT_CHAR   : iarg[iargc++] = (unsigned)argv[i].u.c; break;
      case SNZAT_UCHAR  : iarg[iargc++] = (unsigned)argv[i].u.uc; break;
      case SNZAT_SCHAR  : iarg[iargc++] = (unsigned)argv[i].u.sc; break;
      case SNZAT_LONG   : larg[largc++] = (unsigned long)argv[i].u.l; k += 1; 
	break; 
      case SNZAT_ULONG  : larg[largc++] = (unsigned long)argv[i].u.ul; k += 1; 
	break; 
      case SNZAT_CHARPTR:
	if (sizeof( char * ) == sizeof( int  ))
	  /* Ignore warnings here */
	  iarg[iargc++] = (unsigned)argv[i].u.str.cp;
	else {
	  larg[largc++] = (unsigned long)argv[i].u.str.cp; k+=1;
	}
	break;
    }
  }

  fprintf( stderr, "Call User Function: %lx %d\n", (unsigned long)fptr,argc );
  fprintf( stderr, "  %x %x %x %x\n", iarg[0],iarg[1],iarg[2],iarg[3] );
  fprintf( stderr, "  %lx %lx %lx %lx\n", larg[0],larg[1],larg[2],larg[3] );

  /* The following code is auto-generated by the program "gencall.sch".
   * It can be extended to handle other cases, too, but the amount
   * of code will grow rather quickly.
   */
  switch (k) {
  case 0:
    ((void (*)(unsigned,unsigned,unsigned,unsigned))fptr)(iarg[0],iarg[1],iarg[2],iarg[3]); break;
  case 1:
    ((void (*)(unsigned,unsigned,unsigned,unsigned long))fptr)(iarg[0],iarg[1],iarg[2],larg[0]); break;
  case 2:
    ((void (*)(unsigned,unsigned,unsigned long,unsigned))fptr)(iarg[0],iarg[1],larg[0],iarg[2]); break;
  case 3:
    ((void (*)(unsigned,unsigned,unsigned long,unsigned long))fptr)(iarg[0],iarg[1],larg[0],larg[1]); break;
  case 4:
    ((void (*)(unsigned,unsigned long,unsigned,unsigned))fptr)(iarg[0],larg[0],iarg[1],iarg[2]); break;
  case 5:
    ((void (*)(unsigned,unsigned long,unsigned,unsigned long))fptr)(iarg[0],larg[0],iarg[1],larg[1]); break;
  case 6:
    ((void (*)(unsigned,unsigned long,unsigned long,unsigned))fptr)(iarg[0],larg[0],larg[1],iarg[1]); break;
  case 7:
    ((void (*)(unsigned,unsigned long,unsigned long,unsigned long))fptr)(iarg[0],larg[0],larg[1],larg[2]); break;
  case 8:
    ((void (*)(unsigned long,unsigned,unsigned,unsigned))fptr)(larg[0],iarg[0],iarg[1],iarg[2]); break;
  case 9:
    ((void (*)(unsigned long,unsigned,unsigned,unsigned long))fptr)(larg[0],iarg[0],iarg[1],larg[1]); break;
  case 10:
    ((void (*)(unsigned long,unsigned,unsigned long,unsigned))fptr)(larg[0],iarg[0],larg[1],iarg[1]); break;
  case 11:
    ((void (*)(unsigned long,unsigned,unsigned long,unsigned long))fptr)(larg[0],iarg[0],larg[1],larg[2]); break;
  case 12:
    ((void (*)(unsigned long,unsigned long,unsigned,unsigned))fptr)(larg[0],larg[1],iarg[0],iarg[1]); break;
  case 13:
    ((void (*)(unsigned long,unsigned long,unsigned,unsigned long))fptr)(larg[0],larg[1],iarg[0],larg[2]); break;
  case 14:
    ((void (*)(unsigned long,unsigned long,unsigned long,unsigned))fptr)(larg[0],larg[1],larg[2],iarg[0]); break;
  case 15:
    ((void (*)(unsigned long,unsigned long,unsigned long,unsigned long))fptr)(larg[0],larg[1],larg[2],larg[3]); break;
  }
}


/* On an SPMD system, the code spaces are unshared, and breakpoints are
 * handled without fuss: there is a per-address-space table which maps
 * breakpoint addresses to EIDs.
 *
 * How to handle multiple breakpoints on the same address?
 */
static void cmd_setbrkpt( snz_sconn_t *c, snz_addr_t ip, int eid )
{
#if 0
  _snz_arch_setbrkpt( ip );  /* Just set it */
  _snz_associate_brkpt( ip, eid );
#endif
}


/* eof */
