/* s e x p r . c
 * $Id: sexpr.c,v 1.1 1996/05/01 20:10:03 lth Exp $
 *
 * Copyright (C) 1995 The University of Oregon
 *
 *   Author: Lars Thomas Hansen [lth@cs.uoregon.edu]
 *  Created: October 25, 1995
 * Modified:
 *
 * Simple S-expression parser for Sneezy's Lisp Syntax mode.
 *
 * BUGS
 *  - It would be useful to abstract away from the use of FILE * and 
 *    move to a sneezy-specific FILE-like structure which could be as
 *    efficient (would map 1-1 onto FILEs in the trivial case, including
 *    macros).
 */ 

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#if 0 /* handy for debugging */
#undef getc
static int getc( FILE *fp )
{
  int c = fgetc( fp );
  if (c == EOF) puts( "<EOF>" ); else putchar( c ); 
  fflush( stdout );
  return c;
}
#endif


#if !defined(CLIENT_SIDE) && !defined(SERVER_SIDE)
#error "Must have either CLIENT_SIDE or SERVER_SIDE"
#endif

#ifdef CLIENT_SIDE
#include "client-api.h"
#endif

#ifdef SERVER_SIDE
#include "server-types.h"
#endif

#define FULL_TYPECHECKS
#include "sexpr.h"

#define SEXPR_SEG_SIZE 32

struct sexpr_seg_t {
  sexpr_obj_t mem[ SEXPR_SEG_SIZE ];
  sexpr_seg_t *next;
};

static sexpr_seg_t *allocsegment( void );
static void addsegment( sexpr_mem_t *mem, sexpr_seg_t *seg );
#ifdef CLIENT_SIDE
static sexpr_obj_t *parse_list( snz_conn_t *, sexpr_mem_t *mem );
static sexpr_obj_t *readatom( snz_conn_t *, sexpr_mem_t *mem );
#else
static sexpr_obj_t *parse_list( FILE *, sexpr_mem_t *mem );
static sexpr_obj_t *readatom( FILE *, sexpr_mem_t *mem );
#endif

/* These are used for their addresses exclusively */
static sexpr_obj_t atom_lp, atom_rp, atom_error;

sexpr_mem_t *sexpr_alloc_mem( void )
{
  sexpr_mem_t *p;

  if ((p = (sexpr_mem_t*)malloc( sizeof( sexpr_mem_t ) )) == 0)
    snz_perror_panic( "malloc" );
  p->freelist = 0;
  p->segments = 0;
  addsegment( p, allocsegment() );
  return p;
}

void sexpr_clear_mem( sexpr_mem_t *mem )
{
  sexpr_seg_t *q, *n;

  mem->freelist = 0;
  q = mem->segments;
  mem->segments = 0;
  while (q != 0) {
    n = q->next;
    addsegment( mem, q );
    q = n;
  }
}

void sexpr_release_mem( sexpr_mem_t *mem )
{
  sexpr_seg_t *p, *n;

  p = mem->segments;
  while (p != 0) {
    n = p->next;
    free( p );
    p = n;
  }
  mem->segments = 0;
  mem->freelist = 0;
}

static sexpr_obj_t *alloc( sexpr_mem_t *mem )
{
  sexpr_obj_t *p;

  if (mem->freelist == 0)
    addsegment( mem, allocsegment() );
  if (mem->freelist == 0)
    snz_panic( "freelist consistency problem." );
  p = mem->freelist;
  mem->freelist = car(mem->freelist);
  return p;
}

#ifdef CLIENT_SIDE
sexpr_obj_t *sexpr_parse( snz_conn_t *conn, sexpr_mem_t *mem )
#else
sexpr_obj_t *sexpr_parse( FILE *conn, sexpr_mem_t *mem )
#endif
{
  sexpr_obj_t *atom;
  int c;

  atom_error.tag = TAG_ERROR;
  atom_lp.tag = TAG_LP;
  atom_rp.tag = TAG_RP;
  atom = readatom( conn, mem );
  if (atom == &atom_lp)
    atom = parse_list( conn, mem );
  else
    atom = 0;
  if (atom && atom->tag == TAG_PAIR) return atom;

#ifdef CLIENT_SIDE
  c = getc( conn->u.socket.fp_in );
#else
  c = getc( conn );
#endif
  if (c != EOF)
    snz_panic( "sexpr_parse: illegal char (ASCII %d) on input stream", c );

  return 0; /*EOF*/
}

#ifdef CLIENT_SIDE
static sexpr_obj_t *parse_list( snz_conn_t *c, sexpr_mem_t *mem )
#else
static sexpr_obj_t *parse_list( FILE *c, sexpr_mem_t *mem )
#endif
{
  sexpr_obj_t *first = 0, *prev = 0, *p, *atom;

  do {
    atom = readatom( c, mem );
    if (atom == &atom_error) return atom;
    if (atom != &atom_rp) {
      if (atom == &atom_lp) 
	atom = parse_list( c, mem );
      if (atom == &atom_error) return atom;
      p = alloc( mem );
      p->tag = TAG_PAIR;
      p->u.pair.car = atom;
      if (first == 0)
	first = prev = p;
      else
	prev = prev->u.pair.cdr = p;
    }
  } while (atom != &atom_rp);

  if (prev == 0)
    return 0;
  else {
    prev->u.pair.cdr = 0;
    return first;
  }
}

#ifdef CLIENT_SIDE
static sexpr_obj_t *readatom( snz_conn_t *conn, sexpr_mem_t *mem )
#else
static sexpr_obj_t *readatom( FILE *fp, sexpr_mem_t *mem )
#endif
{
  int c;
  sexpr_obj_t *p;
#ifdef CLIENT_SIDE
  FILE *fp = conn->u.socket.fp_in;  /* Hmmmm. */
#endif

  while ((c = getc( fp )) != EOF && isspace(c))
    ;
  if (c == EOF)
    return &atom_error;
  else if (c == '(')
    return &atom_lp;
  else if (c == ')')
    return &atom_rp;
  else if (isdigit( c ) || c == '-') {
    char n[ 21 ];  /* ceil(log10(2^64))+1 */ /* FIXME */
    char *q = n;

    *q++ = c;
    while ((c=getc( fp )) != EOF && isdigit(c))
      *q++ = c;
    ungetc( c, fp );
    *q = 0;
    p = alloc( mem );
    p->tag = TAG_NUMBER;
#ifdef CLIENT_SIDE
    p->u.num = snz_str2long( conn, n );
#else
    p->u.num = snz_str2long( n );
#endif
    return p;
  }
  else if (c == '#') {
    c = getc( fp );
    if (c == 'x' || c == 'X') {
      char n[ 22 ]; /* FIXME */
      char *q = n;
      
      *q++ = '#'; /* [sic] */
      *q++ = 'x';
      if ((c = getc( fp )) == '-') *q++ = '-'; else ungetc( c, fp );
      while ((c=getc( fp )) != EOF && isxdigit( c ))
	*q++ = c;
      ungetc( c, fp );
      *q = 0;
      p = alloc( mem );
      p->tag = TAG_NUMBER;
#ifdef CLIENT_SIDE
      p->u.num = snz_str2long( conn, n ); 
#else
      p->u.num = snz_str2long( n );
#endif
      return p;
    }
    else 
      snz_panic( "readatom: SYNTAX: #%c not supported.", c );
  }
  else if (c == '"') {
    /* Parse string, alloc, return.
     *
     * Assumption: both machines have bytes of same size, and both 
     * have same character set.
     *
     * The worst-case performance of this is pretty bad, but should be OK
     * in most cases (not too many reallocs requiring copying).
     */
    int bufmax = 512;
    char *buf;
    int i = 0;

    if ((buf = (char*)malloc( bufmax )) == 0)
      snz_perror_panic( "malloc" );
    while ((c = getc( fp )) != '"' && c != EOF) {
      if (c == '\\') {
	c = getc( fp );
	if (c == EOF) return &atom_error;
      }
      buf[i++] = c;
      if (i == bufmax) {
	bufmax *= 2;
	if ((buf = (char*)realloc( buf, bufmax )) == 0)
	  snz_perror_panic( "realloc" );
      }
    }
    if (c == EOF)
      return &atom_error;
    buf[i] = 0;
    p = alloc( mem );
    p->tag = TAG_STRING;
    p->u.str.length = i;
    p->u.str.data = buf;
    return p;
  }
  else {
    ungetc( c, fp );
    return &atom_error;
  }
  /*NOTREACHED*/
  return &atom_error;
}

static sexpr_seg_t *allocsegment( void )
{
  sexpr_seg_t *p;
  if ((p = (sexpr_seg_t*)malloc( sizeof( sexpr_seg_t ) )) == 0)
    snz_perror_panic( "malloc" );
  return p;
}

static void addsegment( sexpr_mem_t *mem, sexpr_seg_t *seg )
{
  int i;

  for (i = 0 ; i < SEXPR_SEG_SIZE ; i++ ) { /* index: [sic] */
    seg->mem[i].u.pair.car = &(seg->mem[i+1]);
    seg->mem[i].u.pair.cdr = 0;
    seg->mem[i].tag = TAG_PAIR;
  }
  seg->mem[SEXPR_SEG_SIZE-1].u.pair.car = mem->freelist;
  seg->next = mem->segments;
  mem->freelist = seg->mem;
}


int length( sexpr_obj_t *q )
{
  int l = 0;
  while (q != 0) {
    q = cdr(q);
    l++;
  }
  return l;
}

sexpr_obj_t *_sexpr_panic( const char *str, int tag, sexpr_obj_t *p )
{
  if (p == 0)
    snz_panic( "%s: Null pointer.", str );
  else
    snz_panic( "Illegal type to operation %s, tag is %d", str, p->tag );
  /*NOTREACHED*/
  return 0;
}

/**************************************************************************
 *
 * Encode and decode raw data for transmission as S-expression.
 * Can do this any number of ways; for now, encode each byte as a two-byte
 * textual hex number, and 0-terminate.  In addition, encode length as
 * two-byte big-endian number at beginning.  Return buffer; buffer must
 * be explicitly freed by calling free().
 *
 * FIXME: should allow more data.
 * FIXME: more efficient encoding. Things to consider include hints of the
 *        type "almost always text" (i.e., encode as ASCII with escape chars
 *        when necessary), or uuencoding (which results in less bloat).
 *        One could also extend the parser to cope with raw binary data, but
 *        that would make it unusable in the context of "real" Lisp readers.
 */

static char *hex = "0123456789abcdef";

#define putbyte( p, b )  \
  do { *p++ = hex[((b)>>4)&15]; *p++=hex[(b)&15]; } while(0)

#define unhex( x )  ((x)<='9' ? (x)-'0' : (x)-'a'+10)

#define getbyte( p, b ) \
  do { (b) = (unhex( *(p) ) << 4) | unhex( *((p)+1) ); p+=2; } while(0)

char *sexpr_encode_data( unsigned char *data, unsigned len )
{
  char *buf, *s;
  int i;

  if (len > 65535)
    snz_panic( "Too much data!" );
  buf = (char*)malloc( (len*2)+5 );
  if (!buf)
    snz_perror_panic( "malloc" );
  s = buf;
  putbyte( s, len/256 );
  putbyte( s, len%256 );
  for ( i = 0 ; i < len ; i++, data++ )
    putbyte( s, *data );
  *s = 0;
  return buf;
}

/* Decode data. */
/* Sometimes this can be done in-line, which would save allocation.
 * An extra parameter could encode this possibility.  FIXME.
 */
unsigned char *sexpr_decode_data( char *data, unsigned *plen )
{
  unsigned char *buf, *s;
  unsigned l, l1, l2;
  int i;

  getbyte( data, l1 );
  getbyte( data, l2 );
  l = l1*256+l2;
  if (plen)
    *plen = l;
  buf = (unsigned char *)malloc( l );
  if (buf == 0)
    snz_perror_panic( "malloc" );
  s = buf;
  for ( i = 0 ; i < l ; i++ ) {
    getbyte( data, l1 );
    *s++ = l1;
  }
  return buf;
}

/* eof */
