/* s o c k e t . c
 * $Id: socket.c,v 1.1 1996/05/01 20:10:07 lth Exp $
 *
 * Copyright (C) 1995 The University of Oregon.
 *
 *   Author: Lars Thomas Hansen [lth@cs.uoregon.edu]
 *  Created: November 9, 1995
 * Modified: January 24, 1996 / lth -- read_bytes and write_bytes
 *
 * Code for BSD sockets. 
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <stropts.h>
#include <poll.h>
#include <unistd.h>
#include <errno.h>

#include "socket.h"

static void set_socket_options( int fd );

/* These are arbitrary -- I can't remember (and can't find) which port 
 * number ranges are actually available to the user.
 */
#define FIRST_PORT 2000
#define LAST_PORT  4999

#define WAIT_TIME      10     /* seconds */

/* Open a server socket on the local host; return file descriptor and
 * port number.  If *portno is initially 0, then this procedure will attempt
 * to find a vacant port in the range [FIRST_PORT..LAST_PORT].
 * If it can't, it will panic.
 * Returns the fd, but also puts the fd and the port in the value-result 
 * parameters.
 * 
 * FIXME: if we're not iterating on the port space then we should retry
 * if the first bind fails.
 */
int open_server_socket( int *portno, int *outfd )
{
  int fd, scanports = 0, alen;
  struct sockaddr_in addr;

  fprintf( stderr, "enter: server %d\n", *portno ); fflush( stdout );
  if (*portno == 0) {
    scanports = 1;
    *portno = FIRST_PORT;
  }

  fd = socket( AF_INET, SOCK_STREAM, 0 );
  if (fd == 0) 
    snz_perror_panic( "open_server_socket: socket" );

  memset((char *) &addr, 0, sizeof(addr));
  addr.sin_family = AF_INET;
  addr.sin_addr.s_addr=htonl(INADDR_ANY);

  alen = sizeof( addr );

  /* Search for an available port. */
  /* If we can find none the first time around, we should sleep and retry */
  while (*portno <= LAST_PORT || !scanports) {
    addr.sin_port = htons(*portno);
    if (bind( fd, (struct sockaddr*)&addr, alen ) != -1) break;
    if (!scanports)
      snz_perror_panic( "open_server_socket: bind (port %d)", *portno );
    *portno += 1;
  }

  set_socket_options( fd );

  if (listen( fd, 5 ) == -1) 
    snz_perror_panic( "open_server_socket: listen" );

  *outfd = fd;
  return fd;
}


/* Connect to a server and return the fd for the connection. */
int connect_to_server( char *hostname, int portnumber )
{
  struct sockaddr_in sin;
  struct hostent *hp;
  int fd, i;

  for (i = 0 ; i < 5 ; i++ ) {

    if ((fd = socket( AF_INET, SOCK_STREAM, 0 )) < 0)
      snz_perror_panic( "snz_connect_to_server: socket" );

    sin.sin_family = AF_INET;
    sin.sin_port = htons( portnumber );

    if ((hp = gethostbyname( hostname )) != NULL)
      sin.sin_addr.s_addr = *(long *)(hp->h_addr);
    else
      sin.sin_addr.s_addr = htonl( inet_addr( hostname ) );

    if (connect( fd, (struct sockaddr*)&sin, sizeof(sin)) < 0) {
      snz_message( "Failed to connect; sleeping..." );
      close( fd );
      sleep( 5 );
      continue;
    }

    set_socket_options( fd );
    return fd;
  }
  snz_panic( "snz_connect_to_server: can't connect to %s %d.", 
	    hostname, portnumber );
  return -1;
}  


/* Accept connection from a client and return the fd for the connection. */
/* FIXME: this could be made to retry if necessary */
int accept_connection( int fd )
{
  struct sockaddr_in addr;
  int addrsize, newfd, one = 1;

  addrsize = sizeof( addr );

 AGAIN:
  if ((newfd = accept( fd, &addr, &addrsize )) < 0) {
    if (errno != EINTR && errno != EAGAIN)
      snz_perror_panic( "accept" );
    else {
      snz_message( "accept: %s; repeating.", errno==EINTR?"EINTR":"EAGAIN" );
      goto AGAIN;
    }
  }

  set_socket_options( newfd );

  return newfd;
}

char *myaddr( void )
{
  static char buf[ 256 ];

  if (gethostname( buf, sizeof( buf ) ) == 0)
    return buf;
  else
    return 0;
}


/* WARNING: the definition of hasbuffered() is not portable. What we want
 * to discover is whether there is any data in the FILE pointer's buffer.
 * We need this information to implement the semantics of sock_wait().
 * The FILE structure on many Unix systems have a member _cnt which
 * is the number of chars left in the buffer.  (Sometimes it's called 
 * __cnt.  Look in /usr/include/stdio.h if you're not sure.)
 * On systems where this is not true (Linux, for example), some other, 
 * equivalent method usually exists to allow getc() to be implemented as 
 * a macro.  If no such method exists you're out of luck... in principle,
 * the socket level should not use FILE pointers (even though it's very
 * convenient) but do its own buffering on top of open(), read(), and write().
 */
#define hasbuffered( fp )  ((fp)->_cnt > 0)

/* Return 1 if there is input ready on the FILE *. */

int ready( FILE *fp, int how )
{
  struct pollfd f[1];
  int r;

  if (hasbuffered( fp )) return 1;
  if (how == RDY_TRIVIAL) return 0;

  f[0].fd = fileno( fp );
  f[0].events = POLLIN|POLLPRI|POLLRDNORM|POLLRDBAND;
  do {
    r = poll( f, 1, 0 );
  } while (r == -1 && errno == EINTR);
  return r != 0;
}


static void set_socket_options( int fd )
{
  int one = 1;
  struct protoent *pp;

  if ((pp = getprotobyname( "tcp" )) == 0)
    snz_perror_panic( "getprotobyname" );

  /* Make port address available immediately */
  if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (char*)&one, sizeof(one)) == -1)
    snz_perror_panic("setsockopt: SO_REUSEADDR");

  /* Do not wait for more output to arrive before sending */
  if (setsockopt(fd, pp->p_proto, TCP_NODELAY, (char*)&one, sizeof(one)) == -1)
    snz_perror_panic("setsockopt: TCP_NODELAY");
}


/* Read n bytes from the file descriptor. 
   On error, print error message and return 0.
   On eof during read, return 0.
   Otherwise, return 1.
*/
int read_bytes(int fd, void *buf, int nbytes)
{
  char *ptr = (char*)buf;

  while (nbytes) {
    int n;

    n = read( fd, ptr, nbytes );
    if (n < 0) {
      perror( "read" );
      return 0;
    }

    if (n == 0)
      return 0;  /* EOF */
      
    nbytes -= n;
    ptr    += n;
  }
  return 1;
}

/* Write n bytes to the file descriptor.
   On error, print error message and return 0.
   Otherwise, return 1.
 */
int write_bytes(int fd, void *buf, int nbytes)
{
  char *ptr = (char*)buf;

  while (nbytes) {
    int n;

    n = write(fd, ptr, nbytes );
    if (n <= 0) {
      perror( "write" );
      return 0;
    }
    nbytes -= n;
    ptr   += n;
  }
  return 1;
}

/* eof */
