; c l i e n t . s c h
; $Id: client.sch,v 1.1 1996/05/01 20:11:11 lth Exp $
;
; Sneezy client for connection-based communication.
; Lars Thomas Hansen [lth@cs.uoregon.edu]
;
; Created June 19 - July 6, 1995 / lth
; Modified November 30 1995 / lth
; Modified March 12, 1996 / lth -- added call-user-function.
;
; A handler procedure is a procedure of four arguments: etype, cid, tid,
; and the argument data.
;
; USAGE
;  (client.initialize hostname portnumber ready-handler) => connection
;    Where a connection is a data type defined in connect.sch.
;
;  (client.data key . more)  =>  data
;    Retrieve data about the server. 'Key' is a symbol, one of 
;      'server-context-count  - return number of contexts
;      'all-contexts          - return list of all context numbers
;      'threads-in-context    - given context number, return number of threads
;
;  (client.cleanup)
;    Shut down the client.
;
;  (client.control key) => data
;    Control the client's behavior. 'Key' is a symbol, one of
;      'monitor         - monitor all communication with debug msgs.
;      'nomonitor       - turn off monitoring.
;      'peek            - return representation of client's state.
;    Client.control can be called _before_ initialize to turn on
;    monitoring. Monitoring is off by default.
;
;  (client.continue cid tid)
;
;  (client.terminate cid tid)
;
;  (client.handle etype cid tid event-handler) => event-number
;    Where an event-number is an integer which identifies the event
;    code, usually not useful for anything.
;
;  (client.unhandle etype cid tid)
;    Remove the installed handler for the given context and thread.
;
;  (client.read cid tid gptr len)
;    Read data; return as standard-encoding string.
;
;  (client.write cid tid gptr string)
;    Write data; data must be supplied as standard-encoding string.
;
;  (client.call-user-function cid tid fptr arglist)
;    Call the indicated user function with the given arguments.  The argument
;    list is in the standard argument list encoding.  The procedure blocks
;    until the call returns and then returns the list of all data PUSHed
;    by the program during the call to the function.  There is no data
;    associated with the reply.
;
; DESCRIPTION
;   A call to client.initialize starts a background event loop and then
;   returns.  The background loop listens on the connection for events,
;   command replies, and data pushes.  When an event arrives, all handlers
;   for the given event code are fetched from the event table and are
;   then run on the thread of the background loop in some unspecified
;   order.  No events or other messages will be processed until the 
;   event handler returns or calls a blocking command procedure.  
;   Long-running event procedures should spawn threads or hand off work
;   to running threads, and return.
;
; BUGS
;   Deadlock possibility in RPC code.  FIX IT!!
;
; TODO
; - Consider allowing multiple connections: parameterize!
; - Error handling - perhaps by installing an error handler?
; - It would be easy to generalize this to use any communication package,
;   all we need is to modularize open, send, receive, and close and the
;   connection data passed to 'initialize' (or perhaps make initialize
;   a procedure which is called with the generic initialize procedure
;   and whatever arguments it's given).
; - Allow one outstanding request per thread rather than per context.

(require "sneezy.sch")
(require "module.sch")
(require "connect.sch")
(require "synch.sch")
(require "sleep.sch")

(module client 
  (initialize
   data
   cleanup
   control 
   (cmd:continue continue)
   (cmd:terminate terminate)
   (cmd:handle handle)
   (cmd:unhandle unhandle)
   (cmd:read read)
   (cmd:write write)
   (cmd:call-user-function call-user-function)
   )

  (define mutex #f)                  ; mutex protects the following vars.
  (define server-connection #f)
  (define server-port-number #f)
  (define server-host-name #f)
  (define ready-handler #f)
  (define monitoring #f)

  (define server-contexts 0)
  (define all-contexts '())
  (define threads-in-context '#())


  ; The ready handler must be given (it is handled specially).

  (define (initialize host port ready)
    (set! mutex (make-mutex))
    (call-with-mutex mutex
      (lambda ()
	(initialize-event-registry)
	(open-server-connection host port)
	(read-server-data)
	(initialize-rpc)
	(set! ready-handler ready)
	(start-event-loop)
	#t)))


  ; Open a connection to the server.

  (define (open-server-connection hostname portno)
    (sneezy.msg "Waiting for server to come up...")
    (set! server-port-number portno)
    (set! server-host-name hostname)
    (let ((conn (let loop ()
		  (let ((c (connect-to hostname portno)))
		    (if c
			c
			(begin (sneezy.msg "No connection; sleeping...")
			       (sleep 5 'seconds)
			       (loop)))))))
      (set! server-connection conn)
      conn))


  ; Read a data structure from the server with the following data:
  ;   (version ((ctxno number-of-threads) ...)

  (define (read-server-data)
    (sneezy.msg "Reading server data...")
    (let ((serverdata (receive)))
      (let ((version    (car serverdata)))
	(if (not (= version 1))
	    (sneezy.error "client.initialize: wrong protocol version: ~a" 
			  version)
	    (let ((d (cadr serverdata)))
	      (set! server-contexts (length d))
	      (set! all-contexts (map car d))
	      (set! threads-in-context (list->vector (map cadr d))))))))


  ; Return data about the server.

  (define (data which . rest)
    (call-with-mutex mutex
      (lambda ()
	(cond ((eq? which 'server-context-count) server-contexts)
	      ((eq? which 'all-contexts) all-contexts)
	      ((eq? which 'threads-in-context)
	       (if (and (not (null? rest))
			(exact? (car rest))
			(integer? (car rest))
			(< -1 (car rest) server-contexts))
		   (vector-ref threads-in-context (car rest))
		   (sneezy.error "client.data: invalid context number: ~a"
				 (if (null? rest)
				     '()
				     (car rest)))))
	      (else
	       (sneezy.error "client.data: invalid selector: ~a" which))))))
      

  ; Close connections and shut down processes as much as possible.
  ; A little convoluted to avoid race condition between internally
  ; initiated cleanup and debugger-initiated cleanup.

  (define (cleanup)
    (let ((mutex mutex))
      (if mutex
	  (begin
	    (call-with-mutex mutex
	      (lambda ()
		(if server-connection
		    (begin (close-connection server-connection)
			   (set! server-connection #f)))))
	    (set! mutex #f)
	    #t))))


  ; Control the internal state of the client (a little). The peek method
  ; does not use the mutex to allow peeking to work in a hung state.
  ; It is OK to call client.control before the client has been initialized.

  (define (control item)
    (cond ((eq? item 'monitor)
	   (if mutex
	       (call-with-mutex mutex
		 (lambda ()
		   (set! monitoring #t)))
	       (set! monitoring #t)))
	  ((eq? item 'nomonitor)
	   (if mutex
	       (call-with-mutex mutex
                (lambda ()
		  (set! monitoring #f)))
	       (set! monitoring #f)))
	  ((eq? item 'peek)
	   `((connection  . ,server-connection)
	     (port-number . ,server-port-number)
	     (host-name   . ,server-host-name)
	     (monitoring  . ,monitoring)
	     (mutex       . ,mutex)))
	  (else
	   (sneezy.msg "client.control: options: monitor, nomonitor, peek."))))


  ; Continue the context/thread

  (define (cmd:continue cid tid)
    (send (list sneezy.cmd:continue cid tid '())))


  ; Terminate the context/thread

  (define (cmd:terminate cid tid)
    (send (list sneezy.cmd:terminate cid tid '())))


  ; Simple implementation of 'handle' which requires all arguments.

  (define (cmd:handle etype cid tid proc)
    (let ((etype (if (symbol? etype)
		     (sneezy.event-type etype)
		     etype)))
      (register-event etype cid tid proc)))


  ; Do not handle the event for the given etype/cid/tid combination.
  ; Must remove the handler from the server if the call to unhandle
  ; removes the last handler for the event.

  (define (cmd:unhandle etype cid tid)
    (let ((etype (if (symbol? etype)
		     (sneezy.event-type etype)
		     etype)))
      (unregister-event etype cid tid)
      #t))


  ; Read data from a global pointer.

  (define (cmd:read cid tid gptr length)
    (rpc-request sneezy.cmd:read cid tid (list (car gptr)
					       (cadr gptr)
					       length)))

  ; Write data to the address space.
  ; The data is a standard-encoding string.
  ;
  ; The length field is redundant but required for the time being;
  ; the protocol should be revised.

  (define (cmd:write cid tid gptr data)
    (let ((length (string->number (substring data 0 4) 16)))
      (send (list sneezy.cmd:write cid tid (list (car gptr)
						 (cadr gptr)
						 length
						 data)))))

  ; Call the user function.  fptr is a function pointer, and the arglist
  ; is an list of lists where the car of a sublist is a type tag for the
  ; argument type, and the cadr is the value.
  ;
  ; Type tags are integers.  Values are either integers, characters,
  ; or strings.  If characters, the type tag must be a character tag.
  ; If a string, the tag must be a character pointer.  Otherwise, the
  ; integer encodes whatever the tag indicates.

  (define (cmd:call-user-function cid tid fptr arglist)
    (let ((data '()))
      (rpc-request sneezy.cmd:call-user-function 
		   cid 
		   tid 
		   (cons fptr (cons (length arglist) arglist))
		   (lambda (datum)
		     (set! data (cons datum data))))
      (reverse data)))

  ; There is exactly one event loop running in every client. It reads an
  ; event from the server connection and runs all the installed handlers for
  ; the event. The handler may not block for an unbounded length of time,
  ; but it may validly be preempted.
  ; There is no guarantee on the order of the handlers' execution.

  (define (start-event-loop)
    (sneezy.msg "Starting client event loop.") (newline)
    (spawn event-loop))


  ; Runs single-threaded and does not need to use the mutex.

  (define (event-loop)
    (let ((e (receive)))
      (if (eof-object? e)
	  (begin (sneezy.msg "Event loop terminated.")
		 (cleanup))
	  (begin (process-event e)
		 (event-loop)))))


  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  ;
  ; Communication.
  ;

  ; Send a message to the server. Must lock, because there are many
  ; senders.

  (define (send msg)
    (call-with-mutex mutex
      (lambda ()
	(if monitoring (sneezy.msg "SEND: ~a" msg))
	(connection-send server-connection msg))))

  ; Receive a message from the server. Must not lock, since there is only
  ; one reader.

  (define (receive)
    (let ((m (connection-receive server-connection)))
      (if monitoring (sneezy.msg "RECV: ~a" m))
      m))


  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  ;
  ; Event registry and processing

  (define event-mutex  #f)           ; mutex protects the following vars
  (define event-lookup #f)
  (define event-table  #f)
  (define event-next   #f)

  (define (initialize-event-registry)
    (set! event-lookup (make-vector sneezy.event-count '()))
    (set! event-table (make-vector 10 '()))
    (set! event-next 0)
    (set! event-mutex (make-mutex)))


  ; Register-event: add a new event to the event table, and register it with
  ; the server. Return the ID for the event.

  (define (register-event etype cid tid proc)
    (let* ((e   (install-handler etype cid tid proc)))
      (send (list sneezy.cmd:install-handler cid tid (list etype e)))))

  (define (unregister-event etype cid tid)
    (remove-handler etype cid tid)
    (send (list sneezy.cmd:uninstall-handler cid tid etype)))

  ; Install a handler for the event named by etype. 

  (define (install-handler etype cid tid proc)
    (let ((eid (etype->eid etype cid tid)))
      (call-with-mutex event-mutex
	(lambda ()
	  (vector-set! event-table eid
		       (cons (cons etype proc)
			     (vector-ref event-table eid)))))
      (+ eid sneezy.eid-base)))

  ; Remove a handler for the event named by etype.

  (define (remove-handler etype cid tid)
    (let ((eid (etype->eid etype cid tid)))
      (call-with-mutex event-mutex
	(lambda ()
	  (vector-set! event-table eid
		       (remassoc etype (vector-ref event-table eid)))))))

  ; An eid is unique for a given combination of etype, cid, and tid.
  ; The array event-lookup is a vector of alists indexed by etype, each
  ; alist keyed by a pair of the cid and tid. 
  ;
  ; The table's maximal size is 
  ;    (* num-etypes num-contexts (avg num-tid-per-ctx))
  ; and perhaps we'd be better off simply allocating that?
  ;
  ; FIXME: could use cid-tid table for the etype map.

  (define (etype->eid etype cid tid)

    ; Allocate a new eid and place it in the correct tables.

    (define (new-eid etype)
      (let ((e event-next))
	(set! event-next (+ event-next 1))
	(if (>= e (vector-length event-table))
	    (set! event-table (extend-vector event-table (+ e 1) '())))
	(vector-set! event-lookup etype e)
	e))

    (call-with-mutex event-mutex
      (lambda ()
	(let* ((alist (vector-ref event-lookup etype))
	       (probe (assoc (cons cid tid) alist)))
	  (if probe
	      (cdr probe)
	      (let ((eid (new-eid etype)))
		(vector-set! event-lookup etype
			     (cons (cons (cons cid tid) eid) alist))
		eid))))))


  ; Process-event: Given a raw event structure, fix it up and dispatch on it.
  ; An event structure is a list (eid cid tid edata).

  (define (process-event e)
    (let ((eid (car e)))
      (if (< eid sneezy.eid-base)
	  (cond ((= eid sneezy.eid:ready)
		 (apply ready-handler e))
		((= eid sneezy.eid:reply)
		 (apply handle-reply e))
		((= eid sneezy.eid:push)
		 (apply handle-push e))
		(else
		 (error "Unknown predefined EID! ~a" eid)))
	  (let ((events (call-with-mutex event-mutex
			  (lambda ()
			     (vector-ref event-table
					 (- eid sneezy.eid-base))))))
	    (for-each (lambda (ev)
			((cdr ev) (car ev) (cadr e) (caddr e) (cadddr e)))
		      events)))))


  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  ;
  ; RPC requests.
  ;
  ; An RPC request is a command that has a reply. In order to associate
  ; a reply with a request, only one reply-giving request is outstanding 
  ; for any given context at any time. This can be generalized to having
  ; one outstanding request for any context/thread combination. This is
  ; still not as efficient as having several outstanding requests at the
  ; same time, but it is simple, and avoids having to use request tokens
  ; (which make the C++ code more complicated and should be avoided).

  ; BUG!! FIXME!! This will not process events while waiting for a 
  ; reply if called, from an event handler, but will deadlock!  It needs
  ; to set up a thread to handle more events while it is waiting for
  ; the reply.  The new thread should become the event loop.  The reply
  ; handler would be executed by the new thread.  It would unblock the
  ; original thread and kill itself, assuming that there is ever only one
  ; reply handler to be executed (not a list, like for event handlers).

  (define rpc-mutex   #f)
  (define rpc-queues  #f)
  (define reply-procs #f)
  (define push-procs #f)

  ;rpc-mutex:notheld
  (define (initialize-rpc)
    (set! rpc-mutex   (make-mutex))
    (set! rpc-queues  (make-vector server-contexts #f))
    (set! reply-procs (make-vector server-contexts #f))
    (set! push-procs  (make-vector server-contexts #f))
    (do ((i (- server-contexts 1) (- i 1)))
	((< i 0) #t)
      (vector-set! rpc-queues i (make-basic-queue)))
    #t)

  ; Initiate an rpc request and wait for the data to return.
  ; If rest is non-null, then it contains a data collector procedure.

  ;rpc-mutex:notheld
  (define (rpc-request cmd cid tid datum . rest)

    ;rpc-mutex:held
    (define (perform-request cmd cid tid datum proc pushproc)
      (vector-set! reply-procs cid proc)
      (vector-set! push-procs cid pushproc)
      (send (list cmd cid tid datum)))

    (let ((sem   (make-semaphore 0))
	  (data  #f))
      (initiate-rpc-request 
       cid
       (lambda ()
	 (perform-request cmd cid tid datum 
			  (lambda (d)
			    (set! data d)
			    (semaphore-up sem))
			  (if (null? rest)
			      (lambda (d)
				(error "Unhandled PUSH: " d))
			      (car rest)))))
      (semaphore-down sem)
      data))

  ; Initiate an RPC request. Proc is a procedure of one argument which
  ; will be called with the edata part of the message when the reply
  ; arrives.

  ;rpc-mutex:notheld
  (define (initiate-rpc-request cid thunk)
    (call-with-mutex rpc-mutex
      (lambda ()
	(let ((q (vector-ref rpc-queues cid)))
	  (if (qempty? q)
	      (thunk)
	      (qput q thunk))))))

  ;rpc-mutex:notheld
  (define (handle-reply eid cid tid edata)
    (call-with-mutex rpc-mutex
      (lambda ()		     
	((vector-ref reply-procs cid) edata)
	(vector-set! reply-procs cid #f)
	(vector-set! push-procs cid #f)
	(let ((q (vector-ref rpc-queues cid)))
	  (if (not (qempty? q))
	      ((qget q)))))))

  ;rpc-mutex:notheld
  (define (handle-push eid cid tid edata)
    ((vector-ref push-procs cid) edata))

)

(provide "client.sch")

; eof
