; Library of useful utility functions for Sneezy.
; lth@cs.uoregon.edu / June 26, 1995
;
; $Id: utils.sch,v 1.1 1996/05/01 20:11:15 lth Exp $
;
; Requires: hash tables.

(define (assert expr location)
  (if (not expr)
      (error "Failed assertion in ~a" location)))

; Given vector, extend it to accomodate 'count' elements, returning the 
; new vector.

(define (extend-vector v count init)
  (let ((u (make-vector count init)))
    (do ((i (- (vector-length v) 1) (- i 1)))
	((< i 0) u)
      (vector-set! u i (vector-ref v i)))))


; Extensible vector data type
;
; (make-extensible-vector length . init)  => evector
; (extensible-vector? obj)  => boolean
; (evector-length evector) => index    ; length (highest index+1)
; (evector-ref evector idx) => value
; (evector-set! evector idx value) => unspecified
;
; evector-ref is valid for any positive index, returning the default
; element if the element has not been set.

(define %evector-tag (list 'evector))

(define (make-extensible-vector initial-length . rest)
  (let ((default (if (null? rest) #f (car rest))))
    (vector %evector-tag
	    (make-vector initial-length default) 
	    default 
	    initial-length)))

(define (extensible-vector? obj)
  (and (vector? obj)
       (> (length obj) 0)
       (eq? (car obj) %evector-tag)))

(define (evector-length obj)
  (vector-ref obj 3))

(define (evector-ref obj idx)
  (if (>= idx (vector-length (vector-ref obj 1)))
      (vector-ref obj 2)
      (vector-ref (vector-ref obj 1) idx)))

(define (evector-set! obj idx val)
  (if (>= idx (vector-length (vector-ref obj 1)))
      (let ((newv (extend-vector (vector-ref obj 1)
				 (max (+ idx 1)
				      (* (vector-length (vector-ref obj 1)) 2))
				 (vector-ref obj 2))))
	(vector-set! obj 1 newv)))
  (vector-set! (vector-ref obj 1) idx val)
  (vector-set! obj 3 (max (vector-ref obj 3) (+ idx 1))))


; Given assoc list, create a hash table and initialize it with members
; of assoc list.

(define (alist->hash-table l)
  (let ((h (make-hash-table)))
    (for-each (lambda (i)
		(hash-table-put! h (car i) (cdr i)))
	      l)
    h))


; Given predicate and list, select elements of list satisfying predicate.
; (Wanted to call it just 'select' but STk uses that name.)

(define (select-if pred? list)
  (define (s l r)
    (cond ((null? l) (reverse! r))
	  ((pred? (car l))
	   (s (cdr l) (cons (car l) r)))
	  (else
	   (s (cdr l) r))))
  (s list '()))

; Given predicate and list, select first element that satisfies predicate,
; and return it.  If none match, return #f.

(define (select-first pred? l)
  (cond ((null? l) #f)
	((pred? (car l)) (car l))
	(else (select-first pred? (cdr l)))))

; Procedure takes element and index

(define (vector-for-each p v)
  (let ((limit (vector-length v)))
    (do ((i 0 (+ i 1)))
	((= i limit) #t)
      (p i (vector-ref v i)))))

; Return list of (0 ... n-1)

(define (iota n)
  (define (loop k)
    (if (= k n)
	'()
	(cons k (loop (+ k 1)))))
  (loop 0))


; Common Scheme procedures not present in STk 2.1.7

(define (list-copy l)
  (append l '()))

(define (remove p l)
  (cond ((null? l)          l)
	((equal? p (car l)) (remove p (cdr l)))
	(else               (cons (car l) (remove p (cdr l))))))

(define (remove-first! p l)
  (cond ((null? l) l)
	((equal? p (car l)) (cdr l))
	(else (set-cdr! l (remove-first! p (cdr l)))
	      l)))

(define (remassoc p l)
  (cond ((null? l)           l)
	((equal? p (caar l)) (remassoc p (cdr l)))
	(else                (cons (car l) (remassoc p (cdr l))))))
	 
(define (append! a b)
  (cond ((null? a)
	 b)
	(else (set-cdr! (last-pair a) b)
	      a)))

(define (last-pair n)
  (if (null? (cdr n))
      n
      (last-pair (cdr n))))

(define (reverse! l)
  (do ((prev '() curr)
       (curr l (let ((next (cdr curr)))
                 (set-cdr! curr prev)
                 next)))
      ((null? curr) prev)))

(define (but-last l)
  (if (null? l)
      (error "butlast: null list.")
      (reverse! (cdr (reverse l)))))

; Decode standard decoding of data.

(define (decode-data str)

  (define (xval c)
    (if (char-numeric? c)
	(- (char->integer c) (char->integer #\0))
	(- (char->integer (char-downcase c)) (char->integer #\a) -10)))

  (define (get-byte str idx)
    (+ (* 16 (xval (string-ref str (* 2 idx))))
       (xval (string-ref str (+ (* 2 idx) 1)))))


  (let* ((len (+ (* 256 (get-byte str 0))
		 (get-byte str 1)))
	 (s   (make-string len)))
    (do ((i 0 (+ i 1))
	 (j 2 (+ j 1)))
	((= i len) s)
      (string-set! s i (integer->char (get-byte str j))))))

; Encode data in standard encoding.

(define (encode-data str)
  (let* ((len (string-length str))
	 (s   (make-string (+ 4 (* len 2)))))

    (define digits "0123456789abcdef")

    (define (put-byte s i b)
      (string-set! s (* i 2) (string-ref digits (quotient b 16)))
      (string-set! s (+ (* i 2) 1) (string-ref digits (remainder b 16))))

    (put-byte s 0 (quotient len 256))
    (put-byte s 1 (remainder len 256))
    (do ((i 0 (+ i 1))
	 (j 2 (+ j 1)))
	((= i len) s)
      (put-byte s j (char->integer (string-ref str i))))))

(define (every? proc l1 . rest)

  (define (loop len lists)
    (cond ((zero? len) #t)
	  ((apply proc (map car lists)) (loop (- len 1) (map cdr lists)))
	  (else #f)))

  (loop (length l1) (cons l1 rest)))
  
(provide "utils.sch")

; eof
