c*******************  BUK.F *************************
c
c      This program generates a disordered sequence of N
c      integers using the random number generator given
c      by randlc (written by D. Bailey) and then uses a 
c      bucket sort to sort it.
c
c      Prog: Leo Dagum
c            NASA Ames
c      Date: Aug 1990.
c
c********************************************************
      program

      integer               N
      integer               MAXKEY      
      integer               m
      integer               numreps
      double precision      A
      double precision      S

      parameter            (N = 1048576/16)
      parameter            (MAXKEY = 32768/16)
      parameter            (m = MAXKEY/4)
      parameter            (numreps = 10)
      parameter            (A = 1220703125.D0, S = 314159265.D0)

      integer               key(0:N-1)
      integer               key2(0:N-1)
      integer               rank(0:N-1)
      integer               keyden(0:MAXKEY-1)
      integer               i,j,k
      double precision      x
      double precision      seed
      double precision      randlc
      real                  etime
      real                  tarray(2)
      real                  second
      real                  time1, time2
      
*
* Initialize seed for random number generator
*
      seed = S
*
* Initialize keys with a Gaussian distribution in key densities
*
      do  2 i=0, N-1
            x = randlc(seed, A)      
            x = x + randlc(seed, A)
            x = x + randlc(seed, A)
            x = x + randlc(seed, A)
            key(i) = m*x
 2      continue
*
* Timing should begin here.  The keys will be ranked
* numreps times, this is only to make sure that the
* elapsed time is large enough to measure with a stop
* watch (in case the system provided timer is significantly 
* off).
*
*      time1 = second()
      time1 = etime(tarray)
      do  5 i=1, numreps
            print *, 'Begin iteration ',i
            key(i) =  i
	    key(i+numreps) = MAXKEY - i
            call bucksort(key, rank, keyden, N, MAXKEY)
*
* For full scale benchmark, use 'partbig' for partial verification test.
* For scaled down benchmark, use 'partsmal' for partial verification test.
*
	    if (N .eq. 1048576*8 .and. MAXKEY .eq. 32768*16) then
            	call partbig(N, rank, i)
	    else if (N .eq. 1048576/16 .and. MAXKEY .eq. 32768/16) then
            	call partsmal(N, rank, i)
	    else
		print *,'No partial test for this size problem.'
	    endif
5      continue
*
* Timing should end here.
*
*      time2 = second()
      time2 = etime(tarray)
      print *,'Performed ',numreps,' rankings in time: ', time2-time1
*
* Move keys to their ranked positions, key2 now stores the sorted sequence
*
      do 20 i=0, N-1
            k       = rank(i)
            key2(k) = key(i)
 20       continue
*
*  Check that things got sorted ok (sorted here means non-descending order)
*
      j = 0
      do 30 i=0, N-2
            if (key2(i+1) .lt.  key2(i)) then
                  j = j + 1
            endif
 30      continue

      if (j .eq. 0) then
            print *, 'PASSED:   0 out of place.'
      else
            print *, 'FAILED:  ', j, 'out of place.'
      endif
      
      end
*******************************************************************
*
*       BUCKET_SORT
*      This subroutine actually only ranks the keys (i.e. it does
*      not move the keys to their sorted locations).  Ranking is
*      done using the bucket sorting algorithm.  
*
*******************************************************************
      subroutine bucksort(key, rank, keyden, N, MAXKEY)


      integer            N
      integer            MAXKEY      

      integer            key(0:N-1)
      integer            rank(0:N-1)
      integer            keyden(0:MAXKEY-1)
      integer            i, j, k
*
*  Zero the keyden array
*
      do 40 i=0, MAXKEY-1
            keyden(i) = 0
 40      continue
*
*  Count occurrences of each key (the 'key density')
*
      do 60 i=0, N-1
            k = key(i)
            keyden(k) = keyden(k) + 1
 60      continue
*
*  Create running sum (i.e. starting index) of keyden array
*
      do 80 i=1, MAXKEY-1
            keyden(i) = keyden(i) + keyden(i-1)
 80      continue
*
*  Compute rank for each key
*

      do  110 i=0, N-1
            k = key(i)
            keyden(k) = keyden(k) - 1
            rank(i)    = keyden(k) 
 110      continue

      return
      end
*******************************************************************
*       PARTLSMAL
*      This routine performs the partial verification test on the
*      ranking.  The numbers are correct for the scaled down benchmark
*      given here.  For the full scale benchmark apply the numbers
*      provided in the benchmark description (and given here in 'partbig').
*******************************************************************
      subroutine partsmal(N, rank, iter)

      integer            N
      integer            rank(0:N-1)
      integer            iter
      integer            i

      i = iter - 1
      if (rank(48427) .ne. (1+i)) then
           print *, 'FAILED partial verification test'
      elseif (rank(17148) .ne. (19+i)) then
           print *, 'FAILED partial verification test'
      elseif (rank(23627) .ne. (347+i)) then
           print *, 'FAILED partial verification test'
      elseif (rank(62548) .ne. (64916-i)) then
           print *, 'FAILED partial verification test'
      elseif (rank(4431) .ne. (65462-i)) then
           print *, 'FAILED partial verification test'
      else
           print *, 'PASSED partial verification test'
	   print *, 'Key Index= 48427, rank= ',rank(48427)
	   print *, 'Key Index= 17148, rank= ',rank(17148)
	   print *, 'Key Index= 23627, rank= ',rank(23627)
	   print *, 'Key Index= 62548, rank= ',rank(62548)
	   print *, 'Key Index= 4431,  rank= ',rank(4431)
      endif
      return
      end
*******************************************************************
*      PARTBIG
*      This routine performs the partial verification test on the
*      ranking.  The numbers are correct for the full scale benchmark
*      (i.e. N=2^23 and MAXKEY=2^19).
*******************************************************************
      subroutine partbig(N, rank, iter)

      integer            N
      integer            rank(0:N-1)
      integer            iter
      integer            i

      i = iter - 1
      if (rank(2112377) .ne. (104+i)) then
           print *, 'FAILED partial verification test'
      elseif (rank(662041) .ne. (17523+i)) then
           print *, 'FAILED partial verification test'
      elseif (rank(5336171) .ne. (123928+i)) then
           print *, 'FAILED partial verification test'
      elseif (rank(3642833) .ne. (8288932-i)) then
           print *, 'FAILED partial verification test'
      elseif (rank(4250760) .ne. (8388264-i)) then
           print *, 'FAILED partial verification test'
      else
           print *, 'PASSED partial verification test'
	   print *, 'Key Index= 2112377, rank= ',rank(2112377)
	   print *, 'Key Index= 662041, rank= ',rank(662041)
	   print *, 'Key Index= 5336171, rank= ',rank(5336171)
	   print *, 'Key Index= 3642833, rank= ',rank(3642833)
	   print *, 'Key Index= 4250760,  rank= ',rank(4250760)
      endif
      return
      end
******************************************************************************
C
      DOUBLE PRECISION FUNCTION RANDLC (X, A)
C
C   This routine returns a uniform pseudorandom double precision number in the
C   range (0, 1) by using the linear congruential generator
C
C   x_{k+1} = a x_k  (mod 2^46)
C
C   where 0 < x_k < 2^46 and 0 < a < 2^46.  This scheme generates 2^44 numbers
C   before repeating.  The argument A is the same as 'a' in the above formula,
C   and X is the same as x_0.  A and X must be odd double precision integers
C   in the range (1, 2^46).  The returned value RANDLC is normalized to be
C   between 0 and 1, i.e. RANDLC = 2^(-46) * x_1.  X is updated to contain
C   the new seed x_1, so that subsequent calls to RANDLC using the same
C   arguments will generate a continuous sequence.
C
C   This routine should produce the same results on any computer with at least
C   48 mantissa bits in double precision floating point data.  On 64 bit
C   systems, double precision should be disabled.
C
C   David H. Bailey     October 26, 1990
C
      IMPLICIT DOUBLE PRECISION (A-H, O-Z)
      SAVE KS, R23, R46, T23, T46
      DATA KS/0/
C
C   If this is the first call to RANDLC, compute R23 = 2 ^ -23, R46 = 2 ^ -46,
C   T23 = 2 ^ 23, and T46 = 2 ^ 46.  These are computed in loops, rather than
C   by merely using the ** operator, in order to insure that the results are
C   exact on all systems.  This code assumes that 0.5D0 is represented exactly.
C
      IF (KS .EQ. 0) THEN
        R23 = 1.D0
        R46 = 1.D0
        T23 = 1.D0
        T46 = 1.D0
C
        DO 100 I = 1, 23
          R23 = 0.5D0 * R23
          T23 = 2.D0 * T23
 100    CONTINUE
C
        DO 110 I = 1, 46
          R46 = 0.5D0 * R46
          T46 = 2.D0 * T46
 110    CONTINUE
C
        KS = 1
      ENDIF
C
C   Break A into two parts such that A = 2^23 * A1 + A2.
C
      T1 = R23 * A
      A1 = AINT (T1)
      A2 = A - T23 * A1
C
C   Break X into two parts such that X = 2^23 * X1 + X2, compute
C   Z = A1 * X2 + A2 * X1  (mod 2^23), and then
C   X = 2^23 * Z + A2 * X2  (mod 2^46).
C
      T1 = R23 * X
      X1 = AINT (T1)
      X2 = X - T23 * X1
      T1 = A1 * X2 + A2 * X1
      T2 = AINT (R23 * T1)
      Z = T1 - T23 * T2
      T3 = T23 * Z + A2 * X2
      T4 = AINT (R46 * T3)
      X = T3 - T46 * T4
      RANDLC = R46 * X
C
      RETURN
      END

