#!/usr/local/bin/perl
# *-*-perl-*-*

# dep2html, v1.0
#
#  author:  Juan Villacis
# project:  pC++/Sage++
# started:  04-29-95
# updated:  09-26-95
#
# This program converts a dep file (produced by the Sage++ compiler tools)
# into html by way of Bernd Mohr's cgm tool.  As a by-product, a litter of
# "htmlized" files are created within the given (sub)directories.
#
# CAVEAT:  this program relies on correctness of Bernd's cgm tool (which
#          in turn, relies on the correctness of the Sage++ compiler tools);
#          this really should parse the dep file itself ...
#

# ************************ system dependent stuff *****************************

# SYSDEPENDENTSTUFFGOESHERE

# *************************** some constants **********************************

$LINENUMBERS = 'true';          # if non-null, then show line numbers
$HLEVERY = 10;                  # highlight every nnn line number


# ************************** filename stuff ***********************************

sub basename { $_=shift(@_); (/^.+\/([^\/]+)$/) ? $1  : $_; }

sub dirname { $_=shift(@_) ; (/^(.+)\/[^\/]+$/) ? $1  : "."; }

sub depName { sprintf("%s.dep",shift(@_)); } # dep file format

sub cgmName { sprintf("%s.cgm",shift(@_)); } # Bernd's cgm file format

sub dbsName { sprintf("%s.dbs",shift(@_)); } # database format


# *********************** Top-level GROUP names *******************************

sub hsdp { sprintf("%s",shift(@_)); }   # HTML subdirectory prefix

# the groups listed below should form a global partition over all the 
# identifiers (IDs); however, IDs in the "ERROR" group are those that failed
# to get into the other groups; if the "ERROR" group is non-empty, the 
# criteria for membership in the other groups should be re-examined

# SYSLIB group:  any ID whose associated file (determined from the CGM
#                entry "ffile:  (..) FILE (...)") matches the SYSLIB_DIRS
#                specified above
$SYS_DIR = &hsdp($GRP_SYS="SYSLIB");

# INCLUDE group:  any ID whose associated file is 1) NOT in the SYSLIB group,
#                 and 2) the filename ends in a ".h"
$INC_DIR = &hsdp($GRP_INC="INCLUDE");

# SOURCE group:  any ID whose assocated file is NOT in the INCLUDE group
$SRC_DIR = &hsdp($GRP_SRC="SOURCE");

# EXTERNAL group:  any ID which 1) had no file associated with it, and
#                  2) had the CGM entry "fclass: - - -2" ;
$EXT_DIR = &hsdp($GRP_EXT="EXTERNAL");

# ERROR group:  any ID which failed to fall within one of the above groups
$ERR_DIR = &hsdp($GRP_ERR="ERROR");


# *********************** Low-level GROUP names *******************************

# the groups listed below form a different partition over the space of IDs;

$GRP_CLS = "CLASS";

$GRP_COL = "COLLECTION";

$GRP_FUN = "FUNCTION";  # in C, all functions which are neither a class nor
                        # a collection nor a method; in Fortran, normal

$GRP_SUB = "SUBROUTINE";  # in Fortran, anything that is not a main program
                          # nor a function


# ************************** I/O related routines *****************************

sub printV {
    print @_ if ($VERBOSE);
}

sub printD {
    print @_ if ($DEBUG);
}

sub ss {
    &printD("System:  ",@_,"\n");
    system(@_);
}

sub clearDir {
    local($dir)=sprintf("%s/%s",$baseDir,shift(@_));
    
    (-e $dir) ? &ss("$RM -f $dir/*.html") : &ss("$MKDIR $dir");
    $dir;
}


# *************************** sort routines ***********************************

sub numerically { $a <=> $b ; } # for sorting numerically

sub byPos { $ID_START{$a} <=> $ID_START{$b}; }

sub byname {
    local($ia,$ib)=($ID_NAME{$a},$ID_NAME{$b});

    local($la,$lb)=($ia,$ib);
    $la=~y/A-Z/a-z/; $lb=~y/A-Z/a-z/;
    $la cmp $lb;
}

sub byArgName {       # we assume all these name+args are to be DISTINCT!
    local($ia)=sprintf("%s%s",$ID_NAME{$a},$ID_ARGLIST{$a});
    local($ib)=sprintf("%s%s",$ID_NAME{$b},$ID_ARGLIST{$b});

    local($la,$lb)=($ia,$ib);
    $la=~y/A-Z/a-z/; $lb=~y/A-Z/a-z/;
    $la cmp $lb;
}


# ************************* non-generic routines ******************************

sub getPos {
    $pos=shift(@_);
    $pos=~/\s*(\d+)\.(\d+)\s*(\d+)\.(\d+)/;
    ($1,$2,$3,$4);
}

sub resolveInclude {            # INTENT: if an include directive is found
                                # (e.g., #INCLUDE <foo.h>), attempt to make
                                # a hyperlink to it
    local($foo)=shift(@_);      # NOTE: cannot do the job of the linker or
                                #       compiler ... will leave that up to
    return $foo;                #       those programs
}

# ****************************** read CGM file ********************************

sub readCGM {

    sub parseFCFILE {
        local($id,$start,$f,$pos)=@_;

        return unless ($id);
        # determine if there is some file associated with $id ...
        do {
            $ID_START{$id}=$start;
            die "ID=$id, pos=$pos \n" unless
                ($pos=~/(\d+\.\d+)\s+(\d+\.\d+)\s+(\d+\.\d+)\s+(\d+\.\d+)/);
            $ID_POS_NAME{$id} = "$1 $2";
            $ID_POS_BODY{$id} = "$3 $4";
            # add $id to FILE_IDs($f) only if it is not already in there
            $FILE_IDs{$f} .= " $id " unless ($FILE_IDs{$f}=~/\s+$id\s+/);
            # determine which GROUP_{SYSLIB,INCLUDE,SOURCE} $f belongs to ...
            local($d,$b) = ($f=~/^(.+)\/([^\/]+)$/) ? ($1,$2) : ("",$f);
            foreach $dir (@SYSLIB_DIRS) {
                ($d=~m/$dir/) &&
                    ( $GROUP_SYSLIB{$f} = " ", last) ;
            }
            do {
                ($b=~/\.h$/i) ? ($GROUP_INCLUDE{$f} = " ")
                    : ($GROUP_SOURCE{$f} .= " ")
            } unless ($GROUP_SYSLIB{$f});
        } unless ($start == 0);
    }

    sub parseFPOS {
        local($fPOS)=shift(@_);
        local(@fIDs)=@_;
        
        while ($fPOS=~/\{([^}]+)\}/) {
            $fPOS="$'"; $fID=shift(@fIDs); $fp=$1;
            $ID_FCALLS{$ID} .= sprintf("%s %s",$fID,$fp) unless ($fp=~/- -/); 
        }
    }

    sub parseCMPOS {
        local($cmPOS)=shift(@_);
        
        while ($cmPOS=~/\{\s+(\d+)\s+(\d+\.\d+\s+\d+\.\d+)\s+\}/) {
            $cmPOS = "$'";
            $ID_FCALLS{$ID} .= sprintf("%s %s ",$1,$2);
        }
    }

    sub parseFCLASS {
        local($coll,$cid,$id)=@_;

        return unless (($cid) && ($id));

	   if ($coll eq "COLL") {
		  do {
			 $GROUP_COLLECTION{$cid} .= " $id ";
			 $MEMBER_OF{$id}=$cid;
		  } unless ($GROUP_COLLECTION{$cid}=~/\s$id\s/);
	   } else {
		  do {
			 $GROUP_CLASS{$cid} .= " $id ";
			 $MEMBER_OF{$id}=$cid;
		  } unless ($GROUP_CLASS{$cid}=~/\s$id\s/);
	   }
    }

    # commence reading the CGM file ..

    print "Reading CGM file \`",&basename($cgmFile),"\' "; &printV("($cgmFile) "); print "...";

    open(CGM,"$cgmFile");

    while (<CGM>) {
      CASE: {
      # function block
          /s*ftag:\s*(\d+)\s+(\S+)/   &&
              ( $ID=$1,
               $ID_NAME{$ID} = $2,
               last CASE );
          /\s*ffile:\s*(\d+)\s+([^@]+)\s+\@\s+(\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+)/  && 
              ( &parseFCFILE($ID,$1,$2,$3),
          last CASE );
          /\s*ffile:\s*0\s+\-\s*/  && 
              ( $POSSIBLE_GROUP_EXTERNAL{$ID} = ' ', # MAY BE externally defined ... (must check fclass below before making final decision!)
               last CASE );
          /\s*fcalls:\s*(\d+)\s+\-\s*(.+)/ &&
              ( $fIDs = "$2",
               last CASE );
          /\s*fpos:\s*(.+)/  && 
              ( &parseFPOS($1,split(/\s/,$fIDs)),
               last CASE );
          /\s*fclass:\s*\-\s+\-\s+\-2\s*/  && ($POSSIBLE_GROUP_EXTERNAL{$ID}) &&
              ( $GROUP_EXTERNAL{$ID} = ' ', # yep, it is externally defined
               last CASE );
          /\s*fclass:\s*(\w+)\s+(COLL|-)\s+(\d+)/ && 
              ( &parseFCLASS($2,$3,$ID),
               last CASE );
          /\s*ftype:\s*(\S+)\s+(\S+)\s+(\S+)/i &&
               ( $ID_TYPE{$ID}="$1 $2 $3",
                $ID="",
                last CASE );
      # class/collection block
          /\s*ctag:\s*(\d+)\s+(\w+)\s*(-|COLL)?/   &&
              ( $ID=$1, $name=$2,
               $ID_NAME{$ID} = $name,
               ($3 eq "COLL") ? $GROUP_COLLECTION{$ID} .=' ' :
               $GROUP_CLASS{$ID} .= ' ',
               last CASE );
          /\s*cfile:\s*(\d+)\s+([^@]+)\s+\@\s+(\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+)/  &&
              ( &parseFCFILE($ID,$1,$2,$3),
               last CASE );
          /\s*cbase:\s*\{\s*(\d+)\s+(\d+\.\d+\s+\d+\.\d+)\s*\}/  &&
              ( ($2) && ($ID_FCALLS{$ID} = "$1 $2 "),
               $MEMBER_OF{$ID}=$1,
               last CASE );
          /\s*cpos:\s*(.+)/   &&     # not used
              ( last CASE );
          /\s*cmpos:\s*(.+)/  &&
              ( &parseCMPOS($1), 
               $ID="",
               last CASE );
#         print "No match: <$_>\n" if ($_);
      }
    }
    close(CGM);

    # update the file GROUPS to contain the IDs (no repetitions) ...

    foreach $key (keys %GROUP_SYSLIB) {
        $GROUP_SYSLIB{$key} = $FILE_IDs{$key};
    }
    
    foreach $key (keys %GROUP_INCLUDE) {
        $GROUP_INCLUDE{$key} = $FILE_IDs{$key};
    }
    
    foreach $key (keys %GROUP_SOURCE) {
        $GROUP_SOURCE{$key} = $FILE_IDs{$key};
    }

    # add to the EXTERNAL group if an ID has no body associated with it ...

    foreach $key (keys %ID_NAME) {
        $GROUP_EXTERNAL{$key} = ' ' unless ($ID_POS_BODY{$key});
    }

    print " done.\n";
}

# **************************** database routines ******************************

sub printDB {

    print "Creating database \`",&basename($dbaseName),"\' "; &printV("($dbaseName) ");

    open(DBASE,"> $dbaseName");

    print DBASE "# Database for \`$project\' (used by csearch, edoc, mccp) \n";
    print DBASE "# \n";
    print DBASE "# created:  ",`/bin/date`;
    print DBASE "#    user:  ",$ENV{"USER"},"\n";
    print DBASE "#  author:  $0\n";
    print DBASE "# \n\n";

    print DBASE "# Note:  the following should be set so that user www (or\n";
    print DBASE "#        whoever is executing mccp at your web site) has\n";
    print DBASE "#        access to this directory!\n";
    print DBASE "\$BASE_DIR = \"$rootDir\";\n\n"; # fully-qualified baseDir ...

    print DBASE "%ID_NAME = ( \n";
    foreach $key (sort numerically (keys %ID_NAME)) {
        print DBASE sprintf(" \'\%s\',  \'%s\',\n",$key,$ID_NAME{$key});
    }
    print DBASE " ); \n\n"; print ".";
    
    print DBASE "%GROUP_$GRP_SRC = ( \n";
    foreach $key (keys %GROUP_SOURCE) {
        print DBASE sprintf(" \'\%s\',  \'%s\',\n",$key,$GROUP_SOURCE{$key});
    }
    print DBASE " ); \n\n"; print ".";

    print DBASE "%GROUP_$GRP_INC = ( \n";
    foreach $key (keys %GROUP_INCLUDE) {
        print DBASE sprintf(" \'\%s\',  \'%s\',\n",$key,$GROUP_INCLUDE{$key});
    }
    print DBASE " ); \n\n"; print ".";
    
    print DBASE "%GROUP_$GRP_SYS = ( \n";
    foreach $key (keys %GROUP_SYSLIB) {
        print DBASE sprintf(" \'\%s\',  \'%s\',\n",$key,$GROUP_SYSLIB{$key});
    }
    print DBASE " ); \n\n"; print ".";

    print DBASE "%GROUP_$GRP_EXT = ( \n";
    foreach $key (keys %GROUP_EXTERNAL) {
        print DBASE sprintf(" \'\%s\',  \'%s\',\n",$key,$GROUP_EXTERNAL{$key});
    }
    print DBASE " ); \n\n"; print ".";

    close(DBASE);

    print " done.\n";
}

sub updateDB {

    &printV("Updating $project database ($dbaseName) ... ");

    open(DBASE,">> $dbaseName");

    print DBASE "\n\$LANGUAGE=\"",(($FORTRAN) ? "Fortran" : "C/C++"), "\";\n\n";

    if ($FORTRAN) {

        print DBASE "%GROUP_MAINPROG = ( \n";
        foreach $key (keys %GROUP_MAINPROG) {
            print DBASE sprintf("  \'\%s\',  \'%s\',\n",$key,$GROUP_MAINPROG{$key});
        }
        print DBASE " ); \n\n";

        print DBASE "%GROUP_$GRP_SUB = ( \n";
        foreach $key (keys %GROUP_SUBROUTINE) {
            print DBASE sprintf("  \'\%s\',  \'%s\',\n",$key,$GROUP_SUBROUTINE{$key});
        }
        print DBASE " ); \n\n";

        print DBASE "%GROUP_$GRP_FUN = ( \n";
        foreach $key (keys %GROUP_FUNCTION) {
            print DBASE sprintf("  \'\%s\',  \'%s\',\n",$key,$GROUP_FUNCTION{$key});
        }
        print DBASE " ); \n\n";

    } else {

    # "nflist" is used to make determining who belongs to the FUNCTION group
    # a little easier ...
        print DBASE "%GROUP_$GRP_CLS = ( \n";
        foreach $key (keys %GROUP_CLASS) {
            $nflist .= " $key $GROUP_CLASS{$key} ";
            print DBASE sprintf("  \'\%s\',  \'%s\',\n",$key,$GROUP_CLASS{$key});
        }
        print DBASE " ); \n\n";
    
        print DBASE "%GROUP_$GRP_COL = ( \n";
        foreach $key (keys %GROUP_COLLECTION) {
            $nflist .= " $key $GROUP_COLLECTION{$key} ";
            print DBASE sprintf("  \'\%s\',  \'%s\',\n",$key,$GROUP_COLLECTION{$key});
        }
        print DBASE " ); \n\n";

        print DBASE "%GROUP_$GRP_FUN = ( \n";
        foreach $key (keys %ID_NAME) {
            ($nflist=~/\s$key\s/) ||
                print DBASE sprintf("   \'\%s\',  \' \',\n",$key);
        }
        print DBASE " ); \n\n";
    }
        
    print DBASE "%ID_ARGLIST = ( \n";
    foreach $key (keys %ID_ARGLIST) {
        print DBASE sprintf(" \'\%s\',  \'%s\',\n",$key,$ID_ARGLIST{$key});
    }
    print DBASE " ); \n\n1;\n\n";

    close(DBASE);

    &printV("done.\n");
}

# ***************************** htmlize code **********************************

sub htmlizeCode {

    sub htmlize {
        local($subdir)=shift(@_);
        local($subtitle)=shift(@_);
        local(*GROUP)=shift(@_);

        # define some variables local to this procedure ...
        local($initFH,$file);
        local(@FHs)=();

        sub ahref {
            local($b,$x,$n,$a)=@_;

            $x = ($x=~/\.\.\//) ? $x : "../$subdir/$x";
            sprintf("%s<A HREF=\"./%s.html\">%s</A>%s",$b,$x,$n,$a);
        }

        sub cac {          # check and correct (only do this when you KNOW
            local($e)=@_;       # the @_ is not <A HREF=..>, etc.

            $e=~s/\</\&lt\;/g; # some things which don't translate
            $e=~s/\>/\&gt\;/g; # well during conversion to HTML
                               # NOTE: will need to change before anno!!!
            return ($e);
        }

        sub dirID {        # return the correct subdir where cid lives
            local($cid)=@_;

            sub check {
                local(*GROUP)=shift(@_);
                local($nm) = shift(@_);

                local($d)="";
                foreach $f (keys %GROUP) {
                    last if (($GROUP{$f} =~ /\s+($cid)\s+/) && ($d="$nm"));
                }
                $d;
            }

            local($ts) = ($GROUP_EXTERNAL{$cid} && $EXT_DIR);

            return sprintf("../%s/index.html#%s",$ts,$cid) if ($ts);

            $ts = (&check(*GROUP_SYSLIB,$SYS_DIR)) ||
                  (&check(*GROUP_INCLUDE,$INC_DIR)) ||
                  (&check(*GROUP_SOURCE,$SRC_DIR)) ;

            if (! $ts) {
# should never happen if all top-level groups cover the space of IDs ...
                $GROUP_ERROR{$cid} ="<A NAME=\"$cid.html\"></A><TT>ID=<B>$cid</B>, file=$file, line=$ln0, col=$offset : <P> <DL><B><A HREF=\"../$subdir/$BID.html\">$line</A></B></DL><P></TT>" unless ($GROUP_ERROR{$cid});
                $ts="../$ERR_DIR/index.html#$cid";
            } elsif ($subdir eq $ts) {
                $ts=sprintf("%s",$cid);
            } else {
                $ts=sprintf("../%s/%s",$ts,$cid);
            }
        }


        sub printNR {
            local($IOO,$h1,$id,$h2)=@_;
            print $IOO &ahref($h1,&dirID($id),($ID_NAME{$id} || "ID=$id"),$h2);
		  if ($ID_ARGLIST{$id}) {
			 print $IOO "<TT>",&cac($ID_ARGLIST{$id}),"</TT>\n";
		  } else {
			 print $IOO "<BR>\n";
		  }
        }


        sub printSub {

            foreach $fH (@FHs) {
                print $fH @_;
            }
        }

        sub printHTML {

            print HTML @_;
        }

        sub printAll {

            &printHTML(@_);
            &printSub(@_);
        }

        # regarding linenumbers (LNs) : we want LNs to show up in the main
        # html document, but not in the little extracted subfiles, and not
        # in the function argument lists (which could be several lines long); 
        # so a few hacks are required, particularly for href ...

        sub addLN {           # generate a line number format for 'n'
            local($n)=shift(@_);

            local($PB,$PE) = (($n % $HLEVERY)==0) ? ("<B>","</B>") : ('','');
            sprintf("$PB%${min_width}.${max_width}d%s$PE: ",$n);
        }

        sub removeLN {        # remove generated line numbers from 'x'
            local($x)=@_;

            ($LINENUMBERS) && ($x=~s/\n(\<[^\>]+\>)?\d+(\<[^\>]+\>)?\:\s/\n/g);
            $x;
        }

        sub removeFC {        # remove Fortran Continuation lines from 'x'
            local($x)=@_;

		  $x =~ s/\s\s\s\s\s\s\S(\s)/$1/g;
            &removeLN($x);
        }

        sub skip {    # NOTE:  Assuming $foL <= $ln0 <= $toL
            local($foL,$foC,$toL,$toC)=@_;

            # return if we've gone past $toL or we're given bad input ...
            do {
                return;
            } if ( ($ln0>$toL) || ($foL>$toL) );

            # case:  $foL==$toL
            do {
			 &printAll(&cac(sprintf("%s",substr($line,$foC,$toC-$foC))))
				if ( ($foC>=0) && ($toC>=$foC) );
                $offset=$toC;
                return;
            } if ($foL==$toL);

            # case:  $foL<$toL
            &printAll(&cac(sprintf("%s",substr($line,$foC))));

            # skip up to (but NOT including) $toL ...
            while ($ln0<$toL-1) {
                $line=<FOO>; $ln0++;
                &printHTML(&addLN($ln0)) if ($LINENUMBERS);
                &printAll(&cac($line));
            }

            # print only $toC columns of line number $toL ...
            $line=<FOO>; $ln0++; 
            &printHTML(&addLN($ln0)) if ($LINENUMBERS);
            &printAll(&cac(substr($line,0,$toC))) if ($toC>0);

            # at this point, $ln0==$toL, so set offset ...
            $offset=$toC;
        }

        sub href {
            local($id,$nbl,$nbc,$nel,$nec,$showLink)=@_;

            #parse the first line(s) : before,match,after
            $before = substr($line,$offset,$nbc-$offset);
            if ($nbl==$nel) {   # id definition lies on one line
                $match = substr($line,$nbc,$nec-$nbc);
            } else {
                $match = substr($line,$nbc);
                while (($line=<FOO>) && ($ln0++<$nel)) {
               	$match .= &addLN($ln0) if ($LINENUMBERS);
				$match .= $line;
			 }
			 $match .= &addLN($ln0) if ($LINENUMBERS);
			 $match .= substr($line,0,$nec);
		  }

		  $after = "";
		  $mname=$ID_NAME{$id};  # in order to use $mname in search patterns,
		  $mname=~s/(\W)/\\\1/g;	#   we must quote any metacharacters therein

# check if we're dealing with Fortran programs ..
		  if ($FORTRAN) {
				# check for fortran keyword defs ...
			 if ($match=~/(\s*)(program|subroutine|function)(\s*)($mname)(\s*)(\([^\)]*\))?/i) {
				$FORTRAN_TYPE="$2";
				$before .= &cac("$`$1$2$3");
				$match = &cac("$4");
				# note:  only the stuff up to and including the arglist
				#        needs to be checked ...
				$after = "$'";
				$ID_ARGLIST{$id} .= &removeFC("$6");
				$after = sprintf("%s%s","$5$6",$after);

				if ($FORTRAN_TYPE=~/subroutine/i) {
				    $GROUP_SUBROUTINE{$id} = ' ';
				} elsif ($FORTRAN_TYPE=~/function/i) {
				    $GROUP_FUNCTION{$id} = ' ';
				} else {
				    $GROUP_MAINPROG{$id} = ' ';
				}

				# otherwise, punt ...
			 } else {
				$after ="";
				$ID_ARGLIST{$id} .= &removeFC("$2") if ($match=~/($mname)\s*(\([^\)]*\))?/);
			 }
# else we're dealing with C programs ..
		  } else {
			 # check for C++ style Collection defs ...
			 if ($match=~/(\s*Collection\s+)($mname)(\s*:\s*)(public\s+)?/) {
				$before .= &cac(sprintf("%s%s",$`,$1));
				$match = &cac($2);
				$after = &cac("$3$4"); # don't include $' as CGM does

				# check for extended class defs ...
			 } elsif ($match=~/(\s*class\s+)($mname)(\s*:\s*public\s*)/) {
				$before .= &cac("$`$1");
				$match = &cac($2);
				$after = "$3";  # don't include $' cuz CGM gets that later

				# check for extended C++ style class method defs ...
			 } elsif ($match=~/([^\s]+)(\s*::\s*)($mname)(\s*)(\([^\)]*\))?(\s*:\s*)([^\(]+)/) {
				$moid = $MEMBER_OF{$id};
				$before .= &ahref(&cac($`),&dirID($moid),$1,$2);
				$match = &cac($3);
				$arglist=&cac($5);
				$ID_ARGLIST{$id} .= &removeLN($arglist);
				# note:  only the stuff up to and including the arglist
				#        needs to be checked ...
				$after = "$'";
				$after = ($MEMBER_OF{$moid}) ? &ahref(&cac("$4$arglist$6"),&dirID($MEMBER_OF{$moid}),&cac($7),$after) : sprintf("%s%s",&cac("$4$arglist$6$7"),$after);

				# check for normal C++ style class method defs ...
			 } elsif ($match=~/([^\s]+)(\s*::\s*)($mname)(\s*)(\([^\)]*\))?/) {
				$before .= &ahref(&cac($`),&dirID($MEMBER_OF{$id}),&cac($1),&cac($2));
				$match = &cac($3);
				$arglist=&cac($5);
				$ID_ARGLIST{$id} .= &removeLN($arglist);
				# note:  only the stuff up to and including the arglist
				#        needs to be checked ...
				$after = "$'";
				$after = sprintf("%s%s","$4$arglist",$after);

				# try to only match the id name, not the surrounding cruft
			 } elsif ($match=~/(\s*)($mname)(\s*)(\([^\)]*\))?/) {
				$before .= &cac("$`$1");
				$match = &cac($2);
				# note:  only the stuff up to and including the arglist
				#        needs to be checked ...
				$after = "$'";
				$arglist=&cac($4);
				$ID_ARGLIST{$id} .= &removeLN($arglist);
				$after = sprintf("%s%s","$3$arglist",$after);

				# otherwise, punt ...
			 } else {
				$after ="";
				$ID_ARGLIST{$id} .= &removeLN("$2") if ($match=~/($mname)\s*(\([^\)]*\))?/);
			 }
		  }

		  # reset the offset ...
		  $offset = $nec;

		  # if no match was found, "hilight" the stuff found AFTER
		  do {
			 $match="<BLINK>$after</BLINK>"; $after="";
		  } unless ($match);

		  # print out what was constructed thus far ...
		  &printHTML(&ahref($before,&dirID($id),$match,$after));
		  # remove LNs from strings before printing to extracted files ...
		  $before = &removeLN($before);
		  $match = &removeLN($match);
		  $after = &removeLN($after);
		  ($showLink) ?
			 &printSub(&ahref($before,&dirID($id),$match,$after)) :
				&printSub(sprintf("%s<B>%s</B>%s",$before,$match,$after) );

	   }

	   sub extract {   # extract an ID's definition into an external file
		  local($id,$nbl,$nbc,$nel,$nec,$bbl,$bbc,$bel,$bec,$fh,$o)=@_;

		  print ".";		  # show that this is "thinking"

		  # skip down to the beginning of the id definition ...
		  &skip($ln0,$offset,$nbl,$o);

		  # construct file to hold id's definition ...
		  $fh++; push(@FHs,$fh);
		  open($fh,"> $bs/$id.html");

		  # print out the header for this file ...
		  $hhhsss=join(' ',@Hs);
		  $hhhsss=~s/\.html\"\>/\.html\#$id\"\>/;
		  print $fh "<TITLE>$project : $hhhsss</TITLE>\n";
		  print $fh "<PRE><H3>", $hhhsss , "</H3><HR>";

		  # put a marker here ...
		  &printHTML("<A NAME=\"$id\"></A>");

		  # htmlize the id name ...
		  &href($id,$nbl,$nbc,$nel,$nec,'');
		  
		  # skip down to the start of ID body ...
		  &skip($nel,$nec,$bbl,$bbc);
		  
		  local($rID,$rPos,$rbl,$rbc,$rel,$rec,$dbl,$dbc,$del,$dec);
		  local($mID,$ebl,$ebc,$eel,$eec);

		  # now examine each ordered reference/definition within ID body ...
		  while ($ID_FCALLS{$id}=~/^(\d+)\s+(\d+\.\d+\s+\d+\.\d+)\s*/) {
			 
			 $rID = "$1"; $rPos = "$2"; $ID_FCALLS{$id} = "$'";

			 # determine whether rID is being referenced or defined ...
			 # NOTE: rID is being defined IF it is in FILE_IDs{$file}
			 #       AND the starting line where it is referenced
			 #       matches the starting column where it is defined

			 # rID's reference position
			 ($rbl,$rbc,$rel,$rec) = &getPos($rPos);
			 # rID's definition position
			 ($dbl,$dbc,$del,$dec) = &getPos($ID_POS_NAME{$rID});
			 
			 # skip to next referenced ID if already seen rID ...
			 next if ($ln0>$rel);

			 if ( ($FILE_IDs{$file}=~/\s$rID\s/) && ($rbl==$dbl) ) {
				# a definition ...
				($ebl,$ebc,$eel,$eec) = &getPos($ID_POS_BODY{$rID});
				&printD("<$ID_NAME{$rID}> ");
				# skip down to the start of rID's definition ...
				&skip($ln0,$offset,$dbl,0);
				$BID=$mID=$MEMBER_OF{$rID};
				push(@Hs,( ($mID) ? &ahref("<FONT SIZE=3>, member of</FONT> ",&dirID($mID),"<TT>" . ($ID_NAME{$mID} || $mID) . "</TT>"," ") : " (<TT><BLINK>Note:</BLINK> <EM>unable to determine membership</EM></TT>)") );
				&extract($rID,$dbl,$dbc,$del,$dec,$ebl,$ebc,$eel,$eec,$fh,$dbc);
				pop(@Hs);
				$fh=pop(@FHs);
				print $fh "</PRE>";
				close($fh);
			 } else {	  # a reference
				# skip down to the start of rID's reference ...
				&skip($ln0,$offset,$rbl,$rbc);
				# htmlize the reference ...
				&href($rID,$rbl,$rbc,$rel,$rec,'TRUE');
			 }
		  }
		  
		  # skip through the remainder of id body (to the end) ...
		  &skip($ln0,$offset,$bel,$bec);
	   }

        # ************************** htmlize **********************************

        # create/clean-out the proper subdirectories
        $bs = &clearDir($subdir);

        # go through each file in GROUP ...
        foreach $file (keys %GROUP) {
		  # determine the language of the source file based on its filename
		  # extension (currently we just check if the file is "Fortran";
		  # otherwise, assume it is "C" (non-Fortran) )
		  $FORTRAN=($file=~/\.$FORTRAN_EXTENSIONS$/);
		  &printD("Htmlizing $subdir file \`$file\' ... \n");
		  print ".";		  # show that this is "thinking"
		  open(FOO,"$file");
		  do {
			 # some code to make the line number format nice ...
			 $min_width=$max_width=0;
			 while (<FOO>) { $min_width++; }
			 close(FOO);
			 $min_width = ($min_width>0) ? $min_width : 1 ;
			 $min_width = int(log($min_width)/log(10)) + 1;
			 $min_width = ($min_width<1) ? 2 : $min_width ;
			 $max_width = $min_width;			 
			 open(FOO,"$file");
		  } if ($LINENUMBERS);
		  # insure that the htmlized file is UNIQUE (since basename isn't)
		  local($bfile)=&basename($file);
		  local($bsfile)=sprintf("%s/%s",$bs,$bfile);
		  local($b)=1;
		  if (-e "$bsfile.html") {
			 while (-e "$bsfile-$b.html") {
				$b++;
			 }
			 $bsfile .= "-$b";
		  }
		  open(HTML,"> $bsfile.html");
		  &printHTML("<TITLE>$project : $subtitle : $bfile</TITLE>\n");
		  &printHTML("<PRE><H3><A HREF=\"index.html\">$subtitle</A>: <I>$bfile</I></H3><HR>");

		  # initialize some stuff ...
		  $line = '';		  # current line within $file
		  $ln0=0;			  # current line number
		  $offset=0;		  # current position within $line
		  $initFH='fh0000';	  # starting file handle name
		  local($nbl,$nbc,$nel,$nec,$bbl,$bbc,$bel,$bec);
		  push(@Hs,"Defined in: <I><A HREF=\"$bfile.html\">$bfile</A></I>");

		  @sortedIDs = (sort byPos (split(/\s+/,$FILE_IDs{$file})));

		  # the following is a hack to catch nested definitions ...
		  # if definition B lies within definition A, then B is put
		  # inside of A's (sorted) function call list

		  @fileIDs=();

		  # grab the first non-empty ID def
		  while ($firstID = shift(@sortedIDs)) {
			 last if ($firstID);
		  }
		  push(@fileIDs,$firstID);

		  # run through the remaining list of IDs, checking to see
		  # whether any of them lie within another (previous) ID's body
		  while ($ID=shift(@sortedIDs)) {
			 push(@fileIDs,$ID);
			 $oldID=$firstID;
			 ($bbl,$bbc,$bel,$bec)=&getPos($ID_POS_BODY{$firstID});
			 if ($bel >= $ID_START{$ID}) {
				($zbl,$zbc,$zel,$zec)=&getPos($ID_POS_NAME{$ID});
				$fcalls = $ID_FCALLS{$firstID};
				$ID_FCALLS{$firstID} = "";
				$foundItAlready = "";
				while ($fcalls=~/^(\d+)\s+(\d+\.\d+\s+\d+\.\d+)\s*/) {
				    $rID = "$1"; $rPos = "$2"; $fcalls = "$'";
				    ($nbl,$nbc,$nel,$nec)=&getPos($rPos);
				    if ( ($rID eq $ID) && ("$zbl" eq "$nbl") ) {
					   # NOTE: cgm doesn't do column def's correctly
					   #       all the time, so cannot rely on zbc=nbc !
					   $foundItAlready = "yes";
					   $ID_FCALLS{$firstID} .= "$rID $rPos $fcalls";
					   last;
				    } elsif ($nbl >= $ID_START{$ID}) {
					   $ID_FCALLS{$firstID} .= "$ID $ID_POS_NAME{$ID} " ;
					   $foundItAlready = "yes";
				    }
				    $ID_FCALLS{$firstID} .= "$rID $rPos ";
				}
				$ID_FCALLS{$firstID} .= "$ID $ID_POS_NAME{$ID} "
				    unless ($foundItAlready);
			 } else {
				$firstID = $ID;
			 }
		  }

		  # @fileIDs contains the list of IDs which are defined in $file ...
		  foreach $ID (@fileIDs) {
			 next unless ($ID);
			 &printD("[$ID_NAME{$ID}] ");

			 # get ID's position info ...
			 ($nbl,$nbc,$nel,$nec)=&getPos($ID_POS_NAME{$ID});
			 ($bbl,$bbc,$bel,$bec)=&getPos($ID_POS_BODY{$ID});

			 # check if we have already passed this ID (e.g., this ID's
			 # definition could have occured in a previous ID's block) ...
			 next if ($ln0>$bel);

			 # extract the ID from $file, and put it into an external
			 # file; note that each call to &extract pushes a file handle
			 # onto a stack of file handles (@FHs) so that &extract can
			 # be done recursively (e.g., ID defs within ID defs);
			 # also, keeps track of the header for each external file ...

			 $BID = $ID;
			 &extract($ID,$nbl,$nbc,$nel,$nec,$bbl,$bbc,$bel,$bec,$initFH,0);


			 pop(@FHs);	  # harm done if this is not put here ...
		  }

		  pop(@Hs);		  # clear out Header stack ...
		  
		  # just print the rest of the file ...
		  &printHTML(substr($line,$offset)); # flush out last line
		  while ($bar=<FOO>) {	  # print the rest
			 $ln0++;
			 &printHTML(&addLN($ln0)) if ($LINENUMBERS);
			 &printHTML($bar);
		  }

		  # close the htmlized file ...
		  close(FOO);
		  &printHTML("</PRE>");
		  close(HTML);
		  &printD("done.\n");
	   }

	   # generate an index for this GROUP ....
	   print ".";		  # show that this is "thinking"
	   open(IOO,"> $bs/index.html");
	   print IOO "<A NAME=\"TOP\"></A>";
	   print IOO "<TITLE>Index for $project $subtitle</TITLE>\n";
	   print IOO "<H2><A HREF=\"../index.html\">Index</A> <FONT SIZE=4>for</FONT> <FONT SIZE=4><I>$project</I></FONT> <B><TT>$subtitle</TT></B> <FONT SIZE=4>files</FONT></H2><P>\n";
	   if (@keys=(keys %GROUP)) {
		  $kiys=$#keys;
		  do {
			 print IOO "<HR>\nJump to:<I><B><TT>  </TT>";
			 foreach $file (sort (@keys)) {
				$f = &basename($file);
				print IOO "<A HREF=\"#$f\">$f</A>";
				print IOO sprintf("<TT> %s </TT>",($kiys==0) ? "or" : ",")
 if (--$kiys>=0);
			 }
			 print IOO "</B></I>\n";
		  } if ($kiys>0);
		  foreach $file (sort (@keys)) {
			 $f = &basename($file);
			 print IOO "<A NAME=$f></A>\n<HR>\n";
			 print IOO "<H3><I><A HREF=\"$f.html\">$f</A></I></H3>\n";
			 print IOO "<DL><DL>\n";
			 @FIDs=split(/\s+/,$FILE_IDs{$file});
			 %FUNCS=(); %METHS=();
			 # separate the "functions" (functions and class/collection 
			 # defs) from the "methods" (methods of the class/collection)
			 foreach $ID (@FIDs) {
				if ($ID_TYPE{$ID}=~/(mem|moc|moe)/i) {
				    $METHS{$ID} = $MEMBER_OF{$ID};
				} else {
				    $FUNCS{$ID} = " ";
				}
			 }
			 # add the class/collection of methods to the top-level lists
			 foreach $ID (keys %METHS) {
				$FUNCS{$METHS{$ID}} = " ";
			 }
			 # sort and print out according to the following format:
			 #          ...
			 #          aabb              <--------- class definition
			 #             init()
			 #             saveit()
			 #          bbcc              <--------- top level subroutine(s)
			 #          ccdd              
			 #          eeff              <--------- class definition
			 #             error(int i)
			 #             jump()
			 #          ...
			 foreach $ID (sort byArgName (keys %FUNCS)) {
				next unless ($ID);

				print IOO "<DT>";

				# it's a main procedure ...
				do {
				    &printNR(IOO,"<B>",$ID,"</B>");
				    next;
				} if ($ID_TYPE{$ID}=~/seq main used/i);

				# it's a function ...
				do {
				    &printNR(IOO,"",$ID,"");
				    next;
				} if ($ID_TYPE{$ID}=~/(ord|main)/i);

				# it's a class ...
				do {
				    &printNR(IOO,"<B>",$ID,"</B>");
				    @CIDs=split(/\s/,$GROUP_CLASS{$ID});
				} if ($GROUP_CLASS{$ID});

				# it's a collection ...
				do {
				    &printNR(IOO,"<B><I>",$ID,"</B></I>");
				    @CIDs=split(/\s/,$GROUP_COLLECTION{$ID});
				} if ($GROUP_COLLECTION{$ID});

				print IOO "<DL>";

				# run through the members of the class/collection ...
				foreach $sID (sort byArgName (@CIDs)) {
				    next unless ( ($sID) && ($METHS{$sID}) );
				    &printNR(IOO," ",$sID,"");
				    print IOO "<BR>";
				}
				print IOO "</DL>";
			 }
			 print IOO "</DL></DL>\n";
		  }
		  print IOO "<HR><P>Jump to the <A HREF=\"#TOP\">TOP</A> of this page.\n";
	   } else {
		  print IOO "<DL><DL><H3><I>No files found!</I></H3></DL></DL>\n";
	   }
	   close(IOO);
    }

    # **************************** htmlizeCode ********************************

    print "Htmlizing file \`$sourceFile\' ";

    # generate sub-indices and html files ...
    open(IOO,sprintf("> %s/index.html",&clearDir($EXT_DIR)));
    print IOO "<A NAME=\"TOP\"></A>";
    print IOO "<TITLE>Index for $project External functions</TITLE>\n";
    print IOO "<H2><A HREF=\"../index.html\">Index</A> <FONT SIZE=4>for</FONT> <FONT SIZE=4><I>$project</I></FONT> <B><TT>External</TT></B> <FONT SIZE=4>functions</FONT></H2><P><HR>\n";
    if (@keys=(keys %GROUP_EXTERNAL)) {
        local($a,$alist,$b,$c);
        do {
		  %alpha = ();		  # use this to determine existence of alphachar
		  foreach $ID (sort byname (@keys)) {
			 next unless ($ID);
			 $a = substr($ID_NAME{$ID},0,1); $a =~ tr/a-z/A-Z/;  $b = $a;
			 $b =~ s/(\W)/\\\1/g;	# quote metacharacters
			 do {
				$alpha{$b} = ' ';
				$alist .= " <A HREF=\"#$a\">$a</A> ";
			 } unless ($alpha{$b});
		  }
		  print IOO "Jump to: <TT>  </TT> $alist <HR>\n";
	   } if ($#keys>1);
	   print IOO "<DL><DL>\n";
        foreach $ID (sort byname (@keys)) {
		  next unless ($ID);
		  $a = substr($ID_NAME{$ID},0,1); $a =~ tr/a-z/A-Z/;  $b = $a;
		  $b =~ s/(\W)/\\\1/g;	# quote metacharacters
		  do {
			 print IOO "<A NAME=\"$a\"></A>" ;
			 print IOO "<H3>$a</H3><P>";
		  } if (! ($c =~ /\s$b\s/)) && ($c .= " $a ");
		  print IOO "<A NAME=\"$ID.html\"></A>";
		  print IOO "<DT>$ID_NAME{$ID}";
		  $BID = $ID;
		  do {
			 ($w1,$w2,$w3) = ($ID_NAME{$MMMM}) ? ('',&dirID($MMMM),$ID_NAME{$MMMM}) : ('ID=',"../$ERR_DIR/index.html#$MMMM","$MMMM") ;
			 $w3 = "<I>$w3</I>" if ($GROUP_COLLECTION{$MMMM});
			 $w4 = ($ID_POS_NAME{$MMMM}) ? "" : " <FONT SIZE=1>(referenced, but NOT defined)</FONT>";
			 # yet another place to catch errors ...
			 do {
				$newErrors{$MMMM} = "<A NAME=\"$MMMM.html\"></A>ID=<B>$MMMM</B> is a referenced but UNdefined object with the following member function(s):" unless ($newErrors{$MMMM});
				$aprefix = ($GROUP_ERROR{$MMMM}) ? " , " : "" ;
				$GROUP_ERROR{$MMMM} .= "$aprefix<B><A HREF=\"../$EXT_DIR/index.html#$ID.html\">$ID_NAME{$ID}</A></B>";
			 } if ($w4);
			 print IOO sprintf("%s <TT> <FONT SIZE=2>, <I>member of</I> %s</FONT></TT>",$ID_ARGLIST{$ID},&ahref($w1,$w2,$w3,$w4)) ;
		  } if ($MMMM=$MEMBER_OF{$ID});
		  print IOO "\n";
	   }
	   foreach $ID (keys %newErrors) {
	       $GROUP_ERROR{$ID} = sprintf("<TT>%s <P> <DL>%s</DL><P></TT>",$newErrors{$ID},$GROUP_ERROR{$ID});
	   }
        print IOO "</DL></DL>\n";

    } else {
            print IOO "<DL><DL><H3><I>None found!</I></H3></DL></DL>\n";
    }
    close(IOO);

    # generate individual group indices ...
    &htmlize("$SRC_DIR","Source",*GROUP_SOURCE);
    &htmlize("$INC_DIR","Include",*GROUP_INCLUDE);
    &htmlize("$SYS_DIR","SysLib",*GROUP_SYSLIB);

    # generate main index ...
    open(IOO,"> $baseDir/index.html");
    print IOO "<TITLE>Main Index for $project</TITLE>\n";
    print IOO "<H1><A HREF=\"$URL\"><IMG SRC=\"$ICON\"></A> Main Index for <I>$project</I></H1>\n";
    print IOO "<PRE><DL><HR><DL>\n";
    sub htit {
	   local($foo)=sprintf("<A HREF=\"%s/index.html\"><B><TT>%s</TT></B></A>",@_);
	   push(@FINAL_GROUPS,$foo);
	   sprintf("<H2><LI>%s</H2>\n",$foo);
    }
    print IOO &htit($SRC_DIR,"Source") if (%GROUP_SOURCE);
    print IOO &htit($INC_DIR,"Include") if (%GROUP_INCLUDE);
    print IOO &htit($SYS_DIR,"SysLib") if (%GROUP_SYSLIB);
    print IOO &htit($EXT_DIR,"External") if (%GROUP_EXTERNAL);
    print IOO &htit($ERR_DIR,"Errors") if (%GROUP_ERROR);
    pop(@FINAL_GROUPS);  # ignore last one (Errors); see below ...
    print IOO "</DL><HR></DL></PRE>\n";
    print IOO "<ADDRESS>\n";
    print IOO "<A HREF=\"mailto:$MAILTO\">$MAILTO</a>\n";
    print IOO "</ADDRESS>\n";
    close(IOO);

    # last chance to handle errors ...
    do {
        open(IOO,sprintf("> %s/index.html",&clearDir($ERR_DIR)));
	   print IOO "<TITLE>Index of $project Errors</TITLE>\n";
	   print IOO "<H2><A HREF=\"../$base.index.html\">Index</A> <FONT SIZE=4>of</FONT> <FONT SIZE=4><I>$project</I></FONT> <B><TT>Errors</TT></B></H2><P>\n";
        print IOO "<HR><PRE>Groups:  ";
        foreach $g (@FINAL_GROUPS) {
		  $g=~s/HREF=\"/HREF=\"\.\.\//; # fix the location ... 
		  print IOO "<B>$g</B>  "; # who says you can't reuse stuff ???
        }
        print IOO "</PRE><HR><P>\n";
        print IOO "Note: The following items could not be catagorized in any of the above groups:\n<DL><DL>\n";
        foreach $key (sort numerically (keys %GROUP_ERROR)) {
		  print IOO "$GROUP_ERROR{$key}\n";
	   }
        print IOO "</DL></DL>\n";
        close(IOO);
    } if (%GROUP_ERROR);

    print " done.\n";
}


# ***************************** main program **********************************

$usage =
"Usage:  $0 source project dir [ opts ]\n\
 where \
       `source'     is the c/c++/pc++/f source filename\n\
       `project'    is the name of the project (a string) that\
                    will be placed at top of the Main Index\n\
       `dir'        is the target directory where the htmiized\
                    files will be placed\n
       `opts'       is any of the following options:\n\
                       [ -debug ]        print debugging info\
                       [ -v(erbose) ]    turn on verbose mode\
                       [ -cgmtest ]      indicate whether program\
                                         generates valid cgm file\
                       [ -icon webImg ]  web link to an image file\
                       [ -url URL ]      web link to another URL\
                       [ -mailto addr ]  address to send email to\n\n";

# read in command-line arguments

die "$usage \n"
    unless ($sourceFile = shift(@ARGV));

# find base filename regardless of whether user gave "foo" or "foo.x"
$foobar = &basename($sourceFile); # get basename of sourcefile
$barfoo = &dirname($sourceFile);  # get dirname of sourcefile
$base = $1 if ($foobar=~/([^.]+)(\.[^.]+)?$/i); # remove extension from basename
$depFile=&depName("$barfoo/$base"); # create depfilename from dirname and base

die "$usage   Cannot access dep file \'$depFile\'\n"
    unless (-r $depFile);
die "$usage   Must specify a project name! \n" 
    unless ($project=shift(@ARGV));
die "$usage   Must specify a target directory! \n"
    unless ($baseDir=(shift(@ARGV)));

# get options ...
while ($o=shift(@ARGV)) {
    ($o=~/^-v(erbose)?/i) && ($VERBOSE=' ')          && next;
    ($o=~/^-debug/i)      && ($DEBUG=' ')            && next;
    ($o=~/^-cgmtest/i)    && ($CGMTEST=' ')          && next;
    ($o=~/^-icon/i)       && ($ICON=shift(@ARGV))    && next;
    ($o=~/^-lang/i)       && ($LANG=shift(@ARGV))    && next;
    ($o=~/^-url/i)        && ($URL=shift(@ARGV))     && next;
    ($o=~/^-mailto/i)     && ($MAILTO=shift(@ARGV))  && next;
}

# yet another hack for passing language type to cgm ...
$LANG = ($LANG=~/$C_EXTENSIONS/) ? "c" : "f";

# use system defaults if user doesn't provide these ...
$ICON   = $PCXX_ICON unless ($ICON);
$URL    = $LINK_HOME unless ($URL);
$MAILTO = $MAIL_TO   unless ($MAILTO);

# clear out basedir
(-e $baseDir) ? &ss("$RM -f $baseDir/*.html") : &ss("$MKDIR $baseDir");

die "$usage  Cannot determine root directory from relative \'$baseDir\'! \n"
    unless chop($rootDir = `(cd $baseDir; $PWD)`);

$baseDir = &fixDir($baseDir);       # fix directory mount points
$rootDir = &fixDir($rootDir);       # fix directory mount points

$|=1; # cause output to be shown IMMEDIATELY rather than all at once at the end

&printV("sourceFile=$sourceFile, project=$project, baseDir=$baseDir, url=$URL, mailto=$MAILTO\n");

# first we pass the dep file through Bernd's cgm tool to produce a 'cgm' file
# (but only if unable to find a valid 'cgm' file) ...

$cgmFile=sprintf("%s/%s",$baseDir,&cgmName($base));

if ( (-r $cgmFile) && (-s _ ) && (-T _ ) ) {

    print "Re-using $cgmFile \n";

} else {

    &printV("Converting $depFile to $cgmFile ...");

    die "\n *** Cannot find dep file \'$depFile\' *** ABORTED\n"
	   unless (-r $depFile);

    $TMP_CGMFILE = "./TMP-cgm";

    do {
	   &ss("$RM -f $TMP_CGMFILE ./core");
	   die "\n$!\n *** Cannot create cgm file \`$cgmFile\` *** ABORTED\n";
    } unless (&ss("$CGM -dumptxt $depFile > $TMP_CGMFILE")==0);

    if ($CGMTEST) {
	   &printV("\n$sourceFile produced valid CGM file\n");
	   &ss("$RM -rf $baseDir $TMP_CGMFILE");
	   exit 0;
    }

    die "\n$!\n *** Unable to clear directory \`$baseDir\' *** ABORTED\n"
	   unless (((-e $baseDir) ? 
			 &ss("$RM -f $baseDir/*.html") : &ss("$MKDIR $baseDir"))==0);

    &ss("$CP $TMP_CGMFILE $cgmFile ; $RM -f $TMP_CGMFILE");

    &printV(" done.\n");
};

# next, we parse Bernd's CGM format that contains parse information
# about function/class body positions and *some* references therein

&readCGM();

# now we create global database that edoc, mccp, and csearch will use ...

$dbaseName = sprintf("%s/%s",$baseDir,&dbsName($base));

&printDB();

# process HTML file, finding things as they come into existence ...

&htmlizeCode();

# finally, update the global database (add ID_ARGLIST to it) ...

&updateDB();

# The End.

1;
