// This may look like C code, but it is really -*- C++ -*-
/*
Copyright (C) 1988 Free Software Foundation
    written by Doug Lea (dl@rocky.oswego.edu)

This file is part of the GNU C++ Library.  This library is free
software; you can redistribute it and/or modify it under the terms of
the GNU Library General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your
option) any later version.  This library is distributed in the hope
that it will be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE.  See the GNU Library General Public License for more details.
You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free Software
Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

// Modified by Shelby Yang, Indiana University. (5/12/94)
// Complex is for double data type.  ComplexFloat is for float data type.  
// Functionalities for ComplexFloat are incomplete because some functions
// for float types are not implemented in GNU C++ library and because
// pC++ cannot handle C++ function overloading (i.e., functions available
// in Complex class cannot be overloaded in ComplexFloat class.)

#ifndef _Complex_h
#define _Complex_h 1

//#include <iostream.h>
#include <math.h>

class Complex
{
public:
  double           real;
  double           imag;
                   Complex();
                   Complex(const Complex& y);
                   Complex(double r, double i=0);
                  ~Complex();

  Complex&         operator =  (const Complex& y);
  Complex&         operator += (const Complex& y);
  Complex&         operator += (double y);
  Complex&         operator -= (const Complex& y);
  Complex&         operator -= (double y);
  Complex&         operator *= (const Complex& y);
  Complex&         operator *= (double y);
  Complex&         operator /= (const Complex& y); 
  Complex&         operator /= (double y); 

  int              operator == (const Complex& y);
  int              operator == (double y);
  int              operator != (const Complex& y);
  int              operator != (double y);

  Complex          operator + (const Complex& y);
  Complex          operator + (double y);
  Complex          operator - (const Complex& y);
  Complex          operator - (double y);
  Complex          operator * (const Complex& y);
  Complex          operator * (double y);

  void             error(const char* msg) const;
};

// non-inline functions

Complex   operator /  (const Complex& x, const Complex& y);
Complex   operator /  (const Complex& x, double y);
Complex   operator /  (double   x, const Complex& y);

Complex   cos(const Complex& x);
Complex   sin(const Complex& x);

Complex   cosh(const Complex& x);
Complex   sinh(const Complex& x);

Complex   exp(const Complex& x);
Complex   log(const Complex& x);

Complex   pow(const Complex& x, int p);
Complex   pow(const Complex& x, const Complex& p);
Complex   pow(const Complex& x, double y);
Complex   sqrt(const Complex& x);
   
//istream&  operator >> (istream& s, Complex& x);
//ostream&  operator << (ostream& s, const Complex& x);

// other functions defined as inlines

Complex  operator - (const Complex& x);
Complex  operator + (double x, const Complex& y);
Complex  operator - (double x, const Complex& y);
Complex  operator * (double x, const Complex& y);
Complex  conj(const Complex& x);

double  real(const Complex& x);
double  imag(const Complex& x);
double  abs(const Complex& x);
double  norm(const Complex& x);
double  arg(const Complex& x);

Complex  polar(double r, double t = 0.0);

// inline members

inline Complex::Complex() {}
inline Complex::Complex(const Complex& y) :real(y.real), imag(y.imag) {}
inline Complex::Complex(double r, double i) :real(r), imag(i) {}

inline Complex::~Complex() {}

inline Complex&  Complex::operator =  (const Complex& y) 
{ 
  real = y.real; imag  = y.imag; return *this; 
} 

inline Complex&  Complex::operator += (const Complex& y)
{ 
  real += y.real;  imag  += y.imag; return *this; 
}

inline Complex&  Complex::operator += (double y)
{ 
  real += y; return *this; 
}

inline Complex&  Complex::operator -= (const Complex& y)
{ 
  real -= y.real;  imag -= y.imag; return *this; 
}

inline Complex&  Complex::operator -= (double y)
{ 
  real -= y; return *this; 
}

inline Complex&  Complex::operator *= (const Complex& y)
{  
  double r = real* y.real - imag * y.imag;
  imag = real* y.imag + imag * y.real; 
  real = r; 
  return *this; 
}

inline Complex&  Complex::operator *= (double y)
{  
  real *=  y; imag *=  y; return *this; 
}

inline int  Complex::operator == (const Complex& y)
{
  return real == y.real && imag == y.imag;
}

inline int  Complex::operator == (double y)
{
  return imag == 0.0 && real == y;
}

inline int  Complex::operator != (const Complex& y)
{
  return real != y.real || imag != y.imag;
}

inline int  Complex::operator != (double y)
{
  return imag != 0.0 || real != y;
}

inline Complex  Complex::operator + (const Complex& y)
{
  return Complex(real + y.real, imag + y.imag);
}

inline Complex  Complex::operator + (double y)
{
  return Complex(real + y, imag);
}

inline Complex  Complex::operator - (const Complex& y)
{
  return Complex(real - y.real, imag - y.imag);
}

inline Complex  Complex::operator - (double y)
{
  return Complex(real - y, imag);
}

inline Complex  Complex::operator * (const Complex& y)
{
  return Complex(real * y.real - imag * y.imag, 
                 real * y.imag + imag * y.real);
}

inline Complex  Complex::operator * (double y)
{
  return Complex(real * y, imag * y);
}

// functions

inline Complex  operator - (const Complex& x)
{
  return Complex(-x.real, -x.imag);
}

inline Complex  conj(const Complex& x)
{
  return Complex(x.real, -x.imag);
}

inline Complex operator + (double x, const Complex& y)
{
  return Complex(x + y.real, y.imag);
}

inline Complex  operator - (double x, const Complex& y)
{
  return Complex(x - y.real, -y.imag);
}

inline Complex  operator * (double x, const Complex& y)
{
  return Complex(x * y.real, x * y.imag);
}

inline double  real(const Complex& x)
{
  return x.real;
}

inline double  imag(const Complex& x)
{
  return x.imag;
}

inline double  abs(const Complex& x)
{
  return hypot(x.real, x.imag);
}

inline double  norm(const Complex& x)
{
  return (x.real * x.real + x.imag * x.imag);
}

inline double  arg(const Complex& x)
{
  return atan2(x.imag, x.real);
}

inline Complex  polar(double r, double t)
{
  return Complex(r * cos(t), r * sin(t));
}


class ComplexFloat
{
public:
  float           real;
  float           imag;
                   ComplexFloat();
                   ComplexFloat(const ComplexFloat& y);
                   ComplexFloat(float r, float i=0);
                  ~ComplexFloat();

  ComplexFloat&         operator =  (const ComplexFloat& y);
  ComplexFloat&         operator += (const ComplexFloat& y);
  ComplexFloat&         operator += (float y);
  ComplexFloat&         operator -= (const ComplexFloat& y);
  ComplexFloat&         operator -= (float y);
  ComplexFloat&         operator *= (const ComplexFloat& y);
  ComplexFloat&         operator *= (float y);
//  ComplexFloat&         operator /= (const ComplexFloat& y); 
//  ComplexFloat&         operator /= (float y); 

  int              operator == (const ComplexFloat& y);
  int              operator == (float y);
  int              operator != (const ComplexFloat& y);
  int              operator != (float y);

  ComplexFloat          operator + (const ComplexFloat& y);
  ComplexFloat          operator + (float y);
  ComplexFloat          operator - (const ComplexFloat& y);
  ComplexFloat          operator - (float y);
  ComplexFloat          operator * (const ComplexFloat& y);
  ComplexFloat          operator * (float y);

//  void             error(const char* msg) const;
};

// non-inline functions

// the following functions are commented out because there is no 
// corresponding implementation in GNU C++ library for some of the functions
// and also because pC++ parser cannot handle C++ function overloading.

//ComplexFloat   operator /  (const ComplexFloat& x, const ComplexFloat& y);
//ComplexFloat   operator /  (const ComplexFloat& x, float y);
//ComplexFloat   operator /  (float   x, const ComplexFloat& y);

//ComplexFloat   cos(const ComplexFloat& x);
//ComplexFloat   sin(const ComplexFloat& x);

//ComplexFloat   cosh(const ComplexFloat& x);
//ComplexFloat   sinh(const ComplexFloat& x);

//ComplexFloat   exp(const ComplexFloat& x);
//ComplexFloat   log(const ComplexFloat& x);

//ComplexFloat   pow(const ComplexFloat& x, int p);
//ComplexFloat   pow(const ComplexFloat& x, const ComplexFloat& p);
//ComplexFloat   pow(const ComplexFloat& x, float y);
//ComplexFloat   sqrt(const ComplexFloat& x);
   
//istream&  operator >> (istream& s, ComplexFloat& x);
//ostream&  operator << (ostream& s, const ComplexFloat& x);

// other functions defined as inlines

//ComplexFloat  operator - (const ComplexFloat& x);
//ComplexFloat  operator + (float x, const ComplexFloat& y);
//ComplexFloat  operator - (float x, const ComplexFloat& y);
//ComplexFloat  operator * (float x, const ComplexFloat& y);
//ComplexFloat  conj(const ComplexFloat& x);

//float  real(const Complex& x);
//float  imag(const Complex& x);
//float  abs(const ComplexFloat& x);
//float  norm(const ComplexFloat& x);
//float  arg(const ComplexFloat& x);

//ComplexFloat  polar(float r, float t = 0.0);

// inline members

inline ComplexFloat::ComplexFloat() {}
inline ComplexFloat::ComplexFloat(const ComplexFloat& y) :real(y.real), imag(y.imag) {}
inline ComplexFloat::ComplexFloat(float r, float i) :real(r), imag(i) {}

inline ComplexFloat::~ComplexFloat() {}

inline ComplexFloat&  ComplexFloat::operator =  (const ComplexFloat& y) 
{ 
  real = y.real; imag  = y.imag; return *this; 
} 

inline ComplexFloat&  ComplexFloat::operator += (const ComplexFloat& y)
{ 
  real += y.real;  imag  += y.imag; return *this; 
}

inline ComplexFloat&  ComplexFloat::operator += (float y)
{ 
  real += y; return *this; 
}

inline ComplexFloat&  ComplexFloat::operator -= (const ComplexFloat& y)
{ 
  real -= y.real;  imag -= y.imag; return *this; 
}

inline ComplexFloat&  ComplexFloat::operator -= (float y)
{ 
  real -= y; return *this; 
}

inline ComplexFloat&  ComplexFloat::operator *= (const ComplexFloat& y)
{  
  float r = real* y.real - imag * y.imag;
  imag = real* y.imag + imag * y.real; 
  real = r; 
  return *this; 
}

inline ComplexFloat&  ComplexFloat::operator *= (float y)
{  
  real *=  y; imag *=  y; return *this; 
}

inline int  ComplexFloat::operator == (const ComplexFloat& y)
{
  return real == y.real && imag == y.imag;
}

inline int  ComplexFloat::operator == (float y)
{
  return imag == 0.0 && real == y;
}

inline int  ComplexFloat::operator != (const ComplexFloat& y)
{
  return real != y.real || imag != y.imag;
}

inline int  ComplexFloat::operator != (float y)
{
  return imag != 0.0 || real != y;
}

inline ComplexFloat  ComplexFloat::operator + (const ComplexFloat& y)
{
  return ComplexFloat(real + y.real, imag + y.imag);
}

inline ComplexFloat  ComplexFloat::operator + (float y)
{
  return ComplexFloat(real + y, imag);
}

inline ComplexFloat  ComplexFloat::operator - (const ComplexFloat& y)
{
  return ComplexFloat(real - y.real, imag - y.imag);
}

inline ComplexFloat  ComplexFloat::operator - (float y)
{
  return ComplexFloat(real - y, imag);
}

inline ComplexFloat  ComplexFloat::operator * (const ComplexFloat& y)
{
  return ComplexFloat(real * y.real - imag * y.imag, 
                 real * y.imag + imag * y.real);
}

inline ComplexFloat  ComplexFloat::operator * (float y)
{
  return ComplexFloat(real * y, imag * y);
}

// functions

// these functions are commented out because pC++ parser cannot handle
// C++ function overloading and also because there is no implementation
// in GNU C++ library for some of the functions

//inline ComplexFloat  operator - (const ComplexFloat& x)
//{
//  return ComplexFloat(-x.real, -x.imag);
//}

//inline ComplexFloat  conj(const ComplexFloat& x)
//{
// return ComplexFloat(x.real, -x.imag);
//}

//inline ComplexFloat operator + (float x, const ComplexFloat& y)
//{
//  return ComplexFloat(x + y.real, y.imag);
//}

//inline ComplexFloat  operator - (float x, const ComplexFloat& y)
//{
//  return ComplexFloat(x - y.real, -y.imag);
//}

//inline ComplexFloat  operator * (float x, const ComplexFloat& y)
//{
//  return ComplexFloat(x * y.real, x * y.imag);
//}

//inline float  real(const Complex& x)
//{
//  return x.real;
//}

//inline float  imag(const Complex& x)
//{
//  return x.imag;
//}

//inline float  abs(const ComplexFloat& x)
//{
//  return hypot(x.real, x.imag);
//}

//inline float  norm(const ComplexFloat& x)
//{
//  return (x.real * x.real + x.imag * x.imag);
//}

//inline float  arg(const ComplexFloat& x)
//{
//  return atan2(x.imag, x.real);
//}

//inline ComplexFloat  polar(float r, float t)
//{
//  return ComplexFloat(r * cos(t), r * sin(t));
//}

#endif
