/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/


/****************************************************************
 *                                                              *
 *         pC++ DistributedArray Collection                     *
 *         Written by Francois Bodin                            *
 *         Computer Science, Indiana University                 *
 *                                                              *
 *         Revised by Shelby Yang (6/24/93)                     *
 *         Computer Science, Indiana University                 *
 *                                                              *
 *         Notice it now inherits SuperKernel, a                *
 *         collection, instead of Kernel, a class.              *
 *         S.Y. (6/24/93)                                       *
 *                                                              *
 ****************************************************************/

Collection DistributedArray : public SuperKernel {
 public:
  int blocksize;
  DistributedArray(Distribution *T, Align *A);
  void Assign(int i1, ElementType &e);        //assign one element to another
  void Assign(int i1, int i2, ElementType &e);//assign one element to another
  DistributedArray<ElementType> 
      &operator = (DistributedArray<ElementType> &rhs); //does elementwise copy
  void Shift(int dim, int distance);//shift elements along a dimension by dist.
  void Reduce();     //reduce all elements. store result in element 0.
  void ReduceDim1(); //reduce along first dimension. result is in first row.
  void ReduceDim2(); //reduce along second dimension. result is in first column
  void Broadcast();
  void BroadcastDim1();
  void BroadcastDim2();
  double DotProd(DistributedArray<ElementType> &arg1); 
 MethodOfElement:
  void WhereAmI();  //return processor ID assigned to an element 
  /***************************************************************************
     The following two virtal functions are used for reduction and dot-product.
     Their implementation depends on ElementType. A user needs to give real 
     definitions in an element class or in the MethodOfElement part of a 
     derived collection if he wishes to invoke reduction or dot-product 
     functions. He needs to provide dummy definitions to fool the compiler 
     if the functions are not to be invoked. 
  ****************************************************************************/
  /*****************************************************/
  /**** the following functions are used internally ****/
  /*****************************************************/
  ElementType *temporaryelement;
  double dotprodtemp;
  void DeleteTemp();
  void LocalGSum(DistributedArray<ElementType> &array,int step);
  void BroadcastTemp(DistributedArray<ElementType> &array,int step);
  void ResetDotProd(DistributedArray<ElementType> &arg1);
  void LocalCopy(DistributedArray<ElementType> *array);
  void Set_ThisCollection(DistributedArray<ElementType> *C);
  void LocalReduce(DistributedArray<ElementType> *array,int step);
  void LocalReduceDim1(DistributedArray<ElementType> *array,int step);
  void LocalBroadcast(DistributedArray<ElementType> *array,int step);
  void LocalBroadcastDim1(DistributedArray<ElementType> *array,int step);
  void LocalReduceDim2(DistributedArray<ElementType> *array,int step);
  void LocalBroadcastDim2(DistributedArray<ElementType> *array,int step);
  void LocalShiftStep1(int dim, int distance);
  void LocalShiftStep2();
  virtual double Dot(ElementType &arg1);
  virtual ElementType &operator +=(ElementType &rhs);
};   

DistributedArray :: DistributedArray(Distribution *T, Align *A) : SuperKernel(T, A)
{
  if (!T || !A)
    pcxx_Message("No alignment or template for collection initialization.",PCXX_FATAL);
  blocksize = 0;
  this->Set_ThisCollection(this);
  if (A->nbdim >= 3)
      pcxx_Message("3D and higher-dimensional DistributedArray not implemented.",
	    PCXX_FATAL);
};

// Wrap around shift 
void DistributedArray ::Shift(int dim, int distance)
{
  this->LocalShiftStep1(dim, distance);
  this->LocalShiftStep2();
  this->DeleteTemp();
}

void DistributedArray ::LocalShiftStep1(int dim, int distance)
{
  int elem;
  ElementType *pt;
  double *ptc, *pte;
  int i,size;
  if (dim == 1)
    {
      elem = (index1 + distance);
      if (elem >= 0)
        elem = elem % ThisCollection->dim1size;
      else
        {
          while (elem<0)
            elem = elem + ThisCollection->dim1size;
        }
      if (ThisCollection->Is_Local(elem, index2))
        {          
          pt =  (ElementType *) ThisCollection->Get_Element(elem, index2);
          size = (ThisCollection->CollectionInfo.element_size)/(int)sizeof(double);
          this->temporaryelement = (ElementType *) new double[size +1];
	  ptc = (double *) this->temporaryelement;
          pte = (double *) pt;
          for (i = 0; i <= size; i++)
            *ptc++ = *pte++;
        } else
          {
            this->temporaryelement = (ElementType *) 
	      ThisCollection->Get_CopyElem(elem, index2);
          }
    } else
      {
        
        elem = (index2 + distance);
        if (elem >= 0)
          elem = elem % ThisCollection->dim2size;
        else
          {
            while (elem<0)
              elem = elem + ThisCollection->dim2size;
          }        
        if (ThisCollection->Is_Local(index1, elem))
          {
            pt =  (ElementType *) ThisCollection->Get_Element(index1, elem);
            size = (ThisCollection->CollectionInfo.element_size)/
                           (int)sizeof(double);
            this->temporaryelement = (ElementType *) new double[size +1];
	    ptc = (double *) this->temporaryelement;
            pte = (double *) pt;
            for (i = 0; i <= size; i++)
              *ptc++ = *pte++;
          } else
            {
              this->temporaryelement = (ElementType *) 
		ThisCollection->Get_CopyElem(index1, elem);
            }
      }
}

void DistributedArray ::LocalShiftStep2()
{
  *this = *(this->temporaryelement);
}

void DistributedArray ::DeleteTemp()
{
  delete this->temporaryelement;
  this->temporaryelement = 0;
}

void DistributedArray :: Set_ThisCollection(DistributedArray<ElementType> *C)
{
  this->ThisCollection = C;
}

void DistributedArray :: WhereAmI()
{
  printf("Element (%d, %d) on processor %d\n",index1,index2, pcxx_NodeNumber());
  pcxx_Barrier();
}

void DistributedArray::Assign(int i1, ElementType &e)
{
  if(this->Is_Local(i1)) *((*this)(i1)) = e;
}

void DistributedArray::Assign(int i1, int i2, ElementType &e)
{
  if(this->Is_Local(i1,i2)) *((*this)(i1,i2)) = e;
}

void DistributedArray::Reduce()
{
  int steps = 1;
  int borne;
  if (dim2size)
    borne = dim1size*dim2size;
  else
     borne = dim1size;
  while (steps < borne)
    {
      this->LocalReduce(this,steps);
      steps = steps*2;
    }
}

/* Assume a power of two numbers of elements in the collection */
void DistributedArray::ReduceDim1()
{
  int steps = 1;
  while (steps < dim1size)
    {
      this->LocalReduceDim1(this,steps);
      steps = steps*2;
    }
}

void DistributedArray::ReduceDim2()
{
  int steps = 1;
  while (steps < dim2size)
    {
      this->LocalReduceDim2(this,steps);
      steps = steps*2;
    }
}

void DistributedArray::LocalReduce(DistributedArray<ElementType> 
				   *array,int step)
{
  int  ind;
  ind = ThisCollection->ElemIndex(index1,index2);
  if (ind % (2*step) == 0)
    { /* do reduction */
      *this += *((*array)(ind +step));
    } /* else do nothing */
}

void DistributedArray::LocalReduceDim1(DistributedArray<ElementType> 
				       *array,int step)
{
  if (index1 % (2*step) == 0)
    { /* do reduction */
      *this += *((*array)(index1 + step, index2));
    } /* else do nothing */
}

void DistributedArray::LocalReduceDim2(DistributedArray<ElementType> 
				       *array,int step)
{
  if (index2 % (2*step) == 0)
    { /* do reduction */
      *this += *((*array)(index1, index2 + step));
    } /* else do nothing */
}

/* Assume power of two numbers of elements in the collection */
void DistributedArray::Broadcast()
{
  int steps;
  if (dim2size)
    steps = (dim1size*dim2size)/2;
  else
    steps = dim1size/2;
  while (steps)
    {
      this->LocalBroadcast(this,steps);
      if (steps == 1)
        break;
      steps = steps/2;
    }
}

void DistributedArray::BroadcastDim1()
{
  int steps = dim1size/2;
  while (steps)
    {
      this->LocalBroadcastDim1(this,steps);
      if (steps == 1)
        break;
      steps = steps/2;
    }
}

void DistributedArray::BroadcastDim2()
{
  int steps = dim2size/2;
  while (steps)
    {
      this->LocalBroadcastDim2(this,steps);
      if (steps == 1)
        break;
      steps = steps/2;
    }
}

void DistributedArray::LocalBroadcast(DistributedArray<ElementType> 
				      *array, int step)
{
  int ind;
  ind = ThisCollection->ElemIndex(index1,index2);
  if (ind && ((ind % step) == 0) && !((ind % (2*step)) == 0))
    { 
      *this = *((*array)(ind - step));
    } 
}

void DistributedArray::LocalBroadcastDim1(DistributedArray<ElementType> 
					  *array,int step)
{
  if (index1 && ((index1 % step) == 0) && !((index1 % (2*step)) == 0))
    { 
      *this = *((*array)(index1 - step, index2));
    } /* else do nothing */
}

void DistributedArray::LocalBroadcastDim2(DistributedArray<ElementType> 
					  *array,int step)
{
  if (index2 && ((index2 % step) == 0) && !((index2 % (2*step)) == 0))
    { /* broadcast */
      *this = *((*array)(index1, index2 - step));
    } /* else do nothing */
}

DistributedArray<ElementType> &DistributedArray::operator 
                              = (DistributedArray<ElementType> &rhs)
{
  this->LocalCopy(&rhs);
  return *this;
}

void DistributedArray::LocalCopy(DistributedArray<ElementType> *array)
{
  (*this) = *((*array)(index1,index2));
}

void DistributedArray::ResetDotProd(DistributedArray<ElementType> &arg1)
{
  dotprodtemp = this->Dot(*(arg1(index1,index2)));
}

double DistributedArray::DotProd(DistributedArray<ElementType> &arg1)
{
  double total;
  int steps = 1;
  int locali;
  ElementType dummy;
  int *addr1,*addr2;
  int borne;
  double *tmp;

  this->ResetDotProd(arg1);
  borne = dim1size;
  if (dim2size)
    borne = borne*dim2size;    
  while (steps < borne)
    {
      this->LocalGSum(*this,steps);
      steps = steps*2;
    }   
  if (dim2size) 
    steps = (dim1size*dim2size)/2;
  else
    steps = (dim1size)/2;
  while (steps)
    {
      this->BroadcastTemp(*this,steps);
      if (steps == 1)
	break;
      steps = steps/2;
    }
  locali = GetFirstLocalElementIndex(this);
  if (locali >= 0)
    total= (*this)(locali)->dotprodtemp;
  else
    {
      addr1 = (int *) &dummy;
      addr2  = (int *) &(dummy.dotprodtemp);
      tmp = (double *) this->Get_ElementPart((pcxx_NodeNumber() % this->dim1size),
	     (addr2 - addr1), (int)sizeof(double)/(int)sizeof(int));
      total = *tmp;
    }
  pcxx_Barrier();
  return total;
 }

void DistributedArray::LocalGSum(DistributedArray<ElementType> 
				 &array,int step)
{
  int ind;
  ind = ThisCollection->ElemIndex(index1,index2);
  if (ind % (2*step) == 0)
    { 
      int *addr1, *addr2;
      double *tmp;
      addr1 = (int *) this;
      addr2  = (int *) &(this->dotprodtemp);
      tmp = (double *) array.Get_ElementPart(ind + step,
		   (addr2 - addr1), (int)sizeof(double)/(int)sizeof(int));
      dotprodtemp += *tmp;
    } 
}

void DistributedArray::BroadcastTemp(DistributedArray<ElementType> 
				     &array,int step)
{
  int ind;
  ind = ThisCollection->ElemIndex(index1,index2);
  if (ind && ((ind % step) == 0) && !((ind % (2*step)) == 0))
    { 
      int *addr1, *addr2;
      double *tmp;
      addr1 = (int *) this;
      addr2  = (int *) &(this->dotprodtemp);
      tmp = (double *) array.Get_ElementPart(ind - step,
                                       (addr2 - addr1), 
                                       (int)sizeof(double)/(int)sizeof(int));
      this->dotprodtemp = *tmp;
    } 
}

