/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/


/***************************************************************************
 *                                                                         *
 *                                                                         *
 *                pC++ Routine for managing data access                    *
 *                CM5 version                                              *
 *                Written by Shelby Yang, Indiana University               *
 *                to use TMC's CMMD and CMAML package                      *
 *                                                                         *
 *                                                                         *
 *                                                                         *
 ***************************************************************************/
void Message(char * st, int flag);

#include "pcxx_events_def.h"
#include "pcxx_timers.h"
#include "profiler_defs.h"
#include "active.h"
#include <cm/cmmd.h>
#include <cm/cmcom_types.h>
#include <cm/cmrt_types.h>
#include <cmsys/dp.h>
#include "dmrts.h"
#include "common.h" /* Common support routines for EVERY KernelCom.c (phb) */

static int message_flag = 0; /* Same address on all nodes */
static int ask_flag = 0; /* Same address on all nodes */
static int ask_buf = 0;
static int remote_execute_flag = 0;
static int buffer_updated_flag = 0;
static int farray_elem_flag = 0;

extern pcxx_PARAM_t pcxx_param;

/* length has to be more than 5 */
static int length = 40;
static int *info_buffer;

/***************************************************************************
  GENERAL FUNCTIONS
 ***************************************************************************/
/*---------------------------------------------------------------------------
  Caution: According to CMMD Reference Manual, no CMMD/IO functions may
           be called in handlers.  printf, etc. are converted to CMMD/IO
	   functions when compiled and linked, so they are not allowed to
	   appear in handlers.  For the same reason, the following Message 
	   function may not be called in handlers. 
  ---------------------------------------------------------------------------*/
void Message(char * st, int flag)
{
   if (flag == WARNING) {
     WarningCount++;
     fprintf(stderr, "Warning: %s\n",st);
     return;
   }
   if (flag == FATAL) {
     fprintf(stderr, "Fatal error: %s\n",st);
     return;
   }
}

void pcxx_MainBegin(int *argc, char *argv[])
{
  int i, j;

  pcxx_InitParam (argc, argv);
  CMMD_fset_io_mode(stdout, CMMD_independent);
  CMMD_fset_io_mode(stderr, CMMD_independent);
  CMMD_fset_io_mode(stdin, CMMD_independent);
  MyTotalNodes = CMMD_partition_size();
  MyNodeNumber = CMMD_self_address();
  info_buffer = (int*) malloc(MyTotalNodes*length*sizeof(int));

  pcxx_EvInit (pcxx_param.tracefile);
  PCXX_EVENT (PCXX_EC_BASIC, PCXX_MAIN_ENTER, 0);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_NEWPROC_ENTER, 0);

  ResetCollectionTraversalLevel ();

  CMMD_sync_with_nodes();
# ifdef TRACING_ON
  CMMD_sync_with_nodes();
  pcxx_Event (PCXX_SYNC_MARK, 0);
# endif
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_NEWPROC_EXIT, MyNodeNumber);
}
 
void pcxx_MainEnd() 
{
  free((char*)info_buffer);

# ifdef TRACING_ON
  CMMD_sync_with_nodes();
  CMMD_sync_with_nodes();
  pcxx_Event (PCXX_SYNC_MARK, 0);
# endif
  PCXX_EVENT (PCXX_EC_BASIC, PCXX_MAIN_EXIT, 0);
  pcxx_EvClose();
}

int IamtheManager(int id)
{
  return ((id % MyTotalNodes) == MyNodeNumber);
}

/***************************************************************************
 REDUCTION FUNCTIONS
 ***************************************************************************/
/*--------------------------------------------------------------------------
   Note:  The processor with flag = 0 broadcasts. 
          length is the length of the message in bytes.
   Also:  At present, Thinking Machines does not support global operations
          on parallel arrays --- arrays allocated on VU's.  Therefore,
          parallel arrays cannot be broadcasted.
   Also:  need to check if there is a conflict with the polling here.
  --------------------------------------------------------------------------*/
void pcxx_BroadcastBytes(int flag,int length,void * buffer)
{
  if(flag == 0) CMMD_bc_to_nodes(buffer, length);
  else CMMD_receive_bc_from_node(buffer, length);
}

void pcxx_ReduceDoubleAdd(void* buffer,int elem_count)
{
  CMMD_reduce_v(buffer,buffer,CMMD_combiner_dadd,sizeof(double),elem_count);
}

void pcxx_ReduceFloatAdd(void* buffer,int elem_count)
{
  CMMD_reduce_v(buffer,buffer,CMMD_combiner_fadd,sizeof(float),elem_count);
}

void pcxx_ReduceIntAdd(void* buffer,int elem_count)
{
  CMMD_reduce_v(buffer,buffer,CMMD_combiner_add,sizeof(int),elem_count);
}

void pcxx_ReduceDoubleMax(void* buffer,int elem_count)
{
  CMMD_reduce_v(buffer,buffer,CMMD_combiner_dmax,sizeof(double),elem_count);
}

void pcxx_ReduceFloatMax(void* buffer,int elem_count)
{
  CMMD_reduce_v(buffer,buffer,CMMD_combiner_fmax,sizeof(float),elem_count);
}

void pcxx_ReduceIntMax(void* buffer,int elem_count)
{
  CMMD_reduce_v(buffer,buffer,CMMD_combiner_max,sizeof(int),elem_count);
}

void pcxx_ReduceDoubleMin(void* buffer,int elem_count)
{
  CMMD_reduce_v(buffer,buffer,CMMD_combiner_dmin,sizeof(double),elem_count);
}

void pcxx_ReduceFloatMin(void* buffer,int elem_count)
{
  CMMD_reduce_v(buffer,buffer,CMMD_combiner_fmin,sizeof(float),elem_count);
}

void pcxx_ReduceIntMin(void* buffer,int elem_count)
{
  CMMD_reduce_v(buffer,buffer,CMMD_combiner_min,sizeof(int),elem_count);
}


/*******************************************************************
 POLL FUNCTION
 *******************************************************************/
void pcxx_Poll()
{
/*
  There is no reason for this function to exist anymore once interrupts are
  enabled.  But since the current pC++ still generates function calls to
  this function, we will keep it here for the moment.
*/
  PROFILE_ENTER(PROFILE_Poll);

  CMAML_disable_interrupts();
  CMAML_poll();
  CMAML_enable_interrupts();

  PROFILE_EXIT(PROFILE_Poll);
  return;
}

/*******************************************************************
 BARRIER FUNCTION
 *******************************************************************/
void pcxx_Barrier()
{
  PROFILE_ENTER(PROFILE_Barrier);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_BARRIER_ENTER, 0);
  if(remote_execute_flag > 0) {
    CMAML_disable_interrupts();
    CMAML_poll_wait(&remote_execute_flag,0);
    CMAML_enable_interrupts();
  } 
  CMMD_sync_with_nodes();
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_BARRIER_EXIT, 0);
  PROFILE_EXIT(PROFILE_Barrier);
}

/**************************************************************************
 ACTIVE MESSAGE DRIVEN COMMUNICATION FUNCTIONS
 **************************************************************************/
void buffer_updated_handler();
void remote_store_handler(int dest_node,int start_at,int buffer1,int buffer2);

void here_is_data_handler(int proc)
{
  ask_buf = proc;
  ask_flag = 1;
  return;
}

void where_is_data_handler(int node, int index, int info)
{
  int data_is;
  int *addr;

  data_is = TableOfCollection[info]->Where[index];
  CMAML_reply(node, here_is_data_handler, data_is);
  return;
}

int WhereIsData(int id, int info)
{
  static int manager; 

  PROFILE_ENTER(PROFILE_Where);
  CMAML_disable_interrupts();
  manager = 0;
  ask_flag = 0;
  manager = (id % MyTotalNodes); 
  
  /* Send a request to manager which returns with index and collection addr */
  CMAML_request(manager, where_is_data_handler, MyNodeNumber, id, info);
  CMAML_poll_wait(&ask_flag, 1);
  CMAML_enable_interrupts();
  PROFILE_EXIT(PROFILE_Where);
  return ask_buf;
}

void end_of_block_handler(int rport)
{
  message_flag = 1;
  return;
}

void read_block_handler(unsigned index_Coll_Id, unsigned rport_dest_node,
                        int size, int offset)
{
  unsigned mask, dest_node, rport, Coll_Id, index;

  mask = 65536 - 1;

  Coll_Id = index_Coll_Id & mask;
  index = index_Coll_Id - Coll_Id;
  index = index >> 16;
  dest_node = rport_dest_node & mask;
  rport = rport_dest_node - dest_node;
  rport = rport >> 16;

  CMAML_scopy(dest_node, rport, 0, 0, 
	      (int*)(TableOfCollection[Coll_Id]->Table[index]+offset),
	      size, NULL, NULL);
  return;
}

int AskForDataBlock(int where, unsigned index, int offset, unsigned Coll_Id, 
		    int size, int *buffer)
{
  unsigned rport, self, index_Coll_Id, rport_self, limit;

  size = size*sizeof(int);
  CMAML_disable_interrupts();
 
  message_flag = 0;

  rport = CMAML_allocate_rport();
  if (rport == -1) Message("No port available.", FATAL);

  CMAML_set_rport_addr(rport, (void *) buffer);
  CMAML_set_rport_total_nbytes(rport, size);
  CMAML_set_rport_byte_counter(rport, size);
  CMAML_set_rport_handler(rport, end_of_block_handler);
  CMAML_set_rport_handler_arg2(rport, NULL);

  /*
    WARNING: The following only works when rport, index, Coll_Id, and
             self are less than 2^16-1 = 65535.
    Note:    rport = the number of ports opened.
             index <= the total number of elements in a collection.
             Coll_Id <= the total number of collections.
             self <= the total number of processors.
  */

  limit = 65535;
  self = MyNodeNumber;

  if(index > limit || Coll_Id > limit || rport > limit || self > limit) {
    Message("AskForDataBlock(): index, Coll_Id, rport, or self > 65535.",
            FATAL);
  }
  index_Coll_Id = (index<<16) | Coll_Id;
  rport_self = (rport<<16) | self;

  CMAML_request(where,read_block_handler,index_Coll_Id,rport_self,size,offset);
  CMAML_poll_wait(&message_flag, 1);
  CMAML_free_rport(rport);

  CMAML_enable_interrupts();
  return 1;
}

/*-----------------------------------------------------------------------
  Fortran interface related.

  Shelby Yang, Indiana University.  (10/25/93)
  ------------------------------------------------------------------------*/
extern CMRT_desc_t CMRT_allocate_heap_array();
extern CMRT_array_geometry_t CMRT_intern_geometry();
extern CMRT_array_geometry_t CMRT_intern_detailed_geometry();
extern unsigned CMRT_desc_get_cm_location();
extern int CMRT_desc_get_subgrid_dimension();
extern int CMRT_desc_get_offchip_position();
extern int CMRT_desc_get_subgrid_axis_increment();
extern int CMRT_desc_get_subgrid_size();
extern void CMRT_cross_geometry_move();

void end_of_array_handler(int rport)
{
  message_flag = 1;
  return;
}

void read_array_handler(unsigned index_Coll_Id, unsigned rport_dest_node,
                        int size, int offset)
{
  unsigned mask, dest_node, rport, Coll_Id, index;
  CMRT_desc_t buffer;

  mask = 65536 - 1;

  Coll_Id = index_Coll_Id & mask;
  index = index_Coll_Id - Coll_Id;
  index = index >> 16;
  dest_node = rport_dest_node & mask;
  rport = rport_dest_node - dest_node;
  rport = rport >> 16;

  buffer = (CMRT_desc_t)*((long*)(TableOfCollection[Coll_Id]->Table[index]+offset));

  CMAML_pcopy(dest_node, rport, 0, 0,
              (int*)buffer->cm_location,size,NULL,NULL);

  return;
}

int AskForFArray(int where, unsigned index, unsigned Coll_Id, int offset,
		 char *buffer, int size)
{
  unsigned rport, self, index_Coll_Id, rport_self, limit;

  CMAML_disable_interrupts();

  if(size < 0) Message("Negative size.", FATAL);
  message_flag = 0;

  rport = CMAML_allocate_rport();
  if (rport == -1) Message("No port available.", FATAL);

  CMAML_set_rport_addr(rport, (void *) ((CMRT_desc_t)buffer)->cm_location);
  CMAML_set_rport_total_nbytes(rport, size);
  CMAML_set_rport_byte_counter(rport, size);
  CMAML_set_rport_handler(rport, end_of_array_handler);
  CMAML_set_rport_handler_arg2(rport, NULL);

  /*
    WARNING: The following only works when rport, index, Coll_Id, and
             self are less than 2^16-1 = 65535.
    Note:    rport = the number of ports opened.
             index <= the total number of elements in a collection.
             Coll_Id <= the total number of collections.
             self <= the total number of processors.
  */

  limit = 65535;
  self = MyNodeNumber;

  if(index > limit || Coll_Id > limit || rport > limit || self > limit) {
    Message("AskForFArray(): index, Coll_Id, rport, or self > 65535.",
            FATAL);
  }
  index_Coll_Id = (index<<16) | Coll_Id;
  rport_self = (rport<<16) | self;

  CMAML_request(where,read_array_handler,index_Coll_Id,rport_self,size,offset);
  CMAML_poll_wait(&message_flag, 1);
  CMAML_free_rport(rport);

  CMAML_enable_interrupts();
  return 1;
}

/*------------------------------------------------------------------------
  Warning: The following defination may be changed by TMC without notice.
           Check with <cm/cmtypes> for its correctness.
  ------------------------------------------------------------------------*/
#define CM_no_axis -1

void CopyFArray(char *dest,char *source,int rank,int size)
{  
  int4 axis_mapping[10];
  int8 source_axis_coords[10];
  int8 dest_axis_coords[10];
  int i;

  if(rank > 10) 
    Message("CopyFArray(): Vector unit array rank > 10!",FATAL);

  for(i = 0; i < 10; i++) {
    axis_mapping[i] = 0;
    source_axis_coords[i] = 0;
    dest_axis_coords[i] = 0;
  }

  for(i = 0; i < rank; i++) {
    axis_mapping[i] = i;
    source_axis_coords[i] = CM_no_axis;
    dest_axis_coords[i] = CM_no_axis;
  }

  CMRT_cross_geometry_move((CMRT_desc_t)dest,(CMRT_desc_t)source,
			   axis_mapping,source_axis_coords,dest_axis_coords,
			   NULL);
}

static int MyExp2(int k) {
  int tt, i;
  tt = 1; 
  for(i = 0; i < k; i++) 
    tt = tt*2;
  return tt;
}

void CopyFArrayElem(char* source, char *buffer, int elem_size, int rank,
		    int i_len, int i_inc, int i_off,
		    int j_len, int j_inc, int j_off,
		    int k_len, int k_inc, int k_off,
		    int i, int j, int k)
{
  CMRT_desc_t d;
  int node, subgrid_index;
  int i_phy, i_sub;
  int j_phy, j_sub;
  int k_phy, k_sub;
  char *p0, *p1, *p2, *p3;
  int loc, n;

  d = (CMRT_desc_t)source;

  loc = CMRT_desc_get_cm_location(d);
  p0 = (char*)TOGGLE_DPV_SPACE(CHANGE_DP(loc,DP_0));
  p1 = (char*)TOGGLE_DPV_SPACE(CHANGE_DP(loc,DP_1));
  p2 = (char*)TOGGLE_DPV_SPACE(CHANGE_DP(loc,DP_2));
  p3 = (char*)TOGGLE_DPV_SPACE(CHANGE_DP(loc,DP_3));

  switch (rank) {
  case 1:
    i_phy = i / i_len;
    i_sub = i % i_len;

    node = (int)(i_phy*MyExp2(i_off));
    subgrid_index = i_sub*i_inc;
    break;
  case 2:
    i_phy = i / i_len;
    i_sub = i % i_len;
    j_phy = j / j_len;
    j_sub = j % j_len;

    node = (int)(i_phy*MyExp2(i_off))
          +(int)(j_phy*MyExp2(j_off));
    subgrid_index = i_sub*i_inc
                  + j_sub*j_inc;
    break;
  case 3:
    i_phy = i / i_len;
    i_sub = i % i_len;
    j_phy = j / j_len;
    j_sub = j % j_len;
    k_phy = k / k_len;
    k_sub = k % k_len;

    node = (int)(i_phy*MyExp2(i_off))
          +(int)(j_phy*MyExp2(j_off))
          +(int)(k_phy*MyExp2(k_off));

    subgrid_index = i_sub*i_inc
                  + j_sub*j_inc
                  + k_sub*k_inc;
    break;
  }

  switch (node) {
  case 0:
    for(n = 0; n < elem_size; n++) 
      buffer[n] = (p0+subgrid_index*elem_size)[n];
    break;
  case 1:
    for(n = 0; n < elem_size; n++) 
      buffer[n] = (p1+subgrid_index*elem_size)[n];
    break;
  case 2:
    for(n = 0; n < elem_size; n++) 
      buffer[n] = (p2+subgrid_index*elem_size)[n];
    break;
  case 3:
    for(n = 0; n < elem_size; n++) 
      buffer[n] = (p3+subgrid_index*elem_size)[n];
    break;
  }
}

void end_of_farry_elem_handler(int data1,int data2,int data3,int data4)
{
  info_buffer[MyNodeNumber*length+0] = data1;
  info_buffer[MyNodeNumber*length+1] = data2;
  info_buffer[MyNodeNumber*length+2] = data3;
  info_buffer[MyNodeNumber*length+3] = data4;
  farray_elem_flag = 1;
}

void read_farray_element_handler(int dest_node,int i,int j,int k)
{
  int n, loc;
  CMRT_desc_t d;
  int node, subgrid_index;
  int i_len, i_inc, i_off;
  int j_len, j_inc, j_off;
  int k_len, k_inc, k_off;
  int i_phy, i_sub;
  int j_phy, j_sub;
  int k_phy, k_sub;
  char *p0, *p1, *p2, *p3;
  int rank, elem_size, Coll_Id, index, offset;
  int buffer[4];

  /* decode the receiving buffer */

  Coll_Id   = info_buffer[dest_node*length+0];
  index     = info_buffer[dest_node*length+1];
  offset    = info_buffer[dest_node*length+2];
  elem_size = info_buffer[dest_node*length+3];
  rank      = info_buffer[dest_node*length+4];

  d = (CMRT_desc_t)*((long*)(TableOfCollection[Coll_Id]->Table[index]+offset));

  loc = CMRT_desc_get_cm_location(d); 
  p0 = (char*)TOGGLE_DPV_SPACE(CHANGE_DP(loc,DP_0));
  p1 = (char*)TOGGLE_DPV_SPACE(CHANGE_DP(loc,DP_1));
  p2 = (char*)TOGGLE_DPV_SPACE(CHANGE_DP(loc,DP_2));
  p3 = (char*)TOGGLE_DPV_SPACE(CHANGE_DP(loc,DP_3));

  switch (rank) {
  case 1:
    i_len = CMRT_desc_get_subgrid_dimension((char*)d,0);
    i_inc = CMRT_desc_get_subgrid_axis_increment((char*)d,0);
    i_off = CMRT_desc_get_offchip_position((char*)d,0);

    i_phy = i / i_len;
    i_sub = i % i_len;

    node = (int)(i_phy*MyExp2(i_off));
    subgrid_index = i_sub*i_inc;
    break;
  case 2:
    i_len = CMRT_desc_get_subgrid_dimension((char*)d,0);
    i_inc = CMRT_desc_get_subgrid_axis_increment((char*)d,0);
    i_off = CMRT_desc_get_offchip_position((char*)d,0);
    j_len = CMRT_desc_get_subgrid_dimension((char*)d,1);
    j_inc = CMRT_desc_get_subgrid_axis_increment((char*)d,1);
    j_off = CMRT_desc_get_offchip_position((char*)d,1);

    i_phy = i / i_len;
    i_sub = i % i_len;
    j_phy = j / j_len;
    j_sub = j % j_len;

    node = (int)(i_phy*MyExp2(i_off))
          +(int)(j_phy*MyExp2(j_off));
    subgrid_index = i_sub*i_inc
                  + j_sub*j_inc;
    break;
  case 3:
    i_len = CMRT_desc_get_subgrid_dimension((char*)d,0);
    i_inc = CMRT_desc_get_subgrid_axis_increment((char*)d,0);
    i_off = CMRT_desc_get_offchip_position((char*)d,0);
    j_len = CMRT_desc_get_subgrid_dimension((char*)d,1);
    j_inc = CMRT_desc_get_subgrid_axis_increment((char*)d,1);
    j_off = CMRT_desc_get_offchip_position((char*)d,1);
    k_len = CMRT_desc_get_subgrid_dimension((char*)d,2);
    k_inc = CMRT_desc_get_subgrid_axis_increment((char*)d,2);
    k_off = CMRT_desc_get_offchip_position((char*)d,2);

    i_phy = i / i_len;
    i_sub = i % i_len;
    j_phy = j / j_len;
    j_sub = j % j_len;
    k_phy = k / k_len;
    k_sub = k % k_len;

    node = (int)(i_phy*MyExp2(i_off))
          +(int)(j_phy*MyExp2(j_off))
          +(int)(k_phy*MyExp2(k_off));

    subgrid_index = i_sub*i_inc
                  + j_sub*j_inc
                  + k_sub*k_inc;
    break;
  }

  switch (node) {
  case 0:
    for(n = 0; n < elem_size; n++) 
      ((char*)buffer)[n] = (p0+subgrid_index*elem_size)[n];
    break;
  case 1:
    for(n = 0; n < elem_size; n++) 
      ((char*)buffer)[n] = (p1+subgrid_index*elem_size)[n];
    break;
  case 2:
    for(n = 0; n < elem_size; n++)
      ((char*)buffer)[n] = (p2+subgrid_index*elem_size)[n];
    break;
  case 3:
    for(n = 0; n < elem_size; n++) 
      ((char*)buffer)[n] = (p3+subgrid_index*elem_size)[n];
    break;
  }
  
  CMAML_reply(dest_node,end_of_farry_elem_handler,
	      buffer[0], buffer[1], buffer[2], buffer[3]);
  return;
}

int AskForFArrayElem(int where,int index,int Coll_Id,int offset,char *buffer,
		     int elem_size,int rank,
		     int i_len, int i_inc, int i_off,
		     int j_len, int j_inc, int j_off,
		     int k_len, int k_inc, int k_off,
		     int i,int j,int k)
{
  int start_at, hops, n;

  info_buffer[MyNodeNumber*length+0] = Coll_Id;
  info_buffer[MyNodeNumber*length+1] = index;
  info_buffer[MyNodeNumber*length+2] = offset; 
  info_buffer[MyNodeNumber*length+3] = elem_size; 
  info_buffer[MyNodeNumber*length+4] = rank; 

  CMAML_disable_interrupts();

  hops = 3;
  start_at = 0;

  for(n = 0; n < hops; n++) {
    buffer_updated_flag = 0;
    CMAML_request(where, remote_store_handler, MyNodeNumber, start_at, 
		  info_buffer[MyNodeNumber*length+2*n], 
		  info_buffer[MyNodeNumber*length+2*n+1]);
    CMAML_poll_wait(&buffer_updated_flag,1);
    start_at += 2;
  }

  /* read farray element */
  farray_elem_flag = 0;
  CMAML_request(where, read_farray_element_handler, MyNodeNumber, i, j, k);
  CMAML_poll_wait(&farray_elem_flag, 1);

  for(n = 0; n < elem_size; n++) 
    buffer[n] = ((char*)(info_buffer+MyNodeNumber*length))[n];

  CMAML_enable_interrupts();
  return 1;
}

/***************************************************************************
  The following handles memory allocation and memory access on vector 
  units.

  Shelby Yang, Indiana University.  (10/25/93)
  **************************************************************************/
  
char *allocate_vu_array(int *extents,int *weights,
			int *axis_order,int rank,int size,int type)
{
    CMRT_desc_t desc;
    CMRT_array_geometry_t geometry;

    /* Initialize RTS if not already */
    if(!CMRT_initialized()) CMRT_init();

    /* Create a geometry and descriptor, then allocate memory.  The array
       layout is given by axis_order. */
    geometry = CMRT_intern_detailed_geometry(extents,weights,axis_order,
					     rank,size);
    switch (type) {
    case 0:
      desc = CMRT_allocate_heap_array(geometry, CMRT_double_float);
      break;
    case 1:
      desc = CMRT_allocate_heap_array(geometry, CMRT_float);
      break;
    case 2:
      desc = CMRT_allocate_heap_array(geometry, CMRT_double_complex);
      break;
    case 3:
      desc = CMRT_allocate_heap_array(geometry, CMRT_complex);
      break;
    default:
      Message("allocate_vu_array(): Wrong data type",FATAL);
      break;
    }

    /* done! */
    return (char*) desc;
}

void deallocate_vu_array(char *array) 
{
  if(array != 0) {
    CMRT_deallocate_heap_array((CMRT_desc_t)array);
  }
}

char * get_address_on_vu_0(unsigned loc) 
{
  return (char *) TOGGLE_DPV_SPACE(CHANGE_DP(loc,DP_0));
}

char * get_address_on_vu_1(unsigned loc) 
{
  return (char *) TOGGLE_DPV_SPACE(CHANGE_DP(loc,DP_1));
}

char * get_address_on_vu_2(unsigned loc) 
{
  return (char *) TOGGLE_DPV_SPACE(CHANGE_DP(loc,DP_2));
}

char * get_address_on_vu_3(unsigned loc) 
{
  return (char *) TOGGLE_DPV_SPACE(CHANGE_DP(loc,DP_3));
}

unsigned get_cm_location(char *desc)
{
  return CMRT_desc_get_cm_location((CMRT_desc_t)desc);
}

int get_subgrid_dimension(char *desc, int i)
{
  return CMRT_desc_get_subgrid_dimension((CMRT_desc_t)desc,i);
}

int get_offchip_position(char *desc, int i)
{
  return CMRT_desc_get_offchip_position((CMRT_desc_t)desc,i);
}

int get_subgrid_axis_increment(char *desc, int i)
{
  return CMRT_desc_get_subgrid_axis_increment((CMRT_desc_t)desc,i);
}

int get_subgrid_size(char *desc)
{
  return CMRT_desc_get_subgrid_size((CMRT_desc_t)desc);
}

/**************************************************************************
  Remote execution related functions and handlers

  Shelby Yang (3/24/94)
  *************************************************************************/
extern void RemoteFunctionList(void *element_address,int function_number,
			       void *arg_buffer);

void buffer_updated_handler()
{
  buffer_updated_flag = 1;
}

void end_of_remote_handler(int tag_location)
{
  /* 
    this handler is called on the processor that requests remote
    execution.  in the future this handler will also handle the return 
    value of a remote function. 
  */
  int *p;
  if(tag_location != 0) {
    p = (int*)tag_location;
    (*p)++;
  }
  remote_execute_flag = 1;
}

void remote_store_handler(int dest_node,int start_at,int buffer1,int buffer2)
{
  info_buffer[dest_node*length+start_at] = buffer1;
  info_buffer[dest_node*length+start_at+1] = buffer2;
  CMAML_reply(dest_node,buffer_updated_handler);
}

/***************************************************************************
   tmp1 and tmp2 are buffers for saving the states of Sparc floating point 
   registers and vector unit registers.  We MAY also need to save the VU data
   register by calling CMAML_save_dataregs() and CMAML_restore_dataregs(),
   but we need to figure out how to allocate parallel memory and pass its
   pointer to these functions first.  Shelby Yang (9/21/94) 
 ****************************************************************************/
double tmp1[17], tmp2[13];

void remote_execute_handler(int dest_node, int Col_Id, int element_index)
{
  int function_number, tag_location;
  char *p;

  /* save the states of floating point registers and vu registers */
  CMAML_save_fpuregs(tmp1);
  CMAML_save_ctrlregs(tmp2);  

  /* decode the receiving buffer */
  function_number = info_buffer[dest_node*length+0];
  tag_location    = info_buffer[dest_node*length+1];

  p = (char*)&info_buffer[dest_node*length+2];
  /* call compiler generated function list */
  RemoteFunctionList((void*)TableOfCollection[Col_Id]->Table[element_index], 
		     function_number, (void*) p);

  /* reply to indicate remote execution is completed */
  CMAML_reply(dest_node, end_of_remote_handler, tag_location);
  
  /* restore the states of floating point registers and vu registers */
  CMAML_restore_fpuregs(tmp1);
  CMAML_restore_ctrlregs(tmp2);
}

void AskForRemoteExecute(int CollectionNumber,int element_index,
			 int function_number,char *arg_buffer,int size,
			 long tag_location)
{
  /* Request for remote execution. size is the size of arg_buffer in bytes.
    
     Warning: The lesson learned on CM5 is that do not poll in a handler.
              The machine hangs when many messages are sent.  
   */ 

  int where, hops, start_at, i;
  int *p;

  if(size > length*sizeof(int)) {
    Message("AskForRemoteExecute(): Requires a too large buffer\n",FATAL);
  }

  where = TableOfCollection[CollectionNumber]->Where[element_index];

  if(where == MyNodeNumber) {
    /* if element is local, execute the function */
    CMAML_disable_interrupts();
    RemoteFunctionList((void*)TableOfCollection[CollectionNumber]->
		       Table[element_index], 
		       function_number, (void*)arg_buffer);
    if(tag_location != 0) {
      p = (int*)tag_location;
      (*p)++;
    }
    CMAML_enable_interrupts();
    return;
  }
    
  CMAML_disable_interrupts();
  /* copy to info_buffer in case arg_buffer size is not evenly divisible 
     by 2*sizeof(int) */
  memcpy((char*)&info_buffer[MyNodeNumber*length],arg_buffer,size);

  /* send function_number and tag_location */
  start_at = 0;
  buffer_updated_flag = 0;
  CMAML_request(where, remote_store_handler, MyNodeNumber, start_at, 
		function_number, (int)tag_location);
  CMAML_poll_wait(&buffer_updated_flag,1);

  hops = size/(2*sizeof(int));
  if(size%(2*sizeof(int)) != 0) hops++;

  /* send the argument buffer */
  start_at = 2;
  for(i = 0; i < hops; i++) {
    buffer_updated_flag = 0;
    CMAML_request(where, remote_store_handler,
		  MyNodeNumber, start_at, info_buffer[MyNodeNumber*length+2*i],
		  info_buffer[MyNodeNumber*length+2*i+1]);

   /* while(buffer_updated_flag != 1) CMAML_poll(); */
    CMAML_poll_wait(&buffer_updated_flag,1);
    start_at += 2;
  }

  /* remote procedure call */
  remote_execute_flag = 0;
  CMAML_request(where, remote_execute_handler, MyNodeNumber,
		CollectionNumber, element_index);
  CMAML_poll_wait(&remote_execute_flag,1);

  CMAML_enable_interrupts();
}

