/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/****************************************************************
 *                                                              *
 *         pC++ kernel classes and SuperKernel collection       *
 *         This file should be included only after are          *
 *         element classes defined.                             *
 *                                                              *
 *         Originally written by Francois Bodin                 *
 *         Modified often by:                                   *
 *         Pete Beckman, Bernd Mohr, Shelby Yang                *
 *         Computer Science, Indiana University                 *
 *                                                              *
 ****************************************************************/

#include "GlobalStruct.h"
#include "profiler.h"
#include "pcxx_events_def.h"

extern "C" {
  void pcxx_ReduceDoubleAdd(void *buffer, int elem_count);
  void pcxx_ReduceFloatAdd(void *buffer, int elem_count);
  void pcxx_ReduceIntAdd(void *buffer, int elem_count);
  void pcxx_ReduceDoubleMax(void *buffer, int elem_count);
  void pcxx_ReduceFloatMax(void *buffer, int elem_count);
  void pcxx_ReduceIntMax(void *buffer, int elem_count);
  void pcxx_ReduceDoubleMin(void *buffer, int elem_count);
  void pcxx_ReduceFloatMin(void *buffer, int elem_count);
  void pcxx_ReduceIntMin(void *buffer, int elem_count);
}

// Note that "brkRuntime.h" & "brkExecDef.h" are included later in this file.
// They must be included after the class Kernel declaration.

/*******************************************************************/
/* Global Variable to ALL Run-time environments */

int pcxx_MsgBlockSize =  512;  /* size in INTS */
int pcxx_MaxMsgSize   = 16384;  /* size in INTS */
/*******************************************************************/

class Processors{
  public :
    int dimension;
};

class Distribution {
 public:
   int nbdim;
   int nbelement;
   int dim1size, dim2size, dim3size;
   int dist1, dist2, dist3;
   Processors P;
   Distribution();
   Distribution(int d1size, Processors *P, int dt1);
   Distribution(int d1size, int d2size, Processors *P, int dt1, int dt2);
   Distribution(int d1size, int d2size, int d3size, Processors *P, int dt1, 
	    int dt2, int dt3);
   int GiveElementProc(int i);
   int GiveElementProc(int i, int j);
   int GiveElementProc(int i, int j, int k);
};

Distribution::Distribution() {
  nbdim =1;
  nbelement = pcxx_TotalNodes();
  dim1size = pcxx_TotalNodes();
  dist1 = CYCLIC;
}

Distribution::Distribution(int d1size, Processors *, int dt1) {
  nbdim =1;
  nbelement = d1size;
  dim1size = d1size;
  dist1 = dt1;
}

Distribution::Distribution(int d1size, int d2size, Processors *, int dt1, int dt2) {
  nbdim =2;
  nbelement = d1size*d2size;
  dim1size = d1size;
  dim2size = d2size;
  dist1 = dt1;
  dist2 = dt2;
}

Distribution::Distribution(int d1size, int d2size, int d3size, Processors *, 
		   int dt1, int dt2, int dt3) {
  nbdim =2;
  nbelement = d1size*d2size*d3size;
  dim1size = d1size;
  dim2size = d2size;
  dim3size = d3size;
  dist1 = dt1;
  dist2 = dt2;
  dist3 = dt3;
  Message("Distribution(): Sorry, 3D template not implemented.",FATAL);
}
     
int Distribution::GiveElementProc(int i) {
  int proc = 0;
  int p, rem, q;
  switch (dist1) {
    case BLOCK:
/**************************************************************************
Notes from Suresh Srinivas (6/23/93):

New version of Block distribution, tries to keep the allocation
of elements to processors balanced. It tries to divide the data-structure
into contiguous blocks whose sizes are identical, except for the last
block which may be of a smaller size.  

Let `dimsize' be the size of a distributee in a certain dimension
and let `Numproc' be the number of processors in that dimension.
[Note: Integer division (i.e x/y where x and y are integers) gives the
same result as ceil((x*1.0)/(y*1.0))]

Let p   = dimsize/Numproc
    rem = dimsize - Numproc * (dimsize/Numproc)
    q = if (p <= rem) p + 1  else  p

The distribution function `delta' on the index space in that dimension is
delta(i) = { j } for all i, 0 <= i <= (dimsize - 1)
            where  j = i/q
            (note if (j >= Numproc) then j = (Numproc - 1) - j % Numproc
[Note: When j exceeds Numproc; it is better to keep j close to the computed
value by setting j to (Numproc - 1) - j % Numproc instead of trying
j = j % Numproc. The reason being that the former promotes better locality]

Example:
a) dimsize = 32, Numproc = 4.
   p = 8, rem = 0, q = 8.
   Each processor gets 8 elements.
b) dimsize = 33, Numproc = 4.
   p = 8, rem = 1, q = 8.
   Each processor gets 8 except one processor which gets 9.
c) dimsize = 99, Numproc = 50
   p = 1, rem = 49, q = 2.
   Each processor gets 2 except one processor which gets 1.

But beware, when dimsize = 8 and Numproc = 6, each of four processors will 
get 2 and two processors will get nothing.   So don't assume that if n > p 
every processor will get something.   --- Kate.

****************************************************************************/
      p = dim1size/pcxx_TotalNodes();
      rem = dim1size - pcxx_TotalNodes() * p;
      q = (p <= rem) ? p+1 : p;
      proc = i/q;
      if (proc >= pcxx_TotalNodes()) proc = (pcxx_TotalNodes() - 1) - proc % pcxx_TotalNodes();
      break;
    case CYCLIC:
      proc = (i % pcxx_TotalNodes());
      break;
    case WHOLE:
      Message("GiveElementProc()_1D: WHOLE has no meaning in 1D Distribution,",
	      WARNING);
      Message("GiveElementProc()_1D: use BLOCK distribution instead.",
	      WARNING);
      p = dim1size/pcxx_TotalNodes();
      rem = dim1size - pcxx_TotalNodes() * p;
      q = (p <= rem) ? p+1 : p;
      proc = i/q;
      if (proc >= pcxx_TotalNodes()) proc = (pcxx_TotalNodes() - 1) - proc % pcxx_TotalNodes();
      break;
    default:
      Message("GiveElementProc()_1D: Unknow distribution for 1D Distribution.", 
	      FATAL);
    }

  if (proc >= pcxx_TotalNodes()) {
    Message("GiveElementProc()_1D: Error in 1D Distribution object.",WARNING);
    proc = proc %pcxx_TotalNodes();
  }

  return proc;
}

int Distribution::GiveElementProc(int i, int j) {
  int proc = 0;
  int index1,index2, sq, p, rem, q;
  float x;

  if ((dist1 == QBLOCK) && (dist2 == QBLOCK)) {
    q = (int) QGridToProc(i,j,dim1size);
    x = dim1size*dim2size/ (float) pcxx_TotalNodes();
    proc = (int) (q/x);
    if (proc >= pcxx_TotalNodes()) {
      Message("GiveElementProc()_2D: Processor # Out of Bounds)", 
	      WARNING);
      proc = proc %pcxx_TotalNodes();
    }
    return proc;
  }
  
  sq = (int) sqrt((double)  pcxx_TotalNodes());
  if((dist1 == WHOLE) && (dist2 != WHOLE)) sq = pcxx_TotalNodes(); 
  if((dist1 != WHOLE) && (dist2 == WHOLE)) sq = pcxx_TotalNodes(); 
  if((dist1 == WHOLE) && (dist2 == WHOLE)) sq = 1;

  switch (dist1) {
    case BLOCK:
      p = dim1size/sq;
      rem = dim1size - sq * p;
      q = (p <= rem) ? p+1 : p;
      index1 = i/q;
      if (index1 >= sq) index1 = (sq - 1) - index1 % sq;
      break;
    case CYCLIC:
      index1 = (i % sq);
      break;
    case WHOLE:
      index1 = 0;
      break;
    default:
      Message("GiveElementProc()_2D: Unknow distribution for 2D Distribution.", 
	      FATAL);
    }
  switch (dist2) {
    case BLOCK:
      p = dim2size/sq;
      rem = dim2size - sq * p;
      q = (p <= rem) ? p+1 : p;
      index2 = j/q;
      if (index2 >= sq) index2 = (sq - 1) - index2 % sq;
      break;
    case CYCLIC:
      index2 = (j % sq);
      break;
    case WHOLE:
      index2 = 0;
      break;
    default:
      Message("GiveElementProc()_2D: Unknow distribution for 2D Distribution.",
	      FATAL);
    }
  
  if (dist1==WHOLE)
    proc = index2;
  else proc = index1 + index2*(sq);

  if (proc >= pcxx_TotalNodes()) {
    Message("GiveElementProc()_2D: Error in 2D Distribution object.", WARNING);
    proc = proc %pcxx_TotalNodes();
  }
  
  return proc;
}

int Distribution::GiveElementProc(int, int, int) {
  Message("GiveElementProc()_3D: Sorry, 3D template not implemented.",FATAL);
  return 0;
}

class Align {
  public:
    int nbelement;
    int nbdim;
    int dim1size,dim2size,dim3size;
    int ret_index1;
    int ret_index2;
    int ret_index3;
    int chosen;
    Align(int, int, char *) {
      /* fool the compiler, the real one is the next */
    };
    /* the real one */
    Align(int n1, int n2, int choix) {
      nbdim = 2;
      dim1size = n1;
      dim2size = n2;
      nbelement = n1*n2;
      chosen = choix;
    };
    Align(int n1, int choix) {
      nbdim = 1;
      dim1size = n1;
      dim2size = 0;
      nbelement = n1;
      chosen = choix;
    };
    void align_func(int var_i, int var_j); /* avoid shadow */
};

void Align::align_func(int varZZ_i, int varZZ_j)
{
  /* Set ret_index1 and ret_index2 depending on dep2C+
     generated align function (PHB) */
  GenAlignFunc(chosen,varZZ_i,varZZ_j,&ret_index1,&ret_index2);
}

class Kernel {
  public:
    int **ElementLink;  
    Distribution *distribution;
    Align * align;
    struct INFO_HEAD CollectionInfo;
    int CollectionNumber;
    int CollectionNumber_Type; 
    int *buffer;
    Kernel(Distribution *T,  Align *A, int sizelem);
    ~Kernel();
    int Is_Local(int index); 
    int Is_Local(int index1, int index2); 
    int ElemIndex(int i);   
    int ElemIndex(int i, int j);   
    int Get_2DHighIndex(int index); 
    int Get_2DLowIndex(int index); 
    int AllocatedProc(int index);
    int AllocatedProc(int index1, int index2);
    int Install_Collection(Distribution *T, Align *A, int sizeelem);
    int *Get_Element(int index);
    int *Get_Element(int index1, int index2 );
    int *Get_ElementPart(int index,int startof, int size);
    int *Get_ElementPart(int index1, int index2,int startof, int size);
    int Kernel::Get_ElementPart_Int(int index1, int startof);
    float Kernel::Get_ElementPart_Float(int index1, int startof);
    double Kernel::Get_ElementPart_Double(int index1, int startof);
    int Kernel::Get_ElementPart_Int(int index1, int index2, int startof);
    float Kernel::Get_ElementPart_Float(int index1, int index2, int startof);
    double Kernel::Get_ElementPart_Double(int index1, int index2, int startof);
    int *Get_CopyElem(int index);
    int *Get_CopyElem(int index1, int index2 );
    void Fetch_Element(int index,int *buf,int startof, int size);
};


//  Including breakpoint stuff here...;
#include "breezyInc.h"

int  Kernel::AllocatedProc(int index)
{
  // Assign the indexth element to a processor according to the
  // given template and alignment 
  int proc = 0;
  align->align_func(index, 0);
  if(distribution->nbdim == 1)
    proc = distribution->GiveElementProc(align->ret_index1);
  else
    proc = distribution->GiveElementProc(align->ret_index1,align->ret_index2);
  return proc;
}

int Kernel::AllocatedProc(int index1, int index2)
{
  int proc = 0;
  align->align_func(index1, index2);
  proc = distribution->GiveElementProc(align->ret_index1, align->ret_index2);
  return proc;
}

int Kernel::ElemIndex(int index)
{
  return index; 
}

int Kernel::ElemIndex(int index1, int index2)
{
  return (index1+index2*CollectionInfo.dim1size); 
}

int Kernel::Get_2DHighIndex(int index)
{
  return (index/CollectionInfo.dim1size);
}

int Kernel::Get_2DLowIndex(int index)
{
  return (index%CollectionInfo.dim1size);
}

int Kernel::Is_Local(int index)
{
#ifdef _MASTER_SLAVE_
// _EVERYTHING!_ is local to main thread
  if (IamtheMainThread()) return 1;
#endif
  if(CollectionInfo.Where[ElemIndex(index)] != pcxx_NodeNumber())
    return 0;
  else
    return 1; 
}

int Kernel::Is_Local(int index1, int index2)
{
#ifdef _MASTER_SLAVE_
// _EVERYTHING!_ is local to main thread
  if (IamtheMainThread()) return 1;
#endif
  if(CollectionInfo.Where[ElemIndex(index1,index2)] != pcxx_NodeNumber())
    return 0;
  else
    return 1;
}

Kernel :: Kernel(Distribution *T, Align *A, int sizelem) 
{
  /* set breakpointing off until collection is initialized */
  brk_TOGGLE_BREAKPOINT(0);
  if (!T || !A)
    Message("Kernel(): No alignement or template provided.",FATAL);
  distribution = T;
  align = A;
  if(A->nbelement > T->nbelement) {
    Message("Kernel(): Elements in Align are more than elements in Distribution.",
	        WARNING);
  }
  CollectionInfo.type = -1; /* undefined */
  CollectionInfo.number_dimension = A->nbdim; /* the default */
  CollectionInfo.element_size = sizelem;
  CollectionInfo.nb_element = A->nbelement;
  CollectionInfo.dim1size  = A->dim1size;
  CollectionInfo.dim2size  = A->dim2size;
  CollectionInfo.dim3size  = A->dim3size;
  if (!Install_Collection(T,A,sizelem))
    Message("Kernel(): Error in creating collection.",FATAL);
  pcxx_Barrier();
  brk_TOGGLE_BREAKPOINT(1);
}


/* BEGIN MACHINE DEPENDANT SECTION */

#if defined(_MASTER_SLAVE_)
   /* no extra  master/slave runtime system */
   /* use shared memory runtime system definitions */
#  include "kernel-SMRTS.h"
#else
#if defined(_SMRTS_)
   /* use shared memory runtime system definitions for */
   /* BBN TC2000, Sequent Symmetry, KSR1, SGIMP        */
#  include "kernel-SMRTS.h"
#else
   /* cm5, paragon, sp1, pvm3.3                                    */
   /* uniprocessors are regarded as 1-node distributed systems :-) */
#  include "kernel-DMRTS.h"
#endif
#endif

/* END MACHINE DEPENDANT FUNCTIONS */

/* START NON-MACHINE DEPENDANT FUNCTIONS */
/* non-class functions */

int GetFirstLocalElementIndex(Kernel *p)
{
  Kernel *pt;
  int i;
  pt = (Kernel *) p;

#ifdef _MASTER_SLAVE_
// To the main thread, the first local element is always 0
  if (IamtheMainThread()) return 0;
#endif

  for (i = 0; i < pt->CollectionInfo.nb_element; i++) {
    if (pt->CollectionInfo.Where[i] == pcxx_NodeNumber()) return i;
  }   
  return -1;
}

void ResetLocal(Kernel *p)
{
  p->CollectionInfo.LocalTraversalCount = 0;
  SetNextCollectionTraversalLevel((char *) p);
}

#ifdef _MASTER_SLAVE_
int Local(int elem)
#else
int Local(int)
#endif
{
  Kernel *pt;
  int i, node_number;
  pt = (Kernel *) GetCollectionTraversalLevel();

/* This was commented out of the original MS code */
  if (pt->CollectionInfo.LocalTraversalCount > pt->CollectionInfo.nb_element) {
    DecCollectionTraversalLevel();
    return -1;
  }

  node_number = pcxx_NodeNumber();

#ifdef _MASTER_SLAVE_
  for (i = elem+1; i < pt->CollectionInfo.nb_element; i++) {
    if (pt->CollectionInfo.Where[i] == node_number) {
#else
  for (i = pt->CollectionInfo.LocalTraversalCount; i < pt->CollectionInfo.nb_element;
       i++) {
    if (pt->CollectionInfo.Where[i] == node_number) {
      pt->CollectionInfo.LocalTraversalCount = i+1;
#endif
      return i;
    }
  }   
  DecCollectionTraversalLevel();
  return -1;
}

int  FirstLocal ()
{
  return Local(-1);    // -1 instead of 0 (only used by MS anyway)
}


/* functions to deal with subsets for a collection with one dimension */

int FirstInSubset(Kernel *coll, int lb, int ub, int)
{
  if (!coll)
    return -1;
  if (lb == -1)
    return 0;
  if (lb <= ub)
    return lb;
  return -1;
}

int NextInSubset(Kernel *coll, int, int ub, int step, int var)
{
  int bsup;

  if (!coll)
    return -1;

  if (ub == -1)
    bsup = coll->CollectionInfo.dim1size;
  else 
    bsup = ub;

  if (var + step <= bsup)
    return var + step;
  else
    return -1;
}

CheckLocal(Kernel *coll, int index)
{
  if (!coll)
    return 0;
  
  return coll->Is_Local(index);
}

/**************************************************************************
   A kernel collection, SuperKernel, from which all new collections should 
   be derived.  It can also be used directly since basic element access and
   basic element management are provided.  Based on Francios Bodin's code
   in DistributedArray with modifications.

   Shelby Yang, Computer Science, Indiana University, 6/22/93.
***************************************************************************/
#ifdef _SMRTS_
#include <stddef.h>
#endif

Collection SuperKernel : public Kernel {
 public:
  int dim1size;
  int dim2size;
  int dimensions;
  SuperKernel(Distribution *T, Align *A);
  ~SuperKernel();
  ElementType  *operator()(int i1);           //pointer to a collection element
  ElementType  *operator()(int i1, int i2);   //pointer to a collection element

  void BroadcastBufferToField(int proc_id, char *buffer,int offset,int size);
  void BroadcastField(int index,int offset,int size);
  void BroadcastField(int index1,int index2,int offset,int size);

  void ReduceDoubleFieldAdd(int offset,int elem_count);
  void ReduceFloatFieldAdd(int offset,int elem_count);
  void ReduceIntFieldAdd(int offset,int elem_count);

  void ReduceDoubleFieldMax(int offset,int elem_count);
  void ReduceFloatFieldMax(int offset,int elem_count);
  void ReduceIntFieldMax(int offset,int elem_count);

  void ReduceDoubleFieldMin(int offset,int elem_count);
  void ReduceFloatFieldMin(int offset,int elem_count);
  void ReduceIntFieldMin(int offset,int elem_count);

  void RemoteExecute(int index,int function_number,char *arg_buffer,int size,
		     int *tag_location);
  void RemoteExecute(int index,int function_number,char *arg_buffer,int size);
  void RemoteExecute(int index1,int index2,int function_number,
		     char *arg_buffer,int size,int *tag_location);
  void RemoteExecute(int index1,int index2,int function_number,
		     char *arg_buffer,int size);
 MethodOfElement:
  int Ident;
  int index1;
  int index2;
  SuperKernel<ElementType> *ThisCollection;
  ElementType *Self(int i);       //copy an element with an offset i 
  ElementType *Self(int i, int j);//copy an element with an offset i,j
#ifdef _SMRTS_
  void * operator new(size_t size) {
    return (void *)tulip_SharedMalloc(size);
  }
  void operator delete(void *p, size_t ) {
    tulip_SharedFree(p);
  }
#endif
};

SuperKernel::SuperKernel(Distribution *T, Align *A)
                        :Kernel(T, A, (int)sizeof(ElementType))
{
  int i, j;

  /* set breakpointing off until collection is initialized */
  brk_TOGGLE_BREAKPOINT(0);
  if (!T || !A)
    Message("No alignment or template for collection initialization.",FATAL);
  dim1size = A->dim1size;
  dim2size = A->dim2size;
  dimensions = A->nbdim;

  if (A->nbdim == 1) {
    for (i = 0;  i < dim1size; i++) {
      if (Is_Local(i)) {
        ElementLink[ElemIndex(i)] = (int*) new ElementType;
      } 
    }
    pcxx_Barrier();
    for(i = 0; i < dim1size; i++) {
      CollectionInfo.Table[ElemIndex(i)] = (int*) ElementLink[ElemIndex(i)];
      if (Is_Local(i)) {
	(*this)(i)->index1 = i;
	(*this)(i)->index2 = 0;
	(*this)(i)->ThisCollection = (SuperKernel<ElementType> *) this;
	(*this)(i)->Ident = this->CollectionNumber*16777216+i; 
      }
    }
    pcxx_Barrier();
  }

  if (A->nbdim == 2) {
    for (i = 0;  i < dim1size; i++) {
      for (j = 0;  j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  ElementLink[ElemIndex(i,j)] = (int*) new ElementType;
	}
      }
    }
    pcxx_Barrier();
    for (i = 0;  i < dim1size; i++) {
      for (j = 0;  j < dim2size; j++) {
	CollectionInfo.Table[ElemIndex(i,j)] = (int*) ElementLink[ElemIndex(i,j)];
        if (Is_Local(i,j)) {
          (*this)(i,j)->index1 = i;
          (*this)(i,j)->index2 = j;
          (*this)(i,j)->ThisCollection = (SuperKernel<ElementType> *) this;
          (*this)(i,j)->Ident = this->CollectionNumber*16777216+i+j*this->dim1size;
        } 
      }
    }
    pcxx_Barrier();
  }

  if (A->nbdim >= 3)
    Message("3D and higher-dimensional collection not implemented.",FATAL);

#ifdef _SMRTS_
  if(pcxx_NodeNumber() == 0) SMRTSDealloc(ElementLink);
#else
  delete [] ElementLink;
#endif
  brk_TOGGLE_BREAKPOINT(1);

}

SuperKernel::~SuperKernel() 
{
  int i, j;
  if (dimensions == 1) {
    for (i = 0;  i < dim1size; i++) {
      if (AllocatedProc(i) == pcxx_NodeNumber()) {
        delete (ElementType*) CollectionInfo.Table[ElemIndex(i)];
      } 
    }
  }

  if (dimensions == 2) {
    for (i = 0;  i < dim1size; i++) {
      for (j = 0;  j < dim2size; j++) {
        if (AllocatedProc(i, j) == pcxx_NodeNumber()) {
          delete (ElementType*) CollectionInfo.Table[ElemIndex(i,j)];
        } 
      }
    }
  }

  if (dimensions >= 3)
    Message("3D and higher-dimensional collection not implemented.",FATAL);

  brk_DELETE_COLLECTION_ENTRY((void *)this);
}

ElementType * SuperKernel :: operator()(int i1)
{
  ElementType  *pt;
  pt = (ElementType  *) Get_Element(i1);
  return pt;
}

ElementType * SuperKernel :: operator()(int i1, int i2)
{
  ElementType  *pt;
  pt = (ElementType  *) Get_Element(i1,i2);
  return pt;
}

ElementType * SuperKernel :: Self(int i)
{
  return (ElementType  *) ThisCollection->Get_Element(index1 + i);
}

ElementType * SuperKernel :: Self(int i, int j)
{
  return (ElementType  *) ThisCollection->Get_Element(index1 + i, index2 + j);
};

void SuperKernel::RemoteExecute(int index,int function_number,
				char *arg_buffer,int size)
{
  /* the arguments for the remote function should be copied into the
     arg_buffer. index is the index of the element. function_number
     is the number assigned to the function that will be executed. size
     is the size of arg_buffer in bytes.
   */

  int element_index = ElemIndex(index);

  AskForRemoteExecute(CollectionNumber,element_index,function_number,
		      arg_buffer,size,(long)0);
}

void SuperKernel::RemoteExecute(int index,int function_number,
				char *arg_buffer,int size,int *tag_location)
{
  int element_index = ElemIndex(index);

  AskForRemoteExecute(CollectionNumber,element_index,function_number,
		      arg_buffer,size,(long)tag_location);
}

void SuperKernel::RemoteExecute(int index1,int index2,int function_number,
				char *arg_buffer,int size)
{
  /* the arguments for the remote function should be copied into the
     arg_buffer. index1 and index2 are the indices of the element. 
     function_number is the number assigned to the function that will be 
     executed. size is the size of arg_buffer in bytes.
   */

  int element_index = ElemIndex(index1,index2);

  AskForRemoteExecute(CollectionNumber,element_index,function_number,
		      arg_buffer,size,(long)0);
}

void SuperKernel::RemoteExecute(int index1,int index2,int function_number,
				char *arg_buffer,int size,int *tag_location)
{
  int element_index = ElemIndex(index1,index2);

  AskForRemoteExecute(CollectionNumber,element_index,function_number,
		      arg_buffer,size,(long)tag_location);
}


void SuperKernel::BroadcastBufferToField(int proc_id, char *buffer,int offset,
int size)
{
  /************************************************************************ 
    This function broadcasts a buffer located on processor "proc_id" to the
    element class field which is offset by "offset" bytes from the beginning 
    of the element class.  "size" is the size of the buffer given in bytes.
   *************************************************************************/
  int i, j, k, flag;
  char *p;

  flag = 1;
  if(pcxx_NodeNumber() == proc_id) flag = 0;

  pcxx_BroadcastBytes(flag, size, buffer);

  if(dimensions == 1) {
    for(i = 0; i < dim1size; i++) {
      if(Is_Local(i)) {
	p = ((char *)(*this)(i)) + offset;
	for(j = 0; j < size; j++) p[j] = buffer[j];
      }
    }
  }

  if(dimensions == 2) {
    for(i = 0; i < dim1size; i++) {
      for(j = 0; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  p = ((char *)(*this)(i,j)) + offset;
	  for(k = 0; k < size; k++) p[k] = buffer[k];
	}
      }
    }
  }

}

void SuperKernel::BroadcastField(int index,int offset,int size)
{
  /************************************************************************ 
    This function broadcasts a field of an element to the same field of all 
    other elements.  "index" denotes the element that broadcasts.
    The field is offset from the beginning of the element by "offset" bytes.
    "size" is the size of the field in bytes.
   *************************************************************************/
  int i, j;
  char *p, *buffer;
  int flag = 1;

  buffer = new char[size];

  if(Is_Local(index)) {
    /* this processor will broadcast.  set flag to zero. */
    flag = 0;
    p = ((char *)(*this)(index)) + offset;
    for(j = 0; j < size; j++) buffer[j] = p[j]; 
  }
  pcxx_Barrier();

  pcxx_BroadcastBytes(flag, size, buffer);

  for(i = 0; i < dim1size; i++) {
    if(Is_Local(i) && i != index) {
      p = ((char *)(*this)(i)) + offset;
      for(j = 0; j < size; j++) p[j] = buffer[j];
    }
  } 
  
  delete [] buffer;
}

void SuperKernel::BroadcastField(int index1,int index2,int offset,int size)
{
  /************************************************************************ 
    This function broadcasts a field of an element to the same field of all 
    other elements.  "index1" and "index2" denotes the element that broadcasts.
    The field is offset from the beginning of the element by "offset" bytes.
    "size" is the size of the field in bytes.
   *************************************************************************/
  int i, j, k;
  char *p, *buffer;
  int flag = 1;

  buffer = new char[size];

  if(Is_Local(index1,index2)) {
    /* this processor will broadcast.  set flag to zero. */
    flag = 0;
    p = ((char *)(*this)(index1,index2)) + offset;
    for(j = 0; j < size; j++) buffer[j] = p[j]; 
  }

  pcxx_BroadcastBytes(flag, size, buffer);

  for(i = 0; i < dim1size; i++) {
    for(j = 0; j < dim2size; j++) {
      if(Is_Local(i,j) && i!=index1 && j!=index2) {
	p = ((char *)(*this)(i,j)) + offset;
	for(k = 0; k < size; k++) p[k] = buffer[k];
      }
    }
  } 
  
  delete [] buffer;
}

void SuperKernel::ReduceDoubleFieldAdd(int offset,int elem_count)
{
  /**************************************************************************
   Suppose a field of an element class is offset from the beginning of the
   class by "offset" bytes.  This function computes the sum of the field
   from all elements in the collection.  The sum is broadcast back to each 
   element and stored in the same field.  The old content of the field is 
   overwritten.  If the field is an array, corresponding array elements are 
   summed together.  Note "offset" is in bytes and "elem_count" is the number 
   of array elements.
   ***************************************************************************/
  int i, j, k;
  double *p, *buffer;

  buffer = new double[elem_count];

  for(i = 0; i < elem_count; i++) buffer[i] = 0.0;
 
  if(dimensions == 1) {
    for(i = 0; i < dim1size; i++) {
      if(Is_Local(i)) {
	p = (double*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) buffer[j] += p[j];
      }
    }
  } 

  if(dimensions == 2) {
    for(i = 0; i < dim1size; i++) {
      for(j = 0; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  p = (double*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) buffer[k] += p[k];
	}
      }
    }
  }


  /* reduce from and broadcast to processors */ 
  pcxx_ReduceDoubleAdd((void*)buffer,elem_count);

  if(dimensions == 1) {
    for(i = 0; i < dim1size; i++) {
      if(Is_Local(i)) {
	p = (double*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) p[j] = buffer[j];
      }
    }
  }

  if(dimensions == 2) {
    for(i = 0; i < dim1size; i++) {
      for(j = 0; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  p = (double*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) p[k] = buffer[k];
	}
      }
    }
  }

  delete [] buffer;
}

void SuperKernel::ReduceFloatFieldAdd(int offset,int elem_count)
{
  /**************************************************************************
   Suppose a field of an element class is offset from the beginning of the
   class by "offset" bytes.  This function computes the sum of the field
   from all elements in the collection.  The sum is broadcast back to each 
   element and stored in the same field.  The old content of the field is 
   overwritten.  If the field is an array, corresponding array elements are 
   summed together.  Note "offset" is in bytes and "elem_count" is the number 
   of array elements.
   ***************************************************************************/
  int i, j, k;
  float *p, *buffer;

  buffer = new float[elem_count];

  for(i = 0; i < elem_count; i++) buffer[i] = 0.0;
 
  if(dimensions == 1) {
    for(i = 0; i < dim1size; i++) {
      if(Is_Local(i)) {
	p = (float*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) buffer[j] += p[j];
      }
    }
  } 

  if(dimensions == 2) {
    for(i = 0; i < dim1size; i++) {
      for(j = 0; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  p = (float*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) buffer[k] += p[k];
	}
      }
    }
  }

  /* reduce from and broadcast to processors */ 
  pcxx_ReduceFloatAdd((void*)buffer,elem_count);

  if(dimensions == 1) {
    for(i = 0; i < dim1size; i++) {
      if(Is_Local(i)) {
	p = (float*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) p[j] = buffer[j];
      }
    }
  }

  if(dimensions == 2) {
    for(i = 0; i < dim1size; i++) {
      for(j = 0; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  p = (float*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) p[k] = buffer[k];
	}
      }
    }
  }

  delete [] buffer;
}

void SuperKernel::ReduceIntFieldAdd(int offset,int elem_count)
{
  /**************************************************************************
   Suppose a field of an element class is offset from the beginning of the
   class by "offset" bytes.  This function computes the sum of the field
   from all elements in the collection.  The sum is broadcast back to each 
   element and stored in the same field.  The old content of the field is 
   overwritten.  If the field is an array, corresponding array elements are 
   summed together.  Note "offset" is in bytes and "elem_count" is the number 
   of array elements.
   ***************************************************************************/
  int i, j, k;
  int *p, *buffer;

  buffer = new int[elem_count];

  for(i = 0; i < elem_count; i++) buffer[i] = 0;
 
  if(dimensions == 1) {
    for(i = 0; i < dim1size; i++) {
      if(Is_Local(i)) {
	p = (int*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) buffer[j] += p[j];
      }
    }
  }

  if(dimensions == 2) {
    for(i = 0; i < dim1size; i++) {
      for(j = 0; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  p = (int*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) buffer[k] += p[k];
	}
      }
    }
  }

  /* reduce from and broadcast to processors */ 
  pcxx_ReduceIntAdd((void*)buffer,elem_count);

  if(dimensions == 1) {
    for(i = 0; i < dim1size; i++) {
      if(Is_Local(i)) {
	p = (int*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) p[j] = buffer[j];
      }
    }
  } 

  if(dimensions == 2) {
    for(i = 0; i < dim1size; i++) {
      for(j = 0; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  p = (int*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) p[k] = buffer[k];
	}
      }
    }
  }

  delete [] buffer;
}

void SuperKernel::ReduceIntFieldMin(int offset,int elem_count)
{
  /**************************************************************************
   Suppose a field of an element class is offset from the beginning of the
   class by "offset" bytes.  This function finds the minimum of the field
   of all elements in the collection.  The result is broadcast back to each 
   element and stored in the same field.  The old content of the field is 
   overwritten.  If the field is an array, corresponding array elements are 
   compared with each other.  Note "offset" is in bytes and "elem_count" is 
   the number of array elements.
   ***************************************************************************/
  int i, j, k, first_i, first_j, found;
  int *p, *buffer;

  buffer = new int[elem_count];
 
  if(dimensions == 1) {
    for(i = 0; i < dim1size; i++) {
      if(Is_Local(i)) {
	first_i = i;
	p = (int*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) buffer[j] = p[j];
	break;
      }
    }
    for(i = first_i+1; i < dim1size; i++) {
      if(Is_Local(i)) {
	p = (int*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) 
	  if(p[j] < buffer[j]) buffer[j] = p[j];
      }
    }
  }

  found = 0;
  if(dimensions == 2) {
    for(i = 0; i < dim1size; i++) {
      if(found == 1) break;
      for(j = 0; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  first_i = i;
	  first_j = j;
	  p = (int*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) buffer[k] = p[k];
	  found = 1;
	  break;
	}
      }
    }
    for(i = first_i+1; i < dim1size; i++) {
      for(j = first_j+1; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  p = (int*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) 
	    if(p[k] < buffer[k]) buffer[k] = p[k];
	}
      }
    }
  }

  /* reduce from and broadcast to processors */ 
  pcxx_ReduceIntMin((void*)buffer,elem_count);

  if(dimensions == 1) {
    for(i = 0; i < dim1size; i++) {
      if(Is_Local(i)) {
	p = (int*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) p[j] = buffer[j];
      }
    }
  }

  if(dimensions == 2) {
    for(i = 0; i < dim1size; i++) {
      for(j = 0; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  p = (int*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) p[k] = buffer[k];
	}
      }
    }
  }

  delete [] buffer;
}

void SuperKernel::ReduceFloatFieldMin(int offset,int elem_count)
{
  /**************************************************************************
   Suppose a field of an element class is offset from the beginning of the
   class by "offset" bytes.  This function finds the minimum of the field
   of all elements in the collection.  The result is broadcast back to each 
   element and stored in the same field.  The old content of the field is 
   overwritten.  If the field is an array, corresponding array elements are 
   compared with each other.  Note "offset" is in bytes and "elem_count" is 
   the number of array elements.
   ***************************************************************************/
  int i, j, k, first_i, first_j, found;
  float *p, *buffer;

  buffer = new float[elem_count];
 
  if(dimensions == 1) {
    for(i = 0; i < dim1size; i++) {
      if(Is_Local(i)) {
	first_i = i;
	p = (float*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) buffer[j] = p[j];
	break;
      }
    }
    for(i = first_i+1; i < dim1size; i++) {
      if(Is_Local(i)) {
	p = (float*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) 
	  if(p[j] < buffer[j]) buffer[j] = p[j];
      }
    }
  }

  found = 0;
  if(dimensions == 2) {
    for(i = 0; i < dim1size; i++) {
      if(found == 1) break;
      for(j = 0; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  first_i = i;
	  first_j = j;
	  p = (float*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) buffer[k] = p[k];
	  found = 1;
	  break;
	}
      }
    }
    for(i = first_i+1; i < dim1size; i++) {
      for(j = first_j+1; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  p = (float*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) 
	    if(p[k] < buffer[k]) buffer[k] = p[k];
	}
      }
    }
  }

  /* reduce from and broadcast to processors */ 
  pcxx_ReduceFloatMin((void*)buffer,elem_count);

  if(dimensions == 1) {
    for(i = 0; i < dim1size; i++) {
      if(Is_Local(i)) {
	p = (float*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) p[j] = buffer[j];
      }
    }
  }

  if(dimensions == 2) {
    for(i = 0; i < dim1size; i++) {
      for(j = 0; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  p = (float*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) p[k] = buffer[k];
	}
      }
    }
  }

  delete [] buffer;
}

void SuperKernel::ReduceDoubleFieldMin(int offset,int elem_count)
{
  /**************************************************************************
   Suppose a field of an element class is offset from the beginning of the
   class by "offset" bytes.  This function finds the minimum of the field
   of all elements in the collection.  The result is broadcast back to each 
   element and stored in the same field.  The old content of the field is 
   overwritten.  If the field is an array, corresponding array elements are 
   compared with each other.  Note "offset" is in bytes and "elem_count" is 
   the number of array elements.
   ***************************************************************************/
  int i, j, k, first_i, first_j, found;
  double *p, *buffer;

  buffer = new double[elem_count];
 
  if(dimensions == 1) {
    for(i = 0; i < dim1size; i++) {
      if(Is_Local(i)) {
	first_i = i;
	p = (double*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) buffer[j] = p[j];
	break;
      }
    }
    for(i = first_i+1; i < dim1size; i++) {
      if(Is_Local(i)) {
	p = (double*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) 
	  if(p[j] < buffer[j]) buffer[j] = p[j];
      }
    }
  }

  found = 0;
  if(dimensions == 2) {
    for(i = 0; i < dim1size; i++) {
      if(found == 1) break;
      for(j = 0; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  first_i = i;
	  first_j = j;
	  p = (double*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) buffer[k] = p[k];
	  found = 1;
	  break;
	}
      }
    }
    for(i = first_i+1; i < dim1size; i++) {
      for(j = first_j+1; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  p = (double*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) 
	    if(p[k] < buffer[k]) buffer[k] = p[k];
	}
      }
    }
  }

  /* reduce from and broadcast to processors */ 
  pcxx_ReduceDoubleMin((void*)buffer,elem_count);

  if(dimensions == 1) {
    for(i = 0; i < dim1size; i++) {
      if(Is_Local(i)) {
	p = (double*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) p[j] = buffer[j];
      }
    }
  }

  if(dimensions == 2) {
    for(i = 0; i < dim1size; i++) {
      for(j = 0; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  p = (double*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) p[k] = buffer[k];
	}
      }
    }
  }

  delete [] buffer;
}

void SuperKernel::ReduceIntFieldMax(int offset,int elem_count)
{
  /**************************************************************************
   Suppose a field of an element class is offset from the beginning of the
   class by "offset" bytes.  This function finds the maximum of the field
   of all elements in the collection.  The result is broadcast back to each 
   element and stored in the same field.  The old content of the field is 
   overwritten.  If the field is an array, corresponding array elements are 
   compared with each other.  Note "offset" is in bytes and "elem_count" is 
   the number of array elements.
   ***************************************************************************/
  int i, j, k, first_i, first_j, found;
  int *p, *buffer;

  buffer = new int[elem_count];
 
  if(dimensions == 1) {
    for(i = 0; i < dim1size; i++) {
      if(Is_Local(i)) {
	first_i = i;
	p = (int*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) buffer[j] = p[j];
	break;
      }
    }
    for(i = first_i+1; i < dim1size; i++) {
      if(Is_Local(i)) {
	p = (int*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) 
	  if(p[j] > buffer[j]) buffer[j] = p[j];
      }
    }
  }

  found = 0;
  if(dimensions == 2) {
    for(i = 0; i < dim1size; i++) {
    if(found == 1) break;
      for(j = 0; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  first_i = i;
	  first_j = j;
	  p = (int*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) buffer[k] = p[k];
	  found = 1;
	  break;
	}
      }
    }
    for(i = first_i+1; i < dim1size; i++) {
      for(j = first_j+1; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  p = (int*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) 
	    if(p[k] > buffer[k]) buffer[k] = p[k];
	}
      }
    }
  }

  /* reduce from and broadcast to processors */ 
  pcxx_ReduceIntMax((void*)buffer,elem_count);

  if(dimensions == 1) {
    for(i = 0; i < dim1size; i++) {
      if(Is_Local(i)) {
	p = (int*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) p[j] = buffer[j];
      }
    }
  }

  if(dimensions == 2) {
    for(i = 0; i < dim1size; i++) {
      for(j = 0; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  p = (int*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) p[k] = buffer[k];
	}
      }
    }
  }

  delete [] buffer;
}

void SuperKernel::ReduceFloatFieldMax(int offset,int elem_count)
{
  /**************************************************************************
   Suppose a field of an element class is offset from the beginning of the
   class by "offset" bytes.  This function finds the maximum of the field
   of all elements in the collection.  The result is broadcast back to each 
   element and stored in the same field.  The old content of the field is 
   overwritten.  If the field is an array, corresponding array elements are 
   compared with each other.  Note "offset" is in bytes and "elem_count" is 
   the number of array elements.
   ***************************************************************************/
  int i, j, k, first_i, first_j, found;
  float *p, *buffer;

  buffer = new float[elem_count];
 
  if(dimensions == 1) {
    for(i = 0; i < dim1size; i++) {
      if(Is_Local(i)) {
	first_i = i;
	p = (float*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) buffer[j] = p[j];
	break;
      }
    }
    for(i = first_i+1; i < dim1size; i++) {
      if(Is_Local(i)) {
	p = (float*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) 
	  if(p[j] > buffer[j]) buffer[j] = p[j];
      }
    }
  }

  found = 0;
  if(dimensions == 2) {
    for(i = 0; i < dim1size; i++) {
      if(found == 1) break;
      for(j = 0; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  first_i = i;
	  first_j = j;
	  p = (float*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) buffer[k] = p[k];
	  found = 1;
	  break;
	}
      }
    }
    for(i = first_i+1; i < dim1size; i++) {
      for(j = first_j+1; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  p = (float*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) 
	    if(p[k] > buffer[k]) buffer[k] = p[k];
	}
      }
    }
  }

  /* reduce from and broadcast to processors */ 
  pcxx_ReduceFloatMax((void*)buffer,elem_count);

  if(dimensions == 1) {
    for(i = 0; i < dim1size; i++) {
      if(Is_Local(i)) {
	p = (float*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) p[j] = buffer[j];
      }
    }
  }

  if(dimensions == 2) {
    for(i = 0; i < dim1size; i++) {
      for(j = 0; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  p = (float*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) p[k] = buffer[k];
	}
      }
    }
  }

  delete [] buffer;
}

void SuperKernel::ReduceDoubleFieldMax(int offset,int elem_count)
{
  /**************************************************************************
   Suppose a field of an element class is offset from the beginning of the
   class by "offset" bytes.  This function finds the maximum of the field
   of all elements in the collection.  The result is broadcast back to each 
   element and stored in the same field.  The old content of the field is 
   overwritten.  If the field is an array, corresponding array elements are 
   compared with each other.  Note "offset" is in bytes and "elem_count" is 
   the number of array elements.
   ***************************************************************************/
  int i, j, k, first_i, first_j, found;
  double *p, *buffer;

  buffer = new double[elem_count];
 
  if(dimensions == 1) {
    for(i = 0; i < dim1size; i++) {
      if(Is_Local(i)) {
	first_i = i;
	p = (double*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) buffer[j] = p[j];
	break;
      }
    }
    for(i = first_i+1; i < dim1size; i++) {
      if(Is_Local(i)) {
	p = (double*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) 
	  if(p[j] > buffer[j]) buffer[j] = p[j];
      }
    }
  }

  found = 0;
  if(dimensions == 2) {
    for(i = 0; i < dim1size; i++) {
      if(found == 1) break;
      for(j = 0; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  first_i = i;
	  first_j = j;
	  p = (double*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) buffer[k] = p[k];
	  found = 1;
	  break;
	}
      }
    }
    for(i = first_i+1; i < dim1size; i++) {
      for(j = first_j+1; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  p = (double*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) 
	    if(p[k] > buffer[k]) buffer[k] = p[k];
	}
      }
    }
  }

  /* reduce from and broadcast to processors */ 
  pcxx_ReduceDoubleMax((void*)buffer,elem_count);

  if(dimensions == 1) {
    for(i = 0; i < dim1size; i++) {
      if(Is_Local(i)) {
	p = (double*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) p[j] = buffer[j];
      }
    }
  }

  if(dimensions == 2) {
    for(i = 0; i < dim1size; i++) {
      for(j = 0; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  p = (double*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) p[k] = buffer[k];
	}
      }
    }
  }

  delete [] buffer;
}
