/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/*********************************************
   Last overhaul by S. Yang (2/11/94)
                    Pete Beckman (2/15/94)
 *********************************************/

#include <stdarg.h>
#include <stdio.h>

#include "GlobalStruct.h"
#include "profiler_defs.h"

/*************************************************
   Eventually, we need to get rid of all VERBOSE 
   tests, since they slow down the communication.

   VERBOSE KEY:
   0 = No Tracing
   1 = Trace receives
   2 = Trace sends & receives
   3 = Trace sends & receives & "in proc"
   4 = all
 **************************************************/
int VERBOSE = 0;

#define DONE -1
#define DONTCARE -1

#define HereIsMsg 100
#define WhereIsMsg 101
#define ReadBlockMsg 103
#define EndOfBlockMsg 105    
#define BarrierFanOutMsg 107
#define BarrierFanInMsg 108
#define BroadcastMsg 109
#define ReadFArrayMsg 110
#define ReduceMsg 111
#define RemoteExecuteMsg 112
#define EndOfRemoteExecuteMsg 113
#define ReadFArrayElemMsg 114

/********************************************************
 NODE GLOBAL VARIABLES                                   
**********************************************************/
#define MAXCOLLECTION 100
#define MAXBUFSIZE 8192
static volatile int ask_buf[500];
static volatile int ask_flag = 0;
static int WarningCount =0;
static PTR_INFO TableOfCollection[MAXCOLLECTION];
char ReceiveBuffer[MAXBUFSIZE];
char SendBuffer[MAXBUFSIZE];

char *BlockDataBuff;  /* Pointer to where a block of data should be put */
int barrierFlag = 0; 
int IHaveReachedBarrier = 0;
int broadcast_flag = 0;
int reduction_flag = 0;
static long parent_id, left_child, right_child;
static int num_children;
int barrier_msg = 0;
int remote_execute_flag = 0;
int MyTotalNodes, MyNodeNumber;

/*******************************************************************
   Prototypes                                                             
 *******************************************************************/
void Message(char * st, int flag);
void MySend(int type, char * buffer, int len, int ToWhom);
void pcxx_Abort();
void pcxx_Poll();
void here_is_data_handler(int type, int node);
void where_is_data_handler(int type, int node);
void end_of_data_handler(int type, int node);
void end_of_block_handler(int type, int node);
void read_data_handler(int type, int node);
void read_block_handler(int type, int node);
void read_sub_block_handler(int type, int node);
void BarrierFanIn(int node);
void read_farray_handler(int type, int node);
void remote_execute_handler(int type, int node);
void end_of_remote_handler(int type, int node);

/***********************************************************
 Shared source code 
 ************************************************************/
#include "dmrts.h"
#include "common.h" /* Common support routines for EVERY KernelCom.c (phb) */

/*********************************************************************
 GENERAL FUNCTIONS.
 *********************************************************************/
/*---------------------------------------------------------------------
  Message is a general error reporting function that prints a message
  pointed to by "st".  The flag is either WARNING or FATAL.  On FATAL
  it is necesary to abort the program.
  ---------------------------------------------------------------------*/
/*********************************************************************/
void Message(char * st, int flag)
{
  if (flag == WARNING)
    {
      WarningCount++;
      fprintf(stderr,"Warning %s\n",st);
      return;
    }
   if (flag == FATAL)
    {
      fprintf(stderr,"Fatal error %s\n",st);
       pcxx_Abort();
    }
}

/*********************************************************************/
int IamtheManager(int id)
{
  return ((id % MyTotalNodes) == MyNodeNumber);
}

/*********************************************************************
 BARRIER RELATED FUNCTIONS
 *********************************************************************/
/*--------------------------------------------------------------------
  pcxx_Barrier() must use messages, so that pcxx_Poll() can be run 
  periodically.  Otherwise, we could use mpc_sync, which blocks.  
  These routines were originally designed by P. Beckman (SP1 port) based 
  on designs by D. Gannon (Paragon port).  They do a tree fan-out and then 
  fan-in.

  PVM provides a 'probe' function to check if a message has arrived.
  Then a recv can be issued.  Send is non-blocking

  Barrier to synchronize operations in pC++. Processors wait at the
  Barrier until all are there.  Processors need to run pcxx_Poll() while 
  they wait.  This pariticular implementation uses a design by 
  Dennis Gannon.  Processors do a tree fan in to the root node, then a 
  fan out and resume computation.  A fast hardware/OS supported barrier is
  prefered, but the need to run pcxx_Poll() while waiting in the barrier
  excludes most system implementations

  The Barrier works in the following way: The leaves start sending FanInMsg
  to their parents, while their parents poll messages from the children.  
  Once a processor receives messages from all of its children, it sends 
  a message to its parent.  The fan-in completes when processor zero gets 
  the message.  Processor zero then sends FanOutMsg down through the tree.  
  Barrier completes when all processors receive FanOutMsg.    S.Y. (2/11/94)
  --------------------------------------------------------------------*/
/*********************************************************************/
void BarrierFanIn(int from)
{
  /* record the number of children fanned in */
  barrierFlag++;  

  /*If I am not at the barrier, I do not let BarrierFanInMsg funnel through.*/
  if (IHaveReachedBarrier == 0) return;

  if((barrierFlag == 1 && num_children == 1) || 
     (barrierFlag == 2 && num_children == 2) ) {
    if(MyNodeNumber == 0) { 
      barrierFlag = DONE;
    } else {
      MySend(BarrierFanInMsg, SendBuffer, 0, parent_id);
    }
  }

  return;
}

/* this is included so that breakpoint executive is called
 * only when compiled for breakpointing.
 */
 
#include "distApi.h"

/********************************************************************/
void pcxx_Barrier()
{
  PROFILE_ENTER(PROFILE_Barrier);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_BARRIER_ENTER, 0);
  if (MyNodeNumber == 0) {
    brk_BREAKPOINT_EXECUTIVE(); 
  }

  while(remote_execute_flag != 0) pcxx_Poll();
  IHaveReachedBarrier = 1;  /* I have really entered a barrier */

  if (num_children == 0) {
    MySend(BarrierFanInMsg, SendBuffer, 0, parent_id);
  }

  if((barrierFlag == 1 && num_children == 1) || 
     (barrierFlag == 2 && num_children == 2) ) 
    {
      if(MyNodeNumber == 0) { 
	barrierFlag = DONE;
      } else {
	MySend(BarrierFanInMsg, SendBuffer, 0, parent_id);
      }
    }

  while(barrierFlag != DONE)
    pcxx_Poll();

  if (num_children >= 1) {
    MySend(BarrierFanOutMsg, SendBuffer,0,left_child);
  }
  if (num_children == 2) {
    MySend(BarrierFanOutMsg, SendBuffer,0,right_child);
  }

  IHaveReachedBarrier = 0;
  barrierFlag = 0;
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_BARRIER_EXIT, 0);
  PROFILE_EXIT(PROFILE_Barrier);
}

/*********************************************************************
 DATA FETCHING FUNCTIONS
 *********************************************************************/
/*-----------------------------------------------------------------------
  The pC++ system uses a global name space for elements. Each element has 
  a manager processor and an owner processor.
  
  The communication model used by the compiler (and programmer!) is
  to fetch copies of element or parts of element data.  Because these
  data fetches can be to arbitrary processors, it is necessary to have
  a good mechanism that does not require the owner of an element to
  explicitly "send" the data.  Rather we must "interrupt" the owner
  and ask it to gather the data and forward it on to the requester.
  
  The old version of pC++ on the CM-5 used CMAM which required polling
  to replace interrups and the compiler liberally generated Poll() function
  calls in the code to make sure that this properly simulated interrupts.
  The new TMC CMMD has mechanisms for such interrupt driven communication
  as does Intel's NX library, so Polling is not necessary.
  
  However, it may be the case that polling may be faster on some systems
  if the cost of context switches is high.

  Three of the following functions are used to find the owner of an element. 
  The external interfact to this block of code is:
  int WhereIsData(int element_id, int collection_id)                   
  It returns the processor ID that "owns" the element "element_id" in     
  the collection identified by integer "collection_id"                    
                                                                          
  The ask_flag is the number of pending requests asking for the id of 
  the owner of an element.  This should only be 0 or 1 in this 
  implementation.
  
  Every time a manager of an element replies with the id
  of an owner, the ask_flag is decremented.  Every time a request is made
  to a manager the ask_flag is incremented.
  
  release_ask() polls until the pending reqest to a manager is been satisfied.
  It is called only by WhereIsData.
  
  outstanding_ask() does the increment.  It is also only called by 
  WhereIsData. 

  here_is_data_handler is invoked only by where_is_data_handler to
  reply messages.  It saves the reply in a buffer called ask_buf and 
  decrements a flag ask_flag

  where_is_data_handler function replies to a request to identify the 
  location of an element (index) in a given collection (Coll_Id).  It first 
  looks in the table for the given collection and gets the processor ID 
  in "data_is".  It then replies to the requesting processors (node) by 
  asking it to invoke the message handler "here_is_data_handler" with 
  the "data_is" argument.

  To find the owner of a collection element the compiler Kernel Collection
  routines will call WhereIsData.  In the current version, managers are 
  uniformly distribued across the system in a cyclic manner.  That is, 
  element 1 is managed by processor 1, element 2 is managed by processor 2,
  ... element P-1 is managed by processor P-1 and element P is managed 
  by processor 0, etc.  This seems to work as well as anything.  Note:  
  the location of the manager is INDEPENDENT of the owner of the element. 
  The owner of an element is determined by the data alignment and 
  distribution of the collection. Also note: the compiler and the kernel 
  first checks to see if an element is local befor it goes through this step.
  If the element is local no request to find it is made, i.e. "If I own
  the element, I do not need to ask the manager where it is".  This means
  that local requests do not generate message trafic.
  
  The way this works is as follows:
  WhereIsData computes the manager processors by modular arithmetic.
  it then sends an "Active Message" to the owner asking it to invoke
  the where_is_data_handler function for the collection indentified by
  "Coll_Id" and for the element identified by "Id".
  
  Befor doing this it 
  1. looks at the ask_flag and polls until the flag 
  is 0. (this should not be needed in this implementation).
  
  2. posts an oustanding requests by incrementing the ask_flag
     
  After sending the "Active Message",  the processor polls for
  the ask flag to be set to 0, but calling release_ask() again.
  
  On the manager side, when the "where_is_data_handler" handler message
  is recieved and invoked it looks up the element owner and replies
  with an "active message" to invoke the here_is_data_handler function.
  This function decrements the ask_flag and set the reply in
  ask_buf.
  
  When the flag is decremented the WhereIsData function can return
  the value of ask_buf.
                                                                          
  Once we have the owner of an element we may ask to get data.           
  The basic communication functions to get data are:                     
  
  int AskForData(int where, int index, int offset,                       
                 int Coll_Id, int *data1,int *data2,int *data3,int *data4)  
                                                                          
  which goes to processor "where" and to element "index"  in collection 
  "Coll_Id", to fetch four words (16 bytes) at offset.  Offset is a     
  an integer number of words to add to the top of the element to get    
  the desired location.                                                 
  
  For extracting larger chunks of data we use:                           
                                                                          
  int AskForDataBlock(int where, int index, int Coll_Id, int *buffer)  
                                                                          
  which returns the entire elemement into buffer from processor "where" 
                                                                          
  The data transfers follow the same style protocal as above.
  a request is made to the owner of the data through AskForData
  or AskForDataBlock.  this sends a message to the owner which
  invokes 

  end_of_data_handle is invoked only by a reply from read_data_handler.  
  It loads four integers into the "integer message buffers" and 
  decrements the sm_flag which indicates that a transfer is complete.

  read_data_handler is invoked by the receipt of a ReadDataMsg from a processor
  requesting 4 ints of data.  It gathers the data and replies with a
  EndOfDataMsg.  A call to AskForData results in a call to read_data_handler
  at the receiving processor side. 
  ---------------------------------------------------------------------------*/

/*------------------------------------------------------------------------
  The following three functions are used for finding out on which processor
  the wanted data are located.
 -------------------------------------------------------------------------*/
/*********************************************************************/
void here_is_data_handler(int type, int node)
{
  /*printf("here_is_data_handler on proc %d \n",pcxx_NodeNumber());*/
  ask_buf[node] = ((int *) ReceiveBuffer)[0];
  ask_flag = 0;
  pcxx_Poll();
  return;
}

/*********************************************************************/
void where_is_data_handler(int type, int node)
{
  int data_is;
  int index;
  int Coll_Id;

  index   = ((int *) ReceiveBuffer)[0];
  Coll_Id = ((int *) ReceiveBuffer)[1];

  data_is = TableOfCollection[Coll_Id]->Where[index]; /* identity */
  ((int *) SendBuffer)[node] = data_is;

  /*printf("%d sending HereIsMsg to %d\n",MyNodeNumber,node);*/

  MySend(HereIsMsg, (char *) &(((int *) SendBuffer)[node]), sizeof(int), node);
  return;
}

/*********************************************************************/
int WhereIsData(int id, int Coll_Id)
{
  static int manager = 0;
  int rval;

  PROFILE_ENTER(PROFILE_Where);

  /* Calculate who the manager is, using a _SIMPLE_ round robin distribution */
  manager = (id % MyTotalNodes); 
  
  /* If I am the manager, I do NOT need to send a message...*/
  /* I can just look inthe TableoFCollection myself (PHB) */
  if (manager == MyNodeNumber) {
    PROFILE_EXIT(PROFILE_Where);
    return (TableOfCollection[Coll_Id]->Where[id]);
  }

  /* else, the manager must be on a different node */
  ((int *) SendBuffer)[0] = id;
  ((int *) SendBuffer)[1] = Coll_Id;

  /*printf("%d sending WhereIsMsg to %d\n",MyNodeNumber,manager);*/
  ask_flag = 1;
  MySend(WhereIsMsg, SendBuffer, 2*sizeof(int), manager);
  while (ask_flag != 0) pcxx_Poll();
  rval = ask_buf[manager];

  PROFILE_EXIT(PROFILE_Where);
  return rval;
}

/*------------------------------------------------------------------------
  The following three functions are used for fetching four integers 
  at one time.
  ------------------------------------------------------------------------*/
/*********************************************************************/
void end_of_block_handler(int type, int node)
{
  ask_flag = 0;
  pcxx_Poll();
  return ;
}

/*--------------------------------------------------------------------------
  The following two functions are used for fetching a subblock of 
  a collection element at one time.
  --------------------------------------------------------------------------*/
/*********************************************************************/
void read_block_handler(int type, int node)
{
  int Coll_Id, offset, size, index;

  index   = ((int *) ReceiveBuffer)[0];
  offset  = ((int *) ReceiveBuffer)[1]; /* Size in INTs */
  Coll_Id = ((int *) ReceiveBuffer)[2];
  size    = ((int *) ReceiveBuffer)[3]; /* Size in INTs */

  /*printf("%d sending EndOfBlockMsg to %d\n",MyNodeNumber,node);*/

  MySend(EndOfBlockMsg, 
	 (char *) (TableOfCollection[Coll_Id]->Table[index]+offset), 
	 size*sizeof(int), node);
  return;
}

/*********************************************************************/
int AskForDataBlock(int where, int index, int offset, 
		       int Coll_Id, int size, int *buffer)
{
  BlockDataBuff = (char*)buffer;
  ask_flag = 1;

  ((int *) SendBuffer)[0] = index;
  ((int *) SendBuffer)[1] = offset;   /* size is in INTs */
  ((int *) SendBuffer)[2] = Coll_Id;
  ((int *) SendBuffer)[3] = size;     /* size is in INTs */

  MySend(ReadBlockMsg, SendBuffer, 4*sizeof(int), where);

  /*wait for the answer set by the read_block_handler routine*/
  while(ask_flag != 0) pcxx_Poll();

  return 1;
}

/*--------------------------------------------------------------------------
  The following three functions are used for fetching an FArray.
  --------------------------------------------------------------------------*/
/*********************************************************************/
void CopyFArray(char *dest, char *source, int rank, int size)
{
  /* memcpy(dst,src,size); */
  int i;
  for(i = 0; i < size; i++) dest[i] = source[i];
}

/*********************************************************************/
void read_farray_handler(int type,int dest_node)
{
  int index,offset,Coll_Id,size;
  char *buffer;

  index   = ((int *) ReceiveBuffer)[0];
  Coll_Id = ((int *) ReceiveBuffer)[1];
  offset  = ((int *) ReceiveBuffer)[2];
  size    = ((int *) ReceiveBuffer)[3];

  buffer = (char*)*((long*)(TableOfCollection[Coll_Id]->Table[index]+offset));
  MySend(EndOfBlockMsg,buffer,size,dest_node);
  return;
}

/*********************************************************************/
int AskForFArray(int where, unsigned index, unsigned Coll_Id, int offset,
		 char *buffer, int size)
{
  /* size is in bytes.  however offset is in integers a legacy left by 
     Francios Bodin */

  BlockDataBuff = (char*)buffer;
  ask_flag = 1;

  ((int *) SendBuffer)[0] = index;
  ((int *) SendBuffer)[1] = Coll_Id;
  ((int *) SendBuffer)[2] = offset;
  ((int *) SendBuffer)[3] = size;

  MySend(ReadFArrayMsg, SendBuffer, 4*sizeof(int), where);
  while(ask_flag != 0) pcxx_Poll();
  /* wait for the answer set by the read_farray_handler routine */
  return 1;
}

void CopyFArrayElem(char* source, char *buffer,
                  int elem_size, int rank,
                  int i_len, int i_inc, int i_off, 
                  int j_len, int j_inc, int j_off,
                  int k_len, int k_inc, int k_off, 
                  int i, int j, int k)
{
  int m;
  char *pt;
  pt = source + (i + j*i_len + k*i_len*j_len)*elem_size;
  for(m = 0; m < elem_size; m++) buffer[m] = pt[m];
}

/*********************************************************************/
void read_farray_elem_handler(int type,int dest_node)
{
  int index,offset,Coll_Id,elem_size;
  char *buffer, *pt;
  int i_len, j_len, k_len, i, j, k;

  index     = ((int *) ReceiveBuffer)[0];
  Coll_Id   = ((int *) ReceiveBuffer)[1];
  offset    = ((int *) ReceiveBuffer)[2];
  elem_size = ((int *) ReceiveBuffer)[3];
  i_len     = ((int *) ReceiveBuffer)[4];
  j_len     = ((int *) ReceiveBuffer)[5];
  k_len     = ((int *) ReceiveBuffer)[6];
  i         = ((int *) ReceiveBuffer)[7];
  j         = ((int *) ReceiveBuffer)[8];
  k         = ((int *) ReceiveBuffer)[9];

  pt = (char*)*((long*)(TableOfCollection[Coll_Id]->Table[index]+offset));
  buffer = pt + (i + j*i_len + k*i_len*j_len)*elem_size;
  MySend(EndOfBlockMsg,buffer,elem_size,dest_node);
  return;
}

int AskForFArrayElem(int where,int index,int Coll_Id,int offset,char *buffer,
                   int elem_size, int rank,
                   int i_len, int i_inc, int i_off, 
                   int j_len, int j_inc, int j_off,
                   int k_len, int k_inc, int k_off, 
                   int i, int j, int k)
{
  /* elem_size is in bytes.  however offset is in integers a legacy left by 
     Francios Bodin */

  BlockDataBuff = (char*)buffer;
  ask_flag = 1;

  ((int *) SendBuffer)[0] = index;
  ((int *) SendBuffer)[1] = Coll_Id;
  ((int *) SendBuffer)[2] = offset;
  ((int *) SendBuffer)[3] = elem_size;
  ((int *) SendBuffer)[4] = i_len;
  ((int *) SendBuffer)[5] = j_len;
  ((int *) SendBuffer)[6] = k_len;
  ((int *) SendBuffer)[7] = i;
  ((int *) SendBuffer)[8] = j;
  ((int *) SendBuffer)[9] = k;

  MySend(ReadFArrayElemMsg, SendBuffer, 10*sizeof(int), where);
  while(ask_flag != 0) pcxx_Poll();
  /* wait for the answer set by the read_farray_elem_handler routine */
  return 1;
}

/*********************************************************************/
/*********************************************************************/
/************************** ALIEN FUNCTIONS **************************/
/*********************************************************************/
/*********************************************************************/

/*-------------------------------------------------------------------------
Some machines have their own "Native" Broadcast and Reduction
functions.  For those machines, pC++ uses the faster, often hardware
supported, functions calls.  To do this, #defines are used to call the
Native functions that live in KernelCom.c for that machine.  All
machines get both "Native" and "Alien" routines so that the functions
exist for timing purposes, but if no Native routine exists, it is just
mapped to the Alien routine.  The Alien routines use our hand-coded
solutions to Broadcast and Reduce.  On a machine that does not have
any special Broadcast or Reduce functions, pcxx_Broadcast will simply
be mapped to use pcxx_Alien* which relys on tree fan in and fan out
versions.
--------------------------------------------------------------------------*/

/* Forward declarations for Alien functions */
void pcxx_AlienBroadcastBytes(int flag, int length, void * buffer);

void pcxx_AlienReduceIntAdd(void* buffer,int elem_count);
void pcxx_AlienReduceFloatAdd(void* buffer,int elem_count);
void pcxx_AlienReduceDoubleAdd(void* buffer,int elem_count);

void pcxx_AlienReduceIntMax(void* buffer,int elem_count);
void pcxx_AlienReduceFloatMax(void* buffer,int elem_count);
void pcxx_AlienReduceDoubleMax(void* buffer,int elem_count);

void pcxx_AlienReduceIntMin(void* buffer,int elem_count);
void pcxx_AlienReduceFloatMin(void* buffer,int elem_count);
void pcxx_AlienReduceDoubleMin(void* buffer,int elem_count);

/**********************************************************************/
#ifndef __PARAGON__
/* Links to Alien function */
void pcxx_BroadcastBytes(int flag, int length, void * buffer) {
  pcxx_AlienBroadcastBytes(flag, length, buffer); }
  
void pcxx_ReduceIntAdd(void* buffer,int elem_count) {
  pcxx_AlienReduceIntAdd(buffer,elem_count); }

void pcxx_ReduceFloatAdd(void* buffer,int elem_count) {
  pcxx_AlienReduceFloatAdd(buffer,elem_count); }

void pcxx_ReduceDoubleAdd(void* buffer,int elem_count) {
  pcxx_AlienReduceDoubleAdd(buffer,elem_count); }

void pcxx_ReduceIntMax(void* buffer,int elem_count) {
  pcxx_AlienReduceIntMax(buffer,elem_count); }

void pcxx_ReduceFloatMax(void* buffer,int elem_count) {
  pcxx_AlienReduceFloatMax(buffer,elem_count); }

void pcxx_ReduceDoubleMax(void* buffer,int elem_count) {
  pcxx_AlienReduceDoubleMax(buffer,elem_count); }

void pcxx_ReduceIntMin(void* buffer,int elem_count) {
  pcxx_AlienReduceIntMin(buffer,elem_count); }

void pcxx_ReduceFloatMin(void* buffer,int elem_count) {
  pcxx_AlienReduceFloatMin(buffer,elem_count); }

void pcxx_ReduceDoubleMin(void* buffer,int elem_count) {
  pcxx_AlienReduceDoubleMin(buffer,elem_count); }

/**************** MAKE SURE NATIVE IS DEFINED TOO! ********************/
void pcxx_NativeBroadcastBytes(int flag, int length, void * buffer) {
  pcxx_AlienBroadcastBytes(flag, length, buffer); }
  
void pcxx_NativeReduceIntAdd(void* buffer,int elem_count) {
  pcxx_AlienReduceIntAdd(buffer,elem_count); }

void pcxx_NativeReduceFloatAdd(void* buffer,int elem_count) {
  pcxx_AlienReduceFloatAdd(buffer,elem_count); }

void pcxx_NativeReduceDoubleAdd(void* buffer,int elem_count) {
  pcxx_AlienReduceDoubleAdd(buffer,elem_count); }

void pcxx_NativeReduceIntMax(void* buffer,int elem_count) {
  pcxx_AlienReduceIntMax(buffer,elem_count); }

void pcxx_NativeReduceFloatMax(void* buffer,int elem_count) {
  pcxx_AlienReduceFloatMax(buffer,elem_count); }

void pcxx_NativeReduceDoubleMax(void* buffer,int elem_count) {
  pcxx_AlienReduceDoubleMax(buffer,elem_count); }

void pcxx_NativeReduceIntMin(void* buffer,int elem_count) {
  pcxx_AlienReduceIntMin(buffer,elem_count); }

void pcxx_NativeReduceFloatMin(void* buffer,int elem_count) {
  pcxx_AlienReduceFloatMin(buffer,elem_count); }

void pcxx_NativeReduceDoubleMin(void* buffer,int elem_count) {
  pcxx_AlienReduceDoubleMin(buffer,elem_count); }
#else
/**********************************************************************/

/*-------------------------------------------------------------------------
 Paragon Native functions are defined in Paragon KernelCom.c 
 --------------------------------------------------------------------------*/
extern void pcxx_NativeBroadcastBytes(int flag, int length, void * buffer);
extern void pcxx_NativeReduceIntAdd(void* buffer,int elem_count);
extern void pcxx_NativeReduceFloatAdd(void* buffer,int elem_count);
extern void pcxx_NativeReduceDoubleAdd(void* buffer,int elem_count);

extern void pcxx_NativeReduceIntMax(void* buffer,int elem_count);
extern void pcxx_NativeReduceFloatMax(void* buffer,int elem_count);
extern void pcxx_NativeReduceDoubleMax(void* buffer,int elem_count);

extern void pcxx_NativeReduceIntMin(void* buffer,int elem_count);
extern void pcxx_NativeReduceFloatMin(void* buffer,int elem_count);
extern void pcxx_NativeReduceDoubleMin(void* buffer,int elem_count);

void pcxx_BroadcastBytes(int flag, int length, void * buffer) {
  pcxx_NativeBroadcastBytes(flag, length, buffer); }
  
void pcxx_ReduceIntAdd(void* buffer,int elem_count) {
  pcxx_NativeReduceIntAdd(buffer,elem_count); }

void pcxx_ReduceFloatAdd(void* buffer,int elem_count) {
  pcxx_NativeReduceFloatAdd(buffer,elem_count); }

void pcxx_ReduceDoubleAdd(void* buffer,int elem_count) {
  pcxx_NativeReduceDoubleAdd(buffer,elem_count); }

void pcxx_ReduceIntMax(void* buffer,int elem_count) {
  pcxx_NativeReduceIntMax(buffer,elem_count); }

void pcxx_ReduceFloatMax(void* buffer,int elem_count) {
  pcxx_NativeReduceFloatMax(buffer,elem_count); }

void pcxx_ReduceDoubleMax(void* buffer,int elem_count) {
  pcxx_NativeReduceDoubleMax(buffer,elem_count); }

void pcxx_ReduceIntMin(void* buffer,int elem_count) {
  pcxx_NativeReduceIntMin(buffer,elem_count); }

void pcxx_ReduceFloatMin(void* buffer,int elem_count) {
  pcxx_NativeReduceFloatMin(buffer,elem_count); }

void pcxx_ReduceDoubleMin(void* buffer,int elem_count) {
  pcxx_NativeReduceDoubleMin(buffer,elem_count); }
#endif

/*********************************************************************
BROADCAST FUNCTION
*********************************************************************/
/*---------------------------------------------------------------------
pCxx_AlienBroadcastBytes sends messages to all other nodes from the
node with flag = 0.  the message is in the buffer and the length is
known.
---------------------------------------------------------------------*/
/*********************************************************************/
void pcxx_AlienBroadcastBytes(int flag, int length, void * buffer) {

  BlockDataBuff = (char*)buffer;

  /* The following barrier is very important.  First, it prevents a receiver
     from receiving more than one broadcast during poll which may
     overwrite the receiving buffer.  Second, it prevents
     a receiver from receiving a broadcast when it is not ready to receive
     the broadcast, i.e., when it doest not have a buffer to save the
     broadcast.  Also, do not move the above assignment statement to anywhere
     else.         --- S. Y. (2/11/94)
   */
  pcxx_Barrier();

  if(flag == 0 && MyNodeNumber != 0) {
    /*Send buffer to root, if MyNodeNumber != 0 and I want to broadcast.*/
    MySend(BroadcastMsg,buffer,length,0);
  }
  if(MyNodeNumber == 0 && flag != 0) {
    while (broadcast_flag == 0) pcxx_Poll();
  }

  /* Start the msg propagating, from the root (processor 0)*/
  if (MyNodeNumber == 0) {
    if (num_children >= 1) {
      MySend(BroadcastMsg,buffer,length,left_child);	
    }
    if (num_children == 2) {
       MySend(BroadcastMsg,buffer,length,right_child);	
    }
    broadcast_flag = 0;
    return;
  }

  /* Processors other than 0 wait till they receive the broadcast */
  while (broadcast_flag == 0) pcxx_Poll();

  /* Now send the copy I got to my children If I have any */
  if (num_children >= 1) {
    MySend(BroadcastMsg,buffer,length,left_child);	
  }
  if (num_children == 2) {
    MySend(BroadcastMsg,buffer,length,right_child);	
  }

  /* reset flag */
  broadcast_flag = 0;
}

/*-------------------------------------------------------------
  pcxx_AlienReduceDoubleAdd adds together corresponding buffer elements 
  and broadcasts the resulting buffer to all nodes.  The original
  buffer content is overwritten.
 --------------------------------------------------------------*/
/*********************************************************************/
void pcxx_AlienReduceDoubleAdd(void* buffer,int elem_count)
{
   int i, flag;

   pcxx_Barrier();

   /* get totals from children first */
   if(num_children > 0) {
     while(reduction_flag == 0) pcxx_Poll();
     for(i = 0; i < elem_count; i++) { 
       ((double*)buffer)[i] += ((double*)ReceiveBuffer)[i];
     }
   }
   if(num_children > 1) {
     while(reduction_flag < 2) pcxx_Poll();
     for(i = 0; i < elem_count; i++) {
       ((double*)buffer)[i] += ((double*)ReceiveBuffer)[i];
     }
   }
   reduction_flag = 0;

   /* send totals to my parent. */
   if(MyNodeNumber != 0)
     MySend(ReduceMsg, buffer, elem_count*sizeof(double), parent_id);

   /* now share the totals.  processor zero broadcasts */
   flag = 1;
   if(MyNodeNumber == 0) flag = 0;
   pcxx_BroadcastBytes(flag, elem_count*sizeof(double), buffer);   
}

/*********************************************************************/
void pcxx_AlienReduceFloatAdd(void* buffer,int elem_count)
{
   int i, flag;

   pcxx_Barrier();

   /* get totals from children first */
   if(num_children > 0){
     while(reduction_flag == 0) pcxx_Poll();
     for(i = 0; i < elem_count; i++) {
       ((float*)buffer)[i] += ((float*)ReceiveBuffer)[i];
     }
   }
   if(num_children > 1){
     while(reduction_flag < 2) pcxx_Poll();
     for(i = 0; i < elem_count; i++) {
       ((float*)buffer)[i] += ((float*)ReceiveBuffer)[i];
     }
   }
   reduction_flag = 0;

   /* send totals to my parent. */
   if(MyNodeNumber != 0)
     MySend(ReduceMsg, buffer, elem_count*sizeof(float), parent_id);

   /* now share the totals.  processor zero broadcasts */
   flag = 1;
   if(MyNodeNumber == 0) flag = 0;
   pcxx_BroadcastBytes(flag, elem_count*sizeof(float), buffer);   
}

/*********************************************************************/
void pcxx_AlienReduceIntAdd(void* buffer,int elem_count)
{
   int i, flag;

   pcxx_Barrier();

   /* get totals from children first */
   if(num_children > 0){
     while(reduction_flag == 0) pcxx_Poll();
     for(i = 0; i < elem_count; i++) {
       ((int*)buffer)[i] += ((int*)ReceiveBuffer)[i];
     }
   }
   if(num_children > 1){
     while(reduction_flag < 2) pcxx_Poll();
     for(i = 0; i < elem_count; i++) {
       ((int*)buffer)[i] += ((int*)ReceiveBuffer)[i];
     }
   }
   reduction_flag = 0;

   /* send totals to my parent. */
   if(MyNodeNumber != 0)
     MySend(ReduceMsg, buffer, elem_count*sizeof(int), parent_id);

   /* now share the totals.  processor zero broadcasts */
   flag = 1;
   if(MyNodeNumber == 0) flag = 0;
   pcxx_BroadcastBytes(flag, elem_count*sizeof(int), buffer);   
}

/*********************************************************************/
void pcxx_AlienReduceIntMax(void* buffer,int elem_count)
{
   int i, flag;

   pcxx_Barrier();

   /* get result from children first */
   if(num_children > 0){
     while(reduction_flag == 0) pcxx_Poll();
     for(i = 0; i < elem_count; i++) {
       if(((int*)buffer)[i] < ((int*)ReceiveBuffer)[i]) 
	 ((int*)buffer)[i] = ((int*)ReceiveBuffer)[i];
     }
   }
   if(num_children > 1){
     while(reduction_flag < 2) pcxx_Poll();
     for(i = 0; i < elem_count; i++) {
       if(((int*)buffer)[i] < ((int*)ReceiveBuffer)[i]) 
	 ((int*)buffer)[i] = ((int*)ReceiveBuffer)[i];
     }
   }
   reduction_flag = 0;

   /* send result to my parent. */
   if(MyNodeNumber != 0)
     MySend(ReduceMsg, buffer, elem_count*sizeof(int), parent_id);

   /* now share the result.  processor zero broadcasts */
   flag = 1;
   if(MyNodeNumber == 0) flag = 0;
   pcxx_BroadcastBytes(flag, elem_count*sizeof(int), buffer);   
}

/*********************************************************************/
void pcxx_AlienReduceFloatMax(void* buffer,int elem_count)
{
   int i, flag;

   pcxx_Barrier();

   /* get result from children first */
   if(num_children > 0){
     while(reduction_flag == 0) pcxx_Poll();
     for(i = 0; i < elem_count; i++) {
       if(((float*)buffer)[i] < ((float*)ReceiveBuffer)[i]) 
	 ((float*)buffer)[i] = ((float*)ReceiveBuffer)[i];
     }
   }
   if(num_children > 1){
     while(reduction_flag < 2) pcxx_Poll();
     for(i = 0; i < elem_count; i++) {
       if(((float*)buffer)[i] < ((float*)ReceiveBuffer)[i]) 
	 ((float*)buffer)[i] = ((float*)ReceiveBuffer)[i];
     }
   }
   reduction_flag = 0;

   /* send result to my parent. */
   if(MyNodeNumber != 0)
     MySend(ReduceMsg, buffer, elem_count*sizeof(float), parent_id);

   /* now share the result.  processor zero broadcasts */
   flag = 1;
   if(MyNodeNumber == 0) flag = 0;
   pcxx_BroadcastBytes(flag, elem_count*sizeof(float), buffer);   
}


/*********************************************************************/
void pcxx_AlienReduceDoubleMax(void* buffer,int elem_count)
{
   int i, flag;

   pcxx_Barrier();

   /* get totals from children first */
   if(num_children > 0){
     while(reduction_flag == 0) pcxx_Poll();
     for(i = 0; i < elem_count; i++) {
       if(((double*)buffer)[i] < ((double*)ReceiveBuffer)[i]) 
	 ((double*)buffer)[i] = ((double*)ReceiveBuffer)[i];
     }
   }
   if(num_children > 1){
     while(reduction_flag < 2) pcxx_Poll();
     for(i = 0; i < elem_count; i++) {
       if(((double*)buffer)[i] < ((double*)ReceiveBuffer)[i]) 
	 ((double*)buffer)[i] = ((double*)ReceiveBuffer)[i];
     }
   }
   reduction_flag = 0;

   /* send result to my parent. */
   if(MyNodeNumber != 0)
     MySend(ReduceMsg, buffer, elem_count*sizeof(double), parent_id);

   /* now share the result.  processor zero broadcasts */
   flag = 1;
   if(MyNodeNumber == 0) flag = 0;
   pcxx_BroadcastBytes(flag, elem_count*sizeof(double), buffer);   
}


/*********************************************************************/
void pcxx_AlienReduceIntMin(void* buffer,int elem_count)
{
   int i, flag;

   pcxx_Barrier();

   /* get result from children first */
   if(num_children > 0){
     while(reduction_flag == 0) pcxx_Poll();
     for(i = 0; i < elem_count; i++) {
       if(((int*)buffer)[i] > ((int*)ReceiveBuffer)[i]) 
	 ((int*)buffer)[i] = ((int*)ReceiveBuffer)[i];
     }
   }
   if(num_children > 1){
     while(reduction_flag < 2) pcxx_Poll();
     for(i = 0; i < elem_count; i++) {
       if(((int*)buffer)[i] > ((int*)ReceiveBuffer)[i]) 
	 ((int*)buffer)[i] = ((int*)ReceiveBuffer)[i];
     }
   }
   reduction_flag = 0;

   /* send result to my parent. */
   if(MyNodeNumber != 0)
     MySend(ReduceMsg, buffer, elem_count*sizeof(int), parent_id);

   /* now share the result.  processor zero broadcasts */
   flag = 1;
   if(MyNodeNumber == 0) flag = 0;
   pcxx_BroadcastBytes(flag, elem_count*sizeof(int), buffer);   
}

/*********************************************************************/
void pcxx_AlienReduceFloatMin(void* buffer,int elem_count)
{
   int i, flag;

   pcxx_Barrier();

   /* get result from children first */
   if(num_children > 0){
     while(reduction_flag == 0) pcxx_Poll();
     for(i = 0; i < elem_count; i++) {
       if(((float*)buffer)[i] > ((float*)ReceiveBuffer)[i]) 
	 ((float*)buffer)[i] = ((float*)ReceiveBuffer)[i];
     }
   }
   if(num_children > 1){
     while(reduction_flag < 2) pcxx_Poll();
     for(i = 0; i < elem_count; i++) {
       if(((float*)buffer)[i] > ((float*)ReceiveBuffer)[i]) 
	 ((float*)buffer)[i] = ((float*)ReceiveBuffer)[i];
     }
   }
   reduction_flag = 0;

   /* send result to my parent. */
   if(MyNodeNumber != 0)
     MySend(ReduceMsg, buffer, elem_count*sizeof(float), parent_id);

   /* now share the result.  processor zero broadcasts */
   flag = 1;
   if(MyNodeNumber == 0) flag = 0;
   pcxx_BroadcastBytes(flag, elem_count*sizeof(float), buffer);   
}


/*********************************************************************/
void pcxx_AlienReduceDoubleMin(void* buffer,int elem_count)
{
   int i, flag;

   pcxx_Barrier();

   /* get totals from children first */
   if(num_children > 0){
     while(reduction_flag == 0) pcxx_Poll();
     for(i = 0; i < elem_count; i++) {
       if(((double*)buffer)[i] > ((double*)ReceiveBuffer)[i]) 
	 ((double*)buffer)[i] = ((double*)ReceiveBuffer)[i];
     }
   }
   if(num_children > 1){
     while(reduction_flag < 2) pcxx_Poll();
     for(i = 0; i < elem_count; i++) {
       if(((double*)buffer)[i] > ((double*)ReceiveBuffer)[i]) 
	 ((double*)buffer)[i] = ((double*)ReceiveBuffer)[i];
     }
   }
   reduction_flag = 0;

   /* send result to my parent. */
   if(MyNodeNumber != 0)
     MySend(ReduceMsg, buffer, elem_count*sizeof(double), parent_id);

   /* now share the result.  processor zero broadcasts */
   flag = 1;
   if(MyNodeNumber == 0) flag = 0;
   pcxx_BroadcastBytes(flag, elem_count*sizeof(double), buffer);   
}


/*-----------------------------------------------------------------------
  Remote execution related functions and handlers

  Shelby Yang (3/24/94)
  ------------------------------------------------------------------------*/
extern void RemoteFunctionList(void *element_address,int function_number,
			       void *arg_buffer);

void end_of_remote_handler(int type, int from)
{
  /* 
    this handler is called on the processor that requests remote
    execution.  in the future this handler will also handle the return 
    value of a remote function. 
  */
  long tag_location;
  int *p;

  tag_location = *((long *) ReceiveBuffer);

  if(tag_location != 0) {
    p = (int*)tag_location;
    (*p)++;
  }

  remote_execute_flag --;
}

void remote_execute_handler(int type,int from)
{
  int dest_node, element_index; 
  int Col_Id, function_number;
  long tag_location;
  char *p;

  /* decode the receiving buffer */
  dest_node       = ((int*)ReceiveBuffer)[0];
  Col_Id          = ((int*)ReceiveBuffer)[1];
  element_index   = ((int*)ReceiveBuffer)[2];
  function_number = ((int*)ReceiveBuffer)[3];

  p = (char*)(ReceiveBuffer+6*(sizeof(int)/sizeof(char)));
  /* call compiler generated function list */
  RemoteFunctionList((void*)TableOfCollection[Col_Id]->Table[element_index], 
		     function_number, (void*) p);

  tag_location = *((long*)(ReceiveBuffer+4*sizeof(int)));
  /* reply to indicate remote execution is completed */
  MySend(EndOfRemoteExecuteMsg, (char*)&tag_location, sizeof(long), dest_node);
}

void AskForRemoteExecute(int CollectionNumber,int element_index,
			 int function_number,char *arg_buffer,int size,
			 long tag_location)
{
  /* request for remote execution. size is the size of send_buffer in bytes */
  int where, buffer_size, *p;
  int *pt;

  where = TableOfCollection[CollectionNumber]->Where[element_index];

  if(where == MyNodeNumber) {
    /* if element is local, execute the function */
    RemoteFunctionList((void*)TableOfCollection[CollectionNumber]->
                       Table[element_index], 
                       function_number, (void*)arg_buffer);
    if(tag_location != 0) {
      pt = (int*)tag_location;
      (*pt)++;
    }
    return;
  }

  p = (int*)SendBuffer;
  p[0] = MyNodeNumber;
  p[1] = CollectionNumber;
  p[2] = element_index;
  p[3] = function_number;

  *(long*)((char*)SendBuffer+4*sizeof(int)) = tag_location;
 
  buffer_size = size + 6*sizeof(int);
  if(buffer_size > MAXBUFSIZE) {
    Message("AskForRemoteExecute(): Reqires a too large buffer\n",FATAL);
  }
  /* copy arg_buffer to SendBuffer. make sure buffer is 8 bytes packed */
  memcpy(SendBuffer+6*sizeof(int),arg_buffer,size);

  remote_execute_flag ++;
  MySend(RemoteExecuteMsg, SendBuffer, buffer_size, where);
}

