/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/*==============================================================*/
/* ksrlib : KSR1 master-slave run-time support library          */
/*==============================================================*/

#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>

__private int MyNodeNumber;

#include "GlobalStruct.h"
#include "pcxx_events_def.h"
#include "pcxx_timers.h"
#include "profiler_defs.h"

/* these numbers have to be the same as they are in KernelCom.c for
   shared memory machines.  MAX_PROC has to be >= MyTotalNodes */

#define MAXBUFF 2048
#define MAX_PROC 128

extern pcxx_PARAM_t pcxx_param;
extern void New_Main(int);
void SpecialSyncbarrier();

char *BroadcastBuff = NULL;         /* a place to point to broadcast values */
char *ReduceBuff[MAX_PROC];

int   MyTotalNodes = -1;            /* number of processors */
int   pcaw_inseq   = 1;


pthread_barrier_t KSR_gen_barr;  /* generic barrier */
pthread_mutex_t  gate;           /* Lock that restricts access to in_barrier */
pthread_cond_t   latch;          /* Condition variable for workers*/
volatile int in_barrier;         /* Condition the workers have to wait for */
processor_name_t proclist[MAX_PROC];


/*
 @ SMRTSSetup () 
 */

int SMRTSSetup ()
{
  int i;
  int avail;

  pcxx_EvInit (pcxx_param.tracefile);
  PCXX_EVENT (PCXX_EC_BASIC, PCXX_SETUP_ENTER, pcxx_param.numproc);
  
  /* set number of processors */
  avail = pthread_procsetinfo (proclist, MAX_PROC);
  if ( pcxx_param.numproc > avail ) {
    fprintf (stderr, "warning: only %d cells available\n", avail);
    pcxx_param.numproc = avail;
  }
  MyTotalNodes = pcxx_param.numproc;
  
  /* initialize */ 
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_SETUP_INIT, 0);

  pthread_barrier_init(&KSR_gen_barr, pthread_barrierattr_default, MyTotalNodes);
  pthread_mutex_init (&gate, (pthread_mutexattr_t) pthread_attr_default);
  pthread_cond_init (&latch, (pthread_condattr_t) pthread_attr_default);
  in_barrier = 0;

   /* allocate shared reduction buffer */
  for(i = 0; i < MyTotalNodes; i++) {
    ReduceBuff[i] = (char*) malloc(MAXBUFF);
  }
 
  PCXX_EVENT (PCXX_EC_BASIC, PCXX_SETUP_EXIT, MyTotalNodes);
  return (0);
}


/*
 MSNewProc : book-keeping for newly-created process
 */

void MSNewProc (int id)
{
  /* record my ID */
  MyNodeNumber = id;

  pcxx_EvInit (pcxx_param.tracefile);
  PCXX_EVENT (PCXX_EC_BASIC, PCXX_MAIN_ENTER, 0);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_NEWPROC_ENTER, 0);

  SpecialSyncbarrier();

  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_NEWPROC_EXIT, id);
}


/*
 MSFinishProc : book-keeping for terminating process
 */

void MSFinishProc (void)
{
  PCXX_EVENT (PCXX_EC_BASIC, PCXX_MAIN_EXIT, 0);
  pcxx_EvClose ();
}


/*
 @ SMRTSFork : start parallel processing
 */

int SMRTSFork (argc, argv)
int argc;
char **argv;
{
  pthread_t new_p[MAX_PROC];
  int i;
  pthread_attr_t attr;

  /* fork */
  PCXX_EVENT (PCXX_EC_BASIC, PCXX_FORK_ENTER, 0);

  pcxx_InitClocks ();

  pthread_attr_create(&attr);
  pthread_attr_setstacksize(&attr, pcxx_param.stacksize);

  /* startup the threads */
  for (i=0; i<MyTotalNodes; i++) {
    PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_FORK_FORK, i);
    if ( pthread_create (&(new_p[i]), attr, New_Main, (void *)i) == -1) {
      perror ("create");
      exit (1);
    }
    PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_FORK_BIND, (long) proclist[i]);
    if ( pthread_move (new_p[i], proclist[i], TH_BOUND_PROCESSOR) == -1 ) {
      perror ("move");
      exit (1);
    }
  }
 
  /* wait for workers to finish */
  for (i=0; i<MyTotalNodes; i++) {
    if ( pthread_join (new_p[i], NULL) == -1 ) {
      perror ("join");
      exit (1);
    }
    PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_FORK_JOIN, i);
  }

  PCXX_EVENT (PCXX_EC_BASIC, PCXX_FORK_EXIT, 0);
  pcxx_EvClose ();
  return (0);
}

/*
 @ pcxx_NodeNumber : which processor I am
 */

int pcxx_NodeNumber()
{
  return MyNodeNumber;
}

/*
 @ IamtheMainThread 
 */

int IamtheMainThread()
{
  return pcaw_inseq;
}

/*
 @ barrier : wait at general barrier
 */

void SpecialSyncbarrier()
{
# ifdef TRACING_ON
  pthread_mutex_lock (&gate);
  pcxx_Event (PCXX_IN_BARRIER, in_barrier);
  in_barrier++;
  if ( in_barrier == MyTotalNodes ) {
    in_barrier=0;
    pthread_cond_broadcast (&latch);
  }
  else {
    pthread_cond_wait (&latch, &gate);
  };
  pthread_mutex_unlock (&gate);
# endif
}

void pcaw_Barrier_Uninstrumented()
{
  pthread_mutex_lock (&gate);
  in_barrier++;
  if ( in_barrier == MyTotalNodes ) {
    in_barrier=0;
    pthread_cond_broadcast (&latch);
  }
  else {
    pthread_cond_wait (&latch, &gate);
  };
  pthread_mutex_unlock (&gate);
  /*pthread_barrier_checkout (&KSR_gen_barr, MyNodeNumber);
  pthread_barrier_checkin  (&KSR_gen_barr, MyNodeNumber);*/
}

void pcaw_Barrier()
{
  PROFILE_ENTER(PROFILE_Barrier);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_BARRIER_ENTER, 0);
  pcaw_Barrier_Uninstrumented();
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_BARRIER_EXIT, 0);
  PROFILE_EXIT(PROFILE_Barrier);
}

void pcxx_Barrier()
{
  if(!pcaw_inseq)
    pcaw_Barrier();
}


/*
 pcxx_BroadcastBytes 
 */

void pcxx_BroadcastBytes(source, length, buffer)
int source;
int length;
void * buffer;
{
  int i;
  char *c, *p;

  /* This barrier is needed to prevent BroadcastBuff from being overwritten
     when two consecutive pcxx_BroadcastBytes are called */
  pcxx_Barrier();

  p = (char *) buffer;

  if (source == 0) BroadcastBuff = buffer;
  pcxx_Barrier();
  if (source != 0){
    c = BroadcastBuff;
    for (i=0; i<length; i++) p[i] = c[i];
  }
}

