/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/*==============================================================*/
/* ksrlib : KSR1 run-time support library			*/
/*==============================================================*/

#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>

#include "GlobalStruct.h"
#include "pcxx_events_def.h"
#include "pcxx_timers.h"
#include "profiler_defs.h"

extern pcxx_PARAM_t pcxx_param;

/* these numbers have to be the same as they are in KernelCom.c for 
   shared memory machines.  MAX_PROC has to be >= MyTotalNodes */
#define MAXBUFF 4096
#define MAX_PROC 128

#ifndef PCXX_MAXPROCS
#  define PCXX_MAXPROCS 1088
#endif

extern void New_Main (int);
extern void SpecialSyncbarrier();

/* Why does this need to be shared? */
pthread_barrier_t KSR_gen_barr;  /* generic barrier */
pthread_mutex_t  gate;           /* Lock that restricts access to in_barrier */
pthread_cond_t   latch;          /* Condition variable for workers*/
volatile int in_barrier;         /* Condition the workers have to wait for */
int   MyTotalNodes = -1;            /* number of processors */
char BroadcastBuff[MAXBUFF];    /* a place to store broacast values */
char *ReduceBuff[MAX_PROC];     /* a place to store reduction values */

__private int MyNodeNumber = PCXX_MAXPROCS-1;
__private int KSR_Loc_I =  1;       /* loc array index */

processor_name_t proclist[PCXX_MAXPROCS];

/*
 @ SMRTSSetup () : set up to run on the KSR
 */

int SMRTSSetup ()
{
  int avail, i;

  MyNodeNumber = PCXX_MAXPROCS-1;

  pcxx_EvInit (pcxx_param.tracefile);
  PCXX_EVENT (PCXX_EC_BASIC, PCXX_SETUP_ENTER, pcxx_param.numproc);

  /* set number of processors */
  avail = pthread_procsetinfo (proclist, PCXX_MAXPROCS);
  if ( pcxx_param.numproc > avail ) {
    fprintf (stderr, "warning: only %d cells available\n");
    pcxx_param.numproc = avail;
  }
  MyTotalNodes = pcxx_param.numproc;

  /* initialise synchronization variables */
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_SETUP_INIT, 0);
  pthread_barrier_init(&KSR_gen_barr, pthread_barrierattr_default, MyTotalNodes);
  pthread_mutex_init (&gate, (pthread_mutexattr_t) pthread_attr_default);
  pthread_cond_init (&latch, (pthread_condattr_t) pthread_attr_default);
  in_barrier = 0;

  /* allocate shared reduction buffer */
  if(MyTotalNodes > MAX_PROC) {
    fprintf(stderr,"Error in sgilib.c: MAX_PROC < number of processors.\n");
    exit(0);
  }
  for(i = 0; i < MyTotalNodes; i++) {
    ReduceBuff[i] = (char*) malloc(MAXBUFF);
  }

  PCXX_EVENT (PCXX_EC_BASIC, PCXX_SETUP_EXIT, MyTotalNodes);
  return (0);
}

/*
 @ SMRTSFork : start parallel processing
 */

int SMRTSFork (argc, argv)
int argc;
char **argv;
{
  pthread_t new_p[PCXX_MAXPROCS];
  int i;

  /* fork */
  PCXX_EVENT (PCXX_EC_BASIC, PCXX_FORK_ENTER, 0);
  for (i=0; i<MyTotalNodes; i++) {
    PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_FORK_FORK, i);
    if ( pthread_create (&(new_p[i]),
                         pthread_attr_default, New_Main, (void *)i) == -1) {
      perror ("create"), exit (1);
    }
    PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_FORK_BIND, (long) proclist[i]);
    if ( pthread_move (new_p[i], proclist[i], TH_BOUND_PROCESSOR) == -1 ) {
      perror ("move"), exit (1);
    }
  }

  /* wait for workers to finish */
  for (i=0; i<MyTotalNodes; i++) {
    if ( pthread_join (new_p[i], NULL) == -1 ) {
      perror ("join"), exit (1);
    }
    PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_FORK_JOIN, i);
  }

  PCXX_EVENT (PCXX_EC_BASIC, PCXX_FORK_EXIT, 0);
  pcxx_EvClose ();

  return (0);
}

/*
 @ SMRTSNewProc : more book-keeping for newly-created process
 */

int SMRTSNewProc (int id)
{
  /* record my ID */
  MyNodeNumber = id;

  pcxx_EvInit (pcxx_param.tracefile);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_NEWPROC_ENTER, 0);

  SpecialSyncbarrier();
  SpecialSyncbarrier();

  KSR_Loc_I = 1;

  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_NEWPROC_EXIT, MyNodeNumber);
  return (MyNodeNumber);
}

/*
 @ barrier : wait at general barrier
 */

void SpecialSyncbarrier()
{
# ifdef TRACING_ON
  pthread_mutex_lock (&gate);
  pcxx_Event (PCXX_IN_BARRIER, in_barrier);
  in_barrier++;
  if ( in_barrier == MyTotalNodes ) {
    in_barrier=0;
    pthread_cond_broadcast (&latch);
  }
  else {
    pthread_cond_wait (&latch, &gate);
  };
  pthread_mutex_unlock (&gate);
# endif
}

void pcxx_Barrier()
{
  PROFILE_ENTER(PROFILE_Barrier);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_BARRIER_ENTER, 0);
  pthread_barrier_checkout (&KSR_gen_barr, MyNodeNumber);
  pthread_barrier_checkin  (&KSR_gen_barr, MyNodeNumber);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_BARRIER_EXIT, 0);
  PROFILE_EXIT(PROFILE_Barrier);
}

/*
 @ pcxx_BrodcastBytes : sends messages from 0 to all other nodes
 */

void pcxx_BroadcastBytes(int source, int length, void * buffer)
{
  int i;
  char *c, *p;

  /* This barrier is needed to prevent BroadcastBuff from being overwritten
     when two consecutive pcxx_BroadcastBytes are called */
  pcxx_Barrier();

  if ( length > MAXBUFF ) Message("Broadcast Too BIG!! ",FATAL);

  c = BroadcastBuff; p = (char *) buffer;

  if (source == 0) for (i=0; i<length; i++) c[i] = p[i];
  pcxx_Barrier();
  if (source != 0) for (i=0; i<length; i++) p[i] = c[i];
}


void SMRTSDealloc (ptr)
void *ptr;
{
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_DEALLOC_ENTER, (long) ptr);
  free (ptr);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_DEALLOC_EXIT, 0);
}

