/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/*==============================================================*/
/* ptlib : pthreads run-time support library                    */
/*==============================================================*/

#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>

#include "GlobalStruct.h"
#include "pcxx_events_def.h"
#include "pcxx_timers.h"
#include "profiler_defs.h"

/* these numbers have to be the same as they are in KernelCom.c for 
   shared memory machines.  MAX_PROC has to be >= MyTotalNodes */

#define MAXBUFF 2048
#define MAX_PROC 128

#define pthread_attr_default NULL

extern pcxx_PARAM_t pcxx_param;
extern void New_Main(int);

char *BroadcastBuff = NULL;         /* a place to point to broadcast values */
char *ReduceBuff[MAX_PROC];   

int   MyTotalNodes = -1;            /* number of processors */
int   pcaw_inseq   = 1;


pthread_mutex_t  gate;           /* Lock that restricts access to in_barrier */
pthread_cond_t   latch;          /* Condition variable for workers*/
volatile int in_barrier;         /* Condition the workers have to wait for */

pthread_key_t pcpt_id_key;


/*
 @ SMRTSSetup () 
 */

int SMRTSSetup ()
{
  int avail = MAX_PROC;
  int i;

  /* uniproc version of timer and event-stuff */
  pcxx_InitClocks ();
  pcxx_EvInit (pcxx_param.tracefile);

  PCXX_EVENT (PCXX_EC_BASIC, PCXX_SETUP_ENTER, pcxx_param.numproc);

  /* set number of processors */
  if ( pcxx_param.numproc > avail ) {
    fprintf (stderr, "warning: only %d cells available\n", avail);
    pcxx_param.numproc = avail;
  }
  MyTotalNodes = pcxx_param.numproc;
  
  /* initialize */ 
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_SETUP_INIT, 0);

  pthread_key_create(&pcpt_id_key, NULL);
  pthread_mutex_init (&gate, pthread_attr_default);
  pthread_cond_init (&latch, pthread_attr_default);
  in_barrier = 0;

  /* allocate shared reduction buffer */
  for(i = 0; i < MyTotalNodes; i++) {
    ReduceBuff[i] = (char*) malloc(MAXBUFF);
  }
  
  PCXX_EVENT (PCXX_EC_BASIC, PCXX_SETUP_EXIT, MyTotalNodes);
  return (0);
}


/*
 MSNewProc : book-keeping for newly-created process
 */

void MSNewProc (int id)
{
  /* record my ID */
  pthread_setspecific(pcpt_id_key, (any_t)id);

  PCXX_EVENT (PCXX_EC_BASIC, PCXX_MAIN_ENTER, 0);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_NEWPROC_ENTER, 0);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_NEWPROC_EXIT, id);
}


/*
 MSFinishProc : book-keeping for terminating process
 */

void MSFinishProc (void)
{
  PCXX_EVENT (PCXX_EC_BASIC, PCXX_MAIN_EXIT, 0);
}


/*
 @ SMRTSFork : start parallel processing
 */

int SMRTSFork (argc, argv)
int argc;
char **argv;
{
  pthread_t new_p[MAX_PROC];
  int i;
  pthread_attr_t attr;

  /* fork */
  PCXX_EVENT (PCXX_EC_BASIC, PCXX_FORK_ENTER, 0);

  pthread_attr_init(&attr);
  pthread_attr_setstacksize(&attr, pcxx_param.stacksize);

  /* startup the threads */
  for (i=1; i<MyTotalNodes; i++) {
    if ( pthread_create (&(new_p[i]), &attr,New_Main, (void *)i) == -1) {
      perror ("create");
      exit (1);
    }
  }

  New_Main(0);

  /* wait for workers to finish */
  for (i=1; i<MyTotalNodes; i++) {
    if ( pthread_join (new_p[i], NULL) == -1 ) {
      perror ("join");
      exit (1);
    }
  }

  PCXX_EVENT (PCXX_EC_BASIC, PCXX_FORK_EXIT, 0);
  pcxx_EvClose ();
  return (0);
}


/*
 @ pcxx_NodeNumber : which processor I am
 */

int pcxx_NodeNumber()
{
  int my_id=0;

  if(!pcaw_inseq) {
    pthread_getspecific(pcpt_id_key, (any_t *)&my_id);
  }
  return my_id;
}

/*
 @ IamtheMainThread 
 */

int IamtheMainThread()
{
  return pcaw_inseq;
}

/*
   To be used within New_Main before calling MSNewProc, because
   until then, node id is not set, so instrumentation cannot be
   used.
*/
void pcaw_Barrier_Uninstrumented()
{
  pthread_mutex_lock (&gate);
  in_barrier++;
  if ( in_barrier == MyTotalNodes ) {
    in_barrier=0;
    pthread_cond_broadcast (&latch);
  }
  else {
    pthread_cond_wait (&latch, &gate);
  };
  pthread_mutex_unlock (&gate);
}

/*
 barrier : wait at general barrier, must be called only in parallel sections
 */

void pcaw_Barrier()
{
  PROFILE_ENTER(PROFILE_Barrier);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_BARRIER_ENTER, 0);
  pcaw_Barrier_Uninstrumented();
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_BARRIER_EXIT, 0);
  PROFILE_EXIT(PROFILE_Barrier);
}

/* Can be used even in sequential sections */
void pcxx_Barrier()
{
  if(!pcaw_inseq)
    pcaw_Barrier();
}


/*
 pcxx_BroadcastBytes 
 */

void pcxx_BroadcastBytes(source, length, buffer)
int source;
int length;
void * buffer;
{
  int i;
  char *c, *p;

  /* This barrier is needed to prevent BroadcastBuff from being overwritten
     when two consecutive pcxx_BroadcastBytes are called */
  pcxx_Barrier();

  p = (char *) buffer;

  if (source == 0) BroadcastBuff = buffer;
  pcxx_Barrier();
  if (source != 0){
    c = BroadcastBuff;
    for (i=0; i<length; i++) p[i] = c[i];
  }
}

