/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/*==============================================================*/
/* seqlib : Sequent run-time support library			*/
/*==============================================================*/

#include <stdio.h>
#include <parallel/parallel.h>
#include <parallel/microtask.h>

#include "GlobalStruct.h"
#include "pcxx_events_def.h"
#include "pcxx_timers.h"
#include "profiler_defs.h"
#include "brkExecInc.h" 

extern pcxx_PARAM_t pcxx_param;

extern void New_Main();

/* these numbers have to be the same as they are in KernelCom.c for 
   shared memory machines.  MAX_PROC has to be >= MyTotalNodes */
#define MAXBUFF 4096
#define MAX_PROC 128

# ifdef sequent
  /* -- sequent running Dynix -- */
  shared sbarrier_t Seq_Gen_Barr;      /* general-purpose barrier */
  shared char BroadcastBuff[MAXBUFF];  /* a place to store broacast values */
  char *ReduceBuff[MAX_PROC];          /* a place to store reduction values */
# else
  /* -- sequent running Dynix/ptx -- */
# pragma sequent_shared
  sbarrier_t Seq_Gen_Barr;	       /* general-purpose barrier */
  char BroadcastBuff[MAXBUFF];         /* a place to store broacast values */
  char *ReduceBuff[MAX_PROC];          /* a place to store reduction values */
# pragma sequent_shared_end

  extern void *memcpy();
# endif

int  MyTotalNodes = -1;	/* number of processors */
int  MyNodeNumber = -1;	/* who I am */

/*
 @ SMRTSSetup() : set up to run on the Sequent
 */

int SMRTSSetup()
{
  int i;
  MyNodeNumber = m_get_myid();
  pcxx_EvInit (pcxx_param.tracefile);
  PCXX_EVENT (PCXX_EC_BASIC, PCXX_SETUP_ENTER, pcxx_param.numproc);

  /* allocate processors */
  if (pcxx_param.numproc >= cpus_online()){
    MyTotalNodes = cpus_online() - 1;
  } else {
    MyTotalNodes = pcxx_param.numproc;
  }
  m_set_procs(MyTotalNodes);

  /* initialise barriers */
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_SETUP_INIT, 0);
  s_init_barrier(&Seq_Gen_Barr, MyTotalNodes);

  /* allocate shared reduction buffer */
  if(MyTotalNodes > MAX_PROC) {
    fprintf(stderr,"Error in sgilib.c: MAX_PROC < number of processors.\n");
    exit(0);
  }
  for(i = 0; i < MyTotalNodes; i++) {
    ReduceBuff[i] = (char*) shmalloc(MAXBUFF);
  }

  PCXX_EVENT (PCXX_EC_BASIC, PCXX_SETUP_EXIT, MyTotalNodes);
  return(0);
}

/*
 @ SMRTSFork : start parallel processing
 */

int SMRTSFork(argc, argv)
  int argc;
  char **argv;
{
  /* fork */
  PCXX_EVENT (PCXX_EC_BASIC, PCXX_FORK_ENTER, 0);
  if (m_fork(New_Main, 0) < 0){
    fprintf(stderr, "m_fork failed\n");
    exit(1);
  }

  /* finished */
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_FORK_KILL, 0);
  m_kill_procs();
  PCXX_EVENT (PCXX_EC_BASIC, PCXX_FORK_EXIT, 0);
  pcxx_EvClose ();
  return(0);
}

/*
 @ SMRTSNewProc : more book-keeping for newly-created process
 */

int SMRTSNewProc(num)
  int num;
{
  /* record my ID */
  MyNodeNumber = m_get_myid();
  pcxx_InitClocks ();

  pcxx_EvInit (pcxx_param.tracefile);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_NEWPROC_ENTER, 0);

  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_NEWPROC_EXIT, MyNodeNumber);
  return(MyNodeNumber);
}

/*
 @ barrier : wait at general barrier
 */

void pcxx_Barrier()
{
  PROFILE_ENTER(PROFILE_Barrier);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_BARRIER_ENTER, 0);
  brk_BREAKPOINT_EXECUTIVE(); 
  s_wait_barrier(&Seq_Gen_Barr);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_BARRIER_EXIT, 0);
  PROFILE_EXIT(PROFILE_Barrier);
}

/*
 @ pcxx_BrodcastBytes : sends messages from 0 to all other nodes
 */

void pcxx_BroadcastBytes(source, length, buffer)
int source;
int length;
void * buffer;
{
  int i;
  char *c, *p;

  /* This barrier is needed to prevent BroadcastBuff from being overwritten
     when two consecutive pcxx_BroadcastBytes are called */
  pcxx_Barrier();

  if ( length > MAXBUFF ) Message("Broadcast Too BIG!! ",FATAL);

  c = BroadcastBuff; p = (char *) buffer;

  if (source == 0) for (i=0; i<length; i++) c[i] = p[i];
  pcxx_Barrier();
  if (source != 0) for (i=0; i<length; i++) p[i] = c[i];
}

/*
 @ SMRTSDealloc : deallocate shared memory pointers
 */

void SMRTSDealloc(ptr)
void *ptr;
{
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_DEALLOC_ENTER, (long) ptr);
  shfree(ptr);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_DEALLOC_EXIT, 0);
}

