/*
 * sbarrier.c: this file contains routines which implement a super-barrier.
 *
 */

/****************************************************************
Includes
****************************************************************/

#include <stdio.h>
#include <mach.h>
#include <sys/file.h>
#include <sys/kern_return.h>
#include <errno.h>


/****************************************************************
Defines
****************************************************************/

#define SBARRIER_ADDRESS  0x1e000000 /* location of the flag data */

#define LOGLEVELS   3
#define LOGTDESCS   8
#define LOGNPROCS   7

#define MAXLEVELS   (1 << LOGLEVELS)
#define MAXTDESCS   (1 << LOGTDESCS)
#define MAXNPROCS   (1 << LOGNPROCS)

#define LEVELSHIFT  2
#define TDESCSHIFT  (LEVELSHIFT + LOGLEVELS)
#define PHASESHIFT  (TDESCSHIFT + LOGTDESCS)
#define IPROCSHIFT  (PHASESHIFT + 1)


/****************************************************************
Interface Functions
****************************************************************/

void sbarrier_init();		/* initialize at startup */
void sbarrier_half();		/* a half barrier */


/****************************************************************
sbarrier_init() - this function must be called by all participating
processes before any other sbarrier calls.

arguments: nprocs  - the number of processors which will participate
                     in the barrier.
           nodemap - an array of integers which contains the cluster
	             relative node id for each processor in the parallel
		     job.  For many parallel jobs, this will be a unity
		     mapping.  For jobs under the gang scheduler which
		     "fills" holes in the gang scheduled cluster, this
		     will not be the case.
****************************************************************/

static char filename[256];

void sbarrier_init(nprocs, nodemap, barId)
register int nprocs;
register int *nodemap;
         int barId;
{
    int  fd;
    char *vaddr;
    int  size;
    int  offset;
    int  flags;
    register int node;
    kern_return_t kernret;
    int  clunid;

    /* Build a filename for the named shared memory file used by
     * the barrier code.
     */
    sprintf(filename, "/tmp/barrier-%d.mem", barId);

    /* Open the file */
    fd = open(filename, O_RDWR | O_CREAT, 0666);
    if (fd <= 0)
    {
	fprintf(stderr, "barrier create open: \"%s\"", filename);
	perror("");
	exit(1);
    }

    /* Construct a contigous block of memory with each subsequent
     * chunks on subsequent nodes.
     */
    vaddr = (char *) SBARRIER_ADDRESS;
    size = (1 << IPROCSHIFT);
    offset = 0;
#ifdef VM_MAPMEM_ANON
    flags = VM_MAPMEM_ANON | VM_MAPMEM_TRUNCATE | VM_MAPMEM_NOBORROW;
#else
    flags = VM_MAPMEM_TRUNCATE | VM_MAPMEM_NOBORROW;
#endif
    for (node = 0; node < nprocs; node++)
    {
	/* Map the node if necessary */
	clunid = (nodemap == NULL) ? node : nodemap[node];

	errno = 0;
	kernret = vm_mapmem(task_self(), (vm_addr_t*) &vaddr, 
			    size, flags, fd, offset, clunid);
	if (kernret != KERN_SUCCESS)
	{
	    mach_error("barrier create vm_mapmem", kernret);
	    if (errno != 0) perror("vm_mapmem");
	    exit(1);
	}

	vaddr += size;
	offset += size;
    }

}   /* end sbarrier_init */

/****************************************************************
sbarrier_shutdown() - this function should be called once only
                 to shutdown the barrier code, remove memory mapped files.

****************************************************************/

void sbarrier_shutdown()
{

    /* 
     * Build the filename for the named shared memory file used by
     * the barrier code. Remove this file.
     */
    (void) unlink(filename);

}   /* end sbarrier_shutdown */


/****************************************************************
sbarrier_half() - half of a full barrier.  The half barrier takes a
parity argument of 0 (even) or 1 (odd).  Since the half barrier is not
reentry protected each half barrier MUST be followed by a half barrier
of the opposite parity.
****************************************************************/

void sbarrier_half(tdesc, tindex, tsize, iproc, phase)
register int tdesc;
register int tindex;
register int tsize;
register int iproc;
register int phase;
{
    register int *targetflag;	/* a pointer to the target flag */
    register int *myflag;	/* a pointer to my current flag */
    register int base;		/* a base value into the memory */
    register int level;		/* current barrier level */
    register int achieved;	/* number of processes "in" so far */
    register int end;		/* the first proc index out of our team */
    register int notify;	/* the other proc index to notify */

    base = SBARRIER_ADDRESS + (tdesc << TDESCSHIFT) +
	   (phase << PHASESHIFT);

    myflag = (int *)(base + (iproc << IPROCSHIFT));

    level = 0;			/* start w/ level 0 */
    achieved = 1;		/* you start with your self */
    end = iproc - tindex + tsize; /* this is just outside the team */

    /* Rendezvous until the number of unique processes you have
     * synchronized with has exceeded the total number of processes
     * in the team.
     */
    while (achieved < tsize)
    {
	/* Which processor should we notify on this level? */
	notify = iproc + achieved;
	if (notify >= end)
	    notify -= tsize;

	/* Notify the other processor */
	targetflag = (int *)(base + (notify << IPROCSHIFT) +
			     (level << LEVELSHIFT));
	*targetflag = 1;

	/* Wait to be notified */
	while (*myflag == 0)
	    ;

	/* Clear the notification */
	*myflag = 0;

	/* On to the next level */
	myflag++;		/* point to the next level's flag */
	level++;
	achieved <<= 1;		/* each rendezvous doubles the count */
    }

}   /* end sbarrier_half */
