/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/*********************************************************************/
/*                Parallel run-time library                          */
/*                                                                   */
/*                Passive Message Passing                            */
/*                                                                   */
/*                      Pete Beckman                                 */
/*                                                                   */
/* Contributors: Francois Bodin *** Dennis Gannon                    */
/*                Bernd Mohr *** Shelby Yang                         */
/*********************************************************************/

/*****************************/
/* System Includes           */
/*****************************/
#include <stdarg.h>
#include <stdio.h>

/*****************************/
/* Prototypes (forward decl) */
/*****************************/
static void MyRecv(int type, int mid, char *addr, int size, int from_whom);
static void MySend(int type, int mid, char * buffer, int len, int ToWhom);
static void MySSend(int type, int mid, char * buffer, int len, int ToWhom);
void tulip_Abort();
void tulip_Poll();

/*****************************/
/* Our Include Files         */
/*****************************/
#include "pcxx_machines.h"

/*****************************/
/* System Macros             */
/*****************************/
#define DONE              -1
#define NOMID              0

/*****************************/
/* Message Types             */
/*****************************/
#define BarrierFanOutMsg 100 /* Message to notify that work can resume */
#define BarrierFanInMsg  101 /* Fan into one node to sync */
#define BroadcastMsg     110 /* All processors listen and receive... */
#define ReduceMsg        120 /* Reduce a value across all processors */

#define AskForDataMsg    130 /* Send this type to remote node to get data */
#define FetchBlockMsg    140 /* Partial reply to earlier AskForDataMsg */
#define EndOfFetchMsg    141 /* Completed reply to earlier AskForDataMsg */

#define PrefixStoreMsg   150 /* Send this type to prepare remote for store */
#define StoreBlockMsg    160 /* Partial block of store */
#define EndOfStoreMsg    161 /* Final block of a store message */

#define RemoteActionMsg  170 /* Send to remote to execute handler */

#define AckReplyMsg      180 /* Reply from remote that store has completed */

#define MAXMSGS          512 /* Maximum "active" message handles */
#define MAXBUFSIZE      16384 /* Maximum buffer size */

/*****************************/
/* External Variables        */
/*****************************/
int tulip_MsgBlockSize = PCXX_SUGGESTED_MSG_SIZE;

/*****************************/
/* Module Global Variables   */
/*****************************/
static char ReceiveBuffer[MAXBUFSIZE], SendBuffer[MAXBUFSIZE];
static tulip_GlobalPointer_t AddressBook[PCXX_MAXPROCS];
static char *PendingMsgs[MAXMSGS];
static int RecvMsgLen[MAXMSGS], FreeMsgIDs[MAXMSGS];
static tulip_ACK_t *AckHandle[MAXMSGS];
static int NextFreeMsgID;

static char *StoreMsgAddr[MAXMSGS*PCXX_MAXPROCS];
static int StoreMsgLen[MAXMSGS*PCXX_MAXPROCS];

static int PendingSysACKs=0;  /* Internal count of unfinished requests */
static int barrierFlag = 0, IHaveReachedBarrier = 0;
static int broadcast_flag = 0, reduction_flag = 0;
static int barrier_msg = 0, remote_execute_flag = 0, num_children;
int MyTotalNodes, MyNodeNumber;
static int parent_id, left_child, right_child;

/*****************************/
/* Structs                   */
/*****************************/
struct msg_params {
  char * address;
  int    length;
};

/*********************************************************************/
/*********************************************************************/
/*********************************************************************/
/* GENERAL FUNCTIONS */
/*********************************************************************/
/*********************************************************************/
/*********************************************************************/

/*********************************************************************/
void * tulip_SharedMalloc(int size) {
  return (void *) malloc(size);
}

/*********************************************************************/
void tulip_SharedFree(void *p) {
  free(p);
}

/*******************************************************************/
int tulip_NumContexts()
{ return MyTotalNodes; }

/*******************************************************************/
int tulip_MyContext() 
{ return MyNodeNumber; }

/*********************************************************************/
/*********************************************************************/
/* BARRIER RELATED FUNCTIONS */
/*********************************************************************/
/*********************************************************************/
/* Barrier to synchronize operations in pC++. Processors wait at the
   Barrier until all are there.  Processors need to run tulip_Poll() while
   they wait.  tulip_Barrier() must use messages on most message passing
   machines, so that tulip_Poll() can be run periodically while waiting at
   the barrier.  These routines were re-worked by P. Beckman (SP1 port)
   based on earlier designs by D.  Gannon (Paragon port).  Processors do
   a tree fan-in to the root node, then a fan-out and resume computation.
   A fast hardware/OS supported barrier is prefered, but the need to run
   tulip_Poll() while waiting in the barrier excludes most system
   implementations
   
   The Barrier works in the following way: The leaves start sending
   FanInMsg to their parents, while their parents poll messages from the
   children.  Once a processor receives messages from all of its
   children, it sends a message to its parent.  The fan-in completes when
   processor zero gets the message.  Processor zero then sends FanOutMsg
   down through the tree.  Barrier completes when all processors receive
   FanOutMsg. (SY & PHB) */

/*********************************************************************/
static void BarrierFanIn(int from)
{
  /* record the number of children fanned in */
  barrierFlag++;  
  
  /*If I am not at the barrier, I do not let BarrierFanInMsg through.*/
  if (IHaveReachedBarrier == 0) return;
  
  if((barrierFlag == 1 && num_children == 1) || 
     (barrierFlag == 2 && num_children == 2) ) {
    if(MyNodeNumber == 0) { 
      barrierFlag = DONE;
    } else {
      MySend(BarrierFanInMsg, NOMID, SendBuffer, 0, parent_id);
    }
  }
  
  return;
}

/*********************************************************************/
void tulip_Barrier()
{
  PROFILE_ENTER(PROFILE_Barrier);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_BARRIER_ENTER, 0);

  /* simply return when only a single processor is used */
  if(num_children == 0 && MyNodeNumber == 0) return;

  while(remote_execute_flag != 0) tulip_Poll();
  
  IHaveReachedBarrier = 1;  /* I have really entered a barrier */
  
  if (num_children == 0) {
    MySend(BarrierFanInMsg, NOMID, SendBuffer, 0, parent_id);
  }
  
  if((barrierFlag == 1 && num_children == 1) || 
     (barrierFlag == 2 && num_children == 2) ) 
    {
      if(MyNodeNumber == 0) { 
	barrierFlag = DONE;
      } else {
	MySend(BarrierFanInMsg, NOMID, SendBuffer, 0, parent_id);
      }
    }
  
  while(barrierFlag != DONE)
    tulip_Poll();
  
  if (num_children >= 1) {
    MySend(BarrierFanOutMsg, NOMID, SendBuffer,0,left_child);
  }
  if (num_children == 2) {
    MySend(BarrierFanOutMsg, NOMID, SendBuffer,0,right_child);
  }
  
  IHaveReachedBarrier = 0;
  barrierFlag = 0;
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_BARRIER_EXIT, 0);
  PROFILE_EXIT(PROFILE_Barrier);
}

/*********************************************************************/
/*********************************************************************/
/* DATA OPERATIONS ON GLOBAL POINTERS */
/*********************************************************************/
/*********************************************************************/

/*********************************************************************/
/* Encode/Decode message type and message ID */
/*********************************************************************/
static int mangle_msgtype(int t, int m)
{  return (t<<8|m); }

/*********************************************************************/
static int mangle_getmid(int t)
{  return (255 & t); }

/*********************************************************************/
static int mangle_gettype(int t)
{  return (t>>8); }

/*********************************************************************/
static void initMsgIDs() 
{
  int i;
  
  /* Create a list of avaible message IDs */
  for (i=0; i<MAXMSGS; i++)
    FreeMsgIDs[i]=i;
  
  NextFreeMsgID = 0;
  
}

/*********************************************************************/
/* Get the next available message ID.  This procedure finds an unused
   message ID (from FreeMsgIDs stack) and returns it.  If all message IDs
   are being used, the program is aborted */

static int NewMessageID()
{
  int i;
  
  /* check for free pointer slot */
  if (NextFreeMsgID == MAXMSGS)
    tulip_Message("Error in NewMessageID(), no free message IDs",TULIP_FATAL);
  
  i = NextFreeMsgID;
  NextFreeMsgID++;
  
  return FreeMsgIDs[i];  /* MID in table slot "i" */
}

/*********************************************************************/
/* Return an old message ID.  This procedure recycles an old, unused,
   message ID, by pushing it back onto the FreeMsgIDs stack.  */

static void RecycleMsgID(int mid)
{
  int i;
  
  if ((NextFreeMsgID-1) < 0)
    tulip_Message("Error in RecycleMsgID(): stack already full (push)",TULIP_FATAL);
  
  i = NextFreeMsgID-1;
  NextFreeMsgID--;
  
  FreeMsgIDs[i] = mid;  /* Set slot to be free MID */
}

/*********************************************************************/
/* A tulip_Get() request has *ARRIVED* from a remote processor
   (AskForDataMsg).  To satisfy the request, this processor must package
   up the requested block of memory into possibly several chunks, and
   send them to the remote node. */

static void fetch_block_handler(int from_whom, int mid)
{
  struct msg_params p;
  
  /* Extract fields from message */
  MyRecv(AskForDataMsg, mid, (char *) &p, (int) sizeof(struct msg_params), from_whom);

  /* If the requested transfer is too big, send in chunks */
  while ((tulip_MsgBlockSize > 0) && (p.length > tulip_MsgBlockSize)) {
    MySSend(FetchBlockMsg, mid, p.address, tulip_MsgBlockSize, from_whom);
    p.length -= tulip_MsgBlockSize;
    p.address += tulip_MsgBlockSize;
  }

  /* Send final chunk (or only chunk if requested block was small) */
  MySSend(EndOfFetchMsg, mid, p.address, p.length, from_whom);

/*  printf("%d: Sent EndOfFetchMsg: %d, %d, %d\n",
  tulip_MyContext(),mangle_msgtype(EndOfFetchMsg,mid),p.length,from_whom);*/
  
  return;
}

/*********************************************************************/
/* A tulip_Put() prefix has arrived (PrefixStoreMsg).  The
   prefix announces the future arrival of a store request.  It's job is
   to set the local address where the subsequent store will be unpacked,
   how big the data block will be, and in the future, it may have flags
   specifying the level of atomicity the store will have. */

static void prefix_store_handler(int from_whom, int mid)
{
  struct msg_params p;
  
/*printf("%d: psh from:%d mid:%d\n",tulip_MyContext,from_whom,mid);*/
  
  MyRecv(PrefixStoreMsg, mid, (char *) &p, (int) sizeof(struct msg_params), from_whom);
  
  /* Store the address and size so later messages will know what to do */
  StoreMsgAddr[ ((from_whom<<10) | mid) ] = p.address;
  StoreMsgLen[ ((from_whom<<10) | mid) ] = p.length;
  
  /* Ok, I know that there will be some messages arriving soon ...*/
  
}

/*********************************************************************/
/* A tulip_Put() request has *ARRIVED* from a remote processor
   (either StoreBlockMsg | EndOfStoreMsg).  To satisfy the request, this
   processor must look-up the address to where the data should be
   unpacked (which it received in a "prefix") , and unpack into the
   correct memory location.  If this is an EndOfStoreMsg, an ACK should
   be returned, and the lookup table cleared */

static void store_block_handler(int from_whom, int mid, int type)
{
  char *localaddr;
  int size;
  
  localaddr =  StoreMsgAddr[ ((from_whom<<10) | mid) ]; /* address */
  
  if (type == EndOfStoreMsg) /* The last piece! */ {
    size = StoreMsgLen[ ((from_whom<<10) | mid) ];
/*printf("%d:  Received last block of %d bytes from %d (storing in %x)\n",
  tulip_MyContext(),size,from_whom,localaddr);*/
    
  }
  else /* StoreBlockMsg, more to come */
    {
      size = tulip_MsgBlockSize;
      /* reduce expected byte yet to receive */
      StoreMsgLen [ ((from_whom<<10) | mid) ] -= tulip_MsgBlockSize;
      StoreMsgAddr[ ((from_whom<<10) | mid) ] += tulip_MsgBlockSize;
/*printf("%d:  Received %d bytes from %d (storing in %x) (with %d expected later)\n",
  tulip_MyContext(),size,from_whom,localaddr,StoreMsgLen[((from_whom<<10) | mid)]);*/
    }
  
  MyRecv(type, mid, localaddr, size, from_whom);
  
  /* Send ACK if it was the last message */
  if (type == EndOfStoreMsg)
    MySend(AckReplyMsg, mid, SendBuffer, 0, from_whom);
  
  return;
}

/*********************************************************************/
/* An acknowledge for a previously completed store has been received.
   Up the local handle with this new, important information and recycle
   the message ID */

static void ack_store_block_handler(int from_whom, int mid)
{
  /* There is no "body" to this message, just an "event" */
  MyRecv(AckReplyMsg, mid, ReceiveBuffer, 0, from_whom);
  
  /* Mark acknowledge */
  tulip_SetACK(AckHandle[mid]);
  
  /* Decrement internal ACK count */
  --PendingSysACKs;
  
  /* recycle old MID */
  RecycleMsgID(mid);
}

/*********************************************************************/
/* A Fetch block request that was previously initiated by this
   processor has returned some (possibly not all) of the data
   (EndOfFetchMsg or FetchBlockMsg).  Grab the fetched data, and if it
   is the last part of the block, notify the handle and recycle the
   message ID */

static void rcvd_fetch_block_handler(int from_whom, int mid, int type)
{
  register int s;
  
  if (type == EndOfFetchMsg) /* The last piece! */
    s = RecvMsgLen[mid];
  else /* FetchBlockMsg, more to come */
    {
      s = tulip_MsgBlockSize;
      /* reduce expected byte yet to receive */
      RecvMsgLen[mid] -= tulip_MsgBlockSize;
    }
  
  MyRecv(type, mid, PendingMsgs[mid], s, from_whom);
  
  /* if it was the last message, notify handle and recycle mid */
  if (type == EndOfFetchMsg) {

    /* Mark acknowledge */
    tulip_SetACK(AckHandle[mid]);
  
    /* Decrement internal ACK count */
    --PendingSysACKs;
    
    /* recycle old MID */
    RecycleMsgID(mid);
  }
  else /* It was part of a block, move pointer forward for future piece */
    PendingMsgs[mid] += tulip_MsgBlockSize;
}

/*********************************************************************/
/* Send a message to a remote processor asking for a block of memory */

int tulip_Get(char *dst, tulip_GlobalPointer_t src, int len, 
	      tulip_ACK_t *ackhandle)
{
  int mid;
  struct msg_params *p;
  
  /* check if src is local */
  if (tulip_ToContext(&src) == MyNodeNumber)
    {
      /* This is real inefficient, and it should have been just a local
	 memory reference, but to keep it correct, I will do the fetch, 
	 even though it is local. */
      
      memcpy(dst, tulip_ToLocalPointer(&src), len);
      return 0;
    }  

if (tulip_ToContext(&src) < 0) {
  printf("%d: ERROR: %d.%x \n",MyNodeNumber,
	 tulip_ToContext(&src),tulip_ToLocalPointer(&src));
  tulip_Abort();
}
  
  mid = NewMessageID();           /* Get new message ID for this Fetch */
  AckHandle[mid] = ackhandle;     /* Store ackhandle */

  tulip_ClearACK(AckHandle[mid]); /* reset ackhandle (if nonnil ptr) */

  PendingMsgs[mid] = dst;         /* Store ptr for incoming msg */
  RecvMsgLen[mid]  = len;         /* length  */
  
  /* Package up message request */
  p = (struct msg_params *) SendBuffer;
  p->address = tulip_ToLocalPointer(&src);
  p->length = len;
  
  /* Post a receive _NOW_ for the data that will be returned */
  MyIRecv(EndOfFetchMsg, mid, dst, len, tulip_ToContext(&src));

  /* Send request for data to remote node */
  MySend(AskForDataMsg, mid, SendBuffer, 
         sizeof(struct msg_params), tulip_ToContext(&src));
  
  /* Increment internal ACK count */
  ++PendingSysACKs;
  
  return 0;
}

/*********************************************************************/
/* Store a block of memory onto a remote processor. */

int tulip_Put(tulip_GlobalPointer_t dst, char *src, int len, 
	      tulip_ACK_t *ackhandle)
{
  int mid;
  struct msg_params p;
  
  /* check if dst is _really_ remote */
  if (tulip_ToContext(&dst) == MyNodeNumber)
    {
      /* This is real inefficient, and it should have been just a local
	 memory reference, but to keep it correct, I will do the store,
	 even though it is local. */
      
      memcpy(tulip_ToLocalPointer(&dst), src, len);
      return 0;
    }  
  
  mid     = NewMessageID();    /* Get a new message ID */
  AckHandle[mid] = ackhandle;    

  tulip_ClearACK(AckHandle[mid]); /* reset ackhandle (if nonnil ptr) */

  PendingMsgs[mid] = src;      
				     
				     /* Package up message request */
  p.address = tulip_ToLocalPointer(&dst);
  p.length = len;
  
  /* Send prefix to remote node to announce the arrival of data... */
  MySend(PrefixStoreMsg, mid, (char *) &p,
	 (int) sizeof(struct msg_params), tulip_ToContext(&dst));
  
  /* Now start streaming (in chunks) the data to store remotely... */
  
/*printf("%d: will store %d bytes to %d (into mem %x)\n",
  tulip_MyContext(),len,tulip_ToContext(&dst),tulip_ToLocalPointer(&dst));*/

  while ((tulip_MsgBlockSize > 0) && (len > tulip_MsgBlockSize)) {
    MySend(StoreBlockMsg, mid, src, tulip_MsgBlockSize, tulip_ToContext(&dst));
    len -= tulip_MsgBlockSize;
    src += tulip_MsgBlockSize;

/*printf("%d: Sent %d to %d, with %d left (into mem %x)\n"
  ,tulip_MyContext(),tulip_MsgBlockSize,tulip_ToContext(&dst),len,tulip_ToLocalPointer(&dst));*/
  }
  
  /* Send final chunk (or only chunk if requested block was small) */
  MySend(EndOfStoreMsg, mid, src, len, tulip_ToContext(&dst));
  
  /* Increment internal ACK count */
  ++PendingSysACKs;

  /* Done!  ACK will arrive later */
  return 0;
}

/*********************************************************************/
/* procedure sends a message to a remote node.  The message causes the
   tulip_RemoteActionHandler to be invoked on that node */

int tulip_RemoteAction(int remote_node, char *buffer, int len, 
		       tulip_ACK_t *ackhandle)
{
  int mid;
  
  mid     = NewMessageID();    /* Get a new message ID */
  
  /* Set up what to do about ack that remote action has completed */
  AckHandle[mid] = ackhandle;    

  tulip_ClearACK(AckHandle[mid]); /* reset ackhandle (if nonnil ptr) */

  MySend(RemoteActionMsg, mid, buffer, len, remote_node);
  
  /* Increment internal ACK count */
  ++PendingSysACKs;
  
  return 0;
}

/*********************************************************************/
/*********************************************************************/
/************************** ALIEN FUNCTIONS **************************/
/*********************************************************************/
/*********************************************************************/

/* Some machines have their own "Native" Broadcast and Reduction
   functions.  For those machines, pC++ uses the faster, often hardware
   supported, functions calls.  To do this, #defines are used to call the
   Native functions that live in KernelCom.c for that machine.  All
   machines get both "Native" and "Alien" routines so that the functions
   exist for timing purposes, but if no Native routine exists, it is just
   mapped to the Alien routine.  The Alien routines use our hand-coded
   solutions to Broadcast and Reduce.  On a machine that does not have
   any special Broadcast or Reduce functions, pcxx_Broadcast will simply
   be mapped to use tulip_Alien* which relys on tree fan in and fan out
   versions. */

/* Forward declarations for Alien functions */
void tulip_AlienBroadcastBytes(int flag, int length, void * buffer);
void tulip_AlienReduceIntAdd(void* buffer,int elem_count);
void tulip_AlienReduceFloatAdd(void* buffer,int elem_count);
void tulip_AlienReduceDoubleAdd(void* buffer,int elem_count);

void tulip_AlienReduceIntMax(void* buffer,int elem_count);
void tulip_AlienReduceFloatMax(void* buffer,int elem_count);
void tulip_AlienReduceDoubleMax(void* buffer,int elem_count);

void tulip_AlienReduceIntMin(void* buffer,int elem_count);
void tulip_AlienReduceFloatMin(void* buffer,int elem_count);
void tulip_AlienReduceDoubleMin(void* buffer,int elem_count);

/**********************************************************************/
#if (defined(__PARAGON__) || defined(__MEIKO__)) && !defined(__PVM__) && !defined(__MPI__)

/**********************************************************************/
/*-------------------------------------------------------------------------
  Paragon Native functions are defined in Paragon KernelCom.c 
 --------------------------------------------------------------------------*/
extern void tulip_NativeBroadcastBytes(int flag, int length, void * buffer);

extern void tulip_NativeReduceIntAdd(void* buffer,int elem_count);
extern void tulip_NativeReduceFloatAdd(void* buffer,int elem_count);
extern void tulip_NativeReduceDoubleAdd(void* buffer,int elem_count);

extern void tulip_NativeReduceIntMax(void* buffer,int elem_count);
extern void tulip_NativeReduceFloatMax(void* buffer,int elem_count);
extern void tulip_NativeReduceDoubleMax(void* buffer,int elem_count);

extern void tulip_NativeReduceIntMin(void* buffer,int elem_count);
extern void tulip_NativeReduceFloatMin(void* buffer,int elem_count);
extern void tulip_NativeReduceDoubleMin(void* buffer,int elem_count);

void tulip_BroadcastBytes(int flag, int length, void * buffer) {
  tulip_NativeBroadcastBytes(flag, length, buffer); }
  
void tulip_ReduceIntAdd(void* buffer,int elem_count) {
  tulip_NativeReduceIntAdd(buffer,elem_count); }

void tulip_ReduceFloatAdd(void* buffer,int elem_count) {
  tulip_NativeReduceFloatAdd(buffer,elem_count); }

void tulip_ReduceDoubleAdd(void* buffer,int elem_count) {
  tulip_NativeReduceDoubleAdd(buffer,elem_count); }

void tulip_ReduceIntMax(void* buffer,int elem_count) {
  tulip_NativeReduceIntMax(buffer,elem_count); }

void tulip_ReduceFloatMax(void* buffer,int elem_count) {
  tulip_NativeReduceFloatMax(buffer,elem_count); }

void tulip_ReduceDoubleMax(void* buffer,int elem_count) {
  tulip_NativeReduceDoubleMax(buffer,elem_count); }

void tulip_ReduceIntMin(void* buffer,int elem_count) {
  tulip_NativeReduceIntMin(buffer,elem_count); }

void tulip_ReduceFloatMin(void* buffer,int elem_count) {
  tulip_NativeReduceFloatMin(buffer,elem_count); }

void tulip_ReduceDoubleMin(void* buffer,int elem_count) {
  tulip_NativeReduceDoubleMin(buffer,elem_count); }

#else

/* Links to Alien function */

void tulip_BroadcastBytes(int flag, int length, void * buffer) {
  tulip_AlienBroadcastBytes(flag, length, buffer); }
     
void tulip_ReduceIntAdd(void* buffer,int elem_count) {
  tulip_AlienReduceIntAdd(buffer,elem_count); }
    
void tulip_ReduceFloatAdd(void* buffer,int elem_count) {
  tulip_AlienReduceFloatAdd(buffer,elem_count); }
     
void tulip_ReduceDoubleAdd(void* buffer,int elem_count) {
  tulip_AlienReduceDoubleAdd(buffer,elem_count); }

void tulip_ReduceIntMax(void* buffer,int elem_count) {
  tulip_AlienReduceIntMax(buffer,elem_count); }
    
void tulip_ReduceFloatMax(void* buffer,int elem_count) {
  tulip_AlienReduceFloatMax(buffer,elem_count); }
     
void tulip_ReduceDoubleMax(void* buffer,int elem_count) {
  tulip_AlienReduceDoubleMax(buffer,elem_count); }

void tulip_ReduceIntMin(void* buffer,int elem_count) {
  tulip_AlienReduceIntMin(buffer,elem_count); }
    
void tulip_ReduceFloatMin(void* buffer,int elem_count) {
  tulip_AlienReduceFloatMin(buffer,elem_count); }
     
void tulip_ReduceDoubleMin(void* buffer,int elem_count) {
  tulip_AlienReduceDoubleMin(buffer,elem_count); }
     
/** MAKE SURE NATIVE IS DEFINED TOO SO THAT WE CAN SWITCH FROM MACHINE 
 **   NATIVE TO USER-CODED NATIVE *************************************/

void tulip_NativeBroadcastBytes(int flag, int length, void * buffer) {
  tulip_AlienBroadcastBytes(flag, length, buffer); }
    
void tulip_NativeReduceIntAdd(void* buffer,int elem_count) {
  tulip_AlienReduceIntAdd(buffer,elem_count); }
     
void tulip_NativeReduceFloatAdd(void* buffer,int elem_count) {
  tulip_AlienReduceFloatAdd(buffer,elem_count); }
     
void tulip_NativeReduceDoubleAdd(void* buffer,int elem_count) {
  tulip_AlienReduceDoubleAdd(buffer,elem_count); }

void tulip_NativeReduceIntMax(void* buffer,int elem_count) {
  tulip_AlienReduceIntMax(buffer,elem_count); }
     
void tulip_NativeReduceFloatMax(void* buffer,int elem_count) {
  tulip_AlienReduceFloatMax(buffer,elem_count); }
     
void tulip_NativeReduceDoubleMax(void* buffer,int elem_count) {
  tulip_AlienReduceDoubleMax(buffer,elem_count); }

void tulip_NativeReduceIntMin(void* buffer,int elem_count) {
  tulip_AlienReduceIntMin(buffer,elem_count); }
     
void tulip_NativeReduceFloatMin(void* buffer,int elem_count) {
  tulip_AlienReduceFloatMin(buffer,elem_count); }
     
void tulip_NativeReduceDoubleMin(void* buffer,int elem_count) {
  tulip_AlienReduceDoubleMin(buffer,elem_count); }

#endif

/*********************************************************************/
/*********************************************************************/
/* BROADCAST FUNCTIONS */
/*********************************************************************/
/*********************************************************************/

/*********************************************************************/
/* pCxx_AlienBroadcastBytes sends messages to all other nodes from the
   node with flag = 0.  the message is in the buffer and the length is
   known. */

void tulip_AlienBroadcastBytes(int flag, int length, void * buffer) 
{
  int i; char *c, *p;

  /* The following barrier is very important.  First, it prevents a receiver
     from receiving more than one broadcast during poll which may
     overwrite the receiving buffer.  Second, it prevents
     a receiver from receiving a broadcast when it is not ready to receive
     the broadcast, i.e., when it doest not have a buffer to save the
     broadcast.   --- S. Y. (2/11/94)
   */
  tulip_Barrier();

  if(flag == 0 && MyNodeNumber != 0) {
    /*Send buffer to root, if MyNodeNumber != 0 and I want to broadcast.*/
    MySend(BroadcastMsg, NOMID, buffer,length,0);
  }
  if(MyNodeNumber == 0 && flag != 0) {
    while (broadcast_flag == 0) tulip_Poll();
    memcpy((char*)buffer, ReceiveBuffer, length);
  }


  /* Start the msg propagating, from the root (processor 0)*/
  if (MyNodeNumber == 0) {
    if (num_children >= 1) {
      MySend(BroadcastMsg, NOMID, buffer,length,left_child);	
    }
    if (num_children == 2) {
       MySend(BroadcastMsg, NOMID, buffer,length,right_child);	
    }
    broadcast_flag = 0;
    return;
  }

  /* Processors other than 0 wait till they receive the broadcast */
  while (broadcast_flag == 0) tulip_Poll();
  memcpy((char*)buffer, ReceiveBuffer, length);

  /* Now send the copy I got to my children If I have any */
  if (num_children >= 1) {
    MySend(BroadcastMsg,NOMID,buffer,length,left_child);	
  }
  if (num_children == 2) {
    MySend(BroadcastMsg,NOMID,buffer,length,right_child);	
  }

  /* reset flag */
  broadcast_flag = 0;
}

/*********************************************************************/
/* tulip_AlienReduceDoubleAdd adds together corresponding buffer elements 
   and broadcasts the resulting buffer to all nodes.  The original
   buffer content is overwritten. */

void tulip_AlienReduceDoubleAdd(void* buffer,int elem_count)
{
   int i, flag;

   tulip_Barrier();

   /* get totals from children first */
   if(num_children > 0) {
     while(reduction_flag == 0) tulip_Poll();
     for(i = 0; i < elem_count; i++) { 
       ((double*)buffer)[i] += ((double*)ReceiveBuffer)[i];
     }
   }
   if(num_children > 1) {
     while(reduction_flag < 2) tulip_Poll();
     for(i = 0; i < elem_count; i++) {
       ((double*)buffer)[i] += ((double*)ReceiveBuffer)[i];
     }
   }
   reduction_flag = 0;

   /* send totals to my parent. */
   if(MyNodeNumber != 0)
     MySend(ReduceMsg, NOMID, buffer, elem_count*sizeof(double), parent_id);

   /* now share the totals.  processor zero broadcasts */
   flag = 1;
   if(MyNodeNumber == 0) flag = 0;
   tulip_BroadcastBytes(flag, elem_count*sizeof(double), buffer);   
}

/*********************************************************************/
void tulip_AlienReduceFloatAdd(void* buffer,int elem_count)
{
   int i, flag;

   tulip_Barrier();

   /* get totals from children first */
   if(num_children > 0){
     while(reduction_flag == 0) tulip_Poll();
     for(i = 0; i < elem_count; i++) {
       ((float*)buffer)[i] += ((float*)ReceiveBuffer)[i];
     }
   }
   if(num_children > 1){
     while(reduction_flag < 2) tulip_Poll();
     for(i = 0; i < elem_count; i++) {
       ((float*)buffer)[i] += ((float*)ReceiveBuffer)[i];
     }
   }
   reduction_flag = 0;

   /* send totals to my parent. */
   if(MyNodeNumber != 0)
     MySend(ReduceMsg, NOMID, buffer, elem_count*sizeof(float), parent_id);

   /* now share the totals.  processor zero broadcasts */
   flag = 1;
   if(MyNodeNumber == 0) flag = 0;
   tulip_BroadcastBytes(flag, elem_count*sizeof(float), buffer);   
}

/*********************************************************************/
void tulip_AlienReduceIntAdd(void* buffer,int elem_count)
{
   int i, flag;

   tulip_Barrier();

   /* get totals from children first */
   if(num_children > 0){
     while(reduction_flag == 0) tulip_Poll();
     for(i = 0; i < elem_count; i++) {
       ((int*)buffer)[i] += ((int*)ReceiveBuffer)[i];
     }
   }
   if(num_children > 1){
     while(reduction_flag < 2) tulip_Poll();
     for(i = 0; i < elem_count; i++) {
       ((int*)buffer)[i] += ((int*)ReceiveBuffer)[i];
     }
   }
   reduction_flag = 0;

   /* send totals to my parent. */
   if(MyNodeNumber != 0)
     MySend(ReduceMsg, NOMID, buffer, elem_count*sizeof(int), parent_id);

   /* now share the totals.  processor zero broadcasts */
   flag = 1;
   if(MyNodeNumber == 0) flag = 0;
   tulip_BroadcastBytes(flag, elem_count*sizeof(int), buffer);   
}

/*********************************************************************/
void tulip_AlienReduceIntMax(void* buffer,int elem_count){
   int i, flag;

   tulip_Barrier();

   /* get result from children first */
   if(num_children > 0){
     while(reduction_flag == 0) tulip_Poll();
     for(i = 0; i < elem_count; i++) {
       if(((int*)buffer)[i] < ((int*)ReceiveBuffer)[i]) {
	 ((int*)buffer)[i] = ((int*)ReceiveBuffer)[i];
       }
     }
   }
   if(num_children > 1){
     while(reduction_flag < 2) tulip_Poll();
     for(i = 0; i < elem_count; i++) {
       if(((int*)buffer)[i] < ((int*)ReceiveBuffer)[i]) {
	 ((int*)buffer)[i] = ((int*)ReceiveBuffer)[i];
       }
     }
   }
   reduction_flag = 0;

   /* send result to my parent. */
   if(MyNodeNumber != 0)
     MySend(ReduceMsg, NOMID, buffer, elem_count*sizeof(int), parent_id);

   /* now share the result.  processor zero broadcasts */
   flag = 1;
   if(MyNodeNumber == 0) flag = 0;
   tulip_BroadcastBytes(flag, elem_count*sizeof(int), buffer);   
}

/*********************************************************************/
void tulip_AlienReduceFloatMax(void* buffer,int elem_count){
   int i, flag;

   tulip_Barrier();

   /* get result from children first */
   if(num_children > 0){
     while(reduction_flag == 0) tulip_Poll();
     for(i = 0; i < elem_count; i++) {
       if(((float*)buffer)[i] < ((float*)ReceiveBuffer)[i]) 
	 ((float*)buffer)[i] = ((float*)ReceiveBuffer)[i];
     }
   }
   if(num_children > 1){
     while(reduction_flag < 2) tulip_Poll();
     for(i = 0; i < elem_count; i++) {
       if(((float*)buffer)[i] < ((float*)ReceiveBuffer)[i]) 
	 ((float*)buffer)[i] = ((float*)ReceiveBuffer)[i];
     }
   }
   reduction_flag = 0;

   /* send result to my parent. */
   if(MyNodeNumber != 0)
     MySend(ReduceMsg, NOMID, buffer, elem_count*sizeof(float), parent_id);

   /* now share the result.  processor zero broadcasts */
   flag = 1;
   if(MyNodeNumber == 0) flag = 0;
   tulip_BroadcastBytes(flag, elem_count*sizeof(float), buffer);   
}


/*********************************************************************/
void tulip_AlienReduceDoubleMax(void* buffer,int elem_count)
{
   int i, flag;

   tulip_Barrier();

   /* get totals from children first */
   if(num_children > 0){
     while(reduction_flag == 0) tulip_Poll();
     for(i = 0; i < elem_count; i++) {
       if(((double*)buffer)[i] < ((double*)ReceiveBuffer)[i]) 
	 ((double*)buffer)[i] = ((double*)ReceiveBuffer)[i];
     }
   }
   if(num_children > 1){
     while(reduction_flag < 2) tulip_Poll();
     for(i = 0; i < elem_count; i++) {
       if(((double*)buffer)[i] < ((double*)ReceiveBuffer)[i]) 
	 ((double*)buffer)[i] = ((double*)ReceiveBuffer)[i];
     }
   }
   reduction_flag = 0;

   /* send result to my parent. */
   if(MyNodeNumber != 0)
     MySend(ReduceMsg, NOMID, buffer, elem_count*sizeof(double), parent_id);

   /* now share the result.  processor zero broadcasts */
   flag = 1;
   if(MyNodeNumber == 0) flag = 0;
   tulip_BroadcastBytes(flag, elem_count*sizeof(double), buffer);   
}


/*********************************************************************/
void tulip_AlienReduceIntMin(void* buffer,int elem_count)
{
   int i, flag;

   tulip_Barrier();

   /* get result from children first */
   if(num_children > 0){
     while(reduction_flag == 0) tulip_Poll();
     for(i = 0; i < elem_count; i++) {
       if(((int*)buffer)[i] > ((int*)ReceiveBuffer)[i]) 
	 ((int*)buffer)[i] = ((int*)ReceiveBuffer)[i];
     }
   }
   if(num_children > 1){
     while(reduction_flag < 2) tulip_Poll();
     for(i = 0; i < elem_count; i++) {
       if(((int*)buffer)[i] > ((int*)ReceiveBuffer)[i]) 
	 ((int*)buffer)[i] = ((int*)ReceiveBuffer)[i];
     }
   }
   reduction_flag = 0;

   /* send result to my parent. */
   if(MyNodeNumber != 0)
     MySend(ReduceMsg, NOMID, buffer, elem_count*sizeof(int), parent_id);

   /* now share the result.  processor zero broadcasts */
   flag = 1;
   if(MyNodeNumber == 0) flag = 0;
   tulip_BroadcastBytes(flag, elem_count*sizeof(int), buffer);   
}

/*********************************************************************/
void tulip_AlienReduceFloatMin(void* buffer,int elem_count)
{
   int i, flag;

   tulip_Barrier();

   /* get result from children first */
   if(num_children > 0){
     while(reduction_flag == 0) tulip_Poll();
     for(i = 0; i < elem_count; i++) {
       if(((float*)buffer)[i] > ((float*)ReceiveBuffer)[i]) 
	 ((float*)buffer)[i] = ((float*)ReceiveBuffer)[i];
     }
   }
   if(num_children > 1){
     while(reduction_flag < 2) tulip_Poll();
     for(i = 0; i < elem_count; i++) {
       if(((float*)buffer)[i] > ((float*)ReceiveBuffer)[i]) 
	 ((float*)buffer)[i] = ((float*)ReceiveBuffer)[i];
     }
   }
   reduction_flag = 0;

   /* send result to my parent. */
   if(MyNodeNumber != 0)
     MySend(ReduceMsg, NOMID, buffer, elem_count*sizeof(float), parent_id);

   /* now share the result.  processor zero broadcasts */
   flag = 1;
   if(MyNodeNumber == 0) flag = 0;
   tulip_BroadcastBytes(flag, elem_count*sizeof(float), buffer);   
}


/*********************************************************************/
void tulip_AlienReduceDoubleMin(void* buffer,int elem_count)
{
   int i, flag;

   tulip_Barrier();

   /* get totals from children first */
   if(num_children > 0){
     while(reduction_flag == 0) tulip_Poll();
     for(i = 0; i < elem_count; i++) {
       if(((double*)buffer)[i] > ((double*)ReceiveBuffer)[i]) 
	 ((double*)buffer)[i] = ((double*)ReceiveBuffer)[i];
     }
   }
   if(num_children > 1){
     while(reduction_flag < 2) tulip_Poll();
     for(i = 0; i < elem_count; i++) {
       if(((double*)buffer)[i] > ((double*)ReceiveBuffer)[i]) 
	 ((double*)buffer)[i] = ((double*)ReceiveBuffer)[i];
     }
   }
   reduction_flag = 0;

   /* send result to my parent. */
   if(MyNodeNumber != 0)
     MySend(ReduceMsg, NOMID, buffer, elem_count*sizeof(double), parent_id);

   /* now share the result.  processor zero broadcasts */
   flag = 1;
   if(MyNodeNumber == 0) flag = 0;
   tulip_BroadcastBytes(flag, elem_count*sizeof(double), buffer);   
}
