/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/*********************************************************************/
/*                Parallel run-time library                          */
/*                                                                   */
/*                T3E Version                                        */
/*                                                                   */
/* Contributions to this module have been made by:                   */
/*                                                                   */
/*     Pete Beckman *** Francois Bodin *** Dennis Gannon             */
/*                Bernd Mohr *** Shelby Yang                         */
/*********************************************************************/

#include "tulip.h"
#include "profiler_defs.h"

#include "ndma.h"

int tulip_MsgBlockSize = -1;

/*****************************/
/* Structs                   */
/*****************************/
#define MAXSLOTS 4
#define MAX_RA_BUFF 64
#define MaxSendBuff 16384

/* structure for remote action */
struct RA_params {
  long   buffer[MAX_RA_BUFF];
  long   length;
  long * ackhandle;
  long   ack;
};
struct RA_params MyRAQueue[PCXX_MAXPROCS][MAXSLOTS]; 

/* structure for remote get */
struct Send_Params {
  char *dst;
  char *src;
  int length;
};
struct Send_Params SendRequest[PCXX_MAXPROCS];

/* buffer for remote get */
char SendBuffer[MaxSendBuff];

/*********************************************************************/
/* EXTERNAL, C-LINKAGE FUNCTION, THAT THE USER MUST PROVIDE: */
extern void tulip_RemoteActionHandler(int from_whom, char *buffer, int len);

extern pcxx_PARAM_t pcxx_param;

/*********************************************************************/
/* MACHINE DEPENDANT CODE */
/*********************************************************************/

extern void shmem_get (long *target, long *source, 
                      unsigned int len, unsigned int pe);
 
extern void shmem_put (long *target, long *source, 
                      unsigned int len, unsigned int pe);

extern void shmem_udcflush();
extern void shmem_set_cache_inv();

extern long test_barrier();
extern void set_barrier();
extern int _MPP_MY_PE;
extern int _MPP_N_PES;

/**********************************************************************/
void tulip_SharedFree(void* p) {
  free(p);
}

/**********************************************************************/
void *tulip_SharedMalloc(int size) {
  return (void *) malloc(size);
}

/**********************************************************************/
long *AlignBefore(char * addr) {
  return (long *) (((unsigned long) addr >> 3) << 3);
}

/**************************************************************************
   WARNING!!! Use this function ONLY to "put" data into a remote buffer 
   which is statically allocated.  "Operand rang error" and core dump may 
   result otherwise.

   Do a remote "put" based on bytes rather than on CRAY words (8
   bytes).  This routine has been programmed to be most efficient for
   word aligned data, with unaligned data a special case.
***************************************************************************/

void PutBytes(char *dst, char *src, int len, int remote)
{
  int FirstExtraBytes, LastExtraBytes, BetweenWords;
  int t;
  char *newsrc, *origmalloc, *k, *l;
  long temp;  /* 8 byte word */

  /* Trivial case */
  if (len == 0) return;

  /* FirstExtraBytes" of 0 means the beginning address is word aligned */
  FirstExtraBytes = (8 - (int) ((unsigned long) dst & 7)) & 7;
  t               = (8 - (int) ((unsigned long) src & 7)) & 7;

  /* Quick check, is this a perfectly word-aligned "put" */
  if ((FirstExtraBytes == 0) && (t == 0) && ((len & 7) == 0)) {
    /*printf("%d: shmem_put %x %x %d %d\n",MyNodeNumber,dst,src,(len>>3),
      remote);*/
    shmem_put((long *) dst, (long *) src, (len>>3), remote);
    return;
  }

  /* Check for alignment shift: If The src and dst alignments (with
     respect to the nearest word) are shifted, we can do NO aggregate
     shmem_put.  I will _COPY_ the src buffer into a dst-aligned temp
     buffer, and then use the aggregate shmem_put for the "middle" bytes. */
  
  if (FirstExtraBytes != t) {
    /* BAD SRC/DST ALIGNMENT !!! */

    /* Allocate dst-aligned Buffer... */
    newsrc = (char *) malloc(len+32); /* padded for later adjustment... */
    if (newsrc == 0) tulip_Message("PutBytes:  Out Of Memory!",TULIP_FATAL);

    /* Adjust the buffer pointer to have the same offset as the dst */
    origmalloc = newsrc;
    newsrc = (char *) ((unsigned long) newsrc + 
		       (unsigned long) 8 - (unsigned long) FirstExtraBytes);

    /* Move the src */
    memcpy(newsrc,src,len);
  }
  else
    newsrc = src;    /* alignment shift not required */
  
  /* At this point, newsrc and the dst have the same word offsets */

  /* There are now 3 parts, first bytes, middle, last bytes */

  /* Do Part 1 */
  /*************************************************************/
  if (FirstExtraBytes > 0) {
    /* Get original word contents */
    shmem_get(&temp, AlignBefore(dst), 1, remote);

    /* Calculate the number of bytes to replace in first word */
    if (FirstExtraBytes > len) 
      /* in case first part does not go to end of word */
      t = len;
    else
      t = FirstExtraBytes;

    /* Swap in the bytes to modify */
    memcpy((char *) &temp, newsrc, t);
    
    /* Now replace word */
    shmem_put(AlignBefore(dst), &temp, 1, remote);
  }
  
  /* OK, first "sub-word" has completed, now do COMPLETE middle words */
  LastExtraBytes = ((len - FirstExtraBytes) & 7);
  BetweenWords = (len - FirstExtraBytes - LastExtraBytes)>>3;

  /* Do Part 2 */
  /*************************************************************/
  if (BetweenWords > 0) {
    /* We have some real WORDS to transfer */
    shmem_put((long *) ((unsigned long) FirstExtraBytes + dst),
	      (long *) ((unsigned long) FirstExtraBytes + newsrc),
	      BetweenWords,remote);
  }
  
  /* Ok, now just the last bytes in the last word (if any) */
  
  /* Do Part 3 */
  /*************************************************************/
  if (LastExtraBytes > 0) {

    /* Calculate Word _before_ end of destination */
    k = (char *) ((unsigned long) dst    + (unsigned long) len -
		  (unsigned long) LastExtraBytes);
    l = (char *) ((unsigned long) newsrc + (unsigned long) len -
		  (unsigned long) LastExtraBytes);
    
    /* Get last word, that will be partially modified */
    shmem_get(&temp, (long *) k, 1, remote);

    /* Swap in the bytes to modify */
    memcpy((char *) &temp, l, LastExtraBytes);
    
    /* Now replace word */
    shmem_put((long *) k, &temp, 1, remote);
  }

  if (origmalloc != src) free(origmalloc);
}

/****************************************************************************
   This function replaces the original GetBytes function which is now called 
   xxxxGetBytes.  The reason is that memory allocated by malloc sometimes
   cannot be accessed by remote processors.  An access by shmem_get or 
   shmem_put will result in core dump and "operand range error."  To fix this
   problem, we use a static buffer, SendBuffer, and split a request into 
   many requests.  In its current form, GetBytes actually sends a request to a
   remote processor.  Up on receiving the request, the remote processor sends 
   data back to the static buffer, and then the content of the static buffer 
   is copied into the destination buffer.

   Shelby Yang, 8/17/95.
******************************************************************************/

void GetBytes(char *dst, char *src, int len, int remote) 
{
  int blocks, extra, i;
  
  blocks = len / MaxSendBuff;
  extra = len % MaxSendBuff;

  for (i = 0; i < blocks; i++) {
    SendRequest[MyNodeNumber].dst = (char*) &SendBuffer;
    SendRequest[MyNodeNumber].src = src + i*MaxSendBuff;
    SendRequest[MyNodeNumber].length = MaxSendBuff;
  
    /* Send request to the remote node.  The address of local SendRequest 
       is the address on the remote processor. */
    PutBytes((char*)&SendRequest[MyNodeNumber], 
	     (char*)&SendRequest[MyNodeNumber], 
	     sizeof(struct Send_Params), remote);
    while(SendRequest[MyNodeNumber].length != 0) tulip_Poll();
    memcpy((char*)(dst+i*MaxSendBuff), SendBuffer, MaxSendBuff);
  }

  if(extra > 0) {
    SendRequest[MyNodeNumber].dst = (char*)&SendBuffer;
    SendRequest[MyNodeNumber].src = src + blocks*MaxSendBuff;
    SendRequest[MyNodeNumber].length = extra;
  
    /* Send request to the remote node.  Put is preferred since it's twice
       as fast as get. The address of local SendRequest is the address on
       the remote processor. */
    PutBytes((char*)&SendRequest[MyNodeNumber], 
	     (char*)&SendRequest[MyNodeNumber], 
	     sizeof(struct Send_Params), remote);
    while(SendRequest[MyNodeNumber].length != 0) tulip_Poll();
    memcpy((char*)(dst+blocks*MaxSendBuff), SendBuffer, extra);
  }
}

/**********************************************************************/
/* Do a remote "get" based on bytes rather than on CRAY words (8
   bytes).  This routine has been programmed to be most efficient for
   word aligned data, with unaligned data a special case   

   This function is now useless.   Shelby Yang. 8/17/95
***********************************************************************/
void xxxxGetBytes(char *dst, char *src, int len, int remote)
{
  int FirstExtraBytes, LastExtraBytes, BetweenWords;
  int t;
  char *newsrc, *k, *l;
  long temp;  /* 8 byte word */

  /*printf("%d: GetBytes %x %x %d %d\n",MyNodeNumber,dst,src,len,remote);*/

  /* Trivial case */
  if (len == 0) return;

  /* FirstExtraBytes" of 0 means the beginning address is word aligned */
  FirstExtraBytes = (8 - (int) ((unsigned long) src & 7)) & 7;
  t               = (8 - (int) ((unsigned long) dst & 7)) & 7;

  /* Quick check, is this a perfectly word-aligned "put" */
  if ((FirstExtraBytes == 0) && (t == 0) && ((len & 7) == 0)) {
    shmem_get((long *) dst, (long *) src, (len>>3), remote);
    return;
  }
  
  /* Check for alignment shift: If The src and dst alignments (with
   respect to the nearest word) are shifted, we can do NO aggregate
   shmem_get, instead, I _COPY_ the remote src to a local buffer, and
   then install the bytes with a memcpy */

  if (FirstExtraBytes != t) {
    /* BAD SRC/DST ALIGNMENT !!! */

    /* Allocate Buffer... */
    newsrc = (char *) malloc(len+32); /* padded for later adjustment... */
    if (newsrc == 0) tulip_Message("GetBytes:  Out Of Memory!",TULIP_FATAL);
    
    /* calculate number of words to grab */
    t = (len-FirstExtraBytes);
    if ( (t & 7) == 0) /* The tail is word aligned */
      t = t >> 3;  /* Divide by wordsize (8) */
    else
      t = (t >> 3) + 1; /* add extra word for partial last word */
    shmem_get((long *) newsrc, AlignBefore(src), t, remote);

    /* HEY! I've got the remote bytes, now just memcpy them to the dst */
    k = (char *) (((unsigned long) src & 7) + (unsigned long) newsrc);
    memcpy(dst,k,len);

    /* return buffer */
    free(newsrc);
    return;
  }

  /* At this point, src and dst have the same word offsets */

  /* There are now 3 parts, first bytes, middle, last bytes */

  /* Do Part 1 */
  /*************************************************************/
  if (FirstExtraBytes > 0) {
    /* Get first word */
    shmem_get(&temp, AlignBefore(src), 1, remote);

    /* Calculate the number of bytes to replace in first word */
    if (FirstExtraBytes > len) 
      /* in case first part does not go to end of word */
      t = len;
    else
      t = FirstExtraBytes;

    /* Swap in the bytes to modify */
    memcpy(dst, (char *) &temp, t);
  }

  /* OK, first "sub-word" has completed, now do COMPLETE middle words */
  LastExtraBytes = ((len - FirstExtraBytes) & 7);
  BetweenWords = (len - FirstExtraBytes - LastExtraBytes)>>3;
  
  /* Do Part 2 */
  /*************************************************************/
  if (BetweenWords > 0) {
    /* We have some real WORDS to transfer */
    shmem_get((long *) ((unsigned long) FirstExtraBytes + dst),
	      (long *) ((unsigned long) FirstExtraBytes + src),
	      BetweenWords,remote);
  }
  
  /* Ok, now just the last bytes in the last word (if any) */
  
  /* Do Part 3 */
  /*************************************************************/
  if (LastExtraBytes > 0) {
    /* get last word */
    /* Calculate address of last byte */
    k = (char *) ((unsigned long) dst + (unsigned long) (len-1));
    l = (char *) ((unsigned long) src + (unsigned long) (len-1));

    /* Get last word from remote src */
    shmem_get(&temp, AlignBefore(l), 1, remote);
    
    /* copy in required bytes */
    memcpy((char *) AlignBefore(k), (char *) &temp, LastExtraBytes);
  }

}

/*********************************************************************/
void tulip_Barrier()
{
  PROFILE_ENTER(PROFILE_Barrier);
  PCXX_EVENT(PCXX_EC_RUNTIME, PCXX_BARRIER_ENTER, 0);

  set_barrier();
  while (!test_barrier()) tulip_Poll();

  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_BARRIER_EXIT, 0);
  PROFILE_EXIT(PROFILE_Barrier);
}

/*********************************************************************/
/* Fetch some memory from a remote node.  This routine does not
   "start" a fetch, but actually does the fetch -- in its entirety.  This
   is because Cray's T3D software interface does not permit a post and
   continue version of the memory transfer operations. */

int tulip_Get(char *dst, tulip_GlobalPointer_t src, int len, tulip_ACK_t *ackhandle)
{
  /* check if src is local */
  if (tulip_ToContext(&src) == MyNodeNumber)
    {
      /* This is real inefficient, and it should have been just a local
	 memory reference, but to keep it correct, I will do the fetch, 
	 even though it is local. */
      
      memcpy(dst, tulip_ToLocalPointer(&src), len);
      return 0;
    }  
  
  /* #pragma _CRI suppress ????? To invalidate cache? */

  GetBytes( dst, tulip_ToLocalPointer(&src), 
	   (unsigned int) len, (unsigned int) tulip_ToContext(&src));

  if (ackhandle) *ackhandle = 1;  /* Done already */

  return 0;
}

/*********************************************************************/
/* Store a block of memory onto a remote processor. 
   (Read notes for tulip_Get()) */

int tulip_Put(tulip_GlobalPointer_t dst, char *src, int len, tulip_ACK_t *ackhandle)
{

  /* check if dst is _really_ remote */
  if (tulip_ToContext(&dst) == MyNodeNumber)
    {
      /* This is real inefficient, and it should have been just a local
	 memory reference, but to keep it correct, I will do the store,
	 even though it is local. */
      
      memcpy(tulip_ToLocalPointer(&dst), src, len);
      return 0;
    }  
  
  /* #pragma _CRI suppress ????? To invalidate cache? */
  
  GetBytes( tulip_ToLocalPointer(&dst), src,
             (unsigned int) len, (unsigned int) tulip_ToContext(&dst));

  if (ackhandle) *ackhandle = 1;  /* Done already */

  return 0;
}

/*********************************************************************/
/* Procedure queues a RemoteAction task on a remote node.  The
   tulip_RemoteActionHandler will be invoked on that node.  The
   Queue allows MAXSLOTS remote actions to be issued from this 
   processor to every other processor in a pool.  If all slots 
   are full, this processor will wait in a Poll loop.  A slot is
   freed when this processor receives a message acknowledging the
   completion of a RemoteAction request.  Increasing 
   MAXSLOTS may or may not be beneficial, tulip_Poll() will take 
   longer to finish and the Queue requires more memory. */

int tulip_RemoteAction(int remote_node, char *buffer, int len, tulip_ACK_t *ackhandle)
{
  int j;

  if(len > MAX_RA_BUFF*sizeof(long)) {
    tulip_Message(": Remote action buffer too small. Increase buffer size!",
		 TULIP_FATAL);
  }    

  /* check to see if a slot is available */
  j = 0;
  while(j < MAXSLOTS) {
    if(MyRAQueue[MyNodeNumber][j].length == 0) break;
    j++;
  }
  if(j > MAXSLOTS-1) {
    /* no slot is available. poll and wait */
    j = 0;
    while(MyRAQueue[MyNodeNumber][j].length != 0) {
      tulip_Poll();
      j = (j+1)%MAXSLOTS;
    }
  }

  /* initialize */
  MyRAQueue[MyNodeNumber][j].ackhandle = (long*) 0;
  MyRAQueue[MyNodeNumber][j].ack = 0;

  /* reset ackhandle to zero (if not nill) */
  if (ackhandle) {
    *ackhandle = 0;
    MyRAQueue[MyNodeNumber][j].ackhandle = (long*)ackhandle;
  }

  /* copy buffer for remote action request */
  memcpy(MyRAQueue[MyNodeNumber][j].buffer, buffer, len);

  /* add in the other parameters */
  MyRAQueue[MyNodeNumber][j].length = len;

  /* Send everything to the remote node.  Put is preferred since it's twice
     as fast as get. The address of local MyRAQueue can be used as the remote 
     address as the addresses are indentical on both processors. */

  PutBytes((char*) &MyRAQueue[MyNodeNumber][j], 
	   (char*) &MyRAQueue[MyNodeNumber][j], 
	   sizeof(struct RA_params), 
	   remote_node);

  return 0;
}

/*********************************************************************/
void tulip_Abort() { 
  exit(1); 
}


/*********************************************************************/
static void My_MainBegin(int *argc, char ***argv)
{
  int i, j;

  pcxx_InitParam(argc, *argv);
  tulip_InitClocks();

  MyNodeNumber = _MPP_MY_PE;
  MyTotalNodes = _MPP_N_PES;

  pcxx_EvInit(pcxx_param.tracefile);
  PCXX_EVENT(PCXX_EC_BASIC, PCXX_MAIN_ENTER, 0);
  PCXX_EVENT(PCXX_EC_RUNTIME, PCXX_NEWPROC_ENTER, 0);
 
  /* a prcoessor sets up its data cache so that remote writes into its
     local memory will automatically invalidate the corresponding cache
     line */
  shmem_set_cache_inv();

  /* Zero out my RemoteAction scheduling queue */

  for (i = 0; i < PCXX_MAXPROCS; i++) {
    SendRequest[i].length = 0;
    for (j = 0; j < MAXSLOTS; j++) {
      MyRAQueue[i][j].length = 0;
      MyRAQueue[i][j].ack = 0;
    }
  }

  /* Wait for all the other nodes to Zero out their RemoteAction Queues */
  tulip_Barrier();

  PCXX_EVENT(PCXX_EC_RUNTIME, PCXX_NEWPROC_EXIT, MyNodeNumber);
}

/*********************************************************************/
static void My_MainEnd()
{
  PCXX_EVENT(PCXX_EC_BASIC, PCXX_MAIN_EXIT, 0);
  pcxx_EvClose();
}

/*********************************************************************/
/* This routine ONLY need to check for the arrival of RemoteAction requests 
   and acknowledgments returned by RemoteActions.  On a NDMA machine, 
   fetch/store requests are handled without the help of the remote node */

void tulip_Poll()
{
  int i, j;
  const long m = 1;
  long *p;

  PROFILE_ENTER(PROFILE_Poll);

  /* flush the data cache just in case that the cache is not updated */
  /*shmem_udcflush();*/

  /* poll for remote action requests */
  for (i = 0; i < MyTotalNodes; i++) {
    if(i != MyNodeNumber) {
      for(j = 0; j < MAXSLOTS; j++) {
	if (MyRAQueue[i][j].length != 0) {
	  /* Call RA handler */
	  tulip_RemoteActionHandler(i, (char*)MyRAQueue[i][j].buffer, 
				   MyRAQueue[i][j].length);
	  /* Reset the Queue slot */
	  MyRAQueue[i][j].length = 0;
	  /* Done. Send acknowledgment back to Handle */
	  PutBytes((char*)&MyRAQueue[i][j].ack, (char*)&m, 
		   sizeof(long), i);
	}
      }
    }
  }

  /* poll for acknowledgments */
  for(j = 0; j < MAXSLOTS; j++) {
    if(MyRAQueue[MyNodeNumber][j].ack == 1) {
      /* a remote action acknowledgment has arrived */
      if(MyRAQueue[MyNodeNumber][j].ackhandle) 
	*((int*)MyRAQueue[MyNodeNumber][j].ackhandle) = 1;
      MyRAQueue[MyNodeNumber][j].length = 0;
      MyRAQueue[MyNodeNumber][j].ack = 0;
    }
  }


  /* poll for remote send requests */
  for (i = 0; i < MyTotalNodes; i++) {
    if(i != MyNodeNumber) {
      if (SendRequest[i].length != 0) {
	/* send buffer */
	PutBytes(SendRequest[i].dst,
		 SendRequest[i].src,
		 SendRequest[i].length,
		 i);
	/* Reset local and remote length to zero */
	SendRequest[i].length = 0;	
	PutBytes((char*)&SendRequest[i].length,
		 (char*)&SendRequest[i].length,
		 sizeof(int), i);
      }
    }
  }

  PROFILE_EXIT(PROFILE_Poll);
}

#define MAXBUFF 4096
static char Buffer[MAXBUFF];  /*some address on every processor*/

void tulip_BroadcastBytes(int flag, int length, void * buffer) 
{
  /* 
    This function broadcasts to all processors NOT threads. you
    may use the slot number to write a broadcast that broadcasts to
    threads.  This is a sequential version of the broadcast function. 
    S.Y. (2/2/95)
  */
  int node;

  if(length > MAXBUFF) 
    tulip_Message("Broadcast buffer too small!  Increase buffer", TULIP_FATAL);

  if(flag == 0) {
    /* this processor broadcasts. address of Buffer is identical on all nodes,
       can directly put "buffer" into other nodes' "Buffer" */
    for(node = 0; node < MyTotalNodes; node++) {
      if(node != MyNodeNumber) PutBytes(Buffer, buffer, length, node);
    }
  }
  tulip_Barrier();
  /* flush the data cache just in case the cache is not updated */
  /*shmem_udcflush();*/

  if(flag != 0) memcpy(buffer,Buffer,length);

  /* this barrier prevents a following broadcast from overwriting the buffer 
     which this broadcast is using */ 
  tulip_Barrier();
}

void tulip_ReduceIntAdd(void* buffer,int elem_count) 
{
  int self, span, flag, elem, size;

  size = elem_count*sizeof(int);
  if(size > MAXBUFF) 
    tulip_Message("Reduction buffer too small!  Increase buffer", TULIP_FATAL);

  self = MyNodeNumber;

  /* copy my buffer into Buffer so that other nodes can get it */
  memcpy(Buffer,(char*)buffer,size);
  tulip_Barrier();

  /* tree reduction */
  for(span = 1; span < MyTotalNodes; span*=2) {
    if(self % (2*span) == 0 && (self+span) < MyTotalNodes) {
      /*fetch "Buffer" and temporarily store the fetched content in "buffer"*/
      GetBytes((char *)buffer, (char *)Buffer, size, self+span);      
      for(elem = 0; elem < elem_count; elem++) {
	((int*)Buffer)[elem] += ((int*)buffer)[elem];
      }
    }
    tulip_Barrier();
  }
  /* processor zero broadcasts reduction result */
  flag = 1;
  if(self == 0) {
    flag = 0;
    memcpy((char*)buffer,(char*)Buffer,size);
  }
  tulip_BroadcastBytes(flag, size, buffer);
}

void tulip_ReduceDoubleAdd(void *buffer,int elem_count)
{
  int self, span, flag, elem, size;

  size = elem_count*sizeof(double);
  if(size > MAXBUFF) 
    tulip_Message("Reduction buffer too small!  Increase buffer", TULIP_FATAL);

  self = MyNodeNumber;

  /* copy my buffer into Buffer so that other nodes can get it */
  memcpy(Buffer,(char*)buffer,size);
  tulip_Barrier();

  /* tree reduction */
  for(span = 1; span < MyTotalNodes; span*=2) {
    if(self % (2*span) == 0 && (self+span) < MyTotalNodes) {
      /*fetch "Buffer" and temporarily store the fetched content in "buffer"*/
      GetBytes((char *)buffer, (char *)Buffer, size, self+span);      
      for(elem = 0; elem < elem_count; elem++) {
	((double*)Buffer)[elem] += ((double*)buffer)[elem];
      }
    }
    tulip_Barrier();
  }
  /* processor zero broadcasts reduction result */
  flag = 1;
  if(self == 0) {
    flag = 0;
    memcpy((char*)buffer,(char*)Buffer,size);
  }
  tulip_BroadcastBytes(flag, size, buffer);
}

void tulip_ReduceFloatAdd (void *buffer,int elem_count)
{
  int self, span, flag, elem, size;

  size = elem_count*sizeof(float);
  if(size > MAXBUFF) 
    tulip_Message("Reduction buffer too small!  Increase buffer", TULIP_FATAL);

  self = MyNodeNumber;

  /* copy my buffer into Buffer so that other nodes can get it */
  memcpy(Buffer,(char*)buffer,size);
  tulip_Barrier();

  /* tree reduction */
  for(span = 1; span < MyTotalNodes; span*=2) {
    if(self % (2*span) == 0 && (self+span) < MyTotalNodes) {
      /*fetch "Buffer" and temporarily store the fetched content in "buffer"*/
      GetBytes((char *)buffer, (char *)Buffer, size, self+span);      
      for(elem = 0; elem < elem_count; elem++) {
	((float*)Buffer)[elem] += ((float*)buffer)[elem];
      }
    }
    tulip_Barrier();
  }
  /* processor zero broadcasts reduction result */
  flag = 1;
  if(self == 0) {
    flag = 0;
    memcpy((char*)buffer,(char*)Buffer,size);
  }
  tulip_BroadcastBytes(flag, size, buffer);
}

void tulip_ReduceDoubleMax(void *buffer,int elem_count)
{
  int self, span, flag, elem, size;

  size = elem_count*sizeof(double);
  if(size > MAXBUFF) 
    tulip_Message("Reduction buffer too small!  Increase buffer", TULIP_FATAL);

  self = MyNodeNumber;

  /* copy my buffer into Buffer so that other nodes can get it */
  memcpy(Buffer,(char*)buffer,size);
  tulip_Barrier();

  /* tree reduction */
  for(span = 1; span < MyTotalNodes; span*=2) {
    if(self % (2*span) == 0 && (self+span) < MyTotalNodes) {
      /*fetch "Buffer" and temporarily store the fetched content in "buffer"*/
      GetBytes((char *)buffer, (char *)Buffer, size, self+span);      
      for(elem = 0; elem < elem_count; elem++) {
	if(((double*)buffer)[elem] > ((double*)Buffer)[elem])
	  ((double*)Buffer)[elem] = ((double*)buffer)[elem];
      }
    }
    tulip_Barrier();
  }
  /* processor zero broadcasts reduction result */
  flag = 1;
  if(self == 0) {
    flag = 0;
    memcpy((char*)buffer,(char*)Buffer,size);
  }
  tulip_BroadcastBytes(flag, size, buffer);
}

void tulip_ReduceFloatMax (void *buffer,int elem_count)
{
  int self, span, flag, elem, size;

  size = elem_count*sizeof(float);
  if(size > MAXBUFF) 
    tulip_Message("Reduction buffer too small!  Increase buffer", TULIP_FATAL);

  self = MyNodeNumber;

  /* copy my buffer into Buffer so that other nodes can get it */
  memcpy(Buffer,(char*)buffer,size);
  tulip_Barrier();

  /* tree reduction */
  for(span = 1; span < MyTotalNodes; span*=2) {
    if(self % (2*span) == 0 && (self+span) < MyTotalNodes) {
      /*fetch "Buffer" and temporarily store the fetched content in "buffer"*/
      GetBytes((char *)buffer, (char *)Buffer, size, self+span);      
      for(elem = 0; elem < elem_count; elem++) {
	if(((float*)buffer)[elem] > ((float*)Buffer)[elem])
	  ((float*)Buffer)[elem] = ((float*)buffer)[elem];
      }
    }
    tulip_Barrier();
  }
  /* processor zero broadcasts reduction result */
  flag = 1;
  if(self == 0) {
    flag = 0;
    memcpy((char*)buffer,(char*)Buffer,size);
  }
  tulip_BroadcastBytes(flag, size, buffer);
}

void tulip_ReduceIntMax(void *buffer,int elem_count)
{
  int self, span, flag, elem, size;

  size = elem_count*sizeof(int);
  if(size > MAXBUFF) 
    tulip_Message("Reduction buffer too small!  Increase buffer", TULIP_FATAL);

  self = MyNodeNumber;

  /* copy my buffer into Buffer so that other nodes can get it */
  memcpy(Buffer,(char*)buffer,size);
  tulip_Barrier();

  /* tree reduction */
  for(span = 1; span < MyTotalNodes; span*=2) {
    if(self % (2*span) == 0 && (self+span) < MyTotalNodes) {
      /*fetch "Buffer" and temporarily store the fetched content in "buffer"*/
      GetBytes((char *)buffer, (char *)Buffer, size, self+span);      
      for(elem = 0; elem < elem_count; elem++) {
	if(((int*)buffer)[elem] > ((int*)Buffer)[elem])
	  ((int*)Buffer)[elem] = ((int*)buffer)[elem];
      }
    }
    tulip_Barrier();
  }
  /* processor zero broadcasts reduction result */
  flag = 1;
  if(self == 0) {
    flag = 0;
    memcpy((char*)buffer,(char*)Buffer,size);
  }
  tulip_BroadcastBytes(flag, size, buffer);
}

void tulip_ReduceDoubleMin(void *buffer,int elem_count)
{
  int self, span, flag, elem, size;

  size = elem_count*sizeof(double);
  if(size > MAXBUFF) 
    tulip_Message("Reduction buffer too small!  Increase buffer", TULIP_FATAL);

  self = MyNodeNumber;

  /* copy my buffer into Buffer so that other nodes can get it */
  memcpy(Buffer,(char*)buffer,size);
  tulip_Barrier();

  /* tree reduction */
  for(span = 1; span < MyTotalNodes; span*=2) {
    if(self % (2*span) == 0 && (self+span) < MyTotalNodes) {
      /*fetch "Buffer" and temporarily store the fetched content in "buffer"*/
      GetBytes((char *)buffer, (char *)Buffer, size, self+span);      
      for(elem = 0; elem < elem_count; elem++) {
	if(((double*)buffer)[elem] < ((double*)Buffer)[elem])
	  ((double*)Buffer)[elem] = ((double*)buffer)[elem];
      }
    }
    tulip_Barrier();
  }
  /* processor zero broadcasts reduction result */
  flag = 1;
  if(self == 0) {
    flag = 0;
    memcpy((char*)buffer,(char*)Buffer,size);
  }
  tulip_BroadcastBytes(flag, size, buffer);
}

void tulip_ReduceFloatMin (void *buffer,int elem_count)
{
  int self, span, flag, elem, size;

  size = elem_count*sizeof(float);
  if(size > MAXBUFF) 
    tulip_Message("Reduction buffer too small!  Increase buffer", TULIP_FATAL);

  self = MyNodeNumber;

  /* copy my buffer into Buffer so that other nodes can get it */
  memcpy(Buffer,(char*)buffer,size);
  tulip_Barrier();

  /* tree reduction */
  for(span = 1; span < MyTotalNodes; span*=2) {
    if(self % (2*span) == 0 && (self+span) < MyTotalNodes) {
      /*fetch "Buffer" and temporarily store the fetched content in "buffer"*/
      GetBytes((char *)buffer, (char *)Buffer, size, self+span);      
      for(elem = 0; elem < elem_count; elem++) {
	if(((float*)buffer)[elem] < ((float*)Buffer)[elem])
	  ((float*)Buffer)[elem] = ((float*)buffer)[elem];
      }
    }
    tulip_Barrier();
  }
  /* processor zero broadcasts reduction result */
  flag = 1;
  if(self == 0) {
    flag = 0;
    memcpy((char*)buffer,(char*)Buffer,size);
  }
  tulip_BroadcastBytes(flag, size, buffer);
}

void tulip_ReduceIntMin(void *buffer,int elem_count)
{
  int self, span, flag, elem, size;

  size = elem_count*sizeof(int);
  if(size > MAXBUFF) 
    tulip_Message("Reduction buffer too small!  Increase buffer", TULIP_FATAL);

  self = MyNodeNumber;

  /* copy my buffer into Buffer so that other nodes can get it */
  memcpy(Buffer,(char*)buffer,size);
  tulip_Barrier();

  /* tree reduction */
  for(span = 1; span < MyTotalNodes; span*=2) {
    if(self % (2*span) == 0 && (self+span) < MyTotalNodes) {
      /*fetch "Buffer" and temporarily store the fetched content in "buffer"*/
      GetBytes((char *)buffer, (char *)Buffer, size, self+span);      
      for(elem = 0; elem < elem_count; elem++) {
	if(((int*)buffer)[elem] < ((int*)Buffer)[elem])
	  ((int*)Buffer)[elem] = ((int*)buffer)[elem];
      }
    }
    tulip_Barrier();
  }
  /* processor zero broadcasts reduction result */
  flag = 1;
  if(self == 0) {
    flag = 0;
    memcpy((char*)buffer,(char*)Buffer,size);
  }
  tulip_BroadcastBytes(flag, size, buffer);
}

#include "SPMD.h"


