/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/*==============================================================*/
/* thread interface to  pthreads */
/* Basically, this module implements the interface from thread_iface.h */
/*==============================================================*/

#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>

#define MAX_THREADS 128

typedef void (*voidFuncPtr)(void *);

pthread_mutex_t  gate;           /* Lock that restricts access to in_barrier */
pthread_cond_t   latch;          /* Condition variable for workers*/
volatile int in_barrier;         /* Condition the workers have to wait for */
pthread_key_t pcpt_id_key;
pthread_t new_p[MAX_THREADS];

int TotalThreads = 4;
int ThreadStackSize = 16384;

voidFuncPtr  ThreadCoopFunc;
void        *ThreadCoopArg;

void Persistent_Main(void *);

/*********************************************************************/
/* Does a barrier, if ALL the threads participate.  This is bad code,
   but I inherited it.  Optimizations come later. (PHB) */
void thread_Barrier()
{
  /* Lock the gate */
  pthread_mutex_lock (&gate);

  /* I'm in, inc counter of threads inside gate */
  in_barrier++;

  /* Are we all here? */
  if ( in_barrier == TotalThreads ) {
    /* Yes, and I was last, so I reset the counter */
    in_barrier=0;
    /* tell the other threads they can leave */
    pthread_cond_broadcast (&latch);
  }
  /* I was not the last, I must wait for the other threads */
  else {
    pthread_cond_wait (&latch, &gate);
  };

  /* ok, we are all out */
  pthread_mutex_unlock (&gate);
}

/*********************************************************************/
/* Initialize threads, uses argc/argv, and modifying them This
function is run ONCE per context, in other words, once for shared
memory.  It is not run in parallel */

void thread_Init(int arc, char **argv)
{
  int i;
  pthread_attr_t attr;

  /* Make some thread-specific data area (simply for Thread ID) */
  pthread_keycreate(&pcpt_id_key, NULL);

  /* Initialize a mutex (for barriers) */
  pthread_mutex_init (&gate, pthread_mutexattr_default);

  /* Initialize a condition (for barriers) */
  pthread_cond_init (&latch, pthread_condattr_default);

  /* Create a Pthread attribute */
  pthread_attr_create(&attr);

  /* Set the stack size */
  pthread_attr_setstacksize(&attr, ThreadStackSize);

  /* Set the barrier flag to 0, we are not in a barrier */
  in_barrier = 0;

  /* startup the persistent threads */
  for (i=1; i<TotalThreads; i++) {
    if (pthread_create 
	(&(new_p[i]), attr,
	 (pthread_startroutine_t) Persistent_Main, 
	 (pthread_addr_t) i) == -1) {
      perror ("create");
      exit (1);
    }
  }

  /* Set my (thread 0) thread-specific data */
  /* The other threads set theirs in Persistant_Main */
  /* record my thread ID */
  pthread_setspecific(pcpt_id_key, 0);

  /* Make sure all the other threads are now ready for work */
  thread_Barrier();

  /* Ok, all has been set up.  N-1 threads are in Persistant_Main, */
  /* and I am returning control to the caller of thread_Init() */
}


/*********************************************************************/
/* Shut down and wrap up the thread package */
void thread_Term()
{
  int i;

   /* terminate threads */

   /* Set shared thread-argument to 0, indicating time to quit */
   ThreadCoopArg = 0;

   /* Enter barrier, releasing the worker threads */
   thread_Barrier();

  /* OK, at this point, I'm shutting things down */
  /* wait for worker threads to finish */
  for (i=1; i<TotalThreads; i++) {
    if ( pthread_join (new_p[i], NULL) == -1 ) {
      perror ("join");
      exit (1);
    }
  }

}

/*********************************************************************/
/* Fork some work onto some threads, passing a single void* argument */
void thread_Fork( voidFuncPtr function, void *arg )
{
  /* Set the shared-arguments, so the worker threads can find them */
  ThreadCoopFunc = function;
  ThreadCoopArg = arg;

  /* Enter Barrier, and release workers to do their work */
  thread_Barrier();

  /* Go do my portion */
  function(arg);

  /* Wait for everyone to finish */
  thread_Barrier();
}

/*********************************************************************/
/* Am I somehow "distinguished" from the other threads? */
void thread_IamDistinguished()
{}

/*********************************************************************/
int thread_MyID() 
{
  int my_id=0;
  pthread_getspecific(pcpt_id_key, (any_t *)&my_id);
  return my_id;
}

/*********************************************************************/
int thread_NumThreads()
{
return TotalThreads;
}

/*********************************************************************/
/* First parallel section for threads */
void Persistent_Main(void *x)
{
  int id;
  int done = 0;

  id = (int) x;

  /* printf("Thread %d: Starting Persistent_Main()\n",id); */

  /* record my thread ID */
  pthread_setspecific(pcpt_id_key, (any_t)id);

  /* Wait for everyone to get here, ready for work */
  thread_Barrier();

  /* Loop forever, looking for work, until done is true */
  while(!done){
    
    /* This is strange.  I wait here, in a barrier, until the main
       distinguished thread enters the barrier, and I am free to begin
       computing. */
    thread_Barrier();

    /* printf("Thread %d: Starting ThreadCoopFunc\n",id); */

    /* I get the work to do by looking in a global variable */
    if (ThreadCoopArg) {
      /* Go do my portion of the work */
      ThreadCoopFunc(ThreadCoopArg);

      /* Wait for everyone to finish their portion of work */
      thread_Barrier();
    /* printf("Thread %d: Done ThreadCoopFunc\n",id); */

    }

    /* Hey!  there is no work to do! */
    else {
      /* Set my flag, so I can exit loop */
      done = 1;

      /* Put wrapup code here */

      /* Final Barrier before leaving... */
      thread_Barrier();
    }
  }
}




