/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/


/*
 * pprof.c : parallel profile data files printer
 *
 * (c) 1993 Jerry Manic Saftware
 *
 * Version 1.0
 * Version 2.0  added collection profiling
 * Version 2.1  added summary printing
 * Version 2.2  added collection name and types
 * Version 2.3  machine-independent profile data file format
 * Version 2.4  racy support (dump format)
 * Version 2.5  switched to ASCII files to be really portable
 *
 */

# include <stdio.h>
# include <stdlib.h>

# include <math.h>
# include <errno.h>

# ifndef TRUE
#   define FALSE 0
#   define TRUE  1
# endif

# if defined(ultrix) || defined(sequent) || defined(butterfly)
double fmod (double x, double y)
{
  return ( x - floor(x/y) * y );
}
# endif

static struct p_func_descr {
    int     numcalls;
    double  usec;
    double  cumusec;
} *p_func_list = 0;    /* -- function profile data -- */

static struct p_prof_elem {
    char   *name;
    int     tag;
    int     numcalls;
    double  usec;
    double  cumusec;
} *p_prof_tbl  = 0,    /* -- function profile data table -- */
  *p_min_tbl = 0,      /* -- minimal summary function profile data table -- */
  *p_max_tbl = 0,      /* -- maximal summary function profile data table -- */
  *p_total_tbl = 0;    /* -- summary function profile data table -- */

static struct p_coll_descr {
    int numelem;
    int dim;
    int size;
    int localacs;
    int remoteacs;
    char *collname;
    char *elemname;
    char *varname;
} *p_coll_list = 0,    /* -- collection profile data table -- */
  *p_coll_tbl  = 0;    /* -- summary collection profile data table -- */

static char *strsave (char *s)
{
  char *r;

  if ( (r = malloc (strlen(s)+1)) == NULL ) {
    fprintf (stderr, "error: no more memory\n");
    exit (1);
  }
  strcpy (r, s);
  return r;
}

/*
 * ToTimeStr: convert usec to hh:mm:ss.mmm
 */

static char *ToTimeStr (double ti, char timbuf[])
{
  long msec, sec, min, hour;

  msec = fmod (ti / 1.0e3, 1.0e3);
  sec  = fmod (ti / 1.0e6, 60.0);
  min  = fmod (ti / 60.0e6, 60.0);
  hour = ti / 36.0e8;

  if ( hour )
    sprintf (timbuf, "%2d:%02d:%02d.%03d", hour, min, sec, msec);
  else if ( min )
    sprintf (timbuf, "   %2d:%02d.%03d", min, sec, msec);
  else if ( sec )
    sprintf (timbuf, "      %2d.%03d", sec, msec);
  else
    sprintf (timbuf, "         %3d", msec);

  return (timbuf);
}

/*
 * MsecCmp   : compare usec field
 * CumMsecCmp: compare cumusec field
 * CallCmp   : compare numcalls field
 *
 * functions return (unsually) -1 for greater than, 0 for equal, 1 for less than
 * in order to reverse the order of sort (descending instead of ascending)
 * can be changed through sign
 */

static int sign = 1;

static int MsecCmp (const void *left, const void *right)
{
  double l = ((struct p_prof_elem *) left)->usec;
  double r = ((struct p_prof_elem *) right)->usec;

  if ( l < r )
    return sign;
  else if ( l > r )
    return -sign;
  else
    return 0;
}

static int CumMsecCmp (const void *left, const void *right)
{
  double l = ((struct p_prof_elem *) left)->cumusec;
  double r = ((struct p_prof_elem *) right)->cumusec;

  if ( l < r )
    return sign;
  else if ( l > r )
    return -sign;
  else
    return 0;
}

static int CallCmp (const void *left, const void *right)
{
  int l = ((struct p_prof_elem *) left)->numcalls;
  int r = ((struct p_prof_elem *) right)->numcalls;

  if ( l < r )
    return sign;
  else if ( l > r )
    return -sign;
  else
    return 0;
}

static int (* compar)(const void *, const void *) = CumMsecCmp;

/*
 * PrintFuncTab : Print plain function profile data
 * DumpFuncTab  : Dump plain function profile data
 */

static int numfunc;          /* -- number of functions   -- */

static void PrintFuncTab (struct p_prof_elem *tab, double total, int max)
{
  int i;
  int o_numcalls = 0;
  double o_usec = 0.0;
  double o_cumusec = 0.0;
  char buf1[20], buf2[20];

  for (i=max; i<numfunc; i++) {
    o_numcalls += tab[i].numcalls;
    o_usec     += tab[i].usec;
    o_cumusec  += tab[i].cumusec;
  }

  printf ("------------------------------------------------------------\n");
  printf ("%%time         msec   total msec    #call  usec/call name\n");
  printf ("------------------------------------------------------------\n");
  for (i=0; i<numfunc && i<max; i++) {
    if ( tab[i].numcalls ) {
      if ( tab[i].usec > 0.0 ) {
        printf ("%5.1f %s %s %8d %10.0f %s\n",
          tab[i].cumusec / total * 100.0,
          ToTimeStr (tab[i].usec, buf1),
          ToTimeStr (tab[i].cumusec, buf2),
          tab[i].numcalls,
          tab[i].cumusec / tab[i].numcalls,
          tab[i].name);
      }
      else {
        printf ("                                %8d            %s\n",
          tab[i].numcalls, tab[i].name);
      }
    }
  }

  if ( o_numcalls > 0 ) {
    if ( o_usec > 0.0 ) {
      printf ("%5.1f %s %s %8d %10.0f -others-\n",
        o_cumusec / total * 100.0,
        ToTimeStr (o_usec, buf1), ToTimeStr (o_cumusec, buf2),
        o_numcalls, o_cumusec / o_numcalls);
    }
    else {
      printf ("                                %8d            -others-\n",
        o_numcalls);
    }
  }
}

static void DumpFuncTab (struct p_prof_elem *tab, int id, double total,
                         int max, char *order)
{
  int i;
  int o_numcalls = 0;
  double t = 0.0;
  double o_usec = 0.0;
  double o_cumusec = 0.0;
  char buf1[20], buf2[20];

  for (i=0; i<numfunc; i++) {
    if ( tab[i].numcalls ) t += tab[i].usec;
    if ( i >= max ) {
      o_numcalls += tab[i].numcalls;
      o_usec     += tab[i].usec;
      o_cumusec  += tab[i].cumusec;
    }
  }

  for (i=0; i<numfunc && i<max; i++) {
    if ( tab[i].numcalls ) {
      if ( id < 0 ) printf ("%c ", -id); else printf ("%d ", id);
      printf ("%d %s %s ", tab[i].tag, tab[i].name, order);
      if ( order[0] == 'e' )
        printf ("%.16G %4.2f\n", tab[i].usec, tab[i].usec / t * 100.0);
      else if ( order[0] == 'i' )
        printf ("%.16G %4.2f\n", tab[i].cumusec, tab[i].cumusec/total*100.0);
      if ( tab[i].usec > 0.0 ) {
        printf ("%5.1f %s %s %8d %10.0f %s\n",
          tab[i].cumusec / total * 100.0,
          ToTimeStr (tab[i].usec, buf1), ToTimeStr (tab[i].cumusec, buf2),
          tab[i].numcalls, tab[i].cumusec / tab[i].numcalls, tab[i].name);
      }
      else {
        printf ("                                %8d            %s\n",
          tab[i].numcalls, tab[i].name);
      }
    }
  }

  if ( o_numcalls > 0 ) {
    if ( id < 0 ) printf ("%c ", -id); else printf ("%d ", id);
    printf (" -1 -others- %s ", order);
    if ( order[0] == 'e' )
      printf ("%.16G %4.2f\n", o_usec, o_usec / t * 100.0);
    else if ( order[0] == 'i' )
      printf ("%.16G %4.2f\n", o_cumusec, o_cumusec / total * 100.0);
    if ( o_usec > 0.0 ) {
      printf ("%5.1f %s %s %8d %10.0f -others-\n",
        o_cumusec / total * 100.0,
        ToTimeStr (o_usec, buf1), ToTimeStr (o_cumusec, buf2),
        o_numcalls, o_cumusec / o_numcalls);
    }
    else {
      printf ("                                %8d            -others-\n",
        o_numcalls);
    }
  }
}

/*
 * ReadNameTable: read function name table file
 * ProcessFile  : read profile data node file and print results
 */

static char *depfile;        /* -- corresponding .dep file -- */
static char proffile[256];   /* -- profile data file     -- */
static char **funcnamebuf;   /* -- function name table   -- */
static int *functagbuf;      /* -- function tag table    -- */
static int numcoll;          /* -- number of collections -- */
static double total_total;   /* -- total time overall    -- */
static double min_total;
static double max_total;
static int nodeprint = TRUE;
static int dump = FALSE;
static int list = FALSE;
static char lbuf[256];        /* -- temporary line buffer for reads -- */
static char sbuf[128];        /* -- temporary string buffer -- */

static void ReadNameTable (char file[])
{
  int i, tag;
  FILE *in;

  sprintf (proffile, "%s.ftab", file);

  if ( (in = fopen (proffile, "r")) == NULL ) {
    perror (proffile);
    exit (1);
  }

  /* -- read depfile -- */
  if ( fgets (lbuf, 256, in) == NULL ) {
    fprintf (stderr, "%s: cannot read function table\n", proffile);
    exit (1);
  }
  sscanf (lbuf, "%s", sbuf);
  depfile = strsave (sbuf);

  /* -- read number of functions -- */
  fgets (lbuf, 256, in);
  sscanf (lbuf, "%d", &numfunc);
  
  /* -- read function table -- */
  funcnamebuf = malloc (numfunc * sizeof(char *));
  functagbuf  = malloc (numfunc * sizeof(int));
  for (i=0; i<numfunc; i++) {
    fgets (lbuf, 256, in);
    sscanf (lbuf, "%d %s", &tag, sbuf);
    if ( tag < 0 )
      funcnamebuf[i] = NULL;
    else
      funcnamebuf[i] = strsave (sbuf);
    functagbuf[i] = tag;
  }
  fclose (in);
}

static int ProcessFile (int no, int max, char prefix[], int ignore)
{
  int i, n, r, s, d, l;
  int numf;
  int numc;
  FILE *in;
  double t1, t2;
  double total, ct;
  extern int errno;
  char s1[128], s2[128];

  sprintf (proffile, "%s.%d", prefix, no);

  /* -- read profile data file and set profile data tables ------------------ */
  /* ------------------------------------------------------------------------ */
  if ( (in = fopen (proffile, "r")) == NULL ) {
    if ( errno == ENOENT && ignore)
      return (FALSE);
    else {
      perror (proffile);
      exit (1);
    }
  }

  /* -- read number of functions -------------------------------------------- */
  if ( fgets (lbuf, 256, in) == NULL ) {
    fprintf (stderr,"invalid proftablefile: cannot read number of functions\n");
    exit (1);
  }
  sscanf (lbuf, "%d", &numf);

  if ( numf != numfunc ) {
    fprintf (stderr, "%s: number of functions does not match\n", proffile);
    exit (1);
  }

  /* -- read and setup function profile data -------------------------------- */
  p_func_list = malloc (numf * sizeof(struct p_func_descr));
  for (i=0; i<numf; i++) {
    fgets (lbuf, 256, in);

#ifdef __bsdi__
    /* Wierd bogus HACK because bsdi is funky (PHB) */
    sscanf (lbuf, "%d %lf %lf", &n, &t1, &t2);
#else
    sscanf (lbuf, "%d %lG %lG", &n, &t1, &t2);
#endif

    p_func_list[i].numcalls = n;
    p_func_list[i].usec     = t1;
    p_func_list[i].cumusec  = t2;
  }

  /* -- read number of collections ------------------------------------------ */
  if ( fgets (lbuf, 256, in) == NULL ) {
    fprintf (stderr,
                  "invalid proftablefile: cannot read number of collections\n");
    exit (1);
  }
  sscanf (lbuf, "%d", &numc);

  /* -- setup and read collection profile data ------------------------------ */
  if ( numc ) {
    p_coll_list = malloc (numc * sizeof(struct p_coll_descr));

    for (i=0; i<numc; i++) {
      fgets (lbuf, 256, in);
      sscanf (lbuf,"%d %d %d %d %d %s %s %s", &n, &d, &s, &l, &r, sbuf, s1, s2);

      p_coll_list[i].numelem   = n;
      p_coll_list[i].dim       = d;
      p_coll_list[i].size      = s;
      p_coll_list[i].localacs  = l;
      p_coll_list[i].remoteacs = r;
      p_coll_list[i].collname = strsave(sbuf);
      p_coll_list[i].elemname = strsave(s1);
      p_coll_list[i].varname  = strsave(s2);
    }

    if ( !p_coll_tbl ) {
      p_coll_tbl = malloc (numc * sizeof(struct p_coll_descr));
      for (i=0; i<numc; i++) {
        p_coll_tbl[i].numelem   = p_coll_list[i].numelem;
        p_coll_tbl[i].dim       = p_coll_list[i].dim;
        p_coll_tbl[i].size      = p_coll_list[i].size;
        p_coll_tbl[i].localacs  = 0;
        p_coll_tbl[i].remoteacs = 0;
        p_coll_tbl[i].collname = p_coll_list[i].collname;
        p_coll_tbl[i].elemname = p_coll_list[i].elemname;
        p_coll_tbl[i].varname  = p_coll_list[i].varname;
      }
      numcoll = numc;
    }
    else if ( numc != numcoll ) {
      fprintf (stderr, "%s: number of collections does not match\n", proffile);
      exit (1);
    }
  }
  fclose (in);

  /* -- initialize summary function profile data table ---------------------- */
  if ( !p_total_tbl ) {
    p_total_tbl = malloc (numf * sizeof(struct p_prof_elem));
    p_min_tbl   = malloc (numf * sizeof(struct p_prof_elem));
    p_max_tbl   = malloc (numf * sizeof(struct p_prof_elem));
    for (i=0; i<numf; i++) {
      p_total_tbl[i].tag      = functagbuf[i];
      p_total_tbl[i].name     = funcnamebuf[i];
      p_total_tbl[i].usec     = 0.0;
      p_total_tbl[i].cumusec  = 0.0;
      p_total_tbl[i].numcalls = 0;

      p_max_tbl[i].tag      = p_min_tbl[i].tag      = p_total_tbl[i].tag;
      p_max_tbl[i].name     = p_min_tbl[i].name     = p_total_tbl[i].name;
      p_max_tbl[i].usec     = p_min_tbl[i].usec     = p_func_list[i].usec;
      p_max_tbl[i].cumusec  = p_min_tbl[i].cumusec  = p_func_list[i].cumusec;
      p_max_tbl[i].numcalls = p_min_tbl[i].numcalls = p_func_list[i].numcalls;
    }
    total_total = 0.0;
    max_total = min_total = p_func_list[0].cumusec;
  }

  /* -- set function profile data table ------------------------------------- */
  /* -- and update summary function profile data table ---------------------- */
  p_prof_tbl = malloc (numf * sizeof(struct p_prof_elem));
  for (i=0; i<numf; i++) {
    p_prof_tbl[i].tag   = p_total_tbl[i].tag;
    p_prof_tbl[i].name  = p_total_tbl[i].name;
    p_total_tbl[i].usec     += p_prof_tbl[i].usec     = p_func_list[i].usec;
    p_total_tbl[i].cumusec  += p_prof_tbl[i].cumusec  = p_func_list[i].cumusec;
    p_total_tbl[i].numcalls += p_prof_tbl[i].numcalls = p_func_list[i].numcalls;

    if ( p_min_tbl[i].usec     > p_func_list[i].usec )
      p_min_tbl[i].usec     = p_func_list[i].usec;
    if ( p_min_tbl[i].cumusec  > p_func_list[i].cumusec )
      p_min_tbl[i].cumusec  = p_func_list[i].cumusec;
    if ( p_min_tbl[i].numcalls > p_func_list[i].numcalls )
      p_min_tbl[i].numcalls = p_func_list[i].numcalls;

    if ( p_max_tbl[i].usec     < p_func_list[i].usec )
      p_max_tbl[i].usec     = p_func_list[i].usec;
    if ( p_max_tbl[i].cumusec  < p_func_list[i].cumusec )
      p_max_tbl[i].cumusec  = p_func_list[i].cumusec;
    if ( p_max_tbl[i].numcalls < p_func_list[i].numcalls )
      p_max_tbl[i].numcalls = p_func_list[i].numcalls;
  }

  /* -- get total runtime (time of "main" function (always function 0)) ----- */
  total_total += total = p_func_list[0].cumusec;
  if ( min_total > p_func_list[0].cumusec ) min_total = p_func_list[0].cumusec;
  if ( max_total < p_func_list[0].cumusec ) max_total = p_func_list[0].cumusec;

  /* -- print function profile data table ----------------------------------- */
  if ( nodeprint ) {
    if ( dump ) {
      qsort ((void *) p_prof_tbl, numf, sizeof(struct p_prof_elem), MsecCmp);
      DumpFuncTab (p_prof_tbl, no, total, max, "excl");
      qsort ((void *) p_prof_tbl, numf, sizeof(struct p_prof_elem), CumMsecCmp);
      DumpFuncTab (p_prof_tbl, no, total, max, "incl");
    }
    else {
      qsort ((void *) p_prof_tbl, numf, sizeof(struct p_prof_elem), compar);
      printf ("\nNODE %d:\n", no);
      PrintFuncTab (p_prof_tbl, total, max);
    }
  }

  if ( numc ) {
    if ( nodeprint ) {
      if ( ! dump ) {
        printf ("\n   local    remote  collection\n");
        printf (  "accesses  accesses  num   name\n");
      }
    }
    for (i=0; i<numc; i++) {
      if ( nodeprint ) {
        /* -- print collection profile data table --------------------------- */
        if ( dump ) {
          if ( ct = p_coll_list[i].localacs + p_coll_list[i].remoteacs ) {
            printf ("coll %d %d %d %4.2f %d %4.2f\n", no, i, 
                   p_coll_list[i].localacs, p_coll_list[i].localacs/ct*100.0,
                   p_coll_list[i].remoteacs, p_coll_list[i].remoteacs/ct*100.0);
          }
          else {
            printf ("coll %d %d 0 0.0 0 0.0\n", no, i);
          }
        }
        else {
          printf ("%8d  %8d  %3d   %s\n",
            p_coll_list[i].localacs, p_coll_list[i].remoteacs,
            i, p_coll_list[i].varname);
        }
      }

      /* -- compute collection profile summary ------------------------------ */
      p_coll_tbl[i].localacs  += p_coll_list[i].localacs;
      p_coll_tbl[i].remoteacs += p_coll_list[i].remoteacs;
    }
  }

  free (p_coll_list);
  free (p_func_list);
  free (p_prof_tbl);
  return (TRUE);
}

/*
 * PrintFuncSummary: Print summary function profile data
 * PrintSummary    : Print all summary profile information
 */

static void PrintFuncSummary (struct p_prof_elem *tab, double total,
                              int max, char *message, char *ident)
{
  /* -- print summary function profile data table --------------------------- */
  if ( dump ) {
    qsort ((void *) tab, numfunc, sizeof(struct p_prof_elem), MsecCmp);
    DumpFuncTab (tab,  -ident[0], total, max, "excl");
    qsort ((void *) tab, numfunc, sizeof(struct p_prof_elem), CumMsecCmp);
    DumpFuncTab (tab,  -ident[0], total, max, "incl");
  }
  else {
    qsort ((void *) tab, numfunc, sizeof(struct p_prof_elem), compar);
    printf ("\nFUNCTION SUMMARY (%s):\n", message);
    PrintFuncTab (tab, total, max);
  }
}

static void PrintSummary (int max, int numproc)
{
  int i;
  double ct;

  if ( (numproc > 1) || dump ) {
    PrintFuncSummary (p_total_tbl, total_total, max, "total", "t");
    PrintFuncSummary (p_min_tbl, min_total, max, "min", "<");
    PrintFuncSummary (p_max_tbl, max_total, max, "max", ">");

    for (i=0; i<numfunc; i++) {
      p_total_tbl[i].numcalls /= numproc;
      p_total_tbl[i].usec     /= numproc;
      p_total_tbl[i].cumusec  /= numproc;
    }
    total_total /= numproc;

    PrintFuncSummary (p_total_tbl, total_total, max, "mean", "m");
  }

  if ( numcoll ) {
    if ( dump ) {
      for (i=0; i<numcoll; i++) {
        ct = p_coll_tbl[i].localacs + p_coll_tbl[i].remoteacs;
        printf ("cinfo %d %s<%s> %s %d %d %d %d %4.2f %d %4.2f\n",
          i, p_coll_tbl[i].collname, 
          p_coll_tbl[i].elemname, p_coll_tbl[i].varname,
          p_coll_tbl[i].numelem, p_coll_tbl[i].size, p_coll_tbl[i].dim,
          p_coll_tbl[i].localacs, p_coll_tbl[i].localacs/ct*100.0,
          p_coll_tbl[i].remoteacs, p_coll_tbl[i].remoteacs/ct*100.0);
      }
    }
    else {
      printf ("\nCOLLECTION SUMMARY:\n");
      printf ("------------------------------------------------------------\n");
      for (i=0; i<numcoll; i++) {
        printf ("%s<%s> %s, collection #%d\n", p_coll_tbl[i].collname,
          p_coll_tbl[i].elemname, p_coll_tbl[i].varname, i);
        printf ("\t%d elements of size %d, %d-dimensional\n",
          p_coll_tbl[i].numelem, p_coll_tbl[i].size, p_coll_tbl[i].dim);
        printf ("\t%8d local / %8d remote accesses\n",
          p_coll_tbl[i].localacs, p_coll_tbl[i].remoteacs);
      }
    }
  }
}

/*
 * profile main function
 */

int main (int argc, char *argv[])
{
  int argno;
  int ch;
  int i;
  int max = 999999999;
  int errflag;
  char *file = "profile";

  extern char *optarg;
  extern int optind;

  /* -- parse command line arguments ---------------------------------------- */
  errflag = FALSE;
  while ( (ch = getopt (argc, argv, "cdf:lmn:rst")) != EOF ) {
    switch ( ch ) {
    case 'c': /* -- sort according to number of *C*alls -- */
      compar = CallCmp;
      break;
    case 'd': /* -- *D*ump output format (for racy) -- */
      dump = TRUE;
      break;
    case 'f': /* -- profile data *F*ile prefix -- */
      file = optarg;
      break;
    case 'l': /* -- *L*ist function table for debug purposes -- */
      list = TRUE;
      break;
    case 'm': /* -- sort according to *M*illiseconds -- */
      compar = MsecCmp;
      break;
    case 'n': /* -- print only first n *N*umber of funtions -- */
      max = atoi(optarg);
      break;
    case 'r': /* -- *R*everse sorting order -- */
      sign = -1;
      break;
    case 's': /* -- print only *S*ummary profile information -- */
      nodeprint = FALSE;
      break;
    case 't': /* -- sort according to *T*otal milliseconds -- */
      compar = CumMsecCmp;
      break;
    default:
      errflag = TRUE;
      break;
    }
  }
  if ( errflag ) {
    fprintf (stderr, "usage: %s [-c|-m|-t] [-r] [-s] [-n num] [node numbers]\n",
             argv[0]);
    exit (1);
  }

  ReadNameTable (file);
  if ( list ) {
    printf ("%s:\n", depfile);
    for (i=0; i<numfunc; i++) {
      if ( functagbuf[i] > 0 )
        printf ("%5d  %s\n", functagbuf[i], funcnamebuf[i]);
    }
    exit (0);
  } else if ( dump ) {
    printf ("%s\n%d functions\n", depfile, numfunc);
  }

  if ( optind == argc ) {
    for (argno = 0; ProcessFile(argno, max, file, TRUE); argno++);
  }
  else {
    for (argno = optind; argno < argc; argno++)
      ProcessFile (atoi(argv[argno]), max, file, FALSE);
    argno -= optind;
  }

  if ( argno ) PrintSummary (max, argno);

  exit (0);
}
