; s y n c h . s c h
;
;   Copyright (C) 1995 The University of Oregon
;   Copyright (C) 1995 Lars Thomas Hansen <lth@cs.uoregon.edu>
;
;   Permission to use, copy, and/or distribute this software and its
;   documentation for any purpose and with or without fee is hereby granted,
;   provided that both the above copyright notice and this permission notice
;   appear in all copies and derived works.  
;
;   This software is provided ``as is'' without express or implied warranty.
;
; lth@cs.uoregon.edu / June 27, 1995
; $Id: synch.sch,v 1.4 1995/07/26 18:29:59 lth Exp lth $
;
; Synchronization primitives for threads package (uniprocessor).
;
; Note that these are implemented using high-level tasking primitives 
; only: there is no knowledge of the internals of the tasking system.
;
; Mutual exclusion through mutex objects (binary semaphores)
; - 'Make-mutex' creates a mutex.
; - 'Mutex?' tells you whether an object is a mutex or not.
; - 'Mutex-wait' waits on a mutex until it becomes available, then
;   makes it unavailable.
; - 'Mutex-signal' makes a mutex available. No checking is performed
;   to ascertain that the calling task holds the mutex or indeed that
;   the mutex is in the blocking state. Mutex-signal is idempotent: 
;   signalling a mutex which is not waiting does nothing.
; - 'Call-with-mutex' is an obvious abstraction on these.
;
; Mutual exclusion through counting semaphores
; - 'Make-semaphore' takes an initial count (>= 0) and returns a semaphore.
; - 'Semaphore?' tests whether an object is a semaphore.
; - 'Semaphore-down' blocks the task if the value of the semaphore is 0,
;   and otherwise decreases the semaphore's value by 1.
; - 'Semaphore-up' increases the value of the semaphore by 1; if the value
;   was 0 and tasks are blocked a task is released (and the value remains 0).
;
; Condition variables
; - 'Mutex-signal-and-semaphore-down' does a signal on a mutex and a
;   down on a semaphore, atomically, blocking the task if the semaphore's
;   value was 0.  When the semaphore becomes non-zero, the mutex is 
;   re-locked and the task is resumed. Essentially, this implements 
;   condition variables as semaphores.
;
; Barriers synchronize the progress of multiple threads.
; - 'Make-barrier' takes an integer (the number of threads which will
;   eventually enter the barrier) and returns a barrier object.
; - 'Barrier?' tests whether an object is a barrier.
; - 'Enter-barrier' takes a barrier as an argument and goes into the
;   barrier; the thread blocks until all the threads which will enter
;   the barrier have entered the barrier. The barrier then becomes
;   useless (the count is not reset).
;
; Sync objects implement rendezvous.
; - 'Make-sync' returns a sync object.
; - 'Sync?' tests whether an object is a sync object.
; - 'Sync-ref' gets the value of a sync, blocking if there is not yet
;   a value.
; - 'Sync-set!' sets the value of the sync and releases all threads
;   blocking on the sync.
;
; Atomic lambda expressions are shorthands for single-threaded procedures.
; - 'Atomic-lambda' is like 'lambda', but there will only be one thread
;   at a time inside the body of the procedure.
;
; NOTES
;   It is an error to do a wait on a mutex or semaphore while in a critical
;   section. This can be made less restrictive by saying that it is an error
;   to block on a mutex or semaphore in a critical section. Don't know yet
;   what makes more sense. Either way, THIS ERROR IS NOT DETECTED.
;
; BUGS and DISCUSSION
; - These primitives should be moved in the direction of the Modula-3
;   primitives.
; - There is no 'semaphore-broadcast' to release all tasks waiting
;   on the semaphore.
; - It would be meaningful to split this file into two: low-level (semaphores
;   and mutexes) and high-level (barriers, syncs, and atomic-lambdas),
;   since the high-level are implemented entirely in terms of the low-level.
; - Should barriers be reusable?
; - Is mutex-signal-and-semaphore-down meaningful? Semaphores don't forget
;   like condition variables do.

(require "queue.sch")
(require "threads.sch")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Mutexes

; An identifying parameter can be passed to make-mutex for debugging purposes.

(define (make-mutex . rest)
  (vector 'mutex
	  #t
	  (make-basic-queue)
	  (if (null? rest) 'anonymous (car rest))))

(define (mutex? m)
  (and (vector? m) (> (vector-length m) 0) (eq? (vector-ref m 0) 'mutex)))

(define (mutex-wait m)
  (if (not (mutex? m))
      (error "mutex-wait: not a mutex: ~a" m))
  (call-without-interrupts
   (lambda ()
     (if (not (vector-ref m 1))
	 (block-task (lambda (t) (qput (vector-ref m 2) t)))
	 (vector-set! m 1 #f)))))

; This is idempotent for good reasons.

(define (mutex-signal m)
  (if (not (mutex? m))
      (error "mutex-signal: not a mutex: ~a" m))
  (call-without-interrupts
   (lambda ()
     (if (not (vector-ref m 1))
	 (let ((tasks (vector-ref m 2)))
	   (if (not (qempty? tasks))
	       (ready-task (qget tasks))
	       (vector-set! m 1 #t)))))))

(define (call-with-mutex m thunk)
  (mutex-wait m)
  (let ((r (thunk)))
    (mutex-signal m)
    r))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Semaphores

(define (make-semaphore init . rest)
  (vector 'semaphore
	  init
	  (make-basic-queue)
	  (if (null? rest) 'anonymous (car rest))))

(define (semaphore? m)
  (and (vector? m) (> (vector-length m) 0) (eq? (vector-ref m 0) 'semaphore)))

(define (semaphore-down sem)
  (if (not (semaphore? sem))
      (error "semaphore-down: not a semaphore: ~a" m))
  (call-without-interrupts
   (lambda ()
     (if (zero? (vector-ref sem 1))
	 (block-task (lambda (t) (qput (vector-ref sem 2) t)))
	 (vector-set! sem 1 (- (vector-ref sem 1) 1))))))

(define (semaphore-up sem)
  (if (not (semaphore? sem))
      (error "semaphore-up: not a semaphore: ~a" sem))
  (call-without-interrupts
   (lambda ()
     (if (zero? (vector-ref sem 1))
	 (let ((tasks (vector-ref sem 2)))
	   (if (not (qempty? tasks))
	       (ready-task (qget tasks))
	       (vector-set! sem 1 1)))
	 (vector-set! sem 1 (+ (vector-ref sem 1) 1))))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Condition variables

; Signal the mutex m, then do a down on the semaphore s, but do not allow
; task switches inbetween. If the task blocks on s, then when s is released,
; wait on m again before returning.

(define (mutex-signal-and-semaphore-down m s)
  (if (or (not (mutex? m)) (not (semaphore? s)))
      (error "mutex-signal-and-semaphore-down: bad args: " m s))
  (call-without-interrupts
   (lambda ()
     (mutex-signal m)
     (if (zero? (vector-ref s 1))
	 (block-task (lambda (t) (qput (vector-ref s 2) t)))
	 (vector-set! s 1 (- (vector-ref s 1) 1)))))
  ; Interrupts will be enabled when we get to this point.
  (mutex-wait m))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Barrier synchronization
;
; A barrier synchronizes a number of threads, letting them all proceed
; when they have all entered the barrier.
;
; A barrier is a one-time data structure: once used, it must be discarded.
;
; Example:
;   (define b (make-barrier 3))  ; synchronize 3 threads
;   (do ((i 0 (+ i 1)))
;       ((= i 3) #t)
;     (spawn (lambda () (enter-barrier b))))

(define (make-barrier num-members)
  (vector 'barrier 
	  (make-mutex)
	  num-members
	  0
	  (make-semaphore 0)))

(define (barrier? b)
  (and (vector? v)
       (> (vector-length b) 0)
       (eq? (vector-ref b 0) 'barrier)))

(define enter-barrier
  (let ()

    (define (barrier.mutex b) (vector-ref b 1))
    (define (barrier.members b) (vector-ref b 2))
    (define (barrier.waiting b) (vector-ref b 3))
    (define (barrier.waiting! b n) (vector-set! b 3 n))
    (define (barrier.sem b) (vector-ref b 4))

    (lambda (b)
      (mutex-wait (barrier.mutex b))
      (barrier.waiting! b (+ (barrier.waiting b) 1))
      (let ((w (barrier.waiting b))
	    (m (barrier.members b))
	    (s (barrier.sem b)))
	(if (= w m)
	    (begin (do ((i 1 (+ i 1)))
		       ((= i m) #t)
		     (semaphore-up s))
		   (mutex-signal (barrier.mutex b)))
	    (begin (mutex-signal (barrier.mutex b))
		   (semaphore-down s)))))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Sync objects. 
;
; A sync object is an object which initially has no value, and any read
; accesses to the object will block until the object is instantiated. When
; the object gets a value, all waiting readers are unblocked, and all
; subsequent read accesses will succeed immediately.
;
; It is an error to give a value to the sync more than once, but we 
; don't detect this.

(define (make-sync)
  (vector 'sync (make-semaphore 0) #f))

(define (sync? s)
  (and (vector? s)
       (> (vector-length s) 0)
       (eq? (vector-ref s 0) 'sync)))

(define (sync-set! s v)
  (if (sync? s)
      (begin (vector-set! s 2 v)
	     (semaphore-up (vector-ref s 1)))
      (error "sync-set!: not a sync: " s)))

(define (sync-ref s)
  (if (sync? s)
      (let ((v (vector-ref s 2)))
	(if (not (eq? v #f))
	    v
	    (begin (semaphore-down (vector-ref s 1))
		   (let ((v (vector-ref s 2)))
		     (semaphore-up (vector-ref s 1))
		     v))))
      (error "sync-ref: not a sync: " s)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Atomic lambda expressions
;
; An atomic lambda expression is a lambda expression whose body is controlled
; by a mutex: only one thread can be in the body at any one time.
; 
; Example:
;   (define x (atomic-lambda (v i o) (vector-set! v i o)))

(define-macro (atomic-lambda args . body)
  (let ((name (gensym))
	(r    (gensym)))
    `(let ((,name (make-mutex)))
       (lambda ,args
	 (mutex-wait ,name)
	 (let ((,r (begin ,@body)))
	   (mutex-signal ,name)
	   ,r)))))


(provide "synch.sch")

; eof
