/*
 * t a s k i n g . c                      -- Low-level tasking primitives
 *
 *   Copyright (C) 1995 The University of Oregon
 *   Copyright (C) 1995 Lars Thomas Hansen <lth@cs.uoregon.edu>
 *
 *   Permission to use, copy, and/or distribute this software and its
 *   documentation for any purpose and with or without fee is hereby granted,
 *   provided that both the above copyright notice and this permission notice
 *   appear in all copies and derived works.  
 *
 *  This software is provided ``as is'' without express or implied warranty.
 *
 *        Author: Lars Thomas Hansen [lth@cs.uoregon.edu]
 * Creation date: June 22, 1995
 *   Last update: July 24, 1995 / lth
 *
 * $Id$
 *
 * A full tasking package and mutual exclusion facilities can be built
 * using these primitives and first-class continuations.
 *
 * See the file tasking.{tex,ps} for a complete explanation.
 *
 * This code requires select().
 */

/* Could use poll(2) on SunOS, Solaris. */
#ifndef HAVE_SELECT
#error "The tasking system requires select()."
#endif

#include <sys/time.h>
#include <sys/types.h>
#include <tk.h>
#include "stk.h"

static long timer = 0;               /* timer count */
static int  timer_enable = 0;        /* interrupt enable flag */
static int  return_from_polling;     /* flag for polling loop */
static int  tasking_on = 0;          /* 1 when the tasking system is ready */

static SCM  interrupt_handler;       /* thunk for handling interrupts */
static SCM  ioblock_handler;         /* procedure for blocking on filedesc. */
static SCM  blocksym = 0;            /* the symbol for "block" */
static SCM  unblocksym = 0;          /* the symbol for "unblock" */

static void signal_io_available(), setup_io_proc(), suspend_io_proc();

static int largest = 0;              /* largest descriptor in a set */
static fd_set readfds;                /* Set of active descriptors */

#define TRUEP( x ) NEQ( (x), Ntruth )
#define FALSEP( x ) EQ( (x), Ntruth )


/* Called from the evaluator (in Src/eval.c) at every procedure call.
 * If the timer has expired and interrupts are enabled, a timer interrupt
 * will be signalled by calling the installed interrupt handler.
 */
void tasking_timercheck(void)
{
  if (!tasking_on) return;
  if (!timer_enable) return;
  if (--timer) return;
  timer_enable = 0;
  STk_user_eval( LIST1( interrupt_handler ), UNBOUND );
  /* returned value is ignored. */
}


/* Called to determine whether the tasking system is running, i.e., 
 * whether tasking_input_block can safely be called.
 */
int tasking_enabled()
{
  return tasking_on;
}


/* Called from the STk_getc() primitive (in Src/io.c) when input is not
 * available on the file descriptor. An input 'block' event will be signalled
 * by calling the installed input event handler.
 *
 * The interrupt handler must return #f if the interrupt is not handled
 * and the I/O system is to block in the OS (desirable for e.g. engines).
 */
int tasking_input_block( int filedes )
{
  SCM r;

  timer_enable = 0;
  setup_io_proc( filedes );
  r = STk_user_eval( LIST3( ioblock_handler,
			    blocksym,
			    STk_makeinteger(filedes)), 
		    UNBOUND );
  return !FALSEP( r );
}


/***************************************************************************
 * Internal
 */

#ifdef USE_TK
/* This is called from Tk in response to file events */
static void io_proc( ClientData cd, int mask )
{
  signal_io_available( (int)cd );
}
#endif


static void signal_io_available( int filedes )
{
  int old = timer_enable;
  SCM r;

  /* This procedure is sometimes called after tasking is disabled because
   * tasking enabled an IO procedure but did not disable it.
   * In those cases we ignore the procedure.
   */
  if (!tasking_on) {
    suspend_io_proc( filedes );
    return;
  }
  timer_enable = 0;
  suspend_io_proc( filedes );
  STk_eval(LIST3(ioblock_handler,
		 unblocksym,
		 STk_makeinteger(filedes)),
	   UNBOUND );
  timer_enable = old;
}


static void setup_io_proc( int fd )
{
#ifdef USE_TK
  if (Tk_initialized) 
    Tk_CreateFileHandler( fd, TK_READABLE, io_proc, (ClientData)fd );
  else 
#endif
    {
      if (fd > largest) largest = fd;
      FD_SET( fd, &readfds );
    }
}


static void suspend_io_proc( int fd )
{
#ifdef USE_TK
  if (Tk_initialized) 
    Tk_CreateFileHandler( fd, 0, 0, 0 );
  else
#endif
    FD_CLR( fd, &readfds );
}


#ifdef USE_TK
/* Poll for system events when Tk is installed and initialized.
 * Notice that interrupts are turned on and a sizeable time slice
 * is given before a Tk event is handled. This is so that the event
 * handler can run and be preempted, if it runs long enough: a task
 * switch can happen in a Tk event handler (for example, it runs for
 * a long time or blocks on a mutex). If Tk has internal state which
 * it cleans up after the event handler returns (i.e., it is not
 * thread safe) then this will eventually be a problem, and event
 * handlers should then run with interrupts turned off.
 */
static void poll_for_tk( int blockp )
{
  int old_timer_enable = timer_enable;
  int old_timer = timer;

  timer_enable = 1;
  if (blockp) {
    return_from_polling = 0;
    while (!return_from_polling) {
      timer = 10000;                /* not entirely arbitrary */
      Tk_DoOneEvent( 0 );
    }
  }
  else {
    timer = 10000;
    Tk_DoOneEvent( TK_DONT_WAIT );
  }
  timer_enable = old_timer_enable;
  timer = old_timer;
}
#endif


/* Poll for system events when Tk is not installed or not initialized. */
static void poll_the_hard_way( int blockp )
{
  int i;
  struct timeval timeout;
  fd_set newset;

 Again:
  newset = readfds;
  timeout.tv_sec = timeout.tv_usec = 0;
  select( largest+1, &newset, (fd_set*)0, (fd_set*)0,
	 (blockp ? (struct timeval*) 0 : &timeout) );

  /* This is quite inefficient */
  return_from_polling = 0;
  for ( i = 0 ; i < largest ; i++ ) {
    if (FD_ISSET( i, &newset )) 
      signal_io_available( i );
  }
  if (!return_from_polling) goto Again;
}



/***************************************************************************
 *
 * Primitives.
 */

/* Install a handler for timer interrupts. */
static PRIMITIVE setup_tasking( SCM timerhandler, SCM iohandler )
{
  if (!STk_is_thunk( timerhandler ) && NEQ( timerhandler, NIL ) )
    Err( "setup_tasking: not a thunk: ", timerhandler );

  if ( FALSEP( STk_procedurep( iohandler )) && NEQ( iohandler, NIL ) )
    Err( "set-input-event-handler: not a procedure: ", timerhandler );

  interrupt_handler = timerhandler;
  ioblock_handler = iohandler;
  tasking_on = interrupt_handler != NIL && ioblock_handler != NIL;
  return UNDEFINED;
}


/* Poll system events, blocking or nonblocking. */
static PRIMITIVE poll_system_events( SCM blockp )
{
  int blocking = TRUEP( blockp );

#ifdef USE_TK
  if (Tk_initialized)
    poll_for_tk( blocking );
  else
#endif
    poll_the_hard_way( blocking );
  return UNDEFINED;
}


/* Disable interrupts; return either the current timer, or #f if interrupts
 * were already disabled.
 */
static PRIMITIVE disable_interrupts()
{
  SCM r = (timer_enable ? STk_makeinteger( timer ) : Ntruth);
  timer_enable = 0;
  return r;
}


/* Enable interrupts, setting the timer to the given count. */
static PRIMITIVE enable_interrupts( SCM count )
{
  if (STk_integerp( count )) {
    timer_enable = 1;
    timer = STk_integer_value( count );
  }
  else
    Err( "enable-interrupts: illegal value: ", count );
  return UNDEFINED;
}

static PRIMITIVE set_return_from_polling_flag()
{
  return_from_polling = 1;
  return UNDEFINED;
}

void STk_init_tasking()
{
  SCM quotesym = Intern( "quote" );

  blocksym = Intern( "block" );        STk_gc_protect( &blocksym );
  blocksym = LIST2( quotesym, blocksym );

  unblocksym = Intern( "unblock" );    STk_gc_protect( &unblocksym );
  unblocksym = LIST2( quotesym, unblocksym );

  interrupt_handler = NIL;             STk_gc_protect( &interrupt_handler );
  ioblock_handler = NIL;               STk_gc_protect( &ioblock_handler );

  FD_ZERO( &readfds );
  STk_add_new_primitive( "tasking/setup-tasking", tc_subr_2, 
			setup_tasking );
  STk_add_new_primitive( "tasking/poll-system-events", tc_subr_1, 
			poll_system_events );
  STk_add_new_primitive( "tasking/disable-interrupts", tc_subr_0,
			disable_interrupts );
  STk_add_new_primitive( "tasking/enable-interrupts", tc_subr_1,
			enable_interrupts );
  STk_add_new_primitive( "tasking/set-return-from-polling-flag", tc_subr_0,
			set_return_from_polling_flag );
}

/* eof */
