/* 
 * t a s k i o . c                  -- I/O system replacement for tasking.
 *
 * Copyright (C) 1993,1994,1995 Erick Gallesio - I3S-CNRS/ESSI <eg@unice.fr>
 *
 * Permission to use, copy, and/or distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that both the above copyright notice and this permission notice appear in
 * all copies and derived works.  Fees for distribution or use of this
 * software or derived works may only be charged with express written
 * permission of the copyright holder.  
 * This software is provided ``as is'' without express or implied warranty.
 *
 * This software is a derivative work of other copyrighted softwares; the
 * copyright notices of these softwares are placed in the file COPYRIGHTS
 *
 *  Original Author: Erick Gallesio [eg@kaolin.unice.fr]
 * Modifications by: Lars Thomas Hansen [lth@cs.uoregon.edu]
 *    Creation date: 27-Jun-1995 
 * Last file update: 
 *
 * DESCRIPTION
 *   This file is based on the standard Src/io.c distributed with STk 2.1.7,
 *   and has been modified to work with the tasking system. The purpose
 *   of the modifications is to provide an I/O subsystem which will block
 *   only the calling task rather than the entire process.
 *
 *   If Tk is present, Tk events are handled along with file descriptor
 *   events, all using Tks event handling mechanisms.
 *
 * REQUIREMENTS
 *   This file requires select(). It could be rewritten to handle interrupt-
 *   driven I/O or non-blocking busy-waiting I/O, if necessary.
 *
 * USAGE
 *   The idiom for blocking waiting for input is:
 *
 *     while (!input_ready_on_fd( filedes ))
 *       tasking_input_block( filedes );
 *
 *   This idiom should be useful in many cases; consider for example its
 *   use in the code for accept() in the socket extension: the loop would
 *   be inserted before the call to accept(), and only when the loop
 *   terminates will the program go into the system call, and then return
 *   without any major delay.
 *
 * BUGS
 * - So far only input is handled: output is assumed not to block for
 *   significant periods of time. This is clearly false, consider output to
 *   a pipe or socket.
 * - The disappearance of all windows (normally a condition for shutdown)
 *   is not detected if tasking is enabled. This may be desirable, but for
 *   now I consider it a bug.
 *
 * PROBLEMS
 *   The definition if the hasbuffered() predicate needs to snoop in the
 *   structure pointed to by a FILE* to see if the buffer is non-empty.
 *   To do this, it accesses the _cnt member of a FILE structure.
 *   The use of _cnt is almost certainly not ANSI/ISO, but it may be POSIX
 *   (I really don't know). Either way, _cnt seems to be implemented almost
 *   everywhere. Those who do not have _cnt may have __cnt. I hate doing 
 *   this, but we _must_ be able to determine whether the buffer is non-empty.
 *   A better (portable) solution would be to change the I/O system to do 
 *   its own buffering, bypassing the FILE.
 */

#include <sys/time.h>
#include <sys/types.h>
#include "stk.h"
#include "sport.h"

#define BUFFER_SIZE     1024
#define SYSTEM(instr)   {long flag=No_interrupt(1); instr; No_interrupt(flag);}
#define STRING_PORTP(f) (((struct str_iob *) f)->signature == SPORT_SIGNATURE)

#include <stdio.h>

/* f is a FILE* */
#define hasbuffered( f )  ((f)->_cnt > 0)

static int STk_old_getc( FILE * );
static int filled = 0;

#ifdef USE_TK
extern int tk_NumMainWindows;
#endif

static void badport(int read)
{
  Err(read ? "String port is not open for reading"
           : "String port is not open for writing",
      NIL);
}

/* Called to receive a hint as to whether input is available on the given 
 * file descriptor. Note that the answer is 'yes' only if input is available, 
 * but may also be 'no' if input is available. Consider the use of this 
 * procedure an optimization.
 *
 * Snarfed largely from char-ready? in Src/port.c.
 */
int input_ready_on_fd( int filedes )
{
  int result;
  { fd_set readfds;
    struct timeval timeout;

    FD_ZERO(&readfds); 
    FD_SET(filedes, &readfds);
    timeout.tv_sec = timeout.tv_usec = 0;
    result = select(filedes+1, &readfds, NULL, NULL, &timeout) > 0;
  }
/*DEBUG  printf( "input_ready_on_fd(%d) returns %d\n", filedes, result ); */
  return result;
}

/* Called to retrieve a character from the input stream. */
int STk_getc( FILE *f )
{
#ifdef USE_TK
  static int tasking_enabled_last_time = 0;
#endif

  /* Case 1: string input port */
  if (STRING_PORTP(f)) {
    register struct str_iob *g = (struct str_iob *)f;
    if (!(g->flag & READING)) badport(TRUE);
    return (--(g->cnt)>=0? ((int)*g->ptr++): EOF);
  }

  /* Case 2: tasking not enabled */
  if (!tasking_enabled()) {
#ifdef USE_TK
    if (tasking_enabled_last_time)
      Tk_CreateFileHandler(fileno(f),TK_READABLE,STk_StdinProc,(ClientData)0);
    tasking_enabled_last_time = 0;
#endif
    return STk_old_getc( f );
  }
#ifdef USE_TK
  tasking_enabled_last_time = 1;
#endif

 Again:
  /* Case 3: Input is available */
  if (hasbuffered(f) || input_ready_on_fd( fileno(f) )) {
    int result;
    SYSTEM( result = getc( f ) );
    return result;
  }

  /* Case 4: Must block. */
  if (tasking_input_block( fileno( f ) ))
    goto Again;
  else {
    int result;
    SYSTEM( result = getc( f ) );
    return result;
  }
}

void STk_StdinProc(void)
{
  filled = 1;
}

static int STk_old_getc( FILE *f )
{
#ifdef USE_TK
  if (Tk_initialized) {
    if (f == stdin && !hasbuffered(f)) {
      /* It's important to clear 'filled' here. The event loop will
       * always provide current&correct information, whereas 'filled'
       * may be stale.
       */
      filled = 0;
      while (!filled) {
	Tk_DoOneEvent(0);
	if (tk_NumMainWindows <= 0) return EOF;
      }
    }
  }
#endif

  { int result;
    SYSTEM(result=getc(f));
    return result;
  }
}

int STk_ungetc(int c, FILE *f)
{
  if (STRING_PORTP(f)) {
    register struct str_iob *g = (struct str_iob *)f;
    if (g->ptr == g->base) Err("INTERNAL ERROR: cannont unget char", NIL);
    if (!(g->flag & READING)) badport(TRUE);
    g->ptr--;
    g->cnt++;
  }
  else
    SYSTEM(ungetc(c, f));
  return c;
}

int STk_putc(int c, FILE *f)
{
  if (STRING_PORTP(f)) {
    register struct str_iob *g = (struct str_iob *)f;
    register int tmp;

    if (!(g->flag & WRITING)) badport(FALSE);
    if (++g->cnt == g->bufsiz) {
      tmp 	 = g->bufsiz;
      tmp 	+= tmp/2;
      g->base	 = must_realloc(g->base, tmp);
      g->ptr 	 = g->base + g->bufsiz - 1; /* base can have been moved */
      g->bufsiz	 = tmp;
    }
    *g->ptr++ = (char) c;
  }
  else 
    SYSTEM(fputc(c, f));
  return c;
}

int STk_puts(char *s, FILE *f)
{
  if (STRING_PORTP(f)) {
    while (*s) STk_putc(*s++, f);
    return 0;
  }
  else {
    int result;
    SYSTEM(result = fputs(s, f));
    return result;
  }
}

int STk_eof(FILE *f)
{
  if (STRING_PORTP(f))
    return (((struct str_iob *)f)->cnt <= 0);

 Again:
  if (hasbuffered(f) || !tasking_enabled() || input_ready_on_fd( fileno(f) )) {
    int result;
    SYSTEM(result=feof(f));
    return result;
  }

  if (tasking_input_block( fileno( f ) ))
    goto Again;
  else {
    int result;
    SYSTEM(result=feof(f));
    return result;
  }
}

/* eof */
