; t h r e a d s . s c h 
;
;   Copyright (C) 1995 The University of Oregon
;   Copyright (C) 1995 Lars Thomas Hansen <lth@cs.uoregon.edu>
;
;   Permission to use, copy, and/or distribute this software and its
;   documentation for any purpose and with or without fee is hereby granted,
;   provided that both the above copyright notice and this permission notice
;   appear in all copies and derived works.  
;
;   This software is provided ``as is'' without express or implied warranty.
;
; lth@cs.uoregon.edu / June 21, 1995
; $Id$
;
; Simple uniprocessor threads package built on top of tasking primitives.
;
; This package is meant as an example of how a tasking system can be 
; implemented on top of the tasking primitives. More elaborate systems
; can also be build, with thread identifiers, process groups, fancy
; scheduling (this one uses round-robin with one priority level) and so on.
;
; See the file tasking.{tex,ps} for a complete explanation.
;
; TASKING PROCEDURES
; - 'Begin-tasking' is called from your program. It initializes the tasking
;   system and returns to the caller. You must call 'begin-tasking' before
;   you call any other tasking primitives. You can call 'begin-tasking' a
;   second time after ending tasking first with 'end-tasking' (below).
; - 'Spawn' takes a thunk and runs the thunk in a separate thread of control.
;   Spawn returns no value, and the return value of the thunk is thrown
;   away.
; - 'Surrender-timeslice' suspends the calling thread but keeps it on the
;   run queue.
; - 'Halt-task' terminates the calling task.
; - 'Block-task' takes a procedure of one parameter and passes it a
;   data structure representing a blocked task.
; - 'Ready-task' places a task on the run queue but does not run it.
; - 'End-tasking' shuts down the tasking system. The thread which calls
;   'end-tasking' will return from the call and will remain as the only
;   running thread in the system.
;
; BUGS and ANNOYANCES
; - Encapsulation is not so good; would be better to stuff things in 
;   a global (let () ...) to hide some of the global variables and
;   auxiliary procedures. OK for testing purposes, though, and it allows
;   for some inspection of the internals.

(require "tasking.sch")
(require "queue.sch")

; This is better but must be entered on the top level.
;(define reset #f)
;(call-with-current-continuation
;  (lambda (k)
;    (set! reset (lambda ()
;		  (display "[Scheme RESET]")
;		  (newline)
;		  (k #f)))))
(define (reset) (error "RESET"))

; When tasking is enabled, report-error does not bring up a window.

(define report-error
  (let ((r report-error))
    (lambda (head msg obj)
      (if *tasking-enabled*
	  (begin (set! %stack (%get-eval-stack))
		 (format #t "ERROR: ~a: ~s~%" msg obj)
		 (*end-tasking* 'error))
	  (r head msg obj)))))

; 'Begin-tasking' takes an optional procedure of one argument which
; will be called when 'end-tasking' completes, with the value which
; (optionally) is passed to 'end-tasking'.

(define (begin-tasking . rest)
  (if (not *tasking-enabled*)
      (*begin-tasking* (if (null? rest) #f (car rest)))))


; End-tasking takes an optional value which will be passed to the exit 
; procedure.

(define (end-tasking . rest)
  (if *tasking-enabled*
      (*end-tasking* (if (null? rest) #f (car rest)))))


(define (spawn thunk) 
  (if (not *tasking-enabled*)
      (error "spawn: tasking is not enabled."))
  (call-without-interrupts
   (lambda ()
     (*schedule-task* 
      (tasking/create-task 
       (lambda ()
	 (thunk)
	 (halt-task)))))))


; Some people prefer to call this 'Yield'.
(define (surrender-timeslice)
  (debugmsg "Task surrenders timeslice.")
  (if (not *tasking-enabled*)
      (error "surrender-timeslice: tasking is not enabled."))
  (if (not (tasking/disable-interrupts))
      (error "surrender-timeslice: in critical section."))
  (*surrender-timeslice*))


(define (halt-task)
  (debugmsg "Task terminates.")
  (if (not *tasking-enabled*)
      (error "halt-task: tasking is not enabled."))
  (if (not (tasking/disable-interrupts))
      (error "halt-task: in critical section."))
  (tasking/capture-task (lambda (t)
			  (tasking/destroy-task t)
			  (*run-scheduler*))))


(define (block-task proc)
  (tasking/disable-interrupts)
  (tasking/capture-task
   (lambda (t)
     (proc t)
     (*run-scheduler*))))


(define (ready-task t)
  (call-without-interrupts
   (lambda ()
     (*schedule-task* t))))


(define (call-without-interrupts thunk)
  (let ((istate (tasking/disable-interrupts)))
    (let ((r (thunk)))
      (if istate (tasking/enable-interrupts istate))
      r)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Internal procedures. Should always be called with interrupts off.
;
; The timeslice is the number of procedure calls between each timer
; interrupt. 10000 seems to work OK, but for different applications
; and machines the best value almost certainly varies. It can be changed
; at any time and will take effect at the next task switch.

(define *standard-timeslice* 10000)    ; procedure calls between interrupts
(define *standard-poll-interval* 10)   ; tasks run between event polls
(define *tasking-enabled* #f)          ; #t when tasking is enabled
(define *run-queue* #f)                ; queue of runnable tasks
(define *pending* '#())                ; vector[fd] of lists of blocked tasks
(define *exit-proc* #f)                ; continuation for end-tasking
(define *debug* #f)                    ; set to #t to enable some debug msgs.

(define (*begin-tasking* proc)
  (set! *run-queue* (make-basic-queue))
  (tasking/setup-tasking *surrender-timeslice* *input-event-handler*)
  (tasking/enable-interrupts *standard-timeslice*)
  (if proc
      (set! *exit-proc* proc)
      (set! *exit-proc* (lambda (x) x)))
  (set! *tasking-enabled* #t)
  #t)

(define (*end-tasking* val)
  (tasking/disable-interrupts)
  (tasking/setup-tasking '() '())
  (set! *tasking-enabled* #f)
  (display "TASKING HALTED.")
  (newline)
  (*exit-proc* val))

(define *surrender-timeslice*
  (let ((p (lambda (t)
	     (*schedule-task* t)
	     (*run-scheduler*))))
    (lambda ()
      (debugmsg "Event: timer.")
      (tasking/capture-task p))))

(define (*schedule-task* thunk)
  (if (qempty? *run-queue*)
      (tasking/set-return-from-polling-flag))
  (qput *run-queue* thunk))

; It's in principle possible to detect deadlock here. What we need to 
; know is that:
; - there are no runnable threads (trivial).
; - no thread is waiting for an external event (hard).
; As far as this module is concerned, external events are associated with
; input streams only. However, there are things like Tk button presses,
; timers, and all that, for which we do not 'wait' for per se. It seems
; to me that it not possible to easily know whether such events may
; arrive to unblock a task. Therefore, deadlock may not be detectable in
; general. Of course, if there were an fd on which Tk events appear, then
; the wait would be explicit, but would this help?

(define *run-scheduler*
  (let ((time-to-poll *standard-poll-interval*))
    (lambda ()
      (cond ((not *tasking-enabled*)
	     (display "Scheduler: internal inconsistency.")
	     (newline)
	     (reset))
	    ((qempty? *run-queue*)
	     (begin (tasking/poll-system-events #t)
		    (set! time-to-poll *standard-poll-interval*)
		    (*run-scheduler*)))
	    (else
	     (let ((p (qget *run-queue*)))
	       (set! time-to-poll (- time-to-poll 1))
	       (if (zero? time-to-poll)
		   (begin (tasking/poll-system-events #f)
			  (set! time-to-poll *standard-poll-interval*)))
	       (tasking/enable-interrupts *standard-timeslice*)
	       (tasking/resume-task p)))))))

(define *input-event-handler* 
  (let ()

    (define (extend fd)
      (let* ((l (vector-length *pending*))
	     (n (max (+ l 10) (+ fd 10)))
	     (v (make-vector n '())))
	(do ((i (- l 1) (- i 1)))
	    ((< i 0) (set! *pending* v))
	  (vector-set! v i (vector-ref *pending* i)))))

    (define (*register-pending-input* fd)
      (tasking/capture-task
       (lambda (t)
	 (if (>= fd (vector-length *pending*))
	     (extend fd))
	 (vector-set! *pending* 
		      fd 
		      (cons t (vector-ref *pending* fd)))
	 (*run-scheduler*))))

    (define (*wakeup-pending-input* fd)
      (if (< fd (vector-length *pending*))
	  (begin (let ((tasks (vector-ref *pending* fd)))
		   (for-each *schedule-task* tasks))
		 (vector-set! *pending* fd '()))))

    (lambda (event fd)
      (debugmsg "Event: IO: ~a ~a" event fd)
      (cond ((eq? event 'block)
	     (*register-pending-input* fd)
	     ; Return #t to signal that block has been handled.
	     #t)
	    ((eq? event 'unblock)
	     (*wakeup-pending-input* fd))
	    (else
	     (error "*input-event-handler*: unknown event: ~a" event))))))

(define (debugmsg fmt . rest)
  (if *debug*
      (begin (apply format #t (string-append "[debug] " fmt) rest)
	     (newline))))

(provide "threads.sch")

; eof
